/* Generated by Cython 3.0.11 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [],
        "name": "flitter.render.physics",
        "sources": [
            "src/flitter/render/physics.pyx"
        ]
    },
    "module_name": "flitter.render.physics"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#if defined(CYTHON_LIMITED_API) && 0
  #ifndef Py_LIMITED_API
    #if CYTHON_LIMITED_API+0 > 0x03030000
      #define Py_LIMITED_API CYTHON_LIMITED_API
    #else
      #define Py_LIMITED_API 0x03030000
    #endif
  #endif
#endif

#include "Python.h"

    #if PY_MAJOR_VERSION >= 3
      #define __Pyx_PyFloat_FromString(obj)  PyFloat_FromString(obj)
    #else
      #define __Pyx_PyFloat_FromString(obj)  PyFloat_FromString(obj, NULL)
    #endif
    

    #if PY_MAJOR_VERSION <= 2
    #define PyDict_GetItemWithError _PyDict_GetItemWithError
    #endif
    

    #if (PY_VERSION_HEX < 0x030700b1 || (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM < 0x07030600)) && !defined(PyContextVar_Get)
    #define PyContextVar_Get(var, d, v)         ((d) ?             ((void)(var), Py_INCREF(d), (v)[0] = (d), 0) :             ((v)[0] = NULL, 0)         )
    #endif
    
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02070000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.7+ or Python 3.3+.
#else
#if defined(CYTHON_LIMITED_API) && CYTHON_LIMITED_API
#define __PYX_EXTRA_ABI_MODULE_NAME "limited"
#else
#define __PYX_EXTRA_ABI_MODULE_NAME ""
#endif
#define CYTHON_ABI "3_0_11" __PYX_EXTRA_ABI_MODULE_NAME
#define __PYX_ABI_MODULE_NAME "_cython_" CYTHON_ABI
#define __PYX_TYPE_MODULE_PREFIX __PYX_ABI_MODULE_NAME "."
#define CYTHON_HEX_VERSION 0x03000BF0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(_WIN32) && !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #define HAVE_LONG_LONG
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#define __PYX_LIMITED_VERSION_HEX PY_VERSION_HEX
#if defined(GRAALVM_PYTHON)
  /* For very preliminary testing purposes. Most variables are set the same as PyPy.
     The existence of this section does not imply that anything works or is even tested */
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS (PY_MAJOR_VERSION >= 3)
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(PYPY_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS (PY_MAJOR_VERSION >= 3)
  #endif
  #if PY_VERSION_HEX < 0x03090000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1 && PYPY_VERSION_NUM >= 0x07030C00)
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(CYTHON_LIMITED_API)
  #ifdef Py_LIMITED_API
    #undef __PYX_LIMITED_VERSION_HEX
    #define __PYX_LIMITED_VERSION_HEX Py_LIMITED_API
  #endif
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 1
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_CLINE_IN_TRACEBACK
  #define CYTHON_CLINE_IN_TRACEBACK 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 1
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #endif
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 1
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(Py_GIL_DISABLED) || defined(Py_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL 1
  #endif
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS 0
  #endif
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #ifndef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_GIL
    #define CYTHON_FAST_GIL (PY_MAJOR_VERSION < 3 || PY_VERSION_HEX >= 0x03060000 && PY_VERSION_HEX < 0x030C00A6)
  #endif
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL (PY_VERSION_HEX >= 0x030700A1)
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #if PY_VERSION_HEX < 0x030400a1
    #undef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #elif !defined(CYTHON_USE_TP_FINALIZE)
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #if PY_VERSION_HEX < 0x030600B1
    #undef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS 0
  #elif !defined(CYTHON_USE_DICT_VERSIONS)
    #define CYTHON_USE_DICT_VERSIONS  (PY_VERSION_HEX < 0x030C00A5)
  #endif
  #if PY_VERSION_HEX < 0x030700A3
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK 1
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if !defined(CYTHON_VECTORCALL)
#define CYTHON_VECTORCALL  (CYTHON_FAST_PYCCALL && PY_VERSION_HEX >= 0x030800B1)
#endif
#define CYTHON_BACKPORT_VECTORCALL (CYTHON_METH_FASTCALL && PY_VERSION_HEX < 0x030800B1)
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(maybe_unused) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(maybe_unused)
        #define CYTHON_UNUSED [[maybe_unused]]
      #endif
    #endif
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
  #define CYTHON_MAYBE_UNUSED_VAR(x) CYTHON_UNUSED_VAR(x)
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_USE_CPP_STD_MOVE
  #if defined(__cplusplus) && (\
    __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600))
    #define CYTHON_USE_CPP_STD_MOVE 1
  #else
    #define CYTHON_USE_CPP_STD_MOVE 0
  #endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
            typedef unsigned char     uint8_t;
            typedef unsigned short    uint16_t;
            typedef unsigned int      uint32_t;
        #else
            typedef unsigned __int8   uint8_t;
            typedef unsigned __int16  uint16_t;
            typedef unsigned __int32  uint32_t;
        #endif
    #endif
    #if _MSC_VER < 1300
        #ifdef _WIN64
            typedef unsigned long long  __pyx_uintptr_t;
        #else
            typedef unsigned int        __pyx_uintptr_t;
        #endif
    #else
        #ifdef _WIN64
            typedef unsigned __int64    __pyx_uintptr_t;
        #else
            typedef unsigned __int32    __pyx_uintptr_t;
        #endif
    #endif
#else
    #include <stdint.h>
    typedef uintptr_t  __pyx_uintptr_t;
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(fallthrough) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(fallthrough)
        #define CYTHON_FALLTHROUGH [[fallthrough]]
      #endif
    #endif
    #ifndef CYTHON_FALLTHROUGH
      #if __has_cpp_attribute(clang::fallthrough)
        #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
      #elif __has_cpp_attribute(gnu::fallthrough)
        #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
      #endif
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif
#ifdef __cplusplus
  template <typename T>
  struct __PYX_IS_UNSIGNED_IMPL {static const bool value = T(0) < T(-1);};
  #define __PYX_IS_UNSIGNED(type) (__PYX_IS_UNSIGNED_IMPL<type>::value)
#else
  #define __PYX_IS_UNSIGNED(type) (((type)-1) > 0)
#endif
#if CYTHON_COMPILING_IN_PYPY == 1
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x030A0000)
#else
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000)
#endif
#define __PYX_REINTERPRET_FUNCION(func_pointer, other_pointer) ((func_pointer)(void(*)(void))(other_pointer))

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_DefaultClassType PyClass_Type
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if CYTHON_COMPILING_IN_LIMITED_API
    static CYTHON_INLINE PyObject* __Pyx_PyCode_New(int a, int p, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *exception_table = NULL;
        PyObject *types_module=NULL, *code_type=NULL, *result=NULL;
        #if __PYX_LIMITED_VERSION_HEX < 0x030B0000
        PyObject *version_info;
        PyObject *py_minor_version = NULL;
        #endif
        long minor_version = 0;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        #if __PYX_LIMITED_VERSION_HEX >= 0x030B0000
        minor_version = 11;
        #else
        if (!(version_info = PySys_GetObject("version_info"))) goto end;
        if (!(py_minor_version = PySequence_GetItem(version_info, 1))) goto end;
        minor_version = PyLong_AsLong(py_minor_version);
        Py_DECREF(py_minor_version);
        if (minor_version == -1 && PyErr_Occurred()) goto end;
        #endif
        if (!(types_module = PyImport_ImportModule("types"))) goto end;
        if (!(code_type = PyObject_GetAttrString(types_module, "CodeType"))) goto end;
        if (minor_version <= 7) {
            (void)p;
            result = PyObject_CallFunction(code_type, "iiiiiOOOOOOiOO", a, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else if (minor_version <= 10) {
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOiOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else {
            if (!(exception_table = PyBytes_FromStringAndSize(NULL, 0))) goto end;
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOOiOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, name, fline, lnos, exception_table, fv, cell);
        }
    end:
        Py_XDECREF(code_type);
        Py_XDECREF(exception_table);
        Py_XDECREF(types_module);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return result;
    }
    #ifndef CO_OPTIMIZED
    #define CO_OPTIMIZED 0x0001
    #endif
    #ifndef CO_NEWLOCALS
    #define CO_NEWLOCALS 0x0002
    #endif
    #ifndef CO_VARARGS
    #define CO_VARARGS 0x0004
    #endif
    #ifndef CO_VARKEYWORDS
    #define CO_VARKEYWORDS 0x0008
    #endif
    #ifndef CO_ASYNC_GENERATOR
    #define CO_ASYNC_GENERATOR 0x0200
    #endif
    #ifndef CO_GENERATOR
    #define CO_GENERATOR 0x0020
    #endif
    #ifndef CO_COROUTINE
    #define CO_COROUTINE 0x0080
    #endif
#elif PY_VERSION_HEX >= 0x030B0000
  static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int p, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
    PyCodeObject *result;
    PyObject *empty_bytes = PyBytes_FromStringAndSize("", 0);
    if (!empty_bytes) return NULL;
    result =
      #if PY_VERSION_HEX >= 0x030C0000
        PyUnstable_Code_NewWithPosOnlyArgs
      #else
        PyCode_NewWithPosOnlyArgs
      #endif
        (a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, name, fline, lnos, empty_bytes);
    Py_DECREF(empty_bytes);
    return result;
  }
#elif PY_VERSION_HEX >= 0x030800B2 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_NewWithPosOnlyArgs(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
#endif
#if PY_VERSION_HEX >= 0x030900A4 || defined(Py_IS_TYPE)
  #define __Pyx_IS_TYPE(ob, type) Py_IS_TYPE(ob, type)
#else
  #define __Pyx_IS_TYPE(ob, type) (((const PyObject*)ob)->ob_type == (type))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_Is)
  #define __Pyx_Py_Is(x, y)  Py_Is(x, y)
#else
  #define __Pyx_Py_Is(x, y) ((x) == (y))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsNone)
  #define __Pyx_Py_IsNone(ob) Py_IsNone(ob)
#else
  #define __Pyx_Py_IsNone(ob) __Pyx_Py_Is((ob), Py_None)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsTrue)
  #define __Pyx_Py_IsTrue(ob) Py_IsTrue(ob)
#else
  #define __Pyx_Py_IsTrue(ob) __Pyx_Py_Is((ob), Py_True)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsFalse)
  #define __Pyx_Py_IsFalse(ob) Py_IsFalse(ob)
#else
  #define __Pyx_Py_IsFalse(ob) __Pyx_Py_Is((ob), Py_False)
#endif
#define __Pyx_NoneAsNull(obj)  (__Pyx_Py_IsNone(obj) ? NULL : (obj))
#if PY_VERSION_HEX >= 0x030900F0 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyObject_GC_IsFinalized(o) PyObject_GC_IsFinalized(o)
#else
  #define __Pyx_PyObject_GC_IsFinalized(o) _PyGC_FINALIZED(o)
#endif
#ifndef CO_COROUTINE
  #define CO_COROUTINE 0x80
#endif
#ifndef CO_ASYNC_GENERATOR
  #define CO_ASYNC_GENERATOR 0x200
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef Py_TPFLAGS_SEQUENCE
  #define Py_TPFLAGS_SEQUENCE 0
#endif
#ifndef Py_TPFLAGS_MAPPING
  #define Py_TPFLAGS_MAPPING 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #if PY_VERSION_HEX >= 0x030d00A4
  #  define __Pyx_PyCFunctionFast PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords PyCFunctionFastWithKeywords
  #else
  #  define __Pyx_PyCFunctionFast _PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
  #endif
#endif
#if CYTHON_METH_FASTCALL
  #define __Pyx_METH_FASTCALL METH_FASTCALL
  #define __Pyx_PyCFunction_FastCall __Pyx_PyCFunctionFast
  #define __Pyx_PyCFunction_FastCallWithKeywords __Pyx_PyCFunctionFastWithKeywords
#else
  #define __Pyx_METH_FASTCALL METH_VARARGS
  #define __Pyx_PyCFunction_FastCall PyCFunction
  #define __Pyx_PyCFunction_FastCallWithKeywords PyCFunctionWithKeywords
#endif
#if CYTHON_VECTORCALL
  #define __pyx_vectorcallfunc vectorcallfunc
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  PY_VECTORCALL_ARGUMENTS_OFFSET
  #define __Pyx_PyVectorcall_NARGS(n)  PyVectorcall_NARGS((size_t)(n))
#elif CYTHON_BACKPORT_VECTORCALL
  typedef PyObject *(*__pyx_vectorcallfunc)(PyObject *callable, PyObject *const *args,
                                            size_t nargsf, PyObject *kwnames);
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  ((size_t)1 << (8 * sizeof(size_t) - 1))
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(((size_t)(n)) & ~__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET))
#else
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  0
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(n))
#endif
#if PY_MAJOR_VERSION >= 0x030900B1
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_CheckExact(func)
#else
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_Check(func)
#endif
#define __Pyx_CyOrPyCFunction_Check(func)  PyCFunction_Check(func)
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  (((PyCFunctionObject*)(func))->m_ml->ml_meth)
#elif !CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  PyCFunction_GET_FUNCTION(func)
#endif
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FLAGS(func)  (((PyCFunctionObject*)(func))->m_ml->ml_flags)
static CYTHON_INLINE PyObject* __Pyx_CyOrPyCFunction_GET_SELF(PyObject *func) {
    return (__Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_STATIC) ? NULL : ((PyCFunctionObject*)func)->m_self;
}
#endif
static CYTHON_INLINE int __Pyx__IsSameCFunction(PyObject *func, void *cfunc) {
#if CYTHON_COMPILING_IN_LIMITED_API
    return PyCFunction_Check(func) && PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
#else
    return PyCFunction_Check(func) && PyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
#endif
}
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCFunction(func, cfunc)
#if __PYX_LIMITED_VERSION_HEX < 0x030900B1
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  ((void)m, PyType_FromSpecWithBases(s, b))
  typedef PyObject *(*__Pyx_PyCMethod)(PyObject *, PyTypeObject *, PyObject *const *, size_t, PyObject *);
#else
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  PyType_FromModuleAndSpec(m, s, b)
  #define __Pyx_PyCMethod  PyCMethod
#endif
#ifndef METH_METHOD
  #define METH_METHOD 0x200
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyThreadState_Current PyThreadState_Get()
#elif !CYTHON_FAST_THREAD_STATE
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyThreadState_Current PyThreadState_GetUnchecked()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE void *__Pyx_PyModule_GetState(PyObject *op)
{
    void *result;
    result = PyModule_GetState(op);
    if (!result)
        Py_FatalError("Couldn't find the module state");
    return result;
}
#endif
#define __Pyx_PyObject_GetSlot(obj, name, func_ctype)  __Pyx_PyType_GetSlot(Py_TYPE(obj), name, func_ctype)
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((func_ctype) PyType_GetSlot((type), Py_##name))
#else
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((type)->name)
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if PY_MAJOR_VERSION < 3
    #if CYTHON_COMPILING_IN_PYPY
        #if PYPY_VERSION_NUM < 0x07030600
            #if defined(__cplusplus) && __cplusplus >= 201402L
                [[deprecated("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6")]]
            #elif defined(__GNUC__) || defined(__clang__)
                __attribute__ ((__deprecated__("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6")))
            #elif defined(_MSC_VER)
                __declspec(deprecated("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6"))
            #endif
            static CYTHON_INLINE int PyGILState_Check(void) {
                return 0;
            }
        #else  // PYPY_VERSION_NUM < 0x07030600
        #endif  // PYPY_VERSION_NUM < 0x07030600
    #else
        static CYTHON_INLINE int PyGILState_Check(void) {
            PyThreadState * tstate = _PyThreadState_Current;
            return tstate && (tstate == PyGILState_GetThisThreadState());
        }
    #endif
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000 || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX > 0x030600B4 && PY_VERSION_HEX < 0x030d0000 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStrWithError(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStr(PyObject *dict, PyObject *name) {
    PyObject *res = __Pyx_PyDict_GetItemStrWithError(dict, name);
    if (res == NULL) PyErr_Clear();
    return res;
}
#elif PY_MAJOR_VERSION >= 3 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000)
#define __Pyx_PyDict_GetItemStrWithError  PyDict_GetItemWithError
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#else
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStrWithError(PyObject *dict, PyObject *name) {
#if CYTHON_COMPILING_IN_PYPY
    return PyDict_GetItem(dict, name);
#else
    PyDictEntry *ep;
    PyDictObject *mp = (PyDictObject*) dict;
    long hash = ((PyStringObject *) name)->ob_shash;
    assert(hash != -1);
    ep = (mp->ma_lookup)(mp, name, hash);
    if (ep == NULL) {
        return NULL;
    }
    return ep->me_value;
#endif
}
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#endif
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetFlags(tp)   (((PyTypeObject *)tp)->tp_flags)
  #define __Pyx_PyType_HasFeature(type, feature)  ((__Pyx_PyType_GetFlags(type) & (feature)) != 0)
  #define __Pyx_PyObject_GetIterNextFunc(obj)  (Py_TYPE(obj)->tp_iternext)
#else
  #define __Pyx_PyType_GetFlags(tp)   (PyType_GetFlags((PyTypeObject *)tp))
  #define __Pyx_PyType_HasFeature(type, feature)  PyType_HasFeature(type, feature)
  #define __Pyx_PyObject_GetIterNextFunc(obj)  PyIter_Next
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_SetItemOnTypeDict(tp, k, v) PyObject_GenericSetAttr((PyObject*)tp, k, v)
#else
  #define __Pyx_SetItemOnTypeDict(tp, k, v) PyDict_SetItem(tp->tp_dict, k, v)
#endif
#if CYTHON_USE_TYPE_SPECS && PY_VERSION_HEX >= 0x03080000
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  {\
    PyTypeObject *type = Py_TYPE((PyObject*)obj);\
    assert(__Pyx_PyType_HasFeature(type, Py_TPFLAGS_HEAPTYPE));\
    PyObject_GC_Del(obj);\
    Py_DECREF(type);\
}
#else
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  PyObject_GC_Del(obj)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GetLength(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_ReadChar(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((void)u, 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((void)u, (0))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)u)
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)k, PyUnicode_ReadChar((PyObject*)(d), i))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GetLength(u))
#elif PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         ((int)PyUnicode_KIND(u))
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, (Py_UCS4) ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535U : 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((int)sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = (Py_UNICODE) ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #if !defined(PyUnicode_DecodeUnicodeEscape)
    #define PyUnicode_DecodeUnicodeEscape(s, size, errors)  PyUnicode_Decode(s, size, "unicode_escape", errors)
  #endif
  #if !defined(PyUnicode_Contains) || (PY_MAJOR_VERSION == 2 && PYPY_VERSION_NUM < 0x07030500)
    #undef PyUnicode_Contains
    #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
  #endif
  #if !defined(PyByteArray_Check)
    #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
  #endif
  #if !defined(PyObject_Format)
    #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
  #endif
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#if CYTHON_COMPILING_IN_CPYTHON
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && Py_REFCNT(obj) == 1) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#else
  #define __Pyx_PySequence_ListKeepNew(obj)  PySequence_List(obj)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        __Pyx_IS_TYPE(obj, &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_ITEM(o, i) PySequence_ITEM(o, i)
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) (PyTuple_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyList_SET_ITEM(o, i, v) (PyList_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_GET_SIZE(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_GET_SIZE(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_GET_SIZE(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_GET_SIZE(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_GET_SIZE(o)
#else
  #define __Pyx_PySequence_ITEM(o, i) PySequence_GetItem(o, i)
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) PyTuple_SetItem(o, i, v)
  #define __Pyx_PyList_SET_ITEM(o, i, v) PyList_SetItem(o, i, v)
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_Size(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_Size(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_Size(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_Size(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_Size(o)
#endif
#if __PYX_LIMITED_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#else
  static CYTHON_INLINE PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *module = PyImport_AddModule(name);
      Py_XINCREF(module);
      return module;
  }
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define __Pyx_Py3Int_Check(op)       PyLong_Check(op)
  #define __Pyx_Py3Int_CheckExact(op)  PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#else
  #define __Pyx_Py3Int_Check(op)       (PyLong_Check(op) || PyInt_Check(op))
  #define __Pyx_Py3Int_CheckExact(op)  (PyLong_CheckExact(op) || PyInt_CheckExact(op))
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__;  (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifdef CYTHON_EXTERN_C
    #undef __PYX_EXTERN_C
    #define __PYX_EXTERN_C CYTHON_EXTERN_C
#elif defined(__PYX_EXTERN_C)
    #ifdef _MSC_VER
    #pragma message ("Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.")
    #else
    #warning Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.
    #endif
#else
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__flitter__render__physics
#define __PYX_HAVE_API__flitter__render__physics
/* Early includes */
#include <stdint.h>
#include <stddef.h>
#include <math.h>
#include <string.h>
#include <stdio.h>
#include "pythread.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s);
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char*);
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#define __Pyx_PyUnicode_FromOrdinal(o)       PyUnicode_FromOrdinal((int)o)
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_VERSION_HEX >= 0x030C00A7
  #ifndef _PyLong_SIGN_MASK
    #define _PyLong_SIGN_MASK 3
  #endif
  #ifndef _PyLong_NON_SIZE_BITS
    #define _PyLong_NON_SIZE_BITS 3
  #endif
  #define __Pyx_PyLong_Sign(x)  (((PyLongObject*)x)->long_value.lv_tag & _PyLong_SIGN_MASK)
  #define __Pyx_PyLong_IsNeg(x)  ((__Pyx_PyLong_Sign(x) & 2) != 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (!__Pyx_PyLong_IsNeg(x))
  #define __Pyx_PyLong_IsZero(x)  (__Pyx_PyLong_Sign(x) & 1)
  #define __Pyx_PyLong_IsPos(x)  (__Pyx_PyLong_Sign(x) == 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  (__Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  ((Py_ssize_t) (((PyLongObject*)x)->long_value.lv_tag >> _PyLong_NON_SIZE_BITS))
  #define __Pyx_PyLong_SignedDigitCount(x)\
        ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * __Pyx_PyLong_DigitCount(x))
  #if defined(PyUnstable_Long_IsCompact) && defined(PyUnstable_Long_CompactValue)
    #define __Pyx_PyLong_IsCompact(x)     PyUnstable_Long_IsCompact((PyLongObject*) x)
    #define __Pyx_PyLong_CompactValue(x)  PyUnstable_Long_CompactValue((PyLongObject*) x)
  #else
    #define __Pyx_PyLong_IsCompact(x)     (((PyLongObject*)x)->long_value.lv_tag < (2 << _PyLong_NON_SIZE_BITS))
    #define __Pyx_PyLong_CompactValue(x)  ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * (Py_ssize_t) __Pyx_PyLong_Digits(x)[0])
  #endif
  typedef Py_ssize_t  __Pyx_compact_pylong;
  typedef size_t  __Pyx_compact_upylong;
  #else
  #define __Pyx_PyLong_IsNeg(x)  (Py_SIZE(x) < 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (Py_SIZE(x) >= 0)
  #define __Pyx_PyLong_IsZero(x)  (Py_SIZE(x) == 0)
  #define __Pyx_PyLong_IsPos(x)  (Py_SIZE(x) > 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  ((Py_SIZE(x) == 0) ? 0 : __Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  __Pyx_sst_abs(Py_SIZE(x))
  #define __Pyx_PyLong_SignedDigitCount(x)  Py_SIZE(x)
  #define __Pyx_PyLong_IsCompact(x)  (Py_SIZE(x) == 0 || Py_SIZE(x) == 1 || Py_SIZE(x) == -1)
  #define __Pyx_PyLong_CompactValue(x)\
        ((Py_SIZE(x) == 0) ? (sdigit) 0 : ((Py_SIZE(x) < 0) ? -(sdigit)__Pyx_PyLong_Digits(x)[0] : (sdigit)__Pyx_PyLong_Digits(x)[0]))
  typedef sdigit  __Pyx_compact_pylong;
  typedef digit  __Pyx_compact_upylong;
  #endif
  #if PY_VERSION_HEX >= 0x030C00A5
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->long_value.ob_digit)
  #else
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->ob_digit)
  #endif
#endif
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
#include <string.h>
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = (char) c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#include <string.h>
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_m = NULL;
#endif
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm = __FILE__;
static const char *__pyx_filename;

/* Header.proto */
#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif (defined(_Complex_I) && !defined(_MSC_VER)) || ((defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L) && !defined(__STDC_NO_COMPLEX__) && !defined(_MSC_VER))
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif

/* #### Code section: filename_table ### */

static const char *__pyx_f[] = {
  "<stringsource>",
  "src/flitter/render/physics.pyx",
  "contextvars.pxd",
  "src/flitter/model.pxd",
  "src/flitter/language/functions.pxd",
  "type.pxd",
  "bool.pxd",
  "complex.pxd",
};
/* #### Code section: utility_code_proto_before_types ### */
/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* NoFastGil.proto */
#define __Pyx_PyGILState_Ensure PyGILState_Ensure
#define __Pyx_PyGILState_Release PyGILState_Release
#define __Pyx_FastGIL_Remember()
#define __Pyx_FastGIL_Forget()
#define __Pyx_FastGilFuncInit()

/* #### Code section: numeric_typedefs ### */
/* #### Code section: complex_type_declarations ### */
/* Declarations.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif
static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);

/* #### Code section: type_declarations ### */

/*--- Type declarations ---*/
struct __pyx_obj_7flitter_5model_Vector;
struct __pyx_obj_7flitter_5model_Matrix33;
struct __pyx_obj_7flitter_5model_Matrix44;
struct __pyx_obj_7flitter_5model_Quaternion;
struct __pyx_obj_7flitter_5model_Node;
struct __pyx_obj_7flitter_5model_StateDict;
struct __pyx_obj_7flitter_5model_Context;
struct __pyx_obj_7flitter_8language_9functions_uniform;
struct __pyx_obj_7flitter_8language_9functions_beta;
struct __pyx_obj_7flitter_8language_9functions_normal;
struct __pyx_obj_7flitter_6render_7physics_PhysicsGroup;
struct __pyx_obj_7flitter_6render_7physics_Particle;
struct __pyx_obj_7flitter_6render_7physics_Barrier;
struct __pyx_obj_7flitter_6render_7physics_ParticleForceApplier;
struct __pyx_obj_7flitter_6render_7physics_MatrixForceApplier;
struct __pyx_obj_7flitter_6render_7physics_DistanceForceApplier;
struct __pyx_obj_7flitter_6render_7physics_DragForceApplier;
struct __pyx_obj_7flitter_6render_7physics_BuoyancyForceApplier;
struct __pyx_obj_7flitter_6render_7physics_ConstantForceApplier;
struct __pyx_obj_7flitter_6render_7physics_RandomForceApplier;
struct __pyx_obj_7flitter_6render_7physics_FieldForceApplier;
struct __pyx_obj_7flitter_6render_7physics_CollisionForceApplier;
struct __pyx_obj_7flitter_6render_7physics_GravityForceApplier;
struct __pyx_obj_7flitter_6render_7physics_ElectrostaticForceApplier;
struct __pyx_obj_7flitter_6render_7physics_AdhesionForceApplier;
struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem;
struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct__destroy;
struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct_1_purge;
struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct_2_update;
struct __pyx_opt_args_7flitter_5model_6Vector_match;
struct __pyx_opt_args_7flitter_5model_4Node_get;

/* "model.pxd":63
 *     cdef str as_string(self)
 *     cpdef uint64_t hash(self, bint floor_floats)
 *     cpdef object match(self, int64_t n=?, type t=?, default=?)             # <<<<<<<<<<<<<<
 *     cdef str repr(self)
 *     cdef Vector neg(self)
 */
struct __pyx_opt_args_7flitter_5model_6Vector_match {
  int __pyx_n;
  int64_t n;
  PyTypeObject *t;
  PyObject *__pyx_default;
};

/* "model.pxd":215
 *     cpdef void append_vector(self, Vector nodes)
 *     cdef bint _equals(self, Node other)
 *     cpdef object get(self, str name, int n=?, type t=?, object default=?)             # <<<<<<<<<<<<<<
 *     cdef Vector get_vec(self, str name, Vector default)
 *     cdef Vector get_fvec(self, str name, int n, Vector default)
 */
struct __pyx_opt_args_7flitter_5model_4Node_get {
  int __pyx_n;
  int n;
  PyTypeObject *t;
  PyObject *__pyx_default;
};
struct __pyx_opt_args_7cpython_11contextvars_get_value;
struct __pyx_opt_args_7cpython_11contextvars_get_value_no_default;

/* "cpython/contextvars.pxd":112
 * 
 * 
 * cdef inline object get_value(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the default value of the context variable,
 */
struct __pyx_opt_args_7cpython_11contextvars_get_value {
  int __pyx_n;
  PyObject *default_value;
};

/* "cpython/contextvars.pxd":129
 * 
 * 
 * cdef inline object get_value_no_default(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the provided default value if no such value was found.
 */
struct __pyx_opt_args_7cpython_11contextvars_get_value_no_default {
  int __pyx_n;
  PyObject *default_value;
};

/* "model.pxd":35
 * 
 * 
 * cdef class Vector:             # <<<<<<<<<<<<<<
 *     cdef int64_t length
 *     cdef tuple objects
 */
struct __pyx_obj_7flitter_5model_Vector {
  PyObject_HEAD
  struct __pyx_vtabstruct_7flitter_5model_Vector *__pyx_vtab;
  int64_t length;
  PyObject *objects;
  double *numbers;
  double _numbers[16];
  uint64_t _hash;
};


/* "model.pxd":107
 * 
 * 
 * cdef class Matrix33(Vector):             # <<<<<<<<<<<<<<
 *     @staticmethod
 *     cdef Matrix33 _identity()
 */
struct __pyx_obj_7flitter_5model_Matrix33 {
  struct __pyx_obj_7flitter_5model_Vector __pyx_base;
};


/* "model.pxd":130
 * 
 * 
 * cdef class Matrix44(Vector):             # <<<<<<<<<<<<<<
 *     @staticmethod
 *     cdef Matrix44 _identity()
 */
struct __pyx_obj_7flitter_5model_Matrix44 {
  struct __pyx_obj_7flitter_5model_Vector __pyx_base;
};


/* "model.pxd":181
 * 
 * 
 * cdef class Quaternion(Vector):             # <<<<<<<<<<<<<<
 *     @staticmethod
 *     cdef Quaternion _coerce(other)
 */
struct __pyx_obj_7flitter_5model_Quaternion {
  struct __pyx_obj_7flitter_5model_Vector __pyx_base;
};


/* "model.pxd":201
 * 
 * 
 * cdef class Node:             # <<<<<<<<<<<<<<
 *     cdef readonly str kind
 *     cdef set _tags
 */
struct __pyx_obj_7flitter_5model_Node {
  PyObject_HEAD
  struct __pyx_vtabstruct_7flitter_5model_Node *__pyx_vtab;
  PyObject *kind;
  PyObject *_tags;
  PyObject *_attributes;
  int _attributes_shared;
  PyObject *_children;
};


/* "model.pxd":225
 * 
 * 
 * cdef class StateDict:             # <<<<<<<<<<<<<<
 *     cdef set _changed_keys
 *     cdef dict _state
 */
struct __pyx_obj_7flitter_5model_StateDict {
  PyObject_HEAD
  struct __pyx_vtabstruct_7flitter_5model_StateDict *__pyx_vtab;
  PyObject *_changed_keys;
  PyObject *_state;
};


/* "model.pxd":234
 * 
 * 
 * cdef class Context:             # <<<<<<<<<<<<<<
 *     cdef readonly dict names
 *     cdef readonly set captures
 */
struct __pyx_obj_7flitter_5model_Context {
  PyObject_HEAD
  PyObject *names;
  PyObject *captures;
  struct __pyx_obj_7flitter_5model_StateDict *state;
  struct __pyx_obj_7flitter_5model_Node *root;
  PyObject *path;
  struct __pyx_obj_7flitter_5model_Context *parent;
  PyObject *exports;
  PyObject *errors;
  PyObject *logs;
  PyObject *references;
  PyObject *stack;
  PyObject *lnames;
  PyObject *dependencies;
};


/* "language/functions.pxd":5
 * 
 * 
 * cdef class uniform(Vector):             # <<<<<<<<<<<<<<
 *     cdef double _item(self, unsigned long long i) noexcept nogil
 *     cpdef Vector slice(self, Vector index)
 */
struct __pyx_obj_7flitter_8language_9functions_uniform {
  struct __pyx_obj_7flitter_5model_Vector __pyx_base;
};


/* "language/functions.pxd":11
 * 
 * 
 * cdef class beta(uniform):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_7flitter_8language_9functions_beta {
  struct __pyx_obj_7flitter_8language_9functions_uniform __pyx_base;
};


/* "language/functions.pxd":15
 * 
 * 
 * cdef class normal(uniform):             # <<<<<<<<<<<<<<
 *     cdef bint cached
 *     cdef unsigned long long i
 */
struct __pyx_obj_7flitter_8language_9functions_normal {
  struct __pyx_obj_7flitter_8language_9functions_uniform __pyx_base;
  int cached;
  unsigned PY_LONG_LONG i;
  double R;
  double th;
};


/* "flitter/render/physics.pyx":37
 * 
 * 
 * cdef class PhysicsGroup:             # <<<<<<<<<<<<<<
 *     cdef PhysicsGroup parent
 *     cdef list particles
 */
struct __pyx_obj_7flitter_6render_7physics_PhysicsGroup {
  PyObject_HEAD
  struct __pyx_obj_7flitter_6render_7physics_PhysicsGroup *parent;
  PyObject *particles;
  PyObject *particle_forces;
  PyObject *matrix_forces;
  PyObject *distance_forces;
  PyObject *barriers;
};


/* "flitter/render/physics.pyx":54
 * 
 * @cython.final
 * cdef class Particle:             # <<<<<<<<<<<<<<
 *     cdef Vector id
 *     cdef bint non_anchor
 */
struct __pyx_obj_7flitter_6render_7physics_Particle {
  PyObject_HEAD
  struct __pyx_vtabstruct_7flitter_6render_7physics_Particle *__pyx_vtab;
  struct __pyx_obj_7flitter_5model_Vector *id;
  int non_anchor;
  struct __pyx_obj_7flitter_5model_Vector *position_state_key;
  struct __pyx_obj_7flitter_5model_Vector *position;
  struct __pyx_obj_7flitter_5model_Vector *velocity_state_key;
  struct __pyx_obj_7flitter_5model_Vector *velocity;
  struct __pyx_obj_7flitter_5model_Vector *acceleration;
  struct __pyx_obj_7flitter_5model_Vector *initial_force;
  struct __pyx_obj_7flitter_5model_Vector *force;
  double radius;
  double mass;
  double charge;
  double ease;
};


/* "flitter/render/physics.pyx":98
 * 
 * @cython.final
 * cdef class Barrier:             # <<<<<<<<<<<<<<
 *     cdef Vector position
 *     cdef Vector normal
 */
struct __pyx_obj_7flitter_6render_7physics_Barrier {
  PyObject_HEAD
  struct __pyx_vtabstruct_7flitter_6render_7physics_Barrier *__pyx_vtab;
  struct __pyx_obj_7flitter_5model_Vector *position;
  struct __pyx_obj_7flitter_5model_Vector *normal;
  double restitution;
  double strength;
  double power;
  double minimum;
};


/* "flitter/render/physics.pyx":166
 * 
 * 
 * cdef class ParticleForceApplier:             # <<<<<<<<<<<<<<
 *     cdef double strength
 * 
 */
struct __pyx_obj_7flitter_6render_7physics_ParticleForceApplier {
  PyObject_HEAD
  struct __pyx_vtabstruct_7flitter_6render_7physics_ParticleForceApplier *__pyx_vtab;
  double strength;
};


/* "flitter/render/physics.pyx":178
 * 
 * 
 * cdef class MatrixForceApplier:             # <<<<<<<<<<<<<<
 *     cdef double strength
 *     cdef double max_distance
 */
struct __pyx_obj_7flitter_6render_7physics_MatrixForceApplier {
  PyObject_HEAD
  struct __pyx_vtabstruct_7flitter_6render_7physics_MatrixForceApplier *__pyx_vtab;
  double strength;
  double max_distance;
};


/* "flitter/render/physics.pyx":193
 * 
 * @cython.final
 * cdef class DistanceForceApplier:             # <<<<<<<<<<<<<<
 *     cdef double strength
 *     cdef Vector from_particle_id
 */
struct __pyx_obj_7flitter_6render_7physics_DistanceForceApplier {
  PyObject_HEAD
  struct __pyx_vtabstruct_7flitter_6render_7physics_DistanceForceApplier *__pyx_vtab;
  double strength;
  struct __pyx_obj_7flitter_5model_Vector *from_particle_id;
  struct __pyx_obj_7flitter_5model_Vector *to_particle_id;
  double power;
  double minimum;
  double maximum;
};


/* "flitter/render/physics.pyx":225
 * 
 * 
 * cdef class DragForceApplier(ParticleForceApplier):             # <<<<<<<<<<<<<<
 *     cdef Vector flow
 * 
 */
struct __pyx_obj_7flitter_6render_7physics_DragForceApplier {
  struct __pyx_obj_7flitter_6render_7physics_ParticleForceApplier __pyx_base;
  struct __pyx_obj_7flitter_5model_Vector *flow;
};


/* "flitter/render/physics.pyx":245
 * 
 * 
 * cdef class BuoyancyForceApplier(ParticleForceApplier):             # <<<<<<<<<<<<<<
 *     cdef double density
 *     cdef Vector gravity
 */
struct __pyx_obj_7flitter_6render_7physics_BuoyancyForceApplier {
  struct __pyx_obj_7flitter_6render_7physics_ParticleForceApplier __pyx_base;
  double density;
  struct __pyx_obj_7flitter_5model_Vector *gravity;
};


/* "flitter/render/physics.pyx":268
 * 
 * 
 * cdef class ConstantForceApplier(ParticleForceApplier):             # <<<<<<<<<<<<<<
 *     cdef Vector force
 *     cdef Vector acceleration
 */
struct __pyx_obj_7flitter_6render_7physics_ConstantForceApplier {
  struct __pyx_obj_7flitter_6render_7physics_ParticleForceApplier __pyx_base;
  struct __pyx_obj_7flitter_5model_Vector *force;
  struct __pyx_obj_7flitter_5model_Vector *acceleration;
};


/* "flitter/render/physics.pyx":286
 * 
 * 
 * cdef class RandomForceApplier(ParticleForceApplier):             # <<<<<<<<<<<<<<
 *     cdef normal random_source
 *     cdef uint64_t index
 */
struct __pyx_obj_7flitter_6render_7physics_RandomForceApplier {
  struct __pyx_obj_7flitter_6render_7physics_ParticleForceApplier __pyx_base;
  struct __pyx_obj_7flitter_8language_9functions_normal *random_source;
  uint64_t index;
};


/* "flitter/render/physics.pyx":302
 * 
 * 
 * cdef class FieldForceApplier(ParticleForceApplier):             # <<<<<<<<<<<<<<
 *     cdef Vector direction
 * 
 */
struct __pyx_obj_7flitter_6render_7physics_FieldForceApplier {
  struct __pyx_obj_7flitter_6render_7physics_ParticleForceApplier __pyx_base;
  struct __pyx_obj_7flitter_5model_Vector *direction;
};


/* "flitter/render/physics.pyx":316
 * 
 * 
 * cdef class CollisionForceApplier(MatrixForceApplier):             # <<<<<<<<<<<<<<
 *     cdef double power
 * 
 */
struct __pyx_obj_7flitter_6render_7physics_CollisionForceApplier {
  struct __pyx_obj_7flitter_6render_7physics_MatrixForceApplier __pyx_base;
  double power;
};


/* "flitter/render/physics.pyx":340
 * 
 * 
 * cdef class GravityForceApplier(MatrixForceApplier):             # <<<<<<<<<<<<<<
 *     @cython.profile(False)
 *     cdef void apply(self, Particle from_particle, Particle to_particle, Vector direction, double distance) noexcept nogil:
 */
struct __pyx_obj_7flitter_6render_7physics_GravityForceApplier {
  struct __pyx_obj_7flitter_6render_7physics_MatrixForceApplier __pyx_base;
};


/* "flitter/render/physics.pyx":354
 * 
 * 
 * cdef class ElectrostaticForceApplier(MatrixForceApplier):             # <<<<<<<<<<<<<<
 *     @cython.profile(False)
 *     cdef void apply(self, Particle from_particle, Particle to_particle, Vector direction, double distance) noexcept nogil:
 */
struct __pyx_obj_7flitter_6render_7physics_ElectrostaticForceApplier {
  struct __pyx_obj_7flitter_6render_7physics_MatrixForceApplier __pyx_base;
};


/* "flitter/render/physics.pyx":368
 * 
 * 
 * cdef class AdhesionForceApplier(MatrixForceApplier):             # <<<<<<<<<<<<<<
 *     cdef double overlap
 * 
 */
struct __pyx_obj_7flitter_6render_7physics_AdhesionForceApplier {
  struct __pyx_obj_7flitter_6render_7physics_MatrixForceApplier __pyx_base;
  double overlap;
};


/* "flitter/render/physics.pyx":392
 * 
 * 
 * cdef class PhysicsSystem:             # <<<<<<<<<<<<<<
 *     cdef set state_keys
 *     cdef int64_t dimensions
 */
struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem {
  PyObject_HEAD
  struct __pyx_vtabstruct_7flitter_6render_7physics_PhysicsSystem *__pyx_vtab;
  PyObject *state_keys;
  int64_t dimensions;
  double resolution;
  double speed_of_light;
};


/* "flitter/render/physics.pyx":401
 *         self.state_keys = set()
 * 
 *     async def destroy(self):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct__destroy {
  PyObject_HEAD
  struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem *__pyx_v_self;
};


/* "flitter/render/physics.pyx":404
 *         pass
 * 
 *     async def purge(self):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct_1_purge {
  PyObject_HEAD
  struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem *__pyx_v_self;
};


/* "flitter/render/physics.pyx":407
 *         pass
 * 
 *     async def update(self, engine, Node node, double time, double performance, **kwargs):             # <<<<<<<<<<<<<<
 *         cdef int64_t run = node.get_int('run', 0)
 *         self.dimensions = node.get_int('dimensions', 0)
 */
struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct_2_update {
  PyObject_HEAD
  PyObject *__pyx_v_clock;
  PyObject *__pyx_v_engine;
  int __pyx_v_extra;
  struct __pyx_obj_7flitter_6render_7physics_PhysicsGroup *__pyx_v_group;
  PyObject *__pyx_v_groups;
  int64_t __pyx_v_i;
  int64_t __pyx_v_iteration;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_iteration_vector;
  PyObject *__pyx_v_kwargs;
  int64_t __pyx_v_last_run;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_last_run_vector;
  double __pyx_v_last_time;
  PyObject *__pyx_v_new_state_keys;
  struct __pyx_obj_7flitter_5model_Node *__pyx_v_node;
  PyObject *__pyx_v_old_state_keys;
  struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_particle;
  PyObject *__pyx_v_particles_by_id;
  double __pyx_v_performance;
  int64_t __pyx_v_run;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_run_vector;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_seed;
  struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem *__pyx_v_self;
  struct __pyx_obj_7flitter_5model_StateDict *__pyx_v_state;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_state_key;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_state_prefix;
  double __pyx_v_time;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_time_vector;
  struct __pyx_obj_7flitter_6render_7physics_PhysicsGroup *__pyx_v_top_group;
};



/* "model.pxd":35
 * 
 * 
 * cdef class Vector:             # <<<<<<<<<<<<<<
 *     cdef int64_t length
 *     cdef tuple objects
 */

struct __pyx_vtabstruct_7flitter_5model_Vector {
  struct __pyx_obj_7flitter_5model_Vector *(*_coerce)(PyObject *);
  struct __pyx_obj_7flitter_5model_Vector *(*_compose)(PyObject *);
  struct __pyx_obj_7flitter_5model_Vector *(*_symbol)(PyObject *);
  int64_t (*allocate_numbers)(struct __pyx_obj_7flitter_5model_Vector *, int64_t);
  void (*deallocate_numbers)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*copy)(struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Vector *(*intern)(struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  void (*fill_range)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  int (*isinstance)(struct __pyx_obj_7flitter_5model_Vector *, PyObject *, int __pyx_skip_dispatch);
  int (*is_finite)(struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  int (*as_bool)(struct __pyx_obj_7flitter_5model_Vector *);
  double (*as_double)(struct __pyx_obj_7flitter_5model_Vector *);
  int64_t (*as_integer)(struct __pyx_obj_7flitter_5model_Vector *);
  PyObject *(*as_string)(struct __pyx_obj_7flitter_5model_Vector *);
  uint64_t (*hash)(struct __pyx_obj_7flitter_5model_Vector *, int, int __pyx_skip_dispatch);
  PyObject *(*match)(struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch, struct __pyx_opt_args_7flitter_5model_6Vector_match *__pyx_optional_args);
  PyObject *(*repr)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*neg)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*pos)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*abs)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*ceil)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*floor)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*fract)(struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Vector *(*round)(struct __pyx_obj_7flitter_5model_Vector *, int64_t);
  struct __pyx_obj_7flitter_5model_Vector *(*add)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*mul_add)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Vector *(*sub)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*mul)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*truediv)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*floordiv)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*mod)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*pow)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*eq)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*ne)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*gt)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*ge)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*lt)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*le)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  int64_t (*compare)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Vector *(*slice)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Vector *(*item)(struct __pyx_obj_7flitter_5model_Vector *, int64_t, int __pyx_skip_dispatch);
  double (*squared_sum)(struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Vector *(*normalize)(struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Vector *(*dot)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Vector *(*cross)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Vector *(*clamp)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  double (*minimum)(struct __pyx_obj_7flitter_5model_Vector *);
  double (*maximum)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*concat)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7flitter_5model_Vector *__pyx_vtabptr_7flitter_5model_Vector;


/* "model.pxd":107
 * 
 * 
 * cdef class Matrix33(Vector):             # <<<<<<<<<<<<<<
 *     @staticmethod
 *     cdef Matrix33 _identity()
 */

struct __pyx_vtabstruct_7flitter_5model_Matrix33 {
  struct __pyx_vtabstruct_7flitter_5model_Vector __pyx_base;
  struct __pyx_obj_7flitter_5model_Matrix33 *(*_identity)(void);
  struct __pyx_obj_7flitter_5model_Matrix33 *(*_translate)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Matrix33 *(*_scale)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Matrix33 *(*_rotate)(double);
  struct __pyx_obj_7flitter_5model_Matrix33 *(*copy)(struct __pyx_obj_7flitter_5model_Matrix33 *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Matrix33 *(*mmul)(struct __pyx_obj_7flitter_5model_Matrix33 *, struct __pyx_obj_7flitter_5model_Matrix33 *);
  struct __pyx_obj_7flitter_5model_Vector *(*vmul)(struct __pyx_obj_7flitter_5model_Matrix33 *, struct __pyx_obj_7flitter_5model_Vector *);
  double (*det)(struct __pyx_obj_7flitter_5model_Matrix33 *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Matrix33 *(*inverse)(struct __pyx_obj_7flitter_5model_Matrix33 *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Matrix33 *(*cofactor)(struct __pyx_obj_7flitter_5model_Matrix33 *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Matrix33 *(*transpose)(struct __pyx_obj_7flitter_5model_Matrix33 *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*matrix44)(struct __pyx_obj_7flitter_5model_Matrix33 *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7flitter_5model_Matrix33 *__pyx_vtabptr_7flitter_5model_Matrix33;


/* "model.pxd":130
 * 
 * 
 * cdef class Matrix44(Vector):             # <<<<<<<<<<<<<<
 *     @staticmethod
 *     cdef Matrix44 _identity()
 */

struct __pyx_vtabstruct_7flitter_5model_Matrix44 {
  struct __pyx_vtabstruct_7flitter_5model_Vector __pyx_base;
  struct __pyx_obj_7flitter_5model_Matrix44 *(*_identity)(void);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*_project)(double, double, double, double);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*_ortho)(double, double, double, double);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*_look)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*_translate)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*_scale)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*_rotate_x)(double);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*_rotate_y)(double);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*_rotate_z)(double);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*_rotate)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*_shear_x)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*_shear_y)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*_shear_z)(struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*copy)(struct __pyx_obj_7flitter_5model_Matrix44 *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*mmul)(struct __pyx_obj_7flitter_5model_Matrix44 *, struct __pyx_obj_7flitter_5model_Matrix44 *);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*immul)(struct __pyx_obj_7flitter_5model_Matrix44 *, struct __pyx_obj_7flitter_5model_Matrix44 *);
  struct __pyx_obj_7flitter_5model_Vector *(*vmul)(struct __pyx_obj_7flitter_5model_Matrix44 *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*inverse)(struct __pyx_obj_7flitter_5model_Matrix44 *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*transpose)(struct __pyx_obj_7flitter_5model_Matrix44 *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Matrix33 *(*inverse_transpose_matrix33)(struct __pyx_obj_7flitter_5model_Matrix44 *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Matrix33 *(*matrix33_cofactor)(struct __pyx_obj_7flitter_5model_Matrix44 *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Matrix33 *(*matrix33)(struct __pyx_obj_7flitter_5model_Matrix44 *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7flitter_5model_Matrix44 *__pyx_vtabptr_7flitter_5model_Matrix44;


/* "model.pxd":181
 * 
 * 
 * cdef class Quaternion(Vector):             # <<<<<<<<<<<<<<
 *     @staticmethod
 *     cdef Quaternion _coerce(other)
 */

struct __pyx_vtabstruct_7flitter_5model_Quaternion {
  struct __pyx_vtabstruct_7flitter_5model_Vector __pyx_base;
  struct __pyx_obj_7flitter_5model_Quaternion *(*_coerce)(PyObject *);
  struct __pyx_obj_7flitter_5model_Quaternion *(*_euler)(struct __pyx_obj_7flitter_5model_Vector *, double);
  struct __pyx_obj_7flitter_5model_Quaternion *(*_between)(struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Quaternion *(*copy)(struct __pyx_obj_7flitter_5model_Quaternion *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Quaternion *(*mmul)(struct __pyx_obj_7flitter_5model_Quaternion *, struct __pyx_obj_7flitter_5model_Quaternion *);
  struct __pyx_obj_7flitter_5model_Vector *(*conjugate)(struct __pyx_obj_7flitter_5model_Quaternion *, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Quaternion *(*normalize)(struct __pyx_obj_7flitter_5model_Quaternion *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Quaternion *(*inverse)(struct __pyx_obj_7flitter_5model_Quaternion *, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Quaternion *(*exponent)(struct __pyx_obj_7flitter_5model_Quaternion *, double, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Quaternion *(*slerp)(struct __pyx_obj_7flitter_5model_Quaternion *, struct __pyx_obj_7flitter_5model_Quaternion *, double, int __pyx_skip_dispatch);
  struct __pyx_obj_7flitter_5model_Matrix44 *(*matrix44)(struct __pyx_obj_7flitter_5model_Quaternion *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7flitter_5model_Quaternion *__pyx_vtabptr_7flitter_5model_Quaternion;


/* "model.pxd":201
 * 
 * 
 * cdef class Node:             # <<<<<<<<<<<<<<
 *     cdef readonly str kind
 *     cdef set _tags
 */

struct __pyx_vtabstruct_7flitter_5model_Node {
  uint64_t (*hash)(struct __pyx_obj_7flitter_5model_Node *);
  struct __pyx_obj_7flitter_5model_Node *(*copy)(struct __pyx_obj_7flitter_5model_Node *, int __pyx_skip_dispatch);
  void (*add_tag)(struct __pyx_obj_7flitter_5model_Node *, PyObject *, int __pyx_skip_dispatch);
  void (*set_attribute)(struct __pyx_obj_7flitter_5model_Node *, PyObject *, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  void (*append)(struct __pyx_obj_7flitter_5model_Node *, struct __pyx_obj_7flitter_5model_Node *, int __pyx_skip_dispatch);
  void (*append_vector)(struct __pyx_obj_7flitter_5model_Node *, struct __pyx_obj_7flitter_5model_Vector *, int __pyx_skip_dispatch);
  int (*_equals)(struct __pyx_obj_7flitter_5model_Node *, struct __pyx_obj_7flitter_5model_Node *);
  PyObject *(*get)(struct __pyx_obj_7flitter_5model_Node *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_7flitter_5model_4Node_get *__pyx_optional_args);
  struct __pyx_obj_7flitter_5model_Vector *(*get_vec)(struct __pyx_obj_7flitter_5model_Node *, PyObject *, struct __pyx_obj_7flitter_5model_Vector *);
  struct __pyx_obj_7flitter_5model_Vector *(*get_fvec)(struct __pyx_obj_7flitter_5model_Node *, PyObject *, int, struct __pyx_obj_7flitter_5model_Vector *);
  double (*get_float)(struct __pyx_obj_7flitter_5model_Node *, PyObject *, double);
  int64_t (*get_int)(struct __pyx_obj_7flitter_5model_Node *, PyObject *, int64_t);
  int (*get_bool)(struct __pyx_obj_7flitter_5model_Node *, PyObject *, int);
  PyObject *(*get_str)(struct __pyx_obj_7flitter_5model_Node *, PyObject *, PyObject *);
  void (*repr)(struct __pyx_obj_7flitter_5model_Node *, PyObject *, int);
};
static struct __pyx_vtabstruct_7flitter_5model_Node *__pyx_vtabptr_7flitter_5model_Node;


/* "model.pxd":225
 * 
 * 
 * cdef class StateDict:             # <<<<<<<<<<<<<<
 *     cdef set _changed_keys
 *     cdef dict _state
 */

struct __pyx_vtabstruct_7flitter_5model_StateDict {
  struct __pyx_obj_7flitter_5model_Vector *(*get_item)(struct __pyx_obj_7flitter_5model_StateDict *, struct __pyx_obj_7flitter_5model_Vector *);
  void (*set_item)(struct __pyx_obj_7flitter_5model_StateDict *, struct __pyx_obj_7flitter_5model_Vector *, struct __pyx_obj_7flitter_5model_Vector *);
  int (*contains)(struct __pyx_obj_7flitter_5model_StateDict *, struct __pyx_obj_7flitter_5model_Vector *);
};
static struct __pyx_vtabstruct_7flitter_5model_StateDict *__pyx_vtabptr_7flitter_5model_StateDict;


/* "language/functions.pxd":5
 * 
 * 
 * cdef class uniform(Vector):             # <<<<<<<<<<<<<<
 *     cdef double _item(self, unsigned long long i) noexcept nogil
 *     cpdef Vector slice(self, Vector index)
 */

struct __pyx_vtabstruct_7flitter_8language_9functions_uniform {
  struct __pyx_vtabstruct_7flitter_5model_Vector __pyx_base;
  double (*_item)(struct __pyx_obj_7flitter_8language_9functions_uniform *, unsigned PY_LONG_LONG);
  int (*as_bool)(struct __pyx_obj_7flitter_8language_9functions_uniform *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7flitter_8language_9functions_uniform *__pyx_vtabptr_7flitter_8language_9functions_uniform;


/* "language/functions.pxd":11
 * 
 * 
 * cdef class beta(uniform):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_7flitter_8language_9functions_beta {
  struct __pyx_vtabstruct_7flitter_8language_9functions_uniform __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_8language_9functions_beta *__pyx_vtabptr_7flitter_8language_9functions_beta;


/* "language/functions.pxd":15
 * 
 * 
 * cdef class normal(uniform):             # <<<<<<<<<<<<<<
 *     cdef bint cached
 *     cdef unsigned long long i
 */

struct __pyx_vtabstruct_7flitter_8language_9functions_normal {
  struct __pyx_vtabstruct_7flitter_8language_9functions_uniform __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_8language_9functions_normal *__pyx_vtabptr_7flitter_8language_9functions_normal;


/* "flitter/render/physics.pyx":54
 * 
 * @cython.final
 * cdef class Particle:             # <<<<<<<<<<<<<<
 *     cdef Vector id
 *     cdef bint non_anchor
 */

struct __pyx_vtabstruct_7flitter_6render_7physics_Particle {
  void (*step)(struct __pyx_obj_7flitter_6render_7physics_Particle *, double, double, double);
};
static struct __pyx_vtabstruct_7flitter_6render_7physics_Particle *__pyx_vtabptr_7flitter_6render_7physics_Particle;
static void __pyx_f_7flitter_6render_7physics_8Particle_step(struct __pyx_obj_7flitter_6render_7physics_Particle *, double, double, double);


/* "flitter/render/physics.pyx":98
 * 
 * @cython.final
 * cdef class Barrier:             # <<<<<<<<<<<<<<
 *     cdef Vector position
 *     cdef Vector normal
 */

struct __pyx_vtabstruct_7flitter_6render_7physics_Barrier {
  void (*apply_distance)(struct __pyx_obj_7flitter_6render_7physics_Barrier *, struct __pyx_obj_7flitter_6render_7physics_Particle *);
  void (*apply_rebound)(struct __pyx_obj_7flitter_6render_7physics_Barrier *, struct __pyx_obj_7flitter_6render_7physics_Particle *, double, double, double);
};
static struct __pyx_vtabstruct_7flitter_6render_7physics_Barrier *__pyx_vtabptr_7flitter_6render_7physics_Barrier;
static void __pyx_f_7flitter_6render_7physics_7Barrier_apply_distance(struct __pyx_obj_7flitter_6render_7physics_Barrier *, struct __pyx_obj_7flitter_6render_7physics_Particle *);
static void __pyx_f_7flitter_6render_7physics_7Barrier_apply_rebound(struct __pyx_obj_7flitter_6render_7physics_Barrier *, struct __pyx_obj_7flitter_6render_7physics_Particle *, double, double, double);


/* "flitter/render/physics.pyx":166
 * 
 * 
 * cdef class ParticleForceApplier:             # <<<<<<<<<<<<<<
 *     cdef double strength
 * 
 */

struct __pyx_vtabstruct_7flitter_6render_7physics_ParticleForceApplier {
  void (*apply)(struct __pyx_obj_7flitter_6render_7physics_ParticleForceApplier *, struct __pyx_obj_7flitter_6render_7physics_Particle *, double);
};
static struct __pyx_vtabstruct_7flitter_6render_7physics_ParticleForceApplier *__pyx_vtabptr_7flitter_6render_7physics_ParticleForceApplier;


/* "flitter/render/physics.pyx":178
 * 
 * 
 * cdef class MatrixForceApplier:             # <<<<<<<<<<<<<<
 *     cdef double strength
 *     cdef double max_distance
 */

struct __pyx_vtabstruct_7flitter_6render_7physics_MatrixForceApplier {
  void (*apply)(struct __pyx_obj_7flitter_6render_7physics_MatrixForceApplier *, struct __pyx_obj_7flitter_6render_7physics_Particle *, struct __pyx_obj_7flitter_6render_7physics_Particle *, struct __pyx_obj_7flitter_5model_Vector *, double);
};
static struct __pyx_vtabstruct_7flitter_6render_7physics_MatrixForceApplier *__pyx_vtabptr_7flitter_6render_7physics_MatrixForceApplier;


/* "flitter/render/physics.pyx":193
 * 
 * @cython.final
 * cdef class DistanceForceApplier:             # <<<<<<<<<<<<<<
 *     cdef double strength
 *     cdef Vector from_particle_id
 */

struct __pyx_vtabstruct_7flitter_6render_7physics_DistanceForceApplier {
  void (*apply)(struct __pyx_obj_7flitter_6render_7physics_DistanceForceApplier *, struct __pyx_obj_7flitter_6render_7physics_Particle *, struct __pyx_obj_7flitter_6render_7physics_Particle *, struct __pyx_obj_7flitter_5model_Vector *, double);
};
static struct __pyx_vtabstruct_7flitter_6render_7physics_DistanceForceApplier *__pyx_vtabptr_7flitter_6render_7physics_DistanceForceApplier;
static void __pyx_f_7flitter_6render_7physics_20DistanceForceApplier_apply(struct __pyx_obj_7flitter_6render_7physics_DistanceForceApplier *, struct __pyx_obj_7flitter_6render_7physics_Particle *, struct __pyx_obj_7flitter_6render_7physics_Particle *, struct __pyx_obj_7flitter_5model_Vector *, double);


/* "flitter/render/physics.pyx":225
 * 
 * 
 * cdef class DragForceApplier(ParticleForceApplier):             # <<<<<<<<<<<<<<
 *     cdef Vector flow
 * 
 */

struct __pyx_vtabstruct_7flitter_6render_7physics_DragForceApplier {
  struct __pyx_vtabstruct_7flitter_6render_7physics_ParticleForceApplier __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_6render_7physics_DragForceApplier *__pyx_vtabptr_7flitter_6render_7physics_DragForceApplier;


/* "flitter/render/physics.pyx":245
 * 
 * 
 * cdef class BuoyancyForceApplier(ParticleForceApplier):             # <<<<<<<<<<<<<<
 *     cdef double density
 *     cdef Vector gravity
 */

struct __pyx_vtabstruct_7flitter_6render_7physics_BuoyancyForceApplier {
  struct __pyx_vtabstruct_7flitter_6render_7physics_ParticleForceApplier __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_6render_7physics_BuoyancyForceApplier *__pyx_vtabptr_7flitter_6render_7physics_BuoyancyForceApplier;


/* "flitter/render/physics.pyx":268
 * 
 * 
 * cdef class ConstantForceApplier(ParticleForceApplier):             # <<<<<<<<<<<<<<
 *     cdef Vector force
 *     cdef Vector acceleration
 */

struct __pyx_vtabstruct_7flitter_6render_7physics_ConstantForceApplier {
  struct __pyx_vtabstruct_7flitter_6render_7physics_ParticleForceApplier __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_6render_7physics_ConstantForceApplier *__pyx_vtabptr_7flitter_6render_7physics_ConstantForceApplier;


/* "flitter/render/physics.pyx":286
 * 
 * 
 * cdef class RandomForceApplier(ParticleForceApplier):             # <<<<<<<<<<<<<<
 *     cdef normal random_source
 *     cdef uint64_t index
 */

struct __pyx_vtabstruct_7flitter_6render_7physics_RandomForceApplier {
  struct __pyx_vtabstruct_7flitter_6render_7physics_ParticleForceApplier __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_6render_7physics_RandomForceApplier *__pyx_vtabptr_7flitter_6render_7physics_RandomForceApplier;


/* "flitter/render/physics.pyx":302
 * 
 * 
 * cdef class FieldForceApplier(ParticleForceApplier):             # <<<<<<<<<<<<<<
 *     cdef Vector direction
 * 
 */

struct __pyx_vtabstruct_7flitter_6render_7physics_FieldForceApplier {
  struct __pyx_vtabstruct_7flitter_6render_7physics_ParticleForceApplier __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_6render_7physics_FieldForceApplier *__pyx_vtabptr_7flitter_6render_7physics_FieldForceApplier;


/* "flitter/render/physics.pyx":316
 * 
 * 
 * cdef class CollisionForceApplier(MatrixForceApplier):             # <<<<<<<<<<<<<<
 *     cdef double power
 * 
 */

struct __pyx_vtabstruct_7flitter_6render_7physics_CollisionForceApplier {
  struct __pyx_vtabstruct_7flitter_6render_7physics_MatrixForceApplier __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_6render_7physics_CollisionForceApplier *__pyx_vtabptr_7flitter_6render_7physics_CollisionForceApplier;


/* "flitter/render/physics.pyx":340
 * 
 * 
 * cdef class GravityForceApplier(MatrixForceApplier):             # <<<<<<<<<<<<<<
 *     @cython.profile(False)
 *     cdef void apply(self, Particle from_particle, Particle to_particle, Vector direction, double distance) noexcept nogil:
 */

struct __pyx_vtabstruct_7flitter_6render_7physics_GravityForceApplier {
  struct __pyx_vtabstruct_7flitter_6render_7physics_MatrixForceApplier __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_6render_7physics_GravityForceApplier *__pyx_vtabptr_7flitter_6render_7physics_GravityForceApplier;


/* "flitter/render/physics.pyx":354
 * 
 * 
 * cdef class ElectrostaticForceApplier(MatrixForceApplier):             # <<<<<<<<<<<<<<
 *     @cython.profile(False)
 *     cdef void apply(self, Particle from_particle, Particle to_particle, Vector direction, double distance) noexcept nogil:
 */

struct __pyx_vtabstruct_7flitter_6render_7physics_ElectrostaticForceApplier {
  struct __pyx_vtabstruct_7flitter_6render_7physics_MatrixForceApplier __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_6render_7physics_ElectrostaticForceApplier *__pyx_vtabptr_7flitter_6render_7physics_ElectrostaticForceApplier;


/* "flitter/render/physics.pyx":368
 * 
 * 
 * cdef class AdhesionForceApplier(MatrixForceApplier):             # <<<<<<<<<<<<<<
 *     cdef double overlap
 * 
 */

struct __pyx_vtabstruct_7flitter_6render_7physics_AdhesionForceApplier {
  struct __pyx_vtabstruct_7flitter_6render_7physics_MatrixForceApplier __pyx_base;
};
static struct __pyx_vtabstruct_7flitter_6render_7physics_AdhesionForceApplier *__pyx_vtabptr_7flitter_6render_7physics_AdhesionForceApplier;


/* "flitter/render/physics.pyx":392
 * 
 * 
 * cdef class PhysicsSystem:             # <<<<<<<<<<<<<<
 *     cdef set state_keys
 *     cdef int64_t dimensions
 */

struct __pyx_vtabstruct_7flitter_6render_7physics_PhysicsSystem {
  void (*collect)(struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem *, PyObject *, struct __pyx_obj_7flitter_6render_7physics_PhysicsGroup *, struct __pyx_obj_7flitter_5model_Node *, PyObject *, PyObject *, PyObject *, struct __pyx_obj_7flitter_5model_StateDict *, struct __pyx_obj_7flitter_5model_Vector *, double, struct __pyx_obj_7flitter_5model_Vector *);
  double (*calculate)(struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem *, PyObject *, PyObject *, int, int, double, double, double, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7flitter_6render_7physics_PhysicsSystem *__pyx_vtabptr_7flitter_6render_7physics_PhysicsSystem;
/* #### Code section: utility_code_proto ### */

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, Py_ssize_t);
    void (*DECREF)(void*, PyObject*, Py_ssize_t);
    void (*GOTREF)(void*, PyObject*, Py_ssize_t);
    void (*GIVEREF)(void*, PyObject*, Py_ssize_t);
    void* (*SetupContext)(const char*, Py_ssize_t, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__))
  #define __Pyx_RefNannyFinishContextNogil() __Pyx_RefNannyFinishContext()
#endif
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_XINCREF(r)  do { if((r) == NULL); else {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) == NULL); else {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) == NULL); else {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) == NULL); else {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContextNogil()
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_Py_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; Py_XDECREF(tmp);\
    } while (0)
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#if PY_VERSION_HEX >= 0x030C00A6
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->current_exception != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->current_exception ? (PyObject*) Py_TYPE(__pyx_tstate->current_exception) : (PyObject*) NULL)
#else
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->curexc_type != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->curexc_type)
#endif
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  (PyErr_Occurred() != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A6
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* TupleAndListFromArray.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n);
static CYTHON_INLINE PyObject* __Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n);
#endif

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* fastcall.proto */
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_Arg_VARARGS(args, i) PySequence_GetItem(args, i)
#elif CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_Arg_VARARGS(args, i) PyTuple_GET_ITEM(args, i)
#else
    #define __Pyx_Arg_VARARGS(args, i) PyTuple_GetItem(args, i)
#endif
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_Arg_NewRef_VARARGS(arg) __Pyx_NewRef(arg)
    #define __Pyx_Arg_XDECREF_VARARGS(arg) Py_XDECREF(arg)
#else
    #define __Pyx_Arg_NewRef_VARARGS(arg) arg
    #define __Pyx_Arg_XDECREF_VARARGS(arg)
#endif
#define __Pyx_NumKwargs_VARARGS(kwds) PyDict_Size(kwds)
#define __Pyx_KwValues_VARARGS(args, nargs) NULL
#define __Pyx_GetKwValue_VARARGS(kw, kwvalues, s) __Pyx_PyDict_GetItemStrWithError(kw, s)
#define __Pyx_KwargsAsDict_VARARGS(kw, kwvalues) PyDict_Copy(kw)
#if CYTHON_METH_FASTCALL
    #define __Pyx_Arg_FASTCALL(args, i) args[i]
    #define __Pyx_NumKwargs_FASTCALL(kwds) PyTuple_GET_SIZE(kwds)
    #define __Pyx_KwValues_FASTCALL(args, nargs) ((args) + (nargs))
    static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
    CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues);
  #else
    #define __Pyx_KwargsAsDict_FASTCALL(kw, kwvalues) _PyStack_AsDict(kwvalues, kw)
  #endif
    #define __Pyx_Arg_NewRef_FASTCALL(arg) arg  /* no-op, __Pyx_Arg_FASTCALL is direct and this needs
                                                   to have the same reference counting */
    #define __Pyx_Arg_XDECREF_FASTCALL(arg)
#else
    #define __Pyx_Arg_FASTCALL __Pyx_Arg_VARARGS
    #define __Pyx_NumKwargs_FASTCALL __Pyx_NumKwargs_VARARGS
    #define __Pyx_KwValues_FASTCALL __Pyx_KwValues_VARARGS
    #define __Pyx_GetKwValue_FASTCALL __Pyx_GetKwValue_VARARGS
    #define __Pyx_KwargsAsDict_FASTCALL __Pyx_KwargsAsDict_VARARGS
    #define __Pyx_Arg_NewRef_FASTCALL(arg) __Pyx_Arg_NewRef_VARARGS(arg)
    #define __Pyx_Arg_XDECREF_FASTCALL(arg) __Pyx_Arg_XDECREF_VARARGS(arg)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) __Pyx_PyTuple_FromArray(&__Pyx_Arg_VARARGS(args, start), stop - start)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) __Pyx_PyTuple_FromArray(&__Pyx_Arg_FASTCALL(args, start), stop - start)
#else
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) PyTuple_GetSlice(args, start, stop)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) PyTuple_GetSlice(args, start, stop)
#endif

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* KeywordStringCheck.proto */
static int __Pyx_CheckKeywordStrings(PyObject *kw, const char* function_name, int kw_allowed);

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject *const *kwvalues,
    PyObject **argnames[],
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,
    const char* function_name);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* RaiseUnexpectedTypeError.proto */
static int __Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely(__Pyx_IS_TYPE(obj, type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#if !CYTHON_VECTORCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if !CYTHON_VECTORCALL
#if PY_VERSION_HEX >= 0x03080000
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()
  #define __Pyx_PyFrame_GetLocalsplus(frame)  ((frame)->f_localsplus)
#else
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif
#endif
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectFastCall.proto */
#define __Pyx_PyObject_FastCall(func, args, nargs)  __Pyx_PyObject_FastCallDict(func, args, (size_t)(nargs), NULL)
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject **args, size_t nargs, PyObject *kwargs);

/* PyObjectCallNoArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* pep479.proto */
static void __Pyx_Generator_Replace_StopIteration(int in_async_gen);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* set_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_set_iterator(PyObject* iterable, int is_set,
                                                  Py_ssize_t* p_orig_length, int* p_source_is_set);
static CYTHON_INLINE int __Pyx_set_iter_next(
        PyObject* iter_obj, Py_ssize_t orig_length,
        Py_ssize_t* ppos, PyObject **value,
        int source_is_set);

/* tp_new.proto */
#define __Pyx_tp_new(type_obj, args) __Pyx_tp_new_kwargs(type_obj, args, NULL)
static CYTHON_INLINE PyObject* __Pyx_tp_new_kwargs(PyObject* type_obj, PyObject* args, PyObject* kwargs) {
    return (PyObject*) (((PyTypeObject*)type_obj)->tp_new((PyTypeObject*)type_obj, args, kwargs));
}

/* IncludeStructmemberH.proto */
#include <structmember.h>

/* FixUpExtensionType.proto */
#if CYTHON_USE_TYPE_SPECS
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type);
#endif

/* FetchSharedCythonModule.proto */
static PyObject *__Pyx_FetchSharedCythonABIModule(void);

/* FetchCommonType.proto */
#if !CYTHON_USE_TYPE_SPECS
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);
#else
static PyTypeObject* __Pyx_FetchCommonTypeFromSpec(PyObject *module, PyType_Spec *spec, PyObject *bases);
#endif

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObjectCall2Args.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* CoroutineBase.proto */
struct __pyx_CoroutineObject;
typedef PyObject *(*__pyx_coroutine_body_t)(struct __pyx_CoroutineObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct __pyx_CoroutineObject {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    PyObject *gi_frame;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Coroutine.proto */
#define __Pyx_Coroutine_USED
#define __Pyx_Coroutine_CheckExact(obj) __Pyx_IS_TYPE(obj, __pyx_CoroutineType)
#define __Pyx_Coroutine_Check(obj) __Pyx_Coroutine_CheckExact(obj)
#define __Pyx_CoroutineAwait_CheckExact(obj) __Pyx_IS_TYPE(obj, __pyx_CoroutineAwaitType)
#define __Pyx_Coroutine_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_CoroutineType, body, code, closure, name, qualname, module_name)
static int __pyx_Coroutine_init(PyObject *module);
static PyObject *__Pyx__Coroutine_await(PyObject *coroutine);
typedef struct {
    PyObject_HEAD
    PyObject *coroutine;
} __pyx_CoroutineAwaitObject;
static PyObject *__Pyx_CoroutineAwait_Close(__pyx_CoroutineAwaitObject *self, PyObject *arg);
static PyObject *__Pyx_CoroutineAwait_Throw(__pyx_CoroutineAwaitObject *self, PyObject *args);

/* GetAwaitIter.proto */
static CYTHON_INLINE PyObject *__Pyx_Coroutine_GetAwaitableIter(PyObject *o);
static PyObject *__Pyx__Coroutine_GetAwaitableIter(PyObject *o);

/* CoroutineYieldFrom.proto */
static CYTHON_INLINE PyObject* __Pyx_Coroutine_Yield_From(__pyx_CoroutineObject *gen, PyObject *source);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
        L->ob_item[len] = x;
        #else
        PyList_SET_ITEM(list, len, x);
        #endif
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* pyfrozenset_new.proto */
static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it);

/* py_set_discard.proto */
static CYTHON_INLINE int __Pyx_PySet_Discard(PyObject *set, PyObject *key);

/* ListExtend.proto */
static CYTHON_INLINE int __Pyx_PyList_Extend(PyObject* L, PyObject* v) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000
    PyObject* none = _PyList_Extend((PyListObject*)L, v);
    if (unlikely(!none))
        return -1;
    Py_DECREF(none);
    return 0;
#else
    return PyList_SetSlice(L, PY_SSIZE_T_MAX, PY_SSIZE_T_MAX, v);
#endif
}

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* CallNextTpTraverse.proto */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse);

/* CallNextTpClear.proto */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_clear);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* ValidateBasesTuple.proto */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases);
#endif

/* PyType_Ready.proto */
CYTHON_UNUSED static int __Pyx_PyType_Ready(PyTypeObject *t);

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetupReduce.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_setup_reduce(PyObject* type_obj);
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyTypeObject* typeptr , void* vtable);

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyTypeObject *type);

/* MergeVTables.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_MergeVtables(PyTypeObject *type);
#endif

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto_3_0_11
#define __PYX_HAVE_RT_ImportType_proto_3_0_11
#if defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L
#include <stdalign.h>
#endif
#if (defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L) || __cplusplus >= 201103L
#define __PYX_GET_STRUCT_ALIGNMENT_3_0_11(s) alignof(s)
#else
#define __PYX_GET_STRUCT_ALIGNMENT_3_0_11(s) sizeof(void*)
#endif
enum __Pyx_ImportType_CheckSize_3_0_11 {
   __Pyx_ImportType_CheckSize_Error_3_0_11 = 0,
   __Pyx_ImportType_CheckSize_Warn_3_0_11 = 1,
   __Pyx_ImportType_CheckSize_Ignore_3_0_11 = 2
};
static PyTypeObject *__Pyx_ImportType_3_0_11(PyObject* module, const char *module_name, const char *class_name, size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_0_11 check_size);
#endif

/* ImportDottedModule.proto */
static PyObject *__Pyx_ImportDottedModule(PyObject *name, PyObject *parts_tuple);
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_ImportDottedModule_WalkParts(PyObject *module, PyObject *name, PyObject *parts_tuple);
#endif

/* PatchInspect.proto */
static PyObject* __Pyx_patch_inspect(PyObject* module);

/* PatchAsyncIO.proto */
static PyObject* __Pyx_patch_asyncio(PyObject* module);

/* PyMethodNew.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    PyObject *typesModule=NULL, *methodType=NULL, *result=NULL;
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    typesModule = PyImport_ImportModule("types");
    if (!typesModule) return NULL;
    methodType = PyObject_GetAttrString(typesModule, "MethodType");
    Py_DECREF(typesModule);
    if (!methodType) return NULL;
    result = PyObject_CallFunctionObjArgs(methodType, func, self, NULL);
    Py_DECREF(methodType);
    return result;
}
#elif PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    return PyMethod_New(func, self);
}
#else
    #define __Pyx_PyMethod_New PyMethod_New
#endif

/* PyVectorcallFastCallDict.proto */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw);
#endif

/* CythonFunctionShared.proto */
#define __Pyx_CyFunction_USED
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CYFUNCTION_COROUTINE     0x08
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_CyFunction_GetClassObj(f)\
      (((__pyx_CyFunctionObject *) (f))->func_classobj)
#else
  #define __Pyx_CyFunction_GetClassObj(f)\
      ((PyObject*) ((PyCMethodObject *) (f))->mm_class)
#endif
#define __Pyx_CyFunction_SetClassObj(f, classobj)\
    __Pyx__CyFunction_SetClassObj((__pyx_CyFunctionObject *) (f), (classobj))
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject_HEAD
    PyObject *func;
#elif PY_VERSION_HEX < 0x030900B1
    PyCFunctionObject func;
#else
    PyCMethodObject func;
#endif
#if CYTHON_BACKPORT_VECTORCALL
    __pyx_vectorcallfunc func_vectorcall;
#endif
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_classobj;
#endif
    void *defaults;
    int defaults_pyobjects;
    size_t defaults_size;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
    PyObject *func_is_coroutine;
} __pyx_CyFunctionObject;
#undef __Pyx_CyOrPyCFunction_Check
#define __Pyx_CyFunction_Check(obj)  __Pyx_TypeCheck(obj, __pyx_CyFunctionType)
#define __Pyx_CyOrPyCFunction_Check(obj)  __Pyx_TypeCheck2(obj, __pyx_CyFunctionType, &PyCFunction_Type)
#define __Pyx_CyFunction_CheckExact(obj)  __Pyx_IS_TYPE(obj, __pyx_CyFunctionType)
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc);
#undef __Pyx_IsSameCFunction
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCyOrCFunction(func, cfunc)
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(PyObject *module);
#if CYTHON_METH_FASTCALL
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#if CYTHON_BACKPORT_VECTORCALL
#define __Pyx_CyFunction_func_vectorcall(f) (((__pyx_CyFunctionObject*)f)->func_vectorcall)
#else
#define __Pyx_CyFunction_func_vectorcall(f) (((PyCFunctionObject*)f)->vectorcall)
#endif
#endif

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);
#endif

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* GCCDiagnostics.proto */
#if !defined(__INTEL_COMPILER) && defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* RealImag.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(__cplusplus) && CYTHON_CCOMPLEX\
        && (defined(_WIN32) || defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 5 || __GNUC__ == 4 && __GNUC_MINOR__ >= 4 )) || __cplusplus >= 201103)
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
    #define __Pyx_c_eq_double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_double(a, b) ((a)-(b))
    #define __Pyx_c_prod_double(a, b) ((a)*(b))
    #define __Pyx_c_quot_double(a, b) ((a)/(b))
    #define __Pyx_c_neg_double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_double(z) ((z)==(double)0)
    #define __Pyx_c_conj_double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (::std::abs(z))
        #define __Pyx_c_pow_double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_double(z) ((z)==0)
    #define __Pyx_c_conj_double(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (cabs(z))
        #define __Pyx_c_pow_double(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int64_t(int64_t value);

/* CIntFromPy.proto */
static CYTHON_INLINE int64_t __Pyx_PyInt_As_int64_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE uint64_t __Pyx_PyInt_As_uint64_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint64_t(uint64_t value);

/* FormatTypeName.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%U"
static __Pyx_TypeName __Pyx_PyType_GetName(PyTypeObject* tp);
#define __Pyx_DECREF_TypeName(obj) Py_XDECREF(obj)
#else
typedef const char *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%.200s"
#define __Pyx_PyType_GetName(tp) ((tp)->tp_name)
#define __Pyx_DECREF_TypeName(obj)
#endif

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) __Pyx_IsAnySubtype2(Py_TYPE(obj), (PyTypeObject *)type1, (PyTypeObject *)type2)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) (PyObject_TypeCheck(obj, (PyTypeObject *)type1) || PyObject_TypeCheck(obj, (PyTypeObject *)type2))
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyErr_ExceptionMatches2(err1, err2)  __Pyx_PyErr_GivenExceptionMatches2(__Pyx_PyErr_CurrentExceptionType(), err1, err2)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static unsigned long __Pyx_get_runtime_version(void);
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer);

/* VoidPtrImport.proto */
static int __Pyx_ImportVoidPtr_3_0_11(PyObject *module, const char *name, void **p, const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

/* #### Code section: module_declarations ### */
static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4real_real(PyComplexObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4imag_imag(PyComplexObject *__pyx_v_self); /* proto*/
static void __pyx_f_7flitter_6render_7physics_8Particle_step(struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_self, double __pyx_v_speed_of_light, double __pyx_v_clock, double __pyx_v_delta); /* proto*/
static void __pyx_f_7flitter_6render_7physics_7Barrier_apply_distance(struct __pyx_obj_7flitter_6render_7physics_Barrier *__pyx_v_self, struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_particle); /* proto*/
static void __pyx_f_7flitter_6render_7physics_7Barrier_apply_rebound(struct __pyx_obj_7flitter_6render_7physics_Barrier *__pyx_v_self, struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_particle, double __pyx_v_speed_of_light, double __pyx_v_clock, double __pyx_v_delta); /* proto*/
static void __pyx_f_7flitter_6render_7physics_20ParticleForceApplier_apply(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_ParticleForceApplier *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_particle, CYTHON_UNUSED double __pyx_v_delta); /* proto*/
static void __pyx_f_7flitter_6render_7physics_18MatrixForceApplier_apply(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_MatrixForceApplier *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_from_particle, CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_to_particle, CYTHON_UNUSED struct __pyx_obj_7flitter_5model_Vector *__pyx_v_direction, CYTHON_UNUSED double __pyx_v_distance); /* proto*/
static void __pyx_f_7flitter_6render_7physics_20DistanceForceApplier_apply(struct __pyx_obj_7flitter_6render_7physics_DistanceForceApplier *__pyx_v_self, struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_from_particle, struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_to_particle, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_direction, double __pyx_v_distance); /* proto*/
static void __pyx_f_7flitter_6render_7physics_16DragForceApplier_apply(struct __pyx_obj_7flitter_6render_7physics_DragForceApplier *__pyx_v_self, struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_particle, double __pyx_v_delta); /* proto*/
static void __pyx_f_7flitter_6render_7physics_20BuoyancyForceApplier_apply(struct __pyx_obj_7flitter_6render_7physics_BuoyancyForceApplier *__pyx_v_self, struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_particle, CYTHON_UNUSED double __pyx_v_delta); /* proto*/
static void __pyx_f_7flitter_6render_7physics_20ConstantForceApplier_apply(struct __pyx_obj_7flitter_6render_7physics_ConstantForceApplier *__pyx_v_self, struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_particle, CYTHON_UNUSED double __pyx_v_delta); /* proto*/
static void __pyx_f_7flitter_6render_7physics_18RandomForceApplier_apply(struct __pyx_obj_7flitter_6render_7physics_RandomForceApplier *__pyx_v_self, struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_particle, CYTHON_UNUSED double __pyx_v_delta); /* proto*/
static void __pyx_f_7flitter_6render_7physics_17FieldForceApplier_apply(struct __pyx_obj_7flitter_6render_7physics_FieldForceApplier *__pyx_v_self, struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_particle, CYTHON_UNUSED double __pyx_v_delta); /* proto*/
static void __pyx_f_7flitter_6render_7physics_21CollisionForceApplier_apply(struct __pyx_obj_7flitter_6render_7physics_CollisionForceApplier *__pyx_v_self, struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_from_particle, struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_to_particle, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_direction, double __pyx_v_distance); /* proto*/
static void __pyx_f_7flitter_6render_7physics_19GravityForceApplier_apply(struct __pyx_obj_7flitter_6render_7physics_GravityForceApplier *__pyx_v_self, struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_from_particle, struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_to_particle, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_direction, double __pyx_v_distance); /* proto*/
static void __pyx_f_7flitter_6render_7physics_25ElectrostaticForceApplier_apply(struct __pyx_obj_7flitter_6render_7physics_ElectrostaticForceApplier *__pyx_v_self, struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_from_particle, struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_to_particle, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_direction, double __pyx_v_distance); /* proto*/
static void __pyx_f_7flitter_6render_7physics_20AdhesionForceApplier_apply(struct __pyx_obj_7flitter_6render_7physics_AdhesionForceApplier *__pyx_v_self, struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_from_particle, struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_to_particle, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_direction, double __pyx_v_distance); /* proto*/
static void __pyx_f_7flitter_6render_7physics_13PhysicsSystem_collect(struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem *__pyx_v_self, PyObject *__pyx_v_groups, struct __pyx_obj_7flitter_6render_7physics_PhysicsGroup *__pyx_v_group, struct __pyx_obj_7flitter_5model_Node *__pyx_v_node, PyObject *__pyx_v_particles_by_id, PyObject *__pyx_v_old_state_keys, PyObject *__pyx_v_new_state_keys, struct __pyx_obj_7flitter_5model_StateDict *__pyx_v_state, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_state_prefix, double __pyx_v_clock, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_seed); /* proto*/
static double __pyx_f_7flitter_6render_7physics_13PhysicsSystem_calculate(struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem *__pyx_v_self, PyObject *__pyx_v_groups, PyObject *__pyx_v_particles_by_id, int __pyx_v_realtime, int __pyx_v_extra, double __pyx_v_time, double __pyx_v_last_time, double __pyx_v_clock, int __pyx_skip_dispatch); /* proto*/

/* Module declarations from "cython" */

/* Module declarations from "libc.stdint" */

/* Module declarations from "libc.stddef" */

/* Module declarations from "cpython.unicode" */

/* Module declarations from "flitter.model" */
static uint64_t *__pyx_vp_7flitter_5model_HASH_START = 0;
#define __pyx_v_7flitter_5model_HASH_START (*__pyx_vp_7flitter_5model_HASH_START)
static struct __pyx_obj_7flitter_5model_Vector **__pyx_vp_7flitter_5model_null_ = 0;
#define __pyx_v_7flitter_5model_null_ (*__pyx_vp_7flitter_5model_null_)
static struct __pyx_obj_7flitter_5model_Vector **__pyx_vp_7flitter_5model_true_ = 0;
#define __pyx_v_7flitter_5model_true_ (*__pyx_vp_7flitter_5model_true_)
static struct __pyx_obj_7flitter_5model_Vector **__pyx_vp_7flitter_5model_false_ = 0;
#define __pyx_v_7flitter_5model_false_ (*__pyx_vp_7flitter_5model_false_)
static struct __pyx_obj_7flitter_5model_Vector **__pyx_vp_7flitter_5model_minusone_ = 0;
#define __pyx_v_7flitter_5model_minusone_ (*__pyx_vp_7flitter_5model_minusone_)
static struct __pyx_obj_7flitter_5model_Vector **__pyx_vp_7flitter_5model_inf_ = 0;
#define __pyx_v_7flitter_5model_inf_ (*__pyx_vp_7flitter_5model_inf_)
static struct __pyx_obj_7flitter_5model_Vector **__pyx_vp_7flitter_5model_nan_ = 0;
#define __pyx_v_7flitter_5model_nan_ (*__pyx_vp_7flitter_5model_nan_)

/* Module declarations from "flitter.language.functions" */

/* Module declarations from "libc.math" */

/* Module declarations from "cpython.version" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.type" */

/* Module declarations from "libc.string" */

/* Module declarations from "libc.stdio" */

/* Module declarations from "cpython.object" */

/* Module declarations from "cpython.ref" */

/* Module declarations from "cpython.exc" */

/* Module declarations from "cpython.module" */

/* Module declarations from "cpython.mem" */

/* Module declarations from "cpython.tuple" */

/* Module declarations from "cpython.list" */

/* Module declarations from "cpython.sequence" */

/* Module declarations from "cpython.mapping" */

/* Module declarations from "cpython.iterator" */

/* Module declarations from "cpython.number" */

/* Module declarations from "cpython.int" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.bool" */

/* Module declarations from "cpython.long" */

/* Module declarations from "cpython.float" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.complex" */

/* Module declarations from "cpython.string" */

/* Module declarations from "cpython.pyport" */

/* Module declarations from "cpython.dict" */

/* Module declarations from "cpython.instance" */

/* Module declarations from "cpython.function" */

/* Module declarations from "cpython.method" */

/* Module declarations from "cpython.weakref" */

/* Module declarations from "cpython.getargs" */

/* Module declarations from "cpython.pythread" */

/* Module declarations from "cpython.pystate" */

/* Module declarations from "cpython.cobject" */

/* Module declarations from "cpython.oldbuffer" */

/* Module declarations from "cpython.set" */

/* Module declarations from "cpython.buffer" */

/* Module declarations from "cpython.bytes" */

/* Module declarations from "cpython.pycapsule" */

/* Module declarations from "cpython.contextvars" */

/* Module declarations from "cpython" */

/* Module declarations from "flitter.render.physics" */
static struct __pyx_obj_7flitter_5model_Vector *__pyx_v_7flitter_6render_7physics_VELOCITY = 0;
static struct __pyx_obj_7flitter_5model_Vector *__pyx_v_7flitter_6render_7physics_LAST = 0;
static struct __pyx_obj_7flitter_5model_Vector *__pyx_v_7flitter_6render_7physics_CLOCK = 0;
static struct __pyx_obj_7flitter_5model_Vector *__pyx_v_7flitter_6render_7physics_RUN = 0;
static struct __pyx_obj_7flitter_5model_Vector *__pyx_v_7flitter_6render_7physics_ITERATION = 0;
static PyObject *__pyx_f_7flitter_6render_7physics___pyx_unpickle_Particle__set_state(struct __pyx_obj_7flitter_6render_7physics_Particle *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_6render_7physics___pyx_unpickle_DistanceForceApplier__set_state(struct __pyx_obj_7flitter_6render_7physics_DistanceForceApplier *, PyObject *); /*proto*/
static PyObject *__pyx_f_7flitter_6render_7physics___pyx_unpickle_PhysicsSystem__set_state(struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem *, PyObject *); /*proto*/
/* #### Code section: typeinfo ### */
/* #### Code section: before_global_var ### */
#define __Pyx_MODULE_NAME "flitter.render.physics"
extern int __pyx_module_is_main_flitter__render__physics;
int __pyx_module_is_main_flitter__render__physics = 0;

/* Implementation of "flitter.render.physics" */
/* #### Code section: global_var ### */
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_NotImplementedError;
static PyObject *__pyx_builtin_enumerate;
/* #### Code section: string_decls ### */
static const char __pyx_k_i[] = "i";
static const char __pyx_k__5[] = ".";
static const char __pyx_k__8[] = "*";
static const char __pyx_k__9[] = "";
static const char __pyx_k_gc[] = "gc";
static const char __pyx_k_id[] = "id";
static const char __pyx_k_to[] = "to";
static const char __pyx_k__52[] = "?";
static const char __pyx_k_max[] = "max";
static const char __pyx_k_min[] = "min";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_run[] = "run";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_drag[] = "drag";
static const char __pyx_k_ease[] = "ease";
static const char __pyx_k_flow[] = "flow";
static const char __pyx_k_from[] = "from";
static const char __pyx_k_last[] = "last";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_mass[] = "mass";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_node[] = "node";
static const char __pyx_k_seed[] = "seed";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_spec[] = "__spec__";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_time[] = "time";
static const char __pyx_k_zero[] = "zero";
static const char __pyx_k_await[] = "__await__";
static const char __pyx_k_clock[] = "clock";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_debug[] = "debug";
static const char __pyx_k_extra[] = "extra";
static const char __pyx_k_field[] = "field";
static const char __pyx_k_fixed[] = "fixed";
static const char __pyx_k_force[] = "force";
static const char __pyx_k_group[] = "group";
static const char __pyx_k_power[] = "power";
static const char __pyx_k_purge[] = "purge";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_state[] = "state";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_trace[] = "trace";
static const char __pyx_k_anchor[] = "anchor";
static const char __pyx_k_charge[] = "charge";
static const char __pyx_k_dict_2[] = "_dict";
static const char __pyx_k_enable[] = "enable";
static const char __pyx_k_engine[] = "engine";
static const char __pyx_k_groups[] = "groups";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_kwargs[] = "kwargs";
static const char __pyx_k_logger[] = "logger";
static const char __pyx_k_loguru[] = "loguru";
static const char __pyx_k_normal[] = "normal";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_radius[] = "radius";
static const char __pyx_k_random[] = "random";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_Barrier[] = "Barrier";
static const char __pyx_k_asyncio[] = "asyncio";
static const char __pyx_k_barrier[] = "barrier";
static const char __pyx_k_density[] = "density";
static const char __pyx_k_destroy[] = "destroy";
static const char __pyx_k_disable[] = "disable";
static const char __pyx_k_gravity[] = "gravity";
static const char __pyx_k_inspect[] = "inspect";
static const char __pyx_k_maximum[] = "maximum";
static const char __pyx_k_minimum[] = "minimum";
static const char __pyx_k_overlap[] = "overlap";
static const char __pyx_k_Particle[] = "Particle";
static const char __pyx_k_adhesion[] = "adhesion";
static const char __pyx_k_buoyancy[] = "buoyancy";
static const char __pyx_k_constant[] = "constant";
static const char __pyx_k_distance[] = "distance";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_last_run[] = "last_run";
static const char __pyx_k_particle[] = "particle";
static const char __pyx_k_position[] = "position";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_realtime[] = "realtime";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_strength[] = "strength";
static const char __pyx_k_velocity[] = "velocity";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_calculate[] = "calculate";
static const char __pyx_k_collision[] = "collision";
static const char __pyx_k_direction[] = "direction";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_isenabled[] = "isenabled";
static const char __pyx_k_iteration[] = "iteration";
static const char __pyx_k_last_time[] = "last_time";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_state_key[] = "state_key";
static const char __pyx_k_to_thread[] = "to_thread";
static const char __pyx_k_top_group[] = "top_group";
static const char __pyx_k_dimensions[] = "dimensions";
static const char __pyx_k_name_patch[] = "name_patch";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_resolution[] = "resolution";
static const char __pyx_k_run_vector[] = "run_vector";
static const char __pyx_k_target_fps[] = "target_fps";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_performance[] = "performance";
static const char __pyx_k_restitution[] = "restitution";
static const char __pyx_k_time_vector[] = "time_vector";
static const char __pyx_k_PhysicsGroup[] = "PhysicsGroup";
static const char __pyx_k_acceleration[] = "acceleration";
static const char __pyx_k_initializing[] = "_initializing";
static const char __pyx_k_is_coroutine[] = "_is_coroutine";
static const char __pyx_k_max_distance[] = "max_distance";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_state_prefix[] = "state_prefix";
static const char __pyx_k_stringsource[] = "<stringsource>";
static const char __pyx_k_use_setstate[] = "use_setstate";
static const char __pyx_k_PhysicsSystem[] = "PhysicsSystem";
static const char __pyx_k_asyncio_tasks[] = "asyncio.tasks";
static const char __pyx_k_electrostatic[] = "electrostatic";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_new_state_keys[] = "new_state_keys";
static const char __pyx_k_old_state_keys[] = "old_state_keys";
static const char __pyx_k_speed_of_light[] = "speed_of_light";
static const char __pyx_k_last_run_vector[] = "last_run_vector";
static const char __pyx_k_particles_by_id[] = "particles_by_id";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_DragForceApplier[] = "DragForceApplier";
static const char __pyx_k_iteration_vector[] = "iteration_vector";
static const char __pyx_k_FieldForceApplier[] = "FieldForceApplier";
static const char __pyx_k_MatrixForceApplier[] = "MatrixForceApplier";
static const char __pyx_k_RandomForceApplier[] = "RandomForceApplier";
static const char __pyx_k_asyncio_coroutines[] = "asyncio.coroutines";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_GravityForceApplier[] = "GravityForceApplier";
static const char __pyx_k_NotImplementedError[] = "NotImplementedError";
static const char __pyx_k_PhysicsSystem_purge[] = "PhysicsSystem.purge";
static const char __pyx_k_AdhesionForceApplier[] = "AdhesionForceApplier";
static const char __pyx_k_BuoyancyForceApplier[] = "BuoyancyForceApplier";
static const char __pyx_k_ConstantForceApplier[] = "ConstantForceApplier";
static const char __pyx_k_DistanceForceApplier[] = "DistanceForceApplier";
static const char __pyx_k_ParticleForceApplier[] = "ParticleForceApplier";
static const char __pyx_k_PhysicsSystem_update[] = "PhysicsSystem.update";
static const char __pyx_k_CollisionForceApplier[] = "CollisionForceApplier";
static const char __pyx_k_PhysicsSystem_destroy[] = "PhysicsSystem.destroy";
static const char __pyx_k_pyx_unpickle_Particle[] = "__pyx_unpickle_Particle";
static const char __pyx_k_Flitter_physics_engine[] = "\nFlitter physics engine\n";
static const char __pyx_k_flitter_render_physics[] = "flitter.render.physics";
static const char __pyx_k_Barrier___reduce_cython[] = "Barrier.__reduce_cython__";
static const char __pyx_k_PhysicsSystem_calculate[] = "PhysicsSystem.calculate";
static const char __pyx_k_Particle___reduce_cython[] = "Particle.__reduce_cython__";
static const char __pyx_k_Barrier___setstate_cython[] = "Barrier.__setstate_cython__";
static const char __pyx_k_ElectrostaticForceApplier[] = "ElectrostaticForceApplier";
static const char __pyx_k_Particle___setstate_cython[] = "Particle.__setstate_cython__";
static const char __pyx_k_pyx_unpickle_PhysicsSystem[] = "__pyx_unpickle_PhysicsSystem";
static const char __pyx_k_PhysicsGroup___reduce_cython[] = "PhysicsGroup.__reduce_cython__";
static const char __pyx_k_PhysicsSystem___reduce_cython[] = "PhysicsSystem.__reduce_cython__";
static const char __pyx_k_PhysicsGroup___setstate_cython[] = "PhysicsGroup.__setstate_cython__";
static const char __pyx_k_Reset_physics_system_r_for_run[] = "Reset physics system {!r} for run {}";
static const char __pyx_k_pyx_unpickle_DistanceForceAppl[] = "__pyx_unpickle_DistanceForceApplier";
static const char __pyx_k_src_flitter_render_physics_pyx[] = "src/flitter/render/physics.pyx";
static const char __pyx_k_AdhesionForceApplier___setstate[] = "AdhesionForceApplier.__setstate_cython__";
static const char __pyx_k_BuoyancyForceApplier___setstate[] = "BuoyancyForceApplier.__setstate_cython__";
static const char __pyx_k_ConstantForceApplier___setstate[] = "ConstantForceApplier.__setstate_cython__";
static const char __pyx_k_DistanceForceApplier___setstate[] = "DistanceForceApplier.__setstate_cython__";
static const char __pyx_k_Group_with_particles_forces_and[] = "Group {} with {} particles, {} forces and {} barriers";
static const char __pyx_k_ParticleForceApplier___setstate[] = "ParticleForceApplier.__setstate_cython__";
static const char __pyx_k_PhysicsSystem___setstate_cython[] = "PhysicsSystem.__setstate_cython__";
static const char __pyx_k_AdhesionForceApplier___reduce_cy[] = "AdhesionForceApplier.__reduce_cython__";
static const char __pyx_k_BuoyancyForceApplier___reduce_cy[] = "BuoyancyForceApplier.__reduce_cython__";
static const char __pyx_k_CollisionForceApplier___reduce_c[] = "CollisionForceApplier.__reduce_cython__";
static const char __pyx_k_CollisionForceApplier___setstate[] = "CollisionForceApplier.__setstate_cython__";
static const char __pyx_k_ConstantForceApplier___reduce_cy[] = "ConstantForceApplier.__reduce_cython__";
static const char __pyx_k_DistanceForceApplier___reduce_cy[] = "DistanceForceApplier.__reduce_cython__";
static const char __pyx_k_DragForceApplier___reduce_cython[] = "DragForceApplier.__reduce_cython__";
static const char __pyx_k_DragForceApplier___setstate_cyth[] = "DragForceApplier.__setstate_cython__";
static const char __pyx_k_ElectrostaticForceApplier___redu[] = "ElectrostaticForceApplier.__reduce_cython__";
static const char __pyx_k_ElectrostaticForceApplier___sets[] = "ElectrostaticForceApplier.__setstate_cython__";
static const char __pyx_k_FieldForceApplier___reduce_cytho[] = "FieldForceApplier.__reduce_cython__";
static const char __pyx_k_FieldForceApplier___setstate_cyt[] = "FieldForceApplier.__setstate_cython__";
static const char __pyx_k_GravityForceApplier___reduce_cyt[] = "GravityForceApplier.__reduce_cython__";
static const char __pyx_k_GravityForceApplier___setstate_c[] = "GravityForceApplier.__setstate_cython__";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0[] = "Incompatible checksums (0x%x vs (0xc9134b9, 0x81fb8cd, 0xb66f251) = (acceleration, charge, ease, force, id, initial_force, mass, non_anchor, position, position_state_key, radius, velocity, velocity_state_key))";
static const char __pyx_k_MatrixForceApplier___reduce_cyth[] = "MatrixForceApplier.__reduce_cython__";
static const char __pyx_k_MatrixForceApplier___setstate_cy[] = "MatrixForceApplier.__setstate_cython__";
static const char __pyx_k_New_D_physics_r_with_particles_f[] = "New {}D physics {!r} with {} particles, {} forces and {} barriers";
static const char __pyx_k_New_D_physics_r_with_particles_o[] = "New {}D physics {!r} with {} particles over {} groups";
static const char __pyx_k_ParticleForceApplier___reduce_cy[] = "ParticleForceApplier.__reduce_cython__";
static const char __pyx_k_RandomForceApplier___reduce_cyth[] = "RandomForceApplier.__reduce_cython__";
static const char __pyx_k_RandomForceApplier___setstate_cy[] = "RandomForceApplier.__setstate_cython__";
static const char __pyx_k_no_default___reduce___due_to_non[] = "no default __reduce__ due to non-trivial __cinit__";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_2[] = "Incompatible checksums (0x%x vs (0xca186ab, 0x9db94ff, 0x9717dcb) = (from_particle_id, maximum, minimum, power, strength, to_particle_id))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_3[] = "Incompatible checksums (0x%x vs (0xf14ade3, 0x885a5ab, 0x3d7be7f) = (dimensions, resolution, speed_of_light, state_keys))";
/* #### Code section: decls ### */
static int __pyx_pf_7flitter_6render_7physics_12PhysicsGroup___cinit__(struct __pyx_obj_7flitter_6render_7physics_PhysicsGroup *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_6render_7physics_12PhysicsGroup_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_PhysicsGroup *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_6render_7physics_12PhysicsGroup_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_PhysicsGroup *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_6render_7physics_8Particle___reduce_cython__(struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_6render_7physics_8Particle_2__setstate_cython__(struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7flitter_6render_7physics_7Barrier___cinit__(struct __pyx_obj_7flitter_6render_7physics_Barrier *__pyx_v_self, struct __pyx_obj_7flitter_5model_Node *__pyx_v_node, double __pyx_v_strength, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_zero); /* proto */
static PyObject *__pyx_pf_7flitter_6render_7physics_7Barrier_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_Barrier *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_6render_7physics_7Barrier_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_Barrier *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7flitter_6render_7physics_20ParticleForceApplier___cinit__(struct __pyx_obj_7flitter_6render_7physics_ParticleForceApplier *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_7flitter_5model_Node *__pyx_v_node, double __pyx_v_strength, CYTHON_UNUSED struct __pyx_obj_7flitter_5model_Vector *__pyx_v_zero); /* proto */
static PyObject *__pyx_pf_7flitter_6render_7physics_20ParticleForceApplier_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_ParticleForceApplier *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_6render_7physics_20ParticleForceApplier_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_ParticleForceApplier *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7flitter_6render_7physics_18MatrixForceApplier___cinit__(struct __pyx_obj_7flitter_6render_7physics_MatrixForceApplier *__pyx_v_self, struct __pyx_obj_7flitter_5model_Node *__pyx_v_node, double __pyx_v_strength, CYTHON_UNUSED struct __pyx_obj_7flitter_5model_Vector *__pyx_v_zero); /* proto */
static PyObject *__pyx_pf_7flitter_6render_7physics_18MatrixForceApplier_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_MatrixForceApplier *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_6render_7physics_18MatrixForceApplier_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_MatrixForceApplier *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_6render_7physics_20DistanceForceApplier___reduce_cython__(struct __pyx_obj_7flitter_6render_7physics_DistanceForceApplier *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_6render_7physics_20DistanceForceApplier_2__setstate_cython__(struct __pyx_obj_7flitter_6render_7physics_DistanceForceApplier *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7flitter_6render_7physics_16DragForceApplier___cinit__(struct __pyx_obj_7flitter_6render_7physics_DragForceApplier *__pyx_v_self, struct __pyx_obj_7flitter_5model_Node *__pyx_v_node, CYTHON_UNUSED double __pyx_v_strength, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_zero); /* proto */
static PyObject *__pyx_pf_7flitter_6render_7physics_16DragForceApplier_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_DragForceApplier *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_6render_7physics_16DragForceApplier_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_DragForceApplier *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7flitter_6render_7physics_20BuoyancyForceApplier___cinit__(struct __pyx_obj_7flitter_6render_7physics_BuoyancyForceApplier *__pyx_v_self, struct __pyx_obj_7flitter_5model_Node *__pyx_v_node, CYTHON_UNUSED double __pyx_v_strength, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_zero); /* proto */
static PyObject *__pyx_pf_7flitter_6render_7physics_20BuoyancyForceApplier_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_BuoyancyForceApplier *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_6render_7physics_20BuoyancyForceApplier_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_BuoyancyForceApplier *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7flitter_6render_7physics_20ConstantForceApplier___cinit__(struct __pyx_obj_7flitter_6render_7physics_ConstantForceApplier *__pyx_v_self, struct __pyx_obj_7flitter_5model_Node *__pyx_v_node, CYTHON_UNUSED double __pyx_v_strength, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_zero); /* proto */
static PyObject *__pyx_pf_7flitter_6render_7physics_20ConstantForceApplier_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_ConstantForceApplier *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_6render_7physics_20ConstantForceApplier_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_ConstantForceApplier *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7flitter_6render_7physics_18RandomForceApplier___cinit__(struct __pyx_obj_7flitter_6render_7physics_RandomForceApplier *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_7flitter_5model_Node *__pyx_v_node, CYTHON_UNUSED double __pyx_v_strength, CYTHON_UNUSED struct __pyx_obj_7flitter_5model_Vector *__pyx_v_zero); /* proto */
static PyObject *__pyx_pf_7flitter_6render_7physics_18RandomForceApplier_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_RandomForceApplier *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_6render_7physics_18RandomForceApplier_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_RandomForceApplier *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7flitter_6render_7physics_17FieldForceApplier___cinit__(struct __pyx_obj_7flitter_6render_7physics_FieldForceApplier *__pyx_v_self, struct __pyx_obj_7flitter_5model_Node *__pyx_v_node, CYTHON_UNUSED double __pyx_v_strength, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_zero); /* proto */
static PyObject *__pyx_pf_7flitter_6render_7physics_17FieldForceApplier_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_FieldForceApplier *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_6render_7physics_17FieldForceApplier_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_FieldForceApplier *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7flitter_6render_7physics_21CollisionForceApplier___cinit__(struct __pyx_obj_7flitter_6render_7physics_CollisionForceApplier *__pyx_v_self, struct __pyx_obj_7flitter_5model_Node *__pyx_v_node, CYTHON_UNUSED double __pyx_v_strength, CYTHON_UNUSED struct __pyx_obj_7flitter_5model_Vector *__pyx_v_zero); /* proto */
static PyObject *__pyx_pf_7flitter_6render_7physics_21CollisionForceApplier_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_CollisionForceApplier *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_6render_7physics_21CollisionForceApplier_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_CollisionForceApplier *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_6render_7physics_19GravityForceApplier___reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_GravityForceApplier *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_6render_7physics_19GravityForceApplier_2__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_GravityForceApplier *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_6render_7physics_25ElectrostaticForceApplier___reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_ElectrostaticForceApplier *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_6render_7physics_25ElectrostaticForceApplier_2__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_ElectrostaticForceApplier *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7flitter_6render_7physics_20AdhesionForceApplier___cinit__(struct __pyx_obj_7flitter_6render_7physics_AdhesionForceApplier *__pyx_v_self, struct __pyx_obj_7flitter_5model_Node *__pyx_v_node, CYTHON_UNUSED double __pyx_v_strength, CYTHON_UNUSED struct __pyx_obj_7flitter_5model_Vector *__pyx_v_zero); /* proto */
static PyObject *__pyx_pf_7flitter_6render_7physics_20AdhesionForceApplier_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_AdhesionForceApplier *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_6render_7physics_20AdhesionForceApplier_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_AdhesionForceApplier *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7flitter_6render_7physics_13PhysicsSystem___init__(struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_7flitter_6render_7physics_13PhysicsSystem_2destroy(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_6render_7physics_13PhysicsSystem_5purge(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_6render_7physics_13PhysicsSystem_8update(struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem *__pyx_v_self, PyObject *__pyx_v_engine, struct __pyx_obj_7flitter_5model_Node *__pyx_v_node, double __pyx_v_time, double __pyx_v_performance, CYTHON_UNUSED PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_7flitter_6render_7physics_13PhysicsSystem_11calculate(struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem *__pyx_v_self, PyObject *__pyx_v_groups, PyObject *__pyx_v_particles_by_id, int __pyx_v_realtime, int __pyx_v_extra, double __pyx_v_time, double __pyx_v_last_time, double __pyx_v_clock); /* proto */
static PyObject *__pyx_pf_7flitter_6render_7physics_13PhysicsSystem_13__reduce_cython__(struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7flitter_6render_7physics_13PhysicsSystem_15__setstate_cython__(struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_6render_7physics___pyx_unpickle_Particle(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_6render_7physics_2__pyx_unpickle_DistanceForceApplier(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7flitter_6render_7physics_4__pyx_unpickle_PhysicsSystem(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_7flitter_6render_7physics_PhysicsGroup(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_6render_7physics_Particle(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_6render_7physics_Barrier(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_6render_7physics_ParticleForceApplier(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_6render_7physics_MatrixForceApplier(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_6render_7physics_DistanceForceApplier(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_6render_7physics_DragForceApplier(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_6render_7physics_BuoyancyForceApplier(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_6render_7physics_ConstantForceApplier(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_6render_7physics_RandomForceApplier(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_6render_7physics_FieldForceApplier(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_6render_7physics_CollisionForceApplier(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_6render_7physics_GravityForceApplier(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_6render_7physics_ElectrostaticForceApplier(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_6render_7physics_AdhesionForceApplier(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_6render_7physics_PhysicsSystem(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_6render_7physics___pyx_scope_struct__destroy(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_6render_7physics___pyx_scope_struct_1_purge(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7flitter_6render_7physics___pyx_scope_struct_2_update(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
/* #### Code section: late_includes ### */
/* #### Code section: module_state ### */
typedef struct {
  PyObject *__pyx_d;
  PyObject *__pyx_b;
  PyObject *__pyx_cython_runtime;
  PyObject *__pyx_empty_tuple;
  PyObject *__pyx_empty_bytes;
  PyObject *__pyx_empty_unicode;
  #ifdef __Pyx_CyFunction_USED
  PyTypeObject *__pyx_CyFunctionType;
  #endif
  #ifdef __Pyx_FusedFunction_USED
  PyTypeObject *__pyx_FusedFunctionType;
  #endif
  #ifdef __Pyx_Generator_USED
  PyTypeObject *__pyx_GeneratorType;
  #endif
  #ifdef __Pyx_IterableCoroutine_USED
  PyTypeObject *__pyx_IterableCoroutineType;
  #endif
  #ifdef __Pyx_Coroutine_USED
  PyTypeObject *__pyx_CoroutineAwaitType;
  #endif
  #ifdef __Pyx_Coroutine_USED
  PyTypeObject *__pyx_CoroutineType;
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7flitter_5model_Vector;
  PyTypeObject *__pyx_ptype_7flitter_5model_Matrix33;
  PyTypeObject *__pyx_ptype_7flitter_5model_Matrix44;
  PyTypeObject *__pyx_ptype_7flitter_5model_Quaternion;
  PyTypeObject *__pyx_ptype_7flitter_5model_Node;
  PyTypeObject *__pyx_ptype_7flitter_5model_StateDict;
  PyTypeObject *__pyx_ptype_7flitter_5model_Context;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7flitter_8language_9functions_uniform;
  PyTypeObject *__pyx_ptype_7flitter_8language_9functions_beta;
  PyTypeObject *__pyx_ptype_7flitter_8language_9functions_normal;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7cpython_4type_type;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7cpython_4bool_bool;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7cpython_7complex_complex;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  PyObject *__pyx_type_7flitter_6render_7physics_PhysicsGroup;
  PyObject *__pyx_type_7flitter_6render_7physics_Particle;
  PyObject *__pyx_type_7flitter_6render_7physics_Barrier;
  PyObject *__pyx_type_7flitter_6render_7physics_ParticleForceApplier;
  PyObject *__pyx_type_7flitter_6render_7physics_MatrixForceApplier;
  PyObject *__pyx_type_7flitter_6render_7physics_DistanceForceApplier;
  PyObject *__pyx_type_7flitter_6render_7physics_DragForceApplier;
  PyObject *__pyx_type_7flitter_6render_7physics_BuoyancyForceApplier;
  PyObject *__pyx_type_7flitter_6render_7physics_ConstantForceApplier;
  PyObject *__pyx_type_7flitter_6render_7physics_RandomForceApplier;
  PyObject *__pyx_type_7flitter_6render_7physics_FieldForceApplier;
  PyObject *__pyx_type_7flitter_6render_7physics_CollisionForceApplier;
  PyObject *__pyx_type_7flitter_6render_7physics_GravityForceApplier;
  PyObject *__pyx_type_7flitter_6render_7physics_ElectrostaticForceApplier;
  PyObject *__pyx_type_7flitter_6render_7physics_AdhesionForceApplier;
  PyObject *__pyx_type_7flitter_6render_7physics_PhysicsSystem;
  PyObject *__pyx_type_7flitter_6render_7physics___pyx_scope_struct__destroy;
  PyObject *__pyx_type_7flitter_6render_7physics___pyx_scope_struct_1_purge;
  PyObject *__pyx_type_7flitter_6render_7physics___pyx_scope_struct_2_update;
  #endif
  PyTypeObject *__pyx_ptype_7flitter_6render_7physics_PhysicsGroup;
  PyTypeObject *__pyx_ptype_7flitter_6render_7physics_Particle;
  PyTypeObject *__pyx_ptype_7flitter_6render_7physics_Barrier;
  PyTypeObject *__pyx_ptype_7flitter_6render_7physics_ParticleForceApplier;
  PyTypeObject *__pyx_ptype_7flitter_6render_7physics_MatrixForceApplier;
  PyTypeObject *__pyx_ptype_7flitter_6render_7physics_DistanceForceApplier;
  PyTypeObject *__pyx_ptype_7flitter_6render_7physics_DragForceApplier;
  PyTypeObject *__pyx_ptype_7flitter_6render_7physics_BuoyancyForceApplier;
  PyTypeObject *__pyx_ptype_7flitter_6render_7physics_ConstantForceApplier;
  PyTypeObject *__pyx_ptype_7flitter_6render_7physics_RandomForceApplier;
  PyTypeObject *__pyx_ptype_7flitter_6render_7physics_FieldForceApplier;
  PyTypeObject *__pyx_ptype_7flitter_6render_7physics_CollisionForceApplier;
  PyTypeObject *__pyx_ptype_7flitter_6render_7physics_GravityForceApplier;
  PyTypeObject *__pyx_ptype_7flitter_6render_7physics_ElectrostaticForceApplier;
  PyTypeObject *__pyx_ptype_7flitter_6render_7physics_AdhesionForceApplier;
  PyTypeObject *__pyx_ptype_7flitter_6render_7physics_PhysicsSystem;
  PyTypeObject *__pyx_ptype_7flitter_6render_7physics___pyx_scope_struct__destroy;
  PyTypeObject *__pyx_ptype_7flitter_6render_7physics___pyx_scope_struct_1_purge;
  PyTypeObject *__pyx_ptype_7flitter_6render_7physics___pyx_scope_struct_2_update;
  PyObject *__pyx_n_s_AdhesionForceApplier;
  PyObject *__pyx_n_s_AdhesionForceApplier___reduce_cy;
  PyObject *__pyx_n_s_AdhesionForceApplier___setstate;
  PyObject *__pyx_n_s_Barrier;
  PyObject *__pyx_n_s_Barrier___reduce_cython;
  PyObject *__pyx_n_s_Barrier___setstate_cython;
  PyObject *__pyx_n_s_BuoyancyForceApplier;
  PyObject *__pyx_n_s_BuoyancyForceApplier___reduce_cy;
  PyObject *__pyx_n_s_BuoyancyForceApplier___setstate;
  PyObject *__pyx_n_s_CollisionForceApplier;
  PyObject *__pyx_n_s_CollisionForceApplier___reduce_c;
  PyObject *__pyx_n_s_CollisionForceApplier___setstate;
  PyObject *__pyx_n_s_ConstantForceApplier;
  PyObject *__pyx_n_s_ConstantForceApplier___reduce_cy;
  PyObject *__pyx_n_s_ConstantForceApplier___setstate;
  PyObject *__pyx_n_s_DistanceForceApplier;
  PyObject *__pyx_n_s_DistanceForceApplier___reduce_cy;
  PyObject *__pyx_n_s_DistanceForceApplier___setstate;
  PyObject *__pyx_n_s_DragForceApplier;
  PyObject *__pyx_n_s_DragForceApplier___reduce_cython;
  PyObject *__pyx_n_s_DragForceApplier___setstate_cyth;
  PyObject *__pyx_n_s_ElectrostaticForceApplier;
  PyObject *__pyx_n_s_ElectrostaticForceApplier___redu;
  PyObject *__pyx_n_s_ElectrostaticForceApplier___sets;
  PyObject *__pyx_n_s_FieldForceApplier;
  PyObject *__pyx_n_s_FieldForceApplier___reduce_cytho;
  PyObject *__pyx_n_s_FieldForceApplier___setstate_cyt;
  PyObject *__pyx_n_s_GravityForceApplier;
  PyObject *__pyx_n_s_GravityForceApplier___reduce_cyt;
  PyObject *__pyx_n_s_GravityForceApplier___setstate_c;
  PyObject *__pyx_kp_u_Group_with_particles_forces_and;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3;
  PyObject *__pyx_n_s_MatrixForceApplier;
  PyObject *__pyx_n_s_MatrixForceApplier___reduce_cyth;
  PyObject *__pyx_n_s_MatrixForceApplier___setstate_cy;
  PyObject *__pyx_kp_u_New_D_physics_r_with_particles_f;
  PyObject *__pyx_kp_u_New_D_physics_r_with_particles_o;
  PyObject *__pyx_n_s_NotImplementedError;
  PyObject *__pyx_n_s_Particle;
  PyObject *__pyx_n_s_ParticleForceApplier;
  PyObject *__pyx_n_s_ParticleForceApplier___reduce_cy;
  PyObject *__pyx_n_s_ParticleForceApplier___setstate;
  PyObject *__pyx_n_s_Particle___reduce_cython;
  PyObject *__pyx_n_s_Particle___setstate_cython;
  PyObject *__pyx_n_s_PhysicsGroup;
  PyObject *__pyx_n_s_PhysicsGroup___reduce_cython;
  PyObject *__pyx_n_s_PhysicsGroup___setstate_cython;
  PyObject *__pyx_n_s_PhysicsSystem;
  PyObject *__pyx_n_s_PhysicsSystem___reduce_cython;
  PyObject *__pyx_n_s_PhysicsSystem___setstate_cython;
  PyObject *__pyx_n_s_PhysicsSystem_calculate;
  PyObject *__pyx_n_s_PhysicsSystem_destroy;
  PyObject *__pyx_n_s_PhysicsSystem_purge;
  PyObject *__pyx_n_s_PhysicsSystem_update;
  PyObject *__pyx_n_s_PickleError;
  PyObject *__pyx_n_s_RandomForceApplier;
  PyObject *__pyx_n_s_RandomForceApplier___reduce_cyth;
  PyObject *__pyx_n_s_RandomForceApplier___setstate_cy;
  PyObject *__pyx_kp_u_Reset_physics_system_r_for_run;
  PyObject *__pyx_n_s_TypeError;
  PyObject *__pyx_kp_u__5;
  PyObject *__pyx_n_s__52;
  PyObject *__pyx_n_s__8;
  PyObject *__pyx_n_s__9;
  PyObject *__pyx_n_u_acceleration;
  PyObject *__pyx_n_u_adhesion;
  PyObject *__pyx_n_u_anchor;
  PyObject *__pyx_n_s_args;
  PyObject *__pyx_n_s_asyncio;
  PyObject *__pyx_n_s_asyncio_coroutines;
  PyObject *__pyx_n_s_asyncio_tasks;
  PyObject *__pyx_n_s_await;
  PyObject *__pyx_n_u_barrier;
  PyObject *__pyx_n_u_buoyancy;
  PyObject *__pyx_n_s_calculate;
  PyObject *__pyx_n_u_charge;
  PyObject *__pyx_n_s_cline_in_traceback;
  PyObject *__pyx_n_s_clock;
  PyObject *__pyx_n_u_clock;
  PyObject *__pyx_n_s_close;
  PyObject *__pyx_n_u_collision;
  PyObject *__pyx_n_u_constant;
  PyObject *__pyx_n_s_debug;
  PyObject *__pyx_n_u_density;
  PyObject *__pyx_n_s_destroy;
  PyObject *__pyx_n_s_dict;
  PyObject *__pyx_n_s_dict_2;
  PyObject *__pyx_n_u_dimensions;
  PyObject *__pyx_n_u_direction;
  PyObject *__pyx_kp_u_disable;
  PyObject *__pyx_n_u_distance;
  PyObject *__pyx_n_u_drag;
  PyObject *__pyx_n_u_ease;
  PyObject *__pyx_n_u_electrostatic;
  PyObject *__pyx_kp_u_enable;
  PyObject *__pyx_n_s_engine;
  PyObject *__pyx_n_s_enumerate;
  PyObject *__pyx_n_s_extra;
  PyObject *__pyx_n_u_field;
  PyObject *__pyx_n_u_fixed;
  PyObject *__pyx_n_s_flitter_render_physics;
  PyObject *__pyx_n_u_flow;
  PyObject *__pyx_n_u_force;
  PyObject *__pyx_n_u_from;
  PyObject *__pyx_kp_u_gc;
  PyObject *__pyx_n_s_getstate;
  PyObject *__pyx_n_u_gravity;
  PyObject *__pyx_n_s_group;
  PyObject *__pyx_n_u_group;
  PyObject *__pyx_n_s_groups;
  PyObject *__pyx_n_s_i;
  PyObject *__pyx_n_u_id;
  PyObject *__pyx_n_s_import;
  PyObject *__pyx_n_s_initializing;
  PyObject *__pyx_n_s_inspect;
  PyObject *__pyx_n_s_is_coroutine;
  PyObject *__pyx_kp_u_isenabled;
  PyObject *__pyx_n_s_iteration;
  PyObject *__pyx_n_u_iteration;
  PyObject *__pyx_n_s_iteration_vector;
  PyObject *__pyx_n_s_kwargs;
  PyObject *__pyx_n_u_last;
  PyObject *__pyx_n_s_last_run;
  PyObject *__pyx_n_s_last_run_vector;
  PyObject *__pyx_n_s_last_time;
  PyObject *__pyx_n_s_logger;
  PyObject *__pyx_n_s_loguru;
  PyObject *__pyx_n_s_main;
  PyObject *__pyx_n_u_mass;
  PyObject *__pyx_n_u_max;
  PyObject *__pyx_n_u_max_distance;
  PyObject *__pyx_n_u_maximum;
  PyObject *__pyx_n_u_min;
  PyObject *__pyx_n_u_minimum;
  PyObject *__pyx_n_s_name;
  PyObject *__pyx_n_s_name_patch;
  PyObject *__pyx_n_s_new;
  PyObject *__pyx_n_s_new_state_keys;
  PyObject *__pyx_kp_s_no_default___reduce___due_to_non;
  PyObject *__pyx_n_s_node;
  PyObject *__pyx_n_u_normal;
  PyObject *__pyx_n_s_old_state_keys;
  PyObject *__pyx_n_u_overlap;
  PyObject *__pyx_n_s_particle;
  PyObject *__pyx_n_u_particle;
  PyObject *__pyx_n_s_particles_by_id;
  PyObject *__pyx_n_s_performance;
  PyObject *__pyx_n_s_pickle;
  PyObject *__pyx_n_u_position;
  PyObject *__pyx_n_u_power;
  PyObject *__pyx_n_s_purge;
  PyObject *__pyx_n_s_pyx_PickleError;
  PyObject *__pyx_n_s_pyx_checksum;
  PyObject *__pyx_n_s_pyx_result;
  PyObject *__pyx_n_s_pyx_state;
  PyObject *__pyx_n_s_pyx_type;
  PyObject *__pyx_n_s_pyx_unpickle_DistanceForceAppl;
  PyObject *__pyx_n_s_pyx_unpickle_Particle;
  PyObject *__pyx_n_s_pyx_unpickle_PhysicsSystem;
  PyObject *__pyx_n_s_pyx_vtable;
  PyObject *__pyx_n_u_radius;
  PyObject *__pyx_n_u_random;
  PyObject *__pyx_n_s_range;
  PyObject *__pyx_n_s_realtime;
  PyObject *__pyx_n_s_reduce;
  PyObject *__pyx_n_s_reduce_cython;
  PyObject *__pyx_n_s_reduce_ex;
  PyObject *__pyx_n_u_resolution;
  PyObject *__pyx_n_u_restitution;
  PyObject *__pyx_n_s_run;
  PyObject *__pyx_n_u_run;
  PyObject *__pyx_n_s_run_vector;
  PyObject *__pyx_n_s_seed;
  PyObject *__pyx_n_s_self;
  PyObject *__pyx_n_s_send;
  PyObject *__pyx_n_s_setstate;
  PyObject *__pyx_n_s_setstate_cython;
  PyObject *__pyx_n_s_spec;
  PyObject *__pyx_n_u_speed_of_light;
  PyObject *__pyx_kp_s_src_flitter_render_physics_pyx;
  PyObject *__pyx_n_s_state;
  PyObject *__pyx_n_u_state;
  PyObject *__pyx_n_s_state_key;
  PyObject *__pyx_n_s_state_prefix;
  PyObject *__pyx_n_s_strength;
  PyObject *__pyx_n_u_strength;
  PyObject *__pyx_kp_s_stringsource;
  PyObject *__pyx_n_s_target_fps;
  PyObject *__pyx_n_s_test;
  PyObject *__pyx_n_s_throw;
  PyObject *__pyx_n_s_time;
  PyObject *__pyx_n_u_time;
  PyObject *__pyx_n_s_time_vector;
  PyObject *__pyx_n_u_to;
  PyObject *__pyx_n_s_to_thread;
  PyObject *__pyx_n_s_top_group;
  PyObject *__pyx_n_s_trace;
  PyObject *__pyx_n_s_update;
  PyObject *__pyx_n_s_use_setstate;
  PyObject *__pyx_n_u_velocity;
  PyObject *__pyx_n_s_zero;
  PyObject *__pyx_int_0;
  PyObject *__pyx_int_1;
  PyObject *__pyx_int_64470655;
  PyObject *__pyx_int_136296653;
  PyObject *__pyx_int_142976427;
  PyObject *__pyx_int_158432715;
  PyObject *__pyx_int_165385471;
  PyObject *__pyx_int_191296081;
  PyObject *__pyx_int_210842809;
  PyObject *__pyx_int_211912363;
  PyObject *__pyx_int_253013475;
  PyObject *__pyx_codeobj_;
  PyObject *__pyx_tuple__4;
  PyObject *__pyx_tuple__6;
  PyObject *__pyx_tuple__7;
  PyObject *__pyx_tuple__10;
  PyObject *__pyx_tuple__12;
  PyObject *__pyx_tuple__14;
  PyObject *__pyx_tuple__43;
  PyObject *__pyx_tuple__44;
  PyObject *__pyx_tuple__48;
  PyObject *__pyx_codeobj__2;
  PyObject *__pyx_codeobj__3;
  PyObject *__pyx_codeobj__11;
  PyObject *__pyx_codeobj__13;
  PyObject *__pyx_codeobj__15;
  PyObject *__pyx_codeobj__16;
  PyObject *__pyx_codeobj__17;
  PyObject *__pyx_codeobj__18;
  PyObject *__pyx_codeobj__19;
  PyObject *__pyx_codeobj__20;
  PyObject *__pyx_codeobj__21;
  PyObject *__pyx_codeobj__22;
  PyObject *__pyx_codeobj__23;
  PyObject *__pyx_codeobj__24;
  PyObject *__pyx_codeobj__25;
  PyObject *__pyx_codeobj__26;
  PyObject *__pyx_codeobj__27;
  PyObject *__pyx_codeobj__28;
  PyObject *__pyx_codeobj__29;
  PyObject *__pyx_codeobj__30;
  PyObject *__pyx_codeobj__31;
  PyObject *__pyx_codeobj__32;
  PyObject *__pyx_codeobj__33;
  PyObject *__pyx_codeobj__34;
  PyObject *__pyx_codeobj__35;
  PyObject *__pyx_codeobj__36;
  PyObject *__pyx_codeobj__37;
  PyObject *__pyx_codeobj__38;
  PyObject *__pyx_codeobj__39;
  PyObject *__pyx_codeobj__40;
  PyObject *__pyx_codeobj__41;
  PyObject *__pyx_codeobj__42;
  PyObject *__pyx_codeobj__45;
  PyObject *__pyx_codeobj__46;
  PyObject *__pyx_codeobj__47;
  PyObject *__pyx_codeobj__49;
  PyObject *__pyx_codeobj__50;
  PyObject *__pyx_codeobj__51;
} __pyx_mstate;

#if CYTHON_USE_MODULE_STATE
#ifdef __cplusplus
namespace {
  extern struct PyModuleDef __pyx_moduledef;
} /* anonymous namespace */
#else
static struct PyModuleDef __pyx_moduledef;
#endif

#define __pyx_mstate(o) ((__pyx_mstate *)__Pyx_PyModule_GetState(o))

#define __pyx_mstate_global (__pyx_mstate(PyState_FindModule(&__pyx_moduledef)))

#define __pyx_m (PyState_FindModule(&__pyx_moduledef))
#else
static __pyx_mstate __pyx_mstate_global_static =
#ifdef __cplusplus
    {};
#else
    {0};
#endif
static __pyx_mstate *__pyx_mstate_global = &__pyx_mstate_global_static;
#endif
/* #### Code section: module_state_clear ### */
#if CYTHON_USE_MODULE_STATE
static int __pyx_m_clear(PyObject *m) {
  __pyx_mstate *clear_module_state = __pyx_mstate(m);
  if (!clear_module_state) return 0;
  Py_CLEAR(clear_module_state->__pyx_d);
  Py_CLEAR(clear_module_state->__pyx_b);
  Py_CLEAR(clear_module_state->__pyx_cython_runtime);
  Py_CLEAR(clear_module_state->__pyx_empty_tuple);
  Py_CLEAR(clear_module_state->__pyx_empty_bytes);
  Py_CLEAR(clear_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_CLEAR(clear_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_CLEAR(clear_module_state->__pyx_FusedFunctionType);
  #endif
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_5model_Vector);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_5model_Matrix33);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_5model_Matrix44);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_5model_Quaternion);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_5model_Node);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_5model_StateDict);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_5model_Context);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_9functions_uniform);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_9functions_beta);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_8language_9functions_normal);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_4type_type);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_4bool_bool);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_7complex_complex);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_6render_7physics_PhysicsGroup);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_6render_7physics_PhysicsGroup);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_6render_7physics_Particle);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_6render_7physics_Particle);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_6render_7physics_Barrier);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_6render_7physics_Barrier);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_6render_7physics_ParticleForceApplier);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_6render_7physics_ParticleForceApplier);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_6render_7physics_MatrixForceApplier);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_6render_7physics_MatrixForceApplier);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_6render_7physics_DistanceForceApplier);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_6render_7physics_DistanceForceApplier);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_6render_7physics_DragForceApplier);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_6render_7physics_DragForceApplier);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_6render_7physics_BuoyancyForceApplier);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_6render_7physics_BuoyancyForceApplier);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_6render_7physics_ConstantForceApplier);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_6render_7physics_ConstantForceApplier);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_6render_7physics_RandomForceApplier);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_6render_7physics_RandomForceApplier);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_6render_7physics_FieldForceApplier);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_6render_7physics_FieldForceApplier);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_6render_7physics_CollisionForceApplier);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_6render_7physics_CollisionForceApplier);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_6render_7physics_GravityForceApplier);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_6render_7physics_GravityForceApplier);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_6render_7physics_ElectrostaticForceApplier);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_6render_7physics_ElectrostaticForceApplier);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_6render_7physics_AdhesionForceApplier);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_6render_7physics_AdhesionForceApplier);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_6render_7physics_PhysicsSystem);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_6render_7physics_PhysicsSystem);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_6render_7physics___pyx_scope_struct__destroy);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_6render_7physics___pyx_scope_struct__destroy);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_6render_7physics___pyx_scope_struct_1_purge);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_6render_7physics___pyx_scope_struct_1_purge);
  Py_CLEAR(clear_module_state->__pyx_ptype_7flitter_6render_7physics___pyx_scope_struct_2_update);
  Py_CLEAR(clear_module_state->__pyx_type_7flitter_6render_7physics___pyx_scope_struct_2_update);
  Py_CLEAR(clear_module_state->__pyx_n_s_AdhesionForceApplier);
  Py_CLEAR(clear_module_state->__pyx_n_s_AdhesionForceApplier___reduce_cy);
  Py_CLEAR(clear_module_state->__pyx_n_s_AdhesionForceApplier___setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_Barrier);
  Py_CLEAR(clear_module_state->__pyx_n_s_Barrier___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Barrier___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_BuoyancyForceApplier);
  Py_CLEAR(clear_module_state->__pyx_n_s_BuoyancyForceApplier___reduce_cy);
  Py_CLEAR(clear_module_state->__pyx_n_s_BuoyancyForceApplier___setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_CollisionForceApplier);
  Py_CLEAR(clear_module_state->__pyx_n_s_CollisionForceApplier___reduce_c);
  Py_CLEAR(clear_module_state->__pyx_n_s_CollisionForceApplier___setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_ConstantForceApplier);
  Py_CLEAR(clear_module_state->__pyx_n_s_ConstantForceApplier___reduce_cy);
  Py_CLEAR(clear_module_state->__pyx_n_s_ConstantForceApplier___setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_DistanceForceApplier);
  Py_CLEAR(clear_module_state->__pyx_n_s_DistanceForceApplier___reduce_cy);
  Py_CLEAR(clear_module_state->__pyx_n_s_DistanceForceApplier___setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_DragForceApplier);
  Py_CLEAR(clear_module_state->__pyx_n_s_DragForceApplier___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_DragForceApplier___setstate_cyth);
  Py_CLEAR(clear_module_state->__pyx_n_s_ElectrostaticForceApplier);
  Py_CLEAR(clear_module_state->__pyx_n_s_ElectrostaticForceApplier___redu);
  Py_CLEAR(clear_module_state->__pyx_n_s_ElectrostaticForceApplier___sets);
  Py_CLEAR(clear_module_state->__pyx_n_s_FieldForceApplier);
  Py_CLEAR(clear_module_state->__pyx_n_s_FieldForceApplier___reduce_cytho);
  Py_CLEAR(clear_module_state->__pyx_n_s_FieldForceApplier___setstate_cyt);
  Py_CLEAR(clear_module_state->__pyx_n_s_GravityForceApplier);
  Py_CLEAR(clear_module_state->__pyx_n_s_GravityForceApplier___reduce_cyt);
  Py_CLEAR(clear_module_state->__pyx_n_s_GravityForceApplier___setstate_c);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Group_with_particles_forces_and);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3);
  Py_CLEAR(clear_module_state->__pyx_n_s_MatrixForceApplier);
  Py_CLEAR(clear_module_state->__pyx_n_s_MatrixForceApplier___reduce_cyth);
  Py_CLEAR(clear_module_state->__pyx_n_s_MatrixForceApplier___setstate_cy);
  Py_CLEAR(clear_module_state->__pyx_kp_u_New_D_physics_r_with_particles_f);
  Py_CLEAR(clear_module_state->__pyx_kp_u_New_D_physics_r_with_particles_o);
  Py_CLEAR(clear_module_state->__pyx_n_s_NotImplementedError);
  Py_CLEAR(clear_module_state->__pyx_n_s_Particle);
  Py_CLEAR(clear_module_state->__pyx_n_s_ParticleForceApplier);
  Py_CLEAR(clear_module_state->__pyx_n_s_ParticleForceApplier___reduce_cy);
  Py_CLEAR(clear_module_state->__pyx_n_s_ParticleForceApplier___setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_Particle___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Particle___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_PhysicsGroup);
  Py_CLEAR(clear_module_state->__pyx_n_s_PhysicsGroup___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_PhysicsGroup___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_PhysicsSystem);
  Py_CLEAR(clear_module_state->__pyx_n_s_PhysicsSystem___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_PhysicsSystem___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_PhysicsSystem_calculate);
  Py_CLEAR(clear_module_state->__pyx_n_s_PhysicsSystem_destroy);
  Py_CLEAR(clear_module_state->__pyx_n_s_PhysicsSystem_purge);
  Py_CLEAR(clear_module_state->__pyx_n_s_PhysicsSystem_update);
  Py_CLEAR(clear_module_state->__pyx_n_s_PickleError);
  Py_CLEAR(clear_module_state->__pyx_n_s_RandomForceApplier);
  Py_CLEAR(clear_module_state->__pyx_n_s_RandomForceApplier___reduce_cyth);
  Py_CLEAR(clear_module_state->__pyx_n_s_RandomForceApplier___setstate_cy);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Reset_physics_system_r_for_run);
  Py_CLEAR(clear_module_state->__pyx_n_s_TypeError);
  Py_CLEAR(clear_module_state->__pyx_kp_u__5);
  Py_CLEAR(clear_module_state->__pyx_n_s__52);
  Py_CLEAR(clear_module_state->__pyx_n_s__8);
  Py_CLEAR(clear_module_state->__pyx_n_s__9);
  Py_CLEAR(clear_module_state->__pyx_n_u_acceleration);
  Py_CLEAR(clear_module_state->__pyx_n_u_adhesion);
  Py_CLEAR(clear_module_state->__pyx_n_u_anchor);
  Py_CLEAR(clear_module_state->__pyx_n_s_args);
  Py_CLEAR(clear_module_state->__pyx_n_s_asyncio);
  Py_CLEAR(clear_module_state->__pyx_n_s_asyncio_coroutines);
  Py_CLEAR(clear_module_state->__pyx_n_s_asyncio_tasks);
  Py_CLEAR(clear_module_state->__pyx_n_s_await);
  Py_CLEAR(clear_module_state->__pyx_n_u_barrier);
  Py_CLEAR(clear_module_state->__pyx_n_u_buoyancy);
  Py_CLEAR(clear_module_state->__pyx_n_s_calculate);
  Py_CLEAR(clear_module_state->__pyx_n_u_charge);
  Py_CLEAR(clear_module_state->__pyx_n_s_cline_in_traceback);
  Py_CLEAR(clear_module_state->__pyx_n_s_clock);
  Py_CLEAR(clear_module_state->__pyx_n_u_clock);
  Py_CLEAR(clear_module_state->__pyx_n_s_close);
  Py_CLEAR(clear_module_state->__pyx_n_u_collision);
  Py_CLEAR(clear_module_state->__pyx_n_u_constant);
  Py_CLEAR(clear_module_state->__pyx_n_s_debug);
  Py_CLEAR(clear_module_state->__pyx_n_u_density);
  Py_CLEAR(clear_module_state->__pyx_n_s_destroy);
  Py_CLEAR(clear_module_state->__pyx_n_s_dict);
  Py_CLEAR(clear_module_state->__pyx_n_s_dict_2);
  Py_CLEAR(clear_module_state->__pyx_n_u_dimensions);
  Py_CLEAR(clear_module_state->__pyx_n_u_direction);
  Py_CLEAR(clear_module_state->__pyx_kp_u_disable);
  Py_CLEAR(clear_module_state->__pyx_n_u_distance);
  Py_CLEAR(clear_module_state->__pyx_n_u_drag);
  Py_CLEAR(clear_module_state->__pyx_n_u_ease);
  Py_CLEAR(clear_module_state->__pyx_n_u_electrostatic);
  Py_CLEAR(clear_module_state->__pyx_kp_u_enable);
  Py_CLEAR(clear_module_state->__pyx_n_s_engine);
  Py_CLEAR(clear_module_state->__pyx_n_s_enumerate);
  Py_CLEAR(clear_module_state->__pyx_n_s_extra);
  Py_CLEAR(clear_module_state->__pyx_n_u_field);
  Py_CLEAR(clear_module_state->__pyx_n_u_fixed);
  Py_CLEAR(clear_module_state->__pyx_n_s_flitter_render_physics);
  Py_CLEAR(clear_module_state->__pyx_n_u_flow);
  Py_CLEAR(clear_module_state->__pyx_n_u_force);
  Py_CLEAR(clear_module_state->__pyx_n_u_from);
  Py_CLEAR(clear_module_state->__pyx_kp_u_gc);
  Py_CLEAR(clear_module_state->__pyx_n_s_getstate);
  Py_CLEAR(clear_module_state->__pyx_n_u_gravity);
  Py_CLEAR(clear_module_state->__pyx_n_s_group);
  Py_CLEAR(clear_module_state->__pyx_n_u_group);
  Py_CLEAR(clear_module_state->__pyx_n_s_groups);
  Py_CLEAR(clear_module_state->__pyx_n_s_i);
  Py_CLEAR(clear_module_state->__pyx_n_u_id);
  Py_CLEAR(clear_module_state->__pyx_n_s_import);
  Py_CLEAR(clear_module_state->__pyx_n_s_initializing);
  Py_CLEAR(clear_module_state->__pyx_n_s_inspect);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_coroutine);
  Py_CLEAR(clear_module_state->__pyx_kp_u_isenabled);
  Py_CLEAR(clear_module_state->__pyx_n_s_iteration);
  Py_CLEAR(clear_module_state->__pyx_n_u_iteration);
  Py_CLEAR(clear_module_state->__pyx_n_s_iteration_vector);
  Py_CLEAR(clear_module_state->__pyx_n_s_kwargs);
  Py_CLEAR(clear_module_state->__pyx_n_u_last);
  Py_CLEAR(clear_module_state->__pyx_n_s_last_run);
  Py_CLEAR(clear_module_state->__pyx_n_s_last_run_vector);
  Py_CLEAR(clear_module_state->__pyx_n_s_last_time);
  Py_CLEAR(clear_module_state->__pyx_n_s_logger);
  Py_CLEAR(clear_module_state->__pyx_n_s_loguru);
  Py_CLEAR(clear_module_state->__pyx_n_s_main);
  Py_CLEAR(clear_module_state->__pyx_n_u_mass);
  Py_CLEAR(clear_module_state->__pyx_n_u_max);
  Py_CLEAR(clear_module_state->__pyx_n_u_max_distance);
  Py_CLEAR(clear_module_state->__pyx_n_u_maximum);
  Py_CLEAR(clear_module_state->__pyx_n_u_min);
  Py_CLEAR(clear_module_state->__pyx_n_u_minimum);
  Py_CLEAR(clear_module_state->__pyx_n_s_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_name_patch);
  Py_CLEAR(clear_module_state->__pyx_n_s_new);
  Py_CLEAR(clear_module_state->__pyx_n_s_new_state_keys);
  Py_CLEAR(clear_module_state->__pyx_kp_s_no_default___reduce___due_to_non);
  Py_CLEAR(clear_module_state->__pyx_n_s_node);
  Py_CLEAR(clear_module_state->__pyx_n_u_normal);
  Py_CLEAR(clear_module_state->__pyx_n_s_old_state_keys);
  Py_CLEAR(clear_module_state->__pyx_n_u_overlap);
  Py_CLEAR(clear_module_state->__pyx_n_s_particle);
  Py_CLEAR(clear_module_state->__pyx_n_u_particle);
  Py_CLEAR(clear_module_state->__pyx_n_s_particles_by_id);
  Py_CLEAR(clear_module_state->__pyx_n_s_performance);
  Py_CLEAR(clear_module_state->__pyx_n_s_pickle);
  Py_CLEAR(clear_module_state->__pyx_n_u_position);
  Py_CLEAR(clear_module_state->__pyx_n_u_power);
  Py_CLEAR(clear_module_state->__pyx_n_s_purge);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_PickleError);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_checksum);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_result);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_DistanceForceAppl);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_Particle);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_PhysicsSystem);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_vtable);
  Py_CLEAR(clear_module_state->__pyx_n_u_radius);
  Py_CLEAR(clear_module_state->__pyx_n_u_random);
  Py_CLEAR(clear_module_state->__pyx_n_s_range);
  Py_CLEAR(clear_module_state->__pyx_n_s_realtime);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce_ex);
  Py_CLEAR(clear_module_state->__pyx_n_u_resolution);
  Py_CLEAR(clear_module_state->__pyx_n_u_restitution);
  Py_CLEAR(clear_module_state->__pyx_n_s_run);
  Py_CLEAR(clear_module_state->__pyx_n_u_run);
  Py_CLEAR(clear_module_state->__pyx_n_s_run_vector);
  Py_CLEAR(clear_module_state->__pyx_n_s_seed);
  Py_CLEAR(clear_module_state->__pyx_n_s_self);
  Py_CLEAR(clear_module_state->__pyx_n_s_send);
  Py_CLEAR(clear_module_state->__pyx_n_s_setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_spec);
  Py_CLEAR(clear_module_state->__pyx_n_u_speed_of_light);
  Py_CLEAR(clear_module_state->__pyx_kp_s_src_flitter_render_physics_pyx);
  Py_CLEAR(clear_module_state->__pyx_n_s_state);
  Py_CLEAR(clear_module_state->__pyx_n_u_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_state_key);
  Py_CLEAR(clear_module_state->__pyx_n_s_state_prefix);
  Py_CLEAR(clear_module_state->__pyx_n_s_strength);
  Py_CLEAR(clear_module_state->__pyx_n_u_strength);
  Py_CLEAR(clear_module_state->__pyx_kp_s_stringsource);
  Py_CLEAR(clear_module_state->__pyx_n_s_target_fps);
  Py_CLEAR(clear_module_state->__pyx_n_s_test);
  Py_CLEAR(clear_module_state->__pyx_n_s_throw);
  Py_CLEAR(clear_module_state->__pyx_n_s_time);
  Py_CLEAR(clear_module_state->__pyx_n_u_time);
  Py_CLEAR(clear_module_state->__pyx_n_s_time_vector);
  Py_CLEAR(clear_module_state->__pyx_n_u_to);
  Py_CLEAR(clear_module_state->__pyx_n_s_to_thread);
  Py_CLEAR(clear_module_state->__pyx_n_s_top_group);
  Py_CLEAR(clear_module_state->__pyx_n_s_trace);
  Py_CLEAR(clear_module_state->__pyx_n_s_update);
  Py_CLEAR(clear_module_state->__pyx_n_s_use_setstate);
  Py_CLEAR(clear_module_state->__pyx_n_u_velocity);
  Py_CLEAR(clear_module_state->__pyx_n_s_zero);
  Py_CLEAR(clear_module_state->__pyx_int_0);
  Py_CLEAR(clear_module_state->__pyx_int_1);
  Py_CLEAR(clear_module_state->__pyx_int_64470655);
  Py_CLEAR(clear_module_state->__pyx_int_136296653);
  Py_CLEAR(clear_module_state->__pyx_int_142976427);
  Py_CLEAR(clear_module_state->__pyx_int_158432715);
  Py_CLEAR(clear_module_state->__pyx_int_165385471);
  Py_CLEAR(clear_module_state->__pyx_int_191296081);
  Py_CLEAR(clear_module_state->__pyx_int_210842809);
  Py_CLEAR(clear_module_state->__pyx_int_211912363);
  Py_CLEAR(clear_module_state->__pyx_int_253013475);
  Py_CLEAR(clear_module_state->__pyx_codeobj_);
  Py_CLEAR(clear_module_state->__pyx_tuple__4);
  Py_CLEAR(clear_module_state->__pyx_tuple__6);
  Py_CLEAR(clear_module_state->__pyx_tuple__7);
  Py_CLEAR(clear_module_state->__pyx_tuple__10);
  Py_CLEAR(clear_module_state->__pyx_tuple__12);
  Py_CLEAR(clear_module_state->__pyx_tuple__14);
  Py_CLEAR(clear_module_state->__pyx_tuple__43);
  Py_CLEAR(clear_module_state->__pyx_tuple__44);
  Py_CLEAR(clear_module_state->__pyx_tuple__48);
  Py_CLEAR(clear_module_state->__pyx_codeobj__2);
  Py_CLEAR(clear_module_state->__pyx_codeobj__3);
  Py_CLEAR(clear_module_state->__pyx_codeobj__11);
  Py_CLEAR(clear_module_state->__pyx_codeobj__13);
  Py_CLEAR(clear_module_state->__pyx_codeobj__15);
  Py_CLEAR(clear_module_state->__pyx_codeobj__16);
  Py_CLEAR(clear_module_state->__pyx_codeobj__17);
  Py_CLEAR(clear_module_state->__pyx_codeobj__18);
  Py_CLEAR(clear_module_state->__pyx_codeobj__19);
  Py_CLEAR(clear_module_state->__pyx_codeobj__20);
  Py_CLEAR(clear_module_state->__pyx_codeobj__21);
  Py_CLEAR(clear_module_state->__pyx_codeobj__22);
  Py_CLEAR(clear_module_state->__pyx_codeobj__23);
  Py_CLEAR(clear_module_state->__pyx_codeobj__24);
  Py_CLEAR(clear_module_state->__pyx_codeobj__25);
  Py_CLEAR(clear_module_state->__pyx_codeobj__26);
  Py_CLEAR(clear_module_state->__pyx_codeobj__27);
  Py_CLEAR(clear_module_state->__pyx_codeobj__28);
  Py_CLEAR(clear_module_state->__pyx_codeobj__29);
  Py_CLEAR(clear_module_state->__pyx_codeobj__30);
  Py_CLEAR(clear_module_state->__pyx_codeobj__31);
  Py_CLEAR(clear_module_state->__pyx_codeobj__32);
  Py_CLEAR(clear_module_state->__pyx_codeobj__33);
  Py_CLEAR(clear_module_state->__pyx_codeobj__34);
  Py_CLEAR(clear_module_state->__pyx_codeobj__35);
  Py_CLEAR(clear_module_state->__pyx_codeobj__36);
  Py_CLEAR(clear_module_state->__pyx_codeobj__37);
  Py_CLEAR(clear_module_state->__pyx_codeobj__38);
  Py_CLEAR(clear_module_state->__pyx_codeobj__39);
  Py_CLEAR(clear_module_state->__pyx_codeobj__40);
  Py_CLEAR(clear_module_state->__pyx_codeobj__41);
  Py_CLEAR(clear_module_state->__pyx_codeobj__42);
  Py_CLEAR(clear_module_state->__pyx_codeobj__45);
  Py_CLEAR(clear_module_state->__pyx_codeobj__46);
  Py_CLEAR(clear_module_state->__pyx_codeobj__47);
  Py_CLEAR(clear_module_state->__pyx_codeobj__49);
  Py_CLEAR(clear_module_state->__pyx_codeobj__50);
  Py_CLEAR(clear_module_state->__pyx_codeobj__51);
  return 0;
}
#endif
/* #### Code section: module_state_traverse ### */
#if CYTHON_USE_MODULE_STATE
static int __pyx_m_traverse(PyObject *m, visitproc visit, void *arg) {
  __pyx_mstate *traverse_module_state = __pyx_mstate(m);
  if (!traverse_module_state) return 0;
  Py_VISIT(traverse_module_state->__pyx_d);
  Py_VISIT(traverse_module_state->__pyx_b);
  Py_VISIT(traverse_module_state->__pyx_cython_runtime);
  Py_VISIT(traverse_module_state->__pyx_empty_tuple);
  Py_VISIT(traverse_module_state->__pyx_empty_bytes);
  Py_VISIT(traverse_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_VISIT(traverse_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_VISIT(traverse_module_state->__pyx_FusedFunctionType);
  #endif
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_5model_Vector);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_5model_Matrix33);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_5model_Matrix44);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_5model_Quaternion);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_5model_Node);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_5model_StateDict);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_5model_Context);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_9functions_uniform);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_9functions_beta);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_8language_9functions_normal);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_4type_type);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_4bool_bool);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_7complex_complex);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_6render_7physics_PhysicsGroup);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_6render_7physics_PhysicsGroup);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_6render_7physics_Particle);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_6render_7physics_Particle);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_6render_7physics_Barrier);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_6render_7physics_Barrier);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_6render_7physics_ParticleForceApplier);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_6render_7physics_ParticleForceApplier);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_6render_7physics_MatrixForceApplier);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_6render_7physics_MatrixForceApplier);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_6render_7physics_DistanceForceApplier);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_6render_7physics_DistanceForceApplier);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_6render_7physics_DragForceApplier);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_6render_7physics_DragForceApplier);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_6render_7physics_BuoyancyForceApplier);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_6render_7physics_BuoyancyForceApplier);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_6render_7physics_ConstantForceApplier);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_6render_7physics_ConstantForceApplier);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_6render_7physics_RandomForceApplier);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_6render_7physics_RandomForceApplier);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_6render_7physics_FieldForceApplier);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_6render_7physics_FieldForceApplier);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_6render_7physics_CollisionForceApplier);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_6render_7physics_CollisionForceApplier);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_6render_7physics_GravityForceApplier);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_6render_7physics_GravityForceApplier);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_6render_7physics_ElectrostaticForceApplier);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_6render_7physics_ElectrostaticForceApplier);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_6render_7physics_AdhesionForceApplier);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_6render_7physics_AdhesionForceApplier);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_6render_7physics_PhysicsSystem);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_6render_7physics_PhysicsSystem);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_6render_7physics___pyx_scope_struct__destroy);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_6render_7physics___pyx_scope_struct__destroy);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_6render_7physics___pyx_scope_struct_1_purge);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_6render_7physics___pyx_scope_struct_1_purge);
  Py_VISIT(traverse_module_state->__pyx_ptype_7flitter_6render_7physics___pyx_scope_struct_2_update);
  Py_VISIT(traverse_module_state->__pyx_type_7flitter_6render_7physics___pyx_scope_struct_2_update);
  Py_VISIT(traverse_module_state->__pyx_n_s_AdhesionForceApplier);
  Py_VISIT(traverse_module_state->__pyx_n_s_AdhesionForceApplier___reduce_cy);
  Py_VISIT(traverse_module_state->__pyx_n_s_AdhesionForceApplier___setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_Barrier);
  Py_VISIT(traverse_module_state->__pyx_n_s_Barrier___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Barrier___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_BuoyancyForceApplier);
  Py_VISIT(traverse_module_state->__pyx_n_s_BuoyancyForceApplier___reduce_cy);
  Py_VISIT(traverse_module_state->__pyx_n_s_BuoyancyForceApplier___setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_CollisionForceApplier);
  Py_VISIT(traverse_module_state->__pyx_n_s_CollisionForceApplier___reduce_c);
  Py_VISIT(traverse_module_state->__pyx_n_s_CollisionForceApplier___setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_ConstantForceApplier);
  Py_VISIT(traverse_module_state->__pyx_n_s_ConstantForceApplier___reduce_cy);
  Py_VISIT(traverse_module_state->__pyx_n_s_ConstantForceApplier___setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_DistanceForceApplier);
  Py_VISIT(traverse_module_state->__pyx_n_s_DistanceForceApplier___reduce_cy);
  Py_VISIT(traverse_module_state->__pyx_n_s_DistanceForceApplier___setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_DragForceApplier);
  Py_VISIT(traverse_module_state->__pyx_n_s_DragForceApplier___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_DragForceApplier___setstate_cyth);
  Py_VISIT(traverse_module_state->__pyx_n_s_ElectrostaticForceApplier);
  Py_VISIT(traverse_module_state->__pyx_n_s_ElectrostaticForceApplier___redu);
  Py_VISIT(traverse_module_state->__pyx_n_s_ElectrostaticForceApplier___sets);
  Py_VISIT(traverse_module_state->__pyx_n_s_FieldForceApplier);
  Py_VISIT(traverse_module_state->__pyx_n_s_FieldForceApplier___reduce_cytho);
  Py_VISIT(traverse_module_state->__pyx_n_s_FieldForceApplier___setstate_cyt);
  Py_VISIT(traverse_module_state->__pyx_n_s_GravityForceApplier);
  Py_VISIT(traverse_module_state->__pyx_n_s_GravityForceApplier___reduce_cyt);
  Py_VISIT(traverse_module_state->__pyx_n_s_GravityForceApplier___setstate_c);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Group_with_particles_forces_and);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3);
  Py_VISIT(traverse_module_state->__pyx_n_s_MatrixForceApplier);
  Py_VISIT(traverse_module_state->__pyx_n_s_MatrixForceApplier___reduce_cyth);
  Py_VISIT(traverse_module_state->__pyx_n_s_MatrixForceApplier___setstate_cy);
  Py_VISIT(traverse_module_state->__pyx_kp_u_New_D_physics_r_with_particles_f);
  Py_VISIT(traverse_module_state->__pyx_kp_u_New_D_physics_r_with_particles_o);
  Py_VISIT(traverse_module_state->__pyx_n_s_NotImplementedError);
  Py_VISIT(traverse_module_state->__pyx_n_s_Particle);
  Py_VISIT(traverse_module_state->__pyx_n_s_ParticleForceApplier);
  Py_VISIT(traverse_module_state->__pyx_n_s_ParticleForceApplier___reduce_cy);
  Py_VISIT(traverse_module_state->__pyx_n_s_ParticleForceApplier___setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_Particle___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Particle___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_PhysicsGroup);
  Py_VISIT(traverse_module_state->__pyx_n_s_PhysicsGroup___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_PhysicsGroup___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_PhysicsSystem);
  Py_VISIT(traverse_module_state->__pyx_n_s_PhysicsSystem___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_PhysicsSystem___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_PhysicsSystem_calculate);
  Py_VISIT(traverse_module_state->__pyx_n_s_PhysicsSystem_destroy);
  Py_VISIT(traverse_module_state->__pyx_n_s_PhysicsSystem_purge);
  Py_VISIT(traverse_module_state->__pyx_n_s_PhysicsSystem_update);
  Py_VISIT(traverse_module_state->__pyx_n_s_PickleError);
  Py_VISIT(traverse_module_state->__pyx_n_s_RandomForceApplier);
  Py_VISIT(traverse_module_state->__pyx_n_s_RandomForceApplier___reduce_cyth);
  Py_VISIT(traverse_module_state->__pyx_n_s_RandomForceApplier___setstate_cy);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Reset_physics_system_r_for_run);
  Py_VISIT(traverse_module_state->__pyx_n_s_TypeError);
  Py_VISIT(traverse_module_state->__pyx_kp_u__5);
  Py_VISIT(traverse_module_state->__pyx_n_s__52);
  Py_VISIT(traverse_module_state->__pyx_n_s__8);
  Py_VISIT(traverse_module_state->__pyx_n_s__9);
  Py_VISIT(traverse_module_state->__pyx_n_u_acceleration);
  Py_VISIT(traverse_module_state->__pyx_n_u_adhesion);
  Py_VISIT(traverse_module_state->__pyx_n_u_anchor);
  Py_VISIT(traverse_module_state->__pyx_n_s_args);
  Py_VISIT(traverse_module_state->__pyx_n_s_asyncio);
  Py_VISIT(traverse_module_state->__pyx_n_s_asyncio_coroutines);
  Py_VISIT(traverse_module_state->__pyx_n_s_asyncio_tasks);
  Py_VISIT(traverse_module_state->__pyx_n_s_await);
  Py_VISIT(traverse_module_state->__pyx_n_u_barrier);
  Py_VISIT(traverse_module_state->__pyx_n_u_buoyancy);
  Py_VISIT(traverse_module_state->__pyx_n_s_calculate);
  Py_VISIT(traverse_module_state->__pyx_n_u_charge);
  Py_VISIT(traverse_module_state->__pyx_n_s_cline_in_traceback);
  Py_VISIT(traverse_module_state->__pyx_n_s_clock);
  Py_VISIT(traverse_module_state->__pyx_n_u_clock);
  Py_VISIT(traverse_module_state->__pyx_n_s_close);
  Py_VISIT(traverse_module_state->__pyx_n_u_collision);
  Py_VISIT(traverse_module_state->__pyx_n_u_constant);
  Py_VISIT(traverse_module_state->__pyx_n_s_debug);
  Py_VISIT(traverse_module_state->__pyx_n_u_density);
  Py_VISIT(traverse_module_state->__pyx_n_s_destroy);
  Py_VISIT(traverse_module_state->__pyx_n_s_dict);
  Py_VISIT(traverse_module_state->__pyx_n_s_dict_2);
  Py_VISIT(traverse_module_state->__pyx_n_u_dimensions);
  Py_VISIT(traverse_module_state->__pyx_n_u_direction);
  Py_VISIT(traverse_module_state->__pyx_kp_u_disable);
  Py_VISIT(traverse_module_state->__pyx_n_u_distance);
  Py_VISIT(traverse_module_state->__pyx_n_u_drag);
  Py_VISIT(traverse_module_state->__pyx_n_u_ease);
  Py_VISIT(traverse_module_state->__pyx_n_u_electrostatic);
  Py_VISIT(traverse_module_state->__pyx_kp_u_enable);
  Py_VISIT(traverse_module_state->__pyx_n_s_engine);
  Py_VISIT(traverse_module_state->__pyx_n_s_enumerate);
  Py_VISIT(traverse_module_state->__pyx_n_s_extra);
  Py_VISIT(traverse_module_state->__pyx_n_u_field);
  Py_VISIT(traverse_module_state->__pyx_n_u_fixed);
  Py_VISIT(traverse_module_state->__pyx_n_s_flitter_render_physics);
  Py_VISIT(traverse_module_state->__pyx_n_u_flow);
  Py_VISIT(traverse_module_state->__pyx_n_u_force);
  Py_VISIT(traverse_module_state->__pyx_n_u_from);
  Py_VISIT(traverse_module_state->__pyx_kp_u_gc);
  Py_VISIT(traverse_module_state->__pyx_n_s_getstate);
  Py_VISIT(traverse_module_state->__pyx_n_u_gravity);
  Py_VISIT(traverse_module_state->__pyx_n_s_group);
  Py_VISIT(traverse_module_state->__pyx_n_u_group);
  Py_VISIT(traverse_module_state->__pyx_n_s_groups);
  Py_VISIT(traverse_module_state->__pyx_n_s_i);
  Py_VISIT(traverse_module_state->__pyx_n_u_id);
  Py_VISIT(traverse_module_state->__pyx_n_s_import);
  Py_VISIT(traverse_module_state->__pyx_n_s_initializing);
  Py_VISIT(traverse_module_state->__pyx_n_s_inspect);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_coroutine);
  Py_VISIT(traverse_module_state->__pyx_kp_u_isenabled);
  Py_VISIT(traverse_module_state->__pyx_n_s_iteration);
  Py_VISIT(traverse_module_state->__pyx_n_u_iteration);
  Py_VISIT(traverse_module_state->__pyx_n_s_iteration_vector);
  Py_VISIT(traverse_module_state->__pyx_n_s_kwargs);
  Py_VISIT(traverse_module_state->__pyx_n_u_last);
  Py_VISIT(traverse_module_state->__pyx_n_s_last_run);
  Py_VISIT(traverse_module_state->__pyx_n_s_last_run_vector);
  Py_VISIT(traverse_module_state->__pyx_n_s_last_time);
  Py_VISIT(traverse_module_state->__pyx_n_s_logger);
  Py_VISIT(traverse_module_state->__pyx_n_s_loguru);
  Py_VISIT(traverse_module_state->__pyx_n_s_main);
  Py_VISIT(traverse_module_state->__pyx_n_u_mass);
  Py_VISIT(traverse_module_state->__pyx_n_u_max);
  Py_VISIT(traverse_module_state->__pyx_n_u_max_distance);
  Py_VISIT(traverse_module_state->__pyx_n_u_maximum);
  Py_VISIT(traverse_module_state->__pyx_n_u_min);
  Py_VISIT(traverse_module_state->__pyx_n_u_minimum);
  Py_VISIT(traverse_module_state->__pyx_n_s_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_name_patch);
  Py_VISIT(traverse_module_state->__pyx_n_s_new);
  Py_VISIT(traverse_module_state->__pyx_n_s_new_state_keys);
  Py_VISIT(traverse_module_state->__pyx_kp_s_no_default___reduce___due_to_non);
  Py_VISIT(traverse_module_state->__pyx_n_s_node);
  Py_VISIT(traverse_module_state->__pyx_n_u_normal);
  Py_VISIT(traverse_module_state->__pyx_n_s_old_state_keys);
  Py_VISIT(traverse_module_state->__pyx_n_u_overlap);
  Py_VISIT(traverse_module_state->__pyx_n_s_particle);
  Py_VISIT(traverse_module_state->__pyx_n_u_particle);
  Py_VISIT(traverse_module_state->__pyx_n_s_particles_by_id);
  Py_VISIT(traverse_module_state->__pyx_n_s_performance);
  Py_VISIT(traverse_module_state->__pyx_n_s_pickle);
  Py_VISIT(traverse_module_state->__pyx_n_u_position);
  Py_VISIT(traverse_module_state->__pyx_n_u_power);
  Py_VISIT(traverse_module_state->__pyx_n_s_purge);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_PickleError);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_checksum);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_result);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_DistanceForceAppl);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_Particle);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_PhysicsSystem);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_vtable);
  Py_VISIT(traverse_module_state->__pyx_n_u_radius);
  Py_VISIT(traverse_module_state->__pyx_n_u_random);
  Py_VISIT(traverse_module_state->__pyx_n_s_range);
  Py_VISIT(traverse_module_state->__pyx_n_s_realtime);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce_ex);
  Py_VISIT(traverse_module_state->__pyx_n_u_resolution);
  Py_VISIT(traverse_module_state->__pyx_n_u_restitution);
  Py_VISIT(traverse_module_state->__pyx_n_s_run);
  Py_VISIT(traverse_module_state->__pyx_n_u_run);
  Py_VISIT(traverse_module_state->__pyx_n_s_run_vector);
  Py_VISIT(traverse_module_state->__pyx_n_s_seed);
  Py_VISIT(traverse_module_state->__pyx_n_s_self);
  Py_VISIT(traverse_module_state->__pyx_n_s_send);
  Py_VISIT(traverse_module_state->__pyx_n_s_setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_spec);
  Py_VISIT(traverse_module_state->__pyx_n_u_speed_of_light);
  Py_VISIT(traverse_module_state->__pyx_kp_s_src_flitter_render_physics_pyx);
  Py_VISIT(traverse_module_state->__pyx_n_s_state);
  Py_VISIT(traverse_module_state->__pyx_n_u_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_state_key);
  Py_VISIT(traverse_module_state->__pyx_n_s_state_prefix);
  Py_VISIT(traverse_module_state->__pyx_n_s_strength);
  Py_VISIT(traverse_module_state->__pyx_n_u_strength);
  Py_VISIT(traverse_module_state->__pyx_kp_s_stringsource);
  Py_VISIT(traverse_module_state->__pyx_n_s_target_fps);
  Py_VISIT(traverse_module_state->__pyx_n_s_test);
  Py_VISIT(traverse_module_state->__pyx_n_s_throw);
  Py_VISIT(traverse_module_state->__pyx_n_s_time);
  Py_VISIT(traverse_module_state->__pyx_n_u_time);
  Py_VISIT(traverse_module_state->__pyx_n_s_time_vector);
  Py_VISIT(traverse_module_state->__pyx_n_u_to);
  Py_VISIT(traverse_module_state->__pyx_n_s_to_thread);
  Py_VISIT(traverse_module_state->__pyx_n_s_top_group);
  Py_VISIT(traverse_module_state->__pyx_n_s_trace);
  Py_VISIT(traverse_module_state->__pyx_n_s_update);
  Py_VISIT(traverse_module_state->__pyx_n_s_use_setstate);
  Py_VISIT(traverse_module_state->__pyx_n_u_velocity);
  Py_VISIT(traverse_module_state->__pyx_n_s_zero);
  Py_VISIT(traverse_module_state->__pyx_int_0);
  Py_VISIT(traverse_module_state->__pyx_int_1);
  Py_VISIT(traverse_module_state->__pyx_int_64470655);
  Py_VISIT(traverse_module_state->__pyx_int_136296653);
  Py_VISIT(traverse_module_state->__pyx_int_142976427);
  Py_VISIT(traverse_module_state->__pyx_int_158432715);
  Py_VISIT(traverse_module_state->__pyx_int_165385471);
  Py_VISIT(traverse_module_state->__pyx_int_191296081);
  Py_VISIT(traverse_module_state->__pyx_int_210842809);
  Py_VISIT(traverse_module_state->__pyx_int_211912363);
  Py_VISIT(traverse_module_state->__pyx_int_253013475);
  Py_VISIT(traverse_module_state->__pyx_codeobj_);
  Py_VISIT(traverse_module_state->__pyx_tuple__4);
  Py_VISIT(traverse_module_state->__pyx_tuple__6);
  Py_VISIT(traverse_module_state->__pyx_tuple__7);
  Py_VISIT(traverse_module_state->__pyx_tuple__10);
  Py_VISIT(traverse_module_state->__pyx_tuple__12);
  Py_VISIT(traverse_module_state->__pyx_tuple__14);
  Py_VISIT(traverse_module_state->__pyx_tuple__43);
  Py_VISIT(traverse_module_state->__pyx_tuple__44);
  Py_VISIT(traverse_module_state->__pyx_tuple__48);
  Py_VISIT(traverse_module_state->__pyx_codeobj__2);
  Py_VISIT(traverse_module_state->__pyx_codeobj__3);
  Py_VISIT(traverse_module_state->__pyx_codeobj__11);
  Py_VISIT(traverse_module_state->__pyx_codeobj__13);
  Py_VISIT(traverse_module_state->__pyx_codeobj__15);
  Py_VISIT(traverse_module_state->__pyx_codeobj__16);
  Py_VISIT(traverse_module_state->__pyx_codeobj__17);
  Py_VISIT(traverse_module_state->__pyx_codeobj__18);
  Py_VISIT(traverse_module_state->__pyx_codeobj__19);
  Py_VISIT(traverse_module_state->__pyx_codeobj__20);
  Py_VISIT(traverse_module_state->__pyx_codeobj__21);
  Py_VISIT(traverse_module_state->__pyx_codeobj__22);
  Py_VISIT(traverse_module_state->__pyx_codeobj__23);
  Py_VISIT(traverse_module_state->__pyx_codeobj__24);
  Py_VISIT(traverse_module_state->__pyx_codeobj__25);
  Py_VISIT(traverse_module_state->__pyx_codeobj__26);
  Py_VISIT(traverse_module_state->__pyx_codeobj__27);
  Py_VISIT(traverse_module_state->__pyx_codeobj__28);
  Py_VISIT(traverse_module_state->__pyx_codeobj__29);
  Py_VISIT(traverse_module_state->__pyx_codeobj__30);
  Py_VISIT(traverse_module_state->__pyx_codeobj__31);
  Py_VISIT(traverse_module_state->__pyx_codeobj__32);
  Py_VISIT(traverse_module_state->__pyx_codeobj__33);
  Py_VISIT(traverse_module_state->__pyx_codeobj__34);
  Py_VISIT(traverse_module_state->__pyx_codeobj__35);
  Py_VISIT(traverse_module_state->__pyx_codeobj__36);
  Py_VISIT(traverse_module_state->__pyx_codeobj__37);
  Py_VISIT(traverse_module_state->__pyx_codeobj__38);
  Py_VISIT(traverse_module_state->__pyx_codeobj__39);
  Py_VISIT(traverse_module_state->__pyx_codeobj__40);
  Py_VISIT(traverse_module_state->__pyx_codeobj__41);
  Py_VISIT(traverse_module_state->__pyx_codeobj__42);
  Py_VISIT(traverse_module_state->__pyx_codeobj__45);
  Py_VISIT(traverse_module_state->__pyx_codeobj__46);
  Py_VISIT(traverse_module_state->__pyx_codeobj__47);
  Py_VISIT(traverse_module_state->__pyx_codeobj__49);
  Py_VISIT(traverse_module_state->__pyx_codeobj__50);
  Py_VISIT(traverse_module_state->__pyx_codeobj__51);
  return 0;
}
#endif
/* #### Code section: module_state_defines ### */
#define __pyx_d __pyx_mstate_global->__pyx_d
#define __pyx_b __pyx_mstate_global->__pyx_b
#define __pyx_cython_runtime __pyx_mstate_global->__pyx_cython_runtime
#define __pyx_empty_tuple __pyx_mstate_global->__pyx_empty_tuple
#define __pyx_empty_bytes __pyx_mstate_global->__pyx_empty_bytes
#define __pyx_empty_unicode __pyx_mstate_global->__pyx_empty_unicode
#ifdef __Pyx_CyFunction_USED
#define __pyx_CyFunctionType __pyx_mstate_global->__pyx_CyFunctionType
#endif
#ifdef __Pyx_FusedFunction_USED
#define __pyx_FusedFunctionType __pyx_mstate_global->__pyx_FusedFunctionType
#endif
#ifdef __Pyx_Generator_USED
#define __pyx_GeneratorType __pyx_mstate_global->__pyx_GeneratorType
#endif
#ifdef __Pyx_IterableCoroutine_USED
#define __pyx_IterableCoroutineType __pyx_mstate_global->__pyx_IterableCoroutineType
#endif
#ifdef __Pyx_Coroutine_USED
#define __pyx_CoroutineAwaitType __pyx_mstate_global->__pyx_CoroutineAwaitType
#endif
#ifdef __Pyx_Coroutine_USED
#define __pyx_CoroutineType __pyx_mstate_global->__pyx_CoroutineType
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7flitter_5model_Vector __pyx_mstate_global->__pyx_ptype_7flitter_5model_Vector
#define __pyx_ptype_7flitter_5model_Matrix33 __pyx_mstate_global->__pyx_ptype_7flitter_5model_Matrix33
#define __pyx_ptype_7flitter_5model_Matrix44 __pyx_mstate_global->__pyx_ptype_7flitter_5model_Matrix44
#define __pyx_ptype_7flitter_5model_Quaternion __pyx_mstate_global->__pyx_ptype_7flitter_5model_Quaternion
#define __pyx_ptype_7flitter_5model_Node __pyx_mstate_global->__pyx_ptype_7flitter_5model_Node
#define __pyx_ptype_7flitter_5model_StateDict __pyx_mstate_global->__pyx_ptype_7flitter_5model_StateDict
#define __pyx_ptype_7flitter_5model_Context __pyx_mstate_global->__pyx_ptype_7flitter_5model_Context
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7flitter_8language_9functions_uniform __pyx_mstate_global->__pyx_ptype_7flitter_8language_9functions_uniform
#define __pyx_ptype_7flitter_8language_9functions_beta __pyx_mstate_global->__pyx_ptype_7flitter_8language_9functions_beta
#define __pyx_ptype_7flitter_8language_9functions_normal __pyx_mstate_global->__pyx_ptype_7flitter_8language_9functions_normal
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7cpython_4type_type __pyx_mstate_global->__pyx_ptype_7cpython_4type_type
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7cpython_4bool_bool __pyx_mstate_global->__pyx_ptype_7cpython_4bool_bool
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7cpython_7complex_complex __pyx_mstate_global->__pyx_ptype_7cpython_7complex_complex
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#define __pyx_type_7flitter_6render_7physics_PhysicsGroup __pyx_mstate_global->__pyx_type_7flitter_6render_7physics_PhysicsGroup
#define __pyx_type_7flitter_6render_7physics_Particle __pyx_mstate_global->__pyx_type_7flitter_6render_7physics_Particle
#define __pyx_type_7flitter_6render_7physics_Barrier __pyx_mstate_global->__pyx_type_7flitter_6render_7physics_Barrier
#define __pyx_type_7flitter_6render_7physics_ParticleForceApplier __pyx_mstate_global->__pyx_type_7flitter_6render_7physics_ParticleForceApplier
#define __pyx_type_7flitter_6render_7physics_MatrixForceApplier __pyx_mstate_global->__pyx_type_7flitter_6render_7physics_MatrixForceApplier
#define __pyx_type_7flitter_6render_7physics_DistanceForceApplier __pyx_mstate_global->__pyx_type_7flitter_6render_7physics_DistanceForceApplier
#define __pyx_type_7flitter_6render_7physics_DragForceApplier __pyx_mstate_global->__pyx_type_7flitter_6render_7physics_DragForceApplier
#define __pyx_type_7flitter_6render_7physics_BuoyancyForceApplier __pyx_mstate_global->__pyx_type_7flitter_6render_7physics_BuoyancyForceApplier
#define __pyx_type_7flitter_6render_7physics_ConstantForceApplier __pyx_mstate_global->__pyx_type_7flitter_6render_7physics_ConstantForceApplier
#define __pyx_type_7flitter_6render_7physics_RandomForceApplier __pyx_mstate_global->__pyx_type_7flitter_6render_7physics_RandomForceApplier
#define __pyx_type_7flitter_6render_7physics_FieldForceApplier __pyx_mstate_global->__pyx_type_7flitter_6render_7physics_FieldForceApplier
#define __pyx_type_7flitter_6render_7physics_CollisionForceApplier __pyx_mstate_global->__pyx_type_7flitter_6render_7physics_CollisionForceApplier
#define __pyx_type_7flitter_6render_7physics_GravityForceApplier __pyx_mstate_global->__pyx_type_7flitter_6render_7physics_GravityForceApplier
#define __pyx_type_7flitter_6render_7physics_ElectrostaticForceApplier __pyx_mstate_global->__pyx_type_7flitter_6render_7physics_ElectrostaticForceApplier
#define __pyx_type_7flitter_6render_7physics_AdhesionForceApplier __pyx_mstate_global->__pyx_type_7flitter_6render_7physics_AdhesionForceApplier
#define __pyx_type_7flitter_6render_7physics_PhysicsSystem __pyx_mstate_global->__pyx_type_7flitter_6render_7physics_PhysicsSystem
#define __pyx_type_7flitter_6render_7physics___pyx_scope_struct__destroy __pyx_mstate_global->__pyx_type_7flitter_6render_7physics___pyx_scope_struct__destroy
#define __pyx_type_7flitter_6render_7physics___pyx_scope_struct_1_purge __pyx_mstate_global->__pyx_type_7flitter_6render_7physics___pyx_scope_struct_1_purge
#define __pyx_type_7flitter_6render_7physics___pyx_scope_struct_2_update __pyx_mstate_global->__pyx_type_7flitter_6render_7physics___pyx_scope_struct_2_update
#endif
#define __pyx_ptype_7flitter_6render_7physics_PhysicsGroup __pyx_mstate_global->__pyx_ptype_7flitter_6render_7physics_PhysicsGroup
#define __pyx_ptype_7flitter_6render_7physics_Particle __pyx_mstate_global->__pyx_ptype_7flitter_6render_7physics_Particle
#define __pyx_ptype_7flitter_6render_7physics_Barrier __pyx_mstate_global->__pyx_ptype_7flitter_6render_7physics_Barrier
#define __pyx_ptype_7flitter_6render_7physics_ParticleForceApplier __pyx_mstate_global->__pyx_ptype_7flitter_6render_7physics_ParticleForceApplier
#define __pyx_ptype_7flitter_6render_7physics_MatrixForceApplier __pyx_mstate_global->__pyx_ptype_7flitter_6render_7physics_MatrixForceApplier
#define __pyx_ptype_7flitter_6render_7physics_DistanceForceApplier __pyx_mstate_global->__pyx_ptype_7flitter_6render_7physics_DistanceForceApplier
#define __pyx_ptype_7flitter_6render_7physics_DragForceApplier __pyx_mstate_global->__pyx_ptype_7flitter_6render_7physics_DragForceApplier
#define __pyx_ptype_7flitter_6render_7physics_BuoyancyForceApplier __pyx_mstate_global->__pyx_ptype_7flitter_6render_7physics_BuoyancyForceApplier
#define __pyx_ptype_7flitter_6render_7physics_ConstantForceApplier __pyx_mstate_global->__pyx_ptype_7flitter_6render_7physics_ConstantForceApplier
#define __pyx_ptype_7flitter_6render_7physics_RandomForceApplier __pyx_mstate_global->__pyx_ptype_7flitter_6render_7physics_RandomForceApplier
#define __pyx_ptype_7flitter_6render_7physics_FieldForceApplier __pyx_mstate_global->__pyx_ptype_7flitter_6render_7physics_FieldForceApplier
#define __pyx_ptype_7flitter_6render_7physics_CollisionForceApplier __pyx_mstate_global->__pyx_ptype_7flitter_6render_7physics_CollisionForceApplier
#define __pyx_ptype_7flitter_6render_7physics_GravityForceApplier __pyx_mstate_global->__pyx_ptype_7flitter_6render_7physics_GravityForceApplier
#define __pyx_ptype_7flitter_6render_7physics_ElectrostaticForceApplier __pyx_mstate_global->__pyx_ptype_7flitter_6render_7physics_ElectrostaticForceApplier
#define __pyx_ptype_7flitter_6render_7physics_AdhesionForceApplier __pyx_mstate_global->__pyx_ptype_7flitter_6render_7physics_AdhesionForceApplier
#define __pyx_ptype_7flitter_6render_7physics_PhysicsSystem __pyx_mstate_global->__pyx_ptype_7flitter_6render_7physics_PhysicsSystem
#define __pyx_ptype_7flitter_6render_7physics___pyx_scope_struct__destroy __pyx_mstate_global->__pyx_ptype_7flitter_6render_7physics___pyx_scope_struct__destroy
#define __pyx_ptype_7flitter_6render_7physics___pyx_scope_struct_1_purge __pyx_mstate_global->__pyx_ptype_7flitter_6render_7physics___pyx_scope_struct_1_purge
#define __pyx_ptype_7flitter_6render_7physics___pyx_scope_struct_2_update __pyx_mstate_global->__pyx_ptype_7flitter_6render_7physics___pyx_scope_struct_2_update
#define __pyx_n_s_AdhesionForceApplier __pyx_mstate_global->__pyx_n_s_AdhesionForceApplier
#define __pyx_n_s_AdhesionForceApplier___reduce_cy __pyx_mstate_global->__pyx_n_s_AdhesionForceApplier___reduce_cy
#define __pyx_n_s_AdhesionForceApplier___setstate __pyx_mstate_global->__pyx_n_s_AdhesionForceApplier___setstate
#define __pyx_n_s_Barrier __pyx_mstate_global->__pyx_n_s_Barrier
#define __pyx_n_s_Barrier___reduce_cython __pyx_mstate_global->__pyx_n_s_Barrier___reduce_cython
#define __pyx_n_s_Barrier___setstate_cython __pyx_mstate_global->__pyx_n_s_Barrier___setstate_cython
#define __pyx_n_s_BuoyancyForceApplier __pyx_mstate_global->__pyx_n_s_BuoyancyForceApplier
#define __pyx_n_s_BuoyancyForceApplier___reduce_cy __pyx_mstate_global->__pyx_n_s_BuoyancyForceApplier___reduce_cy
#define __pyx_n_s_BuoyancyForceApplier___setstate __pyx_mstate_global->__pyx_n_s_BuoyancyForceApplier___setstate
#define __pyx_n_s_CollisionForceApplier __pyx_mstate_global->__pyx_n_s_CollisionForceApplier
#define __pyx_n_s_CollisionForceApplier___reduce_c __pyx_mstate_global->__pyx_n_s_CollisionForceApplier___reduce_c
#define __pyx_n_s_CollisionForceApplier___setstate __pyx_mstate_global->__pyx_n_s_CollisionForceApplier___setstate
#define __pyx_n_s_ConstantForceApplier __pyx_mstate_global->__pyx_n_s_ConstantForceApplier
#define __pyx_n_s_ConstantForceApplier___reduce_cy __pyx_mstate_global->__pyx_n_s_ConstantForceApplier___reduce_cy
#define __pyx_n_s_ConstantForceApplier___setstate __pyx_mstate_global->__pyx_n_s_ConstantForceApplier___setstate
#define __pyx_n_s_DistanceForceApplier __pyx_mstate_global->__pyx_n_s_DistanceForceApplier
#define __pyx_n_s_DistanceForceApplier___reduce_cy __pyx_mstate_global->__pyx_n_s_DistanceForceApplier___reduce_cy
#define __pyx_n_s_DistanceForceApplier___setstate __pyx_mstate_global->__pyx_n_s_DistanceForceApplier___setstate
#define __pyx_n_s_DragForceApplier __pyx_mstate_global->__pyx_n_s_DragForceApplier
#define __pyx_n_s_DragForceApplier___reduce_cython __pyx_mstate_global->__pyx_n_s_DragForceApplier___reduce_cython
#define __pyx_n_s_DragForceApplier___setstate_cyth __pyx_mstate_global->__pyx_n_s_DragForceApplier___setstate_cyth
#define __pyx_n_s_ElectrostaticForceApplier __pyx_mstate_global->__pyx_n_s_ElectrostaticForceApplier
#define __pyx_n_s_ElectrostaticForceApplier___redu __pyx_mstate_global->__pyx_n_s_ElectrostaticForceApplier___redu
#define __pyx_n_s_ElectrostaticForceApplier___sets __pyx_mstate_global->__pyx_n_s_ElectrostaticForceApplier___sets
#define __pyx_n_s_FieldForceApplier __pyx_mstate_global->__pyx_n_s_FieldForceApplier
#define __pyx_n_s_FieldForceApplier___reduce_cytho __pyx_mstate_global->__pyx_n_s_FieldForceApplier___reduce_cytho
#define __pyx_n_s_FieldForceApplier___setstate_cyt __pyx_mstate_global->__pyx_n_s_FieldForceApplier___setstate_cyt
#define __pyx_n_s_GravityForceApplier __pyx_mstate_global->__pyx_n_s_GravityForceApplier
#define __pyx_n_s_GravityForceApplier___reduce_cyt __pyx_mstate_global->__pyx_n_s_GravityForceApplier___reduce_cyt
#define __pyx_n_s_GravityForceApplier___setstate_c __pyx_mstate_global->__pyx_n_s_GravityForceApplier___setstate_c
#define __pyx_kp_u_Group_with_particles_forces_and __pyx_mstate_global->__pyx_kp_u_Group_with_particles_forces_and
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3
#define __pyx_n_s_MatrixForceApplier __pyx_mstate_global->__pyx_n_s_MatrixForceApplier
#define __pyx_n_s_MatrixForceApplier___reduce_cyth __pyx_mstate_global->__pyx_n_s_MatrixForceApplier___reduce_cyth
#define __pyx_n_s_MatrixForceApplier___setstate_cy __pyx_mstate_global->__pyx_n_s_MatrixForceApplier___setstate_cy
#define __pyx_kp_u_New_D_physics_r_with_particles_f __pyx_mstate_global->__pyx_kp_u_New_D_physics_r_with_particles_f
#define __pyx_kp_u_New_D_physics_r_with_particles_o __pyx_mstate_global->__pyx_kp_u_New_D_physics_r_with_particles_o
#define __pyx_n_s_NotImplementedError __pyx_mstate_global->__pyx_n_s_NotImplementedError
#define __pyx_n_s_Particle __pyx_mstate_global->__pyx_n_s_Particle
#define __pyx_n_s_ParticleForceApplier __pyx_mstate_global->__pyx_n_s_ParticleForceApplier
#define __pyx_n_s_ParticleForceApplier___reduce_cy __pyx_mstate_global->__pyx_n_s_ParticleForceApplier___reduce_cy
#define __pyx_n_s_ParticleForceApplier___setstate __pyx_mstate_global->__pyx_n_s_ParticleForceApplier___setstate
#define __pyx_n_s_Particle___reduce_cython __pyx_mstate_global->__pyx_n_s_Particle___reduce_cython
#define __pyx_n_s_Particle___setstate_cython __pyx_mstate_global->__pyx_n_s_Particle___setstate_cython
#define __pyx_n_s_PhysicsGroup __pyx_mstate_global->__pyx_n_s_PhysicsGroup
#define __pyx_n_s_PhysicsGroup___reduce_cython __pyx_mstate_global->__pyx_n_s_PhysicsGroup___reduce_cython
#define __pyx_n_s_PhysicsGroup___setstate_cython __pyx_mstate_global->__pyx_n_s_PhysicsGroup___setstate_cython
#define __pyx_n_s_PhysicsSystem __pyx_mstate_global->__pyx_n_s_PhysicsSystem
#define __pyx_n_s_PhysicsSystem___reduce_cython __pyx_mstate_global->__pyx_n_s_PhysicsSystem___reduce_cython
#define __pyx_n_s_PhysicsSystem___setstate_cython __pyx_mstate_global->__pyx_n_s_PhysicsSystem___setstate_cython
#define __pyx_n_s_PhysicsSystem_calculate __pyx_mstate_global->__pyx_n_s_PhysicsSystem_calculate
#define __pyx_n_s_PhysicsSystem_destroy __pyx_mstate_global->__pyx_n_s_PhysicsSystem_destroy
#define __pyx_n_s_PhysicsSystem_purge __pyx_mstate_global->__pyx_n_s_PhysicsSystem_purge
#define __pyx_n_s_PhysicsSystem_update __pyx_mstate_global->__pyx_n_s_PhysicsSystem_update
#define __pyx_n_s_PickleError __pyx_mstate_global->__pyx_n_s_PickleError
#define __pyx_n_s_RandomForceApplier __pyx_mstate_global->__pyx_n_s_RandomForceApplier
#define __pyx_n_s_RandomForceApplier___reduce_cyth __pyx_mstate_global->__pyx_n_s_RandomForceApplier___reduce_cyth
#define __pyx_n_s_RandomForceApplier___setstate_cy __pyx_mstate_global->__pyx_n_s_RandomForceApplier___setstate_cy
#define __pyx_kp_u_Reset_physics_system_r_for_run __pyx_mstate_global->__pyx_kp_u_Reset_physics_system_r_for_run
#define __pyx_n_s_TypeError __pyx_mstate_global->__pyx_n_s_TypeError
#define __pyx_kp_u__5 __pyx_mstate_global->__pyx_kp_u__5
#define __pyx_n_s__52 __pyx_mstate_global->__pyx_n_s__52
#define __pyx_n_s__8 __pyx_mstate_global->__pyx_n_s__8
#define __pyx_n_s__9 __pyx_mstate_global->__pyx_n_s__9
#define __pyx_n_u_acceleration __pyx_mstate_global->__pyx_n_u_acceleration
#define __pyx_n_u_adhesion __pyx_mstate_global->__pyx_n_u_adhesion
#define __pyx_n_u_anchor __pyx_mstate_global->__pyx_n_u_anchor
#define __pyx_n_s_args __pyx_mstate_global->__pyx_n_s_args
#define __pyx_n_s_asyncio __pyx_mstate_global->__pyx_n_s_asyncio
#define __pyx_n_s_asyncio_coroutines __pyx_mstate_global->__pyx_n_s_asyncio_coroutines
#define __pyx_n_s_asyncio_tasks __pyx_mstate_global->__pyx_n_s_asyncio_tasks
#define __pyx_n_s_await __pyx_mstate_global->__pyx_n_s_await
#define __pyx_n_u_barrier __pyx_mstate_global->__pyx_n_u_barrier
#define __pyx_n_u_buoyancy __pyx_mstate_global->__pyx_n_u_buoyancy
#define __pyx_n_s_calculate __pyx_mstate_global->__pyx_n_s_calculate
#define __pyx_n_u_charge __pyx_mstate_global->__pyx_n_u_charge
#define __pyx_n_s_cline_in_traceback __pyx_mstate_global->__pyx_n_s_cline_in_traceback
#define __pyx_n_s_clock __pyx_mstate_global->__pyx_n_s_clock
#define __pyx_n_u_clock __pyx_mstate_global->__pyx_n_u_clock
#define __pyx_n_s_close __pyx_mstate_global->__pyx_n_s_close
#define __pyx_n_u_collision __pyx_mstate_global->__pyx_n_u_collision
#define __pyx_n_u_constant __pyx_mstate_global->__pyx_n_u_constant
#define __pyx_n_s_debug __pyx_mstate_global->__pyx_n_s_debug
#define __pyx_n_u_density __pyx_mstate_global->__pyx_n_u_density
#define __pyx_n_s_destroy __pyx_mstate_global->__pyx_n_s_destroy
#define __pyx_n_s_dict __pyx_mstate_global->__pyx_n_s_dict
#define __pyx_n_s_dict_2 __pyx_mstate_global->__pyx_n_s_dict_2
#define __pyx_n_u_dimensions __pyx_mstate_global->__pyx_n_u_dimensions
#define __pyx_n_u_direction __pyx_mstate_global->__pyx_n_u_direction
#define __pyx_kp_u_disable __pyx_mstate_global->__pyx_kp_u_disable
#define __pyx_n_u_distance __pyx_mstate_global->__pyx_n_u_distance
#define __pyx_n_u_drag __pyx_mstate_global->__pyx_n_u_drag
#define __pyx_n_u_ease __pyx_mstate_global->__pyx_n_u_ease
#define __pyx_n_u_electrostatic __pyx_mstate_global->__pyx_n_u_electrostatic
#define __pyx_kp_u_enable __pyx_mstate_global->__pyx_kp_u_enable
#define __pyx_n_s_engine __pyx_mstate_global->__pyx_n_s_engine
#define __pyx_n_s_enumerate __pyx_mstate_global->__pyx_n_s_enumerate
#define __pyx_n_s_extra __pyx_mstate_global->__pyx_n_s_extra
#define __pyx_n_u_field __pyx_mstate_global->__pyx_n_u_field
#define __pyx_n_u_fixed __pyx_mstate_global->__pyx_n_u_fixed
#define __pyx_n_s_flitter_render_physics __pyx_mstate_global->__pyx_n_s_flitter_render_physics
#define __pyx_n_u_flow __pyx_mstate_global->__pyx_n_u_flow
#define __pyx_n_u_force __pyx_mstate_global->__pyx_n_u_force
#define __pyx_n_u_from __pyx_mstate_global->__pyx_n_u_from
#define __pyx_kp_u_gc __pyx_mstate_global->__pyx_kp_u_gc
#define __pyx_n_s_getstate __pyx_mstate_global->__pyx_n_s_getstate
#define __pyx_n_u_gravity __pyx_mstate_global->__pyx_n_u_gravity
#define __pyx_n_s_group __pyx_mstate_global->__pyx_n_s_group
#define __pyx_n_u_group __pyx_mstate_global->__pyx_n_u_group
#define __pyx_n_s_groups __pyx_mstate_global->__pyx_n_s_groups
#define __pyx_n_s_i __pyx_mstate_global->__pyx_n_s_i
#define __pyx_n_u_id __pyx_mstate_global->__pyx_n_u_id
#define __pyx_n_s_import __pyx_mstate_global->__pyx_n_s_import
#define __pyx_n_s_initializing __pyx_mstate_global->__pyx_n_s_initializing
#define __pyx_n_s_inspect __pyx_mstate_global->__pyx_n_s_inspect
#define __pyx_n_s_is_coroutine __pyx_mstate_global->__pyx_n_s_is_coroutine
#define __pyx_kp_u_isenabled __pyx_mstate_global->__pyx_kp_u_isenabled
#define __pyx_n_s_iteration __pyx_mstate_global->__pyx_n_s_iteration
#define __pyx_n_u_iteration __pyx_mstate_global->__pyx_n_u_iteration
#define __pyx_n_s_iteration_vector __pyx_mstate_global->__pyx_n_s_iteration_vector
#define __pyx_n_s_kwargs __pyx_mstate_global->__pyx_n_s_kwargs
#define __pyx_n_u_last __pyx_mstate_global->__pyx_n_u_last
#define __pyx_n_s_last_run __pyx_mstate_global->__pyx_n_s_last_run
#define __pyx_n_s_last_run_vector __pyx_mstate_global->__pyx_n_s_last_run_vector
#define __pyx_n_s_last_time __pyx_mstate_global->__pyx_n_s_last_time
#define __pyx_n_s_logger __pyx_mstate_global->__pyx_n_s_logger
#define __pyx_n_s_loguru __pyx_mstate_global->__pyx_n_s_loguru
#define __pyx_n_s_main __pyx_mstate_global->__pyx_n_s_main
#define __pyx_n_u_mass __pyx_mstate_global->__pyx_n_u_mass
#define __pyx_n_u_max __pyx_mstate_global->__pyx_n_u_max
#define __pyx_n_u_max_distance __pyx_mstate_global->__pyx_n_u_max_distance
#define __pyx_n_u_maximum __pyx_mstate_global->__pyx_n_u_maximum
#define __pyx_n_u_min __pyx_mstate_global->__pyx_n_u_min
#define __pyx_n_u_minimum __pyx_mstate_global->__pyx_n_u_minimum
#define __pyx_n_s_name __pyx_mstate_global->__pyx_n_s_name
#define __pyx_n_s_name_patch __pyx_mstate_global->__pyx_n_s_name_patch
#define __pyx_n_s_new __pyx_mstate_global->__pyx_n_s_new
#define __pyx_n_s_new_state_keys __pyx_mstate_global->__pyx_n_s_new_state_keys
#define __pyx_kp_s_no_default___reduce___due_to_non __pyx_mstate_global->__pyx_kp_s_no_default___reduce___due_to_non
#define __pyx_n_s_node __pyx_mstate_global->__pyx_n_s_node
#define __pyx_n_u_normal __pyx_mstate_global->__pyx_n_u_normal
#define __pyx_n_s_old_state_keys __pyx_mstate_global->__pyx_n_s_old_state_keys
#define __pyx_n_u_overlap __pyx_mstate_global->__pyx_n_u_overlap
#define __pyx_n_s_particle __pyx_mstate_global->__pyx_n_s_particle
#define __pyx_n_u_particle __pyx_mstate_global->__pyx_n_u_particle
#define __pyx_n_s_particles_by_id __pyx_mstate_global->__pyx_n_s_particles_by_id
#define __pyx_n_s_performance __pyx_mstate_global->__pyx_n_s_performance
#define __pyx_n_s_pickle __pyx_mstate_global->__pyx_n_s_pickle
#define __pyx_n_u_position __pyx_mstate_global->__pyx_n_u_position
#define __pyx_n_u_power __pyx_mstate_global->__pyx_n_u_power
#define __pyx_n_s_purge __pyx_mstate_global->__pyx_n_s_purge
#define __pyx_n_s_pyx_PickleError __pyx_mstate_global->__pyx_n_s_pyx_PickleError
#define __pyx_n_s_pyx_checksum __pyx_mstate_global->__pyx_n_s_pyx_checksum
#define __pyx_n_s_pyx_result __pyx_mstate_global->__pyx_n_s_pyx_result
#define __pyx_n_s_pyx_state __pyx_mstate_global->__pyx_n_s_pyx_state
#define __pyx_n_s_pyx_type __pyx_mstate_global->__pyx_n_s_pyx_type
#define __pyx_n_s_pyx_unpickle_DistanceForceAppl __pyx_mstate_global->__pyx_n_s_pyx_unpickle_DistanceForceAppl
#define __pyx_n_s_pyx_unpickle_Particle __pyx_mstate_global->__pyx_n_s_pyx_unpickle_Particle
#define __pyx_n_s_pyx_unpickle_PhysicsSystem __pyx_mstate_global->__pyx_n_s_pyx_unpickle_PhysicsSystem
#define __pyx_n_s_pyx_vtable __pyx_mstate_global->__pyx_n_s_pyx_vtable
#define __pyx_n_u_radius __pyx_mstate_global->__pyx_n_u_radius
#define __pyx_n_u_random __pyx_mstate_global->__pyx_n_u_random
#define __pyx_n_s_range __pyx_mstate_global->__pyx_n_s_range
#define __pyx_n_s_realtime __pyx_mstate_global->__pyx_n_s_realtime
#define __pyx_n_s_reduce __pyx_mstate_global->__pyx_n_s_reduce
#define __pyx_n_s_reduce_cython __pyx_mstate_global->__pyx_n_s_reduce_cython
#define __pyx_n_s_reduce_ex __pyx_mstate_global->__pyx_n_s_reduce_ex
#define __pyx_n_u_resolution __pyx_mstate_global->__pyx_n_u_resolution
#define __pyx_n_u_restitution __pyx_mstate_global->__pyx_n_u_restitution
#define __pyx_n_s_run __pyx_mstate_global->__pyx_n_s_run
#define __pyx_n_u_run __pyx_mstate_global->__pyx_n_u_run
#define __pyx_n_s_run_vector __pyx_mstate_global->__pyx_n_s_run_vector
#define __pyx_n_s_seed __pyx_mstate_global->__pyx_n_s_seed
#define __pyx_n_s_self __pyx_mstate_global->__pyx_n_s_self
#define __pyx_n_s_send __pyx_mstate_global->__pyx_n_s_send
#define __pyx_n_s_setstate __pyx_mstate_global->__pyx_n_s_setstate
#define __pyx_n_s_setstate_cython __pyx_mstate_global->__pyx_n_s_setstate_cython
#define __pyx_n_s_spec __pyx_mstate_global->__pyx_n_s_spec
#define __pyx_n_u_speed_of_light __pyx_mstate_global->__pyx_n_u_speed_of_light
#define __pyx_kp_s_src_flitter_render_physics_pyx __pyx_mstate_global->__pyx_kp_s_src_flitter_render_physics_pyx
#define __pyx_n_s_state __pyx_mstate_global->__pyx_n_s_state
#define __pyx_n_u_state __pyx_mstate_global->__pyx_n_u_state
#define __pyx_n_s_state_key __pyx_mstate_global->__pyx_n_s_state_key
#define __pyx_n_s_state_prefix __pyx_mstate_global->__pyx_n_s_state_prefix
#define __pyx_n_s_strength __pyx_mstate_global->__pyx_n_s_strength
#define __pyx_n_u_strength __pyx_mstate_global->__pyx_n_u_strength
#define __pyx_kp_s_stringsource __pyx_mstate_global->__pyx_kp_s_stringsource
#define __pyx_n_s_target_fps __pyx_mstate_global->__pyx_n_s_target_fps
#define __pyx_n_s_test __pyx_mstate_global->__pyx_n_s_test
#define __pyx_n_s_throw __pyx_mstate_global->__pyx_n_s_throw
#define __pyx_n_s_time __pyx_mstate_global->__pyx_n_s_time
#define __pyx_n_u_time __pyx_mstate_global->__pyx_n_u_time
#define __pyx_n_s_time_vector __pyx_mstate_global->__pyx_n_s_time_vector
#define __pyx_n_u_to __pyx_mstate_global->__pyx_n_u_to
#define __pyx_n_s_to_thread __pyx_mstate_global->__pyx_n_s_to_thread
#define __pyx_n_s_top_group __pyx_mstate_global->__pyx_n_s_top_group
#define __pyx_n_s_trace __pyx_mstate_global->__pyx_n_s_trace
#define __pyx_n_s_update __pyx_mstate_global->__pyx_n_s_update
#define __pyx_n_s_use_setstate __pyx_mstate_global->__pyx_n_s_use_setstate
#define __pyx_n_u_velocity __pyx_mstate_global->__pyx_n_u_velocity
#define __pyx_n_s_zero __pyx_mstate_global->__pyx_n_s_zero
#define __pyx_int_0 __pyx_mstate_global->__pyx_int_0
#define __pyx_int_1 __pyx_mstate_global->__pyx_int_1
#define __pyx_int_64470655 __pyx_mstate_global->__pyx_int_64470655
#define __pyx_int_136296653 __pyx_mstate_global->__pyx_int_136296653
#define __pyx_int_142976427 __pyx_mstate_global->__pyx_int_142976427
#define __pyx_int_158432715 __pyx_mstate_global->__pyx_int_158432715
#define __pyx_int_165385471 __pyx_mstate_global->__pyx_int_165385471
#define __pyx_int_191296081 __pyx_mstate_global->__pyx_int_191296081
#define __pyx_int_210842809 __pyx_mstate_global->__pyx_int_210842809
#define __pyx_int_211912363 __pyx_mstate_global->__pyx_int_211912363
#define __pyx_int_253013475 __pyx_mstate_global->__pyx_int_253013475
#define __pyx_codeobj_ __pyx_mstate_global->__pyx_codeobj_
#define __pyx_tuple__4 __pyx_mstate_global->__pyx_tuple__4
#define __pyx_tuple__6 __pyx_mstate_global->__pyx_tuple__6
#define __pyx_tuple__7 __pyx_mstate_global->__pyx_tuple__7
#define __pyx_tuple__10 __pyx_mstate_global->__pyx_tuple__10
#define __pyx_tuple__12 __pyx_mstate_global->__pyx_tuple__12
#define __pyx_tuple__14 __pyx_mstate_global->__pyx_tuple__14
#define __pyx_tuple__43 __pyx_mstate_global->__pyx_tuple__43
#define __pyx_tuple__44 __pyx_mstate_global->__pyx_tuple__44
#define __pyx_tuple__48 __pyx_mstate_global->__pyx_tuple__48
#define __pyx_codeobj__2 __pyx_mstate_global->__pyx_codeobj__2
#define __pyx_codeobj__3 __pyx_mstate_global->__pyx_codeobj__3
#define __pyx_codeobj__11 __pyx_mstate_global->__pyx_codeobj__11
#define __pyx_codeobj__13 __pyx_mstate_global->__pyx_codeobj__13
#define __pyx_codeobj__15 __pyx_mstate_global->__pyx_codeobj__15
#define __pyx_codeobj__16 __pyx_mstate_global->__pyx_codeobj__16
#define __pyx_codeobj__17 __pyx_mstate_global->__pyx_codeobj__17
#define __pyx_codeobj__18 __pyx_mstate_global->__pyx_codeobj__18
#define __pyx_codeobj__19 __pyx_mstate_global->__pyx_codeobj__19
#define __pyx_codeobj__20 __pyx_mstate_global->__pyx_codeobj__20
#define __pyx_codeobj__21 __pyx_mstate_global->__pyx_codeobj__21
#define __pyx_codeobj__22 __pyx_mstate_global->__pyx_codeobj__22
#define __pyx_codeobj__23 __pyx_mstate_global->__pyx_codeobj__23
#define __pyx_codeobj__24 __pyx_mstate_global->__pyx_codeobj__24
#define __pyx_codeobj__25 __pyx_mstate_global->__pyx_codeobj__25
#define __pyx_codeobj__26 __pyx_mstate_global->__pyx_codeobj__26
#define __pyx_codeobj__27 __pyx_mstate_global->__pyx_codeobj__27
#define __pyx_codeobj__28 __pyx_mstate_global->__pyx_codeobj__28
#define __pyx_codeobj__29 __pyx_mstate_global->__pyx_codeobj__29
#define __pyx_codeobj__30 __pyx_mstate_global->__pyx_codeobj__30
#define __pyx_codeobj__31 __pyx_mstate_global->__pyx_codeobj__31
#define __pyx_codeobj__32 __pyx_mstate_global->__pyx_codeobj__32
#define __pyx_codeobj__33 __pyx_mstate_global->__pyx_codeobj__33
#define __pyx_codeobj__34 __pyx_mstate_global->__pyx_codeobj__34
#define __pyx_codeobj__35 __pyx_mstate_global->__pyx_codeobj__35
#define __pyx_codeobj__36 __pyx_mstate_global->__pyx_codeobj__36
#define __pyx_codeobj__37 __pyx_mstate_global->__pyx_codeobj__37
#define __pyx_codeobj__38 __pyx_mstate_global->__pyx_codeobj__38
#define __pyx_codeobj__39 __pyx_mstate_global->__pyx_codeobj__39
#define __pyx_codeobj__40 __pyx_mstate_global->__pyx_codeobj__40
#define __pyx_codeobj__41 __pyx_mstate_global->__pyx_codeobj__41
#define __pyx_codeobj__42 __pyx_mstate_global->__pyx_codeobj__42
#define __pyx_codeobj__45 __pyx_mstate_global->__pyx_codeobj__45
#define __pyx_codeobj__46 __pyx_mstate_global->__pyx_codeobj__46
#define __pyx_codeobj__47 __pyx_mstate_global->__pyx_codeobj__47
#define __pyx_codeobj__49 __pyx_mstate_global->__pyx_codeobj__49
#define __pyx_codeobj__50 __pyx_mstate_global->__pyx_codeobj__50
#define __pyx_codeobj__51 __pyx_mstate_global->__pyx_codeobj__51
/* #### Code section: module_code ### */

/* "model.pxd":12
 * cdef uint64_t HASH_START
 * 
 * cdef inline uint64_t HASH_UPDATE(uint64_t _hash, uint64_t y) noexcept:             # <<<<<<<<<<<<<<
 *     _hash ^= y
 *     _hash += <uint64_t>(0x9e3779b97f4a7c15)
 */

static CYTHON_INLINE uint64_t __pyx_f_7flitter_5model_HASH_UPDATE(uint64_t __pyx_v__hash, uint64_t __pyx_v_y) {
  uint64_t __pyx_r;

  /* "model.pxd":13
 * 
 * cdef inline uint64_t HASH_UPDATE(uint64_t _hash, uint64_t y) noexcept:
 *     _hash ^= y             # <<<<<<<<<<<<<<
 *     _hash += <uint64_t>(0x9e3779b97f4a7c15)
 *     _hash ^= _hash >> 30
 */
  __pyx_v__hash = (__pyx_v__hash ^ __pyx_v_y);

  /* "model.pxd":14
 * cdef inline uint64_t HASH_UPDATE(uint64_t _hash, uint64_t y) noexcept:
 *     _hash ^= y
 *     _hash += <uint64_t>(0x9e3779b97f4a7c15)             # <<<<<<<<<<<<<<
 *     _hash ^= _hash >> 30
 *     _hash *= <uint64_t>(0xbf58476d1ce4e5b9)
 */
  __pyx_v__hash = (__pyx_v__hash + ((uint64_t)0x9e3779b97f4a7c15));

  /* "model.pxd":15
 *     _hash ^= y
 *     _hash += <uint64_t>(0x9e3779b97f4a7c15)
 *     _hash ^= _hash >> 30             # <<<<<<<<<<<<<<
 *     _hash *= <uint64_t>(0xbf58476d1ce4e5b9)
 *     _hash ^= _hash >> 27
 */
  __pyx_v__hash = (__pyx_v__hash ^ (__pyx_v__hash >> 30));

  /* "model.pxd":16
 *     _hash += <uint64_t>(0x9e3779b97f4a7c15)
 *     _hash ^= _hash >> 30
 *     _hash *= <uint64_t>(0xbf58476d1ce4e5b9)             # <<<<<<<<<<<<<<
 *     _hash ^= _hash >> 27
 *     _hash *= <uint64_t>(0x94d049bb133111eb)
 */
  __pyx_v__hash = (__pyx_v__hash * ((uint64_t)0xbf58476d1ce4e5b9));

  /* "model.pxd":17
 *     _hash ^= _hash >> 30
 *     _hash *= <uint64_t>(0xbf58476d1ce4e5b9)
 *     _hash ^= _hash >> 27             # <<<<<<<<<<<<<<
 *     _hash *= <uint64_t>(0x94d049bb133111eb)
 *     _hash ^= _hash >> 31
 */
  __pyx_v__hash = (__pyx_v__hash ^ (__pyx_v__hash >> 27));

  /* "model.pxd":18
 *     _hash *= <uint64_t>(0xbf58476d1ce4e5b9)
 *     _hash ^= _hash >> 27
 *     _hash *= <uint64_t>(0x94d049bb133111eb)             # <<<<<<<<<<<<<<
 *     _hash ^= _hash >> 31
 *     return _hash
 */
  __pyx_v__hash = (__pyx_v__hash * ((uint64_t)0x94d049bb133111eb));

  /* "model.pxd":19
 *     _hash ^= _hash >> 27
 *     _hash *= <uint64_t>(0x94d049bb133111eb)
 *     _hash ^= _hash >> 31             # <<<<<<<<<<<<<<
 *     return _hash
 * 
 */
  __pyx_v__hash = (__pyx_v__hash ^ (__pyx_v__hash >> 31));

  /* "model.pxd":20
 *     _hash *= <uint64_t>(0x94d049bb133111eb)
 *     _hash ^= _hash >> 31
 *     return _hash             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v__hash;
  goto __pyx_L0;

  /* "model.pxd":12
 * cdef uint64_t HASH_START
 * 
 * cdef inline uint64_t HASH_UPDATE(uint64_t _hash, uint64_t y) noexcept:             # <<<<<<<<<<<<<<
 *     _hash ^= y
 *     _hash += <uint64_t>(0x9e3779b97f4a7c15)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "model.pxd":24
 * 
 * # FNV-1a hash algorithm [https://en.wikipedia.org/wiki/FowlerNollVo_hash_function#FNV-1a_hash]
 * cdef inline uint64_t HASH_STRING(str value):             # <<<<<<<<<<<<<<
 *     cdef void* data = PyUnicode_DATA(value)
 *     cdef uint64_t i, n=PyUnicode_GET_LENGTH(value), kind=PyUnicode_KIND(value)
 */

static CYTHON_INLINE uint64_t __pyx_f_7flitter_5model_HASH_STRING(PyObject *__pyx_v_value) {
  void *__pyx_v_data;
  uint64_t __pyx_v_i;
  uint64_t __pyx_v_n;
  uint64_t __pyx_v_kind;
  Py_UCS4 __pyx_v_c;
  uint64_t __pyx_v_y;
  uint64_t __pyx_r;
  uint64_t __pyx_t_1;
  uint64_t __pyx_t_2;
  uint64_t __pyx_t_3;

  /* "model.pxd":25
 * # FNV-1a hash algorithm [https://en.wikipedia.org/wiki/FowlerNollVo_hash_function#FNV-1a_hash]
 * cdef inline uint64_t HASH_STRING(str value):
 *     cdef void* data = PyUnicode_DATA(value)             # <<<<<<<<<<<<<<
 *     cdef uint64_t i, n=PyUnicode_GET_LENGTH(value), kind=PyUnicode_KIND(value)
 *     cdef Py_UCS4 c
 */
  __pyx_v_data = PyUnicode_DATA(__pyx_v_value);

  /* "model.pxd":26
 * cdef inline uint64_t HASH_STRING(str value):
 *     cdef void* data = PyUnicode_DATA(value)
 *     cdef uint64_t i, n=PyUnicode_GET_LENGTH(value), kind=PyUnicode_KIND(value)             # <<<<<<<<<<<<<<
 *     cdef Py_UCS4 c
 *     cdef uint64_t y = <uint64_t>(0xcbf29ce484222325)
 */
  __pyx_v_n = PyUnicode_GET_LENGTH(__pyx_v_value);
  __pyx_v_kind = PyUnicode_KIND(__pyx_v_value);

  /* "model.pxd":28
 *     cdef uint64_t i, n=PyUnicode_GET_LENGTH(value), kind=PyUnicode_KIND(value)
 *     cdef Py_UCS4 c
 *     cdef uint64_t y = <uint64_t>(0xcbf29ce484222325)             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         c = PyUnicode_READ(kind, data, i)
 */
  __pyx_v_y = ((uint64_t)0xcbf29ce484222325);

  /* "model.pxd":29
 *     cdef Py_UCS4 c
 *     cdef uint64_t y = <uint64_t>(0xcbf29ce484222325)
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         c = PyUnicode_READ(kind, data, i)
 *         y = (y ^ <uint64_t>c) * <uint64_t>(0x100000001b3)
 */
  __pyx_t_1 = __pyx_v_n;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "model.pxd":30
 *     cdef uint64_t y = <uint64_t>(0xcbf29ce484222325)
 *     for i in range(n):
 *         c = PyUnicode_READ(kind, data, i)             # <<<<<<<<<<<<<<
 *         y = (y ^ <uint64_t>c) * <uint64_t>(0x100000001b3)
 *     return y
 */
    __pyx_v_c = PyUnicode_READ(__pyx_v_kind, __pyx_v_data, __pyx_v_i);

    /* "model.pxd":31
 *     for i in range(n):
 *         c = PyUnicode_READ(kind, data, i)
 *         y = (y ^ <uint64_t>c) * <uint64_t>(0x100000001b3)             # <<<<<<<<<<<<<<
 *     return y
 * 
 */
    __pyx_v_y = ((__pyx_v_y ^ ((uint64_t)__pyx_v_c)) * ((uint64_t)0x100000001b3));
  }

  /* "model.pxd":32
 *         c = PyUnicode_READ(kind, data, i)
 *         y = (y ^ <uint64_t>c) * <uint64_t>(0x100000001b3)
 *     return y             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_y;
  goto __pyx_L0;

  /* "model.pxd":24
 * 
 * # FNV-1a hash algorithm [https://en.wikipedia.org/wiki/FowlerNollVo_hash_function#FNV-1a_hash]
 * cdef inline uint64_t HASH_STRING(str value):             # <<<<<<<<<<<<<<
 *     cdef void* data = PyUnicode_DATA(value)
 *     cdef uint64_t i, n=PyUnicode_GET_LENGTH(value), kind=PyUnicode_KIND(value)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/complex.pxd":19
 * 
 *         @property
 *         cdef inline double real(self) noexcept:             # <<<<<<<<<<<<<<
 *             return self.cval.real
 * 
 */

static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4real_real(PyComplexObject *__pyx_v_self) {
  double __pyx_r;

  /* "cpython/complex.pxd":20
 *         @property
 *         cdef inline double real(self) noexcept:
 *             return self.cval.real             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = __pyx_v_self->cval.real;
  goto __pyx_L0;

  /* "cpython/complex.pxd":19
 * 
 *         @property
 *         cdef inline double real(self) noexcept:             # <<<<<<<<<<<<<<
 *             return self.cval.real
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/complex.pxd":23
 * 
 *         @property
 *         cdef inline double imag(self) noexcept:             # <<<<<<<<<<<<<<
 *             return self.cval.imag
 * 
 */

static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4imag_imag(PyComplexObject *__pyx_v_self) {
  double __pyx_r;

  /* "cpython/complex.pxd":24
 *         @property
 *         cdef inline double imag(self) noexcept:
 *             return self.cval.imag             # <<<<<<<<<<<<<<
 * 
 *     # PyTypeObject PyComplex_Type
 */
  __pyx_r = __pyx_v_self->cval.imag;
  goto __pyx_L0;

  /* "cpython/complex.pxd":23
 * 
 *         @property
 *         cdef inline double imag(self) noexcept:             # <<<<<<<<<<<<<<
 *             return self.cval.imag
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/contextvars.pxd":112
 * 
 * 
 * cdef inline object get_value(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the default value of the context variable,
 */

static CYTHON_INLINE PyObject *__pyx_f_7cpython_11contextvars_get_value(PyObject *__pyx_v_var, struct __pyx_opt_args_7cpython_11contextvars_get_value *__pyx_optional_args) {
  PyObject *__pyx_v_default_value = ((PyObject *)Py_None);
  PyObject *__pyx_v_value;
  PyObject *__pyx_v_pyvalue = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_value", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_default_value = __pyx_optional_args->default_value;
    }
  }

  /* "cpython/contextvars.pxd":117
 *     or None if no such value or default was found.
 *     """
 *     cdef PyObject *value = NULL             # <<<<<<<<<<<<<<
 *     PyContextVar_Get(var, NULL, &value)
 *     if value is NULL:
 */
  __pyx_v_value = NULL;

  /* "cpython/contextvars.pxd":118
 *     """
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, NULL, &value)             # <<<<<<<<<<<<<<
 *     if value is NULL:
 *         # context variable does not have a default
 */
  __pyx_t_1 = PyContextVar_Get(__pyx_v_var, NULL, (&__pyx_v_value)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(2, 118, __pyx_L1_error)

  /* "cpython/contextvars.pxd":119
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, NULL, &value)
 *     if value is NULL:             # <<<<<<<<<<<<<<
 *         # context variable does not have a default
 *         pyvalue = default_value
 */
  __pyx_t_2 = (__pyx_v_value == NULL);
  if (__pyx_t_2) {

    /* "cpython/contextvars.pxd":121
 *     if value is NULL:
 *         # context variable does not have a default
 *         pyvalue = default_value             # <<<<<<<<<<<<<<
 *     else:
 *         # value or default value of context variable
 */
    __Pyx_INCREF(__pyx_v_default_value);
    __pyx_v_pyvalue = __pyx_v_default_value;

    /* "cpython/contextvars.pxd":119
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, NULL, &value)
 *     if value is NULL:             # <<<<<<<<<<<<<<
 *         # context variable does not have a default
 *         pyvalue = default_value
 */
    goto __pyx_L3;
  }

  /* "cpython/contextvars.pxd":124
 *     else:
 *         # value or default value of context variable
 *         pyvalue = <object>value             # <<<<<<<<<<<<<<
 *         Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue
 */
  /*else*/ {
    __pyx_t_3 = ((PyObject *)__pyx_v_value);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_pyvalue = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "cpython/contextvars.pxd":125
 *         # value or default value of context variable
 *         pyvalue = <object>value
 *         Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'             # <<<<<<<<<<<<<<
 *     return pyvalue
 * 
 */
    Py_XDECREF(__pyx_v_value);
  }
  __pyx_L3:;

  /* "cpython/contextvars.pxd":126
 *         pyvalue = <object>value
 *         Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_pyvalue);
  __pyx_r = __pyx_v_pyvalue;
  goto __pyx_L0;

  /* "cpython/contextvars.pxd":112
 * 
 * 
 * cdef inline object get_value(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the default value of the context variable,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cpython.contextvars.get_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pyvalue);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/contextvars.pxd":129
 * 
 * 
 * cdef inline object get_value_no_default(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the provided default value if no such value was found.
 */

static CYTHON_INLINE PyObject *__pyx_f_7cpython_11contextvars_get_value_no_default(PyObject *__pyx_v_var, struct __pyx_opt_args_7cpython_11contextvars_get_value_no_default *__pyx_optional_args) {
  PyObject *__pyx_v_default_value = ((PyObject *)Py_None);
  PyObject *__pyx_v_value;
  PyObject *__pyx_v_pyvalue = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_value_no_default", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_default_value = __pyx_optional_args->default_value;
    }
  }

  /* "cpython/contextvars.pxd":135
 *     Ignores the default value of the context variable, if any.
 *     """
 *     cdef PyObject *value = NULL             # <<<<<<<<<<<<<<
 *     PyContextVar_Get(var, <PyObject*>default_value, &value)
 *     # value of context variable or 'default_value'
 */
  __pyx_v_value = NULL;

  /* "cpython/contextvars.pxd":136
 *     """
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, <PyObject*>default_value, &value)             # <<<<<<<<<<<<<<
 *     # value of context variable or 'default_value'
 *     pyvalue = <object>value
 */
  __pyx_t_1 = PyContextVar_Get(__pyx_v_var, ((PyObject *)__pyx_v_default_value), (&__pyx_v_value)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(2, 136, __pyx_L1_error)

  /* "cpython/contextvars.pxd":138
 *     PyContextVar_Get(var, <PyObject*>default_value, &value)
 *     # value of context variable or 'default_value'
 *     pyvalue = <object>value             # <<<<<<<<<<<<<<
 *     Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue
 */
  __pyx_t_2 = ((PyObject *)__pyx_v_value);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_pyvalue = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "cpython/contextvars.pxd":139
 *     # value of context variable or 'default_value'
 *     pyvalue = <object>value
 *     Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'             # <<<<<<<<<<<<<<
 *     return pyvalue
 */
  Py_XDECREF(__pyx_v_value);

  /* "cpython/contextvars.pxd":140
 *     pyvalue = <object>value
 *     Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_pyvalue);
  __pyx_r = __pyx_v_pyvalue;
  goto __pyx_L0;

  /* "cpython/contextvars.pxd":129
 * 
 * 
 * cdef inline object get_value_no_default(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the provided default value if no such value was found.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cpython.contextvars.get_value_no_default", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pyvalue);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/render/physics.pyx":45
 *     cdef list barriers
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self.particles = []
 *         self.particle_forces = []
 */

/* Python wrapper */
static int __pyx_pw_7flitter_6render_7physics_12PhysicsGroup_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_6render_7physics_12PhysicsGroup_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, __pyx_nargs); return -1;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_VARARGS(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_7flitter_6render_7physics_12PhysicsGroup___cinit__(((struct __pyx_obj_7flitter_6render_7physics_PhysicsGroup *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_6render_7physics_12PhysicsGroup___cinit__(struct __pyx_obj_7flitter_6render_7physics_PhysicsGroup *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 1);

  /* "flitter/render/physics.pyx":46
 * 
 *     def __cinit__(self):
 *         self.particles = []             # <<<<<<<<<<<<<<
 *         self.particle_forces = []
 *         self.matrix_forces = []
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->particles);
  __Pyx_DECREF(__pyx_v_self->particles);
  __pyx_v_self->particles = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/render/physics.pyx":47
 *     def __cinit__(self):
 *         self.particles = []
 *         self.particle_forces = []             # <<<<<<<<<<<<<<
 *         self.matrix_forces = []
 *         self.distance_forces = []
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->particle_forces);
  __Pyx_DECREF(__pyx_v_self->particle_forces);
  __pyx_v_self->particle_forces = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/render/physics.pyx":48
 *         self.particles = []
 *         self.particle_forces = []
 *         self.matrix_forces = []             # <<<<<<<<<<<<<<
 *         self.distance_forces = []
 *         self.barriers = []
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->matrix_forces);
  __Pyx_DECREF(__pyx_v_self->matrix_forces);
  __pyx_v_self->matrix_forces = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/render/physics.pyx":49
 *         self.particle_forces = []
 *         self.matrix_forces = []
 *         self.distance_forces = []             # <<<<<<<<<<<<<<
 *         self.barriers = []
 * 
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->distance_forces);
  __Pyx_DECREF(__pyx_v_self->distance_forces);
  __pyx_v_self->distance_forces = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/render/physics.pyx":50
 *         self.matrix_forces = []
 *         self.distance_forces = []
 *         self.barriers = []             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->barriers);
  __Pyx_DECREF(__pyx_v_self->barriers);
  __pyx_v_self->barriers = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/render/physics.pyx":45
 *     cdef list barriers
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self.particles = []
 *         self.particle_forces = []
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.render.physics.PhysicsGroup.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_6render_7physics_12PhysicsGroup_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_6render_7physics_12PhysicsGroup_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_12PhysicsGroup_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_6render_7physics_12PhysicsGroup_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_6render_7physics_12PhysicsGroup_2__reduce_cython__(((struct __pyx_obj_7flitter_6render_7physics_PhysicsGroup *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_6render_7physics_12PhysicsGroup_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_PhysicsGroup *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.render.physics.PhysicsGroup.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_6render_7physics_12PhysicsGroup_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_6render_7physics_12PhysicsGroup_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_12PhysicsGroup_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_6render_7physics_12PhysicsGroup_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.render.physics.PhysicsGroup.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_6render_7physics_12PhysicsGroup_4__setstate_cython__(((struct __pyx_obj_7flitter_6render_7physics_PhysicsGroup *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_6render_7physics_12PhysicsGroup_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_PhysicsGroup *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.render.physics.PhysicsGroup.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/render/physics.pyx":70
 * 
 *     @cython.profile(False)
 *     cdef void step(self, double speed_of_light, double clock, double delta) noexcept nogil:             # <<<<<<<<<<<<<<
 *         cdef double speed_squared, v, k
 *         cdef int64_t i, n=self.force.length
 */

static void __pyx_f_7flitter_6render_7physics_8Particle_step(struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_self, double __pyx_v_speed_of_light, double __pyx_v_clock, double __pyx_v_delta) {
  double __pyx_v_speed_squared;
  double __pyx_v_v;
  double __pyx_v_k;
  int64_t __pyx_v_i;
  int64_t __pyx_v_n;
  int64_t __pyx_t_1;
  int64_t __pyx_t_2;
  int64_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int64_t __pyx_t_6;
  int64_t __pyx_t_7;

  /* "flitter/render/physics.pyx":72
 *     cdef void step(self, double speed_of_light, double clock, double delta) noexcept nogil:
 *         cdef double speed_squared, v, k
 *         cdef int64_t i, n=self.force.length             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             self.acceleration.numbers[i] = self.velocity.numbers[i]
 */
  __pyx_t_1 = __pyx_v_self->force->length;
  __pyx_v_n = __pyx_t_1;

  /* "flitter/render/physics.pyx":73
 *         cdef double speed_squared, v, k
 *         cdef int64_t i, n=self.force.length
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             self.acceleration.numbers[i] = self.velocity.numbers[i]
 *         if self.mass:
 */
  __pyx_t_1 = __pyx_v_n;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "flitter/render/physics.pyx":74
 *         cdef int64_t i, n=self.force.length
 *         for i in range(n):
 *             self.acceleration.numbers[i] = self.velocity.numbers[i]             # <<<<<<<<<<<<<<
 *         if self.mass:
 *             for i in range(n):
 */
    (__pyx_v_self->acceleration->numbers[__pyx_v_i]) = (__pyx_v_self->velocity->numbers[__pyx_v_i]);
  }

  /* "flitter/render/physics.pyx":75
 *         for i in range(n):
 *             self.acceleration.numbers[i] = self.velocity.numbers[i]
 *         if self.mass:             # <<<<<<<<<<<<<<
 *             for i in range(n):
 *                 if isinf(self.force.numbers[i]) or isnan(self.force.numbers[i]):
 */
  __pyx_t_4 = (__pyx_v_self->mass != 0);
  if (__pyx_t_4) {

    /* "flitter/render/physics.pyx":76
 *             self.acceleration.numbers[i] = self.velocity.numbers[i]
 *         if self.mass:
 *             for i in range(n):             # <<<<<<<<<<<<<<
 *                 if isinf(self.force.numbers[i]) or isnan(self.force.numbers[i]):
 *                     break
 */
    __pyx_t_1 = __pyx_v_n;
    __pyx_t_2 = __pyx_t_1;
    for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
      __pyx_v_i = __pyx_t_3;

      /* "flitter/render/physics.pyx":77
 *         if self.mass:
 *             for i in range(n):
 *                 if isinf(self.force.numbers[i]) or isnan(self.force.numbers[i]):             # <<<<<<<<<<<<<<
 *                     break
 *             else:
 */
      __pyx_t_5 = (isinf((__pyx_v_self->force->numbers[__pyx_v_i])) != 0);
      if (!__pyx_t_5) {
      } else {
        __pyx_t_4 = __pyx_t_5;
        goto __pyx_L9_bool_binop_done;
      }
      __pyx_t_5 = isnan((__pyx_v_self->force->numbers[__pyx_v_i]));
      __pyx_t_4 = __pyx_t_5;
      __pyx_L9_bool_binop_done:;
      if (__pyx_t_4) {

        /* "flitter/render/physics.pyx":78
 *             for i in range(n):
 *                 if isinf(self.force.numbers[i]) or isnan(self.force.numbers[i]):
 *                     break             # <<<<<<<<<<<<<<
 *             else:
 *                 k = delta / self.mass
 */
        goto __pyx_L7_break;

        /* "flitter/render/physics.pyx":77
 *         if self.mass:
 *             for i in range(n):
 *                 if isinf(self.force.numbers[i]) or isnan(self.force.numbers[i]):             # <<<<<<<<<<<<<<
 *                     break
 *             else:
 */
      }
    }
    /*else*/ {

      /* "flitter/render/physics.pyx":80
 *                     break
 *             else:
 *                 k = delta / self.mass             # <<<<<<<<<<<<<<
 *                 if self.ease > 0 and clock < self.ease:
 *                     k *= clock / self.ease
 */
      __pyx_v_k = (__pyx_v_delta / __pyx_v_self->mass);

      /* "flitter/render/physics.pyx":81
 *             else:
 *                 k = delta / self.mass
 *                 if self.ease > 0 and clock < self.ease:             # <<<<<<<<<<<<<<
 *                     k *= clock / self.ease
 *                 speed_squared = 0
 */
      __pyx_t_5 = (__pyx_v_self->ease > 0.0);
      if (__pyx_t_5) {
      } else {
        __pyx_t_4 = __pyx_t_5;
        goto __pyx_L12_bool_binop_done;
      }
      __pyx_t_5 = (__pyx_v_clock < __pyx_v_self->ease);
      __pyx_t_4 = __pyx_t_5;
      __pyx_L12_bool_binop_done:;
      if (__pyx_t_4) {

        /* "flitter/render/physics.pyx":82
 *                 k = delta / self.mass
 *                 if self.ease > 0 and clock < self.ease:
 *                     k *= clock / self.ease             # <<<<<<<<<<<<<<
 *                 speed_squared = 0
 *                 for i in range(n):
 */
        __pyx_v_k = (__pyx_v_k * (__pyx_v_clock / __pyx_v_self->ease));

        /* "flitter/render/physics.pyx":81
 *             else:
 *                 k = delta / self.mass
 *                 if self.ease > 0 and clock < self.ease:             # <<<<<<<<<<<<<<
 *                     k *= clock / self.ease
 *                 speed_squared = 0
 */
      }

      /* "flitter/render/physics.pyx":83
 *                 if self.ease > 0 and clock < self.ease:
 *                     k *= clock / self.ease
 *                 speed_squared = 0             # <<<<<<<<<<<<<<
 *                 for i in range(n):
 *                     v = self.velocity.numbers[i] + self.force.numbers[i] * k
 */
      __pyx_v_speed_squared = 0.0;

      /* "flitter/render/physics.pyx":84
 *                     k *= clock / self.ease
 *                 speed_squared = 0
 *                 for i in range(n):             # <<<<<<<<<<<<<<
 *                     v = self.velocity.numbers[i] + self.force.numbers[i] * k
 *                     self.velocity.numbers[i] = v
 */
      __pyx_t_3 = __pyx_v_n;
      __pyx_t_6 = __pyx_t_3;
      for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
        __pyx_v_i = __pyx_t_7;

        /* "flitter/render/physics.pyx":85
 *                 speed_squared = 0
 *                 for i in range(n):
 *                     v = self.velocity.numbers[i] + self.force.numbers[i] * k             # <<<<<<<<<<<<<<
 *                     self.velocity.numbers[i] = v
 *                     speed_squared += v * v
 */
        __pyx_v_v = ((__pyx_v_self->velocity->numbers[__pyx_v_i]) + ((__pyx_v_self->force->numbers[__pyx_v_i]) * __pyx_v_k));

        /* "flitter/render/physics.pyx":86
 *                 for i in range(n):
 *                     v = self.velocity.numbers[i] + self.force.numbers[i] * k
 *                     self.velocity.numbers[i] = v             # <<<<<<<<<<<<<<
 *                     speed_squared += v * v
 *                 if speed_of_light > 0 and speed_squared > speed_of_light * speed_of_light:
 */
        (__pyx_v_self->velocity->numbers[__pyx_v_i]) = __pyx_v_v;

        /* "flitter/render/physics.pyx":87
 *                     v = self.velocity.numbers[i] + self.force.numbers[i] * k
 *                     self.velocity.numbers[i] = v
 *                     speed_squared += v * v             # <<<<<<<<<<<<<<
 *                 if speed_of_light > 0 and speed_squared > speed_of_light * speed_of_light:
 *                     k = speed_of_light / sqrt(speed_squared)
 */
        __pyx_v_speed_squared = (__pyx_v_speed_squared + (__pyx_v_v * __pyx_v_v));
      }

      /* "flitter/render/physics.pyx":88
 *                     self.velocity.numbers[i] = v
 *                     speed_squared += v * v
 *                 if speed_of_light > 0 and speed_squared > speed_of_light * speed_of_light:             # <<<<<<<<<<<<<<
 *                     k = speed_of_light / sqrt(speed_squared)
 *                     for i in range(n):
 */
      __pyx_t_5 = (__pyx_v_speed_of_light > 0.0);
      if (__pyx_t_5) {
      } else {
        __pyx_t_4 = __pyx_t_5;
        goto __pyx_L17_bool_binop_done;
      }
      __pyx_t_5 = (__pyx_v_speed_squared > (__pyx_v_speed_of_light * __pyx_v_speed_of_light));
      __pyx_t_4 = __pyx_t_5;
      __pyx_L17_bool_binop_done:;
      if (__pyx_t_4) {

        /* "flitter/render/physics.pyx":89
 *                     speed_squared += v * v
 *                 if speed_of_light > 0 and speed_squared > speed_of_light * speed_of_light:
 *                     k = speed_of_light / sqrt(speed_squared)             # <<<<<<<<<<<<<<
 *                     for i in range(n):
 *                         self.velocity.numbers[i] = self.velocity.numbers[i] * k
 */
        __pyx_v_k = (__pyx_v_speed_of_light / sqrt(__pyx_v_speed_squared));

        /* "flitter/render/physics.pyx":90
 *                 if speed_of_light > 0 and speed_squared > speed_of_light * speed_of_light:
 *                     k = speed_of_light / sqrt(speed_squared)
 *                     for i in range(n):             # <<<<<<<<<<<<<<
 *                         self.velocity.numbers[i] = self.velocity.numbers[i] * k
 *         for i in range(n):
 */
        __pyx_t_3 = __pyx_v_n;
        __pyx_t_6 = __pyx_t_3;
        for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
          __pyx_v_i = __pyx_t_7;

          /* "flitter/render/physics.pyx":91
 *                     k = speed_of_light / sqrt(speed_squared)
 *                     for i in range(n):
 *                         self.velocity.numbers[i] = self.velocity.numbers[i] * k             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             self.acceleration.numbers[i] = (self.velocity.numbers[i] - self.acceleration.numbers[i]) / delta
 */
          (__pyx_v_self->velocity->numbers[__pyx_v_i]) = ((__pyx_v_self->velocity->numbers[__pyx_v_i]) * __pyx_v_k);
        }

        /* "flitter/render/physics.pyx":88
 *                     self.velocity.numbers[i] = v
 *                     speed_squared += v * v
 *                 if speed_of_light > 0 and speed_squared > speed_of_light * speed_of_light:             # <<<<<<<<<<<<<<
 *                     k = speed_of_light / sqrt(speed_squared)
 *                     for i in range(n):
 */
      }
    }
    __pyx_L7_break:;

    /* "flitter/render/physics.pyx":75
 *         for i in range(n):
 *             self.acceleration.numbers[i] = self.velocity.numbers[i]
 *         if self.mass:             # <<<<<<<<<<<<<<
 *             for i in range(n):
 *                 if isinf(self.force.numbers[i]) or isnan(self.force.numbers[i]):
 */
  }

  /* "flitter/render/physics.pyx":92
 *                     for i in range(n):
 *                         self.velocity.numbers[i] = self.velocity.numbers[i] * k
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             self.acceleration.numbers[i] = (self.velocity.numbers[i] - self.acceleration.numbers[i]) / delta
 *             self.position.numbers[i] = self.position.numbers[i] + self.velocity.numbers[i] * delta
 */
  __pyx_t_1 = __pyx_v_n;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "flitter/render/physics.pyx":93
 *                         self.velocity.numbers[i] = self.velocity.numbers[i] * k
 *         for i in range(n):
 *             self.acceleration.numbers[i] = (self.velocity.numbers[i] - self.acceleration.numbers[i]) / delta             # <<<<<<<<<<<<<<
 *             self.position.numbers[i] = self.position.numbers[i] + self.velocity.numbers[i] * delta
 * 
 */
    (__pyx_v_self->acceleration->numbers[__pyx_v_i]) = (((__pyx_v_self->velocity->numbers[__pyx_v_i]) - (__pyx_v_self->acceleration->numbers[__pyx_v_i])) / __pyx_v_delta);

    /* "flitter/render/physics.pyx":94
 *         for i in range(n):
 *             self.acceleration.numbers[i] = (self.velocity.numbers[i] - self.acceleration.numbers[i]) / delta
 *             self.position.numbers[i] = self.position.numbers[i] + self.velocity.numbers[i] * delta             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (__pyx_v_self->position->numbers[__pyx_v_i]) = ((__pyx_v_self->position->numbers[__pyx_v_i]) + ((__pyx_v_self->velocity->numbers[__pyx_v_i]) * __pyx_v_delta));
  }

  /* "flitter/render/physics.pyx":70
 * 
 *     @cython.profile(False)
 *     cdef void step(self, double speed_of_light, double clock, double delta) noexcept nogil:             # <<<<<<<<<<<<<<
 *         cdef double speed_squared, v, k
 *         cdef int64_t i, n=self.force.length
 */

  /* function exit code */
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_6render_7physics_8Particle_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_6render_7physics_8Particle_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_8Particle_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_6render_7physics_8Particle_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_6render_7physics_8Particle___reduce_cython__(((struct __pyx_obj_7flitter_6render_7physics_Particle *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_6render_7physics_8Particle___reduce_cython__(struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.acceleration, self.charge, self.ease, self.force, self.id, self.initial_force, self.mass, self.non_anchor, self.position, self.position_state_key, self.radius, self.velocity, self.velocity_state_key)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->charge); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->ease); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->mass); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_self->non_anchor); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_self->radius); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyTuple_New(13); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF((PyObject *)__pyx_v_self->acceleration);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->acceleration);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, ((PyObject *)__pyx_v_self->acceleration))) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->force);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->force);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 3, ((PyObject *)__pyx_v_self->force))) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->id);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->id);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 4, ((PyObject *)__pyx_v_self->id))) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->initial_force);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->initial_force);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 5, ((PyObject *)__pyx_v_self->initial_force))) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 6, __pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 7, __pyx_t_4)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->position);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->position);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 8, ((PyObject *)__pyx_v_self->position))) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->position_state_key);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->position_state_key);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 9, ((PyObject *)__pyx_v_self->position_state_key))) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 10, __pyx_t_5)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->velocity);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->velocity);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 11, ((PyObject *)__pyx_v_self->velocity))) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->velocity_state_key);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->velocity_state_key);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 12, ((PyObject *)__pyx_v_self->velocity_state_key))) __PYX_ERR(0, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.acceleration, self.charge, self.ease, self.force, self.id, self.initial_force, self.mass, self.non_anchor, self.position, self.position_state_key, self.radius, self.velocity, self.velocity_state_key)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_6 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v__dict = __pyx_t_6;
  __pyx_t_6 = 0;

  /* "(tree fragment)":7
 *     state = (self.acceleration, self.charge, self.ease, self.force, self.id, self.initial_force, self.mass, self.non_anchor, self.position, self.position_state_key, self.radius, self.velocity, self.velocity_state_key)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_7 = (__pyx_v__dict != Py_None);
  if (__pyx_t_7) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v__dict)) __PYX_ERR(0, 8, __pyx_L1_error);
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_5));
    __pyx_t_5 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.acceleration is not None or self.force is not None or self.id is not None or self.initial_force is not None or self.position is not None or self.position_state_key is not None or self.velocity is not None or self.velocity_state_key is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.acceleration, self.charge, self.ease, self.force, self.id, self.initial_force, self.mass, self.non_anchor, self.position, self.position_state_key, self.radius, self.velocity, self.velocity_state_key)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.acceleration is not None or self.force is not None or self.id is not None or self.initial_force is not None or self.position is not None or self.position_state_key is not None or self.velocity is not None or self.velocity_state_key is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Particle, (type(self), 0xc9134b9, None), state
 */
  /*else*/ {
    __pyx_t_8 = (((PyObject *)__pyx_v_self->acceleration) != Py_None);
    if (!__pyx_t_8) {
    } else {
      __pyx_t_7 = __pyx_t_8;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_8 = (((PyObject *)__pyx_v_self->force) != Py_None);
    if (!__pyx_t_8) {
    } else {
      __pyx_t_7 = __pyx_t_8;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_8 = (((PyObject *)__pyx_v_self->id) != Py_None);
    if (!__pyx_t_8) {
    } else {
      __pyx_t_7 = __pyx_t_8;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_8 = (((PyObject *)__pyx_v_self->initial_force) != Py_None);
    if (!__pyx_t_8) {
    } else {
      __pyx_t_7 = __pyx_t_8;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_8 = (((PyObject *)__pyx_v_self->position) != Py_None);
    if (!__pyx_t_8) {
    } else {
      __pyx_t_7 = __pyx_t_8;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_8 = (((PyObject *)__pyx_v_self->position_state_key) != Py_None);
    if (!__pyx_t_8) {
    } else {
      __pyx_t_7 = __pyx_t_8;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_8 = (((PyObject *)__pyx_v_self->velocity) != Py_None);
    if (!__pyx_t_8) {
    } else {
      __pyx_t_7 = __pyx_t_8;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_8 = (((PyObject *)__pyx_v_self->velocity_state_key) != Py_None);
    __pyx_t_7 = __pyx_t_8;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_7;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.acceleration is not None or self.force is not None or self.id is not None or self.initial_force is not None or self.position is not None or self.position_state_key is not None or self.velocity is not None or self.velocity_state_key is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Particle, (type(self), 0xc9134b9, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.acceleration is not None or self.force is not None or self.id is not None or self.initial_force is not None or self.position is not None or self.position_state_key is not None or self.velocity is not None or self.velocity_state_key is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Particle, (type(self), 0xc9134b9, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Particle, (type(self), 0xc9134b9, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_Particle); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_210842809);
    __Pyx_GIVEREF(__pyx_int_210842809);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_int_210842809)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 2, Py_None)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_6);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_6)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_6 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.acceleration is not None or self.force is not None or self.id is not None or self.initial_force is not None or self.position is not None or self.position_state_key is not None or self.velocity is not None or self.velocity_state_key is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Particle, (type(self), 0xc9134b9, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Particle, (type(self), 0xc9134b9, None), state
 *     else:
 *         return __pyx_unpickle_Particle, (type(self), 0xc9134b9, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Particle__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_Particle); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_210842809);
    __Pyx_GIVEREF(__pyx_int_210842809);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_int_210842809)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_v_state)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_6);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_6)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_6 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("flitter.render.physics.Particle.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Particle, (type(self), 0xc9134b9, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Particle__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_6render_7physics_8Particle_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_6render_7physics_8Particle_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_8Particle_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_6render_7physics_8Particle_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.render.physics.Particle.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_6render_7physics_8Particle_2__setstate_cython__(((struct __pyx_obj_7flitter_6render_7physics_Particle *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_6render_7physics_8Particle_2__setstate_cython__(struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Particle, (type(self), 0xc9134b9, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Particle__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_6render_7physics___pyx_unpickle_Particle__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Particle, (type(self), 0xc9134b9, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Particle__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.render.physics.Particle.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/render/physics.pyx":106
 *     cdef double minimum
 * 
 *     @cython.profile(False)             # <<<<<<<<<<<<<<
 *     def __cinit__(self, Node node, double strength, Vector zero):
 *         self.position = node.get_fvec('position', zero.length, zero)
 */

/* Python wrapper */
static int __pyx_pw_7flitter_6render_7physics_7Barrier_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_6render_7physics_7Barrier_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_7flitter_5model_Node *__pyx_v_node = 0;
  double __pyx_v_strength;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_zero = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_node,&__pyx_n_s_strength,&__pyx_n_s_zero,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_node)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 106, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_strength)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 106, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, 1); __PYX_ERR(1, 106, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_zero)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 106, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, 2); __PYX_ERR(1, 106, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__cinit__") < 0)) __PYX_ERR(1, 106, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
      values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
    }
    __pyx_v_node = ((struct __pyx_obj_7flitter_5model_Node *)values[0]);
    __pyx_v_strength = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_strength == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 107, __pyx_L3_error)
    __pyx_v_zero = ((struct __pyx_obj_7flitter_5model_Vector *)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 106, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.render.physics.Barrier.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_node), __pyx_ptype_7flitter_5model_Node, 1, "node", 0))) __PYX_ERR(1, 107, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_zero), __pyx_ptype_7flitter_5model_Vector, 1, "zero", 0))) __PYX_ERR(1, 107, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_6render_7physics_7Barrier___cinit__(((struct __pyx_obj_7flitter_6render_7physics_Barrier *)__pyx_v_self), __pyx_v_node, __pyx_v_strength, __pyx_v_zero);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_6render_7physics_7Barrier___cinit__(struct __pyx_obj_7flitter_6render_7physics_Barrier *__pyx_v_self, struct __pyx_obj_7flitter_5model_Node *__pyx_v_node, double __pyx_v_strength, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_zero) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  long __pyx_t_3;
  double __pyx_t_4;
  long __pyx_t_5;
  double __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 1);

  /* "flitter/render/physics.pyx":108
 *     @cython.profile(False)
 *     def __cinit__(self, Node node, double strength, Vector zero):
 *         self.position = node.get_fvec('position', zero.length, zero)             # <<<<<<<<<<<<<<
 *         self.normal = node.get_fvec('normal', zero.length, zero).normalize()
 *         self.restitution = min(max(0, node.get_float('restitution', 1)), 1)
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Node *)__pyx_v_node->__pyx_vtab)->get_fvec(__pyx_v_node, __pyx_n_u_position, __pyx_v_zero->length, __pyx_v_zero)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->position);
  __Pyx_DECREF((PyObject *)__pyx_v_self->position);
  __pyx_v_self->position = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/render/physics.pyx":109
 *     def __cinit__(self, Node node, double strength, Vector zero):
 *         self.position = node.get_fvec('position', zero.length, zero)
 *         self.normal = node.get_fvec('normal', zero.length, zero).normalize()             # <<<<<<<<<<<<<<
 *         self.restitution = min(max(0, node.get_float('restitution', 1)), 1)
 *         self.strength = strength
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Node *)__pyx_v_node->__pyx_vtab)->get_fvec(__pyx_v_node, __pyx_n_u_normal, __pyx_v_zero->length, __pyx_v_zero)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1)->__pyx_vtab)->normalize(((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1), 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->normal);
  __Pyx_DECREF((PyObject *)__pyx_v_self->normal);
  __pyx_v_self->normal = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "flitter/render/physics.pyx":110
 *         self.position = node.get_fvec('position', zero.length, zero)
 *         self.normal = node.get_fvec('normal', zero.length, zero).normalize()
 *         self.restitution = min(max(0, node.get_float('restitution', 1)), 1)             # <<<<<<<<<<<<<<
 *         self.strength = strength
 *         self.power = max(0, node.get_float('power', 1))
 */
  __pyx_t_3 = 1;
  __pyx_t_4 = ((struct __pyx_vtabstruct_7flitter_5model_Node *)__pyx_v_node->__pyx_vtab)->get_float(__pyx_v_node, __pyx_n_u_restitution, 1.0); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 110, __pyx_L1_error)
  __pyx_t_5 = 0;
  __pyx_t_7 = (__pyx_t_4 > __pyx_t_5);
  if (__pyx_t_7) {
    __pyx_t_6 = __pyx_t_4;
  } else {
    __pyx_t_6 = __pyx_t_5;
  }
  __pyx_t_4 = __pyx_t_6;
  __pyx_t_7 = (__pyx_t_3 < __pyx_t_4);
  if (__pyx_t_7) {
    __pyx_t_6 = __pyx_t_3;
  } else {
    __pyx_t_6 = __pyx_t_4;
  }
  __pyx_v_self->restitution = __pyx_t_6;

  /* "flitter/render/physics.pyx":111
 *         self.normal = node.get_fvec('normal', zero.length, zero).normalize()
 *         self.restitution = min(max(0, node.get_float('restitution', 1)), 1)
 *         self.strength = strength             # <<<<<<<<<<<<<<
 *         self.power = max(0, node.get_float('power', 1))
 *         self.minimum = node.get_float('minimum', node.get_float('min', 0))
 */
  __pyx_v_self->strength = __pyx_v_strength;

  /* "flitter/render/physics.pyx":112
 *         self.restitution = min(max(0, node.get_float('restitution', 1)), 1)
 *         self.strength = strength
 *         self.power = max(0, node.get_float('power', 1))             # <<<<<<<<<<<<<<
 *         self.minimum = node.get_float('minimum', node.get_float('min', 0))
 * 
 */
  __pyx_t_6 = ((struct __pyx_vtabstruct_7flitter_5model_Node *)__pyx_v_node->__pyx_vtab)->get_float(__pyx_v_node, __pyx_n_u_power, 1.0); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 112, __pyx_L1_error)
  __pyx_t_3 = 0;
  __pyx_t_7 = (__pyx_t_6 > __pyx_t_3);
  if (__pyx_t_7) {
    __pyx_t_4 = __pyx_t_6;
  } else {
    __pyx_t_4 = __pyx_t_3;
  }
  __pyx_v_self->power = __pyx_t_4;

  /* "flitter/render/physics.pyx":113
 *         self.strength = strength
 *         self.power = max(0, node.get_float('power', 1))
 *         self.minimum = node.get_float('minimum', node.get_float('min', 0))             # <<<<<<<<<<<<<<
 * 
 *     @cython.profile(False)
 */
  __pyx_t_4 = ((struct __pyx_vtabstruct_7flitter_5model_Node *)__pyx_v_node->__pyx_vtab)->get_float(__pyx_v_node, __pyx_n_u_min, 0.0); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 113, __pyx_L1_error)
  __pyx_t_6 = ((struct __pyx_vtabstruct_7flitter_5model_Node *)__pyx_v_node->__pyx_vtab)->get_float(__pyx_v_node, __pyx_n_u_minimum, __pyx_t_4); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 113, __pyx_L1_error)
  __pyx_v_self->minimum = __pyx_t_6;

  /* "flitter/render/physics.pyx":106
 *     cdef double minimum
 * 
 *     @cython.profile(False)             # <<<<<<<<<<<<<<
 *     def __cinit__(self, Node node, double strength, Vector zero):
 *         self.position = node.get_fvec('position', zero.length, zero)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("flitter.render.physics.Barrier.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/render/physics.pyx":116
 * 
 *     @cython.profile(False)
 *     cdef void apply_distance(self, Particle particle) noexcept nogil:             # <<<<<<<<<<<<<<
 *         if self.normal.length == 0 or self.minimum <= 0:
 *             return
 */

static void __pyx_f_7flitter_6render_7physics_7Barrier_apply_distance(struct __pyx_obj_7flitter_6render_7physics_Barrier *__pyx_v_self, struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_particle) {
  int __pyx_v_i;
  int __pyx_v_dimensions;
  double __pyx_v_n;
  double __pyx_v_k;
  double __pyx_v_distance;
  int __pyx_t_1;
  int __pyx_t_2;
  int64_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;

  /* "flitter/render/physics.pyx":117
 *     @cython.profile(False)
 *     cdef void apply_distance(self, Particle particle) noexcept nogil:
 *         if self.normal.length == 0 or self.minimum <= 0:             # <<<<<<<<<<<<<<
 *             return
 *         cdef int i, dimensions=self.position.length
 */
  __pyx_t_2 = (__pyx_v_self->normal->length == 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->minimum <= 0.0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "flitter/render/physics.pyx":118
 *     cdef void apply_distance(self, Particle particle) noexcept nogil:
 *         if self.normal.length == 0 or self.minimum <= 0:
 *             return             # <<<<<<<<<<<<<<
 *         cdef int i, dimensions=self.position.length
 *         cdef double n, k, distance=0
 */
    goto __pyx_L0;

    /* "flitter/render/physics.pyx":117
 *     @cython.profile(False)
 *     cdef void apply_distance(self, Particle particle) noexcept nogil:
 *         if self.normal.length == 0 or self.minimum <= 0:             # <<<<<<<<<<<<<<
 *             return
 *         cdef int i, dimensions=self.position.length
 */
  }

  /* "flitter/render/physics.pyx":119
 *         if self.normal.length == 0 or self.minimum <= 0:
 *             return
 *         cdef int i, dimensions=self.position.length             # <<<<<<<<<<<<<<
 *         cdef double n, k, distance=0
 *         for i in range(dimensions):
 */
  __pyx_t_3 = __pyx_v_self->position->length;
  __pyx_v_dimensions = __pyx_t_3;

  /* "flitter/render/physics.pyx":120
 *             return
 *         cdef int i, dimensions=self.position.length
 *         cdef double n, k, distance=0             # <<<<<<<<<<<<<<
 *         for i in range(dimensions):
 *             n = self.normal.numbers[i]
 */
  __pyx_v_distance = 0.0;

  /* "flitter/render/physics.pyx":121
 *         cdef int i, dimensions=self.position.length
 *         cdef double n, k, distance=0
 *         for i in range(dimensions):             # <<<<<<<<<<<<<<
 *             n = self.normal.numbers[i]
 *             distance = distance + (particle.position.numbers[i] - self.position.numbers[i]) * n
 */
  __pyx_t_4 = __pyx_v_dimensions;
  __pyx_t_5 = __pyx_t_4;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "flitter/render/physics.pyx":122
 *         cdef double n, k, distance=0
 *         for i in range(dimensions):
 *             n = self.normal.numbers[i]             # <<<<<<<<<<<<<<
 *             distance = distance + (particle.position.numbers[i] - self.position.numbers[i]) * n
 *         if distance >= 0 and distance < self.minimum:
 */
    __pyx_v_n = (__pyx_v_self->normal->numbers[__pyx_v_i]);

    /* "flitter/render/physics.pyx":123
 *         for i in range(dimensions):
 *             n = self.normal.numbers[i]
 *             distance = distance + (particle.position.numbers[i] - self.position.numbers[i]) * n             # <<<<<<<<<<<<<<
 *         if distance >= 0 and distance < self.minimum:
 *             k = self.minimum - distance
 */
    __pyx_v_distance = (__pyx_v_distance + (((__pyx_v_particle->position->numbers[__pyx_v_i]) - (__pyx_v_self->position->numbers[__pyx_v_i])) * __pyx_v_n));
  }

  /* "flitter/render/physics.pyx":124
 *             n = self.normal.numbers[i]
 *             distance = distance + (particle.position.numbers[i] - self.position.numbers[i]) * n
 *         if distance >= 0 and distance < self.minimum:             # <<<<<<<<<<<<<<
 *             k = self.minimum - distance
 *             if self.power != 1:
 */
  __pyx_t_2 = (__pyx_v_distance >= 0.0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L9_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_distance < __pyx_v_self->minimum);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L9_bool_binop_done:;
  if (__pyx_t_1) {

    /* "flitter/render/physics.pyx":125
 *             distance = distance + (particle.position.numbers[i] - self.position.numbers[i]) * n
 *         if distance >= 0 and distance < self.minimum:
 *             k = self.minimum - distance             # <<<<<<<<<<<<<<
 *             if self.power != 1:
 *                 k **= self.power
 */
    __pyx_v_k = (__pyx_v_self->minimum - __pyx_v_distance);

    /* "flitter/render/physics.pyx":126
 *         if distance >= 0 and distance < self.minimum:
 *             k = self.minimum - distance
 *             if self.power != 1:             # <<<<<<<<<<<<<<
 *                 k **= self.power
 *             k *= self.strength
 */
    __pyx_t_1 = (__pyx_v_self->power != 1.0);
    if (__pyx_t_1) {

      /* "flitter/render/physics.pyx":127
 *             k = self.minimum - distance
 *             if self.power != 1:
 *                 k **= self.power             # <<<<<<<<<<<<<<
 *             k *= self.strength
 *             for i in range(dimensions):
 */
      __pyx_v_k = pow(__pyx_v_k, __pyx_v_self->power);

      /* "flitter/render/physics.pyx":126
 *         if distance >= 0 and distance < self.minimum:
 *             k = self.minimum - distance
 *             if self.power != 1:             # <<<<<<<<<<<<<<
 *                 k **= self.power
 *             k *= self.strength
 */
    }

    /* "flitter/render/physics.pyx":128
 *             if self.power != 1:
 *                 k **= self.power
 *             k *= self.strength             # <<<<<<<<<<<<<<
 *             for i in range(dimensions):
 *                 particle.force.numbers[i] = particle.force.numbers[i] + self.normal.numbers[i] * k
 */
    __pyx_v_k = (__pyx_v_k * __pyx_v_self->strength);

    /* "flitter/render/physics.pyx":129
 *                 k **= self.power
 *             k *= self.strength
 *             for i in range(dimensions):             # <<<<<<<<<<<<<<
 *                 particle.force.numbers[i] = particle.force.numbers[i] + self.normal.numbers[i] * k
 * 
 */
    __pyx_t_4 = __pyx_v_dimensions;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_i = __pyx_t_6;

      /* "flitter/render/physics.pyx":130
 *             k *= self.strength
 *             for i in range(dimensions):
 *                 particle.force.numbers[i] = particle.force.numbers[i] + self.normal.numbers[i] * k             # <<<<<<<<<<<<<<
 * 
 *     @cython.profile(False)
 */
      (__pyx_v_particle->force->numbers[__pyx_v_i]) = ((__pyx_v_particle->force->numbers[__pyx_v_i]) + ((__pyx_v_self->normal->numbers[__pyx_v_i]) * __pyx_v_k));
    }

    /* "flitter/render/physics.pyx":124
 *             n = self.normal.numbers[i]
 *             distance = distance + (particle.position.numbers[i] - self.position.numbers[i]) * n
 *         if distance >= 0 and distance < self.minimum:             # <<<<<<<<<<<<<<
 *             k = self.minimum - distance
 *             if self.power != 1:
 */
  }

  /* "flitter/render/physics.pyx":116
 * 
 *     @cython.profile(False)
 *     cdef void apply_distance(self, Particle particle) noexcept nogil:             # <<<<<<<<<<<<<<
 *         if self.normal.length == 0 or self.minimum <= 0:
 *             return
 */

  /* function exit code */
  __pyx_L0:;
}

/* "flitter/render/physics.pyx":133
 * 
 *     @cython.profile(False)
 *     cdef void apply_rebound(self, Particle particle, double speed_of_light, double clock, double delta) noexcept nogil:             # <<<<<<<<<<<<<<
 *         if self.normal.length == 0:
 *             return
 */

static void __pyx_f_7flitter_6render_7physics_7Barrier_apply_rebound(struct __pyx_obj_7flitter_6render_7physics_Barrier *__pyx_v_self, struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_particle, double __pyx_v_speed_of_light, double __pyx_v_clock, double __pyx_v_delta) {
  int __pyx_v_i;
  int __pyx_v_dimensions;
  double __pyx_v_n;
  double __pyx_v_velocity;
  double __pyx_v_distance;
  double __pyx_v_acceleration;
  double __pyx_v_rewind_t;
  double __pyx_v_v;
  double __pyx_v_t;
  int __pyx_t_1;
  int64_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  double __pyx_t_7;

  /* "flitter/render/physics.pyx":134
 *     @cython.profile(False)
 *     cdef void apply_rebound(self, Particle particle, double speed_of_light, double clock, double delta) noexcept nogil:
 *         if self.normal.length == 0:             # <<<<<<<<<<<<<<
 *             return
 *         cdef int i, dimensions=self.position.length
 */
  __pyx_t_1 = (__pyx_v_self->normal->length == 0);
  if (__pyx_t_1) {

    /* "flitter/render/physics.pyx":135
 *     cdef void apply_rebound(self, Particle particle, double speed_of_light, double clock, double delta) noexcept nogil:
 *         if self.normal.length == 0:
 *             return             # <<<<<<<<<<<<<<
 *         cdef int i, dimensions=self.position.length
 *         cdef double n, velocity=0, distance=-particle.radius, acceleration=0
 */
    goto __pyx_L0;

    /* "flitter/render/physics.pyx":134
 *     @cython.profile(False)
 *     cdef void apply_rebound(self, Particle particle, double speed_of_light, double clock, double delta) noexcept nogil:
 *         if self.normal.length == 0:             # <<<<<<<<<<<<<<
 *             return
 *         cdef int i, dimensions=self.position.length
 */
  }

  /* "flitter/render/physics.pyx":136
 *         if self.normal.length == 0:
 *             return
 *         cdef int i, dimensions=self.position.length             # <<<<<<<<<<<<<<
 *         cdef double n, velocity=0, distance=-particle.radius, acceleration=0
 *         for i in range(dimensions):
 */
  __pyx_t_2 = __pyx_v_self->position->length;
  __pyx_v_dimensions = __pyx_t_2;

  /* "flitter/render/physics.pyx":137
 *             return
 *         cdef int i, dimensions=self.position.length
 *         cdef double n, velocity=0, distance=-particle.radius, acceleration=0             # <<<<<<<<<<<<<<
 *         for i in range(dimensions):
 *             n = self.normal.numbers[i]
 */
  __pyx_v_velocity = 0.0;
  __pyx_v_distance = (-__pyx_v_particle->radius);
  __pyx_v_acceleration = 0.0;

  /* "flitter/render/physics.pyx":138
 *         cdef int i, dimensions=self.position.length
 *         cdef double n, velocity=0, distance=-particle.radius, acceleration=0
 *         for i in range(dimensions):             # <<<<<<<<<<<<<<
 *             n = self.normal.numbers[i]
 *             distance = distance + (particle.position.numbers[i] - self.position.numbers[i]) * n
 */
  __pyx_t_3 = __pyx_v_dimensions;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "flitter/render/physics.pyx":139
 *         cdef double n, velocity=0, distance=-particle.radius, acceleration=0
 *         for i in range(dimensions):
 *             n = self.normal.numbers[i]             # <<<<<<<<<<<<<<
 *             distance = distance + (particle.position.numbers[i] - self.position.numbers[i]) * n
 *             velocity = velocity + particle.velocity.numbers[i] * n
 */
    __pyx_v_n = (__pyx_v_self->normal->numbers[__pyx_v_i]);

    /* "flitter/render/physics.pyx":140
 *         for i in range(dimensions):
 *             n = self.normal.numbers[i]
 *             distance = distance + (particle.position.numbers[i] - self.position.numbers[i]) * n             # <<<<<<<<<<<<<<
 *             velocity = velocity + particle.velocity.numbers[i] * n
 *             acceleration = acceleration + particle.acceleration.numbers[i] * n
 */
    __pyx_v_distance = (__pyx_v_distance + (((__pyx_v_particle->position->numbers[__pyx_v_i]) - (__pyx_v_self->position->numbers[__pyx_v_i])) * __pyx_v_n));

    /* "flitter/render/physics.pyx":141
 *             n = self.normal.numbers[i]
 *             distance = distance + (particle.position.numbers[i] - self.position.numbers[i]) * n
 *             velocity = velocity + particle.velocity.numbers[i] * n             # <<<<<<<<<<<<<<
 *             acceleration = acceleration + particle.acceleration.numbers[i] * n
 *         if distance >= 0:
 */
    __pyx_v_velocity = (__pyx_v_velocity + ((__pyx_v_particle->velocity->numbers[__pyx_v_i]) * __pyx_v_n));

    /* "flitter/render/physics.pyx":142
 *             distance = distance + (particle.position.numbers[i] - self.position.numbers[i]) * n
 *             velocity = velocity + particle.velocity.numbers[i] * n
 *             acceleration = acceleration + particle.acceleration.numbers[i] * n             # <<<<<<<<<<<<<<
 *         if distance >= 0:
 *             return
 */
    __pyx_v_acceleration = (__pyx_v_acceleration + ((__pyx_v_particle->acceleration->numbers[__pyx_v_i]) * __pyx_v_n));
  }

  /* "flitter/render/physics.pyx":143
 *             velocity = velocity + particle.velocity.numbers[i] * n
 *             acceleration = acceleration + particle.acceleration.numbers[i] * n
 *         if distance >= 0:             # <<<<<<<<<<<<<<
 *             return
 *         cdef double rewind_t = distance/velocity
 */
  __pyx_t_1 = (__pyx_v_distance >= 0.0);
  if (__pyx_t_1) {

    /* "flitter/render/physics.pyx":144
 *             acceleration = acceleration + particle.acceleration.numbers[i] * n
 *         if distance >= 0:
 *             return             # <<<<<<<<<<<<<<
 *         cdef double rewind_t = distance/velocity
 *         if velocity >= 0 or rewind_t > delta:
 */
    goto __pyx_L0;

    /* "flitter/render/physics.pyx":143
 *             velocity = velocity + particle.velocity.numbers[i] * n
 *             acceleration = acceleration + particle.acceleration.numbers[i] * n
 *         if distance >= 0:             # <<<<<<<<<<<<<<
 *             return
 *         cdef double rewind_t = distance/velocity
 */
  }

  /* "flitter/render/physics.pyx":145
 *         if distance >= 0:
 *             return
 *         cdef double rewind_t = distance/velocity             # <<<<<<<<<<<<<<
 *         if velocity >= 0 or rewind_t > delta:
 *             for i in range(dimensions):
 */
  __pyx_v_rewind_t = (__pyx_v_distance / __pyx_v_velocity);

  /* "flitter/render/physics.pyx":146
 *             return
 *         cdef double rewind_t = distance/velocity
 *         if velocity >= 0 or rewind_t > delta:             # <<<<<<<<<<<<<<
 *             for i in range(dimensions):
 *                 particle.position.numbers[i] = particle.position.numbers[i] - distance * self.normal.numbers[i]
 */
  __pyx_t_6 = (__pyx_v_velocity >= 0.0);
  if (!__pyx_t_6) {
  } else {
    __pyx_t_1 = __pyx_t_6;
    goto __pyx_L8_bool_binop_done;
  }
  __pyx_t_6 = (__pyx_v_rewind_t > __pyx_v_delta);
  __pyx_t_1 = __pyx_t_6;
  __pyx_L8_bool_binop_done:;
  if (__pyx_t_1) {

    /* "flitter/render/physics.pyx":147
 *         cdef double rewind_t = distance/velocity
 *         if velocity >= 0 or rewind_t > delta:
 *             for i in range(dimensions):             # <<<<<<<<<<<<<<
 *                 particle.position.numbers[i] = particle.position.numbers[i] - distance * self.normal.numbers[i]
 *             return
 */
    __pyx_t_3 = __pyx_v_dimensions;
    __pyx_t_4 = __pyx_t_3;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
      __pyx_v_i = __pyx_t_5;

      /* "flitter/render/physics.pyx":148
 *         if velocity >= 0 or rewind_t > delta:
 *             for i in range(dimensions):
 *                 particle.position.numbers[i] = particle.position.numbers[i] - distance * self.normal.numbers[i]             # <<<<<<<<<<<<<<
 *             return
 *         cdef double v
 */
      (__pyx_v_particle->position->numbers[__pyx_v_i]) = ((__pyx_v_particle->position->numbers[__pyx_v_i]) - (__pyx_v_distance * (__pyx_v_self->normal->numbers[__pyx_v_i])));
    }

    /* "flitter/render/physics.pyx":149
 *             for i in range(dimensions):
 *                 particle.position.numbers[i] = particle.position.numbers[i] - distance * self.normal.numbers[i]
 *             return             # <<<<<<<<<<<<<<
 *         cdef double v
 *         velocity = 0
 */
    goto __pyx_L0;

    /* "flitter/render/physics.pyx":146
 *             return
 *         cdef double rewind_t = distance/velocity
 *         if velocity >= 0 or rewind_t > delta:             # <<<<<<<<<<<<<<
 *             for i in range(dimensions):
 *                 particle.position.numbers[i] = particle.position.numbers[i] - distance * self.normal.numbers[i]
 */
  }

  /* "flitter/render/physics.pyx":151
 *             return
 *         cdef double v
 *         velocity = 0             # <<<<<<<<<<<<<<
 *         for i in range(dimensions):
 *             particle.position.numbers[i] = particle.position.numbers[i] - rewind_t * particle.velocity.numbers[i]
 */
  __pyx_v_velocity = 0.0;

  /* "flitter/render/physics.pyx":152
 *         cdef double v
 *         velocity = 0
 *         for i in range(dimensions):             # <<<<<<<<<<<<<<
 *             particle.position.numbers[i] = particle.position.numbers[i] - rewind_t * particle.velocity.numbers[i]
 *             v = (particle.velocity.numbers[i] - rewind_t * particle.acceleration.numbers[i]) * self.restitution
 */
  __pyx_t_3 = __pyx_v_dimensions;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "flitter/render/physics.pyx":153
 *         velocity = 0
 *         for i in range(dimensions):
 *             particle.position.numbers[i] = particle.position.numbers[i] - rewind_t * particle.velocity.numbers[i]             # <<<<<<<<<<<<<<
 *             v = (particle.velocity.numbers[i] - rewind_t * particle.acceleration.numbers[i]) * self.restitution
 *             particle.velocity.numbers[i] = v
 */
    (__pyx_v_particle->position->numbers[__pyx_v_i]) = ((__pyx_v_particle->position->numbers[__pyx_v_i]) - (__pyx_v_rewind_t * (__pyx_v_particle->velocity->numbers[__pyx_v_i])));

    /* "flitter/render/physics.pyx":154
 *         for i in range(dimensions):
 *             particle.position.numbers[i] = particle.position.numbers[i] - rewind_t * particle.velocity.numbers[i]
 *             v = (particle.velocity.numbers[i] - rewind_t * particle.acceleration.numbers[i]) * self.restitution             # <<<<<<<<<<<<<<
 *             particle.velocity.numbers[i] = v
 *             velocity = velocity + v * self.normal.numbers[i]
 */
    __pyx_v_v = (((__pyx_v_particle->velocity->numbers[__pyx_v_i]) - (__pyx_v_rewind_t * (__pyx_v_particle->acceleration->numbers[__pyx_v_i]))) * __pyx_v_self->restitution);

    /* "flitter/render/physics.pyx":155
 *             particle.position.numbers[i] = particle.position.numbers[i] - rewind_t * particle.velocity.numbers[i]
 *             v = (particle.velocity.numbers[i] - rewind_t * particle.acceleration.numbers[i]) * self.restitution
 *             particle.velocity.numbers[i] = v             # <<<<<<<<<<<<<<
 *             velocity = velocity + v * self.normal.numbers[i]
 *         for i in range(dimensions):
 */
    (__pyx_v_particle->velocity->numbers[__pyx_v_i]) = __pyx_v_v;

    /* "flitter/render/physics.pyx":156
 *             v = (particle.velocity.numbers[i] - rewind_t * particle.acceleration.numbers[i]) * self.restitution
 *             particle.velocity.numbers[i] = v
 *             velocity = velocity + v * self.normal.numbers[i]             # <<<<<<<<<<<<<<
 *         for i in range(dimensions):
 *             n = self.normal.numbers[i]
 */
    __pyx_v_velocity = (__pyx_v_velocity + (__pyx_v_v * (__pyx_v_self->normal->numbers[__pyx_v_i])));
  }

  /* "flitter/render/physics.pyx":157
 *             particle.velocity.numbers[i] = v
 *             velocity = velocity + v * self.normal.numbers[i]
 *         for i in range(dimensions):             # <<<<<<<<<<<<<<
 *             n = self.normal.numbers[i]
 *             particle.velocity.numbers[i] = particle.velocity.numbers[i] - 2 * velocity * n
 */
  __pyx_t_3 = __pyx_v_dimensions;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "flitter/render/physics.pyx":158
 *             velocity = velocity + v * self.normal.numbers[i]
 *         for i in range(dimensions):
 *             n = self.normal.numbers[i]             # <<<<<<<<<<<<<<
 *             particle.velocity.numbers[i] = particle.velocity.numbers[i] - 2 * velocity * n
 *         cdef double t
 */
    __pyx_v_n = (__pyx_v_self->normal->numbers[__pyx_v_i]);

    /* "flitter/render/physics.pyx":159
 *         for i in range(dimensions):
 *             n = self.normal.numbers[i]
 *             particle.velocity.numbers[i] = particle.velocity.numbers[i] - 2 * velocity * n             # <<<<<<<<<<<<<<
 *         cdef double t
 *         if acceleration < 0 and (t := velocity / acceleration) < rewind_t:
 */
    (__pyx_v_particle->velocity->numbers[__pyx_v_i]) = ((__pyx_v_particle->velocity->numbers[__pyx_v_i]) - ((2.0 * __pyx_v_velocity) * __pyx_v_n));
  }

  /* "flitter/render/physics.pyx":161
 *             particle.velocity.numbers[i] = particle.velocity.numbers[i] - 2 * velocity * n
 *         cdef double t
 *         if acceleration < 0 and (t := velocity / acceleration) < rewind_t:             # <<<<<<<<<<<<<<
 *             rewind_t = t
 *         particle.step(speed_of_light, clock, rewind_t)
 */
  __pyx_t_6 = (__pyx_v_acceleration < 0.0);
  if (__pyx_t_6) {
  } else {
    __pyx_t_1 = __pyx_t_6;
    goto __pyx_L17_bool_binop_done;
  }
  __pyx_t_7 = (__pyx_v_velocity / __pyx_v_acceleration);
  __pyx_v_t = __pyx_t_7;
  __pyx_t_6 = (__pyx_t_7 < __pyx_v_rewind_t);
  __pyx_t_1 = __pyx_t_6;
  __pyx_L17_bool_binop_done:;
  if (__pyx_t_1) {

    /* "flitter/render/physics.pyx":162
 *         cdef double t
 *         if acceleration < 0 and (t := velocity / acceleration) < rewind_t:
 *             rewind_t = t             # <<<<<<<<<<<<<<
 *         particle.step(speed_of_light, clock, rewind_t)
 * 
 */
    __pyx_v_rewind_t = __pyx_v_t;

    /* "flitter/render/physics.pyx":161
 *             particle.velocity.numbers[i] = particle.velocity.numbers[i] - 2 * velocity * n
 *         cdef double t
 *         if acceleration < 0 and (t := velocity / acceleration) < rewind_t:             # <<<<<<<<<<<<<<
 *             rewind_t = t
 *         particle.step(speed_of_light, clock, rewind_t)
 */
  }

  /* "flitter/render/physics.pyx":163
 *         if acceleration < 0 and (t := velocity / acceleration) < rewind_t:
 *             rewind_t = t
 *         particle.step(speed_of_light, clock, rewind_t)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_f_7flitter_6render_7physics_8Particle_step(__pyx_v_particle, __pyx_v_speed_of_light, __pyx_v_clock, __pyx_v_rewind_t);

  /* "flitter/render/physics.pyx":133
 * 
 *     @cython.profile(False)
 *     cdef void apply_rebound(self, Particle particle, double speed_of_light, double clock, double delta) noexcept nogil:             # <<<<<<<<<<<<<<
 *         if self.normal.length == 0:
 *             return
 */

  /* function exit code */
  __pyx_L0:;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_6render_7physics_7Barrier_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_6render_7physics_7Barrier_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_7Barrier_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_6render_7physics_7Barrier_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_6render_7physics_7Barrier_2__reduce_cython__(((struct __pyx_obj_7flitter_6render_7physics_Barrier *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_6render_7physics_7Barrier_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_Barrier *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.render.physics.Barrier.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_6render_7physics_7Barrier_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_6render_7physics_7Barrier_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_7Barrier_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_6render_7physics_7Barrier_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.render.physics.Barrier.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_6render_7physics_7Barrier_4__setstate_cython__(((struct __pyx_obj_7flitter_6render_7physics_Barrier *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_6render_7physics_7Barrier_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_Barrier *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.render.physics.Barrier.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/render/physics.pyx":169
 *     cdef double strength
 * 
 *     @cython.profile(False)             # <<<<<<<<<<<<<<
 *     def __cinit__(self, Node node, double strength, Vector zero):
 *         self.strength = strength
 */

/* Python wrapper */
static int __pyx_pw_7flitter_6render_7physics_20ParticleForceApplier_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_6render_7physics_20ParticleForceApplier_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED struct __pyx_obj_7flitter_5model_Node *__pyx_v_node = 0;
  double __pyx_v_strength;
  CYTHON_UNUSED struct __pyx_obj_7flitter_5model_Vector *__pyx_v_zero = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_node,&__pyx_n_s_strength,&__pyx_n_s_zero,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_node)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 169, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_strength)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 169, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, 1); __PYX_ERR(1, 169, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_zero)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 169, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, 2); __PYX_ERR(1, 169, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__cinit__") < 0)) __PYX_ERR(1, 169, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
      values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
    }
    __pyx_v_node = ((struct __pyx_obj_7flitter_5model_Node *)values[0]);
    __pyx_v_strength = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_strength == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 170, __pyx_L3_error)
    __pyx_v_zero = ((struct __pyx_obj_7flitter_5model_Vector *)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 169, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.render.physics.ParticleForceApplier.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_node), __pyx_ptype_7flitter_5model_Node, 1, "node", 0))) __PYX_ERR(1, 170, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_zero), __pyx_ptype_7flitter_5model_Vector, 1, "zero", 0))) __PYX_ERR(1, 170, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_6render_7physics_20ParticleForceApplier___cinit__(((struct __pyx_obj_7flitter_6render_7physics_ParticleForceApplier *)__pyx_v_self), __pyx_v_node, __pyx_v_strength, __pyx_v_zero);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_6render_7physics_20ParticleForceApplier___cinit__(struct __pyx_obj_7flitter_6render_7physics_ParticleForceApplier *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_7flitter_5model_Node *__pyx_v_node, double __pyx_v_strength, CYTHON_UNUSED struct __pyx_obj_7flitter_5model_Vector *__pyx_v_zero) {
  int __pyx_r;

  /* "flitter/render/physics.pyx":171
 *     @cython.profile(False)
 *     def __cinit__(self, Node node, double strength, Vector zero):
 *         self.strength = strength             # <<<<<<<<<<<<<<
 * 
 *     @cython.profile(False)
 */
  __pyx_v_self->strength = __pyx_v_strength;

  /* "flitter/render/physics.pyx":169
 *     cdef double strength
 * 
 *     @cython.profile(False)             # <<<<<<<<<<<<<<
 *     def __cinit__(self, Node node, double strength, Vector zero):
 *         self.strength = strength
 */

  /* function exit code */
  __pyx_r = 0;
  return __pyx_r;
}

/* "flitter/render/physics.pyx":174
 * 
 *     @cython.profile(False)
 *     cdef void apply(self, Particle particle, double delta) noexcept nogil:             # <<<<<<<<<<<<<<
 *         raise NotImplementedError()
 * 
 */

static void __pyx_f_7flitter_6render_7physics_20ParticleForceApplier_apply(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_ParticleForceApplier *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_particle, CYTHON_UNUSED double __pyx_v_delta) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save;
  #endif
  __Pyx_RefNannySetupContext("apply", 1);

  /* "flitter/render/physics.pyx":175
 *     @cython.profile(False)
 *     cdef void apply(self, Particle particle, double delta) noexcept nogil:
 *         raise NotImplementedError()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      /*try:*/ {
        __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_builtin_NotImplementedError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 175, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_Raise(__pyx_t_1, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __PYX_ERR(1, 175, __pyx_L4_error)
      }
      /*finally:*/ {
        __pyx_L4_error: {
          #ifdef WITH_THREAD
          __Pyx_PyGILState_Release(__pyx_gilstate_save);
          #endif
          goto __pyx_L1_error;
        }
      }
  }

  /* "flitter/render/physics.pyx":174
 * 
 *     @cython.profile(False)
 *     cdef void apply(self, Particle particle, double delta) noexcept nogil:             # <<<<<<<<<<<<<<
 *         raise NotImplementedError()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  #ifdef WITH_THREAD
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("flitter.render.physics.ParticleForceApplier.apply", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  __Pyx_RefNannyFinishContextNogil()
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_6render_7physics_20ParticleForceApplier_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_6render_7physics_20ParticleForceApplier_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_20ParticleForceApplier_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_6render_7physics_20ParticleForceApplier_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_6render_7physics_20ParticleForceApplier_2__reduce_cython__(((struct __pyx_obj_7flitter_6render_7physics_ParticleForceApplier *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_6render_7physics_20ParticleForceApplier_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_ParticleForceApplier *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.render.physics.ParticleForceApplier.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_6render_7physics_20ParticleForceApplier_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_6render_7physics_20ParticleForceApplier_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_20ParticleForceApplier_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_6render_7physics_20ParticleForceApplier_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.render.physics.ParticleForceApplier.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_6render_7physics_20ParticleForceApplier_4__setstate_cython__(((struct __pyx_obj_7flitter_6render_7physics_ParticleForceApplier *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_6render_7physics_20ParticleForceApplier_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_ParticleForceApplier *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.render.physics.ParticleForceApplier.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/render/physics.pyx":182
 *     cdef double max_distance
 * 
 *     @cython.profile(False)             # <<<<<<<<<<<<<<
 *     def __cinit__(self, Node node, double strength, Vector zero):
 *         self.strength = strength
 */

/* Python wrapper */
static int __pyx_pw_7flitter_6render_7physics_18MatrixForceApplier_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_6render_7physics_18MatrixForceApplier_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_7flitter_5model_Node *__pyx_v_node = 0;
  double __pyx_v_strength;
  CYTHON_UNUSED struct __pyx_obj_7flitter_5model_Vector *__pyx_v_zero = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_node,&__pyx_n_s_strength,&__pyx_n_s_zero,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_node)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 182, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_strength)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 182, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, 1); __PYX_ERR(1, 182, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_zero)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 182, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, 2); __PYX_ERR(1, 182, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__cinit__") < 0)) __PYX_ERR(1, 182, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
      values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
    }
    __pyx_v_node = ((struct __pyx_obj_7flitter_5model_Node *)values[0]);
    __pyx_v_strength = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_strength == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 183, __pyx_L3_error)
    __pyx_v_zero = ((struct __pyx_obj_7flitter_5model_Vector *)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 182, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.render.physics.MatrixForceApplier.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_node), __pyx_ptype_7flitter_5model_Node, 1, "node", 0))) __PYX_ERR(1, 183, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_zero), __pyx_ptype_7flitter_5model_Vector, 1, "zero", 0))) __PYX_ERR(1, 183, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_6render_7physics_18MatrixForceApplier___cinit__(((struct __pyx_obj_7flitter_6render_7physics_MatrixForceApplier *)__pyx_v_self), __pyx_v_node, __pyx_v_strength, __pyx_v_zero);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_6render_7physics_18MatrixForceApplier___cinit__(struct __pyx_obj_7flitter_6render_7physics_MatrixForceApplier *__pyx_v_self, struct __pyx_obj_7flitter_5model_Node *__pyx_v_node, double __pyx_v_strength, CYTHON_UNUSED struct __pyx_obj_7flitter_5model_Vector *__pyx_v_zero) {
  int __pyx_r;
  double __pyx_t_1;
  long __pyx_t_2;
  double __pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "flitter/render/physics.pyx":184
 *     @cython.profile(False)
 *     def __cinit__(self, Node node, double strength, Vector zero):
 *         self.strength = strength             # <<<<<<<<<<<<<<
 *         self.max_distance = max(0, node.get_float('max_distance', 0))
 * 
 */
  __pyx_v_self->strength = __pyx_v_strength;

  /* "flitter/render/physics.pyx":185
 *     def __cinit__(self, Node node, double strength, Vector zero):
 *         self.strength = strength
 *         self.max_distance = max(0, node.get_float('max_distance', 0))             # <<<<<<<<<<<<<<
 * 
 *     @cython.profile(False)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_7flitter_5model_Node *)__pyx_v_node->__pyx_vtab)->get_float(__pyx_v_node, __pyx_n_u_max_distance, 0.0); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 185, __pyx_L1_error)
  __pyx_t_2 = 0;
  __pyx_t_4 = (__pyx_t_1 > __pyx_t_2);
  if (__pyx_t_4) {
    __pyx_t_3 = __pyx_t_1;
  } else {
    __pyx_t_3 = __pyx_t_2;
  }
  __pyx_v_self->max_distance = __pyx_t_3;

  /* "flitter/render/physics.pyx":182
 *     cdef double max_distance
 * 
 *     @cython.profile(False)             # <<<<<<<<<<<<<<
 *     def __cinit__(self, Node node, double strength, Vector zero):
 *         self.strength = strength
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.render.physics.MatrixForceApplier.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "flitter/render/physics.pyx":188
 * 
 *     @cython.profile(False)
 *     cdef void apply(self, Particle from_particle, Particle to_particle, Vector direction, double distance) noexcept nogil:             # <<<<<<<<<<<<<<
 *         raise NotImplementedError()
 * 
 */

static void __pyx_f_7flitter_6render_7physics_18MatrixForceApplier_apply(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_MatrixForceApplier *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_from_particle, CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_to_particle, CYTHON_UNUSED struct __pyx_obj_7flitter_5model_Vector *__pyx_v_direction, CYTHON_UNUSED double __pyx_v_distance) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save;
  #endif
  __Pyx_RefNannySetupContext("apply", 1);

  /* "flitter/render/physics.pyx":189
 *     @cython.profile(False)
 *     cdef void apply(self, Particle from_particle, Particle to_particle, Vector direction, double distance) noexcept nogil:
 *         raise NotImplementedError()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      /*try:*/ {
        __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_builtin_NotImplementedError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 189, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_Raise(__pyx_t_1, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __PYX_ERR(1, 189, __pyx_L4_error)
      }
      /*finally:*/ {
        __pyx_L4_error: {
          #ifdef WITH_THREAD
          __Pyx_PyGILState_Release(__pyx_gilstate_save);
          #endif
          goto __pyx_L1_error;
        }
      }
  }

  /* "flitter/render/physics.pyx":188
 * 
 *     @cython.profile(False)
 *     cdef void apply(self, Particle from_particle, Particle to_particle, Vector direction, double distance) noexcept nogil:             # <<<<<<<<<<<<<<
 *         raise NotImplementedError()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  #ifdef WITH_THREAD
  __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("flitter.render.physics.MatrixForceApplier.apply", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  __Pyx_RefNannyFinishContextNogil()
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_6render_7physics_18MatrixForceApplier_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_6render_7physics_18MatrixForceApplier_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_18MatrixForceApplier_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_6render_7physics_18MatrixForceApplier_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_6render_7physics_18MatrixForceApplier_2__reduce_cython__(((struct __pyx_obj_7flitter_6render_7physics_MatrixForceApplier *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_6render_7physics_18MatrixForceApplier_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_MatrixForceApplier *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.render.physics.MatrixForceApplier.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_6render_7physics_18MatrixForceApplier_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_6render_7physics_18MatrixForceApplier_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_18MatrixForceApplier_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_6render_7physics_18MatrixForceApplier_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.render.physics.MatrixForceApplier.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_6render_7physics_18MatrixForceApplier_4__setstate_cython__(((struct __pyx_obj_7flitter_6render_7physics_MatrixForceApplier *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_6render_7physics_18MatrixForceApplier_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_MatrixForceApplier *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.render.physics.MatrixForceApplier.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/render/physics.pyx":202
 * 
 *     @cython.profile(False)
 *     cdef void apply(self, Particle from_particle, Particle to_particle, Vector direction, double distance) noexcept nogil:             # <<<<<<<<<<<<<<
 *         cdef double f, k
 *         cdef int64_t i
 */

static void __pyx_f_7flitter_6render_7physics_20DistanceForceApplier_apply(struct __pyx_obj_7flitter_6render_7physics_DistanceForceApplier *__pyx_v_self, struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_from_particle, struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_to_particle, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_direction, double __pyx_v_distance) {
  double __pyx_v_f;
  double __pyx_v_k;
  int64_t __pyx_v_i;
  int __pyx_t_1;
  int __pyx_t_2;
  int64_t __pyx_t_3;
  int64_t __pyx_t_4;
  int64_t __pyx_t_5;

  /* "flitter/render/physics.pyx":205
 *         cdef double f, k
 *         cdef int64_t i
 *         if self.minimum and distance < self.minimum:             # <<<<<<<<<<<<<<
 *             k = self.minimum - distance
 *             if self.power != 1:
 */
  __pyx_t_2 = (__pyx_v_self->minimum != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_distance < __pyx_v_self->minimum);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "flitter/render/physics.pyx":206
 *         cdef int64_t i
 *         if self.minimum and distance < self.minimum:
 *             k = self.minimum - distance             # <<<<<<<<<<<<<<
 *             if self.power != 1:
 *                 k **= self.power
 */
    __pyx_v_k = (__pyx_v_self->minimum - __pyx_v_distance);

    /* "flitter/render/physics.pyx":207
 *         if self.minimum and distance < self.minimum:
 *             k = self.minimum - distance
 *             if self.power != 1:             # <<<<<<<<<<<<<<
 *                 k **= self.power
 *             k *= self.strength
 */
    __pyx_t_1 = (__pyx_v_self->power != 1.0);
    if (__pyx_t_1) {

      /* "flitter/render/physics.pyx":208
 *             k = self.minimum - distance
 *             if self.power != 1:
 *                 k **= self.power             # <<<<<<<<<<<<<<
 *             k *= self.strength
 *             for i in range(direction.length):
 */
      __pyx_v_k = pow(__pyx_v_k, __pyx_v_self->power);

      /* "flitter/render/physics.pyx":207
 *         if self.minimum and distance < self.minimum:
 *             k = self.minimum - distance
 *             if self.power != 1:             # <<<<<<<<<<<<<<
 *                 k **= self.power
 *             k *= self.strength
 */
    }

    /* "flitter/render/physics.pyx":209
 *             if self.power != 1:
 *                 k **= self.power
 *             k *= self.strength             # <<<<<<<<<<<<<<
 *             for i in range(direction.length):
 *                 f = direction.numbers[i] * k
 */
    __pyx_v_k = (__pyx_v_k * __pyx_v_self->strength);

    /* "flitter/render/physics.pyx":210
 *                 k **= self.power
 *             k *= self.strength
 *             for i in range(direction.length):             # <<<<<<<<<<<<<<
 *                 f = direction.numbers[i] * k
 *                 from_particle.force.numbers[i] = from_particle.force.numbers[i] - f
 */
    __pyx_t_3 = __pyx_v_direction->length;
    __pyx_t_4 = __pyx_t_3;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
      __pyx_v_i = __pyx_t_5;

      /* "flitter/render/physics.pyx":211
 *             k *= self.strength
 *             for i in range(direction.length):
 *                 f = direction.numbers[i] * k             # <<<<<<<<<<<<<<
 *                 from_particle.force.numbers[i] = from_particle.force.numbers[i] - f
 *                 to_particle.force.numbers[i] = to_particle.force.numbers[i] + f
 */
      __pyx_v_f = ((__pyx_v_direction->numbers[__pyx_v_i]) * __pyx_v_k);

      /* "flitter/render/physics.pyx":212
 *             for i in range(direction.length):
 *                 f = direction.numbers[i] * k
 *                 from_particle.force.numbers[i] = from_particle.force.numbers[i] - f             # <<<<<<<<<<<<<<
 *                 to_particle.force.numbers[i] = to_particle.force.numbers[i] + f
 *         elif self.maximum and distance > self.maximum:
 */
      (__pyx_v_from_particle->force->numbers[__pyx_v_i]) = ((__pyx_v_from_particle->force->numbers[__pyx_v_i]) - __pyx_v_f);

      /* "flitter/render/physics.pyx":213
 *                 f = direction.numbers[i] * k
 *                 from_particle.force.numbers[i] = from_particle.force.numbers[i] - f
 *                 to_particle.force.numbers[i] = to_particle.force.numbers[i] + f             # <<<<<<<<<<<<<<
 *         elif self.maximum and distance > self.maximum:
 *             k = distance - self.maximum
 */
      (__pyx_v_to_particle->force->numbers[__pyx_v_i]) = ((__pyx_v_to_particle->force->numbers[__pyx_v_i]) + __pyx_v_f);
    }

    /* "flitter/render/physics.pyx":205
 *         cdef double f, k
 *         cdef int64_t i
 *         if self.minimum and distance < self.minimum:             # <<<<<<<<<<<<<<
 *             k = self.minimum - distance
 *             if self.power != 1:
 */
    goto __pyx_L3;
  }

  /* "flitter/render/physics.pyx":214
 *                 from_particle.force.numbers[i] = from_particle.force.numbers[i] - f
 *                 to_particle.force.numbers[i] = to_particle.force.numbers[i] + f
 *         elif self.maximum and distance > self.maximum:             # <<<<<<<<<<<<<<
 *             k = distance - self.maximum
 *             if self.power != 1:
 */
  __pyx_t_2 = (__pyx_v_self->maximum != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L9_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_distance > __pyx_v_self->maximum);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L9_bool_binop_done:;
  if (__pyx_t_1) {

    /* "flitter/render/physics.pyx":215
 *                 to_particle.force.numbers[i] = to_particle.force.numbers[i] + f
 *         elif self.maximum and distance > self.maximum:
 *             k = distance - self.maximum             # <<<<<<<<<<<<<<
 *             if self.power != 1:
 *                 k **= self.power
 */
    __pyx_v_k = (__pyx_v_distance - __pyx_v_self->maximum);

    /* "flitter/render/physics.pyx":216
 *         elif self.maximum and distance > self.maximum:
 *             k = distance - self.maximum
 *             if self.power != 1:             # <<<<<<<<<<<<<<
 *                 k **= self.power
 *             k *= self.strength
 */
    __pyx_t_1 = (__pyx_v_self->power != 1.0);
    if (__pyx_t_1) {

      /* "flitter/render/physics.pyx":217
 *             k = distance - self.maximum
 *             if self.power != 1:
 *                 k **= self.power             # <<<<<<<<<<<<<<
 *             k *= self.strength
 *             for i in range(direction.length):
 */
      __pyx_v_k = pow(__pyx_v_k, __pyx_v_self->power);

      /* "flitter/render/physics.pyx":216
 *         elif self.maximum and distance > self.maximum:
 *             k = distance - self.maximum
 *             if self.power != 1:             # <<<<<<<<<<<<<<
 *                 k **= self.power
 *             k *= self.strength
 */
    }

    /* "flitter/render/physics.pyx":218
 *             if self.power != 1:
 *                 k **= self.power
 *             k *= self.strength             # <<<<<<<<<<<<<<
 *             for i in range(direction.length):
 *                 f = direction.numbers[i] * k
 */
    __pyx_v_k = (__pyx_v_k * __pyx_v_self->strength);

    /* "flitter/render/physics.pyx":219
 *                 k **= self.power
 *             k *= self.strength
 *             for i in range(direction.length):             # <<<<<<<<<<<<<<
 *                 f = direction.numbers[i] * k
 *                 from_particle.force.numbers[i] = from_particle.force.numbers[i] + f
 */
    __pyx_t_3 = __pyx_v_direction->length;
    __pyx_t_4 = __pyx_t_3;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
      __pyx_v_i = __pyx_t_5;

      /* "flitter/render/physics.pyx":220
 *             k *= self.strength
 *             for i in range(direction.length):
 *                 f = direction.numbers[i] * k             # <<<<<<<<<<<<<<
 *                 from_particle.force.numbers[i] = from_particle.force.numbers[i] + f
 *                 to_particle.force.numbers[i] = to_particle.force.numbers[i] - f
 */
      __pyx_v_f = ((__pyx_v_direction->numbers[__pyx_v_i]) * __pyx_v_k);

      /* "flitter/render/physics.pyx":221
 *             for i in range(direction.length):
 *                 f = direction.numbers[i] * k
 *                 from_particle.force.numbers[i] = from_particle.force.numbers[i] + f             # <<<<<<<<<<<<<<
 *                 to_particle.force.numbers[i] = to_particle.force.numbers[i] - f
 * 
 */
      (__pyx_v_from_particle->force->numbers[__pyx_v_i]) = ((__pyx_v_from_particle->force->numbers[__pyx_v_i]) + __pyx_v_f);

      /* "flitter/render/physics.pyx":222
 *                 f = direction.numbers[i] * k
 *                 from_particle.force.numbers[i] = from_particle.force.numbers[i] + f
 *                 to_particle.force.numbers[i] = to_particle.force.numbers[i] - f             # <<<<<<<<<<<<<<
 * 
 * 
 */
      (__pyx_v_to_particle->force->numbers[__pyx_v_i]) = ((__pyx_v_to_particle->force->numbers[__pyx_v_i]) - __pyx_v_f);
    }

    /* "flitter/render/physics.pyx":214
 *                 from_particle.force.numbers[i] = from_particle.force.numbers[i] - f
 *                 to_particle.force.numbers[i] = to_particle.force.numbers[i] + f
 *         elif self.maximum and distance > self.maximum:             # <<<<<<<<<<<<<<
 *             k = distance - self.maximum
 *             if self.power != 1:
 */
  }
  __pyx_L3:;

  /* "flitter/render/physics.pyx":202
 * 
 *     @cython.profile(False)
 *     cdef void apply(self, Particle from_particle, Particle to_particle, Vector direction, double distance) noexcept nogil:             # <<<<<<<<<<<<<<
 *         cdef double f, k
 *         cdef int64_t i
 */

  /* function exit code */
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_6render_7physics_20DistanceForceApplier_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_6render_7physics_20DistanceForceApplier_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_20DistanceForceApplier_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_6render_7physics_20DistanceForceApplier_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_6render_7physics_20DistanceForceApplier___reduce_cython__(((struct __pyx_obj_7flitter_6render_7physics_DistanceForceApplier *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_6render_7physics_20DistanceForceApplier___reduce_cython__(struct __pyx_obj_7flitter_6render_7physics_DistanceForceApplier *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.from_particle_id, self.maximum, self.minimum, self.power, self.strength, self.to_particle_id)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->maximum); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->minimum); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->power); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_self->strength); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF((PyObject *)__pyx_v_self->from_particle_id);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->from_particle_id);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)__pyx_v_self->from_particle_id))) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 4, __pyx_t_4)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->to_particle_id);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->to_particle_id);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 5, ((PyObject *)__pyx_v_self->to_particle_id))) __PYX_ERR(0, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.from_particle_id, self.maximum, self.minimum, self.power, self.strength, self.to_particle_id)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_5 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_v__dict = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "(tree fragment)":7
 *     state = (self.from_particle_id, self.maximum, self.minimum, self.power, self.strength, self.to_particle_id)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_6 = (__pyx_v__dict != Py_None);
  if (__pyx_t_6) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v__dict)) __PYX_ERR(0, 8, __pyx_L1_error);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.from_particle_id is not None or self.to_particle_id is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.from_particle_id, self.maximum, self.minimum, self.power, self.strength, self.to_particle_id)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.from_particle_id is not None or self.to_particle_id is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_DistanceForceApplier, (type(self), 0xca186ab, None), state
 */
  /*else*/ {
    __pyx_t_7 = (((PyObject *)__pyx_v_self->from_particle_id) != Py_None);
    if (!__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_7 = (((PyObject *)__pyx_v_self->to_particle_id) != Py_None);
    __pyx_t_6 = __pyx_t_7;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_6;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.from_particle_id is not None or self.to_particle_id is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_DistanceForceApplier, (type(self), 0xca186ab, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.from_particle_id is not None or self.to_particle_id is not None
 *     if use_setstate:
 *         return __pyx_unpickle_DistanceForceApplier, (type(self), 0xca186ab, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_DistanceForceApplier, (type(self), 0xca186ab, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_DistanceForceAppl); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_211912363);
    __Pyx_GIVEREF(__pyx_int_211912363);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_211912363)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, Py_None)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_5)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.from_particle_id is not None or self.to_particle_id is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_DistanceForceApplier, (type(self), 0xca186ab, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_DistanceForceApplier, (type(self), 0xca186ab, None), state
 *     else:
 *         return __pyx_unpickle_DistanceForceApplier, (type(self), 0xca186ab, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_DistanceForceApplier__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_DistanceForceAppl); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_211912363);
    __Pyx_GIVEREF(__pyx_int_211912363);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_211912363)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_5)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_5 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("flitter.render.physics.DistanceForceApplier.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_DistanceForceApplier, (type(self), 0xca186ab, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_DistanceForceApplier__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_6render_7physics_20DistanceForceApplier_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_6render_7physics_20DistanceForceApplier_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_20DistanceForceApplier_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_6render_7physics_20DistanceForceApplier_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.render.physics.DistanceForceApplier.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_6render_7physics_20DistanceForceApplier_2__setstate_cython__(((struct __pyx_obj_7flitter_6render_7physics_DistanceForceApplier *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_6render_7physics_20DistanceForceApplier_2__setstate_cython__(struct __pyx_obj_7flitter_6render_7physics_DistanceForceApplier *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_DistanceForceApplier, (type(self), 0xca186ab, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_DistanceForceApplier__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_6render_7physics___pyx_unpickle_DistanceForceApplier__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_DistanceForceApplier, (type(self), 0xca186ab, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_DistanceForceApplier__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.render.physics.DistanceForceApplier.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/render/physics.pyx":228
 *     cdef Vector flow
 * 
 *     @cython.profile(False)             # <<<<<<<<<<<<<<
 *     def __cinit__(self, Node node, double strength, Vector zero):
 *         self.flow = node.get_fvec('flow', zero.length, zero)
 */

/* Python wrapper */
static int __pyx_pw_7flitter_6render_7physics_16DragForceApplier_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_6render_7physics_16DragForceApplier_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_7flitter_5model_Node *__pyx_v_node = 0;
  CYTHON_UNUSED double __pyx_v_strength;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_zero = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_node,&__pyx_n_s_strength,&__pyx_n_s_zero,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_node)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 228, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_strength)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 228, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, 1); __PYX_ERR(1, 228, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_zero)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 228, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, 2); __PYX_ERR(1, 228, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__cinit__") < 0)) __PYX_ERR(1, 228, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
      values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
    }
    __pyx_v_node = ((struct __pyx_obj_7flitter_5model_Node *)values[0]);
    __pyx_v_strength = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_strength == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 229, __pyx_L3_error)
    __pyx_v_zero = ((struct __pyx_obj_7flitter_5model_Vector *)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 228, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.render.physics.DragForceApplier.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_node), __pyx_ptype_7flitter_5model_Node, 1, "node", 0))) __PYX_ERR(1, 229, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_zero), __pyx_ptype_7flitter_5model_Vector, 1, "zero", 0))) __PYX_ERR(1, 229, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_6render_7physics_16DragForceApplier___cinit__(((struct __pyx_obj_7flitter_6render_7physics_DragForceApplier *)__pyx_v_self), __pyx_v_node, __pyx_v_strength, __pyx_v_zero);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_6render_7physics_16DragForceApplier___cinit__(struct __pyx_obj_7flitter_6render_7physics_DragForceApplier *__pyx_v_self, struct __pyx_obj_7flitter_5model_Node *__pyx_v_node, CYTHON_UNUSED double __pyx_v_strength, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_zero) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 1);

  /* "flitter/render/physics.pyx":230
 *     @cython.profile(False)
 *     def __cinit__(self, Node node, double strength, Vector zero):
 *         self.flow = node.get_fvec('flow', zero.length, zero)             # <<<<<<<<<<<<<<
 * 
 *     @cython.profile(False)
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Node *)__pyx_v_node->__pyx_vtab)->get_fvec(__pyx_v_node, __pyx_n_u_flow, __pyx_v_zero->length, __pyx_v_zero)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->flow);
  __Pyx_DECREF((PyObject *)__pyx_v_self->flow);
  __pyx_v_self->flow = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/render/physics.pyx":228
 *     cdef Vector flow
 * 
 *     @cython.profile(False)             # <<<<<<<<<<<<<<
 *     def __cinit__(self, Node node, double strength, Vector zero):
 *         self.flow = node.get_fvec('flow', zero.length, zero)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.render.physics.DragForceApplier.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/render/physics.pyx":233
 * 
 *     @cython.profile(False)
 *     cdef void apply(self, Particle particle, double delta) noexcept nogil:             # <<<<<<<<<<<<<<
 *         cdef double speed_squared=0, v, k
 *         cdef int64_t i
 */

static void __pyx_f_7flitter_6render_7physics_16DragForceApplier_apply(struct __pyx_obj_7flitter_6render_7physics_DragForceApplier *__pyx_v_self, struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_particle, double __pyx_v_delta) {
  double __pyx_v_speed_squared;
  double __pyx_v_v;
  double __pyx_v_k;
  int64_t __pyx_v_i;
  int __pyx_t_1;
  int64_t __pyx_t_2;
  int64_t __pyx_t_3;
  int64_t __pyx_t_4;
  double __pyx_t_5;
  double __pyx_t_6;
  double __pyx_t_7;

  /* "flitter/render/physics.pyx":234
 *     @cython.profile(False)
 *     cdef void apply(self, Particle particle, double delta) noexcept nogil:
 *         cdef double speed_squared=0, v, k             # <<<<<<<<<<<<<<
 *         cdef int64_t i
 *         if particle.radius:
 */
  __pyx_v_speed_squared = 0.0;

  /* "flitter/render/physics.pyx":236
 *         cdef double speed_squared=0, v, k
 *         cdef int64_t i
 *         if particle.radius:             # <<<<<<<<<<<<<<
 *             for i in range(particle.velocity.length):
 *                 v = particle.velocity.numbers[i] - self.flow.numbers[i]
 */
  __pyx_t_1 = (__pyx_v_particle->radius != 0);
  if (__pyx_t_1) {

    /* "flitter/render/physics.pyx":237
 *         cdef int64_t i
 *         if particle.radius:
 *             for i in range(particle.velocity.length):             # <<<<<<<<<<<<<<
 *                 v = particle.velocity.numbers[i] - self.flow.numbers[i]
 *                 speed_squared += v * v
 */
    __pyx_t_2 = __pyx_v_particle->velocity->length;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_i = __pyx_t_4;

      /* "flitter/render/physics.pyx":238
 *         if particle.radius:
 *             for i in range(particle.velocity.length):
 *                 v = particle.velocity.numbers[i] - self.flow.numbers[i]             # <<<<<<<<<<<<<<
 *                 speed_squared += v * v
 *             k = min(self.strength * sqrt(speed_squared) * particle.radius**(particle.force.length-1), particle.mass / delta)
 */
      __pyx_v_v = ((__pyx_v_particle->velocity->numbers[__pyx_v_i]) - (__pyx_v_self->flow->numbers[__pyx_v_i]));

      /* "flitter/render/physics.pyx":239
 *             for i in range(particle.velocity.length):
 *                 v = particle.velocity.numbers[i] - self.flow.numbers[i]
 *                 speed_squared += v * v             # <<<<<<<<<<<<<<
 *             k = min(self.strength * sqrt(speed_squared) * particle.radius**(particle.force.length-1), particle.mass / delta)
 *             for i in range(particle.velocity.length):
 */
      __pyx_v_speed_squared = (__pyx_v_speed_squared + (__pyx_v_v * __pyx_v_v));
    }

    /* "flitter/render/physics.pyx":240
 *                 v = particle.velocity.numbers[i] - self.flow.numbers[i]
 *                 speed_squared += v * v
 *             k = min(self.strength * sqrt(speed_squared) * particle.radius**(particle.force.length-1), particle.mass / delta)             # <<<<<<<<<<<<<<
 *             for i in range(particle.velocity.length):
 *                 particle.force.numbers[i] = particle.force.numbers[i] - (particle.velocity.numbers[i] - self.flow.numbers[i]) * k
 */
    __pyx_t_5 = (__pyx_v_particle->mass / __pyx_v_delta);
    __pyx_t_6 = ((__pyx_v_self->__pyx_base.strength * sqrt(__pyx_v_speed_squared)) * pow(__pyx_v_particle->radius, ((double)(__pyx_v_particle->force->length - 1))));
    __pyx_t_1 = (__pyx_t_5 < __pyx_t_6);
    if (__pyx_t_1) {
      __pyx_t_7 = __pyx_t_5;
    } else {
      __pyx_t_7 = __pyx_t_6;
    }
    __pyx_v_k = __pyx_t_7;

    /* "flitter/render/physics.pyx":241
 *                 speed_squared += v * v
 *             k = min(self.strength * sqrt(speed_squared) * particle.radius**(particle.force.length-1), particle.mass / delta)
 *             for i in range(particle.velocity.length):             # <<<<<<<<<<<<<<
 *                 particle.force.numbers[i] = particle.force.numbers[i] - (particle.velocity.numbers[i] - self.flow.numbers[i]) * k
 * 
 */
    __pyx_t_2 = __pyx_v_particle->velocity->length;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_i = __pyx_t_4;

      /* "flitter/render/physics.pyx":242
 *             k = min(self.strength * sqrt(speed_squared) * particle.radius**(particle.force.length-1), particle.mass / delta)
 *             for i in range(particle.velocity.length):
 *                 particle.force.numbers[i] = particle.force.numbers[i] - (particle.velocity.numbers[i] - self.flow.numbers[i]) * k             # <<<<<<<<<<<<<<
 * 
 * 
 */
      (__pyx_v_particle->force->numbers[__pyx_v_i]) = ((__pyx_v_particle->force->numbers[__pyx_v_i]) - (((__pyx_v_particle->velocity->numbers[__pyx_v_i]) - (__pyx_v_self->flow->numbers[__pyx_v_i])) * __pyx_v_k));
    }

    /* "flitter/render/physics.pyx":236
 *         cdef double speed_squared=0, v, k
 *         cdef int64_t i
 *         if particle.radius:             # <<<<<<<<<<<<<<
 *             for i in range(particle.velocity.length):
 *                 v = particle.velocity.numbers[i] - self.flow.numbers[i]
 */
  }

  /* "flitter/render/physics.pyx":233
 * 
 *     @cython.profile(False)
 *     cdef void apply(self, Particle particle, double delta) noexcept nogil:             # <<<<<<<<<<<<<<
 *         cdef double speed_squared=0, v, k
 *         cdef int64_t i
 */

  /* function exit code */
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_6render_7physics_16DragForceApplier_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_6render_7physics_16DragForceApplier_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_16DragForceApplier_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_6render_7physics_16DragForceApplier_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_6render_7physics_16DragForceApplier_2__reduce_cython__(((struct __pyx_obj_7flitter_6render_7physics_DragForceApplier *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_6render_7physics_16DragForceApplier_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_DragForceApplier *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.render.physics.DragForceApplier.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_6render_7physics_16DragForceApplier_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_6render_7physics_16DragForceApplier_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_16DragForceApplier_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_6render_7physics_16DragForceApplier_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.render.physics.DragForceApplier.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_6render_7physics_16DragForceApplier_4__setstate_cython__(((struct __pyx_obj_7flitter_6render_7physics_DragForceApplier *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_6render_7physics_16DragForceApplier_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_DragForceApplier *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.render.physics.DragForceApplier.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/render/physics.pyx":249
 *     cdef Vector gravity
 * 
 *     @cython.profile(False)             # <<<<<<<<<<<<<<
 *     def __cinit__(self, Node node, double strength, Vector zero):
 *         self.density = node.get_float('density', 1)
 */

/* Python wrapper */
static int __pyx_pw_7flitter_6render_7physics_20BuoyancyForceApplier_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_6render_7physics_20BuoyancyForceApplier_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_7flitter_5model_Node *__pyx_v_node = 0;
  CYTHON_UNUSED double __pyx_v_strength;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_zero = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_node,&__pyx_n_s_strength,&__pyx_n_s_zero,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_node)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 249, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_strength)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 249, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, 1); __PYX_ERR(1, 249, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_zero)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 249, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, 2); __PYX_ERR(1, 249, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__cinit__") < 0)) __PYX_ERR(1, 249, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
      values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
    }
    __pyx_v_node = ((struct __pyx_obj_7flitter_5model_Node *)values[0]);
    __pyx_v_strength = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_strength == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 250, __pyx_L3_error)
    __pyx_v_zero = ((struct __pyx_obj_7flitter_5model_Vector *)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 249, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.render.physics.BuoyancyForceApplier.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_node), __pyx_ptype_7flitter_5model_Node, 1, "node", 0))) __PYX_ERR(1, 250, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_zero), __pyx_ptype_7flitter_5model_Vector, 1, "zero", 0))) __PYX_ERR(1, 250, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_6render_7physics_20BuoyancyForceApplier___cinit__(((struct __pyx_obj_7flitter_6render_7physics_BuoyancyForceApplier *)__pyx_v_self), __pyx_v_node, __pyx_v_strength, __pyx_v_zero);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_6render_7physics_20BuoyancyForceApplier___cinit__(struct __pyx_obj_7flitter_6render_7physics_BuoyancyForceApplier *__pyx_v_self, struct __pyx_obj_7flitter_5model_Node *__pyx_v_node, CYTHON_UNUSED double __pyx_v_strength, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_zero) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 1);

  /* "flitter/render/physics.pyx":251
 *     @cython.profile(False)
 *     def __cinit__(self, Node node, double strength, Vector zero):
 *         self.density = node.get_float('density', 1)             # <<<<<<<<<<<<<<
 *         self.gravity = node.get_fvec('gravity', zero.length, zero)
 *         if self.gravity is zero:
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_7flitter_5model_Node *)__pyx_v_node->__pyx_vtab)->get_float(__pyx_v_node, __pyx_n_u_density, 1.0); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 251, __pyx_L1_error)
  __pyx_v_self->density = __pyx_t_1;

  /* "flitter/render/physics.pyx":252
 *     def __cinit__(self, Node node, double strength, Vector zero):
 *         self.density = node.get_float('density', 1)
 *         self.gravity = node.get_fvec('gravity', zero.length, zero)             # <<<<<<<<<<<<<<
 *         if self.gravity is zero:
 *             self.gravity = self.gravity.copy()
 */
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Node *)__pyx_v_node->__pyx_vtab)->get_fvec(__pyx_v_node, __pyx_n_u_gravity, __pyx_v_zero->length, __pyx_v_zero)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->gravity);
  __Pyx_DECREF((PyObject *)__pyx_v_self->gravity);
  __pyx_v_self->gravity = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "flitter/render/physics.pyx":253
 *         self.density = node.get_float('density', 1)
 *         self.gravity = node.get_fvec('gravity', zero.length, zero)
 *         if self.gravity is zero:             # <<<<<<<<<<<<<<
 *             self.gravity = self.gravity.copy()
 *             self.gravity.numbers[zero.length-1] = -1
 */
  __pyx_t_3 = (__pyx_v_self->gravity == __pyx_v_zero);
  if (__pyx_t_3) {

    /* "flitter/render/physics.pyx":254
 *         self.gravity = node.get_fvec('gravity', zero.length, zero)
 *         if self.gravity is zero:
 *             self.gravity = self.gravity.copy()             # <<<<<<<<<<<<<<
 *             self.gravity.numbers[zero.length-1] = -1
 * 
 */
    __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_self->gravity->__pyx_vtab)->copy(__pyx_v_self->gravity, 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 254, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF((PyObject *)__pyx_v_self->gravity);
    __Pyx_DECREF((PyObject *)__pyx_v_self->gravity);
    __pyx_v_self->gravity = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "flitter/render/physics.pyx":255
 *         if self.gravity is zero:
 *             self.gravity = self.gravity.copy()
 *             self.gravity.numbers[zero.length-1] = -1             # <<<<<<<<<<<<<<
 * 
 *     @cython.profile(False)
 */
    (__pyx_v_self->gravity->numbers[(__pyx_v_zero->length - 1)]) = -1.0;

    /* "flitter/render/physics.pyx":253
 *         self.density = node.get_float('density', 1)
 *         self.gravity = node.get_fvec('gravity', zero.length, zero)
 *         if self.gravity is zero:             # <<<<<<<<<<<<<<
 *             self.gravity = self.gravity.copy()
 *             self.gravity.numbers[zero.length-1] = -1
 */
  }

  /* "flitter/render/physics.pyx":249
 *     cdef Vector gravity
 * 
 *     @cython.profile(False)             # <<<<<<<<<<<<<<
 *     def __cinit__(self, Node node, double strength, Vector zero):
 *         self.density = node.get_float('density', 1)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("flitter.render.physics.BuoyancyForceApplier.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/render/physics.pyx":258
 * 
 *     @cython.profile(False)
 *     cdef void apply(self, Particle particle, double delta) noexcept nogil:             # <<<<<<<<<<<<<<
 *         cdef double displaced_mass, k
 *         cdef int64_t i
 */

static void __pyx_f_7flitter_6render_7physics_20BuoyancyForceApplier_apply(struct __pyx_obj_7flitter_6render_7physics_BuoyancyForceApplier *__pyx_v_self, struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_particle, CYTHON_UNUSED double __pyx_v_delta) {
  double __pyx_v_displaced_mass;
  double __pyx_v_k;
  int64_t __pyx_v_i;
  int __pyx_t_1;
  int __pyx_t_2;
  int64_t __pyx_t_3;
  int64_t __pyx_t_4;
  int64_t __pyx_t_5;

  /* "flitter/render/physics.pyx":261
 *         cdef double displaced_mass, k
 *         cdef int64_t i
 *         if particle.radius and particle.mass:             # <<<<<<<<<<<<<<
 *             displaced_mass = particle.radius**particle.force.length * self.density
 *             k = self.strength * (particle.mass - displaced_mass)
 */
  __pyx_t_2 = (__pyx_v_particle->radius != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_particle->mass != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "flitter/render/physics.pyx":262
 *         cdef int64_t i
 *         if particle.radius and particle.mass:
 *             displaced_mass = particle.radius**particle.force.length * self.density             # <<<<<<<<<<<<<<
 *             k = self.strength * (particle.mass - displaced_mass)
 *             for i in range(particle.force.length):
 */
    __pyx_v_displaced_mass = (pow(__pyx_v_particle->radius, ((double)__pyx_v_particle->force->length)) * __pyx_v_self->density);

    /* "flitter/render/physics.pyx":263
 *         if particle.radius and particle.mass:
 *             displaced_mass = particle.radius**particle.force.length * self.density
 *             k = self.strength * (particle.mass - displaced_mass)             # <<<<<<<<<<<<<<
 *             for i in range(particle.force.length):
 *                 particle.force.numbers[i] = particle.force.numbers[i] + self.gravity.numbers[i] * k
 */
    __pyx_v_k = (__pyx_v_self->__pyx_base.strength * (__pyx_v_particle->mass - __pyx_v_displaced_mass));

    /* "flitter/render/physics.pyx":264
 *             displaced_mass = particle.radius**particle.force.length * self.density
 *             k = self.strength * (particle.mass - displaced_mass)
 *             for i in range(particle.force.length):             # <<<<<<<<<<<<<<
 *                 particle.force.numbers[i] = particle.force.numbers[i] + self.gravity.numbers[i] * k
 * 
 */
    __pyx_t_3 = __pyx_v_particle->force->length;
    __pyx_t_4 = __pyx_t_3;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
      __pyx_v_i = __pyx_t_5;

      /* "flitter/render/physics.pyx":265
 *             k = self.strength * (particle.mass - displaced_mass)
 *             for i in range(particle.force.length):
 *                 particle.force.numbers[i] = particle.force.numbers[i] + self.gravity.numbers[i] * k             # <<<<<<<<<<<<<<
 * 
 * 
 */
      (__pyx_v_particle->force->numbers[__pyx_v_i]) = ((__pyx_v_particle->force->numbers[__pyx_v_i]) + ((__pyx_v_self->gravity->numbers[__pyx_v_i]) * __pyx_v_k));
    }

    /* "flitter/render/physics.pyx":261
 *         cdef double displaced_mass, k
 *         cdef int64_t i
 *         if particle.radius and particle.mass:             # <<<<<<<<<<<<<<
 *             displaced_mass = particle.radius**particle.force.length * self.density
 *             k = self.strength * (particle.mass - displaced_mass)
 */
  }

  /* "flitter/render/physics.pyx":258
 * 
 *     @cython.profile(False)
 *     cdef void apply(self, Particle particle, double delta) noexcept nogil:             # <<<<<<<<<<<<<<
 *         cdef double displaced_mass, k
 *         cdef int64_t i
 */

  /* function exit code */
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_6render_7physics_20BuoyancyForceApplier_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_6render_7physics_20BuoyancyForceApplier_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_20BuoyancyForceApplier_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_6render_7physics_20BuoyancyForceApplier_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_6render_7physics_20BuoyancyForceApplier_2__reduce_cython__(((struct __pyx_obj_7flitter_6render_7physics_BuoyancyForceApplier *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_6render_7physics_20BuoyancyForceApplier_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_BuoyancyForceApplier *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.render.physics.BuoyancyForceApplier.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_6render_7physics_20BuoyancyForceApplier_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_6render_7physics_20BuoyancyForceApplier_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_20BuoyancyForceApplier_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_6render_7physics_20BuoyancyForceApplier_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.render.physics.BuoyancyForceApplier.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_6render_7physics_20BuoyancyForceApplier_4__setstate_cython__(((struct __pyx_obj_7flitter_6render_7physics_BuoyancyForceApplier *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_6render_7physics_20BuoyancyForceApplier_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_BuoyancyForceApplier *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.render.physics.BuoyancyForceApplier.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/render/physics.pyx":272
 *     cdef Vector acceleration
 * 
 *     @cython.profile(False)             # <<<<<<<<<<<<<<
 *     def __cinit__(self, Node node, double strength, Vector zero):
 *         self.force = node.get_fvec('force', zero.length, zero)
 */

/* Python wrapper */
static int __pyx_pw_7flitter_6render_7physics_20ConstantForceApplier_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_6render_7physics_20ConstantForceApplier_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_7flitter_5model_Node *__pyx_v_node = 0;
  CYTHON_UNUSED double __pyx_v_strength;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_zero = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_node,&__pyx_n_s_strength,&__pyx_n_s_zero,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_node)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 272, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_strength)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 272, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, 1); __PYX_ERR(1, 272, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_zero)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 272, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, 2); __PYX_ERR(1, 272, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__cinit__") < 0)) __PYX_ERR(1, 272, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
      values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
    }
    __pyx_v_node = ((struct __pyx_obj_7flitter_5model_Node *)values[0]);
    __pyx_v_strength = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_strength == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 273, __pyx_L3_error)
    __pyx_v_zero = ((struct __pyx_obj_7flitter_5model_Vector *)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 272, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.render.physics.ConstantForceApplier.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_node), __pyx_ptype_7flitter_5model_Node, 1, "node", 0))) __PYX_ERR(1, 273, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_zero), __pyx_ptype_7flitter_5model_Vector, 1, "zero", 0))) __PYX_ERR(1, 273, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_6render_7physics_20ConstantForceApplier___cinit__(((struct __pyx_obj_7flitter_6render_7physics_ConstantForceApplier *)__pyx_v_self), __pyx_v_node, __pyx_v_strength, __pyx_v_zero);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_6render_7physics_20ConstantForceApplier___cinit__(struct __pyx_obj_7flitter_6render_7physics_ConstantForceApplier *__pyx_v_self, struct __pyx_obj_7flitter_5model_Node *__pyx_v_node, CYTHON_UNUSED double __pyx_v_strength, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_zero) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 1);

  /* "flitter/render/physics.pyx":274
 *     @cython.profile(False)
 *     def __cinit__(self, Node node, double strength, Vector zero):
 *         self.force = node.get_fvec('force', zero.length, zero)             # <<<<<<<<<<<<<<
 *         self.acceleration = node.get_fvec('acceleration', zero.length, zero)
 * 
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Node *)__pyx_v_node->__pyx_vtab)->get_fvec(__pyx_v_node, __pyx_n_u_force, __pyx_v_zero->length, __pyx_v_zero)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->force);
  __Pyx_DECREF((PyObject *)__pyx_v_self->force);
  __pyx_v_self->force = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/render/physics.pyx":275
 *     def __cinit__(self, Node node, double strength, Vector zero):
 *         self.force = node.get_fvec('force', zero.length, zero)
 *         self.acceleration = node.get_fvec('acceleration', zero.length, zero)             # <<<<<<<<<<<<<<
 * 
 *     @cython.profile(False)
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Node *)__pyx_v_node->__pyx_vtab)->get_fvec(__pyx_v_node, __pyx_n_u_acceleration, __pyx_v_zero->length, __pyx_v_zero)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->acceleration);
  __Pyx_DECREF((PyObject *)__pyx_v_self->acceleration);
  __pyx_v_self->acceleration = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/render/physics.pyx":272
 *     cdef Vector acceleration
 * 
 *     @cython.profile(False)             # <<<<<<<<<<<<<<
 *     def __cinit__(self, Node node, double strength, Vector zero):
 *         self.force = node.get_fvec('force', zero.length, zero)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.render.physics.ConstantForceApplier.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/render/physics.pyx":278
 * 
 *     @cython.profile(False)
 *     cdef void apply(self, Particle particle, double delta) noexcept nogil:             # <<<<<<<<<<<<<<
 *         cdef int64_t i
 *         cdef double f
 */

static void __pyx_f_7flitter_6render_7physics_20ConstantForceApplier_apply(struct __pyx_obj_7flitter_6render_7physics_ConstantForceApplier *__pyx_v_self, struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_particle, CYTHON_UNUSED double __pyx_v_delta) {
  int64_t __pyx_v_i;
  double __pyx_v_f;
  int64_t __pyx_t_1;
  int64_t __pyx_t_2;
  int64_t __pyx_t_3;

  /* "flitter/render/physics.pyx":281
 *         cdef int64_t i
 *         cdef double f
 *         for i in range(self.force.length):             # <<<<<<<<<<<<<<
 *             f = self.force.numbers[i] + self.acceleration.numbers[i] * particle.mass
 *             particle.force.numbers[i] = particle.force.numbers[i] + f*self.strength
 */
  __pyx_t_1 = __pyx_v_self->force->length;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "flitter/render/physics.pyx":282
 *         cdef double f
 *         for i in range(self.force.length):
 *             f = self.force.numbers[i] + self.acceleration.numbers[i] * particle.mass             # <<<<<<<<<<<<<<
 *             particle.force.numbers[i] = particle.force.numbers[i] + f*self.strength
 * 
 */
    __pyx_v_f = ((__pyx_v_self->force->numbers[__pyx_v_i]) + ((__pyx_v_self->acceleration->numbers[__pyx_v_i]) * __pyx_v_particle->mass));

    /* "flitter/render/physics.pyx":283
 *         for i in range(self.force.length):
 *             f = self.force.numbers[i] + self.acceleration.numbers[i] * particle.mass
 *             particle.force.numbers[i] = particle.force.numbers[i] + f*self.strength             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (__pyx_v_particle->force->numbers[__pyx_v_i]) = ((__pyx_v_particle->force->numbers[__pyx_v_i]) + (__pyx_v_f * __pyx_v_self->__pyx_base.strength));
  }

  /* "flitter/render/physics.pyx":278
 * 
 *     @cython.profile(False)
 *     cdef void apply(self, Particle particle, double delta) noexcept nogil:             # <<<<<<<<<<<<<<
 *         cdef int64_t i
 *         cdef double f
 */

  /* function exit code */
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_6render_7physics_20ConstantForceApplier_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_6render_7physics_20ConstantForceApplier_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_20ConstantForceApplier_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_6render_7physics_20ConstantForceApplier_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_6render_7physics_20ConstantForceApplier_2__reduce_cython__(((struct __pyx_obj_7flitter_6render_7physics_ConstantForceApplier *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_6render_7physics_20ConstantForceApplier_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_ConstantForceApplier *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.render.physics.ConstantForceApplier.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_6render_7physics_20ConstantForceApplier_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_6render_7physics_20ConstantForceApplier_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_20ConstantForceApplier_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_6render_7physics_20ConstantForceApplier_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.render.physics.ConstantForceApplier.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_6render_7physics_20ConstantForceApplier_4__setstate_cython__(((struct __pyx_obj_7flitter_6render_7physics_ConstantForceApplier *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_6render_7physics_20ConstantForceApplier_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_ConstantForceApplier *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.render.physics.ConstantForceApplier.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/render/physics.pyx":290
 *     cdef uint64_t index
 * 
 *     @cython.profile(False)             # <<<<<<<<<<<<<<
 *     def __cinit__(self, Node node, double strength, Vector zero):
 *         self.index = 0
 */

/* Python wrapper */
static int __pyx_pw_7flitter_6render_7physics_18RandomForceApplier_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_6render_7physics_18RandomForceApplier_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED struct __pyx_obj_7flitter_5model_Node *__pyx_v_node = 0;
  CYTHON_UNUSED double __pyx_v_strength;
  CYTHON_UNUSED struct __pyx_obj_7flitter_5model_Vector *__pyx_v_zero = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_node,&__pyx_n_s_strength,&__pyx_n_s_zero,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_node)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 290, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_strength)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 290, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, 1); __PYX_ERR(1, 290, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_zero)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 290, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, 2); __PYX_ERR(1, 290, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__cinit__") < 0)) __PYX_ERR(1, 290, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
      values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
    }
    __pyx_v_node = ((struct __pyx_obj_7flitter_5model_Node *)values[0]);
    __pyx_v_strength = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_strength == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 291, __pyx_L3_error)
    __pyx_v_zero = ((struct __pyx_obj_7flitter_5model_Vector *)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 290, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.render.physics.RandomForceApplier.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_node), __pyx_ptype_7flitter_5model_Node, 1, "node", 0))) __PYX_ERR(1, 291, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_zero), __pyx_ptype_7flitter_5model_Vector, 1, "zero", 0))) __PYX_ERR(1, 291, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_6render_7physics_18RandomForceApplier___cinit__(((struct __pyx_obj_7flitter_6render_7physics_RandomForceApplier *)__pyx_v_self), __pyx_v_node, __pyx_v_strength, __pyx_v_zero);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_6render_7physics_18RandomForceApplier___cinit__(struct __pyx_obj_7flitter_6render_7physics_RandomForceApplier *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_7flitter_5model_Node *__pyx_v_node, CYTHON_UNUSED double __pyx_v_strength, CYTHON_UNUSED struct __pyx_obj_7flitter_5model_Vector *__pyx_v_zero) {
  int __pyx_r;

  /* "flitter/render/physics.pyx":292
 *     @cython.profile(False)
 *     def __cinit__(self, Node node, double strength, Vector zero):
 *         self.index = 0             # <<<<<<<<<<<<<<
 * 
 *     @cython.profile(False)
 */
  __pyx_v_self->index = 0;

  /* "flitter/render/physics.pyx":290
 *     cdef uint64_t index
 * 
 *     @cython.profile(False)             # <<<<<<<<<<<<<<
 *     def __cinit__(self, Node node, double strength, Vector zero):
 *         self.index = 0
 */

  /* function exit code */
  __pyx_r = 0;
  return __pyx_r;
}

/* "flitter/render/physics.pyx":295
 * 
 *     @cython.profile(False)
 *     cdef void apply(self, Particle particle, double delta) noexcept nogil:             # <<<<<<<<<<<<<<
 *         cdef int64_t i
 *         for i in range(particle.force.length):
 */

static void __pyx_f_7flitter_6render_7physics_18RandomForceApplier_apply(struct __pyx_obj_7flitter_6render_7physics_RandomForceApplier *__pyx_v_self, struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_particle, CYTHON_UNUSED double __pyx_v_delta) {
  int64_t __pyx_v_i;
  int64_t __pyx_t_1;
  int64_t __pyx_t_2;
  int64_t __pyx_t_3;

  /* "flitter/render/physics.pyx":297
 *     cdef void apply(self, Particle particle, double delta) noexcept nogil:
 *         cdef int64_t i
 *         for i in range(particle.force.length):             # <<<<<<<<<<<<<<
 *             particle.force.numbers[i] = particle.force.numbers[i] + self.strength * self.random_source._item(self.index)
 *             self.index = self.index + 1
 */
  __pyx_t_1 = __pyx_v_particle->force->length;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "flitter/render/physics.pyx":298
 *         cdef int64_t i
 *         for i in range(particle.force.length):
 *             particle.force.numbers[i] = particle.force.numbers[i] + self.strength * self.random_source._item(self.index)             # <<<<<<<<<<<<<<
 *             self.index = self.index + 1
 * 
 */
    (__pyx_v_particle->force->numbers[__pyx_v_i]) = ((__pyx_v_particle->force->numbers[__pyx_v_i]) + (__pyx_v_self->__pyx_base.strength * ((struct __pyx_vtabstruct_7flitter_8language_9functions_normal *)__pyx_v_self->random_source->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base._item(((struct __pyx_obj_7flitter_8language_9functions_uniform *)__pyx_v_self->random_source), __pyx_v_self->index)));

    /* "flitter/render/physics.pyx":299
 *         for i in range(particle.force.length):
 *             particle.force.numbers[i] = particle.force.numbers[i] + self.strength * self.random_source._item(self.index)
 *             self.index = self.index + 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_self->index = (__pyx_v_self->index + 1);
  }

  /* "flitter/render/physics.pyx":295
 * 
 *     @cython.profile(False)
 *     cdef void apply(self, Particle particle, double delta) noexcept nogil:             # <<<<<<<<<<<<<<
 *         cdef int64_t i
 *         for i in range(particle.force.length):
 */

  /* function exit code */
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_6render_7physics_18RandomForceApplier_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_6render_7physics_18RandomForceApplier_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_18RandomForceApplier_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_6render_7physics_18RandomForceApplier_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_6render_7physics_18RandomForceApplier_2__reduce_cython__(((struct __pyx_obj_7flitter_6render_7physics_RandomForceApplier *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_6render_7physics_18RandomForceApplier_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_RandomForceApplier *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.render.physics.RandomForceApplier.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_6render_7physics_18RandomForceApplier_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_6render_7physics_18RandomForceApplier_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_18RandomForceApplier_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_6render_7physics_18RandomForceApplier_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.render.physics.RandomForceApplier.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_6render_7physics_18RandomForceApplier_4__setstate_cython__(((struct __pyx_obj_7flitter_6render_7physics_RandomForceApplier *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_6render_7physics_18RandomForceApplier_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_RandomForceApplier *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.render.physics.RandomForceApplier.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/render/physics.pyx":305
 *     cdef Vector direction
 * 
 *     @cython.profile(False)             # <<<<<<<<<<<<<<
 *     def __cinit__(self, Node node, double strength, Vector zero):
 *         self.direction = node.get_fvec('direction', zero.length, zero)
 */

/* Python wrapper */
static int __pyx_pw_7flitter_6render_7physics_17FieldForceApplier_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_6render_7physics_17FieldForceApplier_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_7flitter_5model_Node *__pyx_v_node = 0;
  CYTHON_UNUSED double __pyx_v_strength;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_zero = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_node,&__pyx_n_s_strength,&__pyx_n_s_zero,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_node)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 305, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_strength)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 305, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, 1); __PYX_ERR(1, 305, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_zero)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 305, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, 2); __PYX_ERR(1, 305, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__cinit__") < 0)) __PYX_ERR(1, 305, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
      values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
    }
    __pyx_v_node = ((struct __pyx_obj_7flitter_5model_Node *)values[0]);
    __pyx_v_strength = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_strength == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 306, __pyx_L3_error)
    __pyx_v_zero = ((struct __pyx_obj_7flitter_5model_Vector *)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 305, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.render.physics.FieldForceApplier.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_node), __pyx_ptype_7flitter_5model_Node, 1, "node", 0))) __PYX_ERR(1, 306, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_zero), __pyx_ptype_7flitter_5model_Vector, 1, "zero", 0))) __PYX_ERR(1, 306, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_6render_7physics_17FieldForceApplier___cinit__(((struct __pyx_obj_7flitter_6render_7physics_FieldForceApplier *)__pyx_v_self), __pyx_v_node, __pyx_v_strength, __pyx_v_zero);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_6render_7physics_17FieldForceApplier___cinit__(struct __pyx_obj_7flitter_6render_7physics_FieldForceApplier *__pyx_v_self, struct __pyx_obj_7flitter_5model_Node *__pyx_v_node, CYTHON_UNUSED double __pyx_v_strength, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_zero) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 1);

  /* "flitter/render/physics.pyx":307
 *     @cython.profile(False)
 *     def __cinit__(self, Node node, double strength, Vector zero):
 *         self.direction = node.get_fvec('direction', zero.length, zero)             # <<<<<<<<<<<<<<
 * 
 *     @cython.profile(False)
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Node *)__pyx_v_node->__pyx_vtab)->get_fvec(__pyx_v_node, __pyx_n_u_direction, __pyx_v_zero->length, __pyx_v_zero)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 307, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->direction);
  __Pyx_DECREF((PyObject *)__pyx_v_self->direction);
  __pyx_v_self->direction = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/render/physics.pyx":305
 *     cdef Vector direction
 * 
 *     @cython.profile(False)             # <<<<<<<<<<<<<<
 *     def __cinit__(self, Node node, double strength, Vector zero):
 *         self.direction = node.get_fvec('direction', zero.length, zero)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.render.physics.FieldForceApplier.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/render/physics.pyx":310
 * 
 *     @cython.profile(False)
 *     cdef void apply(self, Particle particle, double delta) noexcept nogil:             # <<<<<<<<<<<<<<
 *         cdef int64_t i
 *         for i in range(particle.force.length):
 */

static void __pyx_f_7flitter_6render_7physics_17FieldForceApplier_apply(struct __pyx_obj_7flitter_6render_7physics_FieldForceApplier *__pyx_v_self, struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_particle, CYTHON_UNUSED double __pyx_v_delta) {
  int64_t __pyx_v_i;
  int64_t __pyx_t_1;
  int64_t __pyx_t_2;
  int64_t __pyx_t_3;

  /* "flitter/render/physics.pyx":312
 *     cdef void apply(self, Particle particle, double delta) noexcept nogil:
 *         cdef int64_t i
 *         for i in range(particle.force.length):             # <<<<<<<<<<<<<<
 *             particle.force.numbers[i] = particle.force.numbers[i] + self.strength * self.direction.numbers[i]
 * 
 */
  __pyx_t_1 = __pyx_v_particle->force->length;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "flitter/render/physics.pyx":313
 *         cdef int64_t i
 *         for i in range(particle.force.length):
 *             particle.force.numbers[i] = particle.force.numbers[i] + self.strength * self.direction.numbers[i]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (__pyx_v_particle->force->numbers[__pyx_v_i]) = ((__pyx_v_particle->force->numbers[__pyx_v_i]) + (__pyx_v_self->__pyx_base.strength * (__pyx_v_self->direction->numbers[__pyx_v_i])));
  }

  /* "flitter/render/physics.pyx":310
 * 
 *     @cython.profile(False)
 *     cdef void apply(self, Particle particle, double delta) noexcept nogil:             # <<<<<<<<<<<<<<
 *         cdef int64_t i
 *         for i in range(particle.force.length):
 */

  /* function exit code */
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_6render_7physics_17FieldForceApplier_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_6render_7physics_17FieldForceApplier_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_17FieldForceApplier_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_6render_7physics_17FieldForceApplier_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_6render_7physics_17FieldForceApplier_2__reduce_cython__(((struct __pyx_obj_7flitter_6render_7physics_FieldForceApplier *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_6render_7physics_17FieldForceApplier_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_FieldForceApplier *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.render.physics.FieldForceApplier.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_6render_7physics_17FieldForceApplier_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_6render_7physics_17FieldForceApplier_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_17FieldForceApplier_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_6render_7physics_17FieldForceApplier_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.render.physics.FieldForceApplier.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_6render_7physics_17FieldForceApplier_4__setstate_cython__(((struct __pyx_obj_7flitter_6render_7physics_FieldForceApplier *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_6render_7physics_17FieldForceApplier_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_FieldForceApplier *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.render.physics.FieldForceApplier.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/render/physics.pyx":319
 *     cdef double power
 * 
 *     @cython.profile(False)             # <<<<<<<<<<<<<<
 *     def __cinit__(self, Node node, double strength, Vector zero):
 *         self.power = max(0, node.get_float('power', 1))
 */

/* Python wrapper */
static int __pyx_pw_7flitter_6render_7physics_21CollisionForceApplier_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_6render_7physics_21CollisionForceApplier_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_7flitter_5model_Node *__pyx_v_node = 0;
  CYTHON_UNUSED double __pyx_v_strength;
  CYTHON_UNUSED struct __pyx_obj_7flitter_5model_Vector *__pyx_v_zero = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_node,&__pyx_n_s_strength,&__pyx_n_s_zero,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_node)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 319, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_strength)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 319, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, 1); __PYX_ERR(1, 319, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_zero)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 319, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, 2); __PYX_ERR(1, 319, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__cinit__") < 0)) __PYX_ERR(1, 319, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
      values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
    }
    __pyx_v_node = ((struct __pyx_obj_7flitter_5model_Node *)values[0]);
    __pyx_v_strength = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_strength == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 320, __pyx_L3_error)
    __pyx_v_zero = ((struct __pyx_obj_7flitter_5model_Vector *)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 319, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.render.physics.CollisionForceApplier.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_node), __pyx_ptype_7flitter_5model_Node, 1, "node", 0))) __PYX_ERR(1, 320, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_zero), __pyx_ptype_7flitter_5model_Vector, 1, "zero", 0))) __PYX_ERR(1, 320, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_6render_7physics_21CollisionForceApplier___cinit__(((struct __pyx_obj_7flitter_6render_7physics_CollisionForceApplier *)__pyx_v_self), __pyx_v_node, __pyx_v_strength, __pyx_v_zero);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_6render_7physics_21CollisionForceApplier___cinit__(struct __pyx_obj_7flitter_6render_7physics_CollisionForceApplier *__pyx_v_self, struct __pyx_obj_7flitter_5model_Node *__pyx_v_node, CYTHON_UNUSED double __pyx_v_strength, CYTHON_UNUSED struct __pyx_obj_7flitter_5model_Vector *__pyx_v_zero) {
  int __pyx_r;
  double __pyx_t_1;
  long __pyx_t_2;
  double __pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "flitter/render/physics.pyx":321
 *     @cython.profile(False)
 *     def __cinit__(self, Node node, double strength, Vector zero):
 *         self.power = max(0, node.get_float('power', 1))             # <<<<<<<<<<<<<<
 * 
 *     @cython.profile(False)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_7flitter_5model_Node *)__pyx_v_node->__pyx_vtab)->get_float(__pyx_v_node, __pyx_n_u_power, 1.0); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 321, __pyx_L1_error)
  __pyx_t_2 = 0;
  __pyx_t_4 = (__pyx_t_1 > __pyx_t_2);
  if (__pyx_t_4) {
    __pyx_t_3 = __pyx_t_1;
  } else {
    __pyx_t_3 = __pyx_t_2;
  }
  __pyx_v_self->power = __pyx_t_3;

  /* "flitter/render/physics.pyx":319
 *     cdef double power
 * 
 *     @cython.profile(False)             # <<<<<<<<<<<<<<
 *     def __cinit__(self, Node node, double strength, Vector zero):
 *         self.power = max(0, node.get_float('power', 1))
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.render.physics.CollisionForceApplier.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "flitter/render/physics.pyx":324
 * 
 *     @cython.profile(False)
 *     cdef void apply(self, Particle from_particle, Particle to_particle, Vector direction, double distance) noexcept nogil:             # <<<<<<<<<<<<<<
 *         cdef double min_distance, f, k
 *         cdef int64_t i
 */

static void __pyx_f_7flitter_6render_7physics_21CollisionForceApplier_apply(struct __pyx_obj_7flitter_6render_7physics_CollisionForceApplier *__pyx_v_self, struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_from_particle, struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_to_particle, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_direction, double __pyx_v_distance) {
  double __pyx_v_min_distance;
  double __pyx_v_f;
  double __pyx_v_k;
  int64_t __pyx_v_i;
  int __pyx_t_1;
  int __pyx_t_2;
  int64_t __pyx_t_3;
  int64_t __pyx_t_4;
  int64_t __pyx_t_5;

  /* "flitter/render/physics.pyx":327
 *         cdef double min_distance, f, k
 *         cdef int64_t i
 *         if from_particle.radius and to_particle.radius:             # <<<<<<<<<<<<<<
 *             min_distance = from_particle.radius + to_particle.radius
 *             if distance < min_distance:
 */
  __pyx_t_2 = (__pyx_v_from_particle->radius != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_to_particle->radius != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "flitter/render/physics.pyx":328
 *         cdef int64_t i
 *         if from_particle.radius and to_particle.radius:
 *             min_distance = from_particle.radius + to_particle.radius             # <<<<<<<<<<<<<<
 *             if distance < min_distance:
 *                 k = min_distance - distance
 */
    __pyx_v_min_distance = (__pyx_v_from_particle->radius + __pyx_v_to_particle->radius);

    /* "flitter/render/physics.pyx":329
 *         if from_particle.radius and to_particle.radius:
 *             min_distance = from_particle.radius + to_particle.radius
 *             if distance < min_distance:             # <<<<<<<<<<<<<<
 *                 k = min_distance - distance
 *                 if self.power != 1:
 */
    __pyx_t_1 = (__pyx_v_distance < __pyx_v_min_distance);
    if (__pyx_t_1) {

      /* "flitter/render/physics.pyx":330
 *             min_distance = from_particle.radius + to_particle.radius
 *             if distance < min_distance:
 *                 k = min_distance - distance             # <<<<<<<<<<<<<<
 *                 if self.power != 1:
 *                     k **= self.power
 */
      __pyx_v_k = (__pyx_v_min_distance - __pyx_v_distance);

      /* "flitter/render/physics.pyx":331
 *             if distance < min_distance:
 *                 k = min_distance - distance
 *                 if self.power != 1:             # <<<<<<<<<<<<<<
 *                     k **= self.power
 *                 k *= self.strength
 */
      __pyx_t_1 = (__pyx_v_self->power != 1.0);
      if (__pyx_t_1) {

        /* "flitter/render/physics.pyx":332
 *                 k = min_distance - distance
 *                 if self.power != 1:
 *                     k **= self.power             # <<<<<<<<<<<<<<
 *                 k *= self.strength
 *                 for i in range(direction.length):
 */
        __pyx_v_k = pow(__pyx_v_k, __pyx_v_self->power);

        /* "flitter/render/physics.pyx":331
 *             if distance < min_distance:
 *                 k = min_distance - distance
 *                 if self.power != 1:             # <<<<<<<<<<<<<<
 *                     k **= self.power
 *                 k *= self.strength
 */
      }

      /* "flitter/render/physics.pyx":333
 *                 if self.power != 1:
 *                     k **= self.power
 *                 k *= self.strength             # <<<<<<<<<<<<<<
 *                 for i in range(direction.length):
 *                     f = direction.numbers[i] * k
 */
      __pyx_v_k = (__pyx_v_k * __pyx_v_self->__pyx_base.strength);

      /* "flitter/render/physics.pyx":334
 *                     k **= self.power
 *                 k *= self.strength
 *                 for i in range(direction.length):             # <<<<<<<<<<<<<<
 *                     f = direction.numbers[i] * k
 *                     from_particle.force.numbers[i] = from_particle.force.numbers[i] - f
 */
      __pyx_t_3 = __pyx_v_direction->length;
      __pyx_t_4 = __pyx_t_3;
      for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
        __pyx_v_i = __pyx_t_5;

        /* "flitter/render/physics.pyx":335
 *                 k *= self.strength
 *                 for i in range(direction.length):
 *                     f = direction.numbers[i] * k             # <<<<<<<<<<<<<<
 *                     from_particle.force.numbers[i] = from_particle.force.numbers[i] - f
 *                     to_particle.force.numbers[i] = to_particle.force.numbers[i] + f
 */
        __pyx_v_f = ((__pyx_v_direction->numbers[__pyx_v_i]) * __pyx_v_k);

        /* "flitter/render/physics.pyx":336
 *                 for i in range(direction.length):
 *                     f = direction.numbers[i] * k
 *                     from_particle.force.numbers[i] = from_particle.force.numbers[i] - f             # <<<<<<<<<<<<<<
 *                     to_particle.force.numbers[i] = to_particle.force.numbers[i] + f
 * 
 */
        (__pyx_v_from_particle->force->numbers[__pyx_v_i]) = ((__pyx_v_from_particle->force->numbers[__pyx_v_i]) - __pyx_v_f);

        /* "flitter/render/physics.pyx":337
 *                     f = direction.numbers[i] * k
 *                     from_particle.force.numbers[i] = from_particle.force.numbers[i] - f
 *                     to_particle.force.numbers[i] = to_particle.force.numbers[i] + f             # <<<<<<<<<<<<<<
 * 
 * 
 */
        (__pyx_v_to_particle->force->numbers[__pyx_v_i]) = ((__pyx_v_to_particle->force->numbers[__pyx_v_i]) + __pyx_v_f);
      }

      /* "flitter/render/physics.pyx":329
 *         if from_particle.radius and to_particle.radius:
 *             min_distance = from_particle.radius + to_particle.radius
 *             if distance < min_distance:             # <<<<<<<<<<<<<<
 *                 k = min_distance - distance
 *                 if self.power != 1:
 */
    }

    /* "flitter/render/physics.pyx":327
 *         cdef double min_distance, f, k
 *         cdef int64_t i
 *         if from_particle.radius and to_particle.radius:             # <<<<<<<<<<<<<<
 *             min_distance = from_particle.radius + to_particle.radius
 *             if distance < min_distance:
 */
  }

  /* "flitter/render/physics.pyx":324
 * 
 *     @cython.profile(False)
 *     cdef void apply(self, Particle from_particle, Particle to_particle, Vector direction, double distance) noexcept nogil:             # <<<<<<<<<<<<<<
 *         cdef double min_distance, f, k
 *         cdef int64_t i
 */

  /* function exit code */
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_6render_7physics_21CollisionForceApplier_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_6render_7physics_21CollisionForceApplier_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_21CollisionForceApplier_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_6render_7physics_21CollisionForceApplier_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_6render_7physics_21CollisionForceApplier_2__reduce_cython__(((struct __pyx_obj_7flitter_6render_7physics_CollisionForceApplier *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_6render_7physics_21CollisionForceApplier_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_CollisionForceApplier *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.render.physics.CollisionForceApplier.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_6render_7physics_21CollisionForceApplier_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_6render_7physics_21CollisionForceApplier_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_21CollisionForceApplier_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_6render_7physics_21CollisionForceApplier_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.render.physics.CollisionForceApplier.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_6render_7physics_21CollisionForceApplier_4__setstate_cython__(((struct __pyx_obj_7flitter_6render_7physics_CollisionForceApplier *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_6render_7physics_21CollisionForceApplier_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_CollisionForceApplier *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.render.physics.CollisionForceApplier.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/render/physics.pyx":342
 * cdef class GravityForceApplier(MatrixForceApplier):
 *     @cython.profile(False)
 *     cdef void apply(self, Particle from_particle, Particle to_particle, Vector direction, double distance) noexcept nogil:             # <<<<<<<<<<<<<<
 *         cdef double f, k, d=from_particle.radius + to_particle.radius
 *         cdef int64_t i
 */

static void __pyx_f_7flitter_6render_7physics_19GravityForceApplier_apply(struct __pyx_obj_7flitter_6render_7physics_GravityForceApplier *__pyx_v_self, struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_from_particle, struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_to_particle, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_direction, double __pyx_v_distance) {
  double __pyx_v_f;
  double __pyx_v_k;
  double __pyx_v_d;
  int64_t __pyx_v_i;
  int __pyx_t_1;
  int __pyx_t_2;
  double __pyx_t_3;
  double __pyx_t_4;
  double __pyx_t_5;
  int64_t __pyx_t_6;
  int64_t __pyx_t_7;
  int64_t __pyx_t_8;

  /* "flitter/render/physics.pyx":343
 *     @cython.profile(False)
 *     cdef void apply(self, Particle from_particle, Particle to_particle, Vector direction, double distance) noexcept nogil:
 *         cdef double f, k, d=from_particle.radius + to_particle.radius             # <<<<<<<<<<<<<<
 *         cdef int64_t i
 *         if from_particle.mass and to_particle.mass:
 */
  __pyx_v_d = (__pyx_v_from_particle->radius + __pyx_v_to_particle->radius);

  /* "flitter/render/physics.pyx":345
 *         cdef double f, k, d=from_particle.radius + to_particle.radius
 *         cdef int64_t i
 *         if from_particle.mass and to_particle.mass:             # <<<<<<<<<<<<<<
 *             d = max(d, distance)
 *             k = self.strength * from_particle.mass * to_particle.mass / (d * d)
 */
  __pyx_t_2 = (__pyx_v_from_particle->mass != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_to_particle->mass != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "flitter/render/physics.pyx":346
 *         cdef int64_t i
 *         if from_particle.mass and to_particle.mass:
 *             d = max(d, distance)             # <<<<<<<<<<<<<<
 *             k = self.strength * from_particle.mass * to_particle.mass / (d * d)
 *             for i in range(direction.length):
 */
    __pyx_t_3 = __pyx_v_distance;
    __pyx_t_4 = __pyx_v_d;
    __pyx_t_1 = (__pyx_t_3 > __pyx_t_4);
    if (__pyx_t_1) {
      __pyx_t_5 = __pyx_t_3;
    } else {
      __pyx_t_5 = __pyx_t_4;
    }
    __pyx_v_d = __pyx_t_5;

    /* "flitter/render/physics.pyx":347
 *         if from_particle.mass and to_particle.mass:
 *             d = max(d, distance)
 *             k = self.strength * from_particle.mass * to_particle.mass / (d * d)             # <<<<<<<<<<<<<<
 *             for i in range(direction.length):
 *                 f = direction.numbers[i] * k
 */
    __pyx_v_k = (((__pyx_v_self->__pyx_base.strength * __pyx_v_from_particle->mass) * __pyx_v_to_particle->mass) / (__pyx_v_d * __pyx_v_d));

    /* "flitter/render/physics.pyx":348
 *             d = max(d, distance)
 *             k = self.strength * from_particle.mass * to_particle.mass / (d * d)
 *             for i in range(direction.length):             # <<<<<<<<<<<<<<
 *                 f = direction.numbers[i] * k
 *                 from_particle.force.numbers[i] = from_particle.force.numbers[i] + f
 */
    __pyx_t_6 = __pyx_v_direction->length;
    __pyx_t_7 = __pyx_t_6;
    for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
      __pyx_v_i = __pyx_t_8;

      /* "flitter/render/physics.pyx":349
 *             k = self.strength * from_particle.mass * to_particle.mass / (d * d)
 *             for i in range(direction.length):
 *                 f = direction.numbers[i] * k             # <<<<<<<<<<<<<<
 *                 from_particle.force.numbers[i] = from_particle.force.numbers[i] + f
 *                 to_particle.force.numbers[i] = to_particle.force.numbers[i] - f
 */
      __pyx_v_f = ((__pyx_v_direction->numbers[__pyx_v_i]) * __pyx_v_k);

      /* "flitter/render/physics.pyx":350
 *             for i in range(direction.length):
 *                 f = direction.numbers[i] * k
 *                 from_particle.force.numbers[i] = from_particle.force.numbers[i] + f             # <<<<<<<<<<<<<<
 *                 to_particle.force.numbers[i] = to_particle.force.numbers[i] - f
 * 
 */
      (__pyx_v_from_particle->force->numbers[__pyx_v_i]) = ((__pyx_v_from_particle->force->numbers[__pyx_v_i]) + __pyx_v_f);

      /* "flitter/render/physics.pyx":351
 *                 f = direction.numbers[i] * k
 *                 from_particle.force.numbers[i] = from_particle.force.numbers[i] + f
 *                 to_particle.force.numbers[i] = to_particle.force.numbers[i] - f             # <<<<<<<<<<<<<<
 * 
 * 
 */
      (__pyx_v_to_particle->force->numbers[__pyx_v_i]) = ((__pyx_v_to_particle->force->numbers[__pyx_v_i]) - __pyx_v_f);
    }

    /* "flitter/render/physics.pyx":345
 *         cdef double f, k, d=from_particle.radius + to_particle.radius
 *         cdef int64_t i
 *         if from_particle.mass and to_particle.mass:             # <<<<<<<<<<<<<<
 *             d = max(d, distance)
 *             k = self.strength * from_particle.mass * to_particle.mass / (d * d)
 */
  }

  /* "flitter/render/physics.pyx":342
 * cdef class GravityForceApplier(MatrixForceApplier):
 *     @cython.profile(False)
 *     cdef void apply(self, Particle from_particle, Particle to_particle, Vector direction, double distance) noexcept nogil:             # <<<<<<<<<<<<<<
 *         cdef double f, k, d=from_particle.radius + to_particle.radius
 *         cdef int64_t i
 */

  /* function exit code */
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_6render_7physics_19GravityForceApplier_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_6render_7physics_19GravityForceApplier_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_19GravityForceApplier_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_6render_7physics_19GravityForceApplier_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_6render_7physics_19GravityForceApplier___reduce_cython__(((struct __pyx_obj_7flitter_6render_7physics_GravityForceApplier *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_6render_7physics_19GravityForceApplier___reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_GravityForceApplier *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.render.physics.GravityForceApplier.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_6render_7physics_19GravityForceApplier_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_6render_7physics_19GravityForceApplier_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_19GravityForceApplier_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_6render_7physics_19GravityForceApplier_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.render.physics.GravityForceApplier.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_6render_7physics_19GravityForceApplier_2__setstate_cython__(((struct __pyx_obj_7flitter_6render_7physics_GravityForceApplier *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_6render_7physics_19GravityForceApplier_2__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_GravityForceApplier *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.render.physics.GravityForceApplier.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/render/physics.pyx":356
 * cdef class ElectrostaticForceApplier(MatrixForceApplier):
 *     @cython.profile(False)
 *     cdef void apply(self, Particle from_particle, Particle to_particle, Vector direction, double distance) noexcept nogil:             # <<<<<<<<<<<<<<
 *         cdef double f, k, d=from_particle.radius + to_particle.radius
 *         cdef int64_t i
 */

static void __pyx_f_7flitter_6render_7physics_25ElectrostaticForceApplier_apply(struct __pyx_obj_7flitter_6render_7physics_ElectrostaticForceApplier *__pyx_v_self, struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_from_particle, struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_to_particle, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_direction, double __pyx_v_distance) {
  double __pyx_v_f;
  double __pyx_v_k;
  double __pyx_v_d;
  int64_t __pyx_v_i;
  int __pyx_t_1;
  int __pyx_t_2;
  double __pyx_t_3;
  double __pyx_t_4;
  double __pyx_t_5;
  int64_t __pyx_t_6;
  int64_t __pyx_t_7;
  int64_t __pyx_t_8;

  /* "flitter/render/physics.pyx":357
 *     @cython.profile(False)
 *     cdef void apply(self, Particle from_particle, Particle to_particle, Vector direction, double distance) noexcept nogil:
 *         cdef double f, k, d=from_particle.radius + to_particle.radius             # <<<<<<<<<<<<<<
 *         cdef int64_t i
 *         if from_particle.charge and to_particle.charge:
 */
  __pyx_v_d = (__pyx_v_from_particle->radius + __pyx_v_to_particle->radius);

  /* "flitter/render/physics.pyx":359
 *         cdef double f, k, d=from_particle.radius + to_particle.radius
 *         cdef int64_t i
 *         if from_particle.charge and to_particle.charge:             # <<<<<<<<<<<<<<
 *             d = max(d, distance)
 *             k = self.strength * from_particle.charge * to_particle.charge / (d * d)
 */
  __pyx_t_2 = (__pyx_v_from_particle->charge != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_to_particle->charge != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "flitter/render/physics.pyx":360
 *         cdef int64_t i
 *         if from_particle.charge and to_particle.charge:
 *             d = max(d, distance)             # <<<<<<<<<<<<<<
 *             k = self.strength * from_particle.charge * to_particle.charge / (d * d)
 *             for i in range(direction.length):
 */
    __pyx_t_3 = __pyx_v_distance;
    __pyx_t_4 = __pyx_v_d;
    __pyx_t_1 = (__pyx_t_3 > __pyx_t_4);
    if (__pyx_t_1) {
      __pyx_t_5 = __pyx_t_3;
    } else {
      __pyx_t_5 = __pyx_t_4;
    }
    __pyx_v_d = __pyx_t_5;

    /* "flitter/render/physics.pyx":361
 *         if from_particle.charge and to_particle.charge:
 *             d = max(d, distance)
 *             k = self.strength * from_particle.charge * to_particle.charge / (d * d)             # <<<<<<<<<<<<<<
 *             for i in range(direction.length):
 *                 f = direction.numbers[i] * k
 */
    __pyx_v_k = (((__pyx_v_self->__pyx_base.strength * __pyx_v_from_particle->charge) * __pyx_v_to_particle->charge) / (__pyx_v_d * __pyx_v_d));

    /* "flitter/render/physics.pyx":362
 *             d = max(d, distance)
 *             k = self.strength * from_particle.charge * to_particle.charge / (d * d)
 *             for i in range(direction.length):             # <<<<<<<<<<<<<<
 *                 f = direction.numbers[i] * k
 *                 from_particle.force.numbers[i] = from_particle.force.numbers[i] - f
 */
    __pyx_t_6 = __pyx_v_direction->length;
    __pyx_t_7 = __pyx_t_6;
    for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
      __pyx_v_i = __pyx_t_8;

      /* "flitter/render/physics.pyx":363
 *             k = self.strength * from_particle.charge * to_particle.charge / (d * d)
 *             for i in range(direction.length):
 *                 f = direction.numbers[i] * k             # <<<<<<<<<<<<<<
 *                 from_particle.force.numbers[i] = from_particle.force.numbers[i] - f
 *                 to_particle.force.numbers[i] = to_particle.force.numbers[i] + f
 */
      __pyx_v_f = ((__pyx_v_direction->numbers[__pyx_v_i]) * __pyx_v_k);

      /* "flitter/render/physics.pyx":364
 *             for i in range(direction.length):
 *                 f = direction.numbers[i] * k
 *                 from_particle.force.numbers[i] = from_particle.force.numbers[i] - f             # <<<<<<<<<<<<<<
 *                 to_particle.force.numbers[i] = to_particle.force.numbers[i] + f
 * 
 */
      (__pyx_v_from_particle->force->numbers[__pyx_v_i]) = ((__pyx_v_from_particle->force->numbers[__pyx_v_i]) - __pyx_v_f);

      /* "flitter/render/physics.pyx":365
 *                 f = direction.numbers[i] * k
 *                 from_particle.force.numbers[i] = from_particle.force.numbers[i] - f
 *                 to_particle.force.numbers[i] = to_particle.force.numbers[i] + f             # <<<<<<<<<<<<<<
 * 
 * 
 */
      (__pyx_v_to_particle->force->numbers[__pyx_v_i]) = ((__pyx_v_to_particle->force->numbers[__pyx_v_i]) + __pyx_v_f);
    }

    /* "flitter/render/physics.pyx":359
 *         cdef double f, k, d=from_particle.radius + to_particle.radius
 *         cdef int64_t i
 *         if from_particle.charge and to_particle.charge:             # <<<<<<<<<<<<<<
 *             d = max(d, distance)
 *             k = self.strength * from_particle.charge * to_particle.charge / (d * d)
 */
  }

  /* "flitter/render/physics.pyx":356
 * cdef class ElectrostaticForceApplier(MatrixForceApplier):
 *     @cython.profile(False)
 *     cdef void apply(self, Particle from_particle, Particle to_particle, Vector direction, double distance) noexcept nogil:             # <<<<<<<<<<<<<<
 *         cdef double f, k, d=from_particle.radius + to_particle.radius
 *         cdef int64_t i
 */

  /* function exit code */
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_6render_7physics_25ElectrostaticForceApplier_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_6render_7physics_25ElectrostaticForceApplier_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_25ElectrostaticForceApplier_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_6render_7physics_25ElectrostaticForceApplier_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_6render_7physics_25ElectrostaticForceApplier___reduce_cython__(((struct __pyx_obj_7flitter_6render_7physics_ElectrostaticForceApplier *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_6render_7physics_25ElectrostaticForceApplier___reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_ElectrostaticForceApplier *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.render.physics.ElectrostaticForceApplier.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_6render_7physics_25ElectrostaticForceApplier_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_6render_7physics_25ElectrostaticForceApplier_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_25ElectrostaticForceApplier_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_6render_7physics_25ElectrostaticForceApplier_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.render.physics.ElectrostaticForceApplier.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_6render_7physics_25ElectrostaticForceApplier_2__setstate_cython__(((struct __pyx_obj_7flitter_6render_7physics_ElectrostaticForceApplier *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_6render_7physics_25ElectrostaticForceApplier_2__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_ElectrostaticForceApplier *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.render.physics.ElectrostaticForceApplier.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/render/physics.pyx":371
 *     cdef double overlap
 * 
 *     @cython.profile(False)             # <<<<<<<<<<<<<<
 *     def __cinit__(self, Node node, double strength, Vector zero):
 *         self.overlap = min(max(0, node.get_float('overlap', 0.25)), 1)
 */

/* Python wrapper */
static int __pyx_pw_7flitter_6render_7physics_20AdhesionForceApplier_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_6render_7physics_20AdhesionForceApplier_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_7flitter_5model_Node *__pyx_v_node = 0;
  CYTHON_UNUSED double __pyx_v_strength;
  CYTHON_UNUSED struct __pyx_obj_7flitter_5model_Vector *__pyx_v_zero = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_node,&__pyx_n_s_strength,&__pyx_n_s_zero,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_node)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 371, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_strength)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 371, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, 1); __PYX_ERR(1, 371, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_zero)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 371, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, 2); __PYX_ERR(1, 371, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__cinit__") < 0)) __PYX_ERR(1, 371, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
      values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
    }
    __pyx_v_node = ((struct __pyx_obj_7flitter_5model_Node *)values[0]);
    __pyx_v_strength = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_strength == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 372, __pyx_L3_error)
    __pyx_v_zero = ((struct __pyx_obj_7flitter_5model_Vector *)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 371, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.render.physics.AdhesionForceApplier.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_node), __pyx_ptype_7flitter_5model_Node, 1, "node", 0))) __PYX_ERR(1, 372, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_zero), __pyx_ptype_7flitter_5model_Vector, 1, "zero", 0))) __PYX_ERR(1, 372, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_6render_7physics_20AdhesionForceApplier___cinit__(((struct __pyx_obj_7flitter_6render_7physics_AdhesionForceApplier *)__pyx_v_self), __pyx_v_node, __pyx_v_strength, __pyx_v_zero);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_6render_7physics_20AdhesionForceApplier___cinit__(struct __pyx_obj_7flitter_6render_7physics_AdhesionForceApplier *__pyx_v_self, struct __pyx_obj_7flitter_5model_Node *__pyx_v_node, CYTHON_UNUSED double __pyx_v_strength, CYTHON_UNUSED struct __pyx_obj_7flitter_5model_Vector *__pyx_v_zero) {
  int __pyx_r;
  long __pyx_t_1;
  double __pyx_t_2;
  long __pyx_t_3;
  double __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "flitter/render/physics.pyx":373
 *     @cython.profile(False)
 *     def __cinit__(self, Node node, double strength, Vector zero):
 *         self.overlap = min(max(0, node.get_float('overlap', 0.25)), 1)             # <<<<<<<<<<<<<<
 * 
 *     @cython.profile(False)
 */
  __pyx_t_1 = 1;
  __pyx_t_2 = ((struct __pyx_vtabstruct_7flitter_5model_Node *)__pyx_v_node->__pyx_vtab)->get_float(__pyx_v_node, __pyx_n_u_overlap, 0.25); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 373, __pyx_L1_error)
  __pyx_t_3 = 0;
  __pyx_t_5 = (__pyx_t_2 > __pyx_t_3);
  if (__pyx_t_5) {
    __pyx_t_4 = __pyx_t_2;
  } else {
    __pyx_t_4 = __pyx_t_3;
  }
  __pyx_t_2 = __pyx_t_4;
  __pyx_t_5 = (__pyx_t_1 < __pyx_t_2);
  if (__pyx_t_5) {
    __pyx_t_4 = __pyx_t_1;
  } else {
    __pyx_t_4 = __pyx_t_2;
  }
  __pyx_v_self->overlap = __pyx_t_4;

  /* "flitter/render/physics.pyx":371
 *     cdef double overlap
 * 
 *     @cython.profile(False)             # <<<<<<<<<<<<<<
 *     def __cinit__(self, Node node, double strength, Vector zero):
 *         self.overlap = min(max(0, node.get_float('overlap', 0.25)), 1)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.render.physics.AdhesionForceApplier.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "flitter/render/physics.pyx":376
 * 
 *     @cython.profile(False)
 *     cdef void apply(self, Particle from_particle, Particle to_particle, Vector direction, double distance) noexcept nogil:             # <<<<<<<<<<<<<<
 *         cdef double min_distance, max_distance, overlap_distance, adhesion_distance, f, k
 *         cdef int64_t i
 */

static void __pyx_f_7flitter_6render_7physics_20AdhesionForceApplier_apply(struct __pyx_obj_7flitter_6render_7physics_AdhesionForceApplier *__pyx_v_self, struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_from_particle, struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_to_particle, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_direction, double __pyx_v_distance) {
  double __pyx_v_min_distance;
  double __pyx_v_max_distance;
  double __pyx_v_overlap_distance;
  double __pyx_v_adhesion_distance;
  double __pyx_v_f;
  double __pyx_v_k;
  int64_t __pyx_v_i;
  int __pyx_t_1;
  int __pyx_t_2;
  int64_t __pyx_t_3;
  int64_t __pyx_t_4;
  int64_t __pyx_t_5;

  /* "flitter/render/physics.pyx":379
 *         cdef double min_distance, max_distance, overlap_distance, adhesion_distance, f, k
 *         cdef int64_t i
 *         if from_particle.radius and to_particle.radius:             # <<<<<<<<<<<<<<
 *             max_distance = from_particle.radius + to_particle.radius
 *             if distance < max_distance:
 */
  __pyx_t_2 = (__pyx_v_from_particle->radius != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_to_particle->radius != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "flitter/render/physics.pyx":380
 *         cdef int64_t i
 *         if from_particle.radius and to_particle.radius:
 *             max_distance = from_particle.radius + to_particle.radius             # <<<<<<<<<<<<<<
 *             if distance < max_distance:
 *                 overlap_distance = max_distance - distance
 */
    __pyx_v_max_distance = (__pyx_v_from_particle->radius + __pyx_v_to_particle->radius);

    /* "flitter/render/physics.pyx":381
 *         if from_particle.radius and to_particle.radius:
 *             max_distance = from_particle.radius + to_particle.radius
 *             if distance < max_distance:             # <<<<<<<<<<<<<<
 *                 overlap_distance = max_distance - distance
 *                 min_distance = abs(from_particle.radius - to_particle.radius)
 */
    __pyx_t_1 = (__pyx_v_distance < __pyx_v_max_distance);
    if (__pyx_t_1) {

      /* "flitter/render/physics.pyx":382
 *             max_distance = from_particle.radius + to_particle.radius
 *             if distance < max_distance:
 *                 overlap_distance = max_distance - distance             # <<<<<<<<<<<<<<
 *                 min_distance = abs(from_particle.radius - to_particle.radius)
 *                 adhesion_distance = min_distance*self.overlap + max_distance*(1-self.overlap)
 */
      __pyx_v_overlap_distance = (__pyx_v_max_distance - __pyx_v_distance);

      /* "flitter/render/physics.pyx":383
 *             if distance < max_distance:
 *                 overlap_distance = max_distance - distance
 *                 min_distance = abs(from_particle.radius - to_particle.radius)             # <<<<<<<<<<<<<<
 *                 adhesion_distance = min_distance*self.overlap + max_distance*(1-self.overlap)
 *                 k = self.strength * overlap_distance * (distance - adhesion_distance)
 */
      __pyx_v_min_distance = fabs((__pyx_v_from_particle->radius - __pyx_v_to_particle->radius));

      /* "flitter/render/physics.pyx":384
 *                 overlap_distance = max_distance - distance
 *                 min_distance = abs(from_particle.radius - to_particle.radius)
 *                 adhesion_distance = min_distance*self.overlap + max_distance*(1-self.overlap)             # <<<<<<<<<<<<<<
 *                 k = self.strength * overlap_distance * (distance - adhesion_distance)
 *                 for i in range(direction.length):
 */
      __pyx_v_adhesion_distance = ((__pyx_v_min_distance * __pyx_v_self->overlap) + (__pyx_v_max_distance * (1.0 - __pyx_v_self->overlap)));

      /* "flitter/render/physics.pyx":385
 *                 min_distance = abs(from_particle.radius - to_particle.radius)
 *                 adhesion_distance = min_distance*self.overlap + max_distance*(1-self.overlap)
 *                 k = self.strength * overlap_distance * (distance - adhesion_distance)             # <<<<<<<<<<<<<<
 *                 for i in range(direction.length):
 *                     f = direction.numbers[i] * k
 */
      __pyx_v_k = ((__pyx_v_self->__pyx_base.strength * __pyx_v_overlap_distance) * (__pyx_v_distance - __pyx_v_adhesion_distance));

      /* "flitter/render/physics.pyx":386
 *                 adhesion_distance = min_distance*self.overlap + max_distance*(1-self.overlap)
 *                 k = self.strength * overlap_distance * (distance - adhesion_distance)
 *                 for i in range(direction.length):             # <<<<<<<<<<<<<<
 *                     f = direction.numbers[i] * k
 *                     from_particle.force.numbers[i] = from_particle.force.numbers[i] + f
 */
      __pyx_t_3 = __pyx_v_direction->length;
      __pyx_t_4 = __pyx_t_3;
      for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
        __pyx_v_i = __pyx_t_5;

        /* "flitter/render/physics.pyx":387
 *                 k = self.strength * overlap_distance * (distance - adhesion_distance)
 *                 for i in range(direction.length):
 *                     f = direction.numbers[i] * k             # <<<<<<<<<<<<<<
 *                     from_particle.force.numbers[i] = from_particle.force.numbers[i] + f
 *                     to_particle.force.numbers[i] = to_particle.force.numbers[i] - f
 */
        __pyx_v_f = ((__pyx_v_direction->numbers[__pyx_v_i]) * __pyx_v_k);

        /* "flitter/render/physics.pyx":388
 *                 for i in range(direction.length):
 *                     f = direction.numbers[i] * k
 *                     from_particle.force.numbers[i] = from_particle.force.numbers[i] + f             # <<<<<<<<<<<<<<
 *                     to_particle.force.numbers[i] = to_particle.force.numbers[i] - f
 * 
 */
        (__pyx_v_from_particle->force->numbers[__pyx_v_i]) = ((__pyx_v_from_particle->force->numbers[__pyx_v_i]) + __pyx_v_f);

        /* "flitter/render/physics.pyx":389
 *                     f = direction.numbers[i] * k
 *                     from_particle.force.numbers[i] = from_particle.force.numbers[i] + f
 *                     to_particle.force.numbers[i] = to_particle.force.numbers[i] - f             # <<<<<<<<<<<<<<
 * 
 * 
 */
        (__pyx_v_to_particle->force->numbers[__pyx_v_i]) = ((__pyx_v_to_particle->force->numbers[__pyx_v_i]) - __pyx_v_f);
      }

      /* "flitter/render/physics.pyx":381
 *         if from_particle.radius and to_particle.radius:
 *             max_distance = from_particle.radius + to_particle.radius
 *             if distance < max_distance:             # <<<<<<<<<<<<<<
 *                 overlap_distance = max_distance - distance
 *                 min_distance = abs(from_particle.radius - to_particle.radius)
 */
    }

    /* "flitter/render/physics.pyx":379
 *         cdef double min_distance, max_distance, overlap_distance, adhesion_distance, f, k
 *         cdef int64_t i
 *         if from_particle.radius and to_particle.radius:             # <<<<<<<<<<<<<<
 *             max_distance = from_particle.radius + to_particle.radius
 *             if distance < max_distance:
 */
  }

  /* "flitter/render/physics.pyx":376
 * 
 *     @cython.profile(False)
 *     cdef void apply(self, Particle from_particle, Particle to_particle, Vector direction, double distance) noexcept nogil:             # <<<<<<<<<<<<<<
 *         cdef double min_distance, max_distance, overlap_distance, adhesion_distance, f, k
 *         cdef int64_t i
 */

  /* function exit code */
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_6render_7physics_20AdhesionForceApplier_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_6render_7physics_20AdhesionForceApplier_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_20AdhesionForceApplier_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_6render_7physics_20AdhesionForceApplier_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_6render_7physics_20AdhesionForceApplier_2__reduce_cython__(((struct __pyx_obj_7flitter_6render_7physics_AdhesionForceApplier *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_6render_7physics_20AdhesionForceApplier_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_AdhesionForceApplier *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.render.physics.AdhesionForceApplier.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_6render_7physics_20AdhesionForceApplier_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_6render_7physics_20AdhesionForceApplier_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_20AdhesionForceApplier_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_6render_7physics_20AdhesionForceApplier_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.render.physics.AdhesionForceApplier.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_6render_7physics_20AdhesionForceApplier_4__setstate_cython__(((struct __pyx_obj_7flitter_6render_7physics_AdhesionForceApplier *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_6render_7physics_20AdhesionForceApplier_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_AdhesionForceApplier *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.render.physics.AdhesionForceApplier.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/render/physics.pyx":398
 *     cdef double speed_of_light
 * 
 *     def __init__(self, **kwargs):             # <<<<<<<<<<<<<<
 *         self.state_keys = set()
 * 
 */

/* Python wrapper */
static int __pyx_pw_7flitter_6render_7physics_13PhysicsSystem_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7flitter_6render_7physics_13PhysicsSystem_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__init__", 1, 0, 0, __pyx_nargs); return -1;}
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__init__", 1))) return -1;
  __pyx_r = __pyx_pf_7flitter_6render_7physics_13PhysicsSystem___init__(((struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem *)__pyx_v_self), __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7flitter_6render_7physics_13PhysicsSystem___init__(struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_kwargs) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "flitter/render/physics.pyx":399
 * 
 *     def __init__(self, **kwargs):
 *         self.state_keys = set()             # <<<<<<<<<<<<<<
 * 
 *     async def destroy(self):
 */
  __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->state_keys);
  __Pyx_DECREF(__pyx_v_self->state_keys);
  __pyx_v_self->state_keys = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/render/physics.pyx":398
 *     cdef double speed_of_light
 * 
 *     def __init__(self, **kwargs):             # <<<<<<<<<<<<<<
 *         self.state_keys = set()
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.render.physics.PhysicsSystem.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_7flitter_6render_7physics_13PhysicsSystem_4generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "flitter/render/physics.pyx":401
 *         self.state_keys = set()
 * 
 *     async def destroy(self):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_6render_7physics_13PhysicsSystem_3destroy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_6render_7physics_13PhysicsSystem_3destroy = {"destroy", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_13PhysicsSystem_3destroy, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_6render_7physics_13PhysicsSystem_3destroy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("destroy (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("destroy", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "destroy", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_6render_7physics_13PhysicsSystem_2destroy(((struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_6render_7physics_13PhysicsSystem_2destroy(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem *__pyx_v_self) {
  struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct__destroy *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("destroy", 0);
  __pyx_cur_scope = (struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct__destroy *)__pyx_tp_new_7flitter_6render_7physics___pyx_scope_struct__destroy(__pyx_ptype_7flitter_6render_7physics___pyx_scope_struct__destroy, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct__destroy *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(1, 401, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_Coroutine_New((__pyx_coroutine_body_t) __pyx_gb_7flitter_6render_7physics_13PhysicsSystem_4generator, __pyx_codeobj_, (PyObject *) __pyx_cur_scope, __pyx_n_s_destroy, __pyx_n_s_PhysicsSystem_destroy, __pyx_n_s_flitter_render_physics); if (unlikely(!gen)) __PYX_ERR(1, 401, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.render.physics.PhysicsSystem.destroy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_7flitter_6render_7physics_13PhysicsSystem_4generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct__destroy *__pyx_cur_scope = ((struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct__destroy *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("destroy", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(1, 401, __pyx_L1_error)
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_AddTraceback("destroy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_7flitter_6render_7physics_13PhysicsSystem_7generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "flitter/render/physics.pyx":404
 *         pass
 * 
 *     async def purge(self):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_6render_7physics_13PhysicsSystem_6purge(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_6render_7physics_13PhysicsSystem_6purge = {"purge", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_13PhysicsSystem_6purge, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_6render_7physics_13PhysicsSystem_6purge(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("purge (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("purge", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "purge", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_6render_7physics_13PhysicsSystem_5purge(((struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_6render_7physics_13PhysicsSystem_5purge(CYTHON_UNUSED struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem *__pyx_v_self) {
  struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct_1_purge *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("purge", 0);
  __pyx_cur_scope = (struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct_1_purge *)__pyx_tp_new_7flitter_6render_7physics___pyx_scope_struct_1_purge(__pyx_ptype_7flitter_6render_7physics___pyx_scope_struct_1_purge, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct_1_purge *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(1, 404, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_Coroutine_New((__pyx_coroutine_body_t) __pyx_gb_7flitter_6render_7physics_13PhysicsSystem_7generator1, __pyx_codeobj__2, (PyObject *) __pyx_cur_scope, __pyx_n_s_purge, __pyx_n_s_PhysicsSystem_purge, __pyx_n_s_flitter_render_physics); if (unlikely(!gen)) __PYX_ERR(1, 404, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.render.physics.PhysicsSystem.purge", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_7flitter_6render_7physics_13PhysicsSystem_7generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct_1_purge *__pyx_cur_scope = ((struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct_1_purge *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("purge", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(1, 404, __pyx_L1_error)
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_AddTraceback("purge", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_7flitter_6render_7physics_13PhysicsSystem_10generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "flitter/render/physics.pyx":407
 *         pass
 * 
 *     async def update(self, engine, Node node, double time, double performance, **kwargs):             # <<<<<<<<<<<<<<
 *         cdef int64_t run = node.get_int('run', 0)
 *         self.dimensions = node.get_int('dimensions', 0)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_6render_7physics_13PhysicsSystem_9update(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_6render_7physics_13PhysicsSystem_9update = {"update", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_13PhysicsSystem_9update, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_6render_7physics_13PhysicsSystem_9update(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_engine = 0;
  struct __pyx_obj_7flitter_5model_Node *__pyx_v_node = 0;
  double __pyx_v_time;
  double __pyx_v_performance;
  CYTHON_UNUSED PyObject *__pyx_v_kwargs = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("update (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return NULL;
  __Pyx_GOTREF(__pyx_v_kwargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_engine,&__pyx_n_s_node,&__pyx_n_s_time,&__pyx_n_s_performance,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_engine)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 407, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_node)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 407, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("update", 1, 4, 4, 1); __PYX_ERR(1, 407, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_time)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 407, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("update", 1, 4, 4, 2); __PYX_ERR(1, 407, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_performance)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[3]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 407, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("update", 1, 4, 4, 3); __PYX_ERR(1, 407, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, __pyx_v_kwargs, values + 0, kwd_pos_args, "update") < 0)) __PYX_ERR(1, 407, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 4)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
      values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
    }
    __pyx_v_engine = values[0];
    __pyx_v_node = ((struct __pyx_obj_7flitter_5model_Node *)values[1]);
    __pyx_v_time = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_time == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 407, __pyx_L3_error)
    __pyx_v_performance = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_performance == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 407, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("update", 1, 4, 4, __pyx_nargs); __PYX_ERR(1, 407, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_CLEAR(__pyx_v_kwargs);
  __Pyx_AddTraceback("flitter.render.physics.PhysicsSystem.update", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_node), __pyx_ptype_7flitter_5model_Node, 1, "node", 0))) __PYX_ERR(1, 407, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_6render_7physics_13PhysicsSystem_8update(((struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem *)__pyx_v_self), __pyx_v_engine, __pyx_v_node, __pyx_v_time, __pyx_v_performance, __pyx_v_kwargs);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_DECREF(__pyx_v_kwargs);
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_6render_7physics_13PhysicsSystem_8update(struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem *__pyx_v_self, PyObject *__pyx_v_engine, struct __pyx_obj_7flitter_5model_Node *__pyx_v_node, double __pyx_v_time, double __pyx_v_performance, CYTHON_UNUSED PyObject *__pyx_v_kwargs) {
  struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct_2_update *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("update", 0);
  __pyx_cur_scope = (struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct_2_update *)__pyx_tp_new_7flitter_6render_7physics___pyx_scope_struct_2_update(__pyx_ptype_7flitter_6render_7physics___pyx_scope_struct_2_update, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct_2_update *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(1, 407, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_engine = __pyx_v_engine;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_engine);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_engine);
  __pyx_cur_scope->__pyx_v_node = __pyx_v_node;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_node);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_node);
  __pyx_cur_scope->__pyx_v_time = __pyx_v_time;
  __pyx_cur_scope->__pyx_v_performance = __pyx_v_performance;
  __pyx_cur_scope->__pyx_v_kwargs = __pyx_v_kwargs;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_kwargs);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_kwargs);
  {
    __pyx_CoroutineObject *gen = __Pyx_Coroutine_New((__pyx_coroutine_body_t) __pyx_gb_7flitter_6render_7physics_13PhysicsSystem_10generator2, __pyx_codeobj__3, (PyObject *) __pyx_cur_scope, __pyx_n_s_update, __pyx_n_s_PhysicsSystem_update, __pyx_n_s_flitter_render_physics); if (unlikely(!gen)) __PYX_ERR(1, 407, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("flitter.render.physics.PhysicsSystem.update", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_7flitter_6render_7physics_13PhysicsSystem_10generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct_2_update *__pyx_cur_scope = ((struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct_2_update *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int64_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  double __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  double __pyx_t_6;
  long __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  unsigned int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  int __pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  Py_ssize_t __pyx_t_17;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  Py_ssize_t __pyx_t_20;
  PyObject *__pyx_t_21 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("update", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L32_resume_from_await;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(1, 407, __pyx_L1_error)

  /* "flitter/render/physics.pyx":408
 * 
 *     async def update(self, engine, Node node, double time, double performance, **kwargs):
 *         cdef int64_t run = node.get_int('run', 0)             # <<<<<<<<<<<<<<
 *         self.dimensions = node.get_int('dimensions', 0)
 *         if self.dimensions < 1:
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_7flitter_5model_Node *)__pyx_cur_scope->__pyx_v_node->__pyx_vtab)->get_int(__pyx_cur_scope->__pyx_v_node, __pyx_n_u_run, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 408, __pyx_L1_error)
  __pyx_cur_scope->__pyx_v_run = __pyx_t_1;

  /* "flitter/render/physics.pyx":409
 *     async def update(self, engine, Node node, double time, double performance, **kwargs):
 *         cdef int64_t run = node.get_int('run', 0)
 *         self.dimensions = node.get_int('dimensions', 0)             # <<<<<<<<<<<<<<
 *         if self.dimensions < 1:
 *             return
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_7flitter_5model_Node *)__pyx_cur_scope->__pyx_v_node->__pyx_vtab)->get_int(__pyx_cur_scope->__pyx_v_node, __pyx_n_u_dimensions, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 409, __pyx_L1_error)
  __pyx_cur_scope->__pyx_v_self->dimensions = __pyx_t_1;

  /* "flitter/render/physics.pyx":410
 *         cdef int64_t run = node.get_int('run', 0)
 *         self.dimensions = node.get_int('dimensions', 0)
 *         if self.dimensions < 1:             # <<<<<<<<<<<<<<
 *             return
 *         cdef Vector state_prefix = node.get_vec('state', None)
 */
  __pyx_t_2 = (__pyx_cur_scope->__pyx_v_self->dimensions < 1);
  if (__pyx_t_2) {

    /* "flitter/render/physics.pyx":411
 *         self.dimensions = node.get_int('dimensions', 0)
 *         if self.dimensions < 1:
 *             return             # <<<<<<<<<<<<<<
 *         cdef Vector state_prefix = node.get_vec('state', None)
 *         if state_prefix is None:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = NULL;
    goto __pyx_L0;

    /* "flitter/render/physics.pyx":410
 *         cdef int64_t run = node.get_int('run', 0)
 *         self.dimensions = node.get_int('dimensions', 0)
 *         if self.dimensions < 1:             # <<<<<<<<<<<<<<
 *             return
 *         cdef Vector state_prefix = node.get_vec('state', None)
 */
  }

  /* "flitter/render/physics.pyx":412
 *         if self.dimensions < 1:
 *             return
 *         cdef Vector state_prefix = node.get_vec('state', None)             # <<<<<<<<<<<<<<
 *         if state_prefix is None:
 *             return
 */
  __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Node *)__pyx_cur_scope->__pyx_v_node->__pyx_vtab)->get_vec(__pyx_cur_scope->__pyx_v_node, __pyx_n_u_state, ((struct __pyx_obj_7flitter_5model_Vector *)Py_None))); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_cur_scope->__pyx_v_state_prefix = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "flitter/render/physics.pyx":413
 *             return
 *         cdef Vector state_prefix = node.get_vec('state', None)
 *         if state_prefix is None:             # <<<<<<<<<<<<<<
 *             return
 *         time = node.get_float('time', time)
 */
  __pyx_t_2 = (((PyObject *)__pyx_cur_scope->__pyx_v_state_prefix) == Py_None);
  if (__pyx_t_2) {

    /* "flitter/render/physics.pyx":414
 *         cdef Vector state_prefix = node.get_vec('state', None)
 *         if state_prefix is None:
 *             return             # <<<<<<<<<<<<<<
 *         time = node.get_float('time', time)
 *         self.resolution = node.get_float('resolution', 1/engine.target_fps)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = NULL;
    goto __pyx_L0;

    /* "flitter/render/physics.pyx":413
 *             return
 *         cdef Vector state_prefix = node.get_vec('state', None)
 *         if state_prefix is None:             # <<<<<<<<<<<<<<
 *             return
 *         time = node.get_float('time', time)
 */
  }

  /* "flitter/render/physics.pyx":415
 *         if state_prefix is None:
 *             return
 *         time = node.get_float('time', time)             # <<<<<<<<<<<<<<
 *         self.resolution = node.get_float('resolution', 1/engine.target_fps)
 *         if self.resolution <= 0:
 */
  __pyx_t_4 = ((struct __pyx_vtabstruct_7flitter_5model_Node *)__pyx_cur_scope->__pyx_v_node->__pyx_vtab)->get_float(__pyx_cur_scope->__pyx_v_node, __pyx_n_u_time, __pyx_cur_scope->__pyx_v_time); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 415, __pyx_L1_error)
  __pyx_cur_scope->__pyx_v_time = __pyx_t_4;

  /* "flitter/render/physics.pyx":416
 *             return
 *         time = node.get_float('time', time)
 *         self.resolution = node.get_float('resolution', 1/engine.target_fps)             # <<<<<<<<<<<<<<
 *         if self.resolution <= 0:
 *             return
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_engine, __pyx_n_s_target_fps); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 416, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyNumber_Divide(__pyx_int_1, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 416, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_5); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 416, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_6 = ((struct __pyx_vtabstruct_7flitter_5model_Node *)__pyx_cur_scope->__pyx_v_node->__pyx_vtab)->get_float(__pyx_cur_scope->__pyx_v_node, __pyx_n_u_resolution, __pyx_t_4); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 416, __pyx_L1_error)
  __pyx_cur_scope->__pyx_v_self->resolution = __pyx_t_6;

  /* "flitter/render/physics.pyx":417
 *         time = node.get_float('time', time)
 *         self.resolution = node.get_float('resolution', 1/engine.target_fps)
 *         if self.resolution <= 0:             # <<<<<<<<<<<<<<
 *             return
 *         self.speed_of_light = max(0, node.get_float('speed_of_light', 0))
 */
  __pyx_t_2 = (__pyx_cur_scope->__pyx_v_self->resolution <= 0.0);
  if (__pyx_t_2) {

    /* "flitter/render/physics.pyx":418
 *         self.resolution = node.get_float('resolution', 1/engine.target_fps)
 *         if self.resolution <= 0:
 *             return             # <<<<<<<<<<<<<<
 *         self.speed_of_light = max(0, node.get_float('speed_of_light', 0))
 *         cdef StateDict state = engine.state
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = NULL;
    goto __pyx_L0;

    /* "flitter/render/physics.pyx":417
 *         time = node.get_float('time', time)
 *         self.resolution = node.get_float('resolution', 1/engine.target_fps)
 *         if self.resolution <= 0:             # <<<<<<<<<<<<<<
 *             return
 *         self.speed_of_light = max(0, node.get_float('speed_of_light', 0))
 */
  }

  /* "flitter/render/physics.pyx":419
 *         if self.resolution <= 0:
 *             return
 *         self.speed_of_light = max(0, node.get_float('speed_of_light', 0))             # <<<<<<<<<<<<<<
 *         cdef StateDict state = engine.state
 *         cdef Vector time_vector = state.get_item(state_prefix.concat(CLOCK))
 */
  __pyx_t_6 = ((struct __pyx_vtabstruct_7flitter_5model_Node *)__pyx_cur_scope->__pyx_v_node->__pyx_vtab)->get_float(__pyx_cur_scope->__pyx_v_node, __pyx_n_u_speed_of_light, 0.0); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 419, __pyx_L1_error)
  __pyx_t_7 = 0;
  __pyx_t_2 = (__pyx_t_6 > __pyx_t_7);
  if (__pyx_t_2) {
    __pyx_t_4 = __pyx_t_6;
  } else {
    __pyx_t_4 = __pyx_t_7;
  }
  __pyx_cur_scope->__pyx_v_self->speed_of_light = __pyx_t_4;

  /* "flitter/render/physics.pyx":420
 *             return
 *         self.speed_of_light = max(0, node.get_float('speed_of_light', 0))
 *         cdef StateDict state = engine.state             # <<<<<<<<<<<<<<
 *         cdef Vector time_vector = state.get_item(state_prefix.concat(CLOCK))
 *         if time_vector.length == 1 and time_vector.numbers != NULL:
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_engine, __pyx_n_s_state); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_7flitter_5model_StateDict))))) __PYX_ERR(1, 420, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_5);
  __pyx_cur_scope->__pyx_v_state = ((struct __pyx_obj_7flitter_5model_StateDict *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "flitter/render/physics.pyx":421
 *         self.speed_of_light = max(0, node.get_float('speed_of_light', 0))
 *         cdef StateDict state = engine.state
 *         cdef Vector time_vector = state.get_item(state_prefix.concat(CLOCK))             # <<<<<<<<<<<<<<
 *         if time_vector.length == 1 and time_vector.numbers != NULL:
 *             clock = time_vector.numbers[0]
 */
  __pyx_t_5 = ((PyObject *)__pyx_v_7flitter_6render_7physics_CLOCK);
  __Pyx_INCREF(__pyx_t_5);
  __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_cur_scope->__pyx_v_state_prefix->__pyx_vtab)->concat(__pyx_cur_scope->__pyx_v_state_prefix, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_5), 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 421, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_StateDict *)__pyx_cur_scope->__pyx_v_state->__pyx_vtab)->get_item(__pyx_cur_scope->__pyx_v_state, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_3))); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 421, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GIVEREF(__pyx_t_5);
  __pyx_cur_scope->__pyx_v_time_vector = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "flitter/render/physics.pyx":422
 *         cdef StateDict state = engine.state
 *         cdef Vector time_vector = state.get_item(state_prefix.concat(CLOCK))
 *         if time_vector.length == 1 and time_vector.numbers != NULL:             # <<<<<<<<<<<<<<
 *             clock = time_vector.numbers[0]
 *         else:
 */
  __pyx_t_8 = (__pyx_cur_scope->__pyx_v_time_vector->length == 1);
  if (__pyx_t_8) {
  } else {
    __pyx_t_2 = __pyx_t_8;
    goto __pyx_L8_bool_binop_done;
  }
  __pyx_t_8 = (__pyx_cur_scope->__pyx_v_time_vector->numbers != NULL);
  __pyx_t_2 = __pyx_t_8;
  __pyx_L8_bool_binop_done:;
  if (__pyx_t_2) {

    /* "flitter/render/physics.pyx":423
 *         cdef Vector time_vector = state.get_item(state_prefix.concat(CLOCK))
 *         if time_vector.length == 1 and time_vector.numbers != NULL:
 *             clock = time_vector.numbers[0]             # <<<<<<<<<<<<<<
 *         else:
 *             clock = 0
 */
    __pyx_t_5 = PyFloat_FromDouble((__pyx_cur_scope->__pyx_v_time_vector->numbers[0])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 423, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_cur_scope->__pyx_v_clock = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "flitter/render/physics.pyx":422
 *         cdef StateDict state = engine.state
 *         cdef Vector time_vector = state.get_item(state_prefix.concat(CLOCK))
 *         if time_vector.length == 1 and time_vector.numbers != NULL:             # <<<<<<<<<<<<<<
 *             clock = time_vector.numbers[0]
 *         else:
 */
    goto __pyx_L7;
  }

  /* "flitter/render/physics.pyx":425
 *             clock = time_vector.numbers[0]
 *         else:
 *             clock = 0             # <<<<<<<<<<<<<<
 *         cdef double last_time
 *         time_vector = state.get_item(state_prefix.concat(LAST))
 */
  /*else*/ {
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    __pyx_cur_scope->__pyx_v_clock = __pyx_int_0;
  }
  __pyx_L7:;

  /* "flitter/render/physics.pyx":427
 *             clock = 0
 *         cdef double last_time
 *         time_vector = state.get_item(state_prefix.concat(LAST))             # <<<<<<<<<<<<<<
 *         if time_vector.length == 1 and time_vector.numbers != NULL:
 *             last_time = time_vector.numbers[0]
 */
  __pyx_t_5 = ((PyObject *)__pyx_v_7flitter_6render_7physics_LAST);
  __Pyx_INCREF(__pyx_t_5);
  __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_cur_scope->__pyx_v_state_prefix->__pyx_vtab)->concat(__pyx_cur_scope->__pyx_v_state_prefix, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_5), 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 427, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_StateDict *)__pyx_cur_scope->__pyx_v_state->__pyx_vtab)->get_item(__pyx_cur_scope->__pyx_v_state, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_3))); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 427, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GOTREF((PyObject *)__pyx_cur_scope->__pyx_v_time_vector);
  __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_time_vector, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_5));
  __Pyx_GIVEREF(__pyx_t_5);
  __pyx_t_5 = 0;

  /* "flitter/render/physics.pyx":428
 *         cdef double last_time
 *         time_vector = state.get_item(state_prefix.concat(LAST))
 *         if time_vector.length == 1 and time_vector.numbers != NULL:             # <<<<<<<<<<<<<<
 *             last_time = time_vector.numbers[0]
 *         else:
 */
  __pyx_t_8 = (__pyx_cur_scope->__pyx_v_time_vector->length == 1);
  if (__pyx_t_8) {
  } else {
    __pyx_t_2 = __pyx_t_8;
    goto __pyx_L11_bool_binop_done;
  }
  __pyx_t_8 = (__pyx_cur_scope->__pyx_v_time_vector->numbers != NULL);
  __pyx_t_2 = __pyx_t_8;
  __pyx_L11_bool_binop_done:;
  if (__pyx_t_2) {

    /* "flitter/render/physics.pyx":429
 *         time_vector = state.get_item(state_prefix.concat(LAST))
 *         if time_vector.length == 1 and time_vector.numbers != NULL:
 *             last_time = time_vector.numbers[0]             # <<<<<<<<<<<<<<
 *         else:
 *             last_time = time
 */
    __pyx_cur_scope->__pyx_v_last_time = (__pyx_cur_scope->__pyx_v_time_vector->numbers[0]);

    /* "flitter/render/physics.pyx":428
 *         cdef double last_time
 *         time_vector = state.get_item(state_prefix.concat(LAST))
 *         if time_vector.length == 1 and time_vector.numbers != NULL:             # <<<<<<<<<<<<<<
 *             last_time = time_vector.numbers[0]
 *         else:
 */
    goto __pyx_L10;
  }

  /* "flitter/render/physics.pyx":431
 *             last_time = time_vector.numbers[0]
 *         else:
 *             last_time = time             # <<<<<<<<<<<<<<
 *         cdef Vector last_run_vector = state.get_item(state_prefix.concat(RUN))
 *         cdef int64_t last_run
 */
  /*else*/ {
    __pyx_cur_scope->__pyx_v_last_time = __pyx_cur_scope->__pyx_v_time;
  }
  __pyx_L10:;

  /* "flitter/render/physics.pyx":432
 *         else:
 *             last_time = time
 *         cdef Vector last_run_vector = state.get_item(state_prefix.concat(RUN))             # <<<<<<<<<<<<<<
 *         cdef int64_t last_run
 *         if last_run_vector.length == 1 and last_run_vector.numbers != NULL:
 */
  __pyx_t_5 = ((PyObject *)__pyx_v_7flitter_6render_7physics_RUN);
  __Pyx_INCREF(__pyx_t_5);
  __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_cur_scope->__pyx_v_state_prefix->__pyx_vtab)->concat(__pyx_cur_scope->__pyx_v_state_prefix, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_5), 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 432, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_StateDict *)__pyx_cur_scope->__pyx_v_state->__pyx_vtab)->get_item(__pyx_cur_scope->__pyx_v_state, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_3))); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 432, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GIVEREF(__pyx_t_5);
  __pyx_cur_scope->__pyx_v_last_run_vector = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "flitter/render/physics.pyx":434
 *         cdef Vector last_run_vector = state.get_item(state_prefix.concat(RUN))
 *         cdef int64_t last_run
 *         if last_run_vector.length == 1 and last_run_vector.numbers != NULL:             # <<<<<<<<<<<<<<
 *             last_run = <int64_t>floor(last_run_vector.numbers[0])
 *         else:
 */
  __pyx_t_8 = (__pyx_cur_scope->__pyx_v_last_run_vector->length == 1);
  if (__pyx_t_8) {
  } else {
    __pyx_t_2 = __pyx_t_8;
    goto __pyx_L14_bool_binop_done;
  }
  __pyx_t_8 = (__pyx_cur_scope->__pyx_v_last_run_vector->numbers != NULL);
  __pyx_t_2 = __pyx_t_8;
  __pyx_L14_bool_binop_done:;
  if (__pyx_t_2) {

    /* "flitter/render/physics.pyx":435
 *         cdef int64_t last_run
 *         if last_run_vector.length == 1 and last_run_vector.numbers != NULL:
 *             last_run = <int64_t>floor(last_run_vector.numbers[0])             # <<<<<<<<<<<<<<
 *         else:
 *             last_run = run
 */
    __pyx_cur_scope->__pyx_v_last_run = ((int64_t)floor((__pyx_cur_scope->__pyx_v_last_run_vector->numbers[0])));

    /* "flitter/render/physics.pyx":434
 *         cdef Vector last_run_vector = state.get_item(state_prefix.concat(RUN))
 *         cdef int64_t last_run
 *         if last_run_vector.length == 1 and last_run_vector.numbers != NULL:             # <<<<<<<<<<<<<<
 *             last_run = <int64_t>floor(last_run_vector.numbers[0])
 *         else:
 */
    goto __pyx_L13;
  }

  /* "flitter/render/physics.pyx":437
 *             last_run = <int64_t>floor(last_run_vector.numbers[0])
 *         else:
 *             last_run = run             # <<<<<<<<<<<<<<
 *         cdef Vector iteration_vector = state.get_item(state_prefix.concat(ITERATION))
 *         cdef int64_t iteration
 */
  /*else*/ {
    __pyx_cur_scope->__pyx_v_last_run = __pyx_cur_scope->__pyx_v_run;
  }
  __pyx_L13:;

  /* "flitter/render/physics.pyx":438
 *         else:
 *             last_run = run
 *         cdef Vector iteration_vector = state.get_item(state_prefix.concat(ITERATION))             # <<<<<<<<<<<<<<
 *         cdef int64_t iteration
 *         if iteration_vector.length == 1 and iteration_vector.numbers != NULL:
 */
  __pyx_t_5 = ((PyObject *)__pyx_v_7flitter_6render_7physics_ITERATION);
  __Pyx_INCREF(__pyx_t_5);
  __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_cur_scope->__pyx_v_state_prefix->__pyx_vtab)->concat(__pyx_cur_scope->__pyx_v_state_prefix, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_5), 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 438, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_StateDict *)__pyx_cur_scope->__pyx_v_state->__pyx_vtab)->get_item(__pyx_cur_scope->__pyx_v_state, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_3))); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 438, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GIVEREF(__pyx_t_5);
  __pyx_cur_scope->__pyx_v_iteration_vector = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "flitter/render/physics.pyx":440
 *         cdef Vector iteration_vector = state.get_item(state_prefix.concat(ITERATION))
 *         cdef int64_t iteration
 *         if iteration_vector.length == 1 and iteration_vector.numbers != NULL:             # <<<<<<<<<<<<<<
 *             iteration = <int64_t>floor(iteration_vector.numbers[0]) + 1
 *         else:
 */
  __pyx_t_8 = (__pyx_cur_scope->__pyx_v_iteration_vector->length == 1);
  if (__pyx_t_8) {
  } else {
    __pyx_t_2 = __pyx_t_8;
    goto __pyx_L17_bool_binop_done;
  }
  __pyx_t_8 = (__pyx_cur_scope->__pyx_v_iteration_vector->numbers != NULL);
  __pyx_t_2 = __pyx_t_8;
  __pyx_L17_bool_binop_done:;
  if (__pyx_t_2) {

    /* "flitter/render/physics.pyx":441
 *         cdef int64_t iteration
 *         if iteration_vector.length == 1 and iteration_vector.numbers != NULL:
 *             iteration = <int64_t>floor(iteration_vector.numbers[0]) + 1             # <<<<<<<<<<<<<<
 *         else:
 *             iteration = 0
 */
    __pyx_cur_scope->__pyx_v_iteration = (((int64_t)floor((__pyx_cur_scope->__pyx_v_iteration_vector->numbers[0]))) + 1);

    /* "flitter/render/physics.pyx":440
 *         cdef Vector iteration_vector = state.get_item(state_prefix.concat(ITERATION))
 *         cdef int64_t iteration
 *         if iteration_vector.length == 1 and iteration_vector.numbers != NULL:             # <<<<<<<<<<<<<<
 *             iteration = <int64_t>floor(iteration_vector.numbers[0]) + 1
 *         else:
 */
    goto __pyx_L16;
  }

  /* "flitter/render/physics.pyx":443
 *             iteration = <int64_t>floor(iteration_vector.numbers[0]) + 1
 *         else:
 *             iteration = 0             # <<<<<<<<<<<<<<
 *         cdef Vector state_key
 *         if run != last_run:
 */
  /*else*/ {
    __pyx_cur_scope->__pyx_v_iteration = 0;
  }
  __pyx_L16:;

  /* "flitter/render/physics.pyx":445
 *             iteration = 0
 *         cdef Vector state_key
 *         if run != last_run:             # <<<<<<<<<<<<<<
 *             logger.debug("Reset physics system {!r} for run {}", state_prefix, run)
 *             for state_key in self.state_keys:
 */
  __pyx_t_2 = (__pyx_cur_scope->__pyx_v_run != __pyx_cur_scope->__pyx_v_last_run);
  if (__pyx_t_2) {

    /* "flitter/render/physics.pyx":446
 *         cdef Vector state_key
 *         if run != last_run:
 *             logger.debug("Reset physics system {!r} for run {}", state_prefix, run)             # <<<<<<<<<<<<<<
 *             for state_key in self.state_keys:
 *                 state.set_item(state_key, null_)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_logger); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 446, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_debug); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 446, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyInt_From_int64_t(__pyx_cur_scope->__pyx_v_run); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 446, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_10 = NULL;
    __pyx_t_11 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
        __pyx_t_11 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[4] = {__pyx_t_10, __pyx_kp_u_Reset_physics_system_r_for_run, ((PyObject *)__pyx_cur_scope->__pyx_v_state_prefix), __pyx_t_3};
      __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_9, __pyx_callargs+1-__pyx_t_11, 3+__pyx_t_11);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 446, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "flitter/render/physics.pyx":447
 *         if run != last_run:
 *             logger.debug("Reset physics system {!r} for run {}", state_prefix, run)
 *             for state_key in self.state_keys:             # <<<<<<<<<<<<<<
 *                 state.set_item(state_key, null_)
 *             self.state_keys = set()
 */
    __pyx_t_12 = 0;
    __pyx_t_9 = __Pyx_set_iterator(__pyx_cur_scope->__pyx_v_self->state_keys, 1, (&__pyx_t_13), (&__pyx_t_14)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 447, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_XDECREF(__pyx_t_5);
    __pyx_t_5 = __pyx_t_9;
    __pyx_t_9 = 0;
    while (1) {
      __pyx_t_15 = __Pyx_set_iter_next(__pyx_t_5, __pyx_t_13, &__pyx_t_12, &__pyx_t_9, __pyx_t_14);
      if (unlikely(__pyx_t_15 == 0)) break;
      if (unlikely(__pyx_t_15 == -1)) __PYX_ERR(1, 447, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      if (!(likely(((__pyx_t_9) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_9, __pyx_ptype_7flitter_5model_Vector))))) __PYX_ERR(1, 447, __pyx_L1_error)
      __Pyx_XGOTREF((PyObject *)__pyx_cur_scope->__pyx_v_state_key);
      __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_state_key, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_9));
      __Pyx_GIVEREF(__pyx_t_9);
      __pyx_t_9 = 0;

      /* "flitter/render/physics.pyx":448
 *             logger.debug("Reset physics system {!r} for run {}", state_prefix, run)
 *             for state_key in self.state_keys:
 *                 state.set_item(state_key, null_)             # <<<<<<<<<<<<<<
 *             self.state_keys = set()
 *             time_vector = Vector.__new__(Vector)
 */
      __pyx_t_9 = ((PyObject *)__pyx_v_7flitter_5model_null_);
      __Pyx_INCREF(__pyx_t_9);
      ((struct __pyx_vtabstruct_7flitter_5model_StateDict *)__pyx_cur_scope->__pyx_v_state->__pyx_vtab)->set_item(__pyx_cur_scope->__pyx_v_state, __pyx_cur_scope->__pyx_v_state_key, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_9)); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 448, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "flitter/render/physics.pyx":449
 *             for state_key in self.state_keys:
 *                 state.set_item(state_key, null_)
 *             self.state_keys = set()             # <<<<<<<<<<<<<<
 *             time_vector = Vector.__new__(Vector)
 *             time_vector.allocate_numbers(1)
 */
    __pyx_t_5 = PySet_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 449, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_self->state_keys);
    __Pyx_DECREF(__pyx_cur_scope->__pyx_v_self->state_keys);
    __pyx_cur_scope->__pyx_v_self->state_keys = ((PyObject*)__pyx_t_5);
    __pyx_t_5 = 0;

    /* "flitter/render/physics.pyx":450
 *                 state.set_item(state_key, null_)
 *             self.state_keys = set()
 *             time_vector = Vector.__new__(Vector)             # <<<<<<<<<<<<<<
 *             time_vector.allocate_numbers(1)
 *             time_vector.numbers[0] = time
 */
    __pyx_t_5 = __Pyx_tp_new(((PyObject *)__pyx_ptype_7flitter_5model_Vector), __pyx_empty_tuple); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 450, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (!(likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_7flitter_5model_Vector)))) __PYX_ERR(1, 450, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope->__pyx_v_time_vector);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_time_vector, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_5));
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_5 = 0;

    /* "flitter/render/physics.pyx":451
 *             self.state_keys = set()
 *             time_vector = Vector.__new__(Vector)
 *             time_vector.allocate_numbers(1)             # <<<<<<<<<<<<<<
 *             time_vector.numbers[0] = time
 *             state.set_item(state_prefix, time_vector)
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_cur_scope->__pyx_v_time_vector->__pyx_vtab)->allocate_numbers(__pyx_cur_scope->__pyx_v_time_vector, 1); if (unlikely(__pyx_t_1 == ((int64_t)-1L))) __PYX_ERR(1, 451, __pyx_L1_error)

    /* "flitter/render/physics.pyx":452
 *             time_vector = Vector.__new__(Vector)
 *             time_vector.allocate_numbers(1)
 *             time_vector.numbers[0] = time             # <<<<<<<<<<<<<<
 *             state.set_item(state_prefix, time_vector)
 *             clock = 0
 */
    (__pyx_cur_scope->__pyx_v_time_vector->numbers[0]) = __pyx_cur_scope->__pyx_v_time;

    /* "flitter/render/physics.pyx":453
 *             time_vector.allocate_numbers(1)
 *             time_vector.numbers[0] = time
 *             state.set_item(state_prefix, time_vector)             # <<<<<<<<<<<<<<
 *             clock = 0
 *             iteration = 0
 */
    ((struct __pyx_vtabstruct_7flitter_5model_StateDict *)__pyx_cur_scope->__pyx_v_state->__pyx_vtab)->set_item(__pyx_cur_scope->__pyx_v_state, __pyx_cur_scope->__pyx_v_state_prefix, __pyx_cur_scope->__pyx_v_time_vector); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 453, __pyx_L1_error)

    /* "flitter/render/physics.pyx":454
 *             time_vector.numbers[0] = time
 *             state.set_item(state_prefix, time_vector)
 *             clock = 0             # <<<<<<<<<<<<<<
 *             iteration = 0
 *         cdef Vector run_vector = Vector.__new__(Vector)
 */
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_clock);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_clock, __pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);

    /* "flitter/render/physics.pyx":455
 *             state.set_item(state_prefix, time_vector)
 *             clock = 0
 *             iteration = 0             # <<<<<<<<<<<<<<
 *         cdef Vector run_vector = Vector.__new__(Vector)
 *         run_vector.allocate_numbers(1)
 */
    __pyx_cur_scope->__pyx_v_iteration = 0;

    /* "flitter/render/physics.pyx":445
 *             iteration = 0
 *         cdef Vector state_key
 *         if run != last_run:             # <<<<<<<<<<<<<<
 *             logger.debug("Reset physics system {!r} for run {}", state_prefix, run)
 *             for state_key in self.state_keys:
 */
  }

  /* "flitter/render/physics.pyx":456
 *             clock = 0
 *             iteration = 0
 *         cdef Vector run_vector = Vector.__new__(Vector)             # <<<<<<<<<<<<<<
 *         run_vector.allocate_numbers(1)
 *         run_vector.numbers[0] = run
 */
  __pyx_t_5 = __Pyx_tp_new(((PyObject *)__pyx_ptype_7flitter_5model_Vector), __pyx_empty_tuple); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (!(likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_7flitter_5model_Vector)))) __PYX_ERR(1, 456, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_5);
  __pyx_cur_scope->__pyx_v_run_vector = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "flitter/render/physics.pyx":457
 *             iteration = 0
 *         cdef Vector run_vector = Vector.__new__(Vector)
 *         run_vector.allocate_numbers(1)             # <<<<<<<<<<<<<<
 *         run_vector.numbers[0] = run
 *         state.set_item(state_prefix.concat(RUN), run_vector)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_cur_scope->__pyx_v_run_vector->__pyx_vtab)->allocate_numbers(__pyx_cur_scope->__pyx_v_run_vector, 1); if (unlikely(__pyx_t_1 == ((int64_t)-1L))) __PYX_ERR(1, 457, __pyx_L1_error)

  /* "flitter/render/physics.pyx":458
 *         cdef Vector run_vector = Vector.__new__(Vector)
 *         run_vector.allocate_numbers(1)
 *         run_vector.numbers[0] = run             # <<<<<<<<<<<<<<
 *         state.set_item(state_prefix.concat(RUN), run_vector)
 *         iteration_vector = Vector.__new__(Vector)
 */
  (__pyx_cur_scope->__pyx_v_run_vector->numbers[0]) = __pyx_cur_scope->__pyx_v_run;

  /* "flitter/render/physics.pyx":459
 *         run_vector.allocate_numbers(1)
 *         run_vector.numbers[0] = run
 *         state.set_item(state_prefix.concat(RUN), run_vector)             # <<<<<<<<<<<<<<
 *         iteration_vector = Vector.__new__(Vector)
 *         iteration_vector.allocate_numbers(1)
 */
  __pyx_t_5 = ((PyObject *)__pyx_v_7flitter_6render_7physics_RUN);
  __Pyx_INCREF(__pyx_t_5);
  __pyx_t_9 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_cur_scope->__pyx_v_state_prefix->__pyx_vtab)->concat(__pyx_cur_scope->__pyx_v_state_prefix, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_5), 0)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 459, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  ((struct __pyx_vtabstruct_7flitter_5model_StateDict *)__pyx_cur_scope->__pyx_v_state->__pyx_vtab)->set_item(__pyx_cur_scope->__pyx_v_state, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_9), __pyx_cur_scope->__pyx_v_run_vector); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 459, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "flitter/render/physics.pyx":460
 *         run_vector.numbers[0] = run
 *         state.set_item(state_prefix.concat(RUN), run_vector)
 *         iteration_vector = Vector.__new__(Vector)             # <<<<<<<<<<<<<<
 *         iteration_vector.allocate_numbers(1)
 *         iteration_vector.numbers[0] = iteration
 */
  __pyx_t_9 = __Pyx_tp_new(((PyObject *)__pyx_ptype_7flitter_5model_Vector), __pyx_empty_tuple); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 460, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (!(likely(__Pyx_TypeTest(__pyx_t_9, __pyx_ptype_7flitter_5model_Vector)))) __PYX_ERR(1, 460, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_cur_scope->__pyx_v_iteration_vector);
  __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_iteration_vector, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_9));
  __Pyx_GIVEREF(__pyx_t_9);
  __pyx_t_9 = 0;

  /* "flitter/render/physics.pyx":461
 *         state.set_item(state_prefix.concat(RUN), run_vector)
 *         iteration_vector = Vector.__new__(Vector)
 *         iteration_vector.allocate_numbers(1)             # <<<<<<<<<<<<<<
 *         iteration_vector.numbers[0] = iteration
 *         state.set_item(state_prefix.concat(ITERATION), iteration_vector)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_cur_scope->__pyx_v_iteration_vector->__pyx_vtab)->allocate_numbers(__pyx_cur_scope->__pyx_v_iteration_vector, 1); if (unlikely(__pyx_t_1 == ((int64_t)-1L))) __PYX_ERR(1, 461, __pyx_L1_error)

  /* "flitter/render/physics.pyx":462
 *         iteration_vector = Vector.__new__(Vector)
 *         iteration_vector.allocate_numbers(1)
 *         iteration_vector.numbers[0] = iteration             # <<<<<<<<<<<<<<
 *         state.set_item(state_prefix.concat(ITERATION), iteration_vector)
 *         cdef Vector seed = state_prefix.concat(run_vector).concat(iteration_vector)
 */
  (__pyx_cur_scope->__pyx_v_iteration_vector->numbers[0]) = __pyx_cur_scope->__pyx_v_iteration;

  /* "flitter/render/physics.pyx":463
 *         iteration_vector.allocate_numbers(1)
 *         iteration_vector.numbers[0] = iteration
 *         state.set_item(state_prefix.concat(ITERATION), iteration_vector)             # <<<<<<<<<<<<<<
 *         cdef Vector seed = state_prefix.concat(run_vector).concat(iteration_vector)
 *         cdef PhysicsGroup group, top_group=PhysicsGroup.__new__(PhysicsGroup)
 */
  __pyx_t_9 = ((PyObject *)__pyx_v_7flitter_6render_7physics_ITERATION);
  __Pyx_INCREF(__pyx_t_9);
  __pyx_t_5 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_cur_scope->__pyx_v_state_prefix->__pyx_vtab)->concat(__pyx_cur_scope->__pyx_v_state_prefix, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_9), 0)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 463, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  ((struct __pyx_vtabstruct_7flitter_5model_StateDict *)__pyx_cur_scope->__pyx_v_state->__pyx_vtab)->set_item(__pyx_cur_scope->__pyx_v_state, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_5), __pyx_cur_scope->__pyx_v_iteration_vector); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 463, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "flitter/render/physics.pyx":464
 *         iteration_vector.numbers[0] = iteration
 *         state.set_item(state_prefix.concat(ITERATION), iteration_vector)
 *         cdef Vector seed = state_prefix.concat(run_vector).concat(iteration_vector)             # <<<<<<<<<<<<<<
 *         cdef PhysicsGroup group, top_group=PhysicsGroup.__new__(PhysicsGroup)
 *         cdef list groups = [top_group]
 */
  __pyx_t_5 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_cur_scope->__pyx_v_state_prefix->__pyx_vtab)->concat(__pyx_cur_scope->__pyx_v_state_prefix, __pyx_cur_scope->__pyx_v_run_vector, 0)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_5)->__pyx_vtab)->concat(((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_5), __pyx_cur_scope->__pyx_v_iteration_vector, 0)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GIVEREF(__pyx_t_9);
  __pyx_cur_scope->__pyx_v_seed = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_9);
  __pyx_t_9 = 0;

  /* "flitter/render/physics.pyx":465
 *         state.set_item(state_prefix.concat(ITERATION), iteration_vector)
 *         cdef Vector seed = state_prefix.concat(run_vector).concat(iteration_vector)
 *         cdef PhysicsGroup group, top_group=PhysicsGroup.__new__(PhysicsGroup)             # <<<<<<<<<<<<<<
 *         cdef list groups = [top_group]
 *         cdef dict particles_by_id = {}
 */
  __pyx_t_9 = ((PyObject *)__pyx_tp_new_7flitter_6render_7physics_PhysicsGroup(((PyTypeObject *)__pyx_ptype_7flitter_6render_7physics_PhysicsGroup), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 465, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_9);
  __Pyx_GIVEREF((PyObject *)__pyx_t_9);
  __pyx_cur_scope->__pyx_v_top_group = ((struct __pyx_obj_7flitter_6render_7physics_PhysicsGroup *)__pyx_t_9);
  __pyx_t_9 = 0;

  /* "flitter/render/physics.pyx":466
 *         cdef Vector seed = state_prefix.concat(run_vector).concat(iteration_vector)
 *         cdef PhysicsGroup group, top_group=PhysicsGroup.__new__(PhysicsGroup)
 *         cdef list groups = [top_group]             # <<<<<<<<<<<<<<
 *         cdef dict particles_by_id = {}
 *         cdef set old_state_keys = self.state_keys
 */
  __pyx_t_9 = PyList_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_top_group);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_top_group);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_9, 0, ((PyObject *)__pyx_cur_scope->__pyx_v_top_group))) __PYX_ERR(1, 466, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_9);
  __pyx_cur_scope->__pyx_v_groups = ((PyObject*)__pyx_t_9);
  __pyx_t_9 = 0;

  /* "flitter/render/physics.pyx":467
 *         cdef PhysicsGroup group, top_group=PhysicsGroup.__new__(PhysicsGroup)
 *         cdef list groups = [top_group]
 *         cdef dict particles_by_id = {}             # <<<<<<<<<<<<<<
 *         cdef set old_state_keys = self.state_keys
 *         cdef set new_state_keys = set()
 */
  __pyx_t_9 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 467, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_9);
  __pyx_cur_scope->__pyx_v_particles_by_id = ((PyObject*)__pyx_t_9);
  __pyx_t_9 = 0;

  /* "flitter/render/physics.pyx":468
 *         cdef list groups = [top_group]
 *         cdef dict particles_by_id = {}
 *         cdef set old_state_keys = self.state_keys             # <<<<<<<<<<<<<<
 *         cdef set new_state_keys = set()
 *         self.collect(groups, top_group, node, particles_by_id, old_state_keys, new_state_keys, state, state_prefix, clock, seed)
 */
  __pyx_t_9 = __pyx_cur_scope->__pyx_v_self->state_keys;
  __Pyx_INCREF(__pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_9);
  __pyx_cur_scope->__pyx_v_old_state_keys = ((PyObject*)__pyx_t_9);
  __pyx_t_9 = 0;

  /* "flitter/render/physics.pyx":469
 *         cdef dict particles_by_id = {}
 *         cdef set old_state_keys = self.state_keys
 *         cdef set new_state_keys = set()             # <<<<<<<<<<<<<<
 *         self.collect(groups, top_group, node, particles_by_id, old_state_keys, new_state_keys, state, state_prefix, clock, seed)
 *         for state_key in old_state_keys:
 */
  __pyx_t_9 = PySet_New(0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_9);
  __pyx_cur_scope->__pyx_v_new_state_keys = ((PyObject*)__pyx_t_9);
  __pyx_t_9 = 0;

  /* "flitter/render/physics.pyx":470
 *         cdef set old_state_keys = self.state_keys
 *         cdef set new_state_keys = set()
 *         self.collect(groups, top_group, node, particles_by_id, old_state_keys, new_state_keys, state, state_prefix, clock, seed)             # <<<<<<<<<<<<<<
 *         for state_key in old_state_keys:
 *             state.set_item(state_key, null_)
 */
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_cur_scope->__pyx_v_clock); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 470, __pyx_L1_error)
  ((struct __pyx_vtabstruct_7flitter_6render_7physics_PhysicsSystem *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->collect(__pyx_cur_scope->__pyx_v_self, __pyx_cur_scope->__pyx_v_groups, __pyx_cur_scope->__pyx_v_top_group, __pyx_cur_scope->__pyx_v_node, __pyx_cur_scope->__pyx_v_particles_by_id, __pyx_cur_scope->__pyx_v_old_state_keys, __pyx_cur_scope->__pyx_v_new_state_keys, __pyx_cur_scope->__pyx_v_state, __pyx_cur_scope->__pyx_v_state_prefix, __pyx_t_4, __pyx_cur_scope->__pyx_v_seed); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 470, __pyx_L1_error)

  /* "flitter/render/physics.pyx":471
 *         cdef set new_state_keys = set()
 *         self.collect(groups, top_group, node, particles_by_id, old_state_keys, new_state_keys, state, state_prefix, clock, seed)
 *         for state_key in old_state_keys:             # <<<<<<<<<<<<<<
 *             state.set_item(state_key, null_)
 *         self.state_keys = new_state_keys
 */
  __pyx_t_13 = 0;
  __pyx_t_5 = __Pyx_set_iterator(__pyx_cur_scope->__pyx_v_old_state_keys, 1, (&__pyx_t_12), (&__pyx_t_14)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 471, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_9);
  __pyx_t_9 = __pyx_t_5;
  __pyx_t_5 = 0;
  while (1) {
    __pyx_t_15 = __Pyx_set_iter_next(__pyx_t_9, __pyx_t_12, &__pyx_t_13, &__pyx_t_5, __pyx_t_14);
    if (unlikely(__pyx_t_15 == 0)) break;
    if (unlikely(__pyx_t_15 == -1)) __PYX_ERR(1, 471, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_7flitter_5model_Vector))))) __PYX_ERR(1, 471, __pyx_L1_error)
    __Pyx_XGOTREF((PyObject *)__pyx_cur_scope->__pyx_v_state_key);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_state_key, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_5));
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_5 = 0;

    /* "flitter/render/physics.pyx":472
 *         self.collect(groups, top_group, node, particles_by_id, old_state_keys, new_state_keys, state, state_prefix, clock, seed)
 *         for state_key in old_state_keys:
 *             state.set_item(state_key, null_)             # <<<<<<<<<<<<<<
 *         self.state_keys = new_state_keys
 *         time_vector = state.get_item(state_prefix)
 */
    __pyx_t_5 = ((PyObject *)__pyx_v_7flitter_5model_null_);
    __Pyx_INCREF(__pyx_t_5);
    ((struct __pyx_vtabstruct_7flitter_5model_StateDict *)__pyx_cur_scope->__pyx_v_state->__pyx_vtab)->set_item(__pyx_cur_scope->__pyx_v_state, __pyx_cur_scope->__pyx_v_state_key, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_5)); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 472, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "flitter/render/physics.pyx":473
 *         for state_key in old_state_keys:
 *             state.set_item(state_key, null_)
 *         self.state_keys = new_state_keys             # <<<<<<<<<<<<<<
 *         time_vector = state.get_item(state_prefix)
 *         cdef int64_t i
 */
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_new_state_keys);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_new_state_keys);
  __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_self->state_keys);
  __Pyx_DECREF(__pyx_cur_scope->__pyx_v_self->state_keys);
  __pyx_cur_scope->__pyx_v_self->state_keys = __pyx_cur_scope->__pyx_v_new_state_keys;

  /* "flitter/render/physics.pyx":474
 *             state.set_item(state_key, null_)
 *         self.state_keys = new_state_keys
 *         time_vector = state.get_item(state_prefix)             # <<<<<<<<<<<<<<
 *         cdef int64_t i
 *         if time_vector.length != 1 or time_vector.numbers == NULL:
 */
  __pyx_t_9 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_StateDict *)__pyx_cur_scope->__pyx_v_state->__pyx_vtab)->get_item(__pyx_cur_scope->__pyx_v_state, __pyx_cur_scope->__pyx_v_state_prefix)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 474, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GOTREF((PyObject *)__pyx_cur_scope->__pyx_v_time_vector);
  __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_time_vector, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_9));
  __Pyx_GIVEREF(__pyx_t_9);
  __pyx_t_9 = 0;

  /* "flitter/render/physics.pyx":476
 *         time_vector = state.get_item(state_prefix)
 *         cdef int64_t i
 *         if time_vector.length != 1 or time_vector.numbers == NULL:             # <<<<<<<<<<<<<<
 *             if len(groups) == 1:
 *                 logger.debug("New {}D physics {!r} with {} particles, {} forces and {} barriers", self.dimensions, state_prefix, len(top_group.particles),
 */
  __pyx_t_8 = (__pyx_cur_scope->__pyx_v_time_vector->length != 1);
  if (!__pyx_t_8) {
  } else {
    __pyx_t_2 = __pyx_t_8;
    goto __pyx_L25_bool_binop_done;
  }
  __pyx_t_8 = (__pyx_cur_scope->__pyx_v_time_vector->numbers == NULL);
  __pyx_t_2 = __pyx_t_8;
  __pyx_L25_bool_binop_done:;
  if (__pyx_t_2) {

    /* "flitter/render/physics.pyx":477
 *         cdef int64_t i
 *         if time_vector.length != 1 or time_vector.numbers == NULL:
 *             if len(groups) == 1:             # <<<<<<<<<<<<<<
 *                 logger.debug("New {}D physics {!r} with {} particles, {} forces and {} barriers", self.dimensions, state_prefix, len(top_group.particles),
 *                              len(top_group.particle_forces) + len(top_group.matrix_forces) + len(top_group.distance_forces), len(top_group.barriers))
 */
    __pyx_t_12 = __Pyx_PyList_GET_SIZE(__pyx_cur_scope->__pyx_v_groups); if (unlikely(__pyx_t_12 == ((Py_ssize_t)-1))) __PYX_ERR(1, 477, __pyx_L1_error)
    __pyx_t_2 = (__pyx_t_12 == 1);
    if (__pyx_t_2) {

      /* "flitter/render/physics.pyx":478
 *         if time_vector.length != 1 or time_vector.numbers == NULL:
 *             if len(groups) == 1:
 *                 logger.debug("New {}D physics {!r} with {} particles, {} forces and {} barriers", self.dimensions, state_prefix, len(top_group.particles),             # <<<<<<<<<<<<<<
 *                              len(top_group.particle_forces) + len(top_group.matrix_forces) + len(top_group.distance_forces), len(top_group.barriers))
 *             else:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_logger); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 478, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_debug); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 478, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyInt_From_int64_t(__pyx_cur_scope->__pyx_v_self->dimensions); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 478, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_10 = __pyx_cur_scope->__pyx_v_top_group->particles;
      __Pyx_INCREF(__pyx_t_10);
      if (unlikely(__pyx_t_10 == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
        __PYX_ERR(1, 478, __pyx_L1_error)
      }
      __pyx_t_12 = __Pyx_PyList_GET_SIZE(__pyx_t_10); if (unlikely(__pyx_t_12 == ((Py_ssize_t)-1))) __PYX_ERR(1, 478, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = PyInt_FromSsize_t(__pyx_t_12); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 478, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);

      /* "flitter/render/physics.pyx":479
 *             if len(groups) == 1:
 *                 logger.debug("New {}D physics {!r} with {} particles, {} forces and {} barriers", self.dimensions, state_prefix, len(top_group.particles),
 *                              len(top_group.particle_forces) + len(top_group.matrix_forces) + len(top_group.distance_forces), len(top_group.barriers))             # <<<<<<<<<<<<<<
 *             else:
 *                 logger.debug("New {}D physics {!r} with {} particles over {} groups", self.dimensions, state_prefix, len(particles_by_id), len(groups))
 */
      __pyx_t_16 = __pyx_cur_scope->__pyx_v_top_group->particle_forces;
      __Pyx_INCREF(__pyx_t_16);
      if (unlikely(__pyx_t_16 == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
        __PYX_ERR(1, 479, __pyx_L1_error)
      }
      __pyx_t_12 = __Pyx_PyList_GET_SIZE(__pyx_t_16); if (unlikely(__pyx_t_12 == ((Py_ssize_t)-1))) __PYX_ERR(1, 479, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
      __pyx_t_16 = __pyx_cur_scope->__pyx_v_top_group->matrix_forces;
      __Pyx_INCREF(__pyx_t_16);
      if (unlikely(__pyx_t_16 == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
        __PYX_ERR(1, 479, __pyx_L1_error)
      }
      __pyx_t_13 = __Pyx_PyList_GET_SIZE(__pyx_t_16); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(1, 479, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
      __pyx_t_16 = __pyx_cur_scope->__pyx_v_top_group->distance_forces;
      __Pyx_INCREF(__pyx_t_16);
      if (unlikely(__pyx_t_16 == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
        __PYX_ERR(1, 479, __pyx_L1_error)
      }
      __pyx_t_17 = __Pyx_PyList_GET_SIZE(__pyx_t_16); if (unlikely(__pyx_t_17 == ((Py_ssize_t)-1))) __PYX_ERR(1, 479, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
      __pyx_t_16 = PyInt_FromSsize_t(((__pyx_t_12 + __pyx_t_13) + __pyx_t_17)); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 479, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_16);
      __pyx_t_18 = __pyx_cur_scope->__pyx_v_top_group->barriers;
      __Pyx_INCREF(__pyx_t_18);
      if (unlikely(__pyx_t_18 == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
        __PYX_ERR(1, 479, __pyx_L1_error)
      }
      __pyx_t_17 = __Pyx_PyList_GET_SIZE(__pyx_t_18); if (unlikely(__pyx_t_17 == ((Py_ssize_t)-1))) __PYX_ERR(1, 479, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
      __pyx_t_18 = PyInt_FromSsize_t(__pyx_t_17); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 479, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_18);
      __pyx_t_19 = NULL;
      __pyx_t_11 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_19 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_19)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_19);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_11 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[7] = {__pyx_t_19, __pyx_kp_u_New_D_physics_r_with_particles_f, __pyx_t_5, ((PyObject *)__pyx_cur_scope->__pyx_v_state_prefix), __pyx_t_10, __pyx_t_16, __pyx_t_18};
        __pyx_t_9 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_11, 6+__pyx_t_11);
        __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
        if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 478, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "flitter/render/physics.pyx":477
 *         cdef int64_t i
 *         if time_vector.length != 1 or time_vector.numbers == NULL:
 *             if len(groups) == 1:             # <<<<<<<<<<<<<<
 *                 logger.debug("New {}D physics {!r} with {} particles, {} forces and {} barriers", self.dimensions, state_prefix, len(top_group.particles),
 *                              len(top_group.particle_forces) + len(top_group.matrix_forces) + len(top_group.distance_forces), len(top_group.barriers))
 */
      goto __pyx_L27;
    }

    /* "flitter/render/physics.pyx":481
 *                              len(top_group.particle_forces) + len(top_group.matrix_forces) + len(top_group.distance_forces), len(top_group.barriers))
 *             else:
 *                 logger.debug("New {}D physics {!r} with {} particles over {} groups", self.dimensions, state_prefix, len(particles_by_id), len(groups))             # <<<<<<<<<<<<<<
 *                 for i, group in enumerate(groups):
 *                     logger.trace("Group {} with {} particles, {} forces and {} barriers", i, len(group.particles),
 */
    /*else*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_logger); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 481, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_18 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_debug); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 481, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_18);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyInt_From_int64_t(__pyx_cur_scope->__pyx_v_self->dimensions); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 481, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_17 = PyDict_Size(__pyx_cur_scope->__pyx_v_particles_by_id); if (unlikely(__pyx_t_17 == ((Py_ssize_t)-1))) __PYX_ERR(1, 481, __pyx_L1_error)
      __pyx_t_16 = PyInt_FromSsize_t(__pyx_t_17); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 481, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_16);
      __pyx_t_17 = __Pyx_PyList_GET_SIZE(__pyx_cur_scope->__pyx_v_groups); if (unlikely(__pyx_t_17 == ((Py_ssize_t)-1))) __PYX_ERR(1, 481, __pyx_L1_error)
      __pyx_t_10 = PyInt_FromSsize_t(__pyx_t_17); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 481, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_5 = NULL;
      __pyx_t_11 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_18))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_18);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_18);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_18, function);
          __pyx_t_11 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[6] = {__pyx_t_5, __pyx_kp_u_New_D_physics_r_with_particles_o, __pyx_t_3, ((PyObject *)__pyx_cur_scope->__pyx_v_state_prefix), __pyx_t_16, __pyx_t_10};
        __pyx_t_9 = __Pyx_PyObject_FastCall(__pyx_t_18, __pyx_callargs+1-__pyx_t_11, 5+__pyx_t_11);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 481, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
      }
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "flitter/render/physics.pyx":482
 *             else:
 *                 logger.debug("New {}D physics {!r} with {} particles over {} groups", self.dimensions, state_prefix, len(particles_by_id), len(groups))
 *                 for i, group in enumerate(groups):             # <<<<<<<<<<<<<<
 *                     logger.trace("Group {} with {} particles, {} forces and {} barriers", i, len(group.particles),
 *                                  len(group.particle_forces) + len(group.matrix_forces) + len(group.distance_forces), len(group.barriers))
 */
      __pyx_t_1 = 0;
      __pyx_t_9 = __pyx_cur_scope->__pyx_v_groups; __Pyx_INCREF(__pyx_t_9);
      __pyx_t_17 = 0;
      for (;;) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_9);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(1, 482, __pyx_L1_error)
          #endif
          if (__pyx_t_17 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_18 = PyList_GET_ITEM(__pyx_t_9, __pyx_t_17); __Pyx_INCREF(__pyx_t_18); __pyx_t_17++; if (unlikely((0 < 0))) __PYX_ERR(1, 482, __pyx_L1_error)
        #else
        __pyx_t_18 = __Pyx_PySequence_ITEM(__pyx_t_9, __pyx_t_17); __pyx_t_17++; if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 482, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_18);
        #endif
        if (!(likely(((__pyx_t_18) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_18, __pyx_ptype_7flitter_6render_7physics_PhysicsGroup))))) __PYX_ERR(1, 482, __pyx_L1_error)
        __Pyx_XGOTREF((PyObject *)__pyx_cur_scope->__pyx_v_group);
        __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_group, ((struct __pyx_obj_7flitter_6render_7physics_PhysicsGroup *)__pyx_t_18));
        __Pyx_GIVEREF(__pyx_t_18);
        __pyx_t_18 = 0;
        __pyx_cur_scope->__pyx_v_i = __pyx_t_1;
        __pyx_t_1 = (__pyx_t_1 + 1);

        /* "flitter/render/physics.pyx":483
 *                 logger.debug("New {}D physics {!r} with {} particles over {} groups", self.dimensions, state_prefix, len(particles_by_id), len(groups))
 *                 for i, group in enumerate(groups):
 *                     logger.trace("Group {} with {} particles, {} forces and {} barriers", i, len(group.particles),             # <<<<<<<<<<<<<<
 *                                  len(group.particle_forces) + len(group.matrix_forces) + len(group.distance_forces), len(group.barriers))
 *             time_vector = Vector.__new__(Vector)
 */
        __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_logger); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 483, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_trace); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 483, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_10 = __Pyx_PyInt_From_int64_t(__pyx_cur_scope->__pyx_v_i); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 483, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_3 = __pyx_cur_scope->__pyx_v_group->particles;
        __Pyx_INCREF(__pyx_t_3);
        if (unlikely(__pyx_t_3 == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
          __PYX_ERR(1, 483, __pyx_L1_error)
        }
        __pyx_t_13 = __Pyx_PyList_GET_SIZE(__pyx_t_3); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(1, 483, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyInt_FromSsize_t(__pyx_t_13); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 483, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);

        /* "flitter/render/physics.pyx":484
 *                 for i, group in enumerate(groups):
 *                     logger.trace("Group {} with {} particles, {} forces and {} barriers", i, len(group.particles),
 *                                  len(group.particle_forces) + len(group.matrix_forces) + len(group.distance_forces), len(group.barriers))             # <<<<<<<<<<<<<<
 *             time_vector = Vector.__new__(Vector)
 *             time_vector.allocate_numbers(1)
 */
        __pyx_t_5 = __pyx_cur_scope->__pyx_v_group->particle_forces;
        __Pyx_INCREF(__pyx_t_5);
        if (unlikely(__pyx_t_5 == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
          __PYX_ERR(1, 484, __pyx_L1_error)
        }
        __pyx_t_13 = __Pyx_PyList_GET_SIZE(__pyx_t_5); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(1, 484, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_5 = __pyx_cur_scope->__pyx_v_group->matrix_forces;
        __Pyx_INCREF(__pyx_t_5);
        if (unlikely(__pyx_t_5 == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
          __PYX_ERR(1, 484, __pyx_L1_error)
        }
        __pyx_t_12 = __Pyx_PyList_GET_SIZE(__pyx_t_5); if (unlikely(__pyx_t_12 == ((Py_ssize_t)-1))) __PYX_ERR(1, 484, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_5 = __pyx_cur_scope->__pyx_v_group->distance_forces;
        __Pyx_INCREF(__pyx_t_5);
        if (unlikely(__pyx_t_5 == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
          __PYX_ERR(1, 484, __pyx_L1_error)
        }
        __pyx_t_20 = __Pyx_PyList_GET_SIZE(__pyx_t_5); if (unlikely(__pyx_t_20 == ((Py_ssize_t)-1))) __PYX_ERR(1, 484, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_5 = PyInt_FromSsize_t(((__pyx_t_13 + __pyx_t_12) + __pyx_t_20)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 484, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_19 = __pyx_cur_scope->__pyx_v_group->barriers;
        __Pyx_INCREF(__pyx_t_19);
        if (unlikely(__pyx_t_19 == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
          __PYX_ERR(1, 484, __pyx_L1_error)
        }
        __pyx_t_20 = __Pyx_PyList_GET_SIZE(__pyx_t_19); if (unlikely(__pyx_t_20 == ((Py_ssize_t)-1))) __PYX_ERR(1, 484, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
        __pyx_t_19 = PyInt_FromSsize_t(__pyx_t_20); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 484, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_19);
        __pyx_t_21 = NULL;
        __pyx_t_11 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_16))) {
          __pyx_t_21 = PyMethod_GET_SELF(__pyx_t_16);
          if (likely(__pyx_t_21)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_16);
            __Pyx_INCREF(__pyx_t_21);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_16, function);
            __pyx_t_11 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[6] = {__pyx_t_21, __pyx_kp_u_Group_with_particles_forces_and, __pyx_t_10, __pyx_t_3, __pyx_t_5, __pyx_t_19};
          __pyx_t_18 = __Pyx_PyObject_FastCall(__pyx_t_16, __pyx_callargs+1-__pyx_t_11, 5+__pyx_t_11);
          __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
          if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 483, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_18);
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        }
        __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

        /* "flitter/render/physics.pyx":482
 *             else:
 *                 logger.debug("New {}D physics {!r} with {} particles over {} groups", self.dimensions, state_prefix, len(particles_by_id), len(groups))
 *                 for i, group in enumerate(groups):             # <<<<<<<<<<<<<<
 *                     logger.trace("Group {} with {} particles, {} forces and {} barriers", i, len(group.particles),
 *                                  len(group.particle_forces) + len(group.matrix_forces) + len(group.distance_forces), len(group.barriers))
 */
      }
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __pyx_L27:;

    /* "flitter/render/physics.pyx":485
 *                     logger.trace("Group {} with {} particles, {} forces and {} barriers", i, len(group.particles),
 *                                  len(group.particle_forces) + len(group.matrix_forces) + len(group.distance_forces), len(group.barriers))
 *             time_vector = Vector.__new__(Vector)             # <<<<<<<<<<<<<<
 *             time_vector.allocate_numbers(1)
 *             time_vector.numbers[0] = time
 */
    __pyx_t_9 = __Pyx_tp_new(((PyObject *)__pyx_ptype_7flitter_5model_Vector), __pyx_empty_tuple); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 485, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (!(likely(__Pyx_TypeTest(__pyx_t_9, __pyx_ptype_7flitter_5model_Vector)))) __PYX_ERR(1, 485, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope->__pyx_v_time_vector);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_time_vector, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_9));
    __Pyx_GIVEREF(__pyx_t_9);
    __pyx_t_9 = 0;

    /* "flitter/render/physics.pyx":486
 *                                  len(group.particle_forces) + len(group.matrix_forces) + len(group.distance_forces), len(group.barriers))
 *             time_vector = Vector.__new__(Vector)
 *             time_vector.allocate_numbers(1)             # <<<<<<<<<<<<<<
 *             time_vector.numbers[0] = time
 *             state.set_item(state_prefix, time_vector)
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_cur_scope->__pyx_v_time_vector->__pyx_vtab)->allocate_numbers(__pyx_cur_scope->__pyx_v_time_vector, 1); if (unlikely(__pyx_t_1 == ((int64_t)-1L))) __PYX_ERR(1, 486, __pyx_L1_error)

    /* "flitter/render/physics.pyx":487
 *             time_vector = Vector.__new__(Vector)
 *             time_vector.allocate_numbers(1)
 *             time_vector.numbers[0] = time             # <<<<<<<<<<<<<<
 *             state.set_item(state_prefix, time_vector)
 *         cdef bint extra = performance > 1.5
 */
    (__pyx_cur_scope->__pyx_v_time_vector->numbers[0]) = __pyx_cur_scope->__pyx_v_time;

    /* "flitter/render/physics.pyx":488
 *             time_vector.allocate_numbers(1)
 *             time_vector.numbers[0] = time
 *             state.set_item(state_prefix, time_vector)             # <<<<<<<<<<<<<<
 *         cdef bint extra = performance > 1.5
 *         cdef Particle particle
 */
    ((struct __pyx_vtabstruct_7flitter_5model_StateDict *)__pyx_cur_scope->__pyx_v_state->__pyx_vtab)->set_item(__pyx_cur_scope->__pyx_v_state, __pyx_cur_scope->__pyx_v_state_prefix, __pyx_cur_scope->__pyx_v_time_vector); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 488, __pyx_L1_error)

    /* "flitter/render/physics.pyx":476
 *         time_vector = state.get_item(state_prefix)
 *         cdef int64_t i
 *         if time_vector.length != 1 or time_vector.numbers == NULL:             # <<<<<<<<<<<<<<
 *             if len(groups) == 1:
 *                 logger.debug("New {}D physics {!r} with {} particles, {} forces and {} barriers", self.dimensions, state_prefix, len(top_group.particles),
 */
  }

  /* "flitter/render/physics.pyx":489
 *             time_vector.numbers[0] = time
 *             state.set_item(state_prefix, time_vector)
 *         cdef bint extra = performance > 1.5             # <<<<<<<<<<<<<<
 *         cdef Particle particle
 *         if time > last_time:
 */
  __pyx_cur_scope->__pyx_v_extra = (__pyx_cur_scope->__pyx_v_performance > 1.5);

  /* "flitter/render/physics.pyx":491
 *         cdef bint extra = performance > 1.5
 *         cdef Particle particle
 *         if time > last_time:             # <<<<<<<<<<<<<<
 *             clock = await asyncio.to_thread(self.calculate, groups, particles_by_id, engine.realtime, extra, time, last_time, clock)
 *             for group in groups:
 */
  __pyx_t_2 = (__pyx_cur_scope->__pyx_v_time > __pyx_cur_scope->__pyx_v_last_time);
  if (__pyx_t_2) {

    /* "flitter/render/physics.pyx":492
 *         cdef Particle particle
 *         if time > last_time:
 *             clock = await asyncio.to_thread(self.calculate, groups, particles_by_id, engine.realtime, extra, time, last_time, clock)             # <<<<<<<<<<<<<<
 *             for group in groups:
 *                 for particle in group.particles:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_18, __pyx_n_s_asyncio); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 492, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_t_18, __pyx_n_s_to_thread); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 492, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    __pyx_t_18 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_n_s_calculate); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 492, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __pyx_t_19 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_engine, __pyx_n_s_realtime); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 492, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_19);
    __pyx_t_5 = __Pyx_PyBool_FromLong(__pyx_cur_scope->__pyx_v_extra); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 492, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = PyFloat_FromDouble(__pyx_cur_scope->__pyx_v_time); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 492, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_10 = PyFloat_FromDouble(__pyx_cur_scope->__pyx_v_last_time); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 492, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_21 = NULL;
    __pyx_t_11 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_16))) {
      __pyx_t_21 = PyMethod_GET_SELF(__pyx_t_16);
      if (likely(__pyx_t_21)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_16);
        __Pyx_INCREF(__pyx_t_21);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_16, function);
        __pyx_t_11 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[9] = {__pyx_t_21, __pyx_t_18, __pyx_cur_scope->__pyx_v_groups, __pyx_cur_scope->__pyx_v_particles_by_id, __pyx_t_19, __pyx_t_5, __pyx_t_3, __pyx_t_10, __pyx_cur_scope->__pyx_v_clock};
      __pyx_t_9 = __Pyx_PyObject_FastCall(__pyx_t_16, __pyx_callargs+1-__pyx_t_11, 8+__pyx_t_11);
      __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
      __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
      __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 492, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    }
    __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_t_9);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_XGOTREF(__pyx_r);
    if (likely(__pyx_r)) {
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
      /* return from generator, awaiting value */
      __pyx_generator->resume_label = 1;
      return __pyx_r;
      __pyx_L32_resume_from_await:;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(1, 492, __pyx_L1_error)
      __pyx_t_9 = __pyx_sent_value; __Pyx_INCREF(__pyx_t_9);
    } else {
      __pyx_t_9 = NULL;
      if (__Pyx_PyGen_FetchStopIterationValue(&__pyx_t_9) < 0) __PYX_ERR(1, 492, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
    }
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_clock);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_clock, __pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_9);
    __pyx_t_9 = 0;

    /* "flitter/render/physics.pyx":493
 *         if time > last_time:
 *             clock = await asyncio.to_thread(self.calculate, groups, particles_by_id, engine.realtime, extra, time, last_time, clock)
 *             for group in groups:             # <<<<<<<<<<<<<<
 *                 for particle in group.particles:
 *                     if particle.position.is_finite():
 */
    __pyx_t_9 = __pyx_cur_scope->__pyx_v_groups; __Pyx_INCREF(__pyx_t_9);
    __pyx_t_17 = 0;
    for (;;) {
      {
        Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_9);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely((__pyx_temp < 0))) __PYX_ERR(1, 493, __pyx_L1_error)
        #endif
        if (__pyx_t_17 >= __pyx_temp) break;
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_16 = PyList_GET_ITEM(__pyx_t_9, __pyx_t_17); __Pyx_INCREF(__pyx_t_16); __pyx_t_17++; if (unlikely((0 < 0))) __PYX_ERR(1, 493, __pyx_L1_error)
      #else
      __pyx_t_16 = __Pyx_PySequence_ITEM(__pyx_t_9, __pyx_t_17); __pyx_t_17++; if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 493, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_16);
      #endif
      if (!(likely(((__pyx_t_16) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_16, __pyx_ptype_7flitter_6render_7physics_PhysicsGroup))))) __PYX_ERR(1, 493, __pyx_L1_error)
      __Pyx_XGOTREF((PyObject *)__pyx_cur_scope->__pyx_v_group);
      __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_group, ((struct __pyx_obj_7flitter_6render_7physics_PhysicsGroup *)__pyx_t_16));
      __Pyx_GIVEREF(__pyx_t_16);
      __pyx_t_16 = 0;

      /* "flitter/render/physics.pyx":494
 *             clock = await asyncio.to_thread(self.calculate, groups, particles_by_id, engine.realtime, extra, time, last_time, clock)
 *             for group in groups:
 *                 for particle in group.particles:             # <<<<<<<<<<<<<<
 *                     if particle.position.is_finite():
 *                         state.set_item(particle.position_state_key, particle.position)
 */
      if (unlikely(__pyx_cur_scope->__pyx_v_group->particles == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
        __PYX_ERR(1, 494, __pyx_L1_error)
      }
      __pyx_t_16 = __pyx_cur_scope->__pyx_v_group->particles; __Pyx_INCREF(__pyx_t_16);
      __pyx_t_20 = 0;
      for (;;) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_16);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(1, 494, __pyx_L1_error)
          #endif
          if (__pyx_t_20 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_10 = PyList_GET_ITEM(__pyx_t_16, __pyx_t_20); __Pyx_INCREF(__pyx_t_10); __pyx_t_20++; if (unlikely((0 < 0))) __PYX_ERR(1, 494, __pyx_L1_error)
        #else
        __pyx_t_10 = __Pyx_PySequence_ITEM(__pyx_t_16, __pyx_t_20); __pyx_t_20++; if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 494, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        #endif
        if (!(likely(((__pyx_t_10) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_10, __pyx_ptype_7flitter_6render_7physics_Particle))))) __PYX_ERR(1, 494, __pyx_L1_error)
        __Pyx_XGOTREF((PyObject *)__pyx_cur_scope->__pyx_v_particle);
        __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_particle, ((struct __pyx_obj_7flitter_6render_7physics_Particle *)__pyx_t_10));
        __Pyx_GIVEREF(__pyx_t_10);
        __pyx_t_10 = 0;

        /* "flitter/render/physics.pyx":495
 *             for group in groups:
 *                 for particle in group.particles:
 *                     if particle.position.is_finite():             # <<<<<<<<<<<<<<
 *                         state.set_item(particle.position_state_key, particle.position)
 *                     if particle.velocity.is_finite():
 */
        __pyx_t_2 = ((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_cur_scope->__pyx_v_particle->position->__pyx_vtab)->is_finite(__pyx_cur_scope->__pyx_v_particle->position, 0);
        if (__pyx_t_2) {

          /* "flitter/render/physics.pyx":496
 *                 for particle in group.particles:
 *                     if particle.position.is_finite():
 *                         state.set_item(particle.position_state_key, particle.position)             # <<<<<<<<<<<<<<
 *                     if particle.velocity.is_finite():
 *                         state.set_item(particle.velocity_state_key, particle.velocity)
 */
          __pyx_t_10 = ((PyObject *)__pyx_cur_scope->__pyx_v_particle->position_state_key);
          __Pyx_INCREF(__pyx_t_10);
          __pyx_t_3 = ((PyObject *)__pyx_cur_scope->__pyx_v_particle->position);
          __Pyx_INCREF(__pyx_t_3);
          ((struct __pyx_vtabstruct_7flitter_5model_StateDict *)__pyx_cur_scope->__pyx_v_state->__pyx_vtab)->set_item(__pyx_cur_scope->__pyx_v_state, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10), ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_3)); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 496, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "flitter/render/physics.pyx":495
 *             for group in groups:
 *                 for particle in group.particles:
 *                     if particle.position.is_finite():             # <<<<<<<<<<<<<<
 *                         state.set_item(particle.position_state_key, particle.position)
 *                     if particle.velocity.is_finite():
 */
        }

        /* "flitter/render/physics.pyx":497
 *                     if particle.position.is_finite():
 *                         state.set_item(particle.position_state_key, particle.position)
 *                     if particle.velocity.is_finite():             # <<<<<<<<<<<<<<
 *                         state.set_item(particle.velocity_state_key, particle.velocity)
 *         time_vector = Vector.__new__(Vector)
 */
        __pyx_t_2 = ((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_cur_scope->__pyx_v_particle->velocity->__pyx_vtab)->is_finite(__pyx_cur_scope->__pyx_v_particle->velocity, 0);
        if (__pyx_t_2) {

          /* "flitter/render/physics.pyx":498
 *                         state.set_item(particle.position_state_key, particle.position)
 *                     if particle.velocity.is_finite():
 *                         state.set_item(particle.velocity_state_key, particle.velocity)             # <<<<<<<<<<<<<<
 *         time_vector = Vector.__new__(Vector)
 *         time_vector.allocate_numbers(1)
 */
          __pyx_t_3 = ((PyObject *)__pyx_cur_scope->__pyx_v_particle->velocity_state_key);
          __Pyx_INCREF(__pyx_t_3);
          __pyx_t_10 = ((PyObject *)__pyx_cur_scope->__pyx_v_particle->velocity);
          __Pyx_INCREF(__pyx_t_10);
          ((struct __pyx_vtabstruct_7flitter_5model_StateDict *)__pyx_cur_scope->__pyx_v_state->__pyx_vtab)->set_item(__pyx_cur_scope->__pyx_v_state, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_3), ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_10)); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 498, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "flitter/render/physics.pyx":497
 *                     if particle.position.is_finite():
 *                         state.set_item(particle.position_state_key, particle.position)
 *                     if particle.velocity.is_finite():             # <<<<<<<<<<<<<<
 *                         state.set_item(particle.velocity_state_key, particle.velocity)
 *         time_vector = Vector.__new__(Vector)
 */
        }

        /* "flitter/render/physics.pyx":494
 *             clock = await asyncio.to_thread(self.calculate, groups, particles_by_id, engine.realtime, extra, time, last_time, clock)
 *             for group in groups:
 *                 for particle in group.particles:             # <<<<<<<<<<<<<<
 *                     if particle.position.is_finite():
 *                         state.set_item(particle.position_state_key, particle.position)
 */
      }
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

      /* "flitter/render/physics.pyx":493
 *         if time > last_time:
 *             clock = await asyncio.to_thread(self.calculate, groups, particles_by_id, engine.realtime, extra, time, last_time, clock)
 *             for group in groups:             # <<<<<<<<<<<<<<
 *                 for particle in group.particles:
 *                     if particle.position.is_finite():
 */
    }
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "flitter/render/physics.pyx":491
 *         cdef bint extra = performance > 1.5
 *         cdef Particle particle
 *         if time > last_time:             # <<<<<<<<<<<<<<
 *             clock = await asyncio.to_thread(self.calculate, groups, particles_by_id, engine.realtime, extra, time, last_time, clock)
 *             for group in groups:
 */
  }

  /* "flitter/render/physics.pyx":499
 *                     if particle.velocity.is_finite():
 *                         state.set_item(particle.velocity_state_key, particle.velocity)
 *         time_vector = Vector.__new__(Vector)             # <<<<<<<<<<<<<<
 *         time_vector.allocate_numbers(1)
 *         time_vector.numbers[0] = time
 */
  __pyx_t_9 = __Pyx_tp_new(((PyObject *)__pyx_ptype_7flitter_5model_Vector), __pyx_empty_tuple); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (!(likely(__Pyx_TypeTest(__pyx_t_9, __pyx_ptype_7flitter_5model_Vector)))) __PYX_ERR(1, 499, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_cur_scope->__pyx_v_time_vector);
  __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_time_vector, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_9));
  __Pyx_GIVEREF(__pyx_t_9);
  __pyx_t_9 = 0;

  /* "flitter/render/physics.pyx":500
 *                         state.set_item(particle.velocity_state_key, particle.velocity)
 *         time_vector = Vector.__new__(Vector)
 *         time_vector.allocate_numbers(1)             # <<<<<<<<<<<<<<
 *         time_vector.numbers[0] = time
 *         state.set_item(state_prefix.concat(LAST), time_vector)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_cur_scope->__pyx_v_time_vector->__pyx_vtab)->allocate_numbers(__pyx_cur_scope->__pyx_v_time_vector, 1); if (unlikely(__pyx_t_1 == ((int64_t)-1L))) __PYX_ERR(1, 500, __pyx_L1_error)

  /* "flitter/render/physics.pyx":501
 *         time_vector = Vector.__new__(Vector)
 *         time_vector.allocate_numbers(1)
 *         time_vector.numbers[0] = time             # <<<<<<<<<<<<<<
 *         state.set_item(state_prefix.concat(LAST), time_vector)
 *         time_vector = Vector.__new__(Vector)
 */
  (__pyx_cur_scope->__pyx_v_time_vector->numbers[0]) = __pyx_cur_scope->__pyx_v_time;

  /* "flitter/render/physics.pyx":502
 *         time_vector.allocate_numbers(1)
 *         time_vector.numbers[0] = time
 *         state.set_item(state_prefix.concat(LAST), time_vector)             # <<<<<<<<<<<<<<
 *         time_vector = Vector.__new__(Vector)
 *         time_vector.allocate_numbers(1)
 */
  __pyx_t_9 = ((PyObject *)__pyx_v_7flitter_6render_7physics_LAST);
  __Pyx_INCREF(__pyx_t_9);
  __pyx_t_16 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_cur_scope->__pyx_v_state_prefix->__pyx_vtab)->concat(__pyx_cur_scope->__pyx_v_state_prefix, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_9), 0)); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  ((struct __pyx_vtabstruct_7flitter_5model_StateDict *)__pyx_cur_scope->__pyx_v_state->__pyx_vtab)->set_item(__pyx_cur_scope->__pyx_v_state, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_16), __pyx_cur_scope->__pyx_v_time_vector); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 502, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "flitter/render/physics.pyx":503
 *         time_vector.numbers[0] = time
 *         state.set_item(state_prefix.concat(LAST), time_vector)
 *         time_vector = Vector.__new__(Vector)             # <<<<<<<<<<<<<<
 *         time_vector.allocate_numbers(1)
 *         time_vector.numbers[0] = clock
 */
  __pyx_t_16 = __Pyx_tp_new(((PyObject *)__pyx_ptype_7flitter_5model_Vector), __pyx_empty_tuple); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  if (!(likely(__Pyx_TypeTest(__pyx_t_16, __pyx_ptype_7flitter_5model_Vector)))) __PYX_ERR(1, 503, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_cur_scope->__pyx_v_time_vector);
  __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_time_vector, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_16));
  __Pyx_GIVEREF(__pyx_t_16);
  __pyx_t_16 = 0;

  /* "flitter/render/physics.pyx":504
 *         state.set_item(state_prefix.concat(LAST), time_vector)
 *         time_vector = Vector.__new__(Vector)
 *         time_vector.allocate_numbers(1)             # <<<<<<<<<<<<<<
 *         time_vector.numbers[0] = clock
 *         state.set_item(state_prefix.concat(CLOCK), time_vector)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_cur_scope->__pyx_v_time_vector->__pyx_vtab)->allocate_numbers(__pyx_cur_scope->__pyx_v_time_vector, 1); if (unlikely(__pyx_t_1 == ((int64_t)-1L))) __PYX_ERR(1, 504, __pyx_L1_error)

  /* "flitter/render/physics.pyx":505
 *         time_vector = Vector.__new__(Vector)
 *         time_vector.allocate_numbers(1)
 *         time_vector.numbers[0] = clock             # <<<<<<<<<<<<<<
 *         state.set_item(state_prefix.concat(CLOCK), time_vector)
 * 
 */
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_cur_scope->__pyx_v_clock); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 505, __pyx_L1_error)
  (__pyx_cur_scope->__pyx_v_time_vector->numbers[0]) = __pyx_t_4;

  /* "flitter/render/physics.pyx":506
 *         time_vector.allocate_numbers(1)
 *         time_vector.numbers[0] = clock
 *         state.set_item(state_prefix.concat(CLOCK), time_vector)             # <<<<<<<<<<<<<<
 * 
 *     cdef void collect(self, list groups, PhysicsGroup group, Node node, dict particles_by_id, set old_state_keys, set new_state_keys,
 */
  __pyx_t_16 = ((PyObject *)__pyx_v_7flitter_6render_7physics_CLOCK);
  __Pyx_INCREF(__pyx_t_16);
  __pyx_t_9 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_cur_scope->__pyx_v_state_prefix->__pyx_vtab)->concat(__pyx_cur_scope->__pyx_v_state_prefix, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_16), 0)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 506, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  ((struct __pyx_vtabstruct_7flitter_5model_StateDict *)__pyx_cur_scope->__pyx_v_state->__pyx_vtab)->set_item(__pyx_cur_scope->__pyx_v_state, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_9), __pyx_cur_scope->__pyx_v_time_vector); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 506, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "flitter/render/physics.pyx":407
 *         pass
 * 
 *     async def update(self, engine, Node node, double time, double performance, **kwargs):             # <<<<<<<<<<<<<<
 *         cdef int64_t run = node.get_int('run', 0)
 *         self.dimensions = node.get_int('dimensions', 0)
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_21);
  __Pyx_AddTraceback("update", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "flitter/render/physics.pyx":508
 *         state.set_item(state_prefix.concat(CLOCK), time_vector)
 * 
 *     cdef void collect(self, list groups, PhysicsGroup group, Node node, dict particles_by_id, set old_state_keys, set new_state_keys,             # <<<<<<<<<<<<<<
 *                       StateDict state, Vector state_prefix, double clock, Vector seed):
 *         cdef Node child
 */

static void __pyx_f_7flitter_6render_7physics_13PhysicsSystem_collect(struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem *__pyx_v_self, PyObject *__pyx_v_groups, struct __pyx_obj_7flitter_6render_7physics_PhysicsGroup *__pyx_v_group, struct __pyx_obj_7flitter_5model_Node *__pyx_v_node, PyObject *__pyx_v_particles_by_id, PyObject *__pyx_v_old_state_keys, PyObject *__pyx_v_new_state_keys, struct __pyx_obj_7flitter_5model_StateDict *__pyx_v_state, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_state_prefix, double __pyx_v_clock, struct __pyx_obj_7flitter_5model_Vector *__pyx_v_seed) {
  struct __pyx_obj_7flitter_5model_Node *__pyx_v_child = 0;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_particle_id = 0;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_from_particle_id = 0;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_to_particle_id = 0;
  double __pyx_v_strength;
  double __pyx_v_ease;
  struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v_particle = 0;
  struct __pyx_obj_7flitter_6render_7physics_ParticleForceApplier *__pyx_v_particle_force = 0;
  struct __pyx_obj_7flitter_6render_7physics_DistanceForceApplier *__pyx_v_distance_force = 0;
  struct __pyx_obj_7flitter_6render_7physics_Barrier *__pyx_v_barrier = 0;
  struct __pyx_obj_7flitter_6render_7physics_PhysicsGroup *__pyx_v_child_group = 0;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_zero = 0;
  int64_t __pyx_v_i;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_position = 0;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_velocity = 0;
  double __pyx_v_fixed;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int64_t __pyx_t_2;
  int64_t __pyx_t_3;
  int64_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  double __pyx_t_10;
  long __pyx_t_11;
  double __pyx_t_12;
  int __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("collect", 0);
  __Pyx_INCREF((PyObject *)__pyx_v_seed);

  /* "flitter/render/physics.pyx":518
 *         cdef Barrier barrier
 *         cdef PhysicsGroup child_group
 *         cdef Vector zero = Vector.__new__(Vector)             # <<<<<<<<<<<<<<
 *         zero.allocate_numbers(self.dimensions)
 *         cdef int64_t i
 */
  __pyx_t_1 = __Pyx_tp_new(((PyObject *)__pyx_ptype_7flitter_5model_Vector), __pyx_empty_tuple); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_5model_Vector)))) __PYX_ERR(1, 518, __pyx_L1_error)
  __pyx_v_zero = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/render/physics.pyx":519
 *         cdef PhysicsGroup child_group
 *         cdef Vector zero = Vector.__new__(Vector)
 *         zero.allocate_numbers(self.dimensions)             # <<<<<<<<<<<<<<
 *         cdef int64_t i
 *         for i in range(self.dimensions):
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_zero->__pyx_vtab)->allocate_numbers(__pyx_v_zero, __pyx_v_self->dimensions); if (unlikely(__pyx_t_2 == ((int64_t)-1L))) __PYX_ERR(1, 519, __pyx_L1_error)

  /* "flitter/render/physics.pyx":521
 *         zero.allocate_numbers(self.dimensions)
 *         cdef int64_t i
 *         for i in range(self.dimensions):             # <<<<<<<<<<<<<<
 *             zero.numbers[i] = 0
 *         cdef Vector position, velocity
 */
  __pyx_t_2 = __pyx_v_self->dimensions;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "flitter/render/physics.pyx":522
 *         cdef int64_t i
 *         for i in range(self.dimensions):
 *             zero.numbers[i] = 0             # <<<<<<<<<<<<<<
 *         cdef Vector position, velocity
 *         cdef double fixed
 */
    (__pyx_v_zero->numbers[__pyx_v_i]) = 0.0;
  }

  /* "flitter/render/physics.pyx":525
 *         cdef Vector position, velocity
 *         cdef double fixed
 *         for child in node._children:             # <<<<<<<<<<<<<<
 *             if child.kind is 'particle' or child.kind is 'anchor':
 *                 if (particle_id := child.get_vec('id', None)) is not None:
 */
  if (unlikely(__pyx_v_node->_children == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 525, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_node->_children; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_5 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(1, 525, __pyx_L1_error)
      #endif
      if (__pyx_t_5 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_5); __Pyx_INCREF(__pyx_t_6); __pyx_t_5++; if (unlikely((0 < 0))) __PYX_ERR(1, 525, __pyx_L1_error)
    #else
    __pyx_t_6 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 525, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    #endif
    if (!(likely(((__pyx_t_6) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_6, __pyx_ptype_7flitter_5model_Node))))) __PYX_ERR(1, 525, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_child, ((struct __pyx_obj_7flitter_5model_Node *)__pyx_t_6));
    __pyx_t_6 = 0;

    /* "flitter/render/physics.pyx":526
 *         cdef double fixed
 *         for child in node._children:
 *             if child.kind is 'particle' or child.kind is 'anchor':             # <<<<<<<<<<<<<<
 *                 if (particle_id := child.get_vec('id', None)) is not None:
 *                     particle = Particle.__new__(Particle)
 */
    __pyx_t_8 = (__pyx_v_child->kind == __pyx_n_u_particle);
    if (!__pyx_t_8) {
    } else {
      __pyx_t_7 = __pyx_t_8;
      goto __pyx_L8_bool_binop_done;
    }
    __pyx_t_8 = (__pyx_v_child->kind == __pyx_n_u_anchor);
    __pyx_t_7 = __pyx_t_8;
    __pyx_L8_bool_binop_done:;
    if (__pyx_t_7) {

      /* "flitter/render/physics.pyx":527
 *         for child in node._children:
 *             if child.kind is 'particle' or child.kind is 'anchor':
 *                 if (particle_id := child.get_vec('id', None)) is not None:             # <<<<<<<<<<<<<<
 *                     particle = Particle.__new__(Particle)
 *                     particle.id = particle_id.intern()
 */
      __pyx_t_6 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Node *)__pyx_v_child->__pyx_vtab)->get_vec(__pyx_v_child, __pyx_n_u_id, ((struct __pyx_obj_7flitter_5model_Vector *)Py_None))); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 527, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_v_particle_id, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_6));
      __Pyx_INCREF(__pyx_t_6);
      __pyx_t_7 = (((PyObject *)__pyx_t_6) != Py_None);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (__pyx_t_7) {

        /* "flitter/render/physics.pyx":528
 *             if child.kind is 'particle' or child.kind is 'anchor':
 *                 if (particle_id := child.get_vec('id', None)) is not None:
 *                     particle = Particle.__new__(Particle)             # <<<<<<<<<<<<<<
 *                     particle.id = particle_id.intern()
 *                     particle.position_state_key = state_prefix.concat(particle_id).intern()
 */
        __pyx_t_6 = ((PyObject *)__pyx_tp_new_7flitter_6render_7physics_Particle(((PyTypeObject *)__pyx_ptype_7flitter_6render_7physics_Particle), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 528, __pyx_L1_error)
        __Pyx_GOTREF((PyObject *)__pyx_t_6);
        __Pyx_XDECREF_SET(__pyx_v_particle, ((struct __pyx_obj_7flitter_6render_7physics_Particle *)__pyx_t_6));
        __pyx_t_6 = 0;

        /* "flitter/render/physics.pyx":529
 *                 if (particle_id := child.get_vec('id', None)) is not None:
 *                     particle = Particle.__new__(Particle)
 *                     particle.id = particle_id.intern()             # <<<<<<<<<<<<<<
 *                     particle.position_state_key = state_prefix.concat(particle_id).intern()
 *                     particle.velocity_state_key = particle.position_state_key.concat(VELOCITY).intern()
 */
        __pyx_t_6 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_particle_id->__pyx_vtab)->intern(__pyx_v_particle_id, 0)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 529, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_6);
        __Pyx_GOTREF((PyObject *)__pyx_v_particle->id);
        __Pyx_DECREF((PyObject *)__pyx_v_particle->id);
        __pyx_v_particle->id = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_6);
        __pyx_t_6 = 0;

        /* "flitter/render/physics.pyx":530
 *                     particle = Particle.__new__(Particle)
 *                     particle.id = particle_id.intern()
 *                     particle.position_state_key = state_prefix.concat(particle_id).intern()             # <<<<<<<<<<<<<<
 *                     particle.velocity_state_key = particle.position_state_key.concat(VELOCITY).intern()
 *                     particle.non_anchor = child.kind is 'particle'
 */
        __pyx_t_6 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_state_prefix->__pyx_vtab)->concat(__pyx_v_state_prefix, __pyx_v_particle_id, 0)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 530, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_9 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_6)->__pyx_vtab)->intern(((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_6), 0)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 530, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GIVEREF(__pyx_t_9);
        __Pyx_GOTREF((PyObject *)__pyx_v_particle->position_state_key);
        __Pyx_DECREF((PyObject *)__pyx_v_particle->position_state_key);
        __pyx_v_particle->position_state_key = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_9);
        __pyx_t_9 = 0;

        /* "flitter/render/physics.pyx":531
 *                     particle.id = particle_id.intern()
 *                     particle.position_state_key = state_prefix.concat(particle_id).intern()
 *                     particle.velocity_state_key = particle.position_state_key.concat(VELOCITY).intern()             # <<<<<<<<<<<<<<
 *                     particle.non_anchor = child.kind is 'particle'
 *                     if particle.non_anchor:
 */
        __pyx_t_9 = ((PyObject *)__pyx_v_7flitter_6render_7physics_VELOCITY);
        __Pyx_INCREF(__pyx_t_9);
        __pyx_t_6 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_particle->position_state_key->__pyx_vtab)->concat(__pyx_v_particle->position_state_key, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_9), 0)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 531, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_9 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_6)->__pyx_vtab)->intern(((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_6), 0)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 531, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GIVEREF(__pyx_t_9);
        __Pyx_GOTREF((PyObject *)__pyx_v_particle->velocity_state_key);
        __Pyx_DECREF((PyObject *)__pyx_v_particle->velocity_state_key);
        __pyx_v_particle->velocity_state_key = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_9);
        __pyx_t_9 = 0;

        /* "flitter/render/physics.pyx":532
 *                     particle.position_state_key = state_prefix.concat(particle_id).intern()
 *                     particle.velocity_state_key = particle.position_state_key.concat(VELOCITY).intern()
 *                     particle.non_anchor = child.kind is 'particle'             # <<<<<<<<<<<<<<
 *                     if particle.non_anchor:
 *                         position = state.get_item(particle.position_state_key)
 */
        __pyx_t_7 = (__pyx_v_child->kind == __pyx_n_u_particle);
        __pyx_v_particle->non_anchor = __pyx_t_7;

        /* "flitter/render/physics.pyx":533
 *                     particle.velocity_state_key = particle.position_state_key.concat(VELOCITY).intern()
 *                     particle.non_anchor = child.kind is 'particle'
 *                     if particle.non_anchor:             # <<<<<<<<<<<<<<
 *                         position = state.get_item(particle.position_state_key)
 *                         if position.length == zero.length and position.numbers != NULL:
 */
        if (__pyx_v_particle->non_anchor) {

          /* "flitter/render/physics.pyx":534
 *                     particle.non_anchor = child.kind is 'particle'
 *                     if particle.non_anchor:
 *                         position = state.get_item(particle.position_state_key)             # <<<<<<<<<<<<<<
 *                         if position.length == zero.length and position.numbers != NULL:
 *                             particle.position = position.copy()
 */
          __pyx_t_9 = ((PyObject *)__pyx_v_particle->position_state_key);
          __Pyx_INCREF(__pyx_t_9);
          __pyx_t_6 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_StateDict *)__pyx_v_state->__pyx_vtab)->get_item(__pyx_v_state, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_9))); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 534, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_XDECREF_SET(__pyx_v_position, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_6));
          __pyx_t_6 = 0;

          /* "flitter/render/physics.pyx":535
 *                     if particle.non_anchor:
 *                         position = state.get_item(particle.position_state_key)
 *                         if position.length == zero.length and position.numbers != NULL:             # <<<<<<<<<<<<<<
 *                             particle.position = position.copy()
 *                         else:
 */
          __pyx_t_8 = (__pyx_v_position->length == __pyx_v_zero->length);
          if (__pyx_t_8) {
          } else {
            __pyx_t_7 = __pyx_t_8;
            goto __pyx_L13_bool_binop_done;
          }
          __pyx_t_8 = (__pyx_v_position->numbers != NULL);
          __pyx_t_7 = __pyx_t_8;
          __pyx_L13_bool_binop_done:;
          if (__pyx_t_7) {

            /* "flitter/render/physics.pyx":536
 *                         position = state.get_item(particle.position_state_key)
 *                         if position.length == zero.length and position.numbers != NULL:
 *                             particle.position = position.copy()             # <<<<<<<<<<<<<<
 *                         else:
 *                             particle.position = child.get_fvec('position', zero.length, zero).copy()
 */
            __pyx_t_6 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_position->__pyx_vtab)->copy(__pyx_v_position, 0)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 536, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_GIVEREF(__pyx_t_6);
            __Pyx_GOTREF((PyObject *)__pyx_v_particle->position);
            __Pyx_DECREF((PyObject *)__pyx_v_particle->position);
            __pyx_v_particle->position = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_6);
            __pyx_t_6 = 0;

            /* "flitter/render/physics.pyx":535
 *                     if particle.non_anchor:
 *                         position = state.get_item(particle.position_state_key)
 *                         if position.length == zero.length and position.numbers != NULL:             # <<<<<<<<<<<<<<
 *                             particle.position = position.copy()
 *                         else:
 */
            goto __pyx_L12;
          }

          /* "flitter/render/physics.pyx":538
 *                             particle.position = position.copy()
 *                         else:
 *                             particle.position = child.get_fvec('position', zero.length, zero).copy()             # <<<<<<<<<<<<<<
 *                         velocity = state.get_item(particle.velocity_state_key)
 *                         if velocity.length == zero.length and velocity.numbers != NULL:
 */
          /*else*/ {
            __pyx_t_6 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Node *)__pyx_v_child->__pyx_vtab)->get_fvec(__pyx_v_child, __pyx_n_u_position, __pyx_v_zero->length, __pyx_v_zero)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 538, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_6);
            __pyx_t_9 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_6)->__pyx_vtab)->copy(((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_6), 0)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 538, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_9);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_GIVEREF(__pyx_t_9);
            __Pyx_GOTREF((PyObject *)__pyx_v_particle->position);
            __Pyx_DECREF((PyObject *)__pyx_v_particle->position);
            __pyx_v_particle->position = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_9);
            __pyx_t_9 = 0;
          }
          __pyx_L12:;

          /* "flitter/render/physics.pyx":539
 *                         else:
 *                             particle.position = child.get_fvec('position', zero.length, zero).copy()
 *                         velocity = state.get_item(particle.velocity_state_key)             # <<<<<<<<<<<<<<
 *                         if velocity.length == zero.length and velocity.numbers != NULL:
 *                             particle.velocity = velocity.copy()
 */
          __pyx_t_9 = ((PyObject *)__pyx_v_particle->velocity_state_key);
          __Pyx_INCREF(__pyx_t_9);
          __pyx_t_6 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_StateDict *)__pyx_v_state->__pyx_vtab)->get_item(__pyx_v_state, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_9))); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 539, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_XDECREF_SET(__pyx_v_velocity, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_6));
          __pyx_t_6 = 0;

          /* "flitter/render/physics.pyx":540
 *                             particle.position = child.get_fvec('position', zero.length, zero).copy()
 *                         velocity = state.get_item(particle.velocity_state_key)
 *                         if velocity.length == zero.length and velocity.numbers != NULL:             # <<<<<<<<<<<<<<
 *                             particle.velocity = velocity.copy()
 *                         else:
 */
          __pyx_t_8 = (__pyx_v_velocity->length == __pyx_v_zero->length);
          if (__pyx_t_8) {
          } else {
            __pyx_t_7 = __pyx_t_8;
            goto __pyx_L16_bool_binop_done;
          }
          __pyx_t_8 = (__pyx_v_velocity->numbers != NULL);
          __pyx_t_7 = __pyx_t_8;
          __pyx_L16_bool_binop_done:;
          if (__pyx_t_7) {

            /* "flitter/render/physics.pyx":541
 *                         velocity = state.get_item(particle.velocity_state_key)
 *                         if velocity.length == zero.length and velocity.numbers != NULL:
 *                             particle.velocity = velocity.copy()             # <<<<<<<<<<<<<<
 *                         else:
 *                             particle.velocity = child.get_fvec('velocity', zero.length, zero).copy()
 */
            __pyx_t_6 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_velocity->__pyx_vtab)->copy(__pyx_v_velocity, 0)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 541, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_GIVEREF(__pyx_t_6);
            __Pyx_GOTREF((PyObject *)__pyx_v_particle->velocity);
            __Pyx_DECREF((PyObject *)__pyx_v_particle->velocity);
            __pyx_v_particle->velocity = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_6);
            __pyx_t_6 = 0;

            /* "flitter/render/physics.pyx":540
 *                             particle.position = child.get_fvec('position', zero.length, zero).copy()
 *                         velocity = state.get_item(particle.velocity_state_key)
 *                         if velocity.length == zero.length and velocity.numbers != NULL:             # <<<<<<<<<<<<<<
 *                             particle.velocity = velocity.copy()
 *                         else:
 */
            goto __pyx_L15;
          }

          /* "flitter/render/physics.pyx":543
 *                             particle.velocity = velocity.copy()
 *                         else:
 *                             particle.velocity = child.get_fvec('velocity', zero.length, zero).copy()             # <<<<<<<<<<<<<<
 *                     else:
 *                         particle.position = child.get_fvec('position', zero.length, zero).copy()
 */
          /*else*/ {
            __pyx_t_6 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Node *)__pyx_v_child->__pyx_vtab)->get_fvec(__pyx_v_child, __pyx_n_u_velocity, __pyx_v_zero->length, __pyx_v_zero)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 543, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_6);
            __pyx_t_9 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_6)->__pyx_vtab)->copy(((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_6), 0)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 543, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_9);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_GIVEREF(__pyx_t_9);
            __Pyx_GOTREF((PyObject *)__pyx_v_particle->velocity);
            __Pyx_DECREF((PyObject *)__pyx_v_particle->velocity);
            __pyx_v_particle->velocity = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_9);
            __pyx_t_9 = 0;
          }
          __pyx_L15:;

          /* "flitter/render/physics.pyx":533
 *                     particle.velocity_state_key = particle.position_state_key.concat(VELOCITY).intern()
 *                     particle.non_anchor = child.kind is 'particle'
 *                     if particle.non_anchor:             # <<<<<<<<<<<<<<
 *                         position = state.get_item(particle.position_state_key)
 *                         if position.length == zero.length and position.numbers != NULL:
 */
          goto __pyx_L11;
        }

        /* "flitter/render/physics.pyx":545
 *                             particle.velocity = child.get_fvec('velocity', zero.length, zero).copy()
 *                     else:
 *                         particle.position = child.get_fvec('position', zero.length, zero).copy()             # <<<<<<<<<<<<<<
 *                         particle.velocity = child.get_fvec('velocity', zero.length, zero).copy()
 *                     particle.initial_force = child.get_fvec('force', zero.length, zero)
 */
        /*else*/ {
          __pyx_t_9 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Node *)__pyx_v_child->__pyx_vtab)->get_fvec(__pyx_v_child, __pyx_n_u_position, __pyx_v_zero->length, __pyx_v_zero)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 545, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          __pyx_t_6 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_9)->__pyx_vtab)->copy(((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_9), 0)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 545, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_GIVEREF(__pyx_t_6);
          __Pyx_GOTREF((PyObject *)__pyx_v_particle->position);
          __Pyx_DECREF((PyObject *)__pyx_v_particle->position);
          __pyx_v_particle->position = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_6);
          __pyx_t_6 = 0;

          /* "flitter/render/physics.pyx":546
 *                     else:
 *                         particle.position = child.get_fvec('position', zero.length, zero).copy()
 *                         particle.velocity = child.get_fvec('velocity', zero.length, zero).copy()             # <<<<<<<<<<<<<<
 *                     particle.initial_force = child.get_fvec('force', zero.length, zero)
 *                     particle.force = particle.initial_force.copy()
 */
          __pyx_t_6 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Node *)__pyx_v_child->__pyx_vtab)->get_fvec(__pyx_v_child, __pyx_n_u_velocity, __pyx_v_zero->length, __pyx_v_zero)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 546, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_9 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_6)->__pyx_vtab)->copy(((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_6), 0)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 546, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_GIVEREF(__pyx_t_9);
          __Pyx_GOTREF((PyObject *)__pyx_v_particle->velocity);
          __Pyx_DECREF((PyObject *)__pyx_v_particle->velocity);
          __pyx_v_particle->velocity = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_9);
          __pyx_t_9 = 0;
        }
        __pyx_L11:;

        /* "flitter/render/physics.pyx":547
 *                         particle.position = child.get_fvec('position', zero.length, zero).copy()
 *                         particle.velocity = child.get_fvec('velocity', zero.length, zero).copy()
 *                     particle.initial_force = child.get_fvec('force', zero.length, zero)             # <<<<<<<<<<<<<<
 *                     particle.force = particle.initial_force.copy()
 *                     particle.acceleration = zero.copy()
 */
        __pyx_t_9 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Node *)__pyx_v_child->__pyx_vtab)->get_fvec(__pyx_v_child, __pyx_n_u_force, __pyx_v_zero->length, __pyx_v_zero)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 547, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_GIVEREF(__pyx_t_9);
        __Pyx_GOTREF((PyObject *)__pyx_v_particle->initial_force);
        __Pyx_DECREF((PyObject *)__pyx_v_particle->initial_force);
        __pyx_v_particle->initial_force = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_9);
        __pyx_t_9 = 0;

        /* "flitter/render/physics.pyx":548
 *                         particle.velocity = child.get_fvec('velocity', zero.length, zero).copy()
 *                     particle.initial_force = child.get_fvec('force', zero.length, zero)
 *                     particle.force = particle.initial_force.copy()             # <<<<<<<<<<<<<<
 *                     particle.acceleration = zero.copy()
 *                     particle.radius = max(0, child.get_float('radius', 0))
 */
        __pyx_t_9 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_particle->initial_force->__pyx_vtab)->copy(__pyx_v_particle->initial_force, 0)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 548, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_GIVEREF(__pyx_t_9);
        __Pyx_GOTREF((PyObject *)__pyx_v_particle->force);
        __Pyx_DECREF((PyObject *)__pyx_v_particle->force);
        __pyx_v_particle->force = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_9);
        __pyx_t_9 = 0;

        /* "flitter/render/physics.pyx":549
 *                     particle.initial_force = child.get_fvec('force', zero.length, zero)
 *                     particle.force = particle.initial_force.copy()
 *                     particle.acceleration = zero.copy()             # <<<<<<<<<<<<<<
 *                     particle.radius = max(0, child.get_float('radius', 0))
 *                     particle.mass = max(0, child.get_float('mass', 1))
 */
        __pyx_t_9 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_zero->__pyx_vtab)->copy(__pyx_v_zero, 0)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 549, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_GIVEREF(__pyx_t_9);
        __Pyx_GOTREF((PyObject *)__pyx_v_particle->acceleration);
        __Pyx_DECREF((PyObject *)__pyx_v_particle->acceleration);
        __pyx_v_particle->acceleration = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_9);
        __pyx_t_9 = 0;

        /* "flitter/render/physics.pyx":550
 *                     particle.force = particle.initial_force.copy()
 *                     particle.acceleration = zero.copy()
 *                     particle.radius = max(0, child.get_float('radius', 0))             # <<<<<<<<<<<<<<
 *                     particle.mass = max(0, child.get_float('mass', 1))
 *                     particle.charge = child.get_float('charge', 0)
 */
        __pyx_t_10 = ((struct __pyx_vtabstruct_7flitter_5model_Node *)__pyx_v_child->__pyx_vtab)->get_float(__pyx_v_child, __pyx_n_u_radius, 0.0); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 550, __pyx_L1_error)
        __pyx_t_11 = 0;
        __pyx_t_7 = (__pyx_t_10 > __pyx_t_11);
        if (__pyx_t_7) {
          __pyx_t_12 = __pyx_t_10;
        } else {
          __pyx_t_12 = __pyx_t_11;
        }
        __pyx_v_particle->radius = __pyx_t_12;

        /* "flitter/render/physics.pyx":551
 *                     particle.acceleration = zero.copy()
 *                     particle.radius = max(0, child.get_float('radius', 0))
 *                     particle.mass = max(0, child.get_float('mass', 1))             # <<<<<<<<<<<<<<
 *                     particle.charge = child.get_float('charge', 0)
 *                     particle.ease = child.get_float('ease', 0)
 */
        __pyx_t_12 = ((struct __pyx_vtabstruct_7flitter_5model_Node *)__pyx_v_child->__pyx_vtab)->get_float(__pyx_v_child, __pyx_n_u_mass, 1.0); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 551, __pyx_L1_error)
        __pyx_t_11 = 0;
        __pyx_t_7 = (__pyx_t_12 > __pyx_t_11);
        if (__pyx_t_7) {
          __pyx_t_10 = __pyx_t_12;
        } else {
          __pyx_t_10 = __pyx_t_11;
        }
        __pyx_v_particle->mass = __pyx_t_10;

        /* "flitter/render/physics.pyx":552
 *                     particle.radius = max(0, child.get_float('radius', 0))
 *                     particle.mass = max(0, child.get_float('mass', 1))
 *                     particle.charge = child.get_float('charge', 0)             # <<<<<<<<<<<<<<
 *                     particle.ease = child.get_float('ease', 0)
 *                     particles_by_id[particle.id] = particle
 */
        __pyx_t_10 = ((struct __pyx_vtabstruct_7flitter_5model_Node *)__pyx_v_child->__pyx_vtab)->get_float(__pyx_v_child, __pyx_n_u_charge, 0.0); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 552, __pyx_L1_error)
        __pyx_v_particle->charge = __pyx_t_10;

        /* "flitter/render/physics.pyx":553
 *                     particle.mass = max(0, child.get_float('mass', 1))
 *                     particle.charge = child.get_float('charge', 0)
 *                     particle.ease = child.get_float('ease', 0)             # <<<<<<<<<<<<<<
 *                     particles_by_id[particle.id] = particle
 *                     group.particles.append(particle)
 */
        __pyx_t_10 = ((struct __pyx_vtabstruct_7flitter_5model_Node *)__pyx_v_child->__pyx_vtab)->get_float(__pyx_v_child, __pyx_n_u_ease, 0.0); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 553, __pyx_L1_error)
        __pyx_v_particle->ease = __pyx_t_10;

        /* "flitter/render/physics.pyx":554
 *                     particle.charge = child.get_float('charge', 0)
 *                     particle.ease = child.get_float('ease', 0)
 *                     particles_by_id[particle.id] = particle             # <<<<<<<<<<<<<<
 *                     group.particles.append(particle)
 *                     new_state_keys.add(particle.position_state_key)
 */
        if (unlikely(__pyx_v_particles_by_id == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(1, 554, __pyx_L1_error)
        }
        if (unlikely((PyDict_SetItem(__pyx_v_particles_by_id, ((PyObject *)__pyx_v_particle->id), ((PyObject *)__pyx_v_particle)) < 0))) __PYX_ERR(1, 554, __pyx_L1_error)

        /* "flitter/render/physics.pyx":555
 *                     particle.ease = child.get_float('ease', 0)
 *                     particles_by_id[particle.id] = particle
 *                     group.particles.append(particle)             # <<<<<<<<<<<<<<
 *                     new_state_keys.add(particle.position_state_key)
 *                     old_state_keys.discard(particle.position_state_key)
 */
        if (unlikely(__pyx_v_group->particles == Py_None)) {
          PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
          __PYX_ERR(1, 555, __pyx_L1_error)
        }
        __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_group->particles, ((PyObject *)__pyx_v_particle)); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(1, 555, __pyx_L1_error)

        /* "flitter/render/physics.pyx":556
 *                     particles_by_id[particle.id] = particle
 *                     group.particles.append(particle)
 *                     new_state_keys.add(particle.position_state_key)             # <<<<<<<<<<<<<<
 *                     old_state_keys.discard(particle.position_state_key)
 *                     new_state_keys.add(particle.velocity_state_key)
 */
        if (unlikely(__pyx_v_new_state_keys == Py_None)) {
          PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "add");
          __PYX_ERR(1, 556, __pyx_L1_error)
        }
        __pyx_t_9 = ((PyObject *)__pyx_v_particle->position_state_key);
        __Pyx_INCREF(__pyx_t_9);
        __pyx_t_13 = PySet_Add(__pyx_v_new_state_keys, __pyx_t_9); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(1, 556, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "flitter/render/physics.pyx":557
 *                     group.particles.append(particle)
 *                     new_state_keys.add(particle.position_state_key)
 *                     old_state_keys.discard(particle.position_state_key)             # <<<<<<<<<<<<<<
 *                     new_state_keys.add(particle.velocity_state_key)
 *                     old_state_keys.discard(particle.velocity_state_key)
 */
        if (unlikely(__pyx_v_old_state_keys == Py_None)) {
          PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "discard");
          __PYX_ERR(1, 557, __pyx_L1_error)
        }
        __pyx_t_9 = ((PyObject *)__pyx_v_particle->position_state_key);
        __Pyx_INCREF(__pyx_t_9);
        __pyx_t_13 = __Pyx_PySet_Discard(__pyx_v_old_state_keys, __pyx_t_9); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(1, 557, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "flitter/render/physics.pyx":558
 *                     new_state_keys.add(particle.position_state_key)
 *                     old_state_keys.discard(particle.position_state_key)
 *                     new_state_keys.add(particle.velocity_state_key)             # <<<<<<<<<<<<<<
 *                     old_state_keys.discard(particle.velocity_state_key)
 *             elif child.kind is 'group':
 */
        if (unlikely(__pyx_v_new_state_keys == Py_None)) {
          PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "add");
          __PYX_ERR(1, 558, __pyx_L1_error)
        }
        __pyx_t_9 = ((PyObject *)__pyx_v_particle->velocity_state_key);
        __Pyx_INCREF(__pyx_t_9);
        __pyx_t_13 = PySet_Add(__pyx_v_new_state_keys, __pyx_t_9); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(1, 558, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "flitter/render/physics.pyx":559
 *                     old_state_keys.discard(particle.position_state_key)
 *                     new_state_keys.add(particle.velocity_state_key)
 *                     old_state_keys.discard(particle.velocity_state_key)             # <<<<<<<<<<<<<<
 *             elif child.kind is 'group':
 *                 child_group = PhysicsGroup.__new__(PhysicsGroup)
 */
        if (unlikely(__pyx_v_old_state_keys == Py_None)) {
          PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "discard");
          __PYX_ERR(1, 559, __pyx_L1_error)
        }
        __pyx_t_9 = ((PyObject *)__pyx_v_particle->velocity_state_key);
        __Pyx_INCREF(__pyx_t_9);
        __pyx_t_13 = __Pyx_PySet_Discard(__pyx_v_old_state_keys, __pyx_t_9); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(1, 559, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "flitter/render/physics.pyx":527
 *         for child in node._children:
 *             if child.kind is 'particle' or child.kind is 'anchor':
 *                 if (particle_id := child.get_vec('id', None)) is not None:             # <<<<<<<<<<<<<<
 *                     particle = Particle.__new__(Particle)
 *                     particle.id = particle_id.intern()
 */
      }

      /* "flitter/render/physics.pyx":526
 *         cdef double fixed
 *         for child in node._children:
 *             if child.kind is 'particle' or child.kind is 'anchor':             # <<<<<<<<<<<<<<
 *                 if (particle_id := child.get_vec('id', None)) is not None:
 *                     particle = Particle.__new__(Particle)
 */
      goto __pyx_L7;
    }

    /* "flitter/render/physics.pyx":560
 *                     new_state_keys.add(particle.velocity_state_key)
 *                     old_state_keys.discard(particle.velocity_state_key)
 *             elif child.kind is 'group':             # <<<<<<<<<<<<<<
 *                 child_group = PhysicsGroup.__new__(PhysicsGroup)
 *                 child_group.parent = group
 */
    __pyx_t_7 = (__pyx_v_child->kind == __pyx_n_u_group);
    if (__pyx_t_7) {

      /* "flitter/render/physics.pyx":561
 *                     old_state_keys.discard(particle.velocity_state_key)
 *             elif child.kind is 'group':
 *                 child_group = PhysicsGroup.__new__(PhysicsGroup)             # <<<<<<<<<<<<<<
 *                 child_group.parent = group
 *                 groups.append(child_group)
 */
      __pyx_t_9 = ((PyObject *)__pyx_tp_new_7flitter_6render_7physics_PhysicsGroup(((PyTypeObject *)__pyx_ptype_7flitter_6render_7physics_PhysicsGroup), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 561, __pyx_L1_error)
      __Pyx_GOTREF((PyObject *)__pyx_t_9);
      __Pyx_XDECREF_SET(__pyx_v_child_group, ((struct __pyx_obj_7flitter_6render_7physics_PhysicsGroup *)__pyx_t_9));
      __pyx_t_9 = 0;

      /* "flitter/render/physics.pyx":562
 *             elif child.kind is 'group':
 *                 child_group = PhysicsGroup.__new__(PhysicsGroup)
 *                 child_group.parent = group             # <<<<<<<<<<<<<<
 *                 groups.append(child_group)
 *                 self.collect(groups, child_group, child, particles_by_id, old_state_keys, new_state_keys, state, state_prefix, clock, seed)
 */
      __Pyx_INCREF((PyObject *)__pyx_v_group);
      __Pyx_GIVEREF((PyObject *)__pyx_v_group);
      __Pyx_GOTREF((PyObject *)__pyx_v_child_group->parent);
      __Pyx_DECREF((PyObject *)__pyx_v_child_group->parent);
      __pyx_v_child_group->parent = __pyx_v_group;

      /* "flitter/render/physics.pyx":563
 *                 child_group = PhysicsGroup.__new__(PhysicsGroup)
 *                 child_group.parent = group
 *                 groups.append(child_group)             # <<<<<<<<<<<<<<
 *                 self.collect(groups, child_group, child, particles_by_id, old_state_keys, new_state_keys, state, state_prefix, clock, seed)
 *             else:
 */
      if (unlikely(__pyx_v_groups == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
        __PYX_ERR(1, 563, __pyx_L1_error)
      }
      __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_groups, ((PyObject *)__pyx_v_child_group)); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(1, 563, __pyx_L1_error)

      /* "flitter/render/physics.pyx":564
 *                 child_group.parent = group
 *                 groups.append(child_group)
 *                 self.collect(groups, child_group, child, particles_by_id, old_state_keys, new_state_keys, state, state_prefix, clock, seed)             # <<<<<<<<<<<<<<
 *             else:
 *                 strength = child.get_float('strength', 1)
 */
      ((struct __pyx_vtabstruct_7flitter_6render_7physics_PhysicsSystem *)__pyx_v_self->__pyx_vtab)->collect(__pyx_v_self, __pyx_v_groups, __pyx_v_child_group, __pyx_v_child, __pyx_v_particles_by_id, __pyx_v_old_state_keys, __pyx_v_new_state_keys, __pyx_v_state, __pyx_v_state_prefix, __pyx_v_clock, __pyx_v_seed); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 564, __pyx_L1_error)

      /* "flitter/render/physics.pyx":560
 *                     new_state_keys.add(particle.velocity_state_key)
 *                     old_state_keys.discard(particle.velocity_state_key)
 *             elif child.kind is 'group':             # <<<<<<<<<<<<<<
 *                 child_group = PhysicsGroup.__new__(PhysicsGroup)
 *                 child_group.parent = group
 */
      goto __pyx_L7;
    }

    /* "flitter/render/physics.pyx":566
 *                 self.collect(groups, child_group, child, particles_by_id, old_state_keys, new_state_keys, state, state_prefix, clock, seed)
 *             else:
 *                 strength = child.get_float('strength', 1)             # <<<<<<<<<<<<<<
 *                 ease = child.get_float('ease', 0)
 *                 if ease > 0 and ease < clock:
 */
    /*else*/ {
      __pyx_t_10 = ((struct __pyx_vtabstruct_7flitter_5model_Node *)__pyx_v_child->__pyx_vtab)->get_float(__pyx_v_child, __pyx_n_u_strength, 1.0); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 566, __pyx_L1_error)
      __pyx_v_strength = __pyx_t_10;

      /* "flitter/render/physics.pyx":567
 *             else:
 *                 strength = child.get_float('strength', 1)
 *                 ease = child.get_float('ease', 0)             # <<<<<<<<<<<<<<
 *                 if ease > 0 and ease < clock:
 *                     strength *= clock/ease
 */
      __pyx_t_10 = ((struct __pyx_vtabstruct_7flitter_5model_Node *)__pyx_v_child->__pyx_vtab)->get_float(__pyx_v_child, __pyx_n_u_ease, 0.0); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 567, __pyx_L1_error)
      __pyx_v_ease = __pyx_t_10;

      /* "flitter/render/physics.pyx":568
 *                 strength = child.get_float('strength', 1)
 *                 ease = child.get_float('ease', 0)
 *                 if ease > 0 and ease < clock:             # <<<<<<<<<<<<<<
 *                     strength *= clock/ease
 *                 if child.kind is 'distance' and (from_particle_id := child.get_vec('from', None)) is not None \
 */
      __pyx_t_8 = (__pyx_v_ease > 0.0);
      if (__pyx_t_8) {
      } else {
        __pyx_t_7 = __pyx_t_8;
        goto __pyx_L19_bool_binop_done;
      }
      __pyx_t_8 = (__pyx_v_ease < __pyx_v_clock);
      __pyx_t_7 = __pyx_t_8;
      __pyx_L19_bool_binop_done:;
      if (__pyx_t_7) {

        /* "flitter/render/physics.pyx":569
 *                 ease = child.get_float('ease', 0)
 *                 if ease > 0 and ease < clock:
 *                     strength *= clock/ease             # <<<<<<<<<<<<<<
 *                 if child.kind is 'distance' and (from_particle_id := child.get_vec('from', None)) is not None \
 *                         and (to_particle_id := child.get_vec('to', None)) is not None:
 */
        __pyx_v_strength = (__pyx_v_strength * (__pyx_v_clock / __pyx_v_ease));

        /* "flitter/render/physics.pyx":568
 *                 strength = child.get_float('strength', 1)
 *                 ease = child.get_float('ease', 0)
 *                 if ease > 0 and ease < clock:             # <<<<<<<<<<<<<<
 *                     strength *= clock/ease
 *                 if child.kind is 'distance' and (from_particle_id := child.get_vec('from', None)) is not None \
 */
      }

      /* "flitter/render/physics.pyx":570
 *                 if ease > 0 and ease < clock:
 *                     strength *= clock/ease
 *                 if child.kind is 'distance' and (from_particle_id := child.get_vec('from', None)) is not None \             # <<<<<<<<<<<<<<
 *                         and (to_particle_id := child.get_vec('to', None)) is not None:
 *                     distance_force = DistanceForceApplier.__new__(DistanceForceApplier)
 */
      __pyx_t_8 = (__pyx_v_child->kind == __pyx_n_u_distance);
      if (__pyx_t_8) {
      } else {
        __pyx_t_7 = __pyx_t_8;
        goto __pyx_L22_bool_binop_done;
      }

      /* "flitter/render/physics.pyx":571
 *                     strength *= clock/ease
 *                 if child.kind is 'distance' and (from_particle_id := child.get_vec('from', None)) is not None \
 *                         and (to_particle_id := child.get_vec('to', None)) is not None:             # <<<<<<<<<<<<<<
 *                     distance_force = DistanceForceApplier.__new__(DistanceForceApplier)
 *                     distance_force.strength = strength
 */
      __pyx_t_9 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Node *)__pyx_v_child->__pyx_vtab)->get_vec(__pyx_v_child, __pyx_n_u_from, ((struct __pyx_obj_7flitter_5model_Vector *)Py_None))); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 570, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);

      /* "flitter/render/physics.pyx":570
 *                 if ease > 0 and ease < clock:
 *                     strength *= clock/ease
 *                 if child.kind is 'distance' and (from_particle_id := child.get_vec('from', None)) is not None \             # <<<<<<<<<<<<<<
 *                         and (to_particle_id := child.get_vec('to', None)) is not None:
 *                     distance_force = DistanceForceApplier.__new__(DistanceForceApplier)
 */
      __Pyx_XDECREF_SET(__pyx_v_from_particle_id, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_9));
      __Pyx_INCREF(__pyx_t_9);
      __pyx_t_8 = (((PyObject *)__pyx_t_9) != Py_None);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (__pyx_t_8) {
      } else {
        __pyx_t_7 = __pyx_t_8;
        goto __pyx_L22_bool_binop_done;
      }

      /* "flitter/render/physics.pyx":571
 *                     strength *= clock/ease
 *                 if child.kind is 'distance' and (from_particle_id := child.get_vec('from', None)) is not None \
 *                         and (to_particle_id := child.get_vec('to', None)) is not None:             # <<<<<<<<<<<<<<
 *                     distance_force = DistanceForceApplier.__new__(DistanceForceApplier)
 *                     distance_force.strength = strength
 */
      __pyx_t_9 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Node *)__pyx_v_child->__pyx_vtab)->get_vec(__pyx_v_child, __pyx_n_u_to, ((struct __pyx_obj_7flitter_5model_Vector *)Py_None))); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 571, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_XDECREF_SET(__pyx_v_to_particle_id, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_9));
      __Pyx_INCREF(__pyx_t_9);
      __pyx_t_8 = (((PyObject *)__pyx_t_9) != Py_None);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_7 = __pyx_t_8;
      __pyx_L22_bool_binop_done:;

      /* "flitter/render/physics.pyx":570
 *                 if ease > 0 and ease < clock:
 *                     strength *= clock/ease
 *                 if child.kind is 'distance' and (from_particle_id := child.get_vec('from', None)) is not None \             # <<<<<<<<<<<<<<
 *                         and (to_particle_id := child.get_vec('to', None)) is not None:
 *                     distance_force = DistanceForceApplier.__new__(DistanceForceApplier)
 */
      if (__pyx_t_7) {

        /* "flitter/render/physics.pyx":572
 *                 if child.kind is 'distance' and (from_particle_id := child.get_vec('from', None)) is not None \
 *                         and (to_particle_id := child.get_vec('to', None)) is not None:
 *                     distance_force = DistanceForceApplier.__new__(DistanceForceApplier)             # <<<<<<<<<<<<<<
 *                     distance_force.strength = strength
 *                     distance_force.from_particle_id = from_particle_id.intern()
 */
        __pyx_t_9 = ((PyObject *)__pyx_tp_new_7flitter_6render_7physics_DistanceForceApplier(((PyTypeObject *)__pyx_ptype_7flitter_6render_7physics_DistanceForceApplier), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 572, __pyx_L1_error)
        __Pyx_GOTREF((PyObject *)__pyx_t_9);
        __Pyx_XDECREF_SET(__pyx_v_distance_force, ((struct __pyx_obj_7flitter_6render_7physics_DistanceForceApplier *)__pyx_t_9));
        __pyx_t_9 = 0;

        /* "flitter/render/physics.pyx":573
 *                         and (to_particle_id := child.get_vec('to', None)) is not None:
 *                     distance_force = DistanceForceApplier.__new__(DistanceForceApplier)
 *                     distance_force.strength = strength             # <<<<<<<<<<<<<<
 *                     distance_force.from_particle_id = from_particle_id.intern()
 *                     distance_force.to_particle_id = to_particle_id.intern()
 */
        __pyx_v_distance_force->strength = __pyx_v_strength;

        /* "flitter/render/physics.pyx":574
 *                     distance_force = DistanceForceApplier.__new__(DistanceForceApplier)
 *                     distance_force.strength = strength
 *                     distance_force.from_particle_id = from_particle_id.intern()             # <<<<<<<<<<<<<<
 *                     distance_force.to_particle_id = to_particle_id.intern()
 *                     distance_force.power = max(0, child.get_float('power', 1))
 */
        __pyx_t_9 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_from_particle_id->__pyx_vtab)->intern(__pyx_v_from_particle_id, 0)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 574, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_GIVEREF(__pyx_t_9);
        __Pyx_GOTREF((PyObject *)__pyx_v_distance_force->from_particle_id);
        __Pyx_DECREF((PyObject *)__pyx_v_distance_force->from_particle_id);
        __pyx_v_distance_force->from_particle_id = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_9);
        __pyx_t_9 = 0;

        /* "flitter/render/physics.pyx":575
 *                     distance_force.strength = strength
 *                     distance_force.from_particle_id = from_particle_id.intern()
 *                     distance_force.to_particle_id = to_particle_id.intern()             # <<<<<<<<<<<<<<
 *                     distance_force.power = max(0, child.get_float('power', 1))
 *                     if (fixed := child.get_float('fixed', 0)) != 0:
 */
        __pyx_t_9 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_to_particle_id->__pyx_vtab)->intern(__pyx_v_to_particle_id, 0)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 575, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_GIVEREF(__pyx_t_9);
        __Pyx_GOTREF((PyObject *)__pyx_v_distance_force->to_particle_id);
        __Pyx_DECREF((PyObject *)__pyx_v_distance_force->to_particle_id);
        __pyx_v_distance_force->to_particle_id = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_9);
        __pyx_t_9 = 0;

        /* "flitter/render/physics.pyx":576
 *                     distance_force.from_particle_id = from_particle_id.intern()
 *                     distance_force.to_particle_id = to_particle_id.intern()
 *                     distance_force.power = max(0, child.get_float('power', 1))             # <<<<<<<<<<<<<<
 *                     if (fixed := child.get_float('fixed', 0)) != 0:
 *                         distance_force.minimum = fixed
 */
        __pyx_t_10 = ((struct __pyx_vtabstruct_7flitter_5model_Node *)__pyx_v_child->__pyx_vtab)->get_float(__pyx_v_child, __pyx_n_u_power, 1.0); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 576, __pyx_L1_error)
        __pyx_t_11 = 0;
        __pyx_t_7 = (__pyx_t_10 > __pyx_t_11);
        if (__pyx_t_7) {
          __pyx_t_12 = __pyx_t_10;
        } else {
          __pyx_t_12 = __pyx_t_11;
        }
        __pyx_v_distance_force->power = __pyx_t_12;

        /* "flitter/render/physics.pyx":577
 *                     distance_force.to_particle_id = to_particle_id.intern()
 *                     distance_force.power = max(0, child.get_float('power', 1))
 *                     if (fixed := child.get_float('fixed', 0)) != 0:             # <<<<<<<<<<<<<<
 *                         distance_force.minimum = fixed
 *                         distance_force.maximum = fixed
 */
        __pyx_t_12 = ((struct __pyx_vtabstruct_7flitter_5model_Node *)__pyx_v_child->__pyx_vtab)->get_float(__pyx_v_child, __pyx_n_u_fixed, 0.0); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 577, __pyx_L1_error)
        __pyx_v_fixed = __pyx_t_12;
        __pyx_t_7 = (__pyx_t_12 != 0.0);
        if (__pyx_t_7) {

          /* "flitter/render/physics.pyx":578
 *                     distance_force.power = max(0, child.get_float('power', 1))
 *                     if (fixed := child.get_float('fixed', 0)) != 0:
 *                         distance_force.minimum = fixed             # <<<<<<<<<<<<<<
 *                         distance_force.maximum = fixed
 *                     else:
 */
          __pyx_v_distance_force->minimum = __pyx_v_fixed;

          /* "flitter/render/physics.pyx":579
 *                     if (fixed := child.get_float('fixed', 0)) != 0:
 *                         distance_force.minimum = fixed
 *                         distance_force.maximum = fixed             # <<<<<<<<<<<<<<
 *                     else:
 *                         distance_force.minimum = child.get_float('minimum', child.get_float('min', 0))
 */
          __pyx_v_distance_force->maximum = __pyx_v_fixed;

          /* "flitter/render/physics.pyx":577
 *                     distance_force.to_particle_id = to_particle_id.intern()
 *                     distance_force.power = max(0, child.get_float('power', 1))
 *                     if (fixed := child.get_float('fixed', 0)) != 0:             # <<<<<<<<<<<<<<
 *                         distance_force.minimum = fixed
 *                         distance_force.maximum = fixed
 */
          goto __pyx_L25;
        }

        /* "flitter/render/physics.pyx":581
 *                         distance_force.maximum = fixed
 *                     else:
 *                         distance_force.minimum = child.get_float('minimum', child.get_float('min', 0))             # <<<<<<<<<<<<<<
 *                         distance_force.maximum = child.get_float('maximum', child.get_float('max', 0))
 *                     group.distance_forces.append(distance_force)
 */
        /*else*/ {
          __pyx_t_12 = ((struct __pyx_vtabstruct_7flitter_5model_Node *)__pyx_v_child->__pyx_vtab)->get_float(__pyx_v_child, __pyx_n_u_min, 0.0); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 581, __pyx_L1_error)
          __pyx_t_10 = ((struct __pyx_vtabstruct_7flitter_5model_Node *)__pyx_v_child->__pyx_vtab)->get_float(__pyx_v_child, __pyx_n_u_minimum, __pyx_t_12); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 581, __pyx_L1_error)
          __pyx_v_distance_force->minimum = __pyx_t_10;

          /* "flitter/render/physics.pyx":582
 *                     else:
 *                         distance_force.minimum = child.get_float('minimum', child.get_float('min', 0))
 *                         distance_force.maximum = child.get_float('maximum', child.get_float('max', 0))             # <<<<<<<<<<<<<<
 *                     group.distance_forces.append(distance_force)
 *                 elif child.kind is 'drag':
 */
          __pyx_t_10 = ((struct __pyx_vtabstruct_7flitter_5model_Node *)__pyx_v_child->__pyx_vtab)->get_float(__pyx_v_child, __pyx_n_u_max, 0.0); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 582, __pyx_L1_error)
          __pyx_t_12 = ((struct __pyx_vtabstruct_7flitter_5model_Node *)__pyx_v_child->__pyx_vtab)->get_float(__pyx_v_child, __pyx_n_u_maximum, __pyx_t_10); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 582, __pyx_L1_error)
          __pyx_v_distance_force->maximum = __pyx_t_12;
        }
        __pyx_L25:;

        /* "flitter/render/physics.pyx":583
 *                         distance_force.minimum = child.get_float('minimum', child.get_float('min', 0))
 *                         distance_force.maximum = child.get_float('maximum', child.get_float('max', 0))
 *                     group.distance_forces.append(distance_force)             # <<<<<<<<<<<<<<
 *                 elif child.kind is 'drag':
 *                     group.particle_forces.append(DragForceApplier.__new__(DragForceApplier, child, strength, zero))
 */
        if (unlikely(__pyx_v_group->distance_forces == Py_None)) {
          PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
          __PYX_ERR(1, 583, __pyx_L1_error)
        }
        __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_group->distance_forces, ((PyObject *)__pyx_v_distance_force)); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(1, 583, __pyx_L1_error)

        /* "flitter/render/physics.pyx":570
 *                 if ease > 0 and ease < clock:
 *                     strength *= clock/ease
 *                 if child.kind is 'distance' and (from_particle_id := child.get_vec('from', None)) is not None \             # <<<<<<<<<<<<<<
 *                         and (to_particle_id := child.get_vec('to', None)) is not None:
 *                     distance_force = DistanceForceApplier.__new__(DistanceForceApplier)
 */
        goto __pyx_L21;
      }

      /* "flitter/render/physics.pyx":584
 *                         distance_force.maximum = child.get_float('maximum', child.get_float('max', 0))
 *                     group.distance_forces.append(distance_force)
 *                 elif child.kind is 'drag':             # <<<<<<<<<<<<<<
 *                     group.particle_forces.append(DragForceApplier.__new__(DragForceApplier, child, strength, zero))
 *                 elif child.kind is 'buoyancy':
 */
      __pyx_t_7 = (__pyx_v_child->kind == __pyx_n_u_drag);
      if (__pyx_t_7) {

        /* "flitter/render/physics.pyx":585
 *                     group.distance_forces.append(distance_force)
 *                 elif child.kind is 'drag':
 *                     group.particle_forces.append(DragForceApplier.__new__(DragForceApplier, child, strength, zero))             # <<<<<<<<<<<<<<
 *                 elif child.kind is 'buoyancy':
 *                     group.particle_forces.append(BuoyancyForceApplier.__new__(BuoyancyForceApplier, child, strength, zero))
 */
        if (unlikely(__pyx_v_group->particle_forces == Py_None)) {
          PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
          __PYX_ERR(1, 585, __pyx_L1_error)
        }
        __pyx_t_9 = PyFloat_FromDouble(__pyx_v_strength); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 585, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 585, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_INCREF((PyObject *)__pyx_v_child);
        __Pyx_GIVEREF((PyObject *)__pyx_v_child);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, ((PyObject *)__pyx_v_child))) __PYX_ERR(1, 585, __pyx_L1_error);
        __Pyx_GIVEREF(__pyx_t_9);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_9)) __PYX_ERR(1, 585, __pyx_L1_error);
        __Pyx_INCREF((PyObject *)__pyx_v_zero);
        __Pyx_GIVEREF((PyObject *)__pyx_v_zero);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 2, ((PyObject *)__pyx_v_zero))) __PYX_ERR(1, 585, __pyx_L1_error);
        __pyx_t_9 = 0;
        __pyx_t_9 = ((PyObject *)__pyx_tp_new_7flitter_6render_7physics_DragForceApplier(((PyTypeObject *)__pyx_ptype_7flitter_6render_7physics_DragForceApplier), __pyx_t_6, NULL)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 585, __pyx_L1_error)
        __Pyx_GOTREF((PyObject *)__pyx_t_9);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_group->particle_forces, ((PyObject *)__pyx_t_9)); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(1, 585, __pyx_L1_error)
        __Pyx_DECREF((PyObject *)__pyx_t_9); __pyx_t_9 = 0;

        /* "flitter/render/physics.pyx":584
 *                         distance_force.maximum = child.get_float('maximum', child.get_float('max', 0))
 *                     group.distance_forces.append(distance_force)
 *                 elif child.kind is 'drag':             # <<<<<<<<<<<<<<
 *                     group.particle_forces.append(DragForceApplier.__new__(DragForceApplier, child, strength, zero))
 *                 elif child.kind is 'buoyancy':
 */
        goto __pyx_L21;
      }

      /* "flitter/render/physics.pyx":586
 *                 elif child.kind is 'drag':
 *                     group.particle_forces.append(DragForceApplier.__new__(DragForceApplier, child, strength, zero))
 *                 elif child.kind is 'buoyancy':             # <<<<<<<<<<<<<<
 *                     group.particle_forces.append(BuoyancyForceApplier.__new__(BuoyancyForceApplier, child, strength, zero))
 *                 elif child.kind is 'constant':
 */
      __pyx_t_7 = (__pyx_v_child->kind == __pyx_n_u_buoyancy);
      if (__pyx_t_7) {

        /* "flitter/render/physics.pyx":587
 *                     group.particle_forces.append(DragForceApplier.__new__(DragForceApplier, child, strength, zero))
 *                 elif child.kind is 'buoyancy':
 *                     group.particle_forces.append(BuoyancyForceApplier.__new__(BuoyancyForceApplier, child, strength, zero))             # <<<<<<<<<<<<<<
 *                 elif child.kind is 'constant':
 *                     group.particle_forces.append(ConstantForceApplier.__new__(ConstantForceApplier, child, strength, zero))
 */
        if (unlikely(__pyx_v_group->particle_forces == Py_None)) {
          PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
          __PYX_ERR(1, 587, __pyx_L1_error)
        }
        __pyx_t_9 = PyFloat_FromDouble(__pyx_v_strength); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 587, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 587, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_INCREF((PyObject *)__pyx_v_child);
        __Pyx_GIVEREF((PyObject *)__pyx_v_child);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, ((PyObject *)__pyx_v_child))) __PYX_ERR(1, 587, __pyx_L1_error);
        __Pyx_GIVEREF(__pyx_t_9);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_9)) __PYX_ERR(1, 587, __pyx_L1_error);
        __Pyx_INCREF((PyObject *)__pyx_v_zero);
        __Pyx_GIVEREF((PyObject *)__pyx_v_zero);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 2, ((PyObject *)__pyx_v_zero))) __PYX_ERR(1, 587, __pyx_L1_error);
        __pyx_t_9 = 0;
        __pyx_t_9 = ((PyObject *)__pyx_tp_new_7flitter_6render_7physics_BuoyancyForceApplier(((PyTypeObject *)__pyx_ptype_7flitter_6render_7physics_BuoyancyForceApplier), __pyx_t_6, NULL)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 587, __pyx_L1_error)
        __Pyx_GOTREF((PyObject *)__pyx_t_9);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_group->particle_forces, ((PyObject *)__pyx_t_9)); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(1, 587, __pyx_L1_error)
        __Pyx_DECREF((PyObject *)__pyx_t_9); __pyx_t_9 = 0;

        /* "flitter/render/physics.pyx":586
 *                 elif child.kind is 'drag':
 *                     group.particle_forces.append(DragForceApplier.__new__(DragForceApplier, child, strength, zero))
 *                 elif child.kind is 'buoyancy':             # <<<<<<<<<<<<<<
 *                     group.particle_forces.append(BuoyancyForceApplier.__new__(BuoyancyForceApplier, child, strength, zero))
 *                 elif child.kind is 'constant':
 */
        goto __pyx_L21;
      }

      /* "flitter/render/physics.pyx":588
 *                 elif child.kind is 'buoyancy':
 *                     group.particle_forces.append(BuoyancyForceApplier.__new__(BuoyancyForceApplier, child, strength, zero))
 *                 elif child.kind is 'constant':             # <<<<<<<<<<<<<<
 *                     group.particle_forces.append(ConstantForceApplier.__new__(ConstantForceApplier, child, strength, zero))
 *                 elif child.kind is 'random':
 */
      __pyx_t_7 = (__pyx_v_child->kind == __pyx_n_u_constant);
      if (__pyx_t_7) {

        /* "flitter/render/physics.pyx":589
 *                     group.particle_forces.append(BuoyancyForceApplier.__new__(BuoyancyForceApplier, child, strength, zero))
 *                 elif child.kind is 'constant':
 *                     group.particle_forces.append(ConstantForceApplier.__new__(ConstantForceApplier, child, strength, zero))             # <<<<<<<<<<<<<<
 *                 elif child.kind is 'random':
 *                     particle_force = RandomForceApplier.__new__(RandomForceApplier, child, strength, zero)
 */
        if (unlikely(__pyx_v_group->particle_forces == Py_None)) {
          PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
          __PYX_ERR(1, 589, __pyx_L1_error)
        }
        __pyx_t_9 = PyFloat_FromDouble(__pyx_v_strength); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 589, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 589, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_INCREF((PyObject *)__pyx_v_child);
        __Pyx_GIVEREF((PyObject *)__pyx_v_child);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, ((PyObject *)__pyx_v_child))) __PYX_ERR(1, 589, __pyx_L1_error);
        __Pyx_GIVEREF(__pyx_t_9);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_9)) __PYX_ERR(1, 589, __pyx_L1_error);
        __Pyx_INCREF((PyObject *)__pyx_v_zero);
        __Pyx_GIVEREF((PyObject *)__pyx_v_zero);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 2, ((PyObject *)__pyx_v_zero))) __PYX_ERR(1, 589, __pyx_L1_error);
        __pyx_t_9 = 0;
        __pyx_t_9 = ((PyObject *)__pyx_tp_new_7flitter_6render_7physics_ConstantForceApplier(((PyTypeObject *)__pyx_ptype_7flitter_6render_7physics_ConstantForceApplier), __pyx_t_6, NULL)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 589, __pyx_L1_error)
        __Pyx_GOTREF((PyObject *)__pyx_t_9);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_group->particle_forces, ((PyObject *)__pyx_t_9)); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(1, 589, __pyx_L1_error)
        __Pyx_DECREF((PyObject *)__pyx_t_9); __pyx_t_9 = 0;

        /* "flitter/render/physics.pyx":588
 *                 elif child.kind is 'buoyancy':
 *                     group.particle_forces.append(BuoyancyForceApplier.__new__(BuoyancyForceApplier, child, strength, zero))
 *                 elif child.kind is 'constant':             # <<<<<<<<<<<<<<
 *                     group.particle_forces.append(ConstantForceApplier.__new__(ConstantForceApplier, child, strength, zero))
 *                 elif child.kind is 'random':
 */
        goto __pyx_L21;
      }

      /* "flitter/render/physics.pyx":590
 *                 elif child.kind is 'constant':
 *                     group.particle_forces.append(ConstantForceApplier.__new__(ConstantForceApplier, child, strength, zero))
 *                 elif child.kind is 'random':             # <<<<<<<<<<<<<<
 *                     particle_force = RandomForceApplier.__new__(RandomForceApplier, child, strength, zero)
 *                     (<RandomForceApplier>particle_force).random_source = normal(seed)
 */
      __pyx_t_7 = (__pyx_v_child->kind == __pyx_n_u_random);
      if (__pyx_t_7) {

        /* "flitter/render/physics.pyx":591
 *                     group.particle_forces.append(ConstantForceApplier.__new__(ConstantForceApplier, child, strength, zero))
 *                 elif child.kind is 'random':
 *                     particle_force = RandomForceApplier.__new__(RandomForceApplier, child, strength, zero)             # <<<<<<<<<<<<<<
 *                     (<RandomForceApplier>particle_force).random_source = normal(seed)
 *                     seed = seed.concat(false_)
 */
        __pyx_t_9 = PyFloat_FromDouble(__pyx_v_strength); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 591, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 591, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_INCREF((PyObject *)__pyx_v_child);
        __Pyx_GIVEREF((PyObject *)__pyx_v_child);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, ((PyObject *)__pyx_v_child))) __PYX_ERR(1, 591, __pyx_L1_error);
        __Pyx_GIVEREF(__pyx_t_9);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_9)) __PYX_ERR(1, 591, __pyx_L1_error);
        __Pyx_INCREF((PyObject *)__pyx_v_zero);
        __Pyx_GIVEREF((PyObject *)__pyx_v_zero);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 2, ((PyObject *)__pyx_v_zero))) __PYX_ERR(1, 591, __pyx_L1_error);
        __pyx_t_9 = 0;
        __pyx_t_9 = ((PyObject *)__pyx_tp_new_7flitter_6render_7physics_RandomForceApplier(((PyTypeObject *)__pyx_ptype_7flitter_6render_7physics_RandomForceApplier), __pyx_t_6, NULL)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 591, __pyx_L1_error)
        __Pyx_GOTREF((PyObject *)__pyx_t_9);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (!(likely(__Pyx_TypeTest(((PyObject *)__pyx_t_9), __pyx_ptype_7flitter_6render_7physics_ParticleForceApplier)))) __PYX_ERR(1, 591, __pyx_L1_error)
        __Pyx_XDECREF_SET(__pyx_v_particle_force, ((struct __pyx_obj_7flitter_6render_7physics_ParticleForceApplier *)__pyx_t_9));
        __pyx_t_9 = 0;

        /* "flitter/render/physics.pyx":592
 *                 elif child.kind is 'random':
 *                     particle_force = RandomForceApplier.__new__(RandomForceApplier, child, strength, zero)
 *                     (<RandomForceApplier>particle_force).random_source = normal(seed)             # <<<<<<<<<<<<<<
 *                     seed = seed.concat(false_)
 *                     group.particle_forces.append(particle_force)
 */
        __pyx_t_9 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7flitter_8language_9functions_normal), ((PyObject *)__pyx_v_seed)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 592, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_GIVEREF(__pyx_t_9);
        __Pyx_GOTREF((PyObject *)((struct __pyx_obj_7flitter_6render_7physics_RandomForceApplier *)__pyx_v_particle_force)->random_source);
        __Pyx_DECREF((PyObject *)((struct __pyx_obj_7flitter_6render_7physics_RandomForceApplier *)__pyx_v_particle_force)->random_source);
        ((struct __pyx_obj_7flitter_6render_7physics_RandomForceApplier *)__pyx_v_particle_force)->random_source = ((struct __pyx_obj_7flitter_8language_9functions_normal *)__pyx_t_9);
        __pyx_t_9 = 0;

        /* "flitter/render/physics.pyx":593
 *                     particle_force = RandomForceApplier.__new__(RandomForceApplier, child, strength, zero)
 *                     (<RandomForceApplier>particle_force).random_source = normal(seed)
 *                     seed = seed.concat(false_)             # <<<<<<<<<<<<<<
 *                     group.particle_forces.append(particle_force)
 *                 elif child.kind is 'field':
 */
        __pyx_t_9 = ((PyObject *)__pyx_v_7flitter_5model_false_);
        __Pyx_INCREF(__pyx_t_9);
        __pyx_t_6 = ((PyObject *)((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_seed->__pyx_vtab)->concat(__pyx_v_seed, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_9), 0)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 593, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF_SET(__pyx_v_seed, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_6));
        __pyx_t_6 = 0;

        /* "flitter/render/physics.pyx":594
 *                     (<RandomForceApplier>particle_force).random_source = normal(seed)
 *                     seed = seed.concat(false_)
 *                     group.particle_forces.append(particle_force)             # <<<<<<<<<<<<<<
 *                 elif child.kind is 'field':
 *                     group.particle_forces.append(FieldForceApplier.__new__(FieldForceApplier, child, strength, zero))
 */
        if (unlikely(__pyx_v_group->particle_forces == Py_None)) {
          PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
          __PYX_ERR(1, 594, __pyx_L1_error)
        }
        __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_group->particle_forces, ((PyObject *)__pyx_v_particle_force)); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(1, 594, __pyx_L1_error)

        /* "flitter/render/physics.pyx":590
 *                 elif child.kind is 'constant':
 *                     group.particle_forces.append(ConstantForceApplier.__new__(ConstantForceApplier, child, strength, zero))
 *                 elif child.kind is 'random':             # <<<<<<<<<<<<<<
 *                     particle_force = RandomForceApplier.__new__(RandomForceApplier, child, strength, zero)
 *                     (<RandomForceApplier>particle_force).random_source = normal(seed)
 */
        goto __pyx_L21;
      }

      /* "flitter/render/physics.pyx":595
 *                     seed = seed.concat(false_)
 *                     group.particle_forces.append(particle_force)
 *                 elif child.kind is 'field':             # <<<<<<<<<<<<<<
 *                     group.particle_forces.append(FieldForceApplier.__new__(FieldForceApplier, child, strength, zero))
 *                 elif child.kind is 'collision':
 */
      __pyx_t_7 = (__pyx_v_child->kind == __pyx_n_u_field);
      if (__pyx_t_7) {

        /* "flitter/render/physics.pyx":596
 *                     group.particle_forces.append(particle_force)
 *                 elif child.kind is 'field':
 *                     group.particle_forces.append(FieldForceApplier.__new__(FieldForceApplier, child, strength, zero))             # <<<<<<<<<<<<<<
 *                 elif child.kind is 'collision':
 *                     group.matrix_forces.append(CollisionForceApplier.__new__(CollisionForceApplier, child, strength, zero))
 */
        if (unlikely(__pyx_v_group->particle_forces == Py_None)) {
          PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
          __PYX_ERR(1, 596, __pyx_L1_error)
        }
        __pyx_t_6 = PyFloat_FromDouble(__pyx_v_strength); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 596, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 596, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_INCREF((PyObject *)__pyx_v_child);
        __Pyx_GIVEREF((PyObject *)__pyx_v_child);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)__pyx_v_child))) __PYX_ERR(1, 596, __pyx_L1_error);
        __Pyx_GIVEREF(__pyx_t_6);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_6)) __PYX_ERR(1, 596, __pyx_L1_error);
        __Pyx_INCREF((PyObject *)__pyx_v_zero);
        __Pyx_GIVEREF((PyObject *)__pyx_v_zero);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, ((PyObject *)__pyx_v_zero))) __PYX_ERR(1, 596, __pyx_L1_error);
        __pyx_t_6 = 0;
        __pyx_t_6 = ((PyObject *)__pyx_tp_new_7flitter_6render_7physics_FieldForceApplier(((PyTypeObject *)__pyx_ptype_7flitter_6render_7physics_FieldForceApplier), __pyx_t_9, NULL)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 596, __pyx_L1_error)
        __Pyx_GOTREF((PyObject *)__pyx_t_6);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_group->particle_forces, ((PyObject *)__pyx_t_6)); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(1, 596, __pyx_L1_error)
        __Pyx_DECREF((PyObject *)__pyx_t_6); __pyx_t_6 = 0;

        /* "flitter/render/physics.pyx":595
 *                     seed = seed.concat(false_)
 *                     group.particle_forces.append(particle_force)
 *                 elif child.kind is 'field':             # <<<<<<<<<<<<<<
 *                     group.particle_forces.append(FieldForceApplier.__new__(FieldForceApplier, child, strength, zero))
 *                 elif child.kind is 'collision':
 */
        goto __pyx_L21;
      }

      /* "flitter/render/physics.pyx":597
 *                 elif child.kind is 'field':
 *                     group.particle_forces.append(FieldForceApplier.__new__(FieldForceApplier, child, strength, zero))
 *                 elif child.kind is 'collision':             # <<<<<<<<<<<<<<
 *                     group.matrix_forces.append(CollisionForceApplier.__new__(CollisionForceApplier, child, strength, zero))
 *                 elif child.kind is 'gravity':
 */
      __pyx_t_7 = (__pyx_v_child->kind == __pyx_n_u_collision);
      if (__pyx_t_7) {

        /* "flitter/render/physics.pyx":598
 *                     group.particle_forces.append(FieldForceApplier.__new__(FieldForceApplier, child, strength, zero))
 *                 elif child.kind is 'collision':
 *                     group.matrix_forces.append(CollisionForceApplier.__new__(CollisionForceApplier, child, strength, zero))             # <<<<<<<<<<<<<<
 *                 elif child.kind is 'gravity':
 *                     group.matrix_forces.append(GravityForceApplier.__new__(GravityForceApplier, child, strength, zero))
 */
        if (unlikely(__pyx_v_group->matrix_forces == Py_None)) {
          PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
          __PYX_ERR(1, 598, __pyx_L1_error)
        }
        __pyx_t_6 = PyFloat_FromDouble(__pyx_v_strength); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 598, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 598, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_INCREF((PyObject *)__pyx_v_child);
        __Pyx_GIVEREF((PyObject *)__pyx_v_child);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)__pyx_v_child))) __PYX_ERR(1, 598, __pyx_L1_error);
        __Pyx_GIVEREF(__pyx_t_6);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_6)) __PYX_ERR(1, 598, __pyx_L1_error);
        __Pyx_INCREF((PyObject *)__pyx_v_zero);
        __Pyx_GIVEREF((PyObject *)__pyx_v_zero);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, ((PyObject *)__pyx_v_zero))) __PYX_ERR(1, 598, __pyx_L1_error);
        __pyx_t_6 = 0;
        __pyx_t_6 = ((PyObject *)__pyx_tp_new_7flitter_6render_7physics_CollisionForceApplier(((PyTypeObject *)__pyx_ptype_7flitter_6render_7physics_CollisionForceApplier), __pyx_t_9, NULL)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 598, __pyx_L1_error)
        __Pyx_GOTREF((PyObject *)__pyx_t_6);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_group->matrix_forces, ((PyObject *)__pyx_t_6)); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(1, 598, __pyx_L1_error)
        __Pyx_DECREF((PyObject *)__pyx_t_6); __pyx_t_6 = 0;

        /* "flitter/render/physics.pyx":597
 *                 elif child.kind is 'field':
 *                     group.particle_forces.append(FieldForceApplier.__new__(FieldForceApplier, child, strength, zero))
 *                 elif child.kind is 'collision':             # <<<<<<<<<<<<<<
 *                     group.matrix_forces.append(CollisionForceApplier.__new__(CollisionForceApplier, child, strength, zero))
 *                 elif child.kind is 'gravity':
 */
        goto __pyx_L21;
      }

      /* "flitter/render/physics.pyx":599
 *                 elif child.kind is 'collision':
 *                     group.matrix_forces.append(CollisionForceApplier.__new__(CollisionForceApplier, child, strength, zero))
 *                 elif child.kind is 'gravity':             # <<<<<<<<<<<<<<
 *                     group.matrix_forces.append(GravityForceApplier.__new__(GravityForceApplier, child, strength, zero))
 *                 elif child.kind is 'electrostatic':
 */
      __pyx_t_7 = (__pyx_v_child->kind == __pyx_n_u_gravity);
      if (__pyx_t_7) {

        /* "flitter/render/physics.pyx":600
 *                     group.matrix_forces.append(CollisionForceApplier.__new__(CollisionForceApplier, child, strength, zero))
 *                 elif child.kind is 'gravity':
 *                     group.matrix_forces.append(GravityForceApplier.__new__(GravityForceApplier, child, strength, zero))             # <<<<<<<<<<<<<<
 *                 elif child.kind is 'electrostatic':
 *                     group.matrix_forces.append(ElectrostaticForceApplier.__new__(ElectrostaticForceApplier, child, strength, zero))
 */
        if (unlikely(__pyx_v_group->matrix_forces == Py_None)) {
          PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
          __PYX_ERR(1, 600, __pyx_L1_error)
        }
        __pyx_t_6 = PyFloat_FromDouble(__pyx_v_strength); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 600, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 600, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_INCREF((PyObject *)__pyx_v_child);
        __Pyx_GIVEREF((PyObject *)__pyx_v_child);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)__pyx_v_child))) __PYX_ERR(1, 600, __pyx_L1_error);
        __Pyx_GIVEREF(__pyx_t_6);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_6)) __PYX_ERR(1, 600, __pyx_L1_error);
        __Pyx_INCREF((PyObject *)__pyx_v_zero);
        __Pyx_GIVEREF((PyObject *)__pyx_v_zero);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, ((PyObject *)__pyx_v_zero))) __PYX_ERR(1, 600, __pyx_L1_error);
        __pyx_t_6 = 0;
        __pyx_t_6 = ((PyObject *)__pyx_tp_new_7flitter_6render_7physics_GravityForceApplier(((PyTypeObject *)__pyx_ptype_7flitter_6render_7physics_GravityForceApplier), __pyx_t_9, NULL)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 600, __pyx_L1_error)
        __Pyx_GOTREF((PyObject *)__pyx_t_6);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_group->matrix_forces, ((PyObject *)__pyx_t_6)); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(1, 600, __pyx_L1_error)
        __Pyx_DECREF((PyObject *)__pyx_t_6); __pyx_t_6 = 0;

        /* "flitter/render/physics.pyx":599
 *                 elif child.kind is 'collision':
 *                     group.matrix_forces.append(CollisionForceApplier.__new__(CollisionForceApplier, child, strength, zero))
 *                 elif child.kind is 'gravity':             # <<<<<<<<<<<<<<
 *                     group.matrix_forces.append(GravityForceApplier.__new__(GravityForceApplier, child, strength, zero))
 *                 elif child.kind is 'electrostatic':
 */
        goto __pyx_L21;
      }

      /* "flitter/render/physics.pyx":601
 *                 elif child.kind is 'gravity':
 *                     group.matrix_forces.append(GravityForceApplier.__new__(GravityForceApplier, child, strength, zero))
 *                 elif child.kind is 'electrostatic':             # <<<<<<<<<<<<<<
 *                     group.matrix_forces.append(ElectrostaticForceApplier.__new__(ElectrostaticForceApplier, child, strength, zero))
 *                 elif child.kind is 'adhesion':
 */
      __pyx_t_7 = (__pyx_v_child->kind == __pyx_n_u_electrostatic);
      if (__pyx_t_7) {

        /* "flitter/render/physics.pyx":602
 *                     group.matrix_forces.append(GravityForceApplier.__new__(GravityForceApplier, child, strength, zero))
 *                 elif child.kind is 'electrostatic':
 *                     group.matrix_forces.append(ElectrostaticForceApplier.__new__(ElectrostaticForceApplier, child, strength, zero))             # <<<<<<<<<<<<<<
 *                 elif child.kind is 'adhesion':
 *                     group.matrix_forces.append(AdhesionForceApplier.__new__(AdhesionForceApplier, child, strength, zero))
 */
        if (unlikely(__pyx_v_group->matrix_forces == Py_None)) {
          PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
          __PYX_ERR(1, 602, __pyx_L1_error)
        }
        __pyx_t_6 = PyFloat_FromDouble(__pyx_v_strength); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 602, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 602, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_INCREF((PyObject *)__pyx_v_child);
        __Pyx_GIVEREF((PyObject *)__pyx_v_child);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)__pyx_v_child))) __PYX_ERR(1, 602, __pyx_L1_error);
        __Pyx_GIVEREF(__pyx_t_6);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_6)) __PYX_ERR(1, 602, __pyx_L1_error);
        __Pyx_INCREF((PyObject *)__pyx_v_zero);
        __Pyx_GIVEREF((PyObject *)__pyx_v_zero);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, ((PyObject *)__pyx_v_zero))) __PYX_ERR(1, 602, __pyx_L1_error);
        __pyx_t_6 = 0;
        __pyx_t_6 = ((PyObject *)__pyx_tp_new_7flitter_6render_7physics_ElectrostaticForceApplier(((PyTypeObject *)__pyx_ptype_7flitter_6render_7physics_ElectrostaticForceApplier), __pyx_t_9, NULL)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 602, __pyx_L1_error)
        __Pyx_GOTREF((PyObject *)__pyx_t_6);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_group->matrix_forces, ((PyObject *)__pyx_t_6)); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(1, 602, __pyx_L1_error)
        __Pyx_DECREF((PyObject *)__pyx_t_6); __pyx_t_6 = 0;

        /* "flitter/render/physics.pyx":601
 *                 elif child.kind is 'gravity':
 *                     group.matrix_forces.append(GravityForceApplier.__new__(GravityForceApplier, child, strength, zero))
 *                 elif child.kind is 'electrostatic':             # <<<<<<<<<<<<<<
 *                     group.matrix_forces.append(ElectrostaticForceApplier.__new__(ElectrostaticForceApplier, child, strength, zero))
 *                 elif child.kind is 'adhesion':
 */
        goto __pyx_L21;
      }

      /* "flitter/render/physics.pyx":603
 *                 elif child.kind is 'electrostatic':
 *                     group.matrix_forces.append(ElectrostaticForceApplier.__new__(ElectrostaticForceApplier, child, strength, zero))
 *                 elif child.kind is 'adhesion':             # <<<<<<<<<<<<<<
 *                     group.matrix_forces.append(AdhesionForceApplier.__new__(AdhesionForceApplier, child, strength, zero))
 *                 elif child.kind is 'barrier':
 */
      __pyx_t_7 = (__pyx_v_child->kind == __pyx_n_u_adhesion);
      if (__pyx_t_7) {

        /* "flitter/render/physics.pyx":604
 *                     group.matrix_forces.append(ElectrostaticForceApplier.__new__(ElectrostaticForceApplier, child, strength, zero))
 *                 elif child.kind is 'adhesion':
 *                     group.matrix_forces.append(AdhesionForceApplier.__new__(AdhesionForceApplier, child, strength, zero))             # <<<<<<<<<<<<<<
 *                 elif child.kind is 'barrier':
 *                     barrier = Barrier.__new__(Barrier, child, strength, zero)
 */
        if (unlikely(__pyx_v_group->matrix_forces == Py_None)) {
          PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
          __PYX_ERR(1, 604, __pyx_L1_error)
        }
        __pyx_t_6 = PyFloat_FromDouble(__pyx_v_strength); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 604, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 604, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_INCREF((PyObject *)__pyx_v_child);
        __Pyx_GIVEREF((PyObject *)__pyx_v_child);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)__pyx_v_child))) __PYX_ERR(1, 604, __pyx_L1_error);
        __Pyx_GIVEREF(__pyx_t_6);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_6)) __PYX_ERR(1, 604, __pyx_L1_error);
        __Pyx_INCREF((PyObject *)__pyx_v_zero);
        __Pyx_GIVEREF((PyObject *)__pyx_v_zero);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, ((PyObject *)__pyx_v_zero))) __PYX_ERR(1, 604, __pyx_L1_error);
        __pyx_t_6 = 0;
        __pyx_t_6 = ((PyObject *)__pyx_tp_new_7flitter_6render_7physics_AdhesionForceApplier(((PyTypeObject *)__pyx_ptype_7flitter_6render_7physics_AdhesionForceApplier), __pyx_t_9, NULL)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 604, __pyx_L1_error)
        __Pyx_GOTREF((PyObject *)__pyx_t_6);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_group->matrix_forces, ((PyObject *)__pyx_t_6)); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(1, 604, __pyx_L1_error)
        __Pyx_DECREF((PyObject *)__pyx_t_6); __pyx_t_6 = 0;

        /* "flitter/render/physics.pyx":603
 *                 elif child.kind is 'electrostatic':
 *                     group.matrix_forces.append(ElectrostaticForceApplier.__new__(ElectrostaticForceApplier, child, strength, zero))
 *                 elif child.kind is 'adhesion':             # <<<<<<<<<<<<<<
 *                     group.matrix_forces.append(AdhesionForceApplier.__new__(AdhesionForceApplier, child, strength, zero))
 *                 elif child.kind is 'barrier':
 */
        goto __pyx_L21;
      }

      /* "flitter/render/physics.pyx":605
 *                 elif child.kind is 'adhesion':
 *                     group.matrix_forces.append(AdhesionForceApplier.__new__(AdhesionForceApplier, child, strength, zero))
 *                 elif child.kind is 'barrier':             # <<<<<<<<<<<<<<
 *                     barrier = Barrier.__new__(Barrier, child, strength, zero)
 *                     group.barriers.append(barrier)
 */
      __pyx_t_7 = (__pyx_v_child->kind == __pyx_n_u_barrier);
      if (__pyx_t_7) {

        /* "flitter/render/physics.pyx":606
 *                     group.matrix_forces.append(AdhesionForceApplier.__new__(AdhesionForceApplier, child, strength, zero))
 *                 elif child.kind is 'barrier':
 *                     barrier = Barrier.__new__(Barrier, child, strength, zero)             # <<<<<<<<<<<<<<
 *                     group.barriers.append(barrier)
 * 
 */
        __pyx_t_6 = PyFloat_FromDouble(__pyx_v_strength); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 606, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_9 = PyTuple_New(3); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 606, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_INCREF((PyObject *)__pyx_v_child);
        __Pyx_GIVEREF((PyObject *)__pyx_v_child);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 0, ((PyObject *)__pyx_v_child))) __PYX_ERR(1, 606, __pyx_L1_error);
        __Pyx_GIVEREF(__pyx_t_6);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_6)) __PYX_ERR(1, 606, __pyx_L1_error);
        __Pyx_INCREF((PyObject *)__pyx_v_zero);
        __Pyx_GIVEREF((PyObject *)__pyx_v_zero);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_9, 2, ((PyObject *)__pyx_v_zero))) __PYX_ERR(1, 606, __pyx_L1_error);
        __pyx_t_6 = 0;
        __pyx_t_6 = ((PyObject *)__pyx_tp_new_7flitter_6render_7physics_Barrier(((PyTypeObject *)__pyx_ptype_7flitter_6render_7physics_Barrier), __pyx_t_9, NULL)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 606, __pyx_L1_error)
        __Pyx_GOTREF((PyObject *)__pyx_t_6);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_XDECREF_SET(__pyx_v_barrier, ((struct __pyx_obj_7flitter_6render_7physics_Barrier *)__pyx_t_6));
        __pyx_t_6 = 0;

        /* "flitter/render/physics.pyx":607
 *                 elif child.kind is 'barrier':
 *                     barrier = Barrier.__new__(Barrier, child, strength, zero)
 *                     group.barriers.append(barrier)             # <<<<<<<<<<<<<<
 * 
 *     cpdef double calculate(self, list groups, dict particles_by_id, bint realtime, bint extra, double time, double last_time, double clock):
 */
        if (unlikely(__pyx_v_group->barriers == Py_None)) {
          PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
          __PYX_ERR(1, 607, __pyx_L1_error)
        }
        __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_group->barriers, ((PyObject *)__pyx_v_barrier)); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(1, 607, __pyx_L1_error)

        /* "flitter/render/physics.pyx":605
 *                 elif child.kind is 'adhesion':
 *                     group.matrix_forces.append(AdhesionForceApplier.__new__(AdhesionForceApplier, child, strength, zero))
 *                 elif child.kind is 'barrier':             # <<<<<<<<<<<<<<
 *                     barrier = Barrier.__new__(Barrier, child, strength, zero)
 *                     group.barriers.append(barrier)
 */
      }
      __pyx_L21:;
    }
    __pyx_L7:;

    /* "flitter/render/physics.pyx":525
 *         cdef Vector position, velocity
 *         cdef double fixed
 *         for child in node._children:             # <<<<<<<<<<<<<<
 *             if child.kind is 'particle' or child.kind is 'anchor':
 *                 if (particle_id := child.get_vec('id', None)) is not None:
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "flitter/render/physics.pyx":508
 *         state.set_item(state_prefix.concat(CLOCK), time_vector)
 * 
 *     cdef void collect(self, list groups, PhysicsGroup group, Node node, dict particles_by_id, set old_state_keys, set new_state_keys,             # <<<<<<<<<<<<<<
 *                       StateDict state, Vector state_prefix, double clock, Vector seed):
 *         cdef Node child
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("flitter.render.physics.PhysicsSystem.collect", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_child);
  __Pyx_XDECREF((PyObject *)__pyx_v_particle_id);
  __Pyx_XDECREF((PyObject *)__pyx_v_from_particle_id);
  __Pyx_XDECREF((PyObject *)__pyx_v_to_particle_id);
  __Pyx_XDECREF((PyObject *)__pyx_v_particle);
  __Pyx_XDECREF((PyObject *)__pyx_v_particle_force);
  __Pyx_XDECREF((PyObject *)__pyx_v_distance_force);
  __Pyx_XDECREF((PyObject *)__pyx_v_barrier);
  __Pyx_XDECREF((PyObject *)__pyx_v_child_group);
  __Pyx_XDECREF((PyObject *)__pyx_v_zero);
  __Pyx_XDECREF((PyObject *)__pyx_v_position);
  __Pyx_XDECREF((PyObject *)__pyx_v_velocity);
  __Pyx_XDECREF((PyObject *)__pyx_v_seed);
  __Pyx_RefNannyFinishContext();
}

/* "flitter/render/physics.pyx":609
 *                     group.barriers.append(barrier)
 * 
 *     cpdef double calculate(self, list groups, dict particles_by_id, bint realtime, bint extra, double time, double last_time, double clock):             # <<<<<<<<<<<<<<
 *         cdef list all_particles, particles, forces, barriers
 *         cdef int64_t i, j, k, m, n, p
 */

static PyObject *__pyx_pw_7flitter_6render_7physics_13PhysicsSystem_12calculate(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static double __pyx_f_7flitter_6render_7physics_13PhysicsSystem_calculate(struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem *__pyx_v_self, PyObject *__pyx_v_groups, PyObject *__pyx_v_particles_by_id, int __pyx_v_realtime, int __pyx_v_extra, double __pyx_v_time, double __pyx_v_last_time, double __pyx_v_clock, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_all_particles = 0;
  PyObject *__pyx_v_particles = 0;
  PyObject *__pyx_v_forces = 0;
  PyObject *__pyx_v_barriers = 0;
  int64_t __pyx_v_i;
  int64_t __pyx_v_j;
  int64_t __pyx_v_k;
  int64_t __pyx_v_m;
  int64_t __pyx_v_n;
  int64_t __pyx_v_p;
  double __pyx_v_delta;
  struct __pyx_obj_7flitter_5model_Vector *__pyx_v_direction = 0;
  double __pyx_v_d;
  double __pyx_v_distance;
  double __pyx_v_inv_distance;
  PyObject *__pyx_v_from_particle;
  PyObject *__pyx_v_to_particle;
  PyObject *__pyx_v_force;
  struct __pyx_obj_7flitter_6render_7physics_PhysicsGroup *__pyx_v_group = 0;
  int __pyx_v_more;
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  unsigned int __pyx_t_10;
  double __pyx_t_11;
  int64_t __pyx_t_12;
  double __pyx_t_13;
  double __pyx_t_14;
  int __pyx_t_15;
  int __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  int64_t __pyx_t_18;
  int64_t __pyx_t_19;
  int64_t __pyx_t_20;
  int64_t __pyx_t_21;
  int64_t __pyx_t_22;
  int64_t __pyx_t_23;
  int __pyx_t_24;
  int64_t __pyx_t_25;
  int64_t __pyx_t_26;
  int64_t __pyx_t_27;
  PyObject *__pyx_t_28;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("calculate", 1);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_calculate); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 609, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void*) __pyx_pw_7flitter_6render_7physics_13PhysicsSystem_12calculate)) {
        __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_realtime); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 609, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_extra); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 609, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = PyFloat_FromDouble(__pyx_v_time); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 609, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = PyFloat_FromDouble(__pyx_v_last_time); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 609, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = PyFloat_FromDouble(__pyx_v_clock); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 609, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_8 = __pyx_t_1; __pyx_t_9 = NULL;
        __pyx_t_10 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_8))) {
          __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
          if (likely(__pyx_t_9)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
            __Pyx_INCREF(__pyx_t_9);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_8, function);
            __pyx_t_10 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[8] = {__pyx_t_9, __pyx_v_groups, __pyx_v_particles_by_id, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_10, 7+__pyx_t_10);
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 609, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        __pyx_t_11 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_11 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 609, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_11;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "flitter/render/physics.pyx":613
 *         cdef int64_t i, j, k, m, n, p
 *         cdef double delta
 *         cdef Vector direction = Vector.__new__(Vector)             # <<<<<<<<<<<<<<
 *         direction.allocate_numbers(self.dimensions)
 *         cdef double d, distance, inv_distance
 */
  __pyx_t_1 = __Pyx_tp_new(((PyObject *)__pyx_ptype_7flitter_5model_Vector), __pyx_empty_tuple); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7flitter_5model_Vector)))) __PYX_ERR(1, 613, __pyx_L1_error)
  __pyx_v_direction = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "flitter/render/physics.pyx":614
 *         cdef double delta
 *         cdef Vector direction = Vector.__new__(Vector)
 *         direction.allocate_numbers(self.dimensions)             # <<<<<<<<<<<<<<
 *         cdef double d, distance, inv_distance
 *         cdef PyObject* from_particle
 */
  __pyx_t_12 = ((struct __pyx_vtabstruct_7flitter_5model_Vector *)__pyx_v_direction->__pyx_vtab)->allocate_numbers(__pyx_v_direction, __pyx_v_self->dimensions); if (unlikely(__pyx_t_12 == ((int64_t)-1L))) __PYX_ERR(1, 614, __pyx_L1_error)

  /* "flitter/render/physics.pyx":621
 *         cdef PhysicsGroup group
 *         cdef bint more
 *         while True:             # <<<<<<<<<<<<<<
 *             delta = min(self.resolution, time-last_time)
 *             last_time += delta
 */
  while (1) {

    /* "flitter/render/physics.pyx":622
 *         cdef bint more
 *         while True:
 *             delta = min(self.resolution, time-last_time)             # <<<<<<<<<<<<<<
 *             last_time += delta
 *             more = last_time < time and (not realtime or extra)
 */
    __pyx_t_11 = (__pyx_v_time - __pyx_v_last_time);
    __pyx_t_13 = __pyx_v_self->resolution;
    __pyx_t_15 = (__pyx_t_11 < __pyx_t_13);
    if (__pyx_t_15) {
      __pyx_t_14 = __pyx_t_11;
    } else {
      __pyx_t_14 = __pyx_t_13;
    }
    __pyx_v_delta = __pyx_t_14;

    /* "flitter/render/physics.pyx":623
 *         while True:
 *             delta = min(self.resolution, time-last_time)
 *             last_time += delta             # <<<<<<<<<<<<<<
 *             more = last_time < time and (not realtime or extra)
 *             for group in groups:
 */
    __pyx_v_last_time = (__pyx_v_last_time + __pyx_v_delta);

    /* "flitter/render/physics.pyx":624
 *             delta = min(self.resolution, time-last_time)
 *             last_time += delta
 *             more = last_time < time and (not realtime or extra)             # <<<<<<<<<<<<<<
 *             for group in groups:
 *                 forces = group.distance_forces
 */
    __pyx_t_16 = (__pyx_v_last_time < __pyx_v_time);
    if (__pyx_t_16) {
    } else {
      __pyx_t_15 = __pyx_t_16;
      goto __pyx_L5_bool_binop_done;
    }
    __pyx_t_16 = (!__pyx_v_realtime);
    if (!__pyx_t_16) {
    } else {
      __pyx_t_15 = __pyx_t_16;
      goto __pyx_L5_bool_binop_done;
    }
    __pyx_t_15 = __pyx_v_extra;
    __pyx_L5_bool_binop_done:;
    __pyx_v_more = __pyx_t_15;

    /* "flitter/render/physics.pyx":625
 *             last_time += delta
 *             more = last_time < time and (not realtime or extra)
 *             for group in groups:             # <<<<<<<<<<<<<<
 *                 forces = group.distance_forces
 *                 m = PyList_GET_SIZE(forces)
 */
    if (unlikely(__pyx_v_groups == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(1, 625, __pyx_L1_error)
    }
    __pyx_t_1 = __pyx_v_groups; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_17 = 0;
    for (;;) {
      {
        Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely((__pyx_temp < 0))) __PYX_ERR(1, 625, __pyx_L1_error)
        #endif
        if (__pyx_t_17 >= __pyx_temp) break;
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_17); __Pyx_INCREF(__pyx_t_2); __pyx_t_17++; if (unlikely((0 < 0))) __PYX_ERR(1, 625, __pyx_L1_error)
      #else
      __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_17); __pyx_t_17++; if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 625, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      #endif
      if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_6render_7physics_PhysicsGroup))))) __PYX_ERR(1, 625, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_group, ((struct __pyx_obj_7flitter_6render_7physics_PhysicsGroup *)__pyx_t_2));
      __pyx_t_2 = 0;

      /* "flitter/render/physics.pyx":626
 *             more = last_time < time and (not realtime or extra)
 *             for group in groups:
 *                 forces = group.distance_forces             # <<<<<<<<<<<<<<
 *                 m = PyList_GET_SIZE(forces)
 *                 for i in range(m):
 */
      __pyx_t_2 = __pyx_v_group->distance_forces;
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_XDECREF_SET(__pyx_v_forces, ((PyObject*)__pyx_t_2));
      __pyx_t_2 = 0;

      /* "flitter/render/physics.pyx":627
 *             for group in groups:
 *                 forces = group.distance_forces
 *                 m = PyList_GET_SIZE(forces)             # <<<<<<<<<<<<<<
 *                 for i in range(m):
 *                     force = PyList_GET_ITEM(forces, i)
 */
      __pyx_v_m = PyList_GET_SIZE(__pyx_v_forces);

      /* "flitter/render/physics.pyx":628
 *                 forces = group.distance_forces
 *                 m = PyList_GET_SIZE(forces)
 *                 for i in range(m):             # <<<<<<<<<<<<<<
 *                     force = PyList_GET_ITEM(forces, i)
 *                     from_particle = PyDict_GetItem(particles_by_id, (<DistanceForceApplier>force).from_particle_id)
 */
      __pyx_t_12 = __pyx_v_m;
      __pyx_t_18 = __pyx_t_12;
      for (__pyx_t_19 = 0; __pyx_t_19 < __pyx_t_18; __pyx_t_19+=1) {
        __pyx_v_i = __pyx_t_19;

        /* "flitter/render/physics.pyx":629
 *                 m = PyList_GET_SIZE(forces)
 *                 for i in range(m):
 *                     force = PyList_GET_ITEM(forces, i)             # <<<<<<<<<<<<<<
 *                     from_particle = PyDict_GetItem(particles_by_id, (<DistanceForceApplier>force).from_particle_id)
 *                     if from_particle == NULL:
 */
        __pyx_v_force = PyList_GET_ITEM(__pyx_v_forces, __pyx_v_i);

        /* "flitter/render/physics.pyx":630
 *                 for i in range(m):
 *                     force = PyList_GET_ITEM(forces, i)
 *                     from_particle = PyDict_GetItem(particles_by_id, (<DistanceForceApplier>force).from_particle_id)             # <<<<<<<<<<<<<<
 *                     if from_particle == NULL:
 *                         continue
 */
        __pyx_t_2 = ((PyObject *)((struct __pyx_obj_7flitter_6render_7physics_DistanceForceApplier *)__pyx_v_force)->from_particle_id);
        __Pyx_INCREF(__pyx_t_2);
        __pyx_v_from_particle = PyDict_GetItem(__pyx_v_particles_by_id, __pyx_t_2);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "flitter/render/physics.pyx":631
 *                     force = PyList_GET_ITEM(forces, i)
 *                     from_particle = PyDict_GetItem(particles_by_id, (<DistanceForceApplier>force).from_particle_id)
 *                     if from_particle == NULL:             # <<<<<<<<<<<<<<
 *                         continue
 *                     to_particle = PyDict_GetItem(particles_by_id, (<DistanceForceApplier>force).to_particle_id)
 */
        __pyx_t_15 = (__pyx_v_from_particle == NULL);
        if (__pyx_t_15) {

          /* "flitter/render/physics.pyx":632
 *                     from_particle = PyDict_GetItem(particles_by_id, (<DistanceForceApplier>force).from_particle_id)
 *                     if from_particle == NULL:
 *                         continue             # <<<<<<<<<<<<<<
 *                     to_particle = PyDict_GetItem(particles_by_id, (<DistanceForceApplier>force).to_particle_id)
 *                     if to_particle == NULL:
 */
          goto __pyx_L10_continue;

          /* "flitter/render/physics.pyx":631
 *                     force = PyList_GET_ITEM(forces, i)
 *                     from_particle = PyDict_GetItem(particles_by_id, (<DistanceForceApplier>force).from_particle_id)
 *                     if from_particle == NULL:             # <<<<<<<<<<<<<<
 *                         continue
 *                     to_particle = PyDict_GetItem(particles_by_id, (<DistanceForceApplier>force).to_particle_id)
 */
        }

        /* "flitter/render/physics.pyx":633
 *                     if from_particle == NULL:
 *                         continue
 *                     to_particle = PyDict_GetItem(particles_by_id, (<DistanceForceApplier>force).to_particle_id)             # <<<<<<<<<<<<<<
 *                     if to_particle == NULL:
 *                         continue
 */
        __pyx_t_2 = ((PyObject *)((struct __pyx_obj_7flitter_6render_7physics_DistanceForceApplier *)__pyx_v_force)->to_particle_id);
        __Pyx_INCREF(__pyx_t_2);
        __pyx_v_to_particle = PyDict_GetItem(__pyx_v_particles_by_id, __pyx_t_2);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "flitter/render/physics.pyx":634
 *                         continue
 *                     to_particle = PyDict_GetItem(particles_by_id, (<DistanceForceApplier>force).to_particle_id)
 *                     if to_particle == NULL:             # <<<<<<<<<<<<<<
 *                         continue
 *                     distance = 0
 */
        __pyx_t_15 = (__pyx_v_to_particle == NULL);
        if (__pyx_t_15) {

          /* "flitter/render/physics.pyx":635
 *                     to_particle = PyDict_GetItem(particles_by_id, (<DistanceForceApplier>force).to_particle_id)
 *                     if to_particle == NULL:
 *                         continue             # <<<<<<<<<<<<<<
 *                     distance = 0
 *                     for k in range(self.dimensions):
 */
          goto __pyx_L10_continue;

          /* "flitter/render/physics.pyx":634
 *                         continue
 *                     to_particle = PyDict_GetItem(particles_by_id, (<DistanceForceApplier>force).to_particle_id)
 *                     if to_particle == NULL:             # <<<<<<<<<<<<<<
 *                         continue
 *                     distance = 0
 */
        }

        /* "flitter/render/physics.pyx":636
 *                     if to_particle == NULL:
 *                         continue
 *                     distance = 0             # <<<<<<<<<<<<<<
 *                     for k in range(self.dimensions):
 *                         d = (<Particle>to_particle).position.numbers[k] - (<Particle>from_particle).position.numbers[k]
 */
        __pyx_v_distance = 0.0;

        /* "flitter/render/physics.pyx":637
 *                         continue
 *                     distance = 0
 *                     for k in range(self.dimensions):             # <<<<<<<<<<<<<<
 *                         d = (<Particle>to_particle).position.numbers[k] - (<Particle>from_particle).position.numbers[k]
 *                         direction.numbers[k] = d
 */
        __pyx_t_20 = __pyx_v_self->dimensions;
        __pyx_t_21 = __pyx_t_20;
        for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_21; __pyx_t_22+=1) {
          __pyx_v_k = __pyx_t_22;

          /* "flitter/render/physics.pyx":638
 *                     distance = 0
 *                     for k in range(self.dimensions):
 *                         d = (<Particle>to_particle).position.numbers[k] - (<Particle>from_particle).position.numbers[k]             # <<<<<<<<<<<<<<
 *                         direction.numbers[k] = d
 *                         distance += d * d
 */
          __pyx_v_d = ((((struct __pyx_obj_7flitter_6render_7physics_Particle *)__pyx_v_to_particle)->position->numbers[__pyx_v_k]) - (((struct __pyx_obj_7flitter_6render_7physics_Particle *)__pyx_v_from_particle)->position->numbers[__pyx_v_k]));

          /* "flitter/render/physics.pyx":639
 *                     for k in range(self.dimensions):
 *                         d = (<Particle>to_particle).position.numbers[k] - (<Particle>from_particle).position.numbers[k]
 *                         direction.numbers[k] = d             # <<<<<<<<<<<<<<
 *                         distance += d * d
 *                     distance = sqrt(distance)
 */
          (__pyx_v_direction->numbers[__pyx_v_k]) = __pyx_v_d;

          /* "flitter/render/physics.pyx":640
 *                         d = (<Particle>to_particle).position.numbers[k] - (<Particle>from_particle).position.numbers[k]
 *                         direction.numbers[k] = d
 *                         distance += d * d             # <<<<<<<<<<<<<<
 *                     distance = sqrt(distance)
 *                     inv_distance = 1 / distance
 */
          __pyx_v_distance = (__pyx_v_distance + (__pyx_v_d * __pyx_v_d));
        }

        /* "flitter/render/physics.pyx":641
 *                         direction.numbers[k] = d
 *                         distance += d * d
 *                     distance = sqrt(distance)             # <<<<<<<<<<<<<<
 *                     inv_distance = 1 / distance
 *                     for k in range(self.dimensions):
 */
        __pyx_v_distance = sqrt(__pyx_v_distance);

        /* "flitter/render/physics.pyx":642
 *                         distance += d * d
 *                     distance = sqrt(distance)
 *                     inv_distance = 1 / distance             # <<<<<<<<<<<<<<
 *                     for k in range(self.dimensions):
 *                         direction.numbers[k] *= inv_distance
 */
        __pyx_v_inv_distance = (1.0 / __pyx_v_distance);

        /* "flitter/render/physics.pyx":643
 *                     distance = sqrt(distance)
 *                     inv_distance = 1 / distance
 *                     for k in range(self.dimensions):             # <<<<<<<<<<<<<<
 *                         direction.numbers[k] *= inv_distance
 *                     (<DistanceForceApplier>force).apply(<Particle>from_particle, <Particle>to_particle, direction, distance)
 */
        __pyx_t_20 = __pyx_v_self->dimensions;
        __pyx_t_21 = __pyx_t_20;
        for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_21; __pyx_t_22+=1) {
          __pyx_v_k = __pyx_t_22;

          /* "flitter/render/physics.pyx":644
 *                     inv_distance = 1 / distance
 *                     for k in range(self.dimensions):
 *                         direction.numbers[k] *= inv_distance             # <<<<<<<<<<<<<<
 *                     (<DistanceForceApplier>force).apply(<Particle>from_particle, <Particle>to_particle, direction, distance)
 *                 particles = group.particles
 */
          __pyx_t_23 = __pyx_v_k;
          (__pyx_v_direction->numbers[__pyx_t_23]) = ((__pyx_v_direction->numbers[__pyx_t_23]) * __pyx_v_inv_distance);
        }

        /* "flitter/render/physics.pyx":645
 *                     for k in range(self.dimensions):
 *                         direction.numbers[k] *= inv_distance
 *                     (<DistanceForceApplier>force).apply(<Particle>from_particle, <Particle>to_particle, direction, distance)             # <<<<<<<<<<<<<<
 *                 particles = group.particles
 *                 p = PyList_GET_SIZE(particles)
 */
        __pyx_f_7flitter_6render_7physics_20DistanceForceApplier_apply(((struct __pyx_obj_7flitter_6render_7physics_DistanceForceApplier *)__pyx_v_force), ((struct __pyx_obj_7flitter_6render_7physics_Particle *)__pyx_v_from_particle), ((struct __pyx_obj_7flitter_6render_7physics_Particle *)__pyx_v_to_particle), __pyx_v_direction, __pyx_v_distance);
        __pyx_L10_continue:;
      }

      /* "flitter/render/physics.pyx":646
 *                         direction.numbers[k] *= inv_distance
 *                     (<DistanceForceApplier>force).apply(<Particle>from_particle, <Particle>to_particle, direction, distance)
 *                 particles = group.particles             # <<<<<<<<<<<<<<
 *                 p = PyList_GET_SIZE(particles)
 *                 if not p:
 */
      __pyx_t_2 = __pyx_v_group->particles;
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_XDECREF_SET(__pyx_v_particles, ((PyObject*)__pyx_t_2));
      __pyx_t_2 = 0;

      /* "flitter/render/physics.pyx":647
 *                     (<DistanceForceApplier>force).apply(<Particle>from_particle, <Particle>to_particle, direction, distance)
 *                 particles = group.particles
 *                 p = PyList_GET_SIZE(particles)             # <<<<<<<<<<<<<<
 *                 if not p:
 *                     continue
 */
      __pyx_v_p = PyList_GET_SIZE(__pyx_v_particles);

      /* "flitter/render/physics.pyx":648
 *                 particles = group.particles
 *                 p = PyList_GET_SIZE(particles)
 *                 if not p:             # <<<<<<<<<<<<<<
 *                     continue
 *                 all_particles = []
 */
      __pyx_t_15 = (!(__pyx_v_p != 0));
      if (__pyx_t_15) {

        /* "flitter/render/physics.pyx":649
 *                 p = PyList_GET_SIZE(particles)
 *                 if not p:
 *                     continue             # <<<<<<<<<<<<<<
 *                 all_particles = []
 *                 while group is not None:
 */
        goto __pyx_L8_continue;

        /* "flitter/render/physics.pyx":648
 *                 particles = group.particles
 *                 p = PyList_GET_SIZE(particles)
 *                 if not p:             # <<<<<<<<<<<<<<
 *                     continue
 *                 all_particles = []
 */
      }

      /* "flitter/render/physics.pyx":650
 *                 if not p:
 *                     continue
 *                 all_particles = []             # <<<<<<<<<<<<<<
 *                 while group is not None:
 *                     all_particles.extend(group.particles)
 */
      __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 650, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_XDECREF_SET(__pyx_v_all_particles, ((PyObject*)__pyx_t_2));
      __pyx_t_2 = 0;

      /* "flitter/render/physics.pyx":651
 *                     continue
 *                 all_particles = []
 *                 while group is not None:             # <<<<<<<<<<<<<<
 *                     all_particles.extend(group.particles)
 *                     forces = group.matrix_forces
 */
      while (1) {
        __pyx_t_15 = (((PyObject *)__pyx_v_group) != Py_None);
        if (!__pyx_t_15) break;

        /* "flitter/render/physics.pyx":652
 *                 all_particles = []
 *                 while group is not None:
 *                     all_particles.extend(group.particles)             # <<<<<<<<<<<<<<
 *                     forces = group.matrix_forces
 *                     group = group.parent
 */
        __pyx_t_2 = __pyx_v_group->particles;
        __Pyx_INCREF(__pyx_t_2);
        __pyx_t_24 = __Pyx_PyList_Extend(__pyx_v_all_particles, __pyx_t_2); if (unlikely(__pyx_t_24 == ((int)-1))) __PYX_ERR(1, 652, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "flitter/render/physics.pyx":653
 *                 while group is not None:
 *                     all_particles.extend(group.particles)
 *                     forces = group.matrix_forces             # <<<<<<<<<<<<<<
 *                     group = group.parent
 *                     m = PyList_GET_SIZE(forces)
 */
        __pyx_t_2 = __pyx_v_group->matrix_forces;
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_DECREF_SET(__pyx_v_forces, ((PyObject*)__pyx_t_2));
        __pyx_t_2 = 0;

        /* "flitter/render/physics.pyx":654
 *                     all_particles.extend(group.particles)
 *                     forces = group.matrix_forces
 *                     group = group.parent             # <<<<<<<<<<<<<<
 *                     m = PyList_GET_SIZE(forces)
 *                     if not m:
 */
        __pyx_t_2 = ((PyObject *)__pyx_v_group->parent);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_DECREF_SET(__pyx_v_group, ((struct __pyx_obj_7flitter_6render_7physics_PhysicsGroup *)__pyx_t_2));
        __pyx_t_2 = 0;

        /* "flitter/render/physics.pyx":655
 *                     forces = group.matrix_forces
 *                     group = group.parent
 *                     m = PyList_GET_SIZE(forces)             # <<<<<<<<<<<<<<
 *                     if not m:
 *                         continue
 */
        __pyx_v_m = PyList_GET_SIZE(__pyx_v_forces);

        /* "flitter/render/physics.pyx":656
 *                     group = group.parent
 *                     m = PyList_GET_SIZE(forces)
 *                     if not m:             # <<<<<<<<<<<<<<
 *                         continue
 *                     with nogil:
 */
        __pyx_t_15 = (!(__pyx_v_m != 0));
        if (__pyx_t_15) {

          /* "flitter/render/physics.pyx":657
 *                     m = PyList_GET_SIZE(forces)
 *                     if not m:
 *                         continue             # <<<<<<<<<<<<<<
 *                     with nogil:
 *                         n = PyList_GET_SIZE(all_particles)
 */
          goto __pyx_L19_continue;

          /* "flitter/render/physics.pyx":656
 *                     group = group.parent
 *                     m = PyList_GET_SIZE(forces)
 *                     if not m:             # <<<<<<<<<<<<<<
 *                         continue
 *                     with nogil:
 */
        }

        /* "flitter/render/physics.pyx":658
 *                     if not m:
 *                         continue
 *                     with nogil:             # <<<<<<<<<<<<<<
 *                         n = PyList_GET_SIZE(all_particles)
 *                         for i in range(p):
 */
        {
            #ifdef WITH_THREAD
            PyThreadState *_save;
            _save = NULL;
            Py_UNBLOCK_THREADS
            __Pyx_FastGIL_Remember();
            #endif
            /*try:*/ {

              /* "flitter/render/physics.pyx":659
 *                         continue
 *                     with nogil:
 *                         n = PyList_GET_SIZE(all_particles)             # <<<<<<<<<<<<<<
 *                         for i in range(p):
 *                             from_particle = PyList_GET_ITEM(all_particles, i)
 */
              __pyx_v_n = PyList_GET_SIZE(__pyx_v_all_particles);

              /* "flitter/render/physics.pyx":660
 *                     with nogil:
 *                         n = PyList_GET_SIZE(all_particles)
 *                         for i in range(p):             # <<<<<<<<<<<<<<
 *                             from_particle = PyList_GET_ITEM(all_particles, i)
 *                             for j in range(i+1, n):
 */
              __pyx_t_12 = __pyx_v_p;
              __pyx_t_18 = __pyx_t_12;
              for (__pyx_t_19 = 0; __pyx_t_19 < __pyx_t_18; __pyx_t_19+=1) {
                __pyx_v_i = __pyx_t_19;

                /* "flitter/render/physics.pyx":661
 *                         n = PyList_GET_SIZE(all_particles)
 *                         for i in range(p):
 *                             from_particle = PyList_GET_ITEM(all_particles, i)             # <<<<<<<<<<<<<<
 *                             for j in range(i+1, n):
 *                                 to_particle = PyList_GET_ITEM(all_particles, j)
 */
                __pyx_v_from_particle = PyList_GET_ITEM(__pyx_v_all_particles, __pyx_v_i);

                /* "flitter/render/physics.pyx":662
 *                         for i in range(p):
 *                             from_particle = PyList_GET_ITEM(all_particles, i)
 *                             for j in range(i+1, n):             # <<<<<<<<<<<<<<
 *                                 to_particle = PyList_GET_ITEM(all_particles, j)
 *                                 distance = 0
 */
                __pyx_t_20 = __pyx_v_n;
                __pyx_t_21 = __pyx_t_20;
                for (__pyx_t_22 = (__pyx_v_i + 1); __pyx_t_22 < __pyx_t_21; __pyx_t_22+=1) {
                  __pyx_v_j = __pyx_t_22;

                  /* "flitter/render/physics.pyx":663
 *                             from_particle = PyList_GET_ITEM(all_particles, i)
 *                             for j in range(i+1, n):
 *                                 to_particle = PyList_GET_ITEM(all_particles, j)             # <<<<<<<<<<<<<<
 *                                 distance = 0
 *                                 for k in range(self.dimensions):
 */
                  __pyx_v_to_particle = PyList_GET_ITEM(__pyx_v_all_particles, __pyx_v_j);

                  /* "flitter/render/physics.pyx":664
 *                             for j in range(i+1, n):
 *                                 to_particle = PyList_GET_ITEM(all_particles, j)
 *                                 distance = 0             # <<<<<<<<<<<<<<
 *                                 for k in range(self.dimensions):
 *                                     d = (<Particle>to_particle).position.numbers[k] - (<Particle>from_particle).position.numbers[k]
 */
                  __pyx_v_distance = 0.0;

                  /* "flitter/render/physics.pyx":665
 *                                 to_particle = PyList_GET_ITEM(all_particles, j)
 *                                 distance = 0
 *                                 for k in range(self.dimensions):             # <<<<<<<<<<<<<<
 *                                     d = (<Particle>to_particle).position.numbers[k] - (<Particle>from_particle).position.numbers[k]
 *                                     direction.numbers[k] = d
 */
                  __pyx_t_23 = __pyx_v_self->dimensions;
                  __pyx_t_25 = __pyx_t_23;
                  for (__pyx_t_26 = 0; __pyx_t_26 < __pyx_t_25; __pyx_t_26+=1) {
                    __pyx_v_k = __pyx_t_26;

                    /* "flitter/render/physics.pyx":666
 *                                 distance = 0
 *                                 for k in range(self.dimensions):
 *                                     d = (<Particle>to_particle).position.numbers[k] - (<Particle>from_particle).position.numbers[k]             # <<<<<<<<<<<<<<
 *                                     direction.numbers[k] = d
 *                                     distance += d * d
 */
                    __pyx_v_d = ((((struct __pyx_obj_7flitter_6render_7physics_Particle *)__pyx_v_to_particle)->position->numbers[__pyx_v_k]) - (((struct __pyx_obj_7flitter_6render_7physics_Particle *)__pyx_v_from_particle)->position->numbers[__pyx_v_k]));

                    /* "flitter/render/physics.pyx":667
 *                                 for k in range(self.dimensions):
 *                                     d = (<Particle>to_particle).position.numbers[k] - (<Particle>from_particle).position.numbers[k]
 *                                     direction.numbers[k] = d             # <<<<<<<<<<<<<<
 *                                     distance += d * d
 *                                 distance = sqrt(distance)
 */
                    (__pyx_v_direction->numbers[__pyx_v_k]) = __pyx_v_d;

                    /* "flitter/render/physics.pyx":668
 *                                     d = (<Particle>to_particle).position.numbers[k] - (<Particle>from_particle).position.numbers[k]
 *                                     direction.numbers[k] = d
 *                                     distance += d * d             # <<<<<<<<<<<<<<
 *                                 distance = sqrt(distance)
 *                                 inv_distance = 1 / distance
 */
                    __pyx_v_distance = (__pyx_v_distance + (__pyx_v_d * __pyx_v_d));
                  }

                  /* "flitter/render/physics.pyx":669
 *                                     direction.numbers[k] = d
 *                                     distance += d * d
 *                                 distance = sqrt(distance)             # <<<<<<<<<<<<<<
 *                                 inv_distance = 1 / distance
 *                                 for k in range(self.dimensions):
 */
                  __pyx_v_distance = sqrt(__pyx_v_distance);

                  /* "flitter/render/physics.pyx":670
 *                                     distance += d * d
 *                                 distance = sqrt(distance)
 *                                 inv_distance = 1 / distance             # <<<<<<<<<<<<<<
 *                                 for k in range(self.dimensions):
 *                                     direction.numbers[k] *= inv_distance
 */
                  __pyx_v_inv_distance = (1.0 / __pyx_v_distance);

                  /* "flitter/render/physics.pyx":671
 *                                 distance = sqrt(distance)
 *                                 inv_distance = 1 / distance
 *                                 for k in range(self.dimensions):             # <<<<<<<<<<<<<<
 *                                     direction.numbers[k] *= inv_distance
 *                                 for k in range(m):
 */
                  __pyx_t_23 = __pyx_v_self->dimensions;
                  __pyx_t_25 = __pyx_t_23;
                  for (__pyx_t_26 = 0; __pyx_t_26 < __pyx_t_25; __pyx_t_26+=1) {
                    __pyx_v_k = __pyx_t_26;

                    /* "flitter/render/physics.pyx":672
 *                                 inv_distance = 1 / distance
 *                                 for k in range(self.dimensions):
 *                                     direction.numbers[k] *= inv_distance             # <<<<<<<<<<<<<<
 *                                 for k in range(m):
 *                                     force = PyList_GET_ITEM(forces, k)
 */
                    __pyx_t_27 = __pyx_v_k;
                    (__pyx_v_direction->numbers[__pyx_t_27]) = ((__pyx_v_direction->numbers[__pyx_t_27]) * __pyx_v_inv_distance);
                  }

                  /* "flitter/render/physics.pyx":673
 *                                 for k in range(self.dimensions):
 *                                     direction.numbers[k] *= inv_distance
 *                                 for k in range(m):             # <<<<<<<<<<<<<<
 *                                     force = PyList_GET_ITEM(forces, k)
 *                                     if not (<MatrixForceApplier>force).max_distance or distance < (<MatrixForceApplier>force).max_distance:
 */
                  __pyx_t_23 = __pyx_v_m;
                  __pyx_t_25 = __pyx_t_23;
                  for (__pyx_t_26 = 0; __pyx_t_26 < __pyx_t_25; __pyx_t_26+=1) {
                    __pyx_v_k = __pyx_t_26;

                    /* "flitter/render/physics.pyx":674
 *                                     direction.numbers[k] *= inv_distance
 *                                 for k in range(m):
 *                                     force = PyList_GET_ITEM(forces, k)             # <<<<<<<<<<<<<<
 *                                     if not (<MatrixForceApplier>force).max_distance or distance < (<MatrixForceApplier>force).max_distance:
 *                                         (<MatrixForceApplier>force).apply(<Particle>from_particle, <Particle>to_particle, direction, distance)
 */
                    __pyx_v_force = PyList_GET_ITEM(__pyx_v_forces, __pyx_v_k);

                    /* "flitter/render/physics.pyx":675
 *                                 for k in range(m):
 *                                     force = PyList_GET_ITEM(forces, k)
 *                                     if not (<MatrixForceApplier>force).max_distance or distance < (<MatrixForceApplier>force).max_distance:             # <<<<<<<<<<<<<<
 *                                         (<MatrixForceApplier>force).apply(<Particle>from_particle, <Particle>to_particle, direction, distance)
 *             for group in groups:
 */
                    __pyx_t_16 = (!(((struct __pyx_obj_7flitter_6render_7physics_MatrixForceApplier *)__pyx_v_force)->max_distance != 0));
                    if (!__pyx_t_16) {
                    } else {
                      __pyx_t_15 = __pyx_t_16;
                      goto __pyx_L38_bool_binop_done;
                    }
                    __pyx_t_16 = (__pyx_v_distance < ((struct __pyx_obj_7flitter_6render_7physics_MatrixForceApplier *)__pyx_v_force)->max_distance);
                    __pyx_t_15 = __pyx_t_16;
                    __pyx_L38_bool_binop_done:;
                    if (__pyx_t_15) {

                      /* "flitter/render/physics.pyx":676
 *                                     force = PyList_GET_ITEM(forces, k)
 *                                     if not (<MatrixForceApplier>force).max_distance or distance < (<MatrixForceApplier>force).max_distance:
 *                                         (<MatrixForceApplier>force).apply(<Particle>from_particle, <Particle>to_particle, direction, distance)             # <<<<<<<<<<<<<<
 *             for group in groups:
 *                 particles = group.particles
 */
                      ((struct __pyx_vtabstruct_7flitter_6render_7physics_MatrixForceApplier *)((struct __pyx_obj_7flitter_6render_7physics_MatrixForceApplier *)__pyx_v_force)->__pyx_vtab)->apply(((struct __pyx_obj_7flitter_6render_7physics_MatrixForceApplier *)__pyx_v_force), ((struct __pyx_obj_7flitter_6render_7physics_Particle *)__pyx_v_from_particle), ((struct __pyx_obj_7flitter_6render_7physics_Particle *)__pyx_v_to_particle), __pyx_v_direction, __pyx_v_distance);

                      /* "flitter/render/physics.pyx":675
 *                                 for k in range(m):
 *                                     force = PyList_GET_ITEM(forces, k)
 *                                     if not (<MatrixForceApplier>force).max_distance or distance < (<MatrixForceApplier>force).max_distance:             # <<<<<<<<<<<<<<
 *                                         (<MatrixForceApplier>force).apply(<Particle>from_particle, <Particle>to_particle, direction, distance)
 *             for group in groups:
 */
                    }
                  }
                }
              }
            }

            /* "flitter/render/physics.pyx":658
 *                     if not m:
 *                         continue
 *                     with nogil:             # <<<<<<<<<<<<<<
 *                         n = PyList_GET_SIZE(all_particles)
 *                         for i in range(p):
 */
            /*finally:*/ {
              /*normal exit:*/{
                #ifdef WITH_THREAD
                __Pyx_FastGIL_Forget();
                Py_BLOCK_THREADS
                #endif
                goto __pyx_L26;
              }
              __pyx_L26:;
            }
        }
        __pyx_L19_continue:;
      }

      /* "flitter/render/physics.pyx":625
 *             last_time += delta
 *             more = last_time < time and (not realtime or extra)
 *             for group in groups:             # <<<<<<<<<<<<<<
 *                 forces = group.distance_forces
 *                 m = PyList_GET_SIZE(forces)
 */
      __pyx_L8_continue:;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "flitter/render/physics.pyx":677
 *                                     if not (<MatrixForceApplier>force).max_distance or distance < (<MatrixForceApplier>force).max_distance:
 *                                         (<MatrixForceApplier>force).apply(<Particle>from_particle, <Particle>to_particle, direction, distance)
 *             for group in groups:             # <<<<<<<<<<<<<<
 *                 particles = group.particles
 *                 p = PyList_GET_SIZE(particles)
 */
    if (unlikely(__pyx_v_groups == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(1, 677, __pyx_L1_error)
    }
    __pyx_t_1 = __pyx_v_groups; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_17 = 0;
    for (;;) {
      {
        Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely((__pyx_temp < 0))) __PYX_ERR(1, 677, __pyx_L1_error)
        #endif
        if (__pyx_t_17 >= __pyx_temp) break;
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_17); __Pyx_INCREF(__pyx_t_2); __pyx_t_17++; if (unlikely((0 < 0))) __PYX_ERR(1, 677, __pyx_L1_error)
      #else
      __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_17); __pyx_t_17++; if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 677, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      #endif
      if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_7flitter_6render_7physics_PhysicsGroup))))) __PYX_ERR(1, 677, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_group, ((struct __pyx_obj_7flitter_6render_7physics_PhysicsGroup *)__pyx_t_2));
      __pyx_t_2 = 0;

      /* "flitter/render/physics.pyx":678
 *                                         (<MatrixForceApplier>force).apply(<Particle>from_particle, <Particle>to_particle, direction, distance)
 *             for group in groups:
 *                 particles = group.particles             # <<<<<<<<<<<<<<
 *                 p = PyList_GET_SIZE(particles)
 *                 if p == 0:
 */
      __pyx_t_2 = __pyx_v_group->particles;
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_XDECREF_SET(__pyx_v_particles, ((PyObject*)__pyx_t_2));
      __pyx_t_2 = 0;

      /* "flitter/render/physics.pyx":679
 *             for group in groups:
 *                 particles = group.particles
 *                 p = PyList_GET_SIZE(particles)             # <<<<<<<<<<<<<<
 *                 if p == 0:
 *                     continue
 */
      __pyx_v_p = PyList_GET_SIZE(__pyx_v_particles);

      /* "flitter/render/physics.pyx":680
 *                 particles = group.particles
 *                 p = PyList_GET_SIZE(particles)
 *                 if p == 0:             # <<<<<<<<<<<<<<
 *                     continue
 *                 if group.parent is None:
 */
      __pyx_t_15 = (__pyx_v_p == 0);
      if (__pyx_t_15) {

        /* "flitter/render/physics.pyx":681
 *                 p = PyList_GET_SIZE(particles)
 *                 if p == 0:
 *                     continue             # <<<<<<<<<<<<<<
 *                 if group.parent is None:
 *                     forces = group.particle_forces
 */
        goto __pyx_L41_continue;

        /* "flitter/render/physics.pyx":680
 *                 particles = group.particles
 *                 p = PyList_GET_SIZE(particles)
 *                 if p == 0:             # <<<<<<<<<<<<<<
 *                     continue
 *                 if group.parent is None:
 */
      }

      /* "flitter/render/physics.pyx":682
 *                 if p == 0:
 *                     continue
 *                 if group.parent is None:             # <<<<<<<<<<<<<<
 *                     forces = group.particle_forces
 *                     barriers = group.barriers
 */
      __pyx_t_15 = (((PyObject *)__pyx_v_group->parent) == Py_None);
      if (__pyx_t_15) {

        /* "flitter/render/physics.pyx":683
 *                     continue
 *                 if group.parent is None:
 *                     forces = group.particle_forces             # <<<<<<<<<<<<<<
 *                     barriers = group.barriers
 *                 else:
 */
        __pyx_t_2 = __pyx_v_group->particle_forces;
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_XDECREF_SET(__pyx_v_forces, ((PyObject*)__pyx_t_2));
        __pyx_t_2 = 0;

        /* "flitter/render/physics.pyx":684
 *                 if group.parent is None:
 *                     forces = group.particle_forces
 *                     barriers = group.barriers             # <<<<<<<<<<<<<<
 *                 else:
 *                     forces = list(group.particle_forces)
 */
        __pyx_t_2 = __pyx_v_group->barriers;
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_XDECREF_SET(__pyx_v_barriers, ((PyObject*)__pyx_t_2));
        __pyx_t_2 = 0;

        /* "flitter/render/physics.pyx":682
 *                 if p == 0:
 *                     continue
 *                 if group.parent is None:             # <<<<<<<<<<<<<<
 *                     forces = group.particle_forces
 *                     barriers = group.barriers
 */
        goto __pyx_L44;
      }

      /* "flitter/render/physics.pyx":686
 *                     barriers = group.barriers
 *                 else:
 *                     forces = list(group.particle_forces)             # <<<<<<<<<<<<<<
 *                     barriers = list(group.barriers)
 *                     while (group := group.parent) is not None:
 */
      /*else*/ {
        __pyx_t_2 = PySequence_List(__pyx_v_group->particle_forces); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 686, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_XDECREF_SET(__pyx_v_forces, ((PyObject*)__pyx_t_2));
        __pyx_t_2 = 0;

        /* "flitter/render/physics.pyx":687
 *                 else:
 *                     forces = list(group.particle_forces)
 *                     barriers = list(group.barriers)             # <<<<<<<<<<<<<<
 *                     while (group := group.parent) is not None:
 *                         forces.extend(group.particle_forces)
 */
        __pyx_t_2 = PySequence_List(__pyx_v_group->barriers); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 687, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_XDECREF_SET(__pyx_v_barriers, ((PyObject*)__pyx_t_2));
        __pyx_t_2 = 0;

        /* "flitter/render/physics.pyx":688
 *                     forces = list(group.particle_forces)
 *                     barriers = list(group.barriers)
 *                     while (group := group.parent) is not None:             # <<<<<<<<<<<<<<
 *                         forces.extend(group.particle_forces)
 *                         barriers.extend(group.barriers)
 */
        while (1) {
          __pyx_t_2 = ((PyObject *)__pyx_v_group->parent);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_DECREF_SET(__pyx_v_group, ((struct __pyx_obj_7flitter_6render_7physics_PhysicsGroup *)__pyx_t_2));
          __Pyx_INCREF((PyObject *)__pyx_t_2);
          __pyx_t_15 = (((PyObject *)__pyx_t_2) != Py_None);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          if (!__pyx_t_15) break;

          /* "flitter/render/physics.pyx":689
 *                     barriers = list(group.barriers)
 *                     while (group := group.parent) is not None:
 *                         forces.extend(group.particle_forces)             # <<<<<<<<<<<<<<
 *                         barriers.extend(group.barriers)
 *                 with nogil:
 */
          __pyx_t_2 = __pyx_v_group->particle_forces;
          __Pyx_INCREF(__pyx_t_2);
          __pyx_t_24 = __Pyx_PyList_Extend(__pyx_v_forces, __pyx_t_2); if (unlikely(__pyx_t_24 == ((int)-1))) __PYX_ERR(1, 689, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "flitter/render/physics.pyx":690
 *                     while (group := group.parent) is not None:
 *                         forces.extend(group.particle_forces)
 *                         barriers.extend(group.barriers)             # <<<<<<<<<<<<<<
 *                 with nogil:
 *                     m = PyList_GET_SIZE(forces)
 */
          __pyx_t_2 = __pyx_v_group->barriers;
          __Pyx_INCREF(__pyx_t_2);
          __pyx_t_24 = __Pyx_PyList_Extend(__pyx_v_barriers, __pyx_t_2); if (unlikely(__pyx_t_24 == ((int)-1))) __PYX_ERR(1, 690, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }
      }
      __pyx_L44:;

      /* "flitter/render/physics.pyx":691
 *                         forces.extend(group.particle_forces)
 *                         barriers.extend(group.barriers)
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     m = PyList_GET_SIZE(forces)
 *                     n = PyList_GET_SIZE(barriers)
 */
      {
          #ifdef WITH_THREAD
          PyThreadState *_save;
          _save = NULL;
          Py_UNBLOCK_THREADS
          __Pyx_FastGIL_Remember();
          #endif
          /*try:*/ {

            /* "flitter/render/physics.pyx":692
 *                         barriers.extend(group.barriers)
 *                 with nogil:
 *                     m = PyList_GET_SIZE(forces)             # <<<<<<<<<<<<<<
 *                     n = PyList_GET_SIZE(barriers)
 *                     for i in range(p):
 */
            __pyx_v_m = PyList_GET_SIZE(__pyx_v_forces);

            /* "flitter/render/physics.pyx":693
 *                 with nogil:
 *                     m = PyList_GET_SIZE(forces)
 *                     n = PyList_GET_SIZE(barriers)             # <<<<<<<<<<<<<<
 *                     for i in range(p):
 *                         to_particle = PyList_GET_ITEM(particles, i)
 */
            __pyx_v_n = PyList_GET_SIZE(__pyx_v_barriers);

            /* "flitter/render/physics.pyx":694
 *                     m = PyList_GET_SIZE(forces)
 *                     n = PyList_GET_SIZE(barriers)
 *                     for i in range(p):             # <<<<<<<<<<<<<<
 *                         to_particle = PyList_GET_ITEM(particles, i)
 *                         if (<Particle>to_particle).non_anchor:
 */
            __pyx_t_12 = __pyx_v_p;
            __pyx_t_18 = __pyx_t_12;
            for (__pyx_t_19 = 0; __pyx_t_19 < __pyx_t_18; __pyx_t_19+=1) {
              __pyx_v_i = __pyx_t_19;

              /* "flitter/render/physics.pyx":695
 *                     n = PyList_GET_SIZE(barriers)
 *                     for i in range(p):
 *                         to_particle = PyList_GET_ITEM(particles, i)             # <<<<<<<<<<<<<<
 *                         if (<Particle>to_particle).non_anchor:
 *                             for j in range(m):
 */
              __pyx_v_to_particle = PyList_GET_ITEM(__pyx_v_particles, __pyx_v_i);

              /* "flitter/render/physics.pyx":696
 *                     for i in range(p):
 *                         to_particle = PyList_GET_ITEM(particles, i)
 *                         if (<Particle>to_particle).non_anchor:             # <<<<<<<<<<<<<<
 *                             for j in range(m):
 *                                 (<ParticleForceApplier>PyList_GET_ITEM(forces, j)).apply(<Particle>to_particle, delta)
 */
              if (((struct __pyx_obj_7flitter_6render_7physics_Particle *)__pyx_v_to_particle)->non_anchor) {

                /* "flitter/render/physics.pyx":697
 *                         to_particle = PyList_GET_ITEM(particles, i)
 *                         if (<Particle>to_particle).non_anchor:
 *                             for j in range(m):             # <<<<<<<<<<<<<<
 *                                 (<ParticleForceApplier>PyList_GET_ITEM(forces, j)).apply(<Particle>to_particle, delta)
 *                             for j in range(n):
 */
                __pyx_t_20 = __pyx_v_m;
                __pyx_t_21 = __pyx_t_20;
                for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_21; __pyx_t_22+=1) {
                  __pyx_v_j = __pyx_t_22;

                  /* "flitter/render/physics.pyx":698
 *                         if (<Particle>to_particle).non_anchor:
 *                             for j in range(m):
 *                                 (<ParticleForceApplier>PyList_GET_ITEM(forces, j)).apply(<Particle>to_particle, delta)             # <<<<<<<<<<<<<<
 *                             for j in range(n):
 *                                 (<Barrier>PyList_GET_ITEM(barriers, j)).apply_distance(<Particle>to_particle)
 */
                  __pyx_t_28 = PyList_GET_ITEM(__pyx_v_forces, __pyx_v_j);
                  ((struct __pyx_vtabstruct_7flitter_6render_7physics_ParticleForceApplier *)((struct __pyx_obj_7flitter_6render_7physics_ParticleForceApplier *)__pyx_t_28)->__pyx_vtab)->apply(((struct __pyx_obj_7flitter_6render_7physics_ParticleForceApplier *)__pyx_t_28), ((struct __pyx_obj_7flitter_6render_7physics_Particle *)__pyx_v_to_particle), __pyx_v_delta);
                }

                /* "flitter/render/physics.pyx":699
 *                             for j in range(m):
 *                                 (<ParticleForceApplier>PyList_GET_ITEM(forces, j)).apply(<Particle>to_particle, delta)
 *                             for j in range(n):             # <<<<<<<<<<<<<<
 *                                 (<Barrier>PyList_GET_ITEM(barriers, j)).apply_distance(<Particle>to_particle)
 *                             (<Particle>to_particle).step(self.speed_of_light, clock, delta)
 */
                __pyx_t_20 = __pyx_v_n;
                __pyx_t_21 = __pyx_t_20;
                for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_21; __pyx_t_22+=1) {
                  __pyx_v_j = __pyx_t_22;

                  /* "flitter/render/physics.pyx":700
 *                                 (<ParticleForceApplier>PyList_GET_ITEM(forces, j)).apply(<Particle>to_particle, delta)
 *                             for j in range(n):
 *                                 (<Barrier>PyList_GET_ITEM(barriers, j)).apply_distance(<Particle>to_particle)             # <<<<<<<<<<<<<<
 *                             (<Particle>to_particle).step(self.speed_of_light, clock, delta)
 *                             for j in range(n):
 */
                  __pyx_t_28 = PyList_GET_ITEM(__pyx_v_barriers, __pyx_v_j);
                  __pyx_f_7flitter_6render_7physics_7Barrier_apply_distance(((struct __pyx_obj_7flitter_6render_7physics_Barrier *)__pyx_t_28), ((struct __pyx_obj_7flitter_6render_7physics_Particle *)__pyx_v_to_particle));
                }

                /* "flitter/render/physics.pyx":701
 *                             for j in range(n):
 *                                 (<Barrier>PyList_GET_ITEM(barriers, j)).apply_distance(<Particle>to_particle)
 *                             (<Particle>to_particle).step(self.speed_of_light, clock, delta)             # <<<<<<<<<<<<<<
 *                             for j in range(n):
 *                                 (<Barrier>PyList_GET_ITEM(barriers, j)).apply_rebound(<Particle>to_particle, self.speed_of_light, clock, delta)
 */
                __pyx_f_7flitter_6render_7physics_8Particle_step(((struct __pyx_obj_7flitter_6render_7physics_Particle *)__pyx_v_to_particle), __pyx_v_self->speed_of_light, __pyx_v_clock, __pyx_v_delta);

                /* "flitter/render/physics.pyx":702
 *                                 (<Barrier>PyList_GET_ITEM(barriers, j)).apply_distance(<Particle>to_particle)
 *                             (<Particle>to_particle).step(self.speed_of_light, clock, delta)
 *                             for j in range(n):             # <<<<<<<<<<<<<<
 *                                 (<Barrier>PyList_GET_ITEM(barriers, j)).apply_rebound(<Particle>to_particle, self.speed_of_light, clock, delta)
 *                             if more:
 */
                __pyx_t_20 = __pyx_v_n;
                __pyx_t_21 = __pyx_t_20;
                for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_21; __pyx_t_22+=1) {
                  __pyx_v_j = __pyx_t_22;

                  /* "flitter/render/physics.pyx":703
 *                             (<Particle>to_particle).step(self.speed_of_light, clock, delta)
 *                             for j in range(n):
 *                                 (<Barrier>PyList_GET_ITEM(barriers, j)).apply_rebound(<Particle>to_particle, self.speed_of_light, clock, delta)             # <<<<<<<<<<<<<<
 *                             if more:
 *                                 for j in range(self.dimensions):
 */
                  __pyx_t_28 = PyList_GET_ITEM(__pyx_v_barriers, __pyx_v_j);
                  __pyx_f_7flitter_6render_7physics_7Barrier_apply_rebound(((struct __pyx_obj_7flitter_6render_7physics_Barrier *)__pyx_t_28), ((struct __pyx_obj_7flitter_6render_7physics_Particle *)__pyx_v_to_particle), __pyx_v_self->speed_of_light, __pyx_v_clock, __pyx_v_delta);
                }

                /* "flitter/render/physics.pyx":704
 *                             for j in range(n):
 *                                 (<Barrier>PyList_GET_ITEM(barriers, j)).apply_rebound(<Particle>to_particle, self.speed_of_light, clock, delta)
 *                             if more:             # <<<<<<<<<<<<<<
 *                                 for j in range(self.dimensions):
 *                                     (<Particle>to_particle).force.numbers[j] = (<Particle>to_particle).initial_force.numbers[j]
 */
                if (__pyx_v_more) {

                  /* "flitter/render/physics.pyx":705
 *                                 (<Barrier>PyList_GET_ITEM(barriers, j)).apply_rebound(<Particle>to_particle, self.speed_of_light, clock, delta)
 *                             if more:
 *                                 for j in range(self.dimensions):             # <<<<<<<<<<<<<<
 *                                     (<Particle>to_particle).force.numbers[j] = (<Particle>to_particle).initial_force.numbers[j]
 *             clock += delta
 */
                  __pyx_t_20 = __pyx_v_self->dimensions;
                  __pyx_t_21 = __pyx_t_20;
                  for (__pyx_t_22 = 0; __pyx_t_22 < __pyx_t_21; __pyx_t_22+=1) {
                    __pyx_v_j = __pyx_t_22;

                    /* "flitter/render/physics.pyx":706
 *                             if more:
 *                                 for j in range(self.dimensions):
 *                                     (<Particle>to_particle).force.numbers[j] = (<Particle>to_particle).initial_force.numbers[j]             # <<<<<<<<<<<<<<
 *             clock += delta
 *             if not more:
 */
                    (((struct __pyx_obj_7flitter_6render_7physics_Particle *)__pyx_v_to_particle)->force->numbers[__pyx_v_j]) = (((struct __pyx_obj_7flitter_6render_7physics_Particle *)__pyx_v_to_particle)->initial_force->numbers[__pyx_v_j]);
                  }

                  /* "flitter/render/physics.pyx":704
 *                             for j in range(n):
 *                                 (<Barrier>PyList_GET_ITEM(barriers, j)).apply_rebound(<Particle>to_particle, self.speed_of_light, clock, delta)
 *                             if more:             # <<<<<<<<<<<<<<
 *                                 for j in range(self.dimensions):
 *                                     (<Particle>to_particle).force.numbers[j] = (<Particle>to_particle).initial_force.numbers[j]
 */
                }

                /* "flitter/render/physics.pyx":696
 *                     for i in range(p):
 *                         to_particle = PyList_GET_ITEM(particles, i)
 *                         if (<Particle>to_particle).non_anchor:             # <<<<<<<<<<<<<<
 *                             for j in range(m):
 *                                 (<ParticleForceApplier>PyList_GET_ITEM(forces, j)).apply(<Particle>to_particle, delta)
 */
              }
            }
          }

          /* "flitter/render/physics.pyx":691
 *                         forces.extend(group.particle_forces)
 *                         barriers.extend(group.barriers)
 *                 with nogil:             # <<<<<<<<<<<<<<
 *                     m = PyList_GET_SIZE(forces)
 *                     n = PyList_GET_SIZE(barriers)
 */
          /*finally:*/ {
            /*normal exit:*/{
              #ifdef WITH_THREAD
              __Pyx_FastGIL_Forget();
              Py_BLOCK_THREADS
              #endif
              goto __pyx_L51;
            }
            __pyx_L51:;
          }
      }

      /* "flitter/render/physics.pyx":677
 *                                     if not (<MatrixForceApplier>force).max_distance or distance < (<MatrixForceApplier>force).max_distance:
 *                                         (<MatrixForceApplier>force).apply(<Particle>from_particle, <Particle>to_particle, direction, distance)
 *             for group in groups:             # <<<<<<<<<<<<<<
 *                 particles = group.particles
 *                 p = PyList_GET_SIZE(particles)
 */
      __pyx_L41_continue:;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "flitter/render/physics.pyx":707
 *                                 for j in range(self.dimensions):
 *                                     (<Particle>to_particle).force.numbers[j] = (<Particle>to_particle).initial_force.numbers[j]
 *             clock += delta             # <<<<<<<<<<<<<<
 *             if not more:
 *                 break
 */
    __pyx_v_clock = (__pyx_v_clock + __pyx_v_delta);

    /* "flitter/render/physics.pyx":708
 *                                     (<Particle>to_particle).force.numbers[j] = (<Particle>to_particle).initial_force.numbers[j]
 *             clock += delta
 *             if not more:             # <<<<<<<<<<<<<<
 *                 break
 *             extra = False
 */
    __pyx_t_15 = (!__pyx_v_more);
    if (__pyx_t_15) {

      /* "flitter/render/physics.pyx":709
 *             clock += delta
 *             if not more:
 *                 break             # <<<<<<<<<<<<<<
 *             extra = False
 *         return clock
 */
      goto __pyx_L4_break;

      /* "flitter/render/physics.pyx":708
 *                                     (<Particle>to_particle).force.numbers[j] = (<Particle>to_particle).initial_force.numbers[j]
 *             clock += delta
 *             if not more:             # <<<<<<<<<<<<<<
 *                 break
 *             extra = False
 */
    }

    /* "flitter/render/physics.pyx":710
 *             if not more:
 *                 break
 *             extra = False             # <<<<<<<<<<<<<<
 *         return clock
 */
    __pyx_v_extra = 0;
  }
  __pyx_L4_break:;

  /* "flitter/render/physics.pyx":711
 *                 break
 *             extra = False
 *         return clock             # <<<<<<<<<<<<<<
 */
  __pyx_r = __pyx_v_clock;
  goto __pyx_L0;

  /* "flitter/render/physics.pyx":609
 *                     group.barriers.append(barrier)
 * 
 *     cpdef double calculate(self, list groups, dict particles_by_id, bint realtime, bint extra, double time, double last_time, double clock):             # <<<<<<<<<<<<<<
 *         cdef list all_particles, particles, forces, barriers
 *         cdef int64_t i, j, k, m, n, p
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("flitter.render.physics.PhysicsSystem.calculate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_all_particles);
  __Pyx_XDECREF(__pyx_v_particles);
  __Pyx_XDECREF(__pyx_v_forces);
  __Pyx_XDECREF(__pyx_v_barriers);
  __Pyx_XDECREF((PyObject *)__pyx_v_direction);
  __Pyx_XDECREF((PyObject *)__pyx_v_group);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_6render_7physics_13PhysicsSystem_12calculate(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_6render_7physics_13PhysicsSystem_12calculate = {"calculate", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_13PhysicsSystem_12calculate, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_6render_7physics_13PhysicsSystem_12calculate(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_groups = 0;
  PyObject *__pyx_v_particles_by_id = 0;
  int __pyx_v_realtime;
  int __pyx_v_extra;
  double __pyx_v_time;
  double __pyx_v_last_time;
  double __pyx_v_clock;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[7] = {0,0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("calculate (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_groups,&__pyx_n_s_particles_by_id,&__pyx_n_s_realtime,&__pyx_n_s_extra,&__pyx_n_s_time,&__pyx_n_s_last_time,&__pyx_n_s_clock,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  7: values[6] = __Pyx_Arg_FASTCALL(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = __Pyx_Arg_FASTCALL(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = __Pyx_Arg_FASTCALL(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_groups)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 609, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_particles_by_id)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 609, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("calculate", 1, 7, 7, 1); __PYX_ERR(1, 609, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_realtime)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 609, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("calculate", 1, 7, 7, 2); __PYX_ERR(1, 609, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_extra)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[3]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 609, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("calculate", 1, 7, 7, 3); __PYX_ERR(1, 609, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_time)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[4]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 609, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("calculate", 1, 7, 7, 4); __PYX_ERR(1, 609, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_last_time)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[5]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 609, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("calculate", 1, 7, 7, 5); __PYX_ERR(1, 609, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_clock)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[6]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 609, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("calculate", 1, 7, 7, 6); __PYX_ERR(1, 609, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "calculate") < 0)) __PYX_ERR(1, 609, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 7)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
      values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
      values[4] = __Pyx_Arg_FASTCALL(__pyx_args, 4);
      values[5] = __Pyx_Arg_FASTCALL(__pyx_args, 5);
      values[6] = __Pyx_Arg_FASTCALL(__pyx_args, 6);
    }
    __pyx_v_groups = ((PyObject*)values[0]);
    __pyx_v_particles_by_id = ((PyObject*)values[1]);
    __pyx_v_realtime = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_realtime == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 609, __pyx_L3_error)
    __pyx_v_extra = __Pyx_PyObject_IsTrue(values[3]); if (unlikely((__pyx_v_extra == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 609, __pyx_L3_error)
    __pyx_v_time = __pyx_PyFloat_AsDouble(values[4]); if (unlikely((__pyx_v_time == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 609, __pyx_L3_error)
    __pyx_v_last_time = __pyx_PyFloat_AsDouble(values[5]); if (unlikely((__pyx_v_last_time == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 609, __pyx_L3_error)
    __pyx_v_clock = __pyx_PyFloat_AsDouble(values[6]); if (unlikely((__pyx_v_clock == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 609, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("calculate", 1, 7, 7, __pyx_nargs); __PYX_ERR(1, 609, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.render.physics.PhysicsSystem.calculate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_groups), (&PyList_Type), 1, "groups", 1))) __PYX_ERR(1, 609, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_particles_by_id), (&PyDict_Type), 1, "particles_by_id", 1))) __PYX_ERR(1, 609, __pyx_L1_error)
  __pyx_r = __pyx_pf_7flitter_6render_7physics_13PhysicsSystem_11calculate(((struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem *)__pyx_v_self), __pyx_v_groups, __pyx_v_particles_by_id, __pyx_v_realtime, __pyx_v_extra, __pyx_v_time, __pyx_v_last_time, __pyx_v_clock);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_6render_7physics_13PhysicsSystem_11calculate(struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem *__pyx_v_self, PyObject *__pyx_v_groups, PyObject *__pyx_v_particles_by_id, int __pyx_v_realtime, int __pyx_v_extra, double __pyx_v_time, double __pyx_v_last_time, double __pyx_v_clock) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("calculate", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7flitter_6render_7physics_13PhysicsSystem_calculate(__pyx_v_self, __pyx_v_groups, __pyx_v_particles_by_id, __pyx_v_realtime, __pyx_v_extra, __pyx_v_time, __pyx_v_last_time, __pyx_v_clock, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 609, __pyx_L1_error)
  __pyx_t_2 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 609, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("flitter.render.physics.PhysicsSystem.calculate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_6render_7physics_13PhysicsSystem_14__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_6render_7physics_13PhysicsSystem_14__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_13PhysicsSystem_14__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_6render_7physics_13PhysicsSystem_14__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_7flitter_6render_7physics_13PhysicsSystem_13__reduce_cython__(((struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_6render_7physics_13PhysicsSystem_13__reduce_cython__(struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.dimensions, self.resolution, self.speed_of_light, self.state_keys)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_From_int64_t(__pyx_v_self->dimensions); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->resolution); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->speed_of_light); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->state_keys);
  __Pyx_GIVEREF(__pyx_v_self->state_keys);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_v_self->state_keys)) __PYX_ERR(0, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.dimensions, self.resolution, self.speed_of_light, self.state_keys)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_4 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v__dict = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "(tree fragment)":7
 *     state = (self.dimensions, self.resolution, self.speed_of_light, self.state_keys)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_5 = (__pyx_v__dict != Py_None);
  if (__pyx_t_5) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v__dict)) __PYX_ERR(0, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.state_keys is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.dimensions, self.resolution, self.speed_of_light, self.state_keys)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.state_keys is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_PhysicsSystem, (type(self), 0xf14ade3, None), state
 */
  /*else*/ {
    __pyx_t_5 = (__pyx_v_self->state_keys != ((PyObject*)Py_None));
    __pyx_v_use_setstate = __pyx_t_5;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.state_keys is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_PhysicsSystem, (type(self), 0xf14ade3, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.state_keys is not None
 *     if use_setstate:
 *         return __pyx_unpickle_PhysicsSystem, (type(self), 0xf14ade3, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_PhysicsSystem, (type(self), 0xf14ade3, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_PhysicsSystem); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_253013475);
    __Pyx_GIVEREF(__pyx_int_253013475);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_253013475)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, Py_None)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4)) __PYX_ERR(0, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state)) __PYX_ERR(0, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.state_keys is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_PhysicsSystem, (type(self), 0xf14ade3, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_PhysicsSystem, (type(self), 0xf14ade3, None), state
 *     else:
 *         return __pyx_unpickle_PhysicsSystem, (type(self), 0xf14ade3, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_PhysicsSystem__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_PhysicsSystem); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_253013475);
    __Pyx_GIVEREF(__pyx_int_253013475);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_253013475)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2)) __PYX_ERR(0, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4)) __PYX_ERR(0, 15, __pyx_L1_error);
    __pyx_t_2 = 0;
    __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.render.physics.PhysicsSystem.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PhysicsSystem, (type(self), 0xf14ade3, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PhysicsSystem__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_6render_7physics_13PhysicsSystem_16__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_6render_7physics_13PhysicsSystem_16__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_13PhysicsSystem_16__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_6render_7physics_13PhysicsSystem_16__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(0, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.render.physics.PhysicsSystem.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_6render_7physics_13PhysicsSystem_15__setstate_cython__(((struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_6render_7physics_13PhysicsSystem_15__setstate_cython__(struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_PhysicsSystem, (type(self), 0xf14ade3, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_PhysicsSystem__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7flitter_6render_7physics___pyx_unpickle_PhysicsSystem__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PhysicsSystem, (type(self), 0xf14ade3, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PhysicsSystem__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("flitter.render.physics.PhysicsSystem.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Particle(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_6render_7physics_1__pyx_unpickle_Particle(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_6render_7physics_1__pyx_unpickle_Particle = {"__pyx_unpickle_Particle", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_1__pyx_unpickle_Particle, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_6render_7physics_1__pyx_unpickle_Particle(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Particle (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Particle", 1, 3, 3, 1); __PYX_ERR(0, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Particle", 1, 3, 3, 2); __PYX_ERR(0, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_Particle") < 0)) __PYX_ERR(0, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Particle", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.render.physics.__pyx_unpickle_Particle", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_6render_7physics___pyx_unpickle_Particle(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_6render_7physics___pyx_unpickle_Particle(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Particle", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xc9134b9, 0x81fb8cd, 0xb66f251):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc9134b9, 0x81fb8cd, 0xb66f251) = (acceleration, charge, ease, force, id, initial_force, mass, non_anchor, position, position_state_key, radius, velocity, velocity_state_key))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__4, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xc9134b9, 0x81fb8cd, 0xb66f251):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc9134b9, 0x81fb8cd, 0xb66f251) = (acceleration, charge, ease, force, id, initial_force, mass, non_anchor, position, position_state_key, radius, velocity, velocity_state_key))" % __pyx_checksum
 *     __pyx_result = Particle.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(0, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xc9134b9, 0x81fb8cd, 0xb66f251):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc9134b9, 0x81fb8cd, 0xb66f251) = (acceleration, charge, ease, force, id, initial_force, mass, non_anchor, position, position_state_key, radius, velocity, velocity_state_key))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = Particle.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xc9134b9, 0x81fb8cd, 0xb66f251):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc9134b9, 0x81fb8cd, 0xb66f251) = (acceleration, charge, ease, force, id, initial_force, mass, non_anchor, position, position_state_key, radius, velocity, velocity_state_key))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc9134b9, 0x81fb8cd, 0xb66f251) = (acceleration, charge, ease, force, id, initial_force, mass, non_anchor, position, position_state_key, radius, velocity, velocity_state_key))" % __pyx_checksum
 *     __pyx_result = Particle.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Particle__set_state(<Particle> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_6render_7physics_Particle), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc9134b9, 0x81fb8cd, 0xb66f251) = (acceleration, charge, ease, force, id, initial_force, mass, non_anchor, position, position_state_key, radius, velocity, velocity_state_key))" % __pyx_checksum
 *     __pyx_result = Particle.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Particle__set_state(<Particle> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Particle.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Particle__set_state(<Particle> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Particle__set_state(Particle __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_6render_7physics___pyx_unpickle_Particle__set_state(((struct __pyx_obj_7flitter_6render_7physics_Particle *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc9134b9, 0x81fb8cd, 0xb66f251) = (acceleration, charge, ease, force, id, initial_force, mass, non_anchor, position, position_state_key, radius, velocity, velocity_state_key))" % __pyx_checksum
 *     __pyx_result = Particle.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Particle__set_state(<Particle> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Particle__set_state(<Particle> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Particle__set_state(Particle __pyx_result, tuple __pyx_state):
 *     __pyx_result.acceleration = __pyx_state[0]; __pyx_result.charge = __pyx_state[1]; __pyx_result.ease = __pyx_state[2]; __pyx_result.force = __pyx_state[3]; __pyx_result.id = __pyx_state[4]; __pyx_result.initial_force = __pyx_state[5]; __pyx_result.mass = __pyx_state[6]; __pyx_result.non_anchor = __pyx_state[7]; __pyx_result.position = __pyx_state[8]; __pyx_result.position_state_key = __pyx_state[9]; __pyx_result.radius = __pyx_state[10]; __pyx_result.velocity = __pyx_state[11]; __pyx_result.velocity_state_key = __pyx_state[12]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Particle(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.render.physics.__pyx_unpickle_Particle", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Particle__set_state(<Particle> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Particle__set_state(Particle __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.acceleration = __pyx_state[0]; __pyx_result.charge = __pyx_state[1]; __pyx_result.ease = __pyx_state[2]; __pyx_result.force = __pyx_state[3]; __pyx_result.id = __pyx_state[4]; __pyx_result.initial_force = __pyx_state[5]; __pyx_result.mass = __pyx_state[6]; __pyx_result.non_anchor = __pyx_state[7]; __pyx_result.position = __pyx_state[8]; __pyx_result.position_state_key = __pyx_state[9]; __pyx_result.radius = __pyx_state[10]; __pyx_result.velocity = __pyx_state[11]; __pyx_result.velocity_state_key = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_6render_7physics___pyx_unpickle_Particle__set_state(struct __pyx_obj_7flitter_6render_7physics_Particle *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  double __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Particle__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Particle__set_state(Particle __pyx_result, tuple __pyx_state):
 *     __pyx_result.acceleration = __pyx_state[0]; __pyx_result.charge = __pyx_state[1]; __pyx_result.ease = __pyx_state[2]; __pyx_result.force = __pyx_state[3]; __pyx_result.id = __pyx_state[4]; __pyx_result.initial_force = __pyx_state[5]; __pyx_result.mass = __pyx_state[6]; __pyx_result.non_anchor = __pyx_state[7]; __pyx_result.position = __pyx_state[8]; __pyx_result.position_state_key = __pyx_state[9]; __pyx_result.radius = __pyx_state[10]; __pyx_result.velocity = __pyx_state[11]; __pyx_result.velocity_state_key = __pyx_state[12]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[13])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  if (!(likely(((PyTuple_GET_ITEM(__pyx_v___pyx_state, 0)) == Py_None) || likely(__Pyx_TypeTest(PyTuple_GET_ITEM(__pyx_v___pyx_state, 0), __pyx_ptype_7flitter_5model_Vector))))) __PYX_ERR(0, 12, __pyx_L1_error)
  __pyx_t_1 = PyTuple_GET_ITEM(__pyx_v___pyx_state, 0);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->acceleration);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->acceleration);
  __pyx_v___pyx_result->acceleration = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_PyFloat_AsDouble(PyTuple_GET_ITEM(__pyx_v___pyx_state, 1)); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __pyx_v___pyx_result->charge = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_PyFloat_AsDouble(PyTuple_GET_ITEM(__pyx_v___pyx_state, 2)); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __pyx_v___pyx_result->ease = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  if (!(likely(((PyTuple_GET_ITEM(__pyx_v___pyx_state, 3)) == Py_None) || likely(__Pyx_TypeTest(PyTuple_GET_ITEM(__pyx_v___pyx_state, 3), __pyx_ptype_7flitter_5model_Vector))))) __PYX_ERR(0, 12, __pyx_L1_error)
  __pyx_t_1 = PyTuple_GET_ITEM(__pyx_v___pyx_state, 3);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->force);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->force);
  __pyx_v___pyx_result->force = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  if (!(likely(((PyTuple_GET_ITEM(__pyx_v___pyx_state, 4)) == Py_None) || likely(__Pyx_TypeTest(PyTuple_GET_ITEM(__pyx_v___pyx_state, 4), __pyx_ptype_7flitter_5model_Vector))))) __PYX_ERR(0, 12, __pyx_L1_error)
  __pyx_t_1 = PyTuple_GET_ITEM(__pyx_v___pyx_state, 4);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->id);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->id);
  __pyx_v___pyx_result->id = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  if (!(likely(((PyTuple_GET_ITEM(__pyx_v___pyx_state, 5)) == Py_None) || likely(__Pyx_TypeTest(PyTuple_GET_ITEM(__pyx_v___pyx_state, 5), __pyx_ptype_7flitter_5model_Vector))))) __PYX_ERR(0, 12, __pyx_L1_error)
  __pyx_t_1 = PyTuple_GET_ITEM(__pyx_v___pyx_state, 5);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->initial_force);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->initial_force);
  __pyx_v___pyx_result->initial_force = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_PyFloat_AsDouble(PyTuple_GET_ITEM(__pyx_v___pyx_state, 6)); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __pyx_v___pyx_result->mass = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyObject_IsTrue(PyTuple_GET_ITEM(__pyx_v___pyx_state, 7)); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __pyx_v___pyx_result->non_anchor = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  if (!(likely(((PyTuple_GET_ITEM(__pyx_v___pyx_state, 8)) == Py_None) || likely(__Pyx_TypeTest(PyTuple_GET_ITEM(__pyx_v___pyx_state, 8), __pyx_ptype_7flitter_5model_Vector))))) __PYX_ERR(0, 12, __pyx_L1_error)
  __pyx_t_1 = PyTuple_GET_ITEM(__pyx_v___pyx_state, 8);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->position);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->position);
  __pyx_v___pyx_result->position = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  if (!(likely(((PyTuple_GET_ITEM(__pyx_v___pyx_state, 9)) == Py_None) || likely(__Pyx_TypeTest(PyTuple_GET_ITEM(__pyx_v___pyx_state, 9), __pyx_ptype_7flitter_5model_Vector))))) __PYX_ERR(0, 12, __pyx_L1_error)
  __pyx_t_1 = PyTuple_GET_ITEM(__pyx_v___pyx_state, 9);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->position_state_key);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->position_state_key);
  __pyx_v___pyx_result->position_state_key = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_PyFloat_AsDouble(PyTuple_GET_ITEM(__pyx_v___pyx_state, 10)); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __pyx_v___pyx_result->radius = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  if (!(likely(((PyTuple_GET_ITEM(__pyx_v___pyx_state, 11)) == Py_None) || likely(__Pyx_TypeTest(PyTuple_GET_ITEM(__pyx_v___pyx_state, 11), __pyx_ptype_7flitter_5model_Vector))))) __PYX_ERR(0, 12, __pyx_L1_error)
  __pyx_t_1 = PyTuple_GET_ITEM(__pyx_v___pyx_state, 11);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->velocity);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->velocity);
  __pyx_v___pyx_result->velocity = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  if (!(likely(((PyTuple_GET_ITEM(__pyx_v___pyx_state, 12)) == Py_None) || likely(__Pyx_TypeTest(PyTuple_GET_ITEM(__pyx_v___pyx_state, 12), __pyx_ptype_7flitter_5model_Vector))))) __PYX_ERR(0, 12, __pyx_L1_error)
  __pyx_t_1 = PyTuple_GET_ITEM(__pyx_v___pyx_state, 12);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->velocity_state_key);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->velocity_state_key);
  __pyx_v___pyx_result->velocity_state_key = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Particle__set_state(Particle __pyx_result, tuple __pyx_state):
 *     __pyx_result.acceleration = __pyx_state[0]; __pyx_result.charge = __pyx_state[1]; __pyx_result.ease = __pyx_state[2]; __pyx_result.force = __pyx_state[3]; __pyx_result.id = __pyx_state[4]; __pyx_result.initial_force = __pyx_state[5]; __pyx_result.mass = __pyx_state[6]; __pyx_result.non_anchor = __pyx_state[7]; __pyx_result.position = __pyx_state[8]; __pyx_result.position_state_key = __pyx_state[9]; __pyx_result.radius = __pyx_state[10]; __pyx_result.velocity = __pyx_state[11]; __pyx_result.velocity_state_key = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[13])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 13, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 > 13);
  if (__pyx_t_5) {
  } else {
    __pyx_t_3 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_3 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result.acceleration = __pyx_state[0]; __pyx_result.charge = __pyx_state[1]; __pyx_result.ease = __pyx_state[2]; __pyx_result.force = __pyx_state[3]; __pyx_result.id = __pyx_state[4]; __pyx_result.initial_force = __pyx_state[5]; __pyx_result.mass = __pyx_state[6]; __pyx_result.non_anchor = __pyx_state[7]; __pyx_result.position = __pyx_state[8]; __pyx_result.position_state_key = __pyx_state[9]; __pyx_result.radius = __pyx_state[10]; __pyx_result.velocity = __pyx_state[11]; __pyx_result.velocity_state_key = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[13])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 14, __pyx_L1_error)
    }
    __pyx_t_6 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, PyTuple_GET_ITEM(__pyx_v___pyx_state, 13)};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Particle__set_state(Particle __pyx_result, tuple __pyx_state):
 *     __pyx_result.acceleration = __pyx_state[0]; __pyx_result.charge = __pyx_state[1]; __pyx_result.ease = __pyx_state[2]; __pyx_result.force = __pyx_state[3]; __pyx_result.id = __pyx_state[4]; __pyx_result.initial_force = __pyx_state[5]; __pyx_result.mass = __pyx_state[6]; __pyx_result.non_anchor = __pyx_state[7]; __pyx_result.position = __pyx_state[8]; __pyx_result.position_state_key = __pyx_state[9]; __pyx_result.radius = __pyx_state[10]; __pyx_result.velocity = __pyx_state[11]; __pyx_result.velocity_state_key = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[13])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Particle__set_state(<Particle> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Particle__set_state(Particle __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.acceleration = __pyx_state[0]; __pyx_result.charge = __pyx_state[1]; __pyx_result.ease = __pyx_state[2]; __pyx_result.force = __pyx_state[3]; __pyx_result.id = __pyx_state[4]; __pyx_result.initial_force = __pyx_state[5]; __pyx_result.mass = __pyx_state[6]; __pyx_result.non_anchor = __pyx_state[7]; __pyx_result.position = __pyx_state[8]; __pyx_result.position_state_key = __pyx_state[9]; __pyx_result.radius = __pyx_state[10]; __pyx_result.velocity = __pyx_state[11]; __pyx_result.velocity_state_key = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.render.physics.__pyx_unpickle_Particle__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_DistanceForceApplier(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_6render_7physics_3__pyx_unpickle_DistanceForceApplier(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_6render_7physics_3__pyx_unpickle_DistanceForceApplier = {"__pyx_unpickle_DistanceForceApplier", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_3__pyx_unpickle_DistanceForceApplier, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_6render_7physics_3__pyx_unpickle_DistanceForceApplier(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_DistanceForceApplier (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_DistanceForceApplier", 1, 3, 3, 1); __PYX_ERR(0, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_DistanceForceApplier", 1, 3, 3, 2); __PYX_ERR(0, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_DistanceForceApplier") < 0)) __PYX_ERR(0, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_DistanceForceApplier", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.render.physics.__pyx_unpickle_DistanceForceApplier", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_6render_7physics_2__pyx_unpickle_DistanceForceApplier(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_6render_7physics_2__pyx_unpickle_DistanceForceApplier(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_DistanceForceApplier", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xca186ab, 0x9db94ff, 0x9717dcb):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xca186ab, 0x9db94ff, 0x9717dcb) = (from_particle_id, maximum, minimum, power, strength, to_particle_id))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__6, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xca186ab, 0x9db94ff, 0x9717dcb):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xca186ab, 0x9db94ff, 0x9717dcb) = (from_particle_id, maximum, minimum, power, strength, to_particle_id))" % __pyx_checksum
 *     __pyx_result = DistanceForceApplier.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(0, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xca186ab, 0x9db94ff, 0x9717dcb):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xca186ab, 0x9db94ff, 0x9717dcb) = (from_particle_id, maximum, minimum, power, strength, to_particle_id))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = DistanceForceApplier.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xca186ab, 0x9db94ff, 0x9717dcb):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xca186ab, 0x9db94ff, 0x9717dcb) = (from_particle_id, maximum, minimum, power, strength, to_particle_id))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xca186ab, 0x9db94ff, 0x9717dcb) = (from_particle_id, maximum, minimum, power, strength, to_particle_id))" % __pyx_checksum
 *     __pyx_result = DistanceForceApplier.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_DistanceForceApplier__set_state(<DistanceForceApplier> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_6render_7physics_DistanceForceApplier), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xca186ab, 0x9db94ff, 0x9717dcb) = (from_particle_id, maximum, minimum, power, strength, to_particle_id))" % __pyx_checksum
 *     __pyx_result = DistanceForceApplier.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_DistanceForceApplier__set_state(<DistanceForceApplier> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = DistanceForceApplier.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_DistanceForceApplier__set_state(<DistanceForceApplier> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_DistanceForceApplier__set_state(DistanceForceApplier __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_6render_7physics___pyx_unpickle_DistanceForceApplier__set_state(((struct __pyx_obj_7flitter_6render_7physics_DistanceForceApplier *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xca186ab, 0x9db94ff, 0x9717dcb) = (from_particle_id, maximum, minimum, power, strength, to_particle_id))" % __pyx_checksum
 *     __pyx_result = DistanceForceApplier.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_DistanceForceApplier__set_state(<DistanceForceApplier> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_DistanceForceApplier__set_state(<DistanceForceApplier> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_DistanceForceApplier__set_state(DistanceForceApplier __pyx_result, tuple __pyx_state):
 *     __pyx_result.from_particle_id = __pyx_state[0]; __pyx_result.maximum = __pyx_state[1]; __pyx_result.minimum = __pyx_state[2]; __pyx_result.power = __pyx_state[3]; __pyx_result.strength = __pyx_state[4]; __pyx_result.to_particle_id = __pyx_state[5]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_DistanceForceApplier(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.render.physics.__pyx_unpickle_DistanceForceApplier", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_DistanceForceApplier__set_state(<DistanceForceApplier> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_DistanceForceApplier__set_state(DistanceForceApplier __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.from_particle_id = __pyx_state[0]; __pyx_result.maximum = __pyx_state[1]; __pyx_result.minimum = __pyx_state[2]; __pyx_result.power = __pyx_state[3]; __pyx_result.strength = __pyx_state[4]; __pyx_result.to_particle_id = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_6render_7physics___pyx_unpickle_DistanceForceApplier__set_state(struct __pyx_obj_7flitter_6render_7physics_DistanceForceApplier *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  double __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_DistanceForceApplier__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_DistanceForceApplier__set_state(DistanceForceApplier __pyx_result, tuple __pyx_state):
 *     __pyx_result.from_particle_id = __pyx_state[0]; __pyx_result.maximum = __pyx_state[1]; __pyx_result.minimum = __pyx_state[2]; __pyx_result.power = __pyx_state[3]; __pyx_result.strength = __pyx_state[4]; __pyx_result.to_particle_id = __pyx_state[5]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[6])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  if (!(likely(((PyTuple_GET_ITEM(__pyx_v___pyx_state, 0)) == Py_None) || likely(__Pyx_TypeTest(PyTuple_GET_ITEM(__pyx_v___pyx_state, 0), __pyx_ptype_7flitter_5model_Vector))))) __PYX_ERR(0, 12, __pyx_L1_error)
  __pyx_t_1 = PyTuple_GET_ITEM(__pyx_v___pyx_state, 0);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->from_particle_id);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->from_particle_id);
  __pyx_v___pyx_result->from_particle_id = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_PyFloat_AsDouble(PyTuple_GET_ITEM(__pyx_v___pyx_state, 1)); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __pyx_v___pyx_result->maximum = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_PyFloat_AsDouble(PyTuple_GET_ITEM(__pyx_v___pyx_state, 2)); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __pyx_v___pyx_result->minimum = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_PyFloat_AsDouble(PyTuple_GET_ITEM(__pyx_v___pyx_state, 3)); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __pyx_v___pyx_result->power = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_PyFloat_AsDouble(PyTuple_GET_ITEM(__pyx_v___pyx_state, 4)); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __pyx_v___pyx_result->strength = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  if (!(likely(((PyTuple_GET_ITEM(__pyx_v___pyx_state, 5)) == Py_None) || likely(__Pyx_TypeTest(PyTuple_GET_ITEM(__pyx_v___pyx_state, 5), __pyx_ptype_7flitter_5model_Vector))))) __PYX_ERR(0, 12, __pyx_L1_error)
  __pyx_t_1 = PyTuple_GET_ITEM(__pyx_v___pyx_state, 5);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->to_particle_id);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->to_particle_id);
  __pyx_v___pyx_result->to_particle_id = ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_DistanceForceApplier__set_state(DistanceForceApplier __pyx_result, tuple __pyx_state):
 *     __pyx_result.from_particle_id = __pyx_state[0]; __pyx_result.maximum = __pyx_state[1]; __pyx_result.minimum = __pyx_state[2]; __pyx_result.power = __pyx_state[3]; __pyx_result.strength = __pyx_state[4]; __pyx_result.to_particle_id = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[6])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 13, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 > 6);
  if (__pyx_t_5) {
  } else {
    __pyx_t_3 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_3 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result.from_particle_id = __pyx_state[0]; __pyx_result.maximum = __pyx_state[1]; __pyx_result.minimum = __pyx_state[2]; __pyx_result.power = __pyx_state[3]; __pyx_result.strength = __pyx_state[4]; __pyx_result.to_particle_id = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[6])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 14, __pyx_L1_error)
    }
    __pyx_t_6 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, PyTuple_GET_ITEM(__pyx_v___pyx_state, 6)};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_DistanceForceApplier__set_state(DistanceForceApplier __pyx_result, tuple __pyx_state):
 *     __pyx_result.from_particle_id = __pyx_state[0]; __pyx_result.maximum = __pyx_state[1]; __pyx_result.minimum = __pyx_state[2]; __pyx_result.power = __pyx_state[3]; __pyx_result.strength = __pyx_state[4]; __pyx_result.to_particle_id = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[6])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_DistanceForceApplier__set_state(<DistanceForceApplier> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_DistanceForceApplier__set_state(DistanceForceApplier __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.from_particle_id = __pyx_state[0]; __pyx_result.maximum = __pyx_state[1]; __pyx_result.minimum = __pyx_state[2]; __pyx_result.power = __pyx_state[3]; __pyx_result.strength = __pyx_state[4]; __pyx_result.to_particle_id = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("flitter.render.physics.__pyx_unpickle_DistanceForceApplier__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_PhysicsSystem(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_7flitter_6render_7physics_5__pyx_unpickle_PhysicsSystem(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7flitter_6render_7physics_5__pyx_unpickle_PhysicsSystem = {"__pyx_unpickle_PhysicsSystem", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_5__pyx_unpickle_PhysicsSystem, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7flitter_6render_7physics_5__pyx_unpickle_PhysicsSystem(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_PhysicsSystem (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PhysicsSystem", 1, 3, 3, 1); __PYX_ERR(0, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PhysicsSystem", 1, 3, 3, 2); __PYX_ERR(0, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_PhysicsSystem") < 0)) __PYX_ERR(0, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_PhysicsSystem", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("flitter.render.physics.__pyx_unpickle_PhysicsSystem", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7flitter_6render_7physics_4__pyx_unpickle_PhysicsSystem(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7flitter_6render_7physics_4__pyx_unpickle_PhysicsSystem(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_PhysicsSystem", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xf14ade3, 0x885a5ab, 0x3d7be7f):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xf14ade3, 0x885a5ab, 0x3d7be7f) = (dimensions, resolution, speed_of_light, state_keys))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__7, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xf14ade3, 0x885a5ab, 0x3d7be7f):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xf14ade3, 0x885a5ab, 0x3d7be7f) = (dimensions, resolution, speed_of_light, state_keys))" % __pyx_checksum
 *     __pyx_result = PhysicsSystem.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(0, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xf14ade3, 0x885a5ab, 0x3d7be7f):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xf14ade3, 0x885a5ab, 0x3d7be7f) = (dimensions, resolution, speed_of_light, state_keys))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = PhysicsSystem.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xf14ade3, 0x885a5ab, 0x3d7be7f):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xf14ade3, 0x885a5ab, 0x3d7be7f) = (dimensions, resolution, speed_of_light, state_keys))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xf14ade3, 0x885a5ab, 0x3d7be7f) = (dimensions, resolution, speed_of_light, state_keys))" % __pyx_checksum
 *     __pyx_result = PhysicsSystem.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PhysicsSystem__set_state(<PhysicsSystem> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7flitter_6render_7physics_PhysicsSystem), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xf14ade3, 0x885a5ab, 0x3d7be7f) = (dimensions, resolution, speed_of_light, state_keys))" % __pyx_checksum
 *     __pyx_result = PhysicsSystem.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_PhysicsSystem__set_state(<PhysicsSystem> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = PhysicsSystem.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PhysicsSystem__set_state(<PhysicsSystem> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_PhysicsSystem__set_state(PhysicsSystem __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(0, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7flitter_6render_7physics___pyx_unpickle_PhysicsSystem__set_state(((struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xf14ade3, 0x885a5ab, 0x3d7be7f) = (dimensions, resolution, speed_of_light, state_keys))" % __pyx_checksum
 *     __pyx_result = PhysicsSystem.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_PhysicsSystem__set_state(<PhysicsSystem> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_PhysicsSystem__set_state(<PhysicsSystem> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_PhysicsSystem__set_state(PhysicsSystem __pyx_result, tuple __pyx_state):
 *     __pyx_result.dimensions = __pyx_state[0]; __pyx_result.resolution = __pyx_state[1]; __pyx_result.speed_of_light = __pyx_state[2]; __pyx_result.state_keys = __pyx_state[3]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_PhysicsSystem(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("flitter.render.physics.__pyx_unpickle_PhysicsSystem", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_PhysicsSystem__set_state(<PhysicsSystem> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_PhysicsSystem__set_state(PhysicsSystem __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.dimensions = __pyx_state[0]; __pyx_result.resolution = __pyx_state[1]; __pyx_result.speed_of_light = __pyx_state[2]; __pyx_result.state_keys = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_7flitter_6render_7physics___pyx_unpickle_PhysicsSystem__set_state(struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int64_t __pyx_t_1;
  double __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  unsigned int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_PhysicsSystem__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_PhysicsSystem__set_state(PhysicsSystem __pyx_result, tuple __pyx_state):
 *     __pyx_result.dimensions = __pyx_state[0]; __pyx_result.resolution = __pyx_state[1]; __pyx_result.speed_of_light = __pyx_state[2]; __pyx_result.state_keys = __pyx_state[3]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[4])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_As_int64_t(PyTuple_GET_ITEM(__pyx_v___pyx_state, 0)); if (unlikely((__pyx_t_1 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __pyx_v___pyx_result->dimensions = __pyx_t_1;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_PyFloat_AsDouble(PyTuple_GET_ITEM(__pyx_v___pyx_state, 1)); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __pyx_v___pyx_result->resolution = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_PyFloat_AsDouble(PyTuple_GET_ITEM(__pyx_v___pyx_state, 2)); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 12, __pyx_L1_error)
  __pyx_v___pyx_result->speed_of_light = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 12, __pyx_L1_error)
  }
  if (!(likely(PySet_CheckExact(PyTuple_GET_ITEM(__pyx_v___pyx_state, 3)))||((PyTuple_GET_ITEM(__pyx_v___pyx_state, 3)) == Py_None) || __Pyx_RaiseUnexpectedTypeError("set", PyTuple_GET_ITEM(__pyx_v___pyx_state, 3)))) __PYX_ERR(0, 12, __pyx_L1_error)
  __pyx_t_3 = PyTuple_GET_ITEM(__pyx_v___pyx_state, 3);
  __Pyx_INCREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v___pyx_result->state_keys);
  __Pyx_DECREF(__pyx_v___pyx_result->state_keys);
  __pyx_v___pyx_result->state_keys = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_PhysicsSystem__set_state(PhysicsSystem __pyx_result, tuple __pyx_state):
 *     __pyx_result.dimensions = __pyx_state[0]; __pyx_result.resolution = __pyx_state[1]; __pyx_result.speed_of_light = __pyx_state[2]; __pyx_result.state_keys = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[4])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 13, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_6 = (__pyx_t_5 > 4);
  if (__pyx_t_6) {
  } else {
    __pyx_t_4 = __pyx_t_6;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_6 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_t_4 = __pyx_t_6;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "(tree fragment)":14
 *     __pyx_result.dimensions = __pyx_state[0]; __pyx_result.resolution = __pyx_state[1]; __pyx_result.speed_of_light = __pyx_state[2]; __pyx_result.state_keys = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[4])             # <<<<<<<<<<<<<<
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_update); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 14, __pyx_L1_error)
    }
    __pyx_t_7 = NULL;
    __pyx_t_9 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
        __pyx_t_9 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, PyTuple_GET_ITEM(__pyx_v___pyx_state, 4)};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_PhysicsSystem__set_state(PhysicsSystem __pyx_result, tuple __pyx_state):
 *     __pyx_result.dimensions = __pyx_state[0]; __pyx_result.resolution = __pyx_state[1]; __pyx_result.speed_of_light = __pyx_state[2]; __pyx_result.state_keys = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[4])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_PhysicsSystem__set_state(<PhysicsSystem> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_PhysicsSystem__set_state(PhysicsSystem __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.dimensions = __pyx_state[0]; __pyx_result.resolution = __pyx_state[1]; __pyx_result.speed_of_light = __pyx_state[2]; __pyx_result.state_keys = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("flitter.render.physics.__pyx_unpickle_PhysicsSystem__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_tp_new_7flitter_6render_7physics_PhysicsGroup(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_7flitter_6render_7physics_PhysicsGroup *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_7flitter_6render_7physics_PhysicsGroup *)o);
  p->parent = ((struct __pyx_obj_7flitter_6render_7physics_PhysicsGroup *)Py_None); Py_INCREF(Py_None);
  p->particles = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->particle_forces = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->matrix_forces = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->distance_forces = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->barriers = ((PyObject*)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_7flitter_6render_7physics_12PhysicsGroup_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_7flitter_6render_7physics_PhysicsGroup(PyObject *o) {
  struct __pyx_obj_7flitter_6render_7physics_PhysicsGroup *p = (struct __pyx_obj_7flitter_6render_7physics_PhysicsGroup *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_6render_7physics_PhysicsGroup) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->parent);
  Py_CLEAR(p->particles);
  Py_CLEAR(p->particle_forces);
  Py_CLEAR(p->matrix_forces);
  Py_CLEAR(p->distance_forces);
  Py_CLEAR(p->barriers);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_7flitter_6render_7physics_PhysicsGroup(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_6render_7physics_PhysicsGroup *p = (struct __pyx_obj_7flitter_6render_7physics_PhysicsGroup *)o;
  if (p->parent) {
    e = (*v)(((PyObject *)p->parent), a); if (e) return e;
  }
  if (p->particles) {
    e = (*v)(p->particles, a); if (e) return e;
  }
  if (p->particle_forces) {
    e = (*v)(p->particle_forces, a); if (e) return e;
  }
  if (p->matrix_forces) {
    e = (*v)(p->matrix_forces, a); if (e) return e;
  }
  if (p->distance_forces) {
    e = (*v)(p->distance_forces, a); if (e) return e;
  }
  if (p->barriers) {
    e = (*v)(p->barriers, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7flitter_6render_7physics_PhysicsGroup(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7flitter_6render_7physics_PhysicsGroup *p = (struct __pyx_obj_7flitter_6render_7physics_PhysicsGroup *)o;
  tmp = ((PyObject*)p->parent);
  p->parent = ((struct __pyx_obj_7flitter_6render_7physics_PhysicsGroup *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->particles);
  p->particles = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->particle_forces);
  p->particle_forces = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->matrix_forces);
  p->matrix_forces = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->distance_forces);
  p->distance_forces = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->barriers);
  p->barriers = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_7flitter_6render_7physics_PhysicsGroup[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_12PhysicsGroup_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_12PhysicsGroup_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_6render_7physics_PhysicsGroup_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_6render_7physics_PhysicsGroup},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_6render_7physics_PhysicsGroup},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_6render_7physics_PhysicsGroup},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_6render_7physics_PhysicsGroup},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_6render_7physics_PhysicsGroup},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_6render_7physics_PhysicsGroup_spec = {
  "flitter.render.physics.PhysicsGroup",
  sizeof(struct __pyx_obj_7flitter_6render_7physics_PhysicsGroup),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_6render_7physics_PhysicsGroup_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_6render_7physics_PhysicsGroup = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.render.physics.""PhysicsGroup", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_6render_7physics_PhysicsGroup), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_6render_7physics_PhysicsGroup, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_6render_7physics_PhysicsGroup, /*tp_traverse*/
  __pyx_tp_clear_7flitter_6render_7physics_PhysicsGroup, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_6render_7physics_PhysicsGroup, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_6render_7physics_PhysicsGroup, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_6render_7physics_Particle __pyx_vtable_7flitter_6render_7physics_Particle;

static PyObject *__pyx_tp_new_7flitter_6render_7physics_Particle(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_7flitter_6render_7physics_Particle *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  o = (*t->tp_alloc)(t, 0);
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_7flitter_6render_7physics_Particle *)o);
  p->__pyx_vtab = __pyx_vtabptr_7flitter_6render_7physics_Particle;
  p->id = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  p->position_state_key = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  p->position = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  p->velocity_state_key = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  p->velocity = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  p->acceleration = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  p->initial_force = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  p->force = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7flitter_6render_7physics_Particle(PyObject *o) {
  struct __pyx_obj_7flitter_6render_7physics_Particle *p = (struct __pyx_obj_7flitter_6render_7physics_Particle *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->id);
  Py_CLEAR(p->position_state_key);
  Py_CLEAR(p->position);
  Py_CLEAR(p->velocity_state_key);
  Py_CLEAR(p->velocity);
  Py_CLEAR(p->acceleration);
  Py_CLEAR(p->initial_force);
  Py_CLEAR(p->force);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_7flitter_6render_7physics_Particle(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_6render_7physics_Particle *p = (struct __pyx_obj_7flitter_6render_7physics_Particle *)o;
  if (p->id) {
    e = (*v)(((PyObject *)p->id), a); if (e) return e;
  }
  if (p->position_state_key) {
    e = (*v)(((PyObject *)p->position_state_key), a); if (e) return e;
  }
  if (p->position) {
    e = (*v)(((PyObject *)p->position), a); if (e) return e;
  }
  if (p->velocity_state_key) {
    e = (*v)(((PyObject *)p->velocity_state_key), a); if (e) return e;
  }
  if (p->velocity) {
    e = (*v)(((PyObject *)p->velocity), a); if (e) return e;
  }
  if (p->acceleration) {
    e = (*v)(((PyObject *)p->acceleration), a); if (e) return e;
  }
  if (p->initial_force) {
    e = (*v)(((PyObject *)p->initial_force), a); if (e) return e;
  }
  if (p->force) {
    e = (*v)(((PyObject *)p->force), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7flitter_6render_7physics_Particle(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7flitter_6render_7physics_Particle *p = (struct __pyx_obj_7flitter_6render_7physics_Particle *)o;
  tmp = ((PyObject*)p->id);
  p->id = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->position_state_key);
  p->position_state_key = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->position);
  p->position = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->velocity_state_key);
  p->velocity_state_key = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->velocity);
  p->velocity = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->acceleration);
  p->acceleration = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->initial_force);
  p->initial_force = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->force);
  p->force = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_7flitter_6render_7physics_Particle[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_8Particle_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_8Particle_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_6render_7physics_Particle_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_6render_7physics_Particle},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_6render_7physics_Particle},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_6render_7physics_Particle},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_6render_7physics_Particle},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_6render_7physics_Particle},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_6render_7physics_Particle_spec = {
  "flitter.render.physics.Particle",
  sizeof(struct __pyx_obj_7flitter_6render_7physics_Particle),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_6render_7physics_Particle_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_6render_7physics_Particle = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.render.physics.""Particle", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_6render_7physics_Particle), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_6render_7physics_Particle, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_6render_7physics_Particle, /*tp_traverse*/
  __pyx_tp_clear_7flitter_6render_7physics_Particle, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_6render_7physics_Particle, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_6render_7physics_Particle, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_6render_7physics_Barrier __pyx_vtable_7flitter_6render_7physics_Barrier;

static PyObject *__pyx_tp_new_7flitter_6render_7physics_Barrier(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_6render_7physics_Barrier *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  o = (*t->tp_alloc)(t, 0);
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_7flitter_6render_7physics_Barrier *)o);
  p->__pyx_vtab = __pyx_vtabptr_7flitter_6render_7physics_Barrier;
  p->position = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  p->normal = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_7flitter_6render_7physics_7Barrier_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_7flitter_6render_7physics_Barrier(PyObject *o) {
  struct __pyx_obj_7flitter_6render_7physics_Barrier *p = (struct __pyx_obj_7flitter_6render_7physics_Barrier *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->position);
  Py_CLEAR(p->normal);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_7flitter_6render_7physics_Barrier(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_6render_7physics_Barrier *p = (struct __pyx_obj_7flitter_6render_7physics_Barrier *)o;
  if (p->position) {
    e = (*v)(((PyObject *)p->position), a); if (e) return e;
  }
  if (p->normal) {
    e = (*v)(((PyObject *)p->normal), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7flitter_6render_7physics_Barrier(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7flitter_6render_7physics_Barrier *p = (struct __pyx_obj_7flitter_6render_7physics_Barrier *)o;
  tmp = ((PyObject*)p->position);
  p->position = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->normal);
  p->normal = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_7flitter_6render_7physics_Barrier[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_7Barrier_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_7Barrier_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_6render_7physics_Barrier_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_6render_7physics_Barrier},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_6render_7physics_Barrier},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_6render_7physics_Barrier},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_6render_7physics_Barrier},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_6render_7physics_Barrier},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_6render_7physics_Barrier_spec = {
  "flitter.render.physics.Barrier",
  sizeof(struct __pyx_obj_7flitter_6render_7physics_Barrier),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_6render_7physics_Barrier_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_6render_7physics_Barrier = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.render.physics.""Barrier", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_6render_7physics_Barrier), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_6render_7physics_Barrier, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_6render_7physics_Barrier, /*tp_traverse*/
  __pyx_tp_clear_7flitter_6render_7physics_Barrier, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_6render_7physics_Barrier, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_6render_7physics_Barrier, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_6render_7physics_ParticleForceApplier __pyx_vtable_7flitter_6render_7physics_ParticleForceApplier;

static PyObject *__pyx_tp_new_7flitter_6render_7physics_ParticleForceApplier(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_6render_7physics_ParticleForceApplier *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_7flitter_6render_7physics_ParticleForceApplier *)o);
  p->__pyx_vtab = __pyx_vtabptr_7flitter_6render_7physics_ParticleForceApplier;
  if (unlikely(__pyx_pw_7flitter_6render_7physics_20ParticleForceApplier_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_7flitter_6render_7physics_ParticleForceApplier(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_6render_7physics_ParticleForceApplier) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static PyMethodDef __pyx_methods_7flitter_6render_7physics_ParticleForceApplier[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_20ParticleForceApplier_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_20ParticleForceApplier_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_6render_7physics_ParticleForceApplier_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_6render_7physics_ParticleForceApplier},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_6render_7physics_ParticleForceApplier},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_6render_7physics_ParticleForceApplier},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_6render_7physics_ParticleForceApplier_spec = {
  "flitter.render.physics.ParticleForceApplier",
  sizeof(struct __pyx_obj_7flitter_6render_7physics_ParticleForceApplier),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_7flitter_6render_7physics_ParticleForceApplier_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_6render_7physics_ParticleForceApplier = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.render.physics.""ParticleForceApplier", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_6render_7physics_ParticleForceApplier), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_6render_7physics_ParticleForceApplier, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_6render_7physics_ParticleForceApplier, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_6render_7physics_ParticleForceApplier, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_6render_7physics_MatrixForceApplier __pyx_vtable_7flitter_6render_7physics_MatrixForceApplier;

static PyObject *__pyx_tp_new_7flitter_6render_7physics_MatrixForceApplier(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_6render_7physics_MatrixForceApplier *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_7flitter_6render_7physics_MatrixForceApplier *)o);
  p->__pyx_vtab = __pyx_vtabptr_7flitter_6render_7physics_MatrixForceApplier;
  if (unlikely(__pyx_pw_7flitter_6render_7physics_18MatrixForceApplier_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_7flitter_6render_7physics_MatrixForceApplier(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_6render_7physics_MatrixForceApplier) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static PyMethodDef __pyx_methods_7flitter_6render_7physics_MatrixForceApplier[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_18MatrixForceApplier_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_18MatrixForceApplier_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_6render_7physics_MatrixForceApplier_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_6render_7physics_MatrixForceApplier},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_6render_7physics_MatrixForceApplier},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_6render_7physics_MatrixForceApplier},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_6render_7physics_MatrixForceApplier_spec = {
  "flitter.render.physics.MatrixForceApplier",
  sizeof(struct __pyx_obj_7flitter_6render_7physics_MatrixForceApplier),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_7flitter_6render_7physics_MatrixForceApplier_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_6render_7physics_MatrixForceApplier = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.render.physics.""MatrixForceApplier", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_6render_7physics_MatrixForceApplier), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_6render_7physics_MatrixForceApplier, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_6render_7physics_MatrixForceApplier, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_6render_7physics_MatrixForceApplier, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_6render_7physics_DistanceForceApplier __pyx_vtable_7flitter_6render_7physics_DistanceForceApplier;

static PyObject *__pyx_tp_new_7flitter_6render_7physics_DistanceForceApplier(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_7flitter_6render_7physics_DistanceForceApplier *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  o = (*t->tp_alloc)(t, 0);
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_7flitter_6render_7physics_DistanceForceApplier *)o);
  p->__pyx_vtab = __pyx_vtabptr_7flitter_6render_7physics_DistanceForceApplier;
  p->from_particle_id = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  p->to_particle_id = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7flitter_6render_7physics_DistanceForceApplier(PyObject *o) {
  struct __pyx_obj_7flitter_6render_7physics_DistanceForceApplier *p = (struct __pyx_obj_7flitter_6render_7physics_DistanceForceApplier *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->from_particle_id);
  Py_CLEAR(p->to_particle_id);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_7flitter_6render_7physics_DistanceForceApplier(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_6render_7physics_DistanceForceApplier *p = (struct __pyx_obj_7flitter_6render_7physics_DistanceForceApplier *)o;
  if (p->from_particle_id) {
    e = (*v)(((PyObject *)p->from_particle_id), a); if (e) return e;
  }
  if (p->to_particle_id) {
    e = (*v)(((PyObject *)p->to_particle_id), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7flitter_6render_7physics_DistanceForceApplier(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7flitter_6render_7physics_DistanceForceApplier *p = (struct __pyx_obj_7flitter_6render_7physics_DistanceForceApplier *)o;
  tmp = ((PyObject*)p->from_particle_id);
  p->from_particle_id = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->to_particle_id);
  p->to_particle_id = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_7flitter_6render_7physics_DistanceForceApplier[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_20DistanceForceApplier_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_20DistanceForceApplier_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_6render_7physics_DistanceForceApplier_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_6render_7physics_DistanceForceApplier},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_6render_7physics_DistanceForceApplier},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_6render_7physics_DistanceForceApplier},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_6render_7physics_DistanceForceApplier},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_6render_7physics_DistanceForceApplier},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_6render_7physics_DistanceForceApplier_spec = {
  "flitter.render.physics.DistanceForceApplier",
  sizeof(struct __pyx_obj_7flitter_6render_7physics_DistanceForceApplier),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_6render_7physics_DistanceForceApplier_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_6render_7physics_DistanceForceApplier = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.render.physics.""DistanceForceApplier", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_6render_7physics_DistanceForceApplier), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_6render_7physics_DistanceForceApplier, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_6render_7physics_DistanceForceApplier, /*tp_traverse*/
  __pyx_tp_clear_7flitter_6render_7physics_DistanceForceApplier, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_6render_7physics_DistanceForceApplier, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_6render_7physics_DistanceForceApplier, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_6render_7physics_DragForceApplier __pyx_vtable_7flitter_6render_7physics_DragForceApplier;

static PyObject *__pyx_tp_new_7flitter_6render_7physics_DragForceApplier(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_6render_7physics_DragForceApplier *p;
  PyObject *o = __pyx_tp_new_7flitter_6render_7physics_ParticleForceApplier(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_6render_7physics_DragForceApplier *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_6render_7physics_ParticleForceApplier*)__pyx_vtabptr_7flitter_6render_7physics_DragForceApplier;
  p->flow = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_7flitter_6render_7physics_16DragForceApplier_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_7flitter_6render_7physics_DragForceApplier(PyObject *o) {
  struct __pyx_obj_7flitter_6render_7physics_DragForceApplier *p = (struct __pyx_obj_7flitter_6render_7physics_DragForceApplier *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_6render_7physics_DragForceApplier) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->flow);
  __pyx_tp_dealloc_7flitter_6render_7physics_ParticleForceApplier(o);
}

static int __pyx_tp_traverse_7flitter_6render_7physics_DragForceApplier(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_6render_7physics_DragForceApplier *p = (struct __pyx_obj_7flitter_6render_7physics_DragForceApplier *)o;
  e = ((likely(__pyx_ptype_7flitter_6render_7physics_ParticleForceApplier)) ? ((__pyx_ptype_7flitter_6render_7physics_ParticleForceApplier->tp_traverse) ? __pyx_ptype_7flitter_6render_7physics_ParticleForceApplier->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7flitter_6render_7physics_DragForceApplier)); if (e) return e;
  if (p->flow) {
    e = (*v)(((PyObject *)p->flow), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7flitter_6render_7physics_DragForceApplier(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7flitter_6render_7physics_DragForceApplier *p = (struct __pyx_obj_7flitter_6render_7physics_DragForceApplier *)o;
  if (likely(__pyx_ptype_7flitter_6render_7physics_ParticleForceApplier)) { if (__pyx_ptype_7flitter_6render_7physics_ParticleForceApplier->tp_clear) __pyx_ptype_7flitter_6render_7physics_ParticleForceApplier->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7flitter_6render_7physics_DragForceApplier);
  tmp = ((PyObject*)p->flow);
  p->flow = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_7flitter_6render_7physics_DragForceApplier[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_16DragForceApplier_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_16DragForceApplier_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_6render_7physics_DragForceApplier_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_6render_7physics_DragForceApplier},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_6render_7physics_DragForceApplier},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_6render_7physics_DragForceApplier},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_6render_7physics_DragForceApplier},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_6render_7physics_DragForceApplier},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_6render_7physics_DragForceApplier_spec = {
  "flitter.render.physics.DragForceApplier",
  sizeof(struct __pyx_obj_7flitter_6render_7physics_DragForceApplier),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_6render_7physics_DragForceApplier_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_6render_7physics_DragForceApplier = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.render.physics.""DragForceApplier", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_6render_7physics_DragForceApplier), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_6render_7physics_DragForceApplier, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_6render_7physics_DragForceApplier, /*tp_traverse*/
  __pyx_tp_clear_7flitter_6render_7physics_DragForceApplier, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_6render_7physics_DragForceApplier, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_6render_7physics_DragForceApplier, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_6render_7physics_BuoyancyForceApplier __pyx_vtable_7flitter_6render_7physics_BuoyancyForceApplier;

static PyObject *__pyx_tp_new_7flitter_6render_7physics_BuoyancyForceApplier(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_6render_7physics_BuoyancyForceApplier *p;
  PyObject *o = __pyx_tp_new_7flitter_6render_7physics_ParticleForceApplier(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_6render_7physics_BuoyancyForceApplier *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_6render_7physics_ParticleForceApplier*)__pyx_vtabptr_7flitter_6render_7physics_BuoyancyForceApplier;
  p->gravity = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_7flitter_6render_7physics_20BuoyancyForceApplier_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_7flitter_6render_7physics_BuoyancyForceApplier(PyObject *o) {
  struct __pyx_obj_7flitter_6render_7physics_BuoyancyForceApplier *p = (struct __pyx_obj_7flitter_6render_7physics_BuoyancyForceApplier *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_6render_7physics_BuoyancyForceApplier) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->gravity);
  __pyx_tp_dealloc_7flitter_6render_7physics_ParticleForceApplier(o);
}

static int __pyx_tp_traverse_7flitter_6render_7physics_BuoyancyForceApplier(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_6render_7physics_BuoyancyForceApplier *p = (struct __pyx_obj_7flitter_6render_7physics_BuoyancyForceApplier *)o;
  e = ((likely(__pyx_ptype_7flitter_6render_7physics_ParticleForceApplier)) ? ((__pyx_ptype_7flitter_6render_7physics_ParticleForceApplier->tp_traverse) ? __pyx_ptype_7flitter_6render_7physics_ParticleForceApplier->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7flitter_6render_7physics_BuoyancyForceApplier)); if (e) return e;
  if (p->gravity) {
    e = (*v)(((PyObject *)p->gravity), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7flitter_6render_7physics_BuoyancyForceApplier(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7flitter_6render_7physics_BuoyancyForceApplier *p = (struct __pyx_obj_7flitter_6render_7physics_BuoyancyForceApplier *)o;
  if (likely(__pyx_ptype_7flitter_6render_7physics_ParticleForceApplier)) { if (__pyx_ptype_7flitter_6render_7physics_ParticleForceApplier->tp_clear) __pyx_ptype_7flitter_6render_7physics_ParticleForceApplier->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7flitter_6render_7physics_BuoyancyForceApplier);
  tmp = ((PyObject*)p->gravity);
  p->gravity = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_7flitter_6render_7physics_BuoyancyForceApplier[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_20BuoyancyForceApplier_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_20BuoyancyForceApplier_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_6render_7physics_BuoyancyForceApplier_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_6render_7physics_BuoyancyForceApplier},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_6render_7physics_BuoyancyForceApplier},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_6render_7physics_BuoyancyForceApplier},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_6render_7physics_BuoyancyForceApplier},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_6render_7physics_BuoyancyForceApplier},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_6render_7physics_BuoyancyForceApplier_spec = {
  "flitter.render.physics.BuoyancyForceApplier",
  sizeof(struct __pyx_obj_7flitter_6render_7physics_BuoyancyForceApplier),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_6render_7physics_BuoyancyForceApplier_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_6render_7physics_BuoyancyForceApplier = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.render.physics.""BuoyancyForceApplier", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_6render_7physics_BuoyancyForceApplier), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_6render_7physics_BuoyancyForceApplier, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_6render_7physics_BuoyancyForceApplier, /*tp_traverse*/
  __pyx_tp_clear_7flitter_6render_7physics_BuoyancyForceApplier, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_6render_7physics_BuoyancyForceApplier, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_6render_7physics_BuoyancyForceApplier, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_6render_7physics_ConstantForceApplier __pyx_vtable_7flitter_6render_7physics_ConstantForceApplier;

static PyObject *__pyx_tp_new_7flitter_6render_7physics_ConstantForceApplier(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_6render_7physics_ConstantForceApplier *p;
  PyObject *o = __pyx_tp_new_7flitter_6render_7physics_ParticleForceApplier(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_6render_7physics_ConstantForceApplier *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_6render_7physics_ParticleForceApplier*)__pyx_vtabptr_7flitter_6render_7physics_ConstantForceApplier;
  p->force = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  p->acceleration = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_7flitter_6render_7physics_20ConstantForceApplier_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_7flitter_6render_7physics_ConstantForceApplier(PyObject *o) {
  struct __pyx_obj_7flitter_6render_7physics_ConstantForceApplier *p = (struct __pyx_obj_7flitter_6render_7physics_ConstantForceApplier *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_6render_7physics_ConstantForceApplier) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->force);
  Py_CLEAR(p->acceleration);
  __pyx_tp_dealloc_7flitter_6render_7physics_ParticleForceApplier(o);
}

static int __pyx_tp_traverse_7flitter_6render_7physics_ConstantForceApplier(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_6render_7physics_ConstantForceApplier *p = (struct __pyx_obj_7flitter_6render_7physics_ConstantForceApplier *)o;
  e = ((likely(__pyx_ptype_7flitter_6render_7physics_ParticleForceApplier)) ? ((__pyx_ptype_7flitter_6render_7physics_ParticleForceApplier->tp_traverse) ? __pyx_ptype_7flitter_6render_7physics_ParticleForceApplier->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7flitter_6render_7physics_ConstantForceApplier)); if (e) return e;
  if (p->force) {
    e = (*v)(((PyObject *)p->force), a); if (e) return e;
  }
  if (p->acceleration) {
    e = (*v)(((PyObject *)p->acceleration), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7flitter_6render_7physics_ConstantForceApplier(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7flitter_6render_7physics_ConstantForceApplier *p = (struct __pyx_obj_7flitter_6render_7physics_ConstantForceApplier *)o;
  if (likely(__pyx_ptype_7flitter_6render_7physics_ParticleForceApplier)) { if (__pyx_ptype_7flitter_6render_7physics_ParticleForceApplier->tp_clear) __pyx_ptype_7flitter_6render_7physics_ParticleForceApplier->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7flitter_6render_7physics_ConstantForceApplier);
  tmp = ((PyObject*)p->force);
  p->force = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->acceleration);
  p->acceleration = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_7flitter_6render_7physics_ConstantForceApplier[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_20ConstantForceApplier_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_20ConstantForceApplier_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_6render_7physics_ConstantForceApplier_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_6render_7physics_ConstantForceApplier},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_6render_7physics_ConstantForceApplier},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_6render_7physics_ConstantForceApplier},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_6render_7physics_ConstantForceApplier},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_6render_7physics_ConstantForceApplier},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_6render_7physics_ConstantForceApplier_spec = {
  "flitter.render.physics.ConstantForceApplier",
  sizeof(struct __pyx_obj_7flitter_6render_7physics_ConstantForceApplier),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_6render_7physics_ConstantForceApplier_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_6render_7physics_ConstantForceApplier = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.render.physics.""ConstantForceApplier", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_6render_7physics_ConstantForceApplier), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_6render_7physics_ConstantForceApplier, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_6render_7physics_ConstantForceApplier, /*tp_traverse*/
  __pyx_tp_clear_7flitter_6render_7physics_ConstantForceApplier, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_6render_7physics_ConstantForceApplier, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_6render_7physics_ConstantForceApplier, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_6render_7physics_RandomForceApplier __pyx_vtable_7flitter_6render_7physics_RandomForceApplier;

static PyObject *__pyx_tp_new_7flitter_6render_7physics_RandomForceApplier(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_6render_7physics_RandomForceApplier *p;
  PyObject *o = __pyx_tp_new_7flitter_6render_7physics_ParticleForceApplier(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_6render_7physics_RandomForceApplier *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_6render_7physics_ParticleForceApplier*)__pyx_vtabptr_7flitter_6render_7physics_RandomForceApplier;
  p->random_source = ((struct __pyx_obj_7flitter_8language_9functions_normal *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_7flitter_6render_7physics_18RandomForceApplier_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_7flitter_6render_7physics_RandomForceApplier(PyObject *o) {
  struct __pyx_obj_7flitter_6render_7physics_RandomForceApplier *p = (struct __pyx_obj_7flitter_6render_7physics_RandomForceApplier *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_6render_7physics_RandomForceApplier) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->random_source);
  __pyx_tp_dealloc_7flitter_6render_7physics_ParticleForceApplier(o);
}

static int __pyx_tp_traverse_7flitter_6render_7physics_RandomForceApplier(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_6render_7physics_RandomForceApplier *p = (struct __pyx_obj_7flitter_6render_7physics_RandomForceApplier *)o;
  e = ((likely(__pyx_ptype_7flitter_6render_7physics_ParticleForceApplier)) ? ((__pyx_ptype_7flitter_6render_7physics_ParticleForceApplier->tp_traverse) ? __pyx_ptype_7flitter_6render_7physics_ParticleForceApplier->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7flitter_6render_7physics_RandomForceApplier)); if (e) return e;
  if (p->random_source) {
    e = (*v)(((PyObject *)p->random_source), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7flitter_6render_7physics_RandomForceApplier(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7flitter_6render_7physics_RandomForceApplier *p = (struct __pyx_obj_7flitter_6render_7physics_RandomForceApplier *)o;
  if (likely(__pyx_ptype_7flitter_6render_7physics_ParticleForceApplier)) { if (__pyx_ptype_7flitter_6render_7physics_ParticleForceApplier->tp_clear) __pyx_ptype_7flitter_6render_7physics_ParticleForceApplier->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7flitter_6render_7physics_RandomForceApplier);
  tmp = ((PyObject*)p->random_source);
  p->random_source = ((struct __pyx_obj_7flitter_8language_9functions_normal *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_7flitter_6render_7physics_RandomForceApplier[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_18RandomForceApplier_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_18RandomForceApplier_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_6render_7physics_RandomForceApplier_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_6render_7physics_RandomForceApplier},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_6render_7physics_RandomForceApplier},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_6render_7physics_RandomForceApplier},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_6render_7physics_RandomForceApplier},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_6render_7physics_RandomForceApplier},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_6render_7physics_RandomForceApplier_spec = {
  "flitter.render.physics.RandomForceApplier",
  sizeof(struct __pyx_obj_7flitter_6render_7physics_RandomForceApplier),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_6render_7physics_RandomForceApplier_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_6render_7physics_RandomForceApplier = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.render.physics.""RandomForceApplier", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_6render_7physics_RandomForceApplier), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_6render_7physics_RandomForceApplier, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_6render_7physics_RandomForceApplier, /*tp_traverse*/
  __pyx_tp_clear_7flitter_6render_7physics_RandomForceApplier, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_6render_7physics_RandomForceApplier, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_6render_7physics_RandomForceApplier, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_6render_7physics_FieldForceApplier __pyx_vtable_7flitter_6render_7physics_FieldForceApplier;

static PyObject *__pyx_tp_new_7flitter_6render_7physics_FieldForceApplier(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_6render_7physics_FieldForceApplier *p;
  PyObject *o = __pyx_tp_new_7flitter_6render_7physics_ParticleForceApplier(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_6render_7physics_FieldForceApplier *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_6render_7physics_ParticleForceApplier*)__pyx_vtabptr_7flitter_6render_7physics_FieldForceApplier;
  p->direction = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_7flitter_6render_7physics_17FieldForceApplier_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_7flitter_6render_7physics_FieldForceApplier(PyObject *o) {
  struct __pyx_obj_7flitter_6render_7physics_FieldForceApplier *p = (struct __pyx_obj_7flitter_6render_7physics_FieldForceApplier *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_6render_7physics_FieldForceApplier) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->direction);
  __pyx_tp_dealloc_7flitter_6render_7physics_ParticleForceApplier(o);
}

static int __pyx_tp_traverse_7flitter_6render_7physics_FieldForceApplier(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_6render_7physics_FieldForceApplier *p = (struct __pyx_obj_7flitter_6render_7physics_FieldForceApplier *)o;
  e = ((likely(__pyx_ptype_7flitter_6render_7physics_ParticleForceApplier)) ? ((__pyx_ptype_7flitter_6render_7physics_ParticleForceApplier->tp_traverse) ? __pyx_ptype_7flitter_6render_7physics_ParticleForceApplier->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7flitter_6render_7physics_FieldForceApplier)); if (e) return e;
  if (p->direction) {
    e = (*v)(((PyObject *)p->direction), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7flitter_6render_7physics_FieldForceApplier(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7flitter_6render_7physics_FieldForceApplier *p = (struct __pyx_obj_7flitter_6render_7physics_FieldForceApplier *)o;
  if (likely(__pyx_ptype_7flitter_6render_7physics_ParticleForceApplier)) { if (__pyx_ptype_7flitter_6render_7physics_ParticleForceApplier->tp_clear) __pyx_ptype_7flitter_6render_7physics_ParticleForceApplier->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7flitter_6render_7physics_FieldForceApplier);
  tmp = ((PyObject*)p->direction);
  p->direction = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_7flitter_6render_7physics_FieldForceApplier[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_17FieldForceApplier_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_17FieldForceApplier_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_6render_7physics_FieldForceApplier_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_6render_7physics_FieldForceApplier},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_6render_7physics_FieldForceApplier},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_6render_7physics_FieldForceApplier},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_6render_7physics_FieldForceApplier},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_6render_7physics_FieldForceApplier},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_6render_7physics_FieldForceApplier_spec = {
  "flitter.render.physics.FieldForceApplier",
  sizeof(struct __pyx_obj_7flitter_6render_7physics_FieldForceApplier),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_6render_7physics_FieldForceApplier_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_6render_7physics_FieldForceApplier = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.render.physics.""FieldForceApplier", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_6render_7physics_FieldForceApplier), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_6render_7physics_FieldForceApplier, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_6render_7physics_FieldForceApplier, /*tp_traverse*/
  __pyx_tp_clear_7flitter_6render_7physics_FieldForceApplier, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_6render_7physics_FieldForceApplier, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_6render_7physics_FieldForceApplier, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_6render_7physics_CollisionForceApplier __pyx_vtable_7flitter_6render_7physics_CollisionForceApplier;

static PyObject *__pyx_tp_new_7flitter_6render_7physics_CollisionForceApplier(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_6render_7physics_CollisionForceApplier *p;
  PyObject *o = __pyx_tp_new_7flitter_6render_7physics_MatrixForceApplier(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_6render_7physics_CollisionForceApplier *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_6render_7physics_MatrixForceApplier*)__pyx_vtabptr_7flitter_6render_7physics_CollisionForceApplier;
  if (unlikely(__pyx_pw_7flitter_6render_7physics_21CollisionForceApplier_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static PyMethodDef __pyx_methods_7flitter_6render_7physics_CollisionForceApplier[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_21CollisionForceApplier_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_21CollisionForceApplier_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_6render_7physics_CollisionForceApplier_slots[] = {
  {Py_tp_methods, (void *)__pyx_methods_7flitter_6render_7physics_CollisionForceApplier},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_6render_7physics_CollisionForceApplier},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_6render_7physics_CollisionForceApplier_spec = {
  "flitter.render.physics.CollisionForceApplier",
  sizeof(struct __pyx_obj_7flitter_6render_7physics_CollisionForceApplier),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_7flitter_6render_7physics_CollisionForceApplier_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_6render_7physics_CollisionForceApplier = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.render.physics.""CollisionForceApplier", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_6render_7physics_CollisionForceApplier), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_6render_7physics_MatrixForceApplier, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_6render_7physics_CollisionForceApplier, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_6render_7physics_CollisionForceApplier, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_6render_7physics_GravityForceApplier __pyx_vtable_7flitter_6render_7physics_GravityForceApplier;

static PyObject *__pyx_tp_new_7flitter_6render_7physics_GravityForceApplier(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_6render_7physics_GravityForceApplier *p;
  PyObject *o = __pyx_tp_new_7flitter_6render_7physics_MatrixForceApplier(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_6render_7physics_GravityForceApplier *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_6render_7physics_MatrixForceApplier*)__pyx_vtabptr_7flitter_6render_7physics_GravityForceApplier;
  return o;
}

static PyMethodDef __pyx_methods_7flitter_6render_7physics_GravityForceApplier[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_19GravityForceApplier_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_19GravityForceApplier_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_6render_7physics_GravityForceApplier_slots[] = {
  {Py_tp_methods, (void *)__pyx_methods_7flitter_6render_7physics_GravityForceApplier},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_6render_7physics_GravityForceApplier},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_6render_7physics_GravityForceApplier_spec = {
  "flitter.render.physics.GravityForceApplier",
  sizeof(struct __pyx_obj_7flitter_6render_7physics_GravityForceApplier),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_7flitter_6render_7physics_GravityForceApplier_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_6render_7physics_GravityForceApplier = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.render.physics.""GravityForceApplier", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_6render_7physics_GravityForceApplier), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_6render_7physics_MatrixForceApplier, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_6render_7physics_GravityForceApplier, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_6render_7physics_GravityForceApplier, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_6render_7physics_ElectrostaticForceApplier __pyx_vtable_7flitter_6render_7physics_ElectrostaticForceApplier;

static PyObject *__pyx_tp_new_7flitter_6render_7physics_ElectrostaticForceApplier(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_6render_7physics_ElectrostaticForceApplier *p;
  PyObject *o = __pyx_tp_new_7flitter_6render_7physics_MatrixForceApplier(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_6render_7physics_ElectrostaticForceApplier *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_6render_7physics_MatrixForceApplier*)__pyx_vtabptr_7flitter_6render_7physics_ElectrostaticForceApplier;
  return o;
}

static PyMethodDef __pyx_methods_7flitter_6render_7physics_ElectrostaticForceApplier[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_25ElectrostaticForceApplier_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_25ElectrostaticForceApplier_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_6render_7physics_ElectrostaticForceApplier_slots[] = {
  {Py_tp_methods, (void *)__pyx_methods_7flitter_6render_7physics_ElectrostaticForceApplier},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_6render_7physics_ElectrostaticForceApplier},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_6render_7physics_ElectrostaticForceApplier_spec = {
  "flitter.render.physics.ElectrostaticForceApplier",
  sizeof(struct __pyx_obj_7flitter_6render_7physics_ElectrostaticForceApplier),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_7flitter_6render_7physics_ElectrostaticForceApplier_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_6render_7physics_ElectrostaticForceApplier = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.render.physics.""ElectrostaticForceApplier", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_6render_7physics_ElectrostaticForceApplier), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_6render_7physics_MatrixForceApplier, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_6render_7physics_ElectrostaticForceApplier, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_6render_7physics_ElectrostaticForceApplier, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_6render_7physics_AdhesionForceApplier __pyx_vtable_7flitter_6render_7physics_AdhesionForceApplier;

static PyObject *__pyx_tp_new_7flitter_6render_7physics_AdhesionForceApplier(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7flitter_6render_7physics_AdhesionForceApplier *p;
  PyObject *o = __pyx_tp_new_7flitter_6render_7physics_MatrixForceApplier(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7flitter_6render_7physics_AdhesionForceApplier *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7flitter_6render_7physics_MatrixForceApplier*)__pyx_vtabptr_7flitter_6render_7physics_AdhesionForceApplier;
  if (unlikely(__pyx_pw_7flitter_6render_7physics_20AdhesionForceApplier_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static PyMethodDef __pyx_methods_7flitter_6render_7physics_AdhesionForceApplier[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_20AdhesionForceApplier_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_20AdhesionForceApplier_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_6render_7physics_AdhesionForceApplier_slots[] = {
  {Py_tp_methods, (void *)__pyx_methods_7flitter_6render_7physics_AdhesionForceApplier},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_6render_7physics_AdhesionForceApplier},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_6render_7physics_AdhesionForceApplier_spec = {
  "flitter.render.physics.AdhesionForceApplier",
  sizeof(struct __pyx_obj_7flitter_6render_7physics_AdhesionForceApplier),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_7flitter_6render_7physics_AdhesionForceApplier_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_6render_7physics_AdhesionForceApplier = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.render.physics.""AdhesionForceApplier", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_6render_7physics_AdhesionForceApplier), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_6render_7physics_MatrixForceApplier, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_6render_7physics_AdhesionForceApplier, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_6render_7physics_AdhesionForceApplier, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7flitter_6render_7physics_PhysicsSystem __pyx_vtable_7flitter_6render_7physics_PhysicsSystem;

static PyObject *__pyx_tp_new_7flitter_6render_7physics_PhysicsSystem(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem *)o);
  p->__pyx_vtab = __pyx_vtabptr_7flitter_6render_7physics_PhysicsSystem;
  p->state_keys = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7flitter_6render_7physics_PhysicsSystem(PyObject *o) {
  struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem *p = (struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_6render_7physics_PhysicsSystem) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->state_keys);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_7flitter_6render_7physics_PhysicsSystem(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem *p = (struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem *)o;
  if (p->state_keys) {
    e = (*v)(p->state_keys, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7flitter_6render_7physics_PhysicsSystem(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem *p = (struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem *)o;
  tmp = ((PyObject*)p->state_keys);
  p->state_keys = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_7flitter_6render_7physics_PhysicsSystem[] = {
  {"destroy", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_13PhysicsSystem_3destroy, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"purge", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_13PhysicsSystem_6purge, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"update", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_13PhysicsSystem_9update, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_13PhysicsSystem_14__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7flitter_6render_7physics_13PhysicsSystem_16__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_6render_7physics_PhysicsSystem_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_6render_7physics_PhysicsSystem},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_6render_7physics_PhysicsSystem},
  {Py_tp_clear, (void *)__pyx_tp_clear_7flitter_6render_7physics_PhysicsSystem},
  {Py_tp_methods, (void *)__pyx_methods_7flitter_6render_7physics_PhysicsSystem},
  {Py_tp_init, (void *)__pyx_pw_7flitter_6render_7physics_13PhysicsSystem_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_6render_7physics_PhysicsSystem},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_6render_7physics_PhysicsSystem_spec = {
  "flitter.render.physics.PhysicsSystem",
  sizeof(struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7flitter_6render_7physics_PhysicsSystem_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_6render_7physics_PhysicsSystem = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.render.physics.""PhysicsSystem", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_6render_7physics_PhysicsSystem, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_6render_7physics_PhysicsSystem, /*tp_traverse*/
  __pyx_tp_clear_7flitter_6render_7physics_PhysicsSystem, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7flitter_6render_7physics_PhysicsSystem, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7flitter_6render_7physics_13PhysicsSystem_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_6render_7physics_PhysicsSystem, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct__destroy *__pyx_freelist_7flitter_6render_7physics___pyx_scope_struct__destroy[8];
static int __pyx_freecount_7flitter_6render_7physics___pyx_scope_struct__destroy = 0;
#endif

static PyObject *__pyx_tp_new_7flitter_6render_7physics___pyx_scope_struct__destroy(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_7flitter_6render_7physics___pyx_scope_struct__destroy > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct__destroy)))) {
    o = (PyObject*)__pyx_freelist_7flitter_6render_7physics___pyx_scope_struct__destroy[--__pyx_freecount_7flitter_6render_7physics___pyx_scope_struct__destroy];
    memset(o, 0, sizeof(struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct__destroy));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_7flitter_6render_7physics___pyx_scope_struct__destroy(PyObject *o) {
  struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct__destroy *p = (struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct__destroy *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_6render_7physics___pyx_scope_struct__destroy) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_7flitter_6render_7physics___pyx_scope_struct__destroy < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct__destroy)))) {
    __pyx_freelist_7flitter_6render_7physics___pyx_scope_struct__destroy[__pyx_freecount_7flitter_6render_7physics___pyx_scope_struct__destroy++] = ((struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct__destroy *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_7flitter_6render_7physics___pyx_scope_struct__destroy(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct__destroy *p = (struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct__destroy *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_6render_7physics___pyx_scope_struct__destroy_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_6render_7physics___pyx_scope_struct__destroy},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_6render_7physics___pyx_scope_struct__destroy},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_6render_7physics___pyx_scope_struct__destroy},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_6render_7physics___pyx_scope_struct__destroy_spec = {
  "flitter.render.physics.__pyx_scope_struct__destroy",
  sizeof(struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct__destroy),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_7flitter_6render_7physics___pyx_scope_struct__destroy_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_6render_7physics___pyx_scope_struct__destroy = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.render.physics.""__pyx_scope_struct__destroy", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct__destroy), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_6render_7physics___pyx_scope_struct__destroy, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_6render_7physics___pyx_scope_struct__destroy, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_6render_7physics___pyx_scope_struct__destroy, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct_1_purge *__pyx_freelist_7flitter_6render_7physics___pyx_scope_struct_1_purge[8];
static int __pyx_freecount_7flitter_6render_7physics___pyx_scope_struct_1_purge = 0;
#endif

static PyObject *__pyx_tp_new_7flitter_6render_7physics___pyx_scope_struct_1_purge(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_7flitter_6render_7physics___pyx_scope_struct_1_purge > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct_1_purge)))) {
    o = (PyObject*)__pyx_freelist_7flitter_6render_7physics___pyx_scope_struct_1_purge[--__pyx_freecount_7flitter_6render_7physics___pyx_scope_struct_1_purge];
    memset(o, 0, sizeof(struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct_1_purge));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_7flitter_6render_7physics___pyx_scope_struct_1_purge(PyObject *o) {
  struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct_1_purge *p = (struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct_1_purge *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_6render_7physics___pyx_scope_struct_1_purge) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_7flitter_6render_7physics___pyx_scope_struct_1_purge < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct_1_purge)))) {
    __pyx_freelist_7flitter_6render_7physics___pyx_scope_struct_1_purge[__pyx_freecount_7flitter_6render_7physics___pyx_scope_struct_1_purge++] = ((struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct_1_purge *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_7flitter_6render_7physics___pyx_scope_struct_1_purge(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct_1_purge *p = (struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct_1_purge *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_6render_7physics___pyx_scope_struct_1_purge_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_6render_7physics___pyx_scope_struct_1_purge},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_6render_7physics___pyx_scope_struct_1_purge},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_6render_7physics___pyx_scope_struct_1_purge},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_6render_7physics___pyx_scope_struct_1_purge_spec = {
  "flitter.render.physics.__pyx_scope_struct_1_purge",
  sizeof(struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct_1_purge),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_7flitter_6render_7physics___pyx_scope_struct_1_purge_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_6render_7physics___pyx_scope_struct_1_purge = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.render.physics.""__pyx_scope_struct_1_purge", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct_1_purge), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_6render_7physics___pyx_scope_struct_1_purge, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_6render_7physics___pyx_scope_struct_1_purge, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_6render_7physics___pyx_scope_struct_1_purge, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct_2_update *__pyx_freelist_7flitter_6render_7physics___pyx_scope_struct_2_update[8];
static int __pyx_freecount_7flitter_6render_7physics___pyx_scope_struct_2_update = 0;
#endif

static PyObject *__pyx_tp_new_7flitter_6render_7physics___pyx_scope_struct_2_update(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_7flitter_6render_7physics___pyx_scope_struct_2_update > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct_2_update)))) {
    o = (PyObject*)__pyx_freelist_7flitter_6render_7physics___pyx_scope_struct_2_update[--__pyx_freecount_7flitter_6render_7physics___pyx_scope_struct_2_update];
    memset(o, 0, sizeof(struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct_2_update));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_7flitter_6render_7physics___pyx_scope_struct_2_update(PyObject *o) {
  struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct_2_update *p = (struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct_2_update *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7flitter_6render_7physics___pyx_scope_struct_2_update) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_clock);
  Py_CLEAR(p->__pyx_v_engine);
  Py_CLEAR(p->__pyx_v_group);
  Py_CLEAR(p->__pyx_v_groups);
  Py_CLEAR(p->__pyx_v_iteration_vector);
  Py_CLEAR(p->__pyx_v_kwargs);
  Py_CLEAR(p->__pyx_v_last_run_vector);
  Py_CLEAR(p->__pyx_v_new_state_keys);
  Py_CLEAR(p->__pyx_v_node);
  Py_CLEAR(p->__pyx_v_old_state_keys);
  Py_CLEAR(p->__pyx_v_particle);
  Py_CLEAR(p->__pyx_v_particles_by_id);
  Py_CLEAR(p->__pyx_v_run_vector);
  Py_CLEAR(p->__pyx_v_seed);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_state);
  Py_CLEAR(p->__pyx_v_state_key);
  Py_CLEAR(p->__pyx_v_state_prefix);
  Py_CLEAR(p->__pyx_v_time_vector);
  Py_CLEAR(p->__pyx_v_top_group);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_7flitter_6render_7physics___pyx_scope_struct_2_update < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct_2_update)))) {
    __pyx_freelist_7flitter_6render_7physics___pyx_scope_struct_2_update[__pyx_freecount_7flitter_6render_7physics___pyx_scope_struct_2_update++] = ((struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct_2_update *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_7flitter_6render_7physics___pyx_scope_struct_2_update(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct_2_update *p = (struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct_2_update *)o;
  if (p->__pyx_v_clock) {
    e = (*v)(p->__pyx_v_clock, a); if (e) return e;
  }
  if (p->__pyx_v_engine) {
    e = (*v)(p->__pyx_v_engine, a); if (e) return e;
  }
  if (p->__pyx_v_group) {
    e = (*v)(((PyObject *)p->__pyx_v_group), a); if (e) return e;
  }
  if (p->__pyx_v_groups) {
    e = (*v)(p->__pyx_v_groups, a); if (e) return e;
  }
  if (p->__pyx_v_iteration_vector) {
    e = (*v)(((PyObject *)p->__pyx_v_iteration_vector), a); if (e) return e;
  }
  if (p->__pyx_v_kwargs) {
    e = (*v)(p->__pyx_v_kwargs, a); if (e) return e;
  }
  if (p->__pyx_v_last_run_vector) {
    e = (*v)(((PyObject *)p->__pyx_v_last_run_vector), a); if (e) return e;
  }
  if (p->__pyx_v_new_state_keys) {
    e = (*v)(p->__pyx_v_new_state_keys, a); if (e) return e;
  }
  if (p->__pyx_v_node) {
    e = (*v)(((PyObject *)p->__pyx_v_node), a); if (e) return e;
  }
  if (p->__pyx_v_old_state_keys) {
    e = (*v)(p->__pyx_v_old_state_keys, a); if (e) return e;
  }
  if (p->__pyx_v_particle) {
    e = (*v)(((PyObject *)p->__pyx_v_particle), a); if (e) return e;
  }
  if (p->__pyx_v_particles_by_id) {
    e = (*v)(p->__pyx_v_particles_by_id, a); if (e) return e;
  }
  if (p->__pyx_v_run_vector) {
    e = (*v)(((PyObject *)p->__pyx_v_run_vector), a); if (e) return e;
  }
  if (p->__pyx_v_seed) {
    e = (*v)(((PyObject *)p->__pyx_v_seed), a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_state) {
    e = (*v)(((PyObject *)p->__pyx_v_state), a); if (e) return e;
  }
  if (p->__pyx_v_state_key) {
    e = (*v)(((PyObject *)p->__pyx_v_state_key), a); if (e) return e;
  }
  if (p->__pyx_v_state_prefix) {
    e = (*v)(((PyObject *)p->__pyx_v_state_prefix), a); if (e) return e;
  }
  if (p->__pyx_v_time_vector) {
    e = (*v)(((PyObject *)p->__pyx_v_time_vector), a); if (e) return e;
  }
  if (p->__pyx_v_top_group) {
    e = (*v)(((PyObject *)p->__pyx_v_top_group), a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7flitter_6render_7physics___pyx_scope_struct_2_update_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7flitter_6render_7physics___pyx_scope_struct_2_update},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7flitter_6render_7physics___pyx_scope_struct_2_update},
  {Py_tp_new, (void *)__pyx_tp_new_7flitter_6render_7physics___pyx_scope_struct_2_update},
  {0, 0},
};
static PyType_Spec __pyx_type_7flitter_6render_7physics___pyx_scope_struct_2_update_spec = {
  "flitter.render.physics.__pyx_scope_struct_2_update",
  sizeof(struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct_2_update),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_7flitter_6render_7physics___pyx_scope_struct_2_update_slots,
};
#else

static PyTypeObject __pyx_type_7flitter_6render_7physics___pyx_scope_struct_2_update = {
  PyVarObject_HEAD_INIT(0, 0)
  "flitter.render.physics.""__pyx_scope_struct_2_update", /*tp_name*/
  sizeof(struct __pyx_obj_7flitter_6render_7physics___pyx_scope_struct_2_update), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7flitter_6render_7physics___pyx_scope_struct_2_update, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7flitter_6render_7physics___pyx_scope_struct_2_update, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7flitter_6render_7physics___pyx_scope_struct_2_update, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif
/* #### Code section: pystring_table ### */

static int __Pyx_CreateStringTabAndInitStrings(void) {
  __Pyx_StringTabEntry __pyx_string_tab[] = {
    {&__pyx_n_s_AdhesionForceApplier, __pyx_k_AdhesionForceApplier, sizeof(__pyx_k_AdhesionForceApplier), 0, 0, 1, 1},
    {&__pyx_n_s_AdhesionForceApplier___reduce_cy, __pyx_k_AdhesionForceApplier___reduce_cy, sizeof(__pyx_k_AdhesionForceApplier___reduce_cy), 0, 0, 1, 1},
    {&__pyx_n_s_AdhesionForceApplier___setstate, __pyx_k_AdhesionForceApplier___setstate, sizeof(__pyx_k_AdhesionForceApplier___setstate), 0, 0, 1, 1},
    {&__pyx_n_s_Barrier, __pyx_k_Barrier, sizeof(__pyx_k_Barrier), 0, 0, 1, 1},
    {&__pyx_n_s_Barrier___reduce_cython, __pyx_k_Barrier___reduce_cython, sizeof(__pyx_k_Barrier___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Barrier___setstate_cython, __pyx_k_Barrier___setstate_cython, sizeof(__pyx_k_Barrier___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_BuoyancyForceApplier, __pyx_k_BuoyancyForceApplier, sizeof(__pyx_k_BuoyancyForceApplier), 0, 0, 1, 1},
    {&__pyx_n_s_BuoyancyForceApplier___reduce_cy, __pyx_k_BuoyancyForceApplier___reduce_cy, sizeof(__pyx_k_BuoyancyForceApplier___reduce_cy), 0, 0, 1, 1},
    {&__pyx_n_s_BuoyancyForceApplier___setstate, __pyx_k_BuoyancyForceApplier___setstate, sizeof(__pyx_k_BuoyancyForceApplier___setstate), 0, 0, 1, 1},
    {&__pyx_n_s_CollisionForceApplier, __pyx_k_CollisionForceApplier, sizeof(__pyx_k_CollisionForceApplier), 0, 0, 1, 1},
    {&__pyx_n_s_CollisionForceApplier___reduce_c, __pyx_k_CollisionForceApplier___reduce_c, sizeof(__pyx_k_CollisionForceApplier___reduce_c), 0, 0, 1, 1},
    {&__pyx_n_s_CollisionForceApplier___setstate, __pyx_k_CollisionForceApplier___setstate, sizeof(__pyx_k_CollisionForceApplier___setstate), 0, 0, 1, 1},
    {&__pyx_n_s_ConstantForceApplier, __pyx_k_ConstantForceApplier, sizeof(__pyx_k_ConstantForceApplier), 0, 0, 1, 1},
    {&__pyx_n_s_ConstantForceApplier___reduce_cy, __pyx_k_ConstantForceApplier___reduce_cy, sizeof(__pyx_k_ConstantForceApplier___reduce_cy), 0, 0, 1, 1},
    {&__pyx_n_s_ConstantForceApplier___setstate, __pyx_k_ConstantForceApplier___setstate, sizeof(__pyx_k_ConstantForceApplier___setstate), 0, 0, 1, 1},
    {&__pyx_n_s_DistanceForceApplier, __pyx_k_DistanceForceApplier, sizeof(__pyx_k_DistanceForceApplier), 0, 0, 1, 1},
    {&__pyx_n_s_DistanceForceApplier___reduce_cy, __pyx_k_DistanceForceApplier___reduce_cy, sizeof(__pyx_k_DistanceForceApplier___reduce_cy), 0, 0, 1, 1},
    {&__pyx_n_s_DistanceForceApplier___setstate, __pyx_k_DistanceForceApplier___setstate, sizeof(__pyx_k_DistanceForceApplier___setstate), 0, 0, 1, 1},
    {&__pyx_n_s_DragForceApplier, __pyx_k_DragForceApplier, sizeof(__pyx_k_DragForceApplier), 0, 0, 1, 1},
    {&__pyx_n_s_DragForceApplier___reduce_cython, __pyx_k_DragForceApplier___reduce_cython, sizeof(__pyx_k_DragForceApplier___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_DragForceApplier___setstate_cyth, __pyx_k_DragForceApplier___setstate_cyth, sizeof(__pyx_k_DragForceApplier___setstate_cyth), 0, 0, 1, 1},
    {&__pyx_n_s_ElectrostaticForceApplier, __pyx_k_ElectrostaticForceApplier, sizeof(__pyx_k_ElectrostaticForceApplier), 0, 0, 1, 1},
    {&__pyx_n_s_ElectrostaticForceApplier___redu, __pyx_k_ElectrostaticForceApplier___redu, sizeof(__pyx_k_ElectrostaticForceApplier___redu), 0, 0, 1, 1},
    {&__pyx_n_s_ElectrostaticForceApplier___sets, __pyx_k_ElectrostaticForceApplier___sets, sizeof(__pyx_k_ElectrostaticForceApplier___sets), 0, 0, 1, 1},
    {&__pyx_n_s_FieldForceApplier, __pyx_k_FieldForceApplier, sizeof(__pyx_k_FieldForceApplier), 0, 0, 1, 1},
    {&__pyx_n_s_FieldForceApplier___reduce_cytho, __pyx_k_FieldForceApplier___reduce_cytho, sizeof(__pyx_k_FieldForceApplier___reduce_cytho), 0, 0, 1, 1},
    {&__pyx_n_s_FieldForceApplier___setstate_cyt, __pyx_k_FieldForceApplier___setstate_cyt, sizeof(__pyx_k_FieldForceApplier___setstate_cyt), 0, 0, 1, 1},
    {&__pyx_n_s_GravityForceApplier, __pyx_k_GravityForceApplier, sizeof(__pyx_k_GravityForceApplier), 0, 0, 1, 1},
    {&__pyx_n_s_GravityForceApplier___reduce_cyt, __pyx_k_GravityForceApplier___reduce_cyt, sizeof(__pyx_k_GravityForceApplier___reduce_cyt), 0, 0, 1, 1},
    {&__pyx_n_s_GravityForceApplier___setstate_c, __pyx_k_GravityForceApplier___setstate_c, sizeof(__pyx_k_GravityForceApplier___setstate_c), 0, 0, 1, 1},
    {&__pyx_kp_u_Group_with_particles_forces_and, __pyx_k_Group_with_particles_forces_and, sizeof(__pyx_k_Group_with_particles_forces_and), 0, 1, 0, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_k_Incompatible_checksums_0x_x_vs_0, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2, __pyx_k_Incompatible_checksums_0x_x_vs_0_2, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_2), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_k_Incompatible_checksums_0x_x_vs_0_3, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_3), 0, 0, 1, 0},
    {&__pyx_n_s_MatrixForceApplier, __pyx_k_MatrixForceApplier, sizeof(__pyx_k_MatrixForceApplier), 0, 0, 1, 1},
    {&__pyx_n_s_MatrixForceApplier___reduce_cyth, __pyx_k_MatrixForceApplier___reduce_cyth, sizeof(__pyx_k_MatrixForceApplier___reduce_cyth), 0, 0, 1, 1},
    {&__pyx_n_s_MatrixForceApplier___setstate_cy, __pyx_k_MatrixForceApplier___setstate_cy, sizeof(__pyx_k_MatrixForceApplier___setstate_cy), 0, 0, 1, 1},
    {&__pyx_kp_u_New_D_physics_r_with_particles_f, __pyx_k_New_D_physics_r_with_particles_f, sizeof(__pyx_k_New_D_physics_r_with_particles_f), 0, 1, 0, 0},
    {&__pyx_kp_u_New_D_physics_r_with_particles_o, __pyx_k_New_D_physics_r_with_particles_o, sizeof(__pyx_k_New_D_physics_r_with_particles_o), 0, 1, 0, 0},
    {&__pyx_n_s_NotImplementedError, __pyx_k_NotImplementedError, sizeof(__pyx_k_NotImplementedError), 0, 0, 1, 1},
    {&__pyx_n_s_Particle, __pyx_k_Particle, sizeof(__pyx_k_Particle), 0, 0, 1, 1},
    {&__pyx_n_s_ParticleForceApplier, __pyx_k_ParticleForceApplier, sizeof(__pyx_k_ParticleForceApplier), 0, 0, 1, 1},
    {&__pyx_n_s_ParticleForceApplier___reduce_cy, __pyx_k_ParticleForceApplier___reduce_cy, sizeof(__pyx_k_ParticleForceApplier___reduce_cy), 0, 0, 1, 1},
    {&__pyx_n_s_ParticleForceApplier___setstate, __pyx_k_ParticleForceApplier___setstate, sizeof(__pyx_k_ParticleForceApplier___setstate), 0, 0, 1, 1},
    {&__pyx_n_s_Particle___reduce_cython, __pyx_k_Particle___reduce_cython, sizeof(__pyx_k_Particle___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Particle___setstate_cython, __pyx_k_Particle___setstate_cython, sizeof(__pyx_k_Particle___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_PhysicsGroup, __pyx_k_PhysicsGroup, sizeof(__pyx_k_PhysicsGroup), 0, 0, 1, 1},
    {&__pyx_n_s_PhysicsGroup___reduce_cython, __pyx_k_PhysicsGroup___reduce_cython, sizeof(__pyx_k_PhysicsGroup___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_PhysicsGroup___setstate_cython, __pyx_k_PhysicsGroup___setstate_cython, sizeof(__pyx_k_PhysicsGroup___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_PhysicsSystem, __pyx_k_PhysicsSystem, sizeof(__pyx_k_PhysicsSystem), 0, 0, 1, 1},
    {&__pyx_n_s_PhysicsSystem___reduce_cython, __pyx_k_PhysicsSystem___reduce_cython, sizeof(__pyx_k_PhysicsSystem___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_PhysicsSystem___setstate_cython, __pyx_k_PhysicsSystem___setstate_cython, sizeof(__pyx_k_PhysicsSystem___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_PhysicsSystem_calculate, __pyx_k_PhysicsSystem_calculate, sizeof(__pyx_k_PhysicsSystem_calculate), 0, 0, 1, 1},
    {&__pyx_n_s_PhysicsSystem_destroy, __pyx_k_PhysicsSystem_destroy, sizeof(__pyx_k_PhysicsSystem_destroy), 0, 0, 1, 1},
    {&__pyx_n_s_PhysicsSystem_purge, __pyx_k_PhysicsSystem_purge, sizeof(__pyx_k_PhysicsSystem_purge), 0, 0, 1, 1},
    {&__pyx_n_s_PhysicsSystem_update, __pyx_k_PhysicsSystem_update, sizeof(__pyx_k_PhysicsSystem_update), 0, 0, 1, 1},
    {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
    {&__pyx_n_s_RandomForceApplier, __pyx_k_RandomForceApplier, sizeof(__pyx_k_RandomForceApplier), 0, 0, 1, 1},
    {&__pyx_n_s_RandomForceApplier___reduce_cyth, __pyx_k_RandomForceApplier___reduce_cyth, sizeof(__pyx_k_RandomForceApplier___reduce_cyth), 0, 0, 1, 1},
    {&__pyx_n_s_RandomForceApplier___setstate_cy, __pyx_k_RandomForceApplier___setstate_cy, sizeof(__pyx_k_RandomForceApplier___setstate_cy), 0, 0, 1, 1},
    {&__pyx_kp_u_Reset_physics_system_r_for_run, __pyx_k_Reset_physics_system_r_for_run, sizeof(__pyx_k_Reset_physics_system_r_for_run), 0, 1, 0, 0},
    {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
    {&__pyx_kp_u__5, __pyx_k__5, sizeof(__pyx_k__5), 0, 1, 0, 0},
    {&__pyx_n_s__52, __pyx_k__52, sizeof(__pyx_k__52), 0, 0, 1, 1},
    {&__pyx_n_s__8, __pyx_k__8, sizeof(__pyx_k__8), 0, 0, 1, 1},
    {&__pyx_n_s__9, __pyx_k__9, sizeof(__pyx_k__9), 0, 0, 1, 1},
    {&__pyx_n_u_acceleration, __pyx_k_acceleration, sizeof(__pyx_k_acceleration), 0, 1, 0, 1},
    {&__pyx_n_u_adhesion, __pyx_k_adhesion, sizeof(__pyx_k_adhesion), 0, 1, 0, 1},
    {&__pyx_n_u_anchor, __pyx_k_anchor, sizeof(__pyx_k_anchor), 0, 1, 0, 1},
    {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
    {&__pyx_n_s_asyncio, __pyx_k_asyncio, sizeof(__pyx_k_asyncio), 0, 0, 1, 1},
    {&__pyx_n_s_asyncio_coroutines, __pyx_k_asyncio_coroutines, sizeof(__pyx_k_asyncio_coroutines), 0, 0, 1, 1},
    {&__pyx_n_s_asyncio_tasks, __pyx_k_asyncio_tasks, sizeof(__pyx_k_asyncio_tasks), 0, 0, 1, 1},
    {&__pyx_n_s_await, __pyx_k_await, sizeof(__pyx_k_await), 0, 0, 1, 1},
    {&__pyx_n_u_barrier, __pyx_k_barrier, sizeof(__pyx_k_barrier), 0, 1, 0, 1},
    {&__pyx_n_u_buoyancy, __pyx_k_buoyancy, sizeof(__pyx_k_buoyancy), 0, 1, 0, 1},
    {&__pyx_n_s_calculate, __pyx_k_calculate, sizeof(__pyx_k_calculate), 0, 0, 1, 1},
    {&__pyx_n_u_charge, __pyx_k_charge, sizeof(__pyx_k_charge), 0, 1, 0, 1},
    {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
    {&__pyx_n_s_clock, __pyx_k_clock, sizeof(__pyx_k_clock), 0, 0, 1, 1},
    {&__pyx_n_u_clock, __pyx_k_clock, sizeof(__pyx_k_clock), 0, 1, 0, 1},
    {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
    {&__pyx_n_u_collision, __pyx_k_collision, sizeof(__pyx_k_collision), 0, 1, 0, 1},
    {&__pyx_n_u_constant, __pyx_k_constant, sizeof(__pyx_k_constant), 0, 1, 0, 1},
    {&__pyx_n_s_debug, __pyx_k_debug, sizeof(__pyx_k_debug), 0, 0, 1, 1},
    {&__pyx_n_u_density, __pyx_k_density, sizeof(__pyx_k_density), 0, 1, 0, 1},
    {&__pyx_n_s_destroy, __pyx_k_destroy, sizeof(__pyx_k_destroy), 0, 0, 1, 1},
    {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
    {&__pyx_n_s_dict_2, __pyx_k_dict_2, sizeof(__pyx_k_dict_2), 0, 0, 1, 1},
    {&__pyx_n_u_dimensions, __pyx_k_dimensions, sizeof(__pyx_k_dimensions), 0, 1, 0, 1},
    {&__pyx_n_u_direction, __pyx_k_direction, sizeof(__pyx_k_direction), 0, 1, 0, 1},
    {&__pyx_kp_u_disable, __pyx_k_disable, sizeof(__pyx_k_disable), 0, 1, 0, 0},
    {&__pyx_n_u_distance, __pyx_k_distance, sizeof(__pyx_k_distance), 0, 1, 0, 1},
    {&__pyx_n_u_drag, __pyx_k_drag, sizeof(__pyx_k_drag), 0, 1, 0, 1},
    {&__pyx_n_u_ease, __pyx_k_ease, sizeof(__pyx_k_ease), 0, 1, 0, 1},
    {&__pyx_n_u_electrostatic, __pyx_k_electrostatic, sizeof(__pyx_k_electrostatic), 0, 1, 0, 1},
    {&__pyx_kp_u_enable, __pyx_k_enable, sizeof(__pyx_k_enable), 0, 1, 0, 0},
    {&__pyx_n_s_engine, __pyx_k_engine, sizeof(__pyx_k_engine), 0, 0, 1, 1},
    {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
    {&__pyx_n_s_extra, __pyx_k_extra, sizeof(__pyx_k_extra), 0, 0, 1, 1},
    {&__pyx_n_u_field, __pyx_k_field, sizeof(__pyx_k_field), 0, 1, 0, 1},
    {&__pyx_n_u_fixed, __pyx_k_fixed, sizeof(__pyx_k_fixed), 0, 1, 0, 1},
    {&__pyx_n_s_flitter_render_physics, __pyx_k_flitter_render_physics, sizeof(__pyx_k_flitter_render_physics), 0, 0, 1, 1},
    {&__pyx_n_u_flow, __pyx_k_flow, sizeof(__pyx_k_flow), 0, 1, 0, 1},
    {&__pyx_n_u_force, __pyx_k_force, sizeof(__pyx_k_force), 0, 1, 0, 1},
    {&__pyx_n_u_from, __pyx_k_from, sizeof(__pyx_k_from), 0, 1, 0, 1},
    {&__pyx_kp_u_gc, __pyx_k_gc, sizeof(__pyx_k_gc), 0, 1, 0, 0},
    {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
    {&__pyx_n_u_gravity, __pyx_k_gravity, sizeof(__pyx_k_gravity), 0, 1, 0, 1},
    {&__pyx_n_s_group, __pyx_k_group, sizeof(__pyx_k_group), 0, 0, 1, 1},
    {&__pyx_n_u_group, __pyx_k_group, sizeof(__pyx_k_group), 0, 1, 0, 1},
    {&__pyx_n_s_groups, __pyx_k_groups, sizeof(__pyx_k_groups), 0, 0, 1, 1},
    {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
    {&__pyx_n_u_id, __pyx_k_id, sizeof(__pyx_k_id), 0, 1, 0, 1},
    {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
    {&__pyx_n_s_initializing, __pyx_k_initializing, sizeof(__pyx_k_initializing), 0, 0, 1, 1},
    {&__pyx_n_s_inspect, __pyx_k_inspect, sizeof(__pyx_k_inspect), 0, 0, 1, 1},
    {&__pyx_n_s_is_coroutine, __pyx_k_is_coroutine, sizeof(__pyx_k_is_coroutine), 0, 0, 1, 1},
    {&__pyx_kp_u_isenabled, __pyx_k_isenabled, sizeof(__pyx_k_isenabled), 0, 1, 0, 0},
    {&__pyx_n_s_iteration, __pyx_k_iteration, sizeof(__pyx_k_iteration), 0, 0, 1, 1},
    {&__pyx_n_u_iteration, __pyx_k_iteration, sizeof(__pyx_k_iteration), 0, 1, 0, 1},
    {&__pyx_n_s_iteration_vector, __pyx_k_iteration_vector, sizeof(__pyx_k_iteration_vector), 0, 0, 1, 1},
    {&__pyx_n_s_kwargs, __pyx_k_kwargs, sizeof(__pyx_k_kwargs), 0, 0, 1, 1},
    {&__pyx_n_u_last, __pyx_k_last, sizeof(__pyx_k_last), 0, 1, 0, 1},
    {&__pyx_n_s_last_run, __pyx_k_last_run, sizeof(__pyx_k_last_run), 0, 0, 1, 1},
    {&__pyx_n_s_last_run_vector, __pyx_k_last_run_vector, sizeof(__pyx_k_last_run_vector), 0, 0, 1, 1},
    {&__pyx_n_s_last_time, __pyx_k_last_time, sizeof(__pyx_k_last_time), 0, 0, 1, 1},
    {&__pyx_n_s_logger, __pyx_k_logger, sizeof(__pyx_k_logger), 0, 0, 1, 1},
    {&__pyx_n_s_loguru, __pyx_k_loguru, sizeof(__pyx_k_loguru), 0, 0, 1, 1},
    {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
    {&__pyx_n_u_mass, __pyx_k_mass, sizeof(__pyx_k_mass), 0, 1, 0, 1},
    {&__pyx_n_u_max, __pyx_k_max, sizeof(__pyx_k_max), 0, 1, 0, 1},
    {&__pyx_n_u_max_distance, __pyx_k_max_distance, sizeof(__pyx_k_max_distance), 0, 1, 0, 1},
    {&__pyx_n_u_maximum, __pyx_k_maximum, sizeof(__pyx_k_maximum), 0, 1, 0, 1},
    {&__pyx_n_u_min, __pyx_k_min, sizeof(__pyx_k_min), 0, 1, 0, 1},
    {&__pyx_n_u_minimum, __pyx_k_minimum, sizeof(__pyx_k_minimum), 0, 1, 0, 1},
    {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
    {&__pyx_n_s_name_patch, __pyx_k_name_patch, sizeof(__pyx_k_name_patch), 0, 0, 1, 1},
    {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
    {&__pyx_n_s_new_state_keys, __pyx_k_new_state_keys, sizeof(__pyx_k_new_state_keys), 0, 0, 1, 1},
    {&__pyx_kp_s_no_default___reduce___due_to_non, __pyx_k_no_default___reduce___due_to_non, sizeof(__pyx_k_no_default___reduce___due_to_non), 0, 0, 1, 0},
    {&__pyx_n_s_node, __pyx_k_node, sizeof(__pyx_k_node), 0, 0, 1, 1},
    {&__pyx_n_u_normal, __pyx_k_normal, sizeof(__pyx_k_normal), 0, 1, 0, 1},
    {&__pyx_n_s_old_state_keys, __pyx_k_old_state_keys, sizeof(__pyx_k_old_state_keys), 0, 0, 1, 1},
    {&__pyx_n_u_overlap, __pyx_k_overlap, sizeof(__pyx_k_overlap), 0, 1, 0, 1},
    {&__pyx_n_s_particle, __pyx_k_particle, sizeof(__pyx_k_particle), 0, 0, 1, 1},
    {&__pyx_n_u_particle, __pyx_k_particle, sizeof(__pyx_k_particle), 0, 1, 0, 1},
    {&__pyx_n_s_particles_by_id, __pyx_k_particles_by_id, sizeof(__pyx_k_particles_by_id), 0, 0, 1, 1},
    {&__pyx_n_s_performance, __pyx_k_performance, sizeof(__pyx_k_performance), 0, 0, 1, 1},
    {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
    {&__pyx_n_u_position, __pyx_k_position, sizeof(__pyx_k_position), 0, 1, 0, 1},
    {&__pyx_n_u_power, __pyx_k_power, sizeof(__pyx_k_power), 0, 1, 0, 1},
    {&__pyx_n_s_purge, __pyx_k_purge, sizeof(__pyx_k_purge), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_DistanceForceAppl, __pyx_k_pyx_unpickle_DistanceForceAppl, sizeof(__pyx_k_pyx_unpickle_DistanceForceAppl), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_Particle, __pyx_k_pyx_unpickle_Particle, sizeof(__pyx_k_pyx_unpickle_Particle), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_PhysicsSystem, __pyx_k_pyx_unpickle_PhysicsSystem, sizeof(__pyx_k_pyx_unpickle_PhysicsSystem), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
    {&__pyx_n_u_radius, __pyx_k_radius, sizeof(__pyx_k_radius), 0, 1, 0, 1},
    {&__pyx_n_u_random, __pyx_k_random, sizeof(__pyx_k_random), 0, 1, 0, 1},
    {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
    {&__pyx_n_s_realtime, __pyx_k_realtime, sizeof(__pyx_k_realtime), 0, 0, 1, 1},
    {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
    {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
    {&__pyx_n_u_resolution, __pyx_k_resolution, sizeof(__pyx_k_resolution), 0, 1, 0, 1},
    {&__pyx_n_u_restitution, __pyx_k_restitution, sizeof(__pyx_k_restitution), 0, 1, 0, 1},
    {&__pyx_n_s_run, __pyx_k_run, sizeof(__pyx_k_run), 0, 0, 1, 1},
    {&__pyx_n_u_run, __pyx_k_run, sizeof(__pyx_k_run), 0, 1, 0, 1},
    {&__pyx_n_s_run_vector, __pyx_k_run_vector, sizeof(__pyx_k_run_vector), 0, 0, 1, 1},
    {&__pyx_n_s_seed, __pyx_k_seed, sizeof(__pyx_k_seed), 0, 0, 1, 1},
    {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
    {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
    {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
    {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_spec, __pyx_k_spec, sizeof(__pyx_k_spec), 0, 0, 1, 1},
    {&__pyx_n_u_speed_of_light, __pyx_k_speed_of_light, sizeof(__pyx_k_speed_of_light), 0, 1, 0, 1},
    {&__pyx_kp_s_src_flitter_render_physics_pyx, __pyx_k_src_flitter_render_physics_pyx, sizeof(__pyx_k_src_flitter_render_physics_pyx), 0, 0, 1, 0},
    {&__pyx_n_s_state, __pyx_k_state, sizeof(__pyx_k_state), 0, 0, 1, 1},
    {&__pyx_n_u_state, __pyx_k_state, sizeof(__pyx_k_state), 0, 1, 0, 1},
    {&__pyx_n_s_state_key, __pyx_k_state_key, sizeof(__pyx_k_state_key), 0, 0, 1, 1},
    {&__pyx_n_s_state_prefix, __pyx_k_state_prefix, sizeof(__pyx_k_state_prefix), 0, 0, 1, 1},
    {&__pyx_n_s_strength, __pyx_k_strength, sizeof(__pyx_k_strength), 0, 0, 1, 1},
    {&__pyx_n_u_strength, __pyx_k_strength, sizeof(__pyx_k_strength), 0, 1, 0, 1},
    {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
    {&__pyx_n_s_target_fps, __pyx_k_target_fps, sizeof(__pyx_k_target_fps), 0, 0, 1, 1},
    {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
    {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
    {&__pyx_n_s_time, __pyx_k_time, sizeof(__pyx_k_time), 0, 0, 1, 1},
    {&__pyx_n_u_time, __pyx_k_time, sizeof(__pyx_k_time), 0, 1, 0, 1},
    {&__pyx_n_s_time_vector, __pyx_k_time_vector, sizeof(__pyx_k_time_vector), 0, 0, 1, 1},
    {&__pyx_n_u_to, __pyx_k_to, sizeof(__pyx_k_to), 0, 1, 0, 1},
    {&__pyx_n_s_to_thread, __pyx_k_to_thread, sizeof(__pyx_k_to_thread), 0, 0, 1, 1},
    {&__pyx_n_s_top_group, __pyx_k_top_group, sizeof(__pyx_k_top_group), 0, 0, 1, 1},
    {&__pyx_n_s_trace, __pyx_k_trace, sizeof(__pyx_k_trace), 0, 0, 1, 1},
    {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
    {&__pyx_n_s_use_setstate, __pyx_k_use_setstate, sizeof(__pyx_k_use_setstate), 0, 0, 1, 1},
    {&__pyx_n_u_velocity, __pyx_k_velocity, sizeof(__pyx_k_velocity), 0, 1, 0, 1},
    {&__pyx_n_s_zero, __pyx_k_zero, sizeof(__pyx_k_zero), 0, 0, 1, 1},
    {0, 0, 0, 0, 0, 0, 0}
  };
  return __Pyx_InitStrings(__pyx_string_tab);
}
/* #### Code section: cached_builtins ### */
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(1, 73, __pyx_L1_error)
  __pyx_builtin_NotImplementedError = __Pyx_GetBuiltinName(__pyx_n_s_NotImplementedError); if (!__pyx_builtin_NotImplementedError) __PYX_ERR(1, 175, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(1, 482, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cached_constants ### */

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xc9134b9, 0x81fb8cd, 0xb66f251):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc9134b9, 0x81fb8cd, 0xb66f251) = (acceleration, charge, ease, force, id, initial_force, mass, non_anchor, position, position_state_key, radius, velocity, velocity_state_key))" % __pyx_checksum
 */
  __pyx_tuple__4 = PyTuple_Pack(3, __pyx_int_210842809, __pyx_int_136296653, __pyx_int_191296081); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);
  __pyx_tuple__6 = PyTuple_Pack(3, __pyx_int_211912363, __pyx_int_165385471, __pyx_int_158432715); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);
  __pyx_tuple__7 = PyTuple_Pack(3, __pyx_int_253013475, __pyx_int_142976427, __pyx_int_64470655); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);
  __pyx_codeobj__11 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__10, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__11)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_tuple__12 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_pyx_state); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);
  __pyx_codeobj__13 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__12, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__13)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_tuple__14 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_state, __pyx_n_s_dict_2, __pyx_n_s_use_setstate); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);
  __pyx_codeobj__15 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__14, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__15)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Particle, (type(self), 0xc9134b9, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Particle__set_state(self, __pyx_state)
 */
  __pyx_codeobj__16 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__12, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__16)) __PYX_ERR(0, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__17 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__10, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__17)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__18 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__12, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__18)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__19 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__10, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__19)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__20 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__12, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__20)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__21 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__10, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__21)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__22 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__12, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__22)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__23 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__14, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__23)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_DistanceForceApplier, (type(self), 0xca186ab, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_DistanceForceApplier__set_state(self, __pyx_state)
 */
  __pyx_codeobj__24 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__12, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__24)) __PYX_ERR(0, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__25 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__10, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__25)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__26 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__12, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__26)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__27 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__10, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__27)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__28 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__12, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__28)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__29 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__10, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__29)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__30 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__12, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__30)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__31 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__10, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__31)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__32 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__12, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__32)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__33 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__10, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__33)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__34 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__12, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__34)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__35 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__10, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__35)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__36 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__12, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__36)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__37 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__10, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__37)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__38 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__12, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__38)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__39 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__10, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__39)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__40 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__12, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__40)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__41 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__10, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__41)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__42 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__12, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__42)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "flitter/render/physics.pyx":401
 *         self.state_keys = set()
 * 
 *     async def destroy(self):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_codeobj_ = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_COROUTINE, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__10, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_render_physics_pyx, __pyx_n_s_destroy, 401, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj_)) __PYX_ERR(1, 401, __pyx_L1_error)

  /* "flitter/render/physics.pyx":404
 *         pass
 * 
 *     async def purge(self):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_codeobj__2 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_COROUTINE, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__10, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_render_physics_pyx, __pyx_n_s_purge, 404, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2)) __PYX_ERR(1, 404, __pyx_L1_error)

  /* "flitter/render/physics.pyx":407
 *         pass
 * 
 *     async def update(self, engine, Node node, double time, double performance, **kwargs):             # <<<<<<<<<<<<<<
 *         cdef int64_t run = node.get_int('run', 0)
 *         self.dimensions = node.get_int('dimensions', 0)
 */
  __pyx_tuple__43 = PyTuple_Pack(28, __pyx_n_s_self, __pyx_n_s_engine, __pyx_n_s_node, __pyx_n_s_time, __pyx_n_s_performance, __pyx_n_s_kwargs, __pyx_n_s_run, __pyx_n_s_state_prefix, __pyx_n_s_state, __pyx_n_s_time_vector, __pyx_n_s_clock, __pyx_n_s_last_time, __pyx_n_s_last_run_vector, __pyx_n_s_last_run, __pyx_n_s_iteration_vector, __pyx_n_s_iteration, __pyx_n_s_state_key, __pyx_n_s_run_vector, __pyx_n_s_seed, __pyx_n_s_group, __pyx_n_s_top_group, __pyx_n_s_groups, __pyx_n_s_particles_by_id, __pyx_n_s_old_state_keys, __pyx_n_s_new_state_keys, __pyx_n_s_i, __pyx_n_s_extra, __pyx_n_s_particle); if (unlikely(!__pyx_tuple__43)) __PYX_ERR(1, 407, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__43);
  __Pyx_GIVEREF(__pyx_tuple__43);
  __pyx_codeobj__3 = (PyObject*)__Pyx_PyCode_New(5, 0, 0, 28, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARKEYWORDS|CO_COROUTINE, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__43, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_render_physics_pyx, __pyx_n_s_update, 407, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__3)) __PYX_ERR(1, 407, __pyx_L1_error)

  /* "flitter/render/physics.pyx":609
 *                     group.barriers.append(barrier)
 * 
 *     cpdef double calculate(self, list groups, dict particles_by_id, bint realtime, bint extra, double time, double last_time, double clock):             # <<<<<<<<<<<<<<
 *         cdef list all_particles, particles, forces, barriers
 *         cdef int64_t i, j, k, m, n, p
 */
  __pyx_tuple__44 = PyTuple_Pack(8, __pyx_n_s_self, __pyx_n_s_groups, __pyx_n_s_particles_by_id, __pyx_n_s_realtime, __pyx_n_s_extra, __pyx_n_s_time, __pyx_n_s_last_time, __pyx_n_s_clock); if (unlikely(!__pyx_tuple__44)) __PYX_ERR(1, 609, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__44);
  __Pyx_GIVEREF(__pyx_tuple__44);
  __pyx_codeobj__45 = (PyObject*)__Pyx_PyCode_New(8, 0, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__44, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_flitter_render_physics_pyx, __pyx_n_s_calculate, 609, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__45)) __PYX_ERR(1, 609, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__46 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__14, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__46)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PhysicsSystem, (type(self), 0xf14ade3, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PhysicsSystem__set_state(self, __pyx_state)
 */
  __pyx_codeobj__47 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__12, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__47)) __PYX_ERR(0, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __pyx_unpickle_Particle(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__48 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__48)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__48);
  __Pyx_GIVEREF(__pyx_tuple__48);
  __pyx_codeobj__49 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__48, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Particle, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__49)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_codeobj__50 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__48, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_DistanceForceAppl, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__50)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_codeobj__51 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__48, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_PhysicsSystem, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__51)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}
/* #### Code section: init_constants ### */

static CYTHON_SMALL_CODE int __Pyx_InitConstants(void) {
  if (__Pyx_CreateStringTabAndInitStrings() < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_64470655 = PyInt_FromLong(64470655L); if (unlikely(!__pyx_int_64470655)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_136296653 = PyInt_FromLong(136296653L); if (unlikely(!__pyx_int_136296653)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_142976427 = PyInt_FromLong(142976427L); if (unlikely(!__pyx_int_142976427)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_158432715 = PyInt_FromLong(158432715L); if (unlikely(!__pyx_int_158432715)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_165385471 = PyInt_FromLong(165385471L); if (unlikely(!__pyx_int_165385471)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_191296081 = PyInt_FromLong(191296081L); if (unlikely(!__pyx_int_191296081)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_210842809 = PyInt_FromLong(210842809L); if (unlikely(!__pyx_int_210842809)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_211912363 = PyInt_FromLong(211912363L); if (unlikely(!__pyx_int_211912363)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_253013475 = PyInt_FromLong(253013475L); if (unlikely(!__pyx_int_253013475)) __PYX_ERR(1, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_globals ### */

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  return 0;
}
/* #### Code section: init_module ### */

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __pyx_v_7flitter_6render_7physics_VELOCITY = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  __pyx_v_7flitter_6render_7physics_LAST = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  __pyx_v_7flitter_6render_7physics_CLOCK = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  __pyx_v_7flitter_6render_7physics_RUN = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  __pyx_v_7flitter_6render_7physics_ITERATION = ((struct __pyx_obj_7flitter_5model_Vector *)Py_None); Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_7flitter_6render_7physics_PhysicsGroup = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_6render_7physics_PhysicsGroup_spec, NULL); if (unlikely(!__pyx_ptype_7flitter_6render_7physics_PhysicsGroup)) __PYX_ERR(1, 37, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_6render_7physics_PhysicsGroup_spec, __pyx_ptype_7flitter_6render_7physics_PhysicsGroup) < 0) __PYX_ERR(1, 37, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_6render_7physics_PhysicsGroup = &__pyx_type_7flitter_6render_7physics_PhysicsGroup;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_6render_7physics_PhysicsGroup) < 0) __PYX_ERR(1, 37, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_6render_7physics_PhysicsGroup->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_6render_7physics_PhysicsGroup->tp_dictoffset && __pyx_ptype_7flitter_6render_7physics_PhysicsGroup->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_6render_7physics_PhysicsGroup->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_PhysicsGroup, (PyObject *) __pyx_ptype_7flitter_6render_7physics_PhysicsGroup) < 0) __PYX_ERR(1, 37, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_6render_7physics_PhysicsGroup) < 0) __PYX_ERR(1, 37, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_6render_7physics_Particle = &__pyx_vtable_7flitter_6render_7physics_Particle;
  __pyx_vtable_7flitter_6render_7physics_Particle.step = (void (*)(struct __pyx_obj_7flitter_6render_7physics_Particle *, double, double, double))__pyx_f_7flitter_6render_7physics_8Particle_step;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_7flitter_6render_7physics_Particle = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_6render_7physics_Particle_spec, NULL); if (unlikely(!__pyx_ptype_7flitter_6render_7physics_Particle)) __PYX_ERR(1, 54, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_6render_7physics_Particle_spec, __pyx_ptype_7flitter_6render_7physics_Particle) < 0) __PYX_ERR(1, 54, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_6render_7physics_Particle = &__pyx_type_7flitter_6render_7physics_Particle;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_6render_7physics_Particle) < 0) __PYX_ERR(1, 54, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_6render_7physics_Particle->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_6render_7physics_Particle->tp_dictoffset && __pyx_ptype_7flitter_6render_7physics_Particle->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_6render_7physics_Particle->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_6render_7physics_Particle, __pyx_vtabptr_7flitter_6render_7physics_Particle) < 0) __PYX_ERR(1, 54, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_6render_7physics_Particle) < 0) __PYX_ERR(1, 54, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Particle, (PyObject *) __pyx_ptype_7flitter_6render_7physics_Particle) < 0) __PYX_ERR(1, 54, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_6render_7physics_Particle) < 0) __PYX_ERR(1, 54, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_6render_7physics_Barrier = &__pyx_vtable_7flitter_6render_7physics_Barrier;
  __pyx_vtable_7flitter_6render_7physics_Barrier.apply_distance = (void (*)(struct __pyx_obj_7flitter_6render_7physics_Barrier *, struct __pyx_obj_7flitter_6render_7physics_Particle *))__pyx_f_7flitter_6render_7physics_7Barrier_apply_distance;
  __pyx_vtable_7flitter_6render_7physics_Barrier.apply_rebound = (void (*)(struct __pyx_obj_7flitter_6render_7physics_Barrier *, struct __pyx_obj_7flitter_6render_7physics_Particle *, double, double, double))__pyx_f_7flitter_6render_7physics_7Barrier_apply_rebound;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_7flitter_6render_7physics_Barrier = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_6render_7physics_Barrier_spec, NULL); if (unlikely(!__pyx_ptype_7flitter_6render_7physics_Barrier)) __PYX_ERR(1, 98, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_6render_7physics_Barrier_spec, __pyx_ptype_7flitter_6render_7physics_Barrier) < 0) __PYX_ERR(1, 98, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_6render_7physics_Barrier = &__pyx_type_7flitter_6render_7physics_Barrier;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_6render_7physics_Barrier) < 0) __PYX_ERR(1, 98, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_6render_7physics_Barrier->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_6render_7physics_Barrier->tp_dictoffset && __pyx_ptype_7flitter_6render_7physics_Barrier->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_6render_7physics_Barrier->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_6render_7physics_Barrier, __pyx_vtabptr_7flitter_6render_7physics_Barrier) < 0) __PYX_ERR(1, 98, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_6render_7physics_Barrier) < 0) __PYX_ERR(1, 98, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Barrier, (PyObject *) __pyx_ptype_7flitter_6render_7physics_Barrier) < 0) __PYX_ERR(1, 98, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_6render_7physics_Barrier) < 0) __PYX_ERR(1, 98, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_6render_7physics_ParticleForceApplier = &__pyx_vtable_7flitter_6render_7physics_ParticleForceApplier;
  __pyx_vtable_7flitter_6render_7physics_ParticleForceApplier.apply = (void (*)(struct __pyx_obj_7flitter_6render_7physics_ParticleForceApplier *, struct __pyx_obj_7flitter_6render_7physics_Particle *, double))__pyx_f_7flitter_6render_7physics_20ParticleForceApplier_apply;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_7flitter_6render_7physics_ParticleForceApplier = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_6render_7physics_ParticleForceApplier_spec, NULL); if (unlikely(!__pyx_ptype_7flitter_6render_7physics_ParticleForceApplier)) __PYX_ERR(1, 166, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_6render_7physics_ParticleForceApplier_spec, __pyx_ptype_7flitter_6render_7physics_ParticleForceApplier) < 0) __PYX_ERR(1, 166, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_6render_7physics_ParticleForceApplier = &__pyx_type_7flitter_6render_7physics_ParticleForceApplier;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_6render_7physics_ParticleForceApplier) < 0) __PYX_ERR(1, 166, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_6render_7physics_ParticleForceApplier->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_6render_7physics_ParticleForceApplier->tp_dictoffset && __pyx_ptype_7flitter_6render_7physics_ParticleForceApplier->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_6render_7physics_ParticleForceApplier->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_6render_7physics_ParticleForceApplier, __pyx_vtabptr_7flitter_6render_7physics_ParticleForceApplier) < 0) __PYX_ERR(1, 166, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_6render_7physics_ParticleForceApplier) < 0) __PYX_ERR(1, 166, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ParticleForceApplier, (PyObject *) __pyx_ptype_7flitter_6render_7physics_ParticleForceApplier) < 0) __PYX_ERR(1, 166, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_6render_7physics_ParticleForceApplier) < 0) __PYX_ERR(1, 166, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_6render_7physics_MatrixForceApplier = &__pyx_vtable_7flitter_6render_7physics_MatrixForceApplier;
  __pyx_vtable_7flitter_6render_7physics_MatrixForceApplier.apply = (void (*)(struct __pyx_obj_7flitter_6render_7physics_MatrixForceApplier *, struct __pyx_obj_7flitter_6render_7physics_Particle *, struct __pyx_obj_7flitter_6render_7physics_Particle *, struct __pyx_obj_7flitter_5model_Vector *, double))__pyx_f_7flitter_6render_7physics_18MatrixForceApplier_apply;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_7flitter_6render_7physics_MatrixForceApplier = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_6render_7physics_MatrixForceApplier_spec, NULL); if (unlikely(!__pyx_ptype_7flitter_6render_7physics_MatrixForceApplier)) __PYX_ERR(1, 178, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_6render_7physics_MatrixForceApplier_spec, __pyx_ptype_7flitter_6render_7physics_MatrixForceApplier) < 0) __PYX_ERR(1, 178, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_6render_7physics_MatrixForceApplier = &__pyx_type_7flitter_6render_7physics_MatrixForceApplier;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_6render_7physics_MatrixForceApplier) < 0) __PYX_ERR(1, 178, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_6render_7physics_MatrixForceApplier->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_6render_7physics_MatrixForceApplier->tp_dictoffset && __pyx_ptype_7flitter_6render_7physics_MatrixForceApplier->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_6render_7physics_MatrixForceApplier->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_6render_7physics_MatrixForceApplier, __pyx_vtabptr_7flitter_6render_7physics_MatrixForceApplier) < 0) __PYX_ERR(1, 178, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_6render_7physics_MatrixForceApplier) < 0) __PYX_ERR(1, 178, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_MatrixForceApplier, (PyObject *) __pyx_ptype_7flitter_6render_7physics_MatrixForceApplier) < 0) __PYX_ERR(1, 178, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_6render_7physics_MatrixForceApplier) < 0) __PYX_ERR(1, 178, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_6render_7physics_DistanceForceApplier = &__pyx_vtable_7flitter_6render_7physics_DistanceForceApplier;
  __pyx_vtable_7flitter_6render_7physics_DistanceForceApplier.apply = (void (*)(struct __pyx_obj_7flitter_6render_7physics_DistanceForceApplier *, struct __pyx_obj_7flitter_6render_7physics_Particle *, struct __pyx_obj_7flitter_6render_7physics_Particle *, struct __pyx_obj_7flitter_5model_Vector *, double))__pyx_f_7flitter_6render_7physics_20DistanceForceApplier_apply;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_7flitter_6render_7physics_DistanceForceApplier = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_6render_7physics_DistanceForceApplier_spec, NULL); if (unlikely(!__pyx_ptype_7flitter_6render_7physics_DistanceForceApplier)) __PYX_ERR(1, 193, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_6render_7physics_DistanceForceApplier_spec, __pyx_ptype_7flitter_6render_7physics_DistanceForceApplier) < 0) __PYX_ERR(1, 193, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_6render_7physics_DistanceForceApplier = &__pyx_type_7flitter_6render_7physics_DistanceForceApplier;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_6render_7physics_DistanceForceApplier) < 0) __PYX_ERR(1, 193, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_6render_7physics_DistanceForceApplier->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_6render_7physics_DistanceForceApplier->tp_dictoffset && __pyx_ptype_7flitter_6render_7physics_DistanceForceApplier->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_6render_7physics_DistanceForceApplier->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_6render_7physics_DistanceForceApplier, __pyx_vtabptr_7flitter_6render_7physics_DistanceForceApplier) < 0) __PYX_ERR(1, 193, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_6render_7physics_DistanceForceApplier) < 0) __PYX_ERR(1, 193, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_DistanceForceApplier, (PyObject *) __pyx_ptype_7flitter_6render_7physics_DistanceForceApplier) < 0) __PYX_ERR(1, 193, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_6render_7physics_DistanceForceApplier) < 0) __PYX_ERR(1, 193, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_6render_7physics_DragForceApplier = &__pyx_vtable_7flitter_6render_7physics_DragForceApplier;
  __pyx_vtable_7flitter_6render_7physics_DragForceApplier.__pyx_base = *__pyx_vtabptr_7flitter_6render_7physics_ParticleForceApplier;
  __pyx_vtable_7flitter_6render_7physics_DragForceApplier.__pyx_base.apply = (void (*)(struct __pyx_obj_7flitter_6render_7physics_ParticleForceApplier *, struct __pyx_obj_7flitter_6render_7physics_Particle *, double))__pyx_f_7flitter_6render_7physics_16DragForceApplier_apply;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_6render_7physics_ParticleForceApplier); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_6render_7physics_DragForceApplier = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_6render_7physics_DragForceApplier_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_6render_7physics_DragForceApplier)) __PYX_ERR(1, 225, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_6render_7physics_DragForceApplier_spec, __pyx_ptype_7flitter_6render_7physics_DragForceApplier) < 0) __PYX_ERR(1, 225, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_6render_7physics_DragForceApplier = &__pyx_type_7flitter_6render_7physics_DragForceApplier;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_6render_7physics_DragForceApplier->tp_base = __pyx_ptype_7flitter_6render_7physics_ParticleForceApplier;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_6render_7physics_DragForceApplier) < 0) __PYX_ERR(1, 225, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_6render_7physics_DragForceApplier->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_6render_7physics_DragForceApplier->tp_dictoffset && __pyx_ptype_7flitter_6render_7physics_DragForceApplier->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_6render_7physics_DragForceApplier->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_6render_7physics_DragForceApplier, __pyx_vtabptr_7flitter_6render_7physics_DragForceApplier) < 0) __PYX_ERR(1, 225, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_6render_7physics_DragForceApplier) < 0) __PYX_ERR(1, 225, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_DragForceApplier, (PyObject *) __pyx_ptype_7flitter_6render_7physics_DragForceApplier) < 0) __PYX_ERR(1, 225, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_6render_7physics_DragForceApplier) < 0) __PYX_ERR(1, 225, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_6render_7physics_BuoyancyForceApplier = &__pyx_vtable_7flitter_6render_7physics_BuoyancyForceApplier;
  __pyx_vtable_7flitter_6render_7physics_BuoyancyForceApplier.__pyx_base = *__pyx_vtabptr_7flitter_6render_7physics_ParticleForceApplier;
  __pyx_vtable_7flitter_6render_7physics_BuoyancyForceApplier.__pyx_base.apply = (void (*)(struct __pyx_obj_7flitter_6render_7physics_ParticleForceApplier *, struct __pyx_obj_7flitter_6render_7physics_Particle *, double))__pyx_f_7flitter_6render_7physics_20BuoyancyForceApplier_apply;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_6render_7physics_ParticleForceApplier); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_6render_7physics_BuoyancyForceApplier = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_6render_7physics_BuoyancyForceApplier_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_6render_7physics_BuoyancyForceApplier)) __PYX_ERR(1, 245, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_6render_7physics_BuoyancyForceApplier_spec, __pyx_ptype_7flitter_6render_7physics_BuoyancyForceApplier) < 0) __PYX_ERR(1, 245, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_6render_7physics_BuoyancyForceApplier = &__pyx_type_7flitter_6render_7physics_BuoyancyForceApplier;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_6render_7physics_BuoyancyForceApplier->tp_base = __pyx_ptype_7flitter_6render_7physics_ParticleForceApplier;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_6render_7physics_BuoyancyForceApplier) < 0) __PYX_ERR(1, 245, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_6render_7physics_BuoyancyForceApplier->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_6render_7physics_BuoyancyForceApplier->tp_dictoffset && __pyx_ptype_7flitter_6render_7physics_BuoyancyForceApplier->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_6render_7physics_BuoyancyForceApplier->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_6render_7physics_BuoyancyForceApplier, __pyx_vtabptr_7flitter_6render_7physics_BuoyancyForceApplier) < 0) __PYX_ERR(1, 245, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_6render_7physics_BuoyancyForceApplier) < 0) __PYX_ERR(1, 245, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_BuoyancyForceApplier, (PyObject *) __pyx_ptype_7flitter_6render_7physics_BuoyancyForceApplier) < 0) __PYX_ERR(1, 245, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_6render_7physics_BuoyancyForceApplier) < 0) __PYX_ERR(1, 245, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_6render_7physics_ConstantForceApplier = &__pyx_vtable_7flitter_6render_7physics_ConstantForceApplier;
  __pyx_vtable_7flitter_6render_7physics_ConstantForceApplier.__pyx_base = *__pyx_vtabptr_7flitter_6render_7physics_ParticleForceApplier;
  __pyx_vtable_7flitter_6render_7physics_ConstantForceApplier.__pyx_base.apply = (void (*)(struct __pyx_obj_7flitter_6render_7physics_ParticleForceApplier *, struct __pyx_obj_7flitter_6render_7physics_Particle *, double))__pyx_f_7flitter_6render_7physics_20ConstantForceApplier_apply;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_6render_7physics_ParticleForceApplier); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_6render_7physics_ConstantForceApplier = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_6render_7physics_ConstantForceApplier_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_6render_7physics_ConstantForceApplier)) __PYX_ERR(1, 268, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_6render_7physics_ConstantForceApplier_spec, __pyx_ptype_7flitter_6render_7physics_ConstantForceApplier) < 0) __PYX_ERR(1, 268, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_6render_7physics_ConstantForceApplier = &__pyx_type_7flitter_6render_7physics_ConstantForceApplier;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_6render_7physics_ConstantForceApplier->tp_base = __pyx_ptype_7flitter_6render_7physics_ParticleForceApplier;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_6render_7physics_ConstantForceApplier) < 0) __PYX_ERR(1, 268, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_6render_7physics_ConstantForceApplier->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_6render_7physics_ConstantForceApplier->tp_dictoffset && __pyx_ptype_7flitter_6render_7physics_ConstantForceApplier->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_6render_7physics_ConstantForceApplier->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_6render_7physics_ConstantForceApplier, __pyx_vtabptr_7flitter_6render_7physics_ConstantForceApplier) < 0) __PYX_ERR(1, 268, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_6render_7physics_ConstantForceApplier) < 0) __PYX_ERR(1, 268, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ConstantForceApplier, (PyObject *) __pyx_ptype_7flitter_6render_7physics_ConstantForceApplier) < 0) __PYX_ERR(1, 268, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_6render_7physics_ConstantForceApplier) < 0) __PYX_ERR(1, 268, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_6render_7physics_RandomForceApplier = &__pyx_vtable_7flitter_6render_7physics_RandomForceApplier;
  __pyx_vtable_7flitter_6render_7physics_RandomForceApplier.__pyx_base = *__pyx_vtabptr_7flitter_6render_7physics_ParticleForceApplier;
  __pyx_vtable_7flitter_6render_7physics_RandomForceApplier.__pyx_base.apply = (void (*)(struct __pyx_obj_7flitter_6render_7physics_ParticleForceApplier *, struct __pyx_obj_7flitter_6render_7physics_Particle *, double))__pyx_f_7flitter_6render_7physics_18RandomForceApplier_apply;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_6render_7physics_ParticleForceApplier); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_6render_7physics_RandomForceApplier = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_6render_7physics_RandomForceApplier_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_6render_7physics_RandomForceApplier)) __PYX_ERR(1, 286, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_6render_7physics_RandomForceApplier_spec, __pyx_ptype_7flitter_6render_7physics_RandomForceApplier) < 0) __PYX_ERR(1, 286, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_6render_7physics_RandomForceApplier = &__pyx_type_7flitter_6render_7physics_RandomForceApplier;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_6render_7physics_RandomForceApplier->tp_base = __pyx_ptype_7flitter_6render_7physics_ParticleForceApplier;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_6render_7physics_RandomForceApplier) < 0) __PYX_ERR(1, 286, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_6render_7physics_RandomForceApplier->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_6render_7physics_RandomForceApplier->tp_dictoffset && __pyx_ptype_7flitter_6render_7physics_RandomForceApplier->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_6render_7physics_RandomForceApplier->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_6render_7physics_RandomForceApplier, __pyx_vtabptr_7flitter_6render_7physics_RandomForceApplier) < 0) __PYX_ERR(1, 286, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_6render_7physics_RandomForceApplier) < 0) __PYX_ERR(1, 286, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_RandomForceApplier, (PyObject *) __pyx_ptype_7flitter_6render_7physics_RandomForceApplier) < 0) __PYX_ERR(1, 286, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_6render_7physics_RandomForceApplier) < 0) __PYX_ERR(1, 286, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_6render_7physics_FieldForceApplier = &__pyx_vtable_7flitter_6render_7physics_FieldForceApplier;
  __pyx_vtable_7flitter_6render_7physics_FieldForceApplier.__pyx_base = *__pyx_vtabptr_7flitter_6render_7physics_ParticleForceApplier;
  __pyx_vtable_7flitter_6render_7physics_FieldForceApplier.__pyx_base.apply = (void (*)(struct __pyx_obj_7flitter_6render_7physics_ParticleForceApplier *, struct __pyx_obj_7flitter_6render_7physics_Particle *, double))__pyx_f_7flitter_6render_7physics_17FieldForceApplier_apply;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_6render_7physics_ParticleForceApplier); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_6render_7physics_FieldForceApplier = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_6render_7physics_FieldForceApplier_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_6render_7physics_FieldForceApplier)) __PYX_ERR(1, 302, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_6render_7physics_FieldForceApplier_spec, __pyx_ptype_7flitter_6render_7physics_FieldForceApplier) < 0) __PYX_ERR(1, 302, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_6render_7physics_FieldForceApplier = &__pyx_type_7flitter_6render_7physics_FieldForceApplier;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_6render_7physics_FieldForceApplier->tp_base = __pyx_ptype_7flitter_6render_7physics_ParticleForceApplier;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_6render_7physics_FieldForceApplier) < 0) __PYX_ERR(1, 302, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_6render_7physics_FieldForceApplier->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_6render_7physics_FieldForceApplier->tp_dictoffset && __pyx_ptype_7flitter_6render_7physics_FieldForceApplier->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_6render_7physics_FieldForceApplier->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_6render_7physics_FieldForceApplier, __pyx_vtabptr_7flitter_6render_7physics_FieldForceApplier) < 0) __PYX_ERR(1, 302, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_6render_7physics_FieldForceApplier) < 0) __PYX_ERR(1, 302, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_FieldForceApplier, (PyObject *) __pyx_ptype_7flitter_6render_7physics_FieldForceApplier) < 0) __PYX_ERR(1, 302, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_6render_7physics_FieldForceApplier) < 0) __PYX_ERR(1, 302, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_6render_7physics_CollisionForceApplier = &__pyx_vtable_7flitter_6render_7physics_CollisionForceApplier;
  __pyx_vtable_7flitter_6render_7physics_CollisionForceApplier.__pyx_base = *__pyx_vtabptr_7flitter_6render_7physics_MatrixForceApplier;
  __pyx_vtable_7flitter_6render_7physics_CollisionForceApplier.__pyx_base.apply = (void (*)(struct __pyx_obj_7flitter_6render_7physics_MatrixForceApplier *, struct __pyx_obj_7flitter_6render_7physics_Particle *, struct __pyx_obj_7flitter_6render_7physics_Particle *, struct __pyx_obj_7flitter_5model_Vector *, double))__pyx_f_7flitter_6render_7physics_21CollisionForceApplier_apply;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_6render_7physics_MatrixForceApplier); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 316, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_6render_7physics_CollisionForceApplier = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_6render_7physics_CollisionForceApplier_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_6render_7physics_CollisionForceApplier)) __PYX_ERR(1, 316, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_6render_7physics_CollisionForceApplier_spec, __pyx_ptype_7flitter_6render_7physics_CollisionForceApplier) < 0) __PYX_ERR(1, 316, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_6render_7physics_CollisionForceApplier = &__pyx_type_7flitter_6render_7physics_CollisionForceApplier;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_6render_7physics_CollisionForceApplier->tp_base = __pyx_ptype_7flitter_6render_7physics_MatrixForceApplier;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_6render_7physics_CollisionForceApplier) < 0) __PYX_ERR(1, 316, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_6render_7physics_CollisionForceApplier->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_6render_7physics_CollisionForceApplier->tp_dictoffset && __pyx_ptype_7flitter_6render_7physics_CollisionForceApplier->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_6render_7physics_CollisionForceApplier->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_6render_7physics_CollisionForceApplier, __pyx_vtabptr_7flitter_6render_7physics_CollisionForceApplier) < 0) __PYX_ERR(1, 316, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_6render_7physics_CollisionForceApplier) < 0) __PYX_ERR(1, 316, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_CollisionForceApplier, (PyObject *) __pyx_ptype_7flitter_6render_7physics_CollisionForceApplier) < 0) __PYX_ERR(1, 316, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_6render_7physics_CollisionForceApplier) < 0) __PYX_ERR(1, 316, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_6render_7physics_GravityForceApplier = &__pyx_vtable_7flitter_6render_7physics_GravityForceApplier;
  __pyx_vtable_7flitter_6render_7physics_GravityForceApplier.__pyx_base = *__pyx_vtabptr_7flitter_6render_7physics_MatrixForceApplier;
  __pyx_vtable_7flitter_6render_7physics_GravityForceApplier.__pyx_base.apply = (void (*)(struct __pyx_obj_7flitter_6render_7physics_MatrixForceApplier *, struct __pyx_obj_7flitter_6render_7physics_Particle *, struct __pyx_obj_7flitter_6render_7physics_Particle *, struct __pyx_obj_7flitter_5model_Vector *, double))__pyx_f_7flitter_6render_7physics_19GravityForceApplier_apply;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_6render_7physics_MatrixForceApplier); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 340, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_6render_7physics_GravityForceApplier = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_6render_7physics_GravityForceApplier_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_6render_7physics_GravityForceApplier)) __PYX_ERR(1, 340, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_6render_7physics_GravityForceApplier_spec, __pyx_ptype_7flitter_6render_7physics_GravityForceApplier) < 0) __PYX_ERR(1, 340, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_6render_7physics_GravityForceApplier = &__pyx_type_7flitter_6render_7physics_GravityForceApplier;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_6render_7physics_GravityForceApplier->tp_base = __pyx_ptype_7flitter_6render_7physics_MatrixForceApplier;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_6render_7physics_GravityForceApplier) < 0) __PYX_ERR(1, 340, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_6render_7physics_GravityForceApplier->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_6render_7physics_GravityForceApplier->tp_dictoffset && __pyx_ptype_7flitter_6render_7physics_GravityForceApplier->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_6render_7physics_GravityForceApplier->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_6render_7physics_GravityForceApplier, __pyx_vtabptr_7flitter_6render_7physics_GravityForceApplier) < 0) __PYX_ERR(1, 340, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_6render_7physics_GravityForceApplier) < 0) __PYX_ERR(1, 340, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_GravityForceApplier, (PyObject *) __pyx_ptype_7flitter_6render_7physics_GravityForceApplier) < 0) __PYX_ERR(1, 340, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_6render_7physics_GravityForceApplier) < 0) __PYX_ERR(1, 340, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_6render_7physics_ElectrostaticForceApplier = &__pyx_vtable_7flitter_6render_7physics_ElectrostaticForceApplier;
  __pyx_vtable_7flitter_6render_7physics_ElectrostaticForceApplier.__pyx_base = *__pyx_vtabptr_7flitter_6render_7physics_MatrixForceApplier;
  __pyx_vtable_7flitter_6render_7physics_ElectrostaticForceApplier.__pyx_base.apply = (void (*)(struct __pyx_obj_7flitter_6render_7physics_MatrixForceApplier *, struct __pyx_obj_7flitter_6render_7physics_Particle *, struct __pyx_obj_7flitter_6render_7physics_Particle *, struct __pyx_obj_7flitter_5model_Vector *, double))__pyx_f_7flitter_6render_7physics_25ElectrostaticForceApplier_apply;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_6render_7physics_MatrixForceApplier); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_6render_7physics_ElectrostaticForceApplier = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_6render_7physics_ElectrostaticForceApplier_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_6render_7physics_ElectrostaticForceApplier)) __PYX_ERR(1, 354, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_6render_7physics_ElectrostaticForceApplier_spec, __pyx_ptype_7flitter_6render_7physics_ElectrostaticForceApplier) < 0) __PYX_ERR(1, 354, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_6render_7physics_ElectrostaticForceApplier = &__pyx_type_7flitter_6render_7physics_ElectrostaticForceApplier;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_6render_7physics_ElectrostaticForceApplier->tp_base = __pyx_ptype_7flitter_6render_7physics_MatrixForceApplier;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_6render_7physics_ElectrostaticForceApplier) < 0) __PYX_ERR(1, 354, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_6render_7physics_ElectrostaticForceApplier->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_6render_7physics_ElectrostaticForceApplier->tp_dictoffset && __pyx_ptype_7flitter_6render_7physics_ElectrostaticForceApplier->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_6render_7physics_ElectrostaticForceApplier->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_6render_7physics_ElectrostaticForceApplier, __pyx_vtabptr_7flitter_6render_7physics_ElectrostaticForceApplier) < 0) __PYX_ERR(1, 354, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_6render_7physics_ElectrostaticForceApplier) < 0) __PYX_ERR(1, 354, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ElectrostaticForceApplier, (PyObject *) __pyx_ptype_7flitter_6render_7physics_ElectrostaticForceApplier) < 0) __PYX_ERR(1, 354, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_6render_7physics_ElectrostaticForceApplier) < 0) __PYX_ERR(1, 354, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_6render_7physics_AdhesionForceApplier = &__pyx_vtable_7flitter_6render_7physics_AdhesionForceApplier;
  __pyx_vtable_7flitter_6render_7physics_AdhesionForceApplier.__pyx_base = *__pyx_vtabptr_7flitter_6render_7physics_MatrixForceApplier;
  __pyx_vtable_7flitter_6render_7physics_AdhesionForceApplier.__pyx_base.apply = (void (*)(struct __pyx_obj_7flitter_6render_7physics_MatrixForceApplier *, struct __pyx_obj_7flitter_6render_7physics_Particle *, struct __pyx_obj_7flitter_6render_7physics_Particle *, struct __pyx_obj_7flitter_5model_Vector *, double))__pyx_f_7flitter_6render_7physics_20AdhesionForceApplier_apply;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_7flitter_6render_7physics_MatrixForceApplier); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_6render_7physics_AdhesionForceApplier = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_6render_7physics_AdhesionForceApplier_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_ptype_7flitter_6render_7physics_AdhesionForceApplier)) __PYX_ERR(1, 368, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_6render_7physics_AdhesionForceApplier_spec, __pyx_ptype_7flitter_6render_7physics_AdhesionForceApplier) < 0) __PYX_ERR(1, 368, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_6render_7physics_AdhesionForceApplier = &__pyx_type_7flitter_6render_7physics_AdhesionForceApplier;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_7flitter_6render_7physics_AdhesionForceApplier->tp_base = __pyx_ptype_7flitter_6render_7physics_MatrixForceApplier;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_6render_7physics_AdhesionForceApplier) < 0) __PYX_ERR(1, 368, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_6render_7physics_AdhesionForceApplier->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_6render_7physics_AdhesionForceApplier->tp_dictoffset && __pyx_ptype_7flitter_6render_7physics_AdhesionForceApplier->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_6render_7physics_AdhesionForceApplier->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_6render_7physics_AdhesionForceApplier, __pyx_vtabptr_7flitter_6render_7physics_AdhesionForceApplier) < 0) __PYX_ERR(1, 368, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_6render_7physics_AdhesionForceApplier) < 0) __PYX_ERR(1, 368, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_AdhesionForceApplier, (PyObject *) __pyx_ptype_7flitter_6render_7physics_AdhesionForceApplier) < 0) __PYX_ERR(1, 368, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_6render_7physics_AdhesionForceApplier) < 0) __PYX_ERR(1, 368, __pyx_L1_error)
  #endif
  __pyx_vtabptr_7flitter_6render_7physics_PhysicsSystem = &__pyx_vtable_7flitter_6render_7physics_PhysicsSystem;
  __pyx_vtable_7flitter_6render_7physics_PhysicsSystem.collect = (void (*)(struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem *, PyObject *, struct __pyx_obj_7flitter_6render_7physics_PhysicsGroup *, struct __pyx_obj_7flitter_5model_Node *, PyObject *, PyObject *, PyObject *, struct __pyx_obj_7flitter_5model_StateDict *, struct __pyx_obj_7flitter_5model_Vector *, double, struct __pyx_obj_7flitter_5model_Vector *))__pyx_f_7flitter_6render_7physics_13PhysicsSystem_collect;
  __pyx_vtable_7flitter_6render_7physics_PhysicsSystem.calculate = (double (*)(struct __pyx_obj_7flitter_6render_7physics_PhysicsSystem *, PyObject *, PyObject *, int, int, double, double, double, int __pyx_skip_dispatch))__pyx_f_7flitter_6render_7physics_13PhysicsSystem_calculate;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_7flitter_6render_7physics_PhysicsSystem = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_6render_7physics_PhysicsSystem_spec, NULL); if (unlikely(!__pyx_ptype_7flitter_6render_7physics_PhysicsSystem)) __PYX_ERR(1, 392, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_6render_7physics_PhysicsSystem_spec, __pyx_ptype_7flitter_6render_7physics_PhysicsSystem) < 0) __PYX_ERR(1, 392, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_6render_7physics_PhysicsSystem = &__pyx_type_7flitter_6render_7physics_PhysicsSystem;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_6render_7physics_PhysicsSystem) < 0) __PYX_ERR(1, 392, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_6render_7physics_PhysicsSystem->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_6render_7physics_PhysicsSystem->tp_dictoffset && __pyx_ptype_7flitter_6render_7physics_PhysicsSystem->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_6render_7physics_PhysicsSystem->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_7flitter_6render_7physics_PhysicsSystem, __pyx_vtabptr_7flitter_6render_7physics_PhysicsSystem) < 0) __PYX_ERR(1, 392, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_7flitter_6render_7physics_PhysicsSystem) < 0) __PYX_ERR(1, 392, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_PhysicsSystem, (PyObject *) __pyx_ptype_7flitter_6render_7physics_PhysicsSystem) < 0) __PYX_ERR(1, 392, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_7flitter_6render_7physics_PhysicsSystem) < 0) __PYX_ERR(1, 392, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_7flitter_6render_7physics___pyx_scope_struct__destroy = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_6render_7physics___pyx_scope_struct__destroy_spec, NULL); if (unlikely(!__pyx_ptype_7flitter_6render_7physics___pyx_scope_struct__destroy)) __PYX_ERR(1, 401, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_6render_7physics___pyx_scope_struct__destroy_spec, __pyx_ptype_7flitter_6render_7physics___pyx_scope_struct__destroy) < 0) __PYX_ERR(1, 401, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_6render_7physics___pyx_scope_struct__destroy = &__pyx_type_7flitter_6render_7physics___pyx_scope_struct__destroy;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_6render_7physics___pyx_scope_struct__destroy) < 0) __PYX_ERR(1, 401, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_6render_7physics___pyx_scope_struct__destroy->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_6render_7physics___pyx_scope_struct__destroy->tp_dictoffset && __pyx_ptype_7flitter_6render_7physics___pyx_scope_struct__destroy->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_6render_7physics___pyx_scope_struct__destroy->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_7flitter_6render_7physics___pyx_scope_struct_1_purge = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_6render_7physics___pyx_scope_struct_1_purge_spec, NULL); if (unlikely(!__pyx_ptype_7flitter_6render_7physics___pyx_scope_struct_1_purge)) __PYX_ERR(1, 404, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_6render_7physics___pyx_scope_struct_1_purge_spec, __pyx_ptype_7flitter_6render_7physics___pyx_scope_struct_1_purge) < 0) __PYX_ERR(1, 404, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_6render_7physics___pyx_scope_struct_1_purge = &__pyx_type_7flitter_6render_7physics___pyx_scope_struct_1_purge;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_6render_7physics___pyx_scope_struct_1_purge) < 0) __PYX_ERR(1, 404, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_6render_7physics___pyx_scope_struct_1_purge->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_6render_7physics___pyx_scope_struct_1_purge->tp_dictoffset && __pyx_ptype_7flitter_6render_7physics___pyx_scope_struct_1_purge->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_6render_7physics___pyx_scope_struct_1_purge->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_7flitter_6render_7physics___pyx_scope_struct_2_update = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7flitter_6render_7physics___pyx_scope_struct_2_update_spec, NULL); if (unlikely(!__pyx_ptype_7flitter_6render_7physics___pyx_scope_struct_2_update)) __PYX_ERR(1, 407, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7flitter_6render_7physics___pyx_scope_struct_2_update_spec, __pyx_ptype_7flitter_6render_7physics___pyx_scope_struct_2_update) < 0) __PYX_ERR(1, 407, __pyx_L1_error)
  #else
  __pyx_ptype_7flitter_6render_7physics___pyx_scope_struct_2_update = &__pyx_type_7flitter_6render_7physics___pyx_scope_struct_2_update;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_7flitter_6render_7physics___pyx_scope_struct_2_update) < 0) __PYX_ERR(1, 407, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_7flitter_6render_7physics___pyx_scope_struct_2_update->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_7flitter_6render_7physics___pyx_scope_struct_2_update->tp_dictoffset && __pyx_ptype_7flitter_6render_7physics___pyx_scope_struct_2_update->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_7flitter_6render_7physics___pyx_scope_struct_2_update->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule("flitter.model"); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_5model_Vector = __Pyx_ImportType_3_0_11(__pyx_t_1, "flitter.model", "Vector", sizeof(struct __pyx_obj_7flitter_5model_Vector), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_7flitter_5model_Vector),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7flitter_5model_Vector) __PYX_ERR(3, 35, __pyx_L1_error)
  __pyx_vtabptr_7flitter_5model_Vector = (struct __pyx_vtabstruct_7flitter_5model_Vector*)__Pyx_GetVtable(__pyx_ptype_7flitter_5model_Vector); if (unlikely(!__pyx_vtabptr_7flitter_5model_Vector)) __PYX_ERR(3, 35, __pyx_L1_error)
  __pyx_ptype_7flitter_5model_Matrix33 = __Pyx_ImportType_3_0_11(__pyx_t_1, "flitter.model", "Matrix33", sizeof(struct __pyx_obj_7flitter_5model_Matrix33), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_7flitter_5model_Matrix33),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7flitter_5model_Matrix33) __PYX_ERR(3, 107, __pyx_L1_error)
  __pyx_vtabptr_7flitter_5model_Matrix33 = (struct __pyx_vtabstruct_7flitter_5model_Matrix33*)__Pyx_GetVtable(__pyx_ptype_7flitter_5model_Matrix33); if (unlikely(!__pyx_vtabptr_7flitter_5model_Matrix33)) __PYX_ERR(3, 107, __pyx_L1_error)
  __pyx_ptype_7flitter_5model_Matrix44 = __Pyx_ImportType_3_0_11(__pyx_t_1, "flitter.model", "Matrix44", sizeof(struct __pyx_obj_7flitter_5model_Matrix44), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_7flitter_5model_Matrix44),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7flitter_5model_Matrix44) __PYX_ERR(3, 130, __pyx_L1_error)
  __pyx_vtabptr_7flitter_5model_Matrix44 = (struct __pyx_vtabstruct_7flitter_5model_Matrix44*)__Pyx_GetVtable(__pyx_ptype_7flitter_5model_Matrix44); if (unlikely(!__pyx_vtabptr_7flitter_5model_Matrix44)) __PYX_ERR(3, 130, __pyx_L1_error)
  __pyx_ptype_7flitter_5model_Quaternion = __Pyx_ImportType_3_0_11(__pyx_t_1, "flitter.model", "Quaternion", sizeof(struct __pyx_obj_7flitter_5model_Quaternion), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_7flitter_5model_Quaternion),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7flitter_5model_Quaternion) __PYX_ERR(3, 181, __pyx_L1_error)
  __pyx_vtabptr_7flitter_5model_Quaternion = (struct __pyx_vtabstruct_7flitter_5model_Quaternion*)__Pyx_GetVtable(__pyx_ptype_7flitter_5model_Quaternion); if (unlikely(!__pyx_vtabptr_7flitter_5model_Quaternion)) __PYX_ERR(3, 181, __pyx_L1_error)
  __pyx_ptype_7flitter_5model_Node = __Pyx_ImportType_3_0_11(__pyx_t_1, "flitter.model", "Node", sizeof(struct __pyx_obj_7flitter_5model_Node), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_7flitter_5model_Node),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7flitter_5model_Node) __PYX_ERR(3, 201, __pyx_L1_error)
  __pyx_vtabptr_7flitter_5model_Node = (struct __pyx_vtabstruct_7flitter_5model_Node*)__Pyx_GetVtable(__pyx_ptype_7flitter_5model_Node); if (unlikely(!__pyx_vtabptr_7flitter_5model_Node)) __PYX_ERR(3, 201, __pyx_L1_error)
  __pyx_ptype_7flitter_5model_StateDict = __Pyx_ImportType_3_0_11(__pyx_t_1, "flitter.model", "StateDict", sizeof(struct __pyx_obj_7flitter_5model_StateDict), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_7flitter_5model_StateDict),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7flitter_5model_StateDict) __PYX_ERR(3, 225, __pyx_L1_error)
  __pyx_vtabptr_7flitter_5model_StateDict = (struct __pyx_vtabstruct_7flitter_5model_StateDict*)__Pyx_GetVtable(__pyx_ptype_7flitter_5model_StateDict); if (unlikely(!__pyx_vtabptr_7flitter_5model_StateDict)) __PYX_ERR(3, 225, __pyx_L1_error)
  __pyx_ptype_7flitter_5model_Context = __Pyx_ImportType_3_0_11(__pyx_t_1, "flitter.model", "Context", sizeof(struct __pyx_obj_7flitter_5model_Context), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_7flitter_5model_Context),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7flitter_5model_Context) __PYX_ERR(3, 234, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("flitter.language.functions"); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7flitter_8language_9functions_uniform = __Pyx_ImportType_3_0_11(__pyx_t_1, "flitter.language.functions", "uniform", sizeof(struct __pyx_obj_7flitter_8language_9functions_uniform), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_7flitter_8language_9functions_uniform),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7flitter_8language_9functions_uniform) __PYX_ERR(4, 5, __pyx_L1_error)
  __pyx_vtabptr_7flitter_8language_9functions_uniform = (struct __pyx_vtabstruct_7flitter_8language_9functions_uniform*)__Pyx_GetVtable(__pyx_ptype_7flitter_8language_9functions_uniform); if (unlikely(!__pyx_vtabptr_7flitter_8language_9functions_uniform)) __PYX_ERR(4, 5, __pyx_L1_error)
  __pyx_ptype_7flitter_8language_9functions_beta = __Pyx_ImportType_3_0_11(__pyx_t_1, "flitter.language.functions", "beta", sizeof(struct __pyx_obj_7flitter_8language_9functions_beta), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_7flitter_8language_9functions_beta),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7flitter_8language_9functions_beta) __PYX_ERR(4, 11, __pyx_L1_error)
  __pyx_vtabptr_7flitter_8language_9functions_beta = (struct __pyx_vtabstruct_7flitter_8language_9functions_beta*)__Pyx_GetVtable(__pyx_ptype_7flitter_8language_9functions_beta); if (unlikely(!__pyx_vtabptr_7flitter_8language_9functions_beta)) __PYX_ERR(4, 11, __pyx_L1_error)
  __pyx_ptype_7flitter_8language_9functions_normal = __Pyx_ImportType_3_0_11(__pyx_t_1, "flitter.language.functions", "normal", sizeof(struct __pyx_obj_7flitter_8language_9functions_normal), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(struct __pyx_obj_7flitter_8language_9functions_normal),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7flitter_8language_9functions_normal) __PYX_ERR(4, 15, __pyx_L1_error)
  __pyx_vtabptr_7flitter_8language_9functions_normal = (struct __pyx_vtabstruct_7flitter_8language_9functions_normal*)__Pyx_GetVtable(__pyx_ptype_7flitter_8language_9functions_normal); if (unlikely(!__pyx_vtabptr_7flitter_8language_9functions_normal)) __PYX_ERR(4, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType_3_0_11(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(PyTypeObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(5, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4bool_bool = __Pyx_ImportType_3_0_11(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "bool", sizeof(PyBoolObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(PyBoolObject),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7cpython_4bool_bool) __PYX_ERR(6, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_7complex_complex = __Pyx_ImportType_3_0_11(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "complex", sizeof(PyComplexObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_11(PyComplexObject),__Pyx_ImportType_CheckSize_Warn_3_0_11); if (!__pyx_ptype_7cpython_7complex_complex) __PYX_ERR(7, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __pyx_t_1 = PyImport_ImportModule("flitter.model"); if (!__pyx_t_1) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "HASH_START", (void **)&__pyx_vp_7flitter_5model_HASH_START, "uint64_t") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "null_", (void **)&__pyx_vp_7flitter_5model_null_, "struct __pyx_obj_7flitter_5model_Vector *") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "true_", (void **)&__pyx_vp_7flitter_5model_true_, "struct __pyx_obj_7flitter_5model_Vector *") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "false_", (void **)&__pyx_vp_7flitter_5model_false_, "struct __pyx_obj_7flitter_5model_Vector *") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "minusone_", (void **)&__pyx_vp_7flitter_5model_minusone_, "struct __pyx_obj_7flitter_5model_Vector *") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "inf_", (void **)&__pyx_vp_7flitter_5model_inf_, "struct __pyx_obj_7flitter_5model_Vector *") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr_3_0_11(__pyx_t_1, "nan_", (void **)&__pyx_vp_7flitter_5model_nan_, "struct __pyx_obj_7flitter_5model_Vector *") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_physics(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_physics},
  {0, NULL}
};
#endif

#ifdef __cplusplus
namespace {
  struct PyModuleDef __pyx_moduledef =
  #else
  static struct PyModuleDef __pyx_moduledef =
  #endif
  {
      PyModuleDef_HEAD_INIT,
      "physics",
      __pyx_k_Flitter_physics_engine, /* m_doc */
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      0, /* m_size */
    #elif CYTHON_USE_MODULE_STATE
      sizeof(__pyx_mstate), /* m_size */
    #else
      -1, /* m_size */
    #endif
      __pyx_methods /* m_methods */,
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      __pyx_moduledef_slots, /* m_slots */
    #else
      NULL, /* m_reload */
    #endif
    #if CYTHON_USE_MODULE_STATE
      __pyx_m_traverse, /* m_traverse */
      __pyx_m_clear, /* m_clear */
      NULL /* m_free */
    #else
      NULL, /* m_traverse */
      NULL, /* m_clear */
      NULL /* m_free */
    #endif
  };
  #ifdef __cplusplus
} /* anonymous namespace */
#endif
#endif

#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initphysics(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initphysics(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_physics(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_physics(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *module, const char* from_name, const char* to_name, int allow_none)
#else
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none)
#endif
{
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
#if CYTHON_COMPILING_IN_LIMITED_API
            result = PyModule_AddObject(module, to_name, value);
#else
            result = PyDict_SetItemString(moddict, to_name, value);
#endif
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    CYTHON_UNUSED_VAR(def);
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    moddict = module;
#else
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
#endif
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_physics(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  int stringtab_initialized = 0;
  #if CYTHON_USE_MODULE_STATE
  int pystate_addmodule_run = 0;
  #endif
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'physics' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("physics", __pyx_methods, __pyx_k_Flitter_physics_engine, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  if (unlikely(!__pyx_m)) __PYX_ERR(1, 1, __pyx_L1_error)
  #elif CYTHON_USE_MODULE_STATE
  __pyx_t_1 = PyModule_Create(&__pyx_moduledef); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1, __pyx_L1_error)
  {
    int add_module_result = PyState_AddModule(__pyx_t_1, &__pyx_moduledef);
    __pyx_t_1 = 0; /* transfer ownership from __pyx_t_1 to "physics" pseudovariable */
    if (unlikely((add_module_result < 0))) __PYX_ERR(1, 1, __pyx_L1_error)
    pystate_addmodule_run = 1;
  }
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  if (unlikely(!__pyx_m)) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  #endif
  CYTHON_UNUSED_VAR(__pyx_t_1);
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(1, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = __Pyx_PyImport_AddModuleRef(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_cython_runtime = __Pyx_PyImport_AddModuleRef((const char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(1, 1, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_physics(void)", 0);
  if (__Pyx_check_binary_version(__PYX_LIMITED_VERSION_HEX, __Pyx_get_runtime_version(), CYTHON_COMPILING_IN_LIMITED_API) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(1, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init(__pyx_m) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init(__pyx_m) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init(__pyx_m) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init(__pyx_m) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init(__pyx_m) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init(__pyx_m) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitConstants() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  stringtab_initialized = 1;
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_flitter__render__physics) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(1, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "flitter.render.physics")) {
      if (unlikely((PyDict_SetItemString(modules, "flitter.render.physics", __pyx_m) < 0))) __PYX_ERR(1, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely((__Pyx_modinit_type_init_code() < 0))) __PYX_ERR(1, 1, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_type_import_code() < 0))) __PYX_ERR(1, 1, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_variable_import_code() < 0))) __PYX_ERR(1, 1, __pyx_L1_error)
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif

  /* "flitter/render/physics.pyx":7
 * """
 * 
 * import asyncio             # <<<<<<<<<<<<<<
 * from loguru import logger
 * 
 */
  __pyx_t_2 = __Pyx_patch_asyncio(__Pyx_ImportDottedModule(__pyx_n_s_asyncio, NULL)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_asyncio, __pyx_t_2) < 0) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/render/physics.pyx":8
 * 
 * import asyncio
 * from loguru import logger             # <<<<<<<<<<<<<<
 * 
 * from .. import name_patch
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_logger);
  __Pyx_GIVEREF(__pyx_n_s_logger);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_logger)) __PYX_ERR(1, 8, __pyx_L1_error);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_loguru, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_logger, __pyx_t_2) < 0) __PYX_ERR(1, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "flitter/render/physics.pyx":10
 * from loguru import logger
 * 
 * from .. import name_patch             # <<<<<<<<<<<<<<
 * from ..model cimport Vector, Node, StateDict, null_, false_
 * from ..language.functions cimport normal
 */
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_name_patch);
  __Pyx_GIVEREF(__pyx_n_s_name_patch);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_name_patch)) __PYX_ERR(1, 10, __pyx_L1_error);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s__9, __pyx_t_3, 2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_name_patch); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_name_patch, __pyx_t_3) < 0) __PYX_ERR(1, 10, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "flitter/render/physics.pyx":28
 *     PyObject* PyList_GET_ITEM(object list, Py_ssize_t i) nogil
 * 
 * logger = name_patch(logger, __name__)             # <<<<<<<<<<<<<<
 * 
 * cdef Vector VELOCITY = Vector._symbol('velocity')
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_name_patch); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_logger); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(1, 28, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4)) __PYX_ERR(1, 28, __pyx_L1_error);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_logger, __pyx_t_4) < 0) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "flitter/render/physics.pyx":30
 * logger = name_patch(logger, __name__)
 * 
 * cdef Vector VELOCITY = Vector._symbol('velocity')             # <<<<<<<<<<<<<<
 * cdef Vector LAST = Vector._symbol('last')
 * cdef Vector CLOCK = Vector._symbol('clock')
 */
  __pyx_t_4 = ((PyObject *)__pyx_vtabptr_7flitter_5model_Vector->_symbol(__pyx_n_u_velocity)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF((PyObject *)__pyx_v_7flitter_6render_7physics_VELOCITY);
  __Pyx_DECREF_SET(__pyx_v_7flitter_6render_7physics_VELOCITY, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "flitter/render/physics.pyx":31
 * 
 * cdef Vector VELOCITY = Vector._symbol('velocity')
 * cdef Vector LAST = Vector._symbol('last')             # <<<<<<<<<<<<<<
 * cdef Vector CLOCK = Vector._symbol('clock')
 * cdef Vector RUN = Vector._symbol('run')
 */
  __pyx_t_4 = ((PyObject *)__pyx_vtabptr_7flitter_5model_Vector->_symbol(__pyx_n_u_last)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF((PyObject *)__pyx_v_7flitter_6render_7physics_LAST);
  __Pyx_DECREF_SET(__pyx_v_7flitter_6render_7physics_LAST, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "flitter/render/physics.pyx":32
 * cdef Vector VELOCITY = Vector._symbol('velocity')
 * cdef Vector LAST = Vector._symbol('last')
 * cdef Vector CLOCK = Vector._symbol('clock')             # <<<<<<<<<<<<<<
 * cdef Vector RUN = Vector._symbol('run')
 * cdef Vector ITERATION = Vector._symbol('iteration')
 */
  __pyx_t_4 = ((PyObject *)__pyx_vtabptr_7flitter_5model_Vector->_symbol(__pyx_n_u_clock)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF((PyObject *)__pyx_v_7flitter_6render_7physics_CLOCK);
  __Pyx_DECREF_SET(__pyx_v_7flitter_6render_7physics_CLOCK, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "flitter/render/physics.pyx":33
 * cdef Vector LAST = Vector._symbol('last')
 * cdef Vector CLOCK = Vector._symbol('clock')
 * cdef Vector RUN = Vector._symbol('run')             # <<<<<<<<<<<<<<
 * cdef Vector ITERATION = Vector._symbol('iteration')
 * 
 */
  __pyx_t_4 = ((PyObject *)__pyx_vtabptr_7flitter_5model_Vector->_symbol(__pyx_n_u_run)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF((PyObject *)__pyx_v_7flitter_6render_7physics_RUN);
  __Pyx_DECREF_SET(__pyx_v_7flitter_6render_7physics_RUN, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "flitter/render/physics.pyx":34
 * cdef Vector CLOCK = Vector._symbol('clock')
 * cdef Vector RUN = Vector._symbol('run')
 * cdef Vector ITERATION = Vector._symbol('iteration')             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_4 = ((PyObject *)__pyx_vtabptr_7flitter_5model_Vector->_symbol(__pyx_n_u_iteration)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF((PyObject *)__pyx_v_7flitter_6render_7physics_ITERATION);
  __Pyx_DECREF_SET(__pyx_v_7flitter_6render_7physics_ITERATION, ((struct __pyx_obj_7flitter_5model_Vector *)__pyx_t_4));
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_6render_7physics_12PhysicsGroup_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PhysicsGroup___reduce_cython, NULL, __pyx_n_s_flitter_render_physics, __pyx_d, ((PyObject *)__pyx_codeobj__11)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_6render_7physics_12PhysicsGroup_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PhysicsGroup___setstate_cython, NULL, __pyx_n_s_flitter_render_physics, __pyx_d, ((PyObject *)__pyx_codeobj__13)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_6render_7physics_8Particle_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Particle___reduce_cython, NULL, __pyx_n_s_flitter_render_physics, __pyx_d, ((PyObject *)__pyx_codeobj__15)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_6render_7physics_Particle, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_6render_7physics_Particle);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Particle, (type(self), 0xc9134b9, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Particle__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_6render_7physics_8Particle_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Particle___setstate_cython, NULL, __pyx_n_s_flitter_render_physics, __pyx_d, ((PyObject *)__pyx_codeobj__16)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_6render_7physics_Particle, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_6render_7physics_Particle);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_6render_7physics_7Barrier_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Barrier___reduce_cython, NULL, __pyx_n_s_flitter_render_physics, __pyx_d, ((PyObject *)__pyx_codeobj__17)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_6render_7physics_7Barrier_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Barrier___setstate_cython, NULL, __pyx_n_s_flitter_render_physics, __pyx_d, ((PyObject *)__pyx_codeobj__18)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_6render_7physics_20ParticleForceApplier_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ParticleForceApplier___reduce_cy, NULL, __pyx_n_s_flitter_render_physics, __pyx_d, ((PyObject *)__pyx_codeobj__19)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_6render_7physics_20ParticleForceApplier_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ParticleForceApplier___setstate, NULL, __pyx_n_s_flitter_render_physics, __pyx_d, ((PyObject *)__pyx_codeobj__20)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_6render_7physics_18MatrixForceApplier_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_MatrixForceApplier___reduce_cyth, NULL, __pyx_n_s_flitter_render_physics, __pyx_d, ((PyObject *)__pyx_codeobj__21)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_6render_7physics_18MatrixForceApplier_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_MatrixForceApplier___setstate_cy, NULL, __pyx_n_s_flitter_render_physics, __pyx_d, ((PyObject *)__pyx_codeobj__22)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_6render_7physics_20DistanceForceApplier_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_DistanceForceApplier___reduce_cy, NULL, __pyx_n_s_flitter_render_physics, __pyx_d, ((PyObject *)__pyx_codeobj__23)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_6render_7physics_DistanceForceApplier, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_6render_7physics_DistanceForceApplier);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_DistanceForceApplier, (type(self), 0xca186ab, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_DistanceForceApplier__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_6render_7physics_20DistanceForceApplier_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_DistanceForceApplier___setstate, NULL, __pyx_n_s_flitter_render_physics, __pyx_d, ((PyObject *)__pyx_codeobj__24)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_6render_7physics_DistanceForceApplier, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_6render_7physics_DistanceForceApplier);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_6render_7physics_16DragForceApplier_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_DragForceApplier___reduce_cython, NULL, __pyx_n_s_flitter_render_physics, __pyx_d, ((PyObject *)__pyx_codeobj__25)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_6render_7physics_16DragForceApplier_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_DragForceApplier___setstate_cyth, NULL, __pyx_n_s_flitter_render_physics, __pyx_d, ((PyObject *)__pyx_codeobj__26)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_6render_7physics_20BuoyancyForceApplier_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_BuoyancyForceApplier___reduce_cy, NULL, __pyx_n_s_flitter_render_physics, __pyx_d, ((PyObject *)__pyx_codeobj__27)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_6render_7physics_20BuoyancyForceApplier_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_BuoyancyForceApplier___setstate, NULL, __pyx_n_s_flitter_render_physics, __pyx_d, ((PyObject *)__pyx_codeobj__28)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_6render_7physics_20ConstantForceApplier_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ConstantForceApplier___reduce_cy, NULL, __pyx_n_s_flitter_render_physics, __pyx_d, ((PyObject *)__pyx_codeobj__29)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_6render_7physics_20ConstantForceApplier_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ConstantForceApplier___setstate, NULL, __pyx_n_s_flitter_render_physics, __pyx_d, ((PyObject *)__pyx_codeobj__30)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_6render_7physics_18RandomForceApplier_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_RandomForceApplier___reduce_cyth, NULL, __pyx_n_s_flitter_render_physics, __pyx_d, ((PyObject *)__pyx_codeobj__31)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_6render_7physics_18RandomForceApplier_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_RandomForceApplier___setstate_cy, NULL, __pyx_n_s_flitter_render_physics, __pyx_d, ((PyObject *)__pyx_codeobj__32)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_6render_7physics_17FieldForceApplier_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_FieldForceApplier___reduce_cytho, NULL, __pyx_n_s_flitter_render_physics, __pyx_d, ((PyObject *)__pyx_codeobj__33)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_6render_7physics_17FieldForceApplier_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_FieldForceApplier___setstate_cyt, NULL, __pyx_n_s_flitter_render_physics, __pyx_d, ((PyObject *)__pyx_codeobj__34)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_6render_7physics_21CollisionForceApplier_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CollisionForceApplier___reduce_c, NULL, __pyx_n_s_flitter_render_physics, __pyx_d, ((PyObject *)__pyx_codeobj__35)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_6render_7physics_21CollisionForceApplier_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_CollisionForceApplier___setstate, NULL, __pyx_n_s_flitter_render_physics, __pyx_d, ((PyObject *)__pyx_codeobj__36)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_6render_7physics_19GravityForceApplier_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GravityForceApplier___reduce_cyt, NULL, __pyx_n_s_flitter_render_physics, __pyx_d, ((PyObject *)__pyx_codeobj__37)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_6render_7physics_19GravityForceApplier_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GravityForceApplier___setstate_c, NULL, __pyx_n_s_flitter_render_physics, __pyx_d, ((PyObject *)__pyx_codeobj__38)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_6render_7physics_25ElectrostaticForceApplier_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ElectrostaticForceApplier___redu, NULL, __pyx_n_s_flitter_render_physics, __pyx_d, ((PyObject *)__pyx_codeobj__39)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_6render_7physics_25ElectrostaticForceApplier_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ElectrostaticForceApplier___sets, NULL, __pyx_n_s_flitter_render_physics, __pyx_d, ((PyObject *)__pyx_codeobj__40)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_6render_7physics_20AdhesionForceApplier_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AdhesionForceApplier___reduce_cy, NULL, __pyx_n_s_flitter_render_physics, __pyx_d, ((PyObject *)__pyx_codeobj__41)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_6render_7physics_20AdhesionForceApplier_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AdhesionForceApplier___setstate, NULL, __pyx_n_s_flitter_render_physics, __pyx_d, ((PyObject *)__pyx_codeobj__42)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "flitter/render/physics.pyx":401
 *         self.state_keys = set()
 * 
 *     async def destroy(self):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_6render_7physics_13PhysicsSystem_3destroy, __Pyx_CYFUNCTION_CCLASS | __Pyx_CYFUNCTION_COROUTINE, __pyx_n_s_PhysicsSystem_destroy, NULL, __pyx_n_s_flitter_render_physics, __pyx_d, ((PyObject *)__pyx_codeobj_)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 401, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_6render_7physics_PhysicsSystem, __pyx_n_s_destroy, __pyx_t_4) < 0) __PYX_ERR(1, 401, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_6render_7physics_PhysicsSystem);

  /* "flitter/render/physics.pyx":404
 *         pass
 * 
 *     async def purge(self):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_6render_7physics_13PhysicsSystem_6purge, __Pyx_CYFUNCTION_CCLASS | __Pyx_CYFUNCTION_COROUTINE, __pyx_n_s_PhysicsSystem_purge, NULL, __pyx_n_s_flitter_render_physics, __pyx_d, ((PyObject *)__pyx_codeobj__2)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 404, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_6render_7physics_PhysicsSystem, __pyx_n_s_purge, __pyx_t_4) < 0) __PYX_ERR(1, 404, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_6render_7physics_PhysicsSystem);

  /* "flitter/render/physics.pyx":407
 *         pass
 * 
 *     async def update(self, engine, Node node, double time, double performance, **kwargs):             # <<<<<<<<<<<<<<
 *         cdef int64_t run = node.get_int('run', 0)
 *         self.dimensions = node.get_int('dimensions', 0)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_6render_7physics_13PhysicsSystem_9update, __Pyx_CYFUNCTION_CCLASS | __Pyx_CYFUNCTION_COROUTINE, __pyx_n_s_PhysicsSystem_update, NULL, __pyx_n_s_flitter_render_physics, __pyx_d, ((PyObject *)__pyx_codeobj__3)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 407, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_6render_7physics_PhysicsSystem, __pyx_n_s_update, __pyx_t_4) < 0) __PYX_ERR(1, 407, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_6render_7physics_PhysicsSystem);

  /* "flitter/render/physics.pyx":609
 *                     group.barriers.append(barrier)
 * 
 *     cpdef double calculate(self, list groups, dict particles_by_id, bint realtime, bint extra, double time, double last_time, double clock):             # <<<<<<<<<<<<<<
 *         cdef list all_particles, particles, forces, barriers
 *         cdef int64_t i, j, k, m, n, p
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_6render_7physics_13PhysicsSystem_12calculate, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PhysicsSystem_calculate, NULL, __pyx_n_s_flitter_render_physics, __pyx_d, ((PyObject *)__pyx_codeobj__45)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 609, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_6render_7physics_PhysicsSystem, __pyx_n_s_calculate, __pyx_t_4) < 0) __PYX_ERR(1, 609, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_6render_7physics_PhysicsSystem);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_6render_7physics_13PhysicsSystem_14__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PhysicsSystem___reduce_cython, NULL, __pyx_n_s_flitter_render_physics, __pyx_d, ((PyObject *)__pyx_codeobj__46)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_6render_7physics_PhysicsSystem, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_6render_7physics_PhysicsSystem);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_PhysicsSystem, (type(self), 0xf14ade3, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_PhysicsSystem__set_state(self, __pyx_state)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_6render_7physics_13PhysicsSystem_16__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PhysicsSystem___setstate_cython, NULL, __pyx_n_s_flitter_render_physics, __pyx_d, ((PyObject *)__pyx_codeobj__47)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_7flitter_6render_7physics_PhysicsSystem, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_7flitter_6render_7physics_PhysicsSystem);

  /* "(tree fragment)":1
 * def __pyx_unpickle_Particle(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_6render_7physics_1__pyx_unpickle_Particle, 0, __pyx_n_s_pyx_unpickle_Particle, NULL, __pyx_n_s_flitter_render_physics, __pyx_d, ((PyObject *)__pyx_codeobj__49)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Particle, __pyx_t_4) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Particle__set_state(<Particle> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Particle__set_state(Particle __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.acceleration = __pyx_state[0]; __pyx_result.charge = __pyx_state[1]; __pyx_result.ease = __pyx_state[2]; __pyx_result.force = __pyx_state[3]; __pyx_result.id = __pyx_state[4]; __pyx_result.initial_force = __pyx_state[5]; __pyx_result.mass = __pyx_state[6]; __pyx_result.non_anchor = __pyx_state[7]; __pyx_result.position = __pyx_state[8]; __pyx_result.position_state_key = __pyx_state[9]; __pyx_result.radius = __pyx_state[10]; __pyx_result.velocity = __pyx_state[11]; __pyx_result.velocity_state_key = __pyx_state[12]
 *     if len(__pyx_state) > 13 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_6render_7physics_3__pyx_unpickle_DistanceForceApplier, 0, __pyx_n_s_pyx_unpickle_DistanceForceAppl, NULL, __pyx_n_s_flitter_render_physics, __pyx_d, ((PyObject *)__pyx_codeobj__50)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_DistanceForceAppl, __pyx_t_4) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_PhysicsSystem(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_7flitter_6render_7physics_5__pyx_unpickle_PhysicsSystem, 0, __pyx_n_s_pyx_unpickle_PhysicsSystem, NULL, __pyx_n_s_flitter_render_physics, __pyx_d, ((PyObject *)__pyx_codeobj__51)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_PhysicsSystem, __pyx_t_4) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "flitter/render/physics.pyx":1
 * # cython: boundscheck=False, wraparound=False, cdivision=True, cpow=True             # <<<<<<<<<<<<<<
 * 
 * """
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  if (__pyx_m) {
    if (__pyx_d && stringtab_initialized) {
      __Pyx_AddTraceback("init flitter.render.physics", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    #if !CYTHON_USE_MODULE_STATE
    Py_CLEAR(__pyx_m);
    #else
    Py_DECREF(__pyx_m);
    if (pystate_addmodule_run) {
      PyObject *tp, *value, *tb;
      PyErr_Fetch(&tp, &value, &tb);
      PyState_RemoveModule(&__pyx_moduledef);
      PyErr_Restore(tp, value, tb);
    }
    #endif
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init flitter.render.physics");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}
/* #### Code section: cleanup_globals ### */
/* #### Code section: cleanup_module ### */
/* #### Code section: main_method ### */
/* #### Code section: utility_code_pragmas ### */
#ifdef _MSC_VER
#pragma warning( push )
/* Warning 4127: conditional expression is constant
 * Cython uses constant conditional expressions to allow in inline functions to be optimized at
 * compile-time, so this warning is not useful
 */
#pragma warning( disable : 4127 )
#endif



/* #### Code section: utility_code_def ### */

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    int result;
    PyObject *exc_type;
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *current_exception = tstate->current_exception;
    if (unlikely(!current_exception)) return 0;
    exc_type = (PyObject*) Py_TYPE(current_exception);
    if (exc_type == err) return 1;
#else
    exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
#endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(exc_type);
    #endif
    if (unlikely(PyTuple_Check(err))) {
        result = __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    } else {
        result = __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
    }
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(exc_type);
    #endif
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *tmp_value;
    assert(type == NULL || (value != NULL && type == (PyObject*) Py_TYPE(value)));
    if (value) {
        #if CYTHON_COMPILING_IN_CPYTHON
        if (unlikely(((PyBaseExceptionObject*) value)->traceback != tb))
        #endif
            PyException_SetTraceback(value, tb);
    }
    tmp_value = tstate->current_exception;
    tstate->current_exception = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
#else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject* exc_value;
    exc_value = tstate->current_exception;
    tstate->current_exception = 0;
    *value = exc_value;
    *type = NULL;
    *tb = NULL;
    if (exc_value) {
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        #if CYTHON_COMPILING_IN_CPYTHON
        *tb = ((PyBaseExceptionObject*) exc_value)->traceback;
        Py_XINCREF(*tb);
        #else
        *tb = PyException_GetTraceback(exc_value);
        #endif
    }
#else
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#endif
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* PyObjectGetAttrStrNoError */
#if __PYX_LIMITED_VERSION_HEX < 0x030d00A1
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d00A1
    (void) PyObject_GetOptionalAttr(obj, attr_name, &result);
    return result;
#else
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
#endif
}

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStrNoError(__pyx_b, name);
    if (unlikely(!result) && !PyErr_Occurred()) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* TupleAndListFromArray */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject *const *CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        Py_INCREF(__pyx_empty_tuple);
        return __pyx_empty_tuple;
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyTupleObject*)res)->ob_item, n);
    return res;
}
static CYTHON_INLINE PyObject *
__Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return PyList_New(0);
    }
    res = PyList_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyListObject*)res)->ob_item, n);
    return res;
}
#endif

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* fastcall */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s)
{
    Py_ssize_t i, n = PyTuple_GET_SIZE(kwnames);
    for (i = 0; i < n; i++)
    {
        if (s == PyTuple_GET_ITEM(kwnames, i)) return kwvalues[i];
    }
    for (i = 0; i < n; i++)
    {
        int eq = __Pyx_PyUnicode_Equals(s, PyTuple_GET_ITEM(kwnames, i), Py_EQ);
        if (unlikely(eq != 0)) {
            if (unlikely(eq < 0)) return NULL;
            return kwvalues[i];
        }
    }
    return NULL;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues) {
    Py_ssize_t i, nkwargs = PyTuple_GET_SIZE(kwnames);
    PyObject *dict;
    dict = PyDict_New();
    if (unlikely(!dict))
        return NULL;
    for (i=0; i<nkwargs; i++) {
        PyObject *key = PyTuple_GET_ITEM(kwnames, i);
        if (unlikely(PyDict_SetItem(dict, key, kwvalues[i]) < 0))
            goto bad;
    }
    return dict;
bad:
    Py_DECREF(dict);
    return NULL;
}
#endif
#endif

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* KeywordStringCheck */
static int __Pyx_CheckKeywordStrings(
    PyObject *kw,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kw, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kw))) {
        Py_ssize_t kwsize;
#if CYTHON_ASSUME_SAFE_MACROS
        kwsize = PyTuple_GET_SIZE(kw);
#else
        kwsize = PyTuple_Size(kw);
        if (kwsize < 0) return 0;
#endif
        if (unlikely(kwsize == 0))
            return 1;
        if (!kw_allowed) {
#if CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kw, 0);
#else
            key = PyTuple_GetItem(kw, pos);
            if (!key) return 0;
#endif
            goto invalid_keyword;
        }
#if PY_VERSION_HEX < 0x03090000
        for (pos = 0; pos < kwsize; pos++) {
#if CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kw, pos);
#else
            key = PyTuple_GetItem(kw, pos);
            if (!key) return 0;
#endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
        }
#endif
        return 1;
    }
    while (PyDict_Next(kw, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if (!kw_allowed && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    #if PY_MAJOR_VERSION < 3
    PyErr_Format(PyExc_TypeError,
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    __Pyx_PyThreadState_declare
    CYTHON_UNUSED_VAR(cause);
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
      #if PY_VERSION_HEX >= 0x030C00A6
        PyException_SetTraceback(value, tb);
      #elif CYTHON_FAST_THREAD_STATE
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#else
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject *const *kwvalues,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    int kwds_is_tuple = CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds));
    while (1) {
        Py_XDECREF(key); key = NULL;
        Py_XDECREF(value); value = NULL;
        if (kwds_is_tuple) {
            Py_ssize_t size;
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(kwds);
#else
            size = PyTuple_Size(kwds);
            if (size < 0) goto bad;
#endif
            if (pos >= size) break;
#if CYTHON_AVOID_BORROWED_REFS
            key = __Pyx_PySequence_ITEM(kwds, pos);
            if (!key) goto bad;
#elif CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kwds, pos);
#else
            key = PyTuple_GetItem(kwds, pos);
            if (!key) goto bad;
#endif
            value = kwvalues[pos];
            pos++;
        }
        else
        {
            if (!PyDict_Next(kwds, &pos, &key, &value)) break;
#if CYTHON_AVOID_BORROWED_REFS
            Py_INCREF(key);
#endif
        }
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
            Py_INCREF(value);
            Py_DECREF(key);
#endif
            key = NULL;
            value = NULL;
            continue;
        }
#if !CYTHON_AVOID_BORROWED_REFS
        Py_INCREF(key);
#endif
        Py_INCREF(value);
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
                    value = NULL;
#endif
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key)
                );
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
                    value = NULL;
#endif
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    Py_XDECREF(key);
    Py_XDECREF(value);
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    #if PY_MAJOR_VERSION < 3
    PyErr_Format(PyExc_TypeError,
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    Py_XDECREF(key);
    Py_XDECREF(value);
    return -1;
}

/* GetAttr3 */
#if __PYX_LIMITED_VERSION_HEX < 0x030d00A1
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
#endif
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d00A1
    int res = PyObject_GetOptionalAttr(o, n, &r);
    return (res != 0) ? r : __Pyx_NewRef(d);
#else
  #if CYTHON_USE_TYPE_SLOTS
    if (likely(PyString_Check(n))) {
        r = __Pyx_PyObject_GetAttrStrNoError(o, n);
        if (unlikely(!r) && likely(!PyErr_Occurred())) {
            r = __Pyx_NewRef(d);
        }
        return r;
    }
  #endif
    r = PyObject_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
#endif
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && PY_VERSION_HEX < 0x030d0000
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#elif CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(!__pyx_m)) {
        return NULL;
    }
    result = PyObject_GetAttr(__pyx_m, name);
    if (likely(result)) {
        return result;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* RaiseUnexpectedTypeError */
static int
__Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj)
{
    __Pyx_TypeName obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError, "Expected %s, got " __Pyx_FMT_TYPENAME,
                 expected, obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    __Pyx_TypeName type_name;
    __Pyx_TypeName obj_type_name;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    type_name = __Pyx_PyType_GetName(type);
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected " __Pyx_FMT_TYPENAME
        ", got " __Pyx_FMT_TYPENAME ")", name, type_name, obj_type_name);
    __Pyx_DECREF_TypeName(type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL && !CYTHON_VECTORCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object"))) {
        return NULL;
    }
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object"))) {
        return NULL;
    }
    #endif
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    #endif
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = __Pyx_CyOrPyCFunction_GET_FUNCTION(func);
    self = __Pyx_CyOrPyCFunction_GET_SELF(func);
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    #endif
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectFastCall */
#if PY_VERSION_HEX < 0x03090000 || CYTHON_COMPILING_IN_LIMITED_API
static PyObject* __Pyx_PyObject_FastCall_fallback(PyObject *func, PyObject **args, size_t nargs, PyObject *kwargs) {
    PyObject *argstuple;
    PyObject *result = 0;
    size_t i;
    argstuple = PyTuple_New((Py_ssize_t)nargs);
    if (unlikely(!argstuple)) return NULL;
    for (i = 0; i < nargs; i++) {
        Py_INCREF(args[i]);
        if (__Pyx_PyTuple_SET_ITEM(argstuple, (Py_ssize_t)i, args[i]) < 0) goto bad;
    }
    result = __Pyx_PyObject_Call(func, argstuple, kwargs);
  bad:
    Py_DECREF(argstuple);
    return result;
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject **args, size_t _nargs, PyObject *kwargs) {
    Py_ssize_t nargs = __Pyx_PyVectorcall_NARGS(_nargs);
#if CYTHON_COMPILING_IN_CPYTHON
    if (nargs == 0 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_NOARGS))
            return __Pyx_PyObject_CallMethO(func, NULL);
    }
    else if (nargs == 1 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_O))
            return __Pyx_PyObject_CallMethO(func, args[0]);
    }
#endif
    #if PY_VERSION_HEX < 0x030800B1
    #if CYTHON_FAST_PYCCALL
    if (PyCFunction_Check(func)) {
        if (kwargs) {
            return _PyCFunction_FastCallDict(func, args, nargs, kwargs);
        } else {
            return _PyCFunction_FastCallKeywords(func, args, nargs, NULL);
        }
    }
    #if PY_VERSION_HEX >= 0x030700A1
    if (!kwargs && __Pyx_IS_TYPE(func, &PyMethodDescr_Type)) {
        return _PyMethodDescr_FastCallKeywords(func, args, nargs, NULL);
    }
    #endif
    #endif
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs);
    }
    #endif
    #endif
    if (kwargs == NULL) {
        #if CYTHON_VECTORCALL
        #if PY_VERSION_HEX < 0x03090000
        vectorcallfunc f = _PyVectorcall_Function(func);
        #else
        vectorcallfunc f = PyVectorcall_Function(func);
        #endif
        if (f) {
            return f(func, args, (size_t)nargs, NULL);
        }
        #elif defined(__Pyx_CyFunction_USED) && CYTHON_BACKPORT_VECTORCALL
        if (__Pyx_CyFunction_CheckExact(func)) {
            __pyx_vectorcallfunc f = __Pyx_CyFunction_func_vectorcall(func);
            if (f) return f(func, args, (size_t)nargs, NULL);
        }
        #endif
    }
    if (nargs == 0) {
        return __Pyx_PyObject_Call(func, __pyx_empty_tuple, kwargs);
    }
    #if PY_VERSION_HEX >= 0x03090000 && !CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_VectorcallDict(func, args, (size_t)nargs, kwargs);
    #else
    return __Pyx_PyObject_FastCall_fallback(func, args, (size_t)nargs, kwargs);
    #endif
}

/* PyObjectCallNoArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
    PyObject *arg[2] = {NULL, NULL};
    return __Pyx_PyObject_FastCall(func, arg + 1, 0 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
    else state = (PyGILState_STATE)0;
#endif
    CYTHON_UNUSED_VAR(clineno);
    CYTHON_UNUSED_VAR(lineno);
    CYTHON_UNUSED_VAR(filename);
    CYTHON_MAYBE_UNUSED_VAR(nogil);
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(0);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type = NULL, *local_value, *local_tb = NULL;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if PY_VERSION_HEX >= 0x030C00A6
    local_value = tstate->current_exception;
    tstate->current_exception = 0;
    if (likely(local_value)) {
        local_type = (PyObject*) Py_TYPE(local_value);
        Py_INCREF(local_type);
        local_tb = PyException_GetTraceback(local_value);
    }
  #else
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
  #endif
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE && PY_VERSION_HEX >= 0x030C00A6
    if (unlikely(tstate->current_exception))
#elif CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
      #if PY_VERSION_HEX >= 0x030B00a4
        tmp_value = exc_info->exc_value;
        exc_info->exc_value = local_value;
        tmp_type = NULL;
        tmp_tb = NULL;
        Py_XDECREF(local_type);
        Py_XDECREF(local_tb);
      #else
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
      #endif
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* pep479 */
static void __Pyx_Generator_Replace_StopIteration(int in_async_gen) {
    PyObject *exc, *val, *tb, *cur_exc;
    __Pyx_PyThreadState_declare
    #ifdef __Pyx_StopAsyncIteration_USED
    int is_async_stopiteration = 0;
    #endif
    CYTHON_MAYBE_UNUSED_VAR(in_async_gen);
    cur_exc = PyErr_Occurred();
    if (likely(!__Pyx_PyErr_GivenExceptionMatches(cur_exc, PyExc_StopIteration))) {
        #ifdef __Pyx_StopAsyncIteration_USED
        if (in_async_gen && unlikely(__Pyx_PyErr_GivenExceptionMatches(cur_exc, __Pyx_PyExc_StopAsyncIteration))) {
            is_async_stopiteration = 1;
        } else
        #endif
            return;
    }
    __Pyx_PyThreadState_assign
    __Pyx_GetException(&exc, &val, &tb);
    Py_XDECREF(exc);
    Py_XDECREF(val);
    Py_XDECREF(tb);
    PyErr_SetString(PyExc_RuntimeError,
        #ifdef __Pyx_StopAsyncIteration_USED
        is_async_stopiteration ? "async generator raised StopAsyncIteration" :
        in_async_gen ? "async generator raised StopIteration" :
        #endif
        "generator raised StopIteration");
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    __Pyx_TypeName obj_type_name;
    __Pyx_TypeName type_name;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    type_name = __Pyx_PyType_GetName(type);
    PyErr_Format(PyExc_TypeError,
                 "Cannot convert " __Pyx_FMT_TYPENAME " to " __Pyx_FMT_TYPENAME,
                 obj_type_name, type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    __Pyx_DECREF_TypeName(type_name);
    return 0;
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
    PyObject* exc_type;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    exc_type = __Pyx_PyErr_CurrentExceptionType();
    if (unlikely(exc_type)) {
        if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))
            return -1;
        __Pyx_PyErr_Clear();
        return 0;
    }
    return 0;
}

/* set_iter */
static CYTHON_INLINE PyObject* __Pyx_set_iterator(PyObject* iterable, int is_set,
                                                  Py_ssize_t* p_orig_length, int* p_source_is_set) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000
    is_set = is_set || likely(PySet_CheckExact(iterable) || PyFrozenSet_CheckExact(iterable));
    *p_source_is_set = is_set;
    if (likely(is_set)) {
        *p_orig_length = PySet_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
    }
#else
    CYTHON_UNUSED_VAR(is_set);
    *p_source_is_set = 0;
#endif
    *p_orig_length = 0;
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_set_iter_next(
        PyObject* iter_obj, Py_ssize_t orig_length,
        Py_ssize_t* ppos, PyObject **value,
        int source_is_set) {
    if (!CYTHON_COMPILING_IN_CPYTHON || PY_VERSION_HEX >= 0x030d0000 || unlikely(!source_is_set)) {
        *value = PyIter_Next(iter_obj);
        if (unlikely(!*value)) {
            return __Pyx_IterFinish();
        }
        CYTHON_UNUSED_VAR(orig_length);
        CYTHON_UNUSED_VAR(ppos);
        return 1;
    }
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000
    if (unlikely(PySet_GET_SIZE(iter_obj) != orig_length)) {
        PyErr_SetString(
            PyExc_RuntimeError,
            "set changed size during iteration");
        return -1;
    }
    {
        Py_hash_t hash;
        int ret = _PySet_NextEntry(iter_obj, ppos, value, &hash);
        assert (ret != -1);
        if (likely(ret)) {
            Py_INCREF(*value);
            return 1;
        }
    }
#endif
    return 0;
}

/* FixUpExtensionType */
#if CYTHON_USE_TYPE_SPECS
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type) {
#if PY_VERSION_HEX > 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    CYTHON_UNUSED_VAR(spec);
    CYTHON_UNUSED_VAR(type);
#else
    const PyType_Slot *slot = spec->slots;
    while (slot && slot->slot && slot->slot != Py_tp_members)
        slot++;
    if (slot && slot->slot == Py_tp_members) {
        int changed = 0;
#if !(PY_VERSION_HEX <= 0x030900b1 && CYTHON_COMPILING_IN_CPYTHON)
        const
#endif
            PyMemberDef *memb = (PyMemberDef*) slot->pfunc;
        while (memb && memb->name) {
            if (memb->name[0] == '_' && memb->name[1] == '_') {
#if PY_VERSION_HEX < 0x030900b1
                if (strcmp(memb->name, "__weaklistoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_weaklistoffset = memb->offset;
                    changed = 1;
                }
                else if (strcmp(memb->name, "__dictoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_dictoffset = memb->offset;
                    changed = 1;
                }
#if CYTHON_METH_FASTCALL
                else if (strcmp(memb->name, "__vectorcalloffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
#if PY_VERSION_HEX >= 0x030800b4
                    type->tp_vectorcall_offset = memb->offset;
#else
                    type->tp_print = (printfunc) memb->offset;
#endif
                    changed = 1;
                }
#endif
#else
                if ((0));
#endif
#if PY_VERSION_HEX <= 0x030900b1 && CYTHON_COMPILING_IN_CPYTHON
                else if (strcmp(memb->name, "__module__") == 0) {
                    PyObject *descr;
                    assert(memb->type == T_OBJECT);
                    assert(memb->flags == 0 || memb->flags == READONLY);
                    descr = PyDescr_NewMember(type, memb);
                    if (unlikely(!descr))
                        return -1;
                    if (unlikely(PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr) < 0)) {
                        Py_DECREF(descr);
                        return -1;
                    }
                    Py_DECREF(descr);
                    changed = 1;
                }
#endif
            }
            memb++;
        }
        if (changed)
            PyType_Modified(type);
    }
#endif
    return 0;
}
#endif

/* FetchSharedCythonModule */
static PyObject *__Pyx_FetchSharedCythonABIModule(void) {
    return __Pyx_PyImport_AddModuleRef((char*) __PYX_ABI_MODULE_NAME);
}

/* FetchCommonType */
static int __Pyx_VerifyCachedType(PyObject *cached_type,
                               const char *name,
                               Py_ssize_t basicsize,
                               Py_ssize_t expected_basicsize) {
    if (!PyType_Check(cached_type)) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s is not a type object", name);
        return -1;
    }
    if (basicsize != expected_basicsize) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s has the wrong size, try recompiling",
            name);
        return -1;
    }
    return 0;
}
#if !CYTHON_USE_TYPE_SPECS
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* abi_module;
    const char* object_name;
    PyTypeObject *cached_type = NULL;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) return NULL;
    object_name = strrchr(type->tp_name, '.');
    object_name = object_name ? object_name+1 : type->tp_name;
    cached_type = (PyTypeObject*) PyObject_GetAttrString(abi_module, object_name);
    if (cached_type) {
        if (__Pyx_VerifyCachedType(
              (PyObject *)cached_type,
              object_name,
              cached_type->tp_basicsize,
              type->tp_basicsize) < 0) {
            goto bad;
        }
        goto done;
    }
    if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
    PyErr_Clear();
    if (PyType_Ready(type) < 0) goto bad;
    if (PyObject_SetAttrString(abi_module, object_name, (PyObject *)type) < 0)
        goto bad;
    Py_INCREF(type);
    cached_type = type;
done:
    Py_DECREF(abi_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}
#else
static PyTypeObject *__Pyx_FetchCommonTypeFromSpec(PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *abi_module, *cached_type = NULL;
    const char* object_name = strrchr(spec->name, '.');
    object_name = object_name ? object_name+1 : spec->name;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) return NULL;
    cached_type = PyObject_GetAttrString(abi_module, object_name);
    if (cached_type) {
        Py_ssize_t basicsize;
#if CYTHON_COMPILING_IN_LIMITED_API
        PyObject *py_basicsize;
        py_basicsize = PyObject_GetAttrString(cached_type, "__basicsize__");
        if (unlikely(!py_basicsize)) goto bad;
        basicsize = PyLong_AsSsize_t(py_basicsize);
        Py_DECREF(py_basicsize);
        py_basicsize = 0;
        if (unlikely(basicsize == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
#else
        basicsize = likely(PyType_Check(cached_type)) ? ((PyTypeObject*) cached_type)->tp_basicsize : -1;
#endif
        if (__Pyx_VerifyCachedType(
              cached_type,
              object_name,
              basicsize,
              spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    }
    if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
    PyErr_Clear();
    CYTHON_UNUSED_VAR(module);
    cached_type = __Pyx_PyType_FromModuleAndSpec(abi_module, spec, bases);
    if (unlikely(!cached_type)) goto bad;
    if (unlikely(__Pyx_fix_up_extension_type_from_spec(spec, (PyTypeObject *) cached_type) < 0)) goto bad;
    if (PyObject_SetAttrString(abi_module, object_name, cached_type) < 0) goto bad;
done:
    Py_DECREF(abi_module);
    assert(cached_type == NULL || PyType_Check(cached_type));
    return (PyTypeObject *) cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}
#endif

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_value == NULL || exc_info->exc_value == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    PyObject *exc_value = exc_info->exc_value;
    if (exc_value == NULL || exc_value == Py_None) {
        *value = NULL;
        *type = NULL;
        *tb = NULL;
    } else {
        *value = exc_value;
        Py_INCREF(*value);
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        *tb = PyException_GetTraceback(exc_value);
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #endif
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    PyObject *tmp_value = exc_info->exc_value;
    exc_info->exc_value = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
  #else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
  #endif
}
#endif

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_value = exc_info->exc_value;
    exc_info->exc_value = *value;
    if (tmp_value == NULL || tmp_value == Py_None) {
        Py_XDECREF(tmp_value);
        tmp_value = NULL;
        tmp_type = NULL;
        tmp_tb = NULL;
    } else {
        tmp_type = (PyObject*) Py_TYPE(tmp_value);
        Py_INCREF(tmp_type);
        #if CYTHON_COMPILING_IN_CPYTHON
        tmp_tb = ((PyBaseExceptionObject*) tmp_value)->traceback;
        Py_XINCREF(tmp_tb);
        #else
        tmp_tb = PyException_GetTraceback(tmp_value);
        #endif
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
  #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
  #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* PyObjectCall2Args */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args[3] = {NULL, arg1, arg2};
    return __Pyx_PyObject_FastCall(function, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectCallOneArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *args[2] = {NULL, arg};
    return __Pyx_PyObject_FastCall(func, args+1, 1 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    __Pyx_TypeName type_name;
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if defined(Py_TPFLAGS_METHOD_DESCRIPTOR) && Py_TPFLAGS_METHOD_DESCRIPTOR
        if (__Pyx_PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_METHOD_DESCRIPTOR))
#elif PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (likely(descr != NULL)) {
        *method = descr;
        return 0;
    }
    type_name = __Pyx_PyType_GetName(tp);
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, name);
#else
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%.400s'",
                 type_name, PyString_AS_STRING(name));
#endif
    __Pyx_DECREF_TypeName(type_name);
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod1 */
#if !(CYTHON_VECTORCALL && __PYX_LIMITED_VERSION_HEX >= 0x030C00A2)
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
#endif
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
#if CYTHON_VECTORCALL && __PYX_LIMITED_VERSION_HEX >= 0x030C00A2
    PyObject *args[2] = {obj, arg};
    (void) __Pyx_PyObject_GetMethod;
    (void) __Pyx_PyObject_CallOneArg;
    (void) __Pyx_PyObject_Call2Args;
    return PyObject_VectorcallMethod(method_name, args, 2 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#else
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
#endif
}

/* CoroutineBase */
#include <frameobject.h>
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    CYTHON_UNUSED_VAR(__pyx_tstate);
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (likely(__Pyx_IS_TYPE(ev, (PyTypeObject*)PyExc_StopIteration))) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
#if PY_VERSION_HEX >= 0x030B00a4
    Py_CLEAR(exc_state->exc_value);
#else
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
#endif
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(__pyx_CoroutineObject *gen) {
    const char *msg;
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(PyObject *gen) {
    const char *msg;
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(PyObject *gen, PyObject *value, int closing) {
    CYTHON_MAYBE_UNUSED_VAR(gen);
    CYTHON_MAYBE_UNUSED_VAR(closing);
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_value) {
        #if CYTHON_COMPILING_IN_PYPY
        #else
        PyObject *exc_tb;
        #if PY_VERSION_HEX >= 0x030B00a4 && !CYTHON_COMPILING_IN_CPYTHON
        exc_tb = PyException_GetTraceback(exc_state->exc_value);
        #elif PY_VERSION_HEX >= 0x030B00a4
        exc_tb = ((PyBaseExceptionObject*) exc_state->exc_value)->traceback;
        #else
        exc_tb = exc_state->exc_traceback;
        #endif
        if (exc_tb) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
            PyFrameObject *f = tb->tb_frame;
            assert(f->f_back == NULL);
            #if PY_VERSION_HEX >= 0x030B00A1
            f->f_back = PyThreadState_GetFrame(tstate);
            #else
            Py_XINCREF(tstate->frame);
            f->f_back = tstate->frame;
            #endif
            #if PY_VERSION_HEX >= 0x030B00a4 && !CYTHON_COMPILING_IN_CPYTHON
            Py_DECREF(exc_tb);
            #endif
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    self->is_running = 1;
    retval = self->body(self, tstate, value);
    self->is_running = 0;
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
#if CYTHON_COMPILING_IN_PYPY
    CYTHON_UNUSED_VAR(exc_state);
#else
    PyObject *exc_tb;
    #if PY_VERSION_HEX >= 0x030B00a4
    if (!exc_state->exc_value) return;
    exc_tb = PyException_GetTraceback(exc_state->exc_value);
    #else
    exc_tb = exc_state->exc_traceback;
    #endif
    if (likely(exc_tb)) {
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
        #if PY_VERSION_HEX >= 0x030B00a4
        Py_DECREF(exc_tb);
        #endif
    }
#endif
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(PyObject* gen, PyObject *retval) {
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
static CYTHON_INLINE
PyObject *__Pyx_PyGen_Send(PyGenObject *gen, PyObject *arg) {
#if PY_VERSION_HEX <= 0x030A00A1
    return _PyGen_Send(gen, arg);
#else
    PyObject *result;
    if (PyIter_Send((PyObject*)gen, arg ? arg : Py_None, &result) == PYGEN_RETURN) {
        if (PyAsyncGen_CheckExact(gen)) {
            assert(result == Py_None);
            PyErr_SetNone(PyExc_StopAsyncIteration);
        }
        else if (result == Py_None) {
            PyErr_SetNone(PyExc_StopIteration);
        }
        else {
#if PY_VERSION_HEX < 0x030d00A1
            _PyGen_SetStopIterationValue(result);
#else
            if (!PyTuple_Check(result) && !PyExceptionInstance_Check(result)) {
                PyErr_SetObject(PyExc_StopIteration, result);
            } else {
                PyObject *exc = __Pyx_PyObject_CallOneArg(PyExc_StopIteration, result);
                if (likely(exc != NULL)) {
                    PyErr_SetObject(PyExc_StopIteration, exc);
                    Py_DECREF(exc);
                }
            }
#endif
        }
        Py_DECREF(result);
        result = NULL;
    }
    return result;
#endif
}
#endif
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = __Pyx_PyObject_GetIterNextFunc(yf)(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf, NULL);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStrNoError(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (unlikely(PyErr_Occurred())) {
                PyErr_WriteUnraisable(yf);
            }
        } else {
            retval = __Pyx_PyObject_CallNoArg(meth);
            Py_DECREF(meth);
            if (unlikely(!retval))
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = __Pyx_PyObject_GetIterNextFunc(yf)(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, PyObject *arg) {
    CYTHON_UNUSED_VAR(arg);
    return __Pyx_Coroutine_Close(self);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStrNoError(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (unlikely(PyErr_Occurred())) {
                    gen->is_running = 0;
                    return NULL;
                }
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = __Pyx_PyObject_Call(meth, args, NULL);
            } else {
                PyObject *cargs[4] = {NULL, typ, val, tb};
                ret = __Pyx_PyObject_FastCall(meth, cargs+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (unlikely(!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb)))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
#if PY_VERSION_HEX >= 0x030B00a4
    Py_VISIT(exc_state->exc_value);
#else
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
#endif
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_frame);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (unlikely(PyObject_CallFinalizerFromDealloc(self)))
#else
        Py_TYPE(gen)->tp_del(self);
        if (unlikely(Py_REFCNT(self) > 0))
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    __Pyx_PyHeapTypeObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    __Pyx_SET_REFCNT(self, 1);
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(Py_REFCNT(self) > 0);
    if (likely(--self->ob_refcnt == 0)) {
        return;
    }
    {
        Py_ssize_t refcnt = Py_REFCNT(self);
        _Py_NewReference(self);
        __Pyx_SET_REFCNT(self, refcnt);
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(Py_TYPE(self)) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, void *context)
{
    PyObject *name = self->gi_name;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(self->gi_name, value);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, void *context)
{
    PyObject *name = self->gi_qualname;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(self->gi_qualname, value);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_frame(__pyx_CoroutineObject *self, void *context)
{
    PyObject *frame = self->gi_frame;
    CYTHON_UNUSED_VAR(context);
    if (!frame) {
        if (unlikely(!self->gi_code)) {
            Py_RETURN_NONE;
        }
        frame = (PyObject *) PyFrame_New(
            PyThreadState_Get(),            /*PyThreadState *tstate,*/
            (PyCodeObject*) self->gi_code,  /*PyCodeObject *code,*/
            __pyx_d,                 /*PyObject *globals,*/
            0                               /*PyObject *locals*/
        );
        if (unlikely(!frame))
            return NULL;
        self->gi_frame = frame;
    }
    Py_INCREF(frame);
    return frame;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    #if PY_VERSION_HEX >= 0x030B00a4
    gen->gi_exc_state.exc_value = NULL;
    #else
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
    #endif
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    gen->gi_frame = NULL;
    PyObject_GC_Track(gen);
    return gen;
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    __Pyx_TypeName type_name = __Pyx_PyType_GetName(tp);
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, attr_name);
#else
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%.400s'",
                 type_name, PyString_AS_STRING(attr_name));
#endif
    __Pyx_DECREF_TypeName(type_name);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PatchModuleWithCoroutine */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
#ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (unlikely(!module)) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Coroutine */
static void __Pyx_CoroutineAwait_dealloc(PyObject *self) {
    PyObject_GC_UnTrack(self);
    Py_CLEAR(((__pyx_CoroutineAwaitObject*)self)->coroutine);
    __Pyx_PyHeapTypeObject_GC_Del(self);
}
static int __Pyx_CoroutineAwait_traverse(__pyx_CoroutineAwaitObject *self, visitproc visit, void *arg) {
    Py_VISIT(self->coroutine);
    return 0;
}
static int __Pyx_CoroutineAwait_clear(__pyx_CoroutineAwaitObject *self) {
    Py_CLEAR(self->coroutine);
    return 0;
}
static PyObject *__Pyx_CoroutineAwait_Next(__pyx_CoroutineAwaitObject *self) {
    return __Pyx_Generator_Next(self->coroutine);
}
static PyObject *__Pyx_CoroutineAwait_Send(__pyx_CoroutineAwaitObject *self, PyObject *value) {
    return __Pyx_Coroutine_Send(self->coroutine, value);
}
static PyObject *__Pyx_CoroutineAwait_Throw(__pyx_CoroutineAwaitObject *self, PyObject *args) {
    return __Pyx_Coroutine_Throw(self->coroutine, args);
}
static PyObject *__Pyx_CoroutineAwait_Close(__pyx_CoroutineAwaitObject *self, PyObject *arg) {
    CYTHON_UNUSED_VAR(arg);
    return __Pyx_Coroutine_Close(self->coroutine);
}
static PyObject *__Pyx_CoroutineAwait_self(PyObject *self) {
    Py_INCREF(self);
    return self;
}
#if !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_CoroutineAwait_no_new(PyTypeObject *type, PyObject *args, PyObject *kwargs) {
    CYTHON_UNUSED_VAR(type);
    CYTHON_UNUSED_VAR(args);
    CYTHON_UNUSED_VAR(kwargs);
    PyErr_SetString(PyExc_TypeError, "cannot instantiate type, use 'await coroutine' instead");
    return NULL;
}
#endif
static PyObject *__Pyx_CoroutineAwait_reduce_ex(__pyx_CoroutineAwaitObject *self, PyObject *arg) {
    CYTHON_UNUSED_VAR(arg);
    PyErr_Format(PyExc_TypeError, "cannot pickle '%.200s' object",
                         Py_TYPE(self)->tp_name);
    return NULL;
}
static PyMethodDef __pyx_CoroutineAwait_methods[] = {
    {"send", (PyCFunction) __Pyx_CoroutineAwait_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into coroutine,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_CoroutineAwait_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in coroutine,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_CoroutineAwait_Close, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside coroutine.")},
    {"__reduce_ex__", (PyCFunction) __Pyx_CoroutineAwait_reduce_ex, METH_O, 0},
    {"__reduce__", (PyCFunction) __Pyx_CoroutineAwait_reduce_ex, METH_NOARGS, 0},
    {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_CoroutineAwaitType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_CoroutineAwait_dealloc},
    {Py_tp_traverse, (void *)__Pyx_CoroutineAwait_traverse},
    {Py_tp_clear, (void *)__Pyx_CoroutineAwait_clear},
#if !CYTHON_COMPILING_IN_PYPY
    {Py_tp_new, (void *)__Pyx_CoroutineAwait_no_new},
#endif
    {Py_tp_methods, (void *)__pyx_CoroutineAwait_methods},
    {Py_tp_iter, (void *)__Pyx_CoroutineAwait_self},
    {Py_tp_iternext, (void *)__Pyx_CoroutineAwait_Next},
    {0, 0},
};
static PyType_Spec __pyx_CoroutineAwaitType_spec = {
    __PYX_TYPE_MODULE_PREFIX "coroutine_wrapper",
    sizeof(__pyx_CoroutineAwaitObject),
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    __pyx_CoroutineAwaitType_slots
};
#else
static PyTypeObject __pyx_CoroutineAwaitType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    __PYX_TYPE_MODULE_PREFIX "coroutine_wrapper",
    sizeof(__pyx_CoroutineAwaitObject),
    0,
    (destructor) __Pyx_CoroutineAwait_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    PyDoc_STR("A wrapper object implementing __await__ for coroutines."),
    (traverseproc) __Pyx_CoroutineAwait_traverse,
    (inquiry) __Pyx_CoroutineAwait_clear,
    0,
    0,
    __Pyx_CoroutineAwait_self,
    (iternextfunc) __Pyx_CoroutineAwait_Next,
    __pyx_CoroutineAwait_methods,
    0                         ,
    0                      ,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if !CYTHON_COMPILING_IN_PYPY
    __Pyx_CoroutineAwait_no_new,
#else
    0,
#endif
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if __PYX_NEED_TP_PRINT_SLOT
    0,
#endif
#if PY_VERSION_HEX >= 0x030C0000
    0,
#endif
#if PY_VERSION_HEX >= 0x030d00A4
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
    0,
#endif
};
#endif
#if PY_VERSION_HEX < 0x030500B1 || defined(__Pyx_IterableCoroutine_USED) || CYTHON_USE_ASYNC_SLOTS
static CYTHON_INLINE PyObject *__Pyx__Coroutine_await(PyObject *coroutine) {
    __pyx_CoroutineAwaitObject *await = PyObject_GC_New(__pyx_CoroutineAwaitObject, __pyx_CoroutineAwaitType);
    if (unlikely(!await)) return NULL;
    Py_INCREF(coroutine);
    await->coroutine = coroutine;
    PyObject_GC_Track(await);
    return (PyObject*)await;
}
#endif
#if PY_VERSION_HEX < 0x030500B1
static PyObject *__Pyx_Coroutine_await_method(PyObject *coroutine, PyObject *arg) {
    CYTHON_UNUSED_VAR(arg);
    return __Pyx__Coroutine_await(coroutine);
}
#endif
#if defined(__Pyx_IterableCoroutine_USED) || CYTHON_USE_ASYNC_SLOTS
static PyObject *__Pyx_Coroutine_await(PyObject *coroutine) {
    if (unlikely(!coroutine || !__Pyx_Coroutine_Check(coroutine))) {
        PyErr_SetString(PyExc_TypeError, "invalid input, expected coroutine");
        return NULL;
    }
    return __Pyx__Coroutine_await(coroutine);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3 && PY_VERSION_HEX < 0x030500B1
static PyObject *__Pyx_Coroutine_compare(PyObject *obj, PyObject *other, int op) {
    PyObject* result;
    switch (op) {
        case Py_EQ: result = (other == obj) ? Py_True : Py_False; break;
        case Py_NE: result = (other != obj) ? Py_True : Py_False; break;
        default:
            result = Py_NotImplemented;
    }
    Py_INCREF(result);
    return result;
}
#endif
static PyMethodDef __pyx_Coroutine_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into coroutine,\nreturn next iterated value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in coroutine,\nreturn next iterated value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside coroutine.")},
#if PY_VERSION_HEX < 0x030500B1
    {"__await__", (PyCFunction) __Pyx_Coroutine_await_method, METH_NOARGS,
     (char*) PyDoc_STR("__await__() -> return an iterator to be used in await expression.")},
#endif
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Coroutine_memberlist[] = {
    {(char *) "cr_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "cr_await", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being awaited, or None")},
    {(char*) "cr_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {(char *) "__module__", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_modulename), 0, 0},
#if CYTHON_USE_TYPE_SPECS
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CoroutineObject, gi_weakreflist), READONLY, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Coroutine_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the coroutine"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the coroutine"), 0},
    {(char *) "cr_frame", (getter)__Pyx_Coroutine_get_frame, NULL,
     (char*) PyDoc_STR("Frame of the coroutine"), 0},
    {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_CoroutineType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_Coroutine_dealloc},
    {Py_am_await, (void *)&__Pyx_Coroutine_await},
    {Py_tp_traverse, (void *)__Pyx_Coroutine_traverse},
    {Py_tp_methods, (void *)__pyx_Coroutine_methods},
    {Py_tp_members, (void *)__pyx_Coroutine_memberlist},
    {Py_tp_getset, (void *)__pyx_Coroutine_getsets},
    {Py_tp_getattro, (void *) __Pyx_PyObject_GenericGetAttrNoDict},
#if CYTHON_USE_TP_FINALIZE
    {Py_tp_finalize, (void *)__Pyx_Coroutine_del},
#endif
    {0, 0},
};
static PyType_Spec __pyx_CoroutineType_spec = {
    __PYX_TYPE_MODULE_PREFIX "coroutine",
    sizeof(__pyx_CoroutineObject),
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    __pyx_CoroutineType_slots
};
#else
#if CYTHON_USE_ASYNC_SLOTS
static __Pyx_PyAsyncMethodsStruct __pyx_Coroutine_as_async = {
    __Pyx_Coroutine_await,
    0,
    0,
#if PY_VERSION_HEX >= 0x030A00A3
    0,
#endif
};
#endif
static PyTypeObject __pyx_CoroutineType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    __PYX_TYPE_MODULE_PREFIX "coroutine",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
#if CYTHON_USE_ASYNC_SLOTS
    &__pyx_Coroutine_as_async,
#else
    0,
#endif
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
#if CYTHON_USE_ASYNC_SLOTS && CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3 && PY_VERSION_HEX < 0x030500B1
    __Pyx_Coroutine_compare,
#else
    0,
#endif
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    0,
    __pyx_Coroutine_methods,
    __pyx_Coroutine_memberlist,
    __pyx_Coroutine_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if __PYX_NEED_TP_PRINT_SLOT
    0,
#endif
#if PY_VERSION_HEX >= 0x030C0000
    0,
#endif
#if PY_VERSION_HEX >= 0x030d00A4
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
    0,
#endif
};
#endif
static int __pyx_Coroutine_init(PyObject *module) {
    CYTHON_MAYBE_UNUSED_VAR(module);
#if CYTHON_USE_TYPE_SPECS
    __pyx_CoroutineType = __Pyx_FetchCommonTypeFromSpec(module, &__pyx_CoroutineType_spec, NULL);
#else
    __pyx_CoroutineType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_CoroutineType = __Pyx_FetchCommonType(&__pyx_CoroutineType_type);
#endif
    if (unlikely(!__pyx_CoroutineType))
        return -1;
#ifdef __Pyx_IterableCoroutine_USED
    if (unlikely(__pyx_IterableCoroutine_init(module) == -1))
        return -1;
#endif
#if CYTHON_USE_TYPE_SPECS
    __pyx_CoroutineAwaitType = __Pyx_FetchCommonTypeFromSpec(module, &__pyx_CoroutineAwaitType_spec, NULL);
#else
    __pyx_CoroutineAwaitType = __Pyx_FetchCommonType(&__pyx_CoroutineAwaitType_type);
#endif
    if (unlikely(!__pyx_CoroutineAwaitType))
        return -1;
    return 0;
}

/* GetAwaitIter */
static CYTHON_INLINE PyObject *__Pyx_Coroutine_GetAwaitableIter(PyObject *o) {
#ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(o)) {
        return __Pyx_NewRef(o);
    }
#endif
    return __Pyx__Coroutine_GetAwaitableIter(o);
}
static void __Pyx_Coroutine_AwaitableIterError(PyObject *source) {
#if PY_VERSION_HEX >= 0x030600B3 && PY_VERSION_HEX < 0x030d0000 || defined(_PyErr_FormatFromCause)
    __Pyx_TypeName source_type_name = __Pyx_PyType_GetName(Py_TYPE(source));
    _PyErr_FormatFromCause(PyExc_TypeError,
        "'async for' received an invalid object from __anext__: " __Pyx_FMT_TYPENAME, source_type_name);
    __Pyx_DECREF_TypeName(source_type_name);
#elif PY_MAJOR_VERSION >= 3
    PyObject *exc, *val, *val2, *tb;
    __Pyx_TypeName source_type_name = __Pyx_PyType_GetName(Py_TYPE(source));
    assert(PyErr_Occurred());
    PyErr_Fetch(&exc, &val, &tb);
    PyErr_NormalizeException(&exc, &val, &tb);
    if (tb != NULL) {
        PyException_SetTraceback(val, tb);
        Py_DECREF(tb);
    }
    Py_DECREF(exc);
    assert(!PyErr_Occurred());
    PyErr_Format(PyExc_TypeError,
        "'async for' received an invalid object from __anext__: " __Pyx_FMT_TYPENAME, source_type_name);
    __Pyx_DECREF_TypeName(source_type_name);
    PyErr_Fetch(&exc, &val2, &tb);
    PyErr_NormalizeException(&exc, &val2, &tb);
    Py_INCREF(val);
    PyException_SetCause(val2, val);
    PyException_SetContext(val2, val);
    PyErr_Restore(exc, val2, tb);
#else
    CYTHON_UNUSED_VAR(source);
#endif
}
static PyObject *__Pyx__Coroutine_GetAwaitableIter(PyObject *obj) {
    PyObject *res;
#if CYTHON_USE_ASYNC_SLOTS
    __Pyx_PyAsyncMethodsStruct* am = __Pyx_PyType_AsAsync(obj);
    if (likely(am && am->am_await)) {
        res = (*am->am_await)(obj);
    } else
#endif
#if PY_VERSION_HEX >= 0x030500B2 || defined(PyCoro_CheckExact)
    if (PyCoro_CheckExact(obj)) {
        return __Pyx_NewRef(obj);
    } else
#endif
#if CYTHON_COMPILING_IN_CPYTHON && defined(CO_ITERABLE_COROUTINE)
#if PY_VERSION_HEX >= 0x030C00A6
    if (PyGen_CheckExact(obj) && (PyGen_GetCode((PyGenObject*)obj)->co_flags & CO_ITERABLE_COROUTINE)) {
#else
    if (PyGen_CheckExact(obj) && ((PyGenObject*)obj)->gi_code && ((PyCodeObject *)((PyGenObject*)obj)->gi_code)->co_flags & CO_ITERABLE_COROUTINE) {
#endif
        return __Pyx_NewRef(obj);
    } else
#endif
    {
        PyObject *method = NULL;
        int is_method = __Pyx_PyObject_GetMethod(obj, __pyx_n_s_await, &method);
        if (likely(is_method)) {
            res = __Pyx_PyObject_CallOneArg(method, obj);
        } else if (likely(method)) {
            res = __Pyx_PyObject_CallNoArg(method);
        } else
            goto slot_error;
        Py_DECREF(method);
    }
    if (unlikely(!res)) {
        __Pyx_Coroutine_AwaitableIterError(obj);
        goto bad;
    }
    if (unlikely(!PyIter_Check(res))) {
        __Pyx_TypeName res_type_name = __Pyx_PyType_GetName(Py_TYPE(res));
        PyErr_Format(PyExc_TypeError,
            "__await__() returned non-iterator of type '" __Pyx_FMT_TYPENAME "'", res_type_name);
        __Pyx_DECREF_TypeName(res_type_name);
        Py_CLEAR(res);
    } else {
        int is_coroutine = 0;
        #ifdef __Pyx_Coroutine_USED
        is_coroutine |= __Pyx_Coroutine_Check(res);
        #endif
        #if PY_VERSION_HEX >= 0x030500B2 || defined(PyCoro_CheckExact)
        is_coroutine |= PyCoro_CheckExact(res);
        #endif
        if (unlikely(is_coroutine)) {
            /* __await__ must return an *iterator*, not
               a coroutine or another awaitable (see PEP 492) */
            PyErr_SetString(PyExc_TypeError,
                            "__await__() returned a coroutine");
            Py_CLEAR(res);
        }
    }
    return res;
slot_error:
    {
        __Pyx_TypeName obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
        PyErr_Format(PyExc_TypeError,
            "object " __Pyx_FMT_TYPENAME " can't be used in 'await' expression", obj_type_name);
        __Pyx_DECREF_TypeName(obj_type_name);
    }
bad:
    return NULL;
}

/* CoroutineYieldFrom */
  static PyObject* __Pyx__Coroutine_Yield_From_Generic(__pyx_CoroutineObject *gen, PyObject *source) {
    PyObject *retval;
    PyObject *source_gen = __Pyx__Coroutine_GetAwaitableIter(source);
    if (unlikely(!source_gen)) {
        return NULL;
    }
    if (__Pyx_Coroutine_Check(source_gen)) {
        retval = __Pyx_Generator_Next(source_gen);
    } else {
        retval = __Pyx_PyObject_GetIterNextFunc(source_gen)(source_gen);
    }
    if (retval) {
        gen->yieldfrom = source_gen;
        return retval;
    }
    Py_DECREF(source_gen);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_Coroutine_Yield_From(__pyx_CoroutineObject *gen, PyObject *source) {
    PyObject *retval;
    if (__Pyx_Coroutine_Check(source)) {
        if (unlikely(((__pyx_CoroutineObject*)source)->yieldfrom)) {
            PyErr_SetString(
                PyExc_RuntimeError,
                "coroutine is being awaited already");
            return NULL;
        }
        retval = __Pyx_Generator_Next(source);
#ifdef __Pyx_AsyncGen_USED
    } else if (__pyx_PyAsyncGenASend_CheckExact(source)) {
        retval = __Pyx_async_gen_asend_iternext(source);
#endif
    } else {
        return __Pyx__Coroutine_Yield_From_Generic(gen, source);
    }
    if (retval) {
        Py_INCREF(source);
        gen->yieldfrom = source;
    }
    return retval;
}

/* pyfrozenset_new */
  static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it) {
    if (it) {
        PyObject* result;
#if CYTHON_COMPILING_IN_PYPY
        PyObject* args;
        args = PyTuple_Pack(1, it);
        if (unlikely(!args))
            return NULL;
        result = PyObject_Call((PyObject*)&PyFrozenSet_Type, args, NULL);
        Py_DECREF(args);
        return result;
#else
        if (PyFrozenSet_CheckExact(it)) {
            Py_INCREF(it);
            return it;
        }
        result = PyFrozenSet_New(it);
        if (unlikely(!result))
            return NULL;
        if ((PY_VERSION_HEX >= 0x031000A1) || likely(PySet_GET_SIZE(result)))
            return result;
        Py_DECREF(result);
#endif
    }
#if CYTHON_USE_TYPE_SLOTS
    return PyFrozenSet_Type.tp_new(&PyFrozenSet_Type, __pyx_empty_tuple, NULL);
#else
    return PyObject_Call((PyObject*)&PyFrozenSet_Type, __pyx_empty_tuple, NULL);
#endif
}

/* py_set_discard_unhashable */
  static int __Pyx_PySet_DiscardUnhashable(PyObject *set, PyObject *key) {
    PyObject *tmpkey;
    int rv;
    if (likely(!PySet_Check(key) || !PyErr_ExceptionMatches(PyExc_TypeError)))
        return -1;
    PyErr_Clear();
    tmpkey = __Pyx_PyFrozenSet_New(key);
    if (tmpkey == NULL)
        return -1;
    rv = PySet_Discard(set, tmpkey);
    Py_DECREF(tmpkey);
    return rv;
}

/* py_set_discard */
  static CYTHON_INLINE int __Pyx_PySet_Discard(PyObject *set, PyObject *key) {
    int found = PySet_Discard(set, key);
    if (unlikely(found < 0)) {
        found = __Pyx_PySet_DiscardUnhashable(set, key);
    }
    return found;
}

/* Import */
  static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *module = 0;
    PyObject *empty_dict = 0;
    PyObject *empty_list = 0;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (unlikely(!py_import))
        goto bad;
    if (!from_list) {
        empty_list = PyList_New(0);
        if (unlikely(!empty_list))
            goto bad;
        from_list = empty_list;
    }
    #endif
    empty_dict = PyDict_New();
    if (unlikely(!empty_dict))
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.') != NULL) {
                module = PyImport_ImportModuleLevelObject(
                    name, __pyx_d, empty_dict, from_list, 1);
                if (unlikely(!module)) {
                    if (unlikely(!PyErr_ExceptionMatches(PyExc_ImportError)))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (unlikely(!py_level))
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, __pyx_d, empty_dict, from_list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, __pyx_d, empty_dict, from_list, level);
            #endif
        }
    }
bad:
    Py_XDECREF(empty_dict);
    Py_XDECREF(empty_list);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    return module;
}

/* ImportFrom */
  static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        const char* module_name_str = 0;
        PyObject* module_name = 0;
        PyObject* module_dot = 0;
        PyObject* full_name = 0;
        PyErr_Clear();
        module_name_str = PyModule_GetName(module);
        if (unlikely(!module_name_str)) { goto modbad; }
        module_name = PyUnicode_FromString(module_name_str);
        if (unlikely(!module_name)) { goto modbad; }
        module_dot = PyUnicode_Concat(module_name, __pyx_kp_u__5);
        if (unlikely(!module_dot)) { goto modbad; }
        full_name = PyUnicode_Concat(module_dot, name);
        if (unlikely(!full_name)) { goto modbad; }
        #if PY_VERSION_HEX < 0x030700A1 || (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400)
        {
            PyObject *modules = PyImport_GetModuleDict();
            if (unlikely(!modules))
                goto modbad;
            value = PyObject_GetItem(modules, full_name);
        }
        #else
        value = PyImport_GetModule(full_name);
        #endif
      modbad:
        Py_XDECREF(full_name);
        Py_XDECREF(module_dot);
        Py_XDECREF(module_name);
    }
    if (unlikely(!value)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* GetAttr */
  static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* HasAttr */
  static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (!r) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* CallNextTpTraverse */
  static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse) {
    PyTypeObject* type = Py_TYPE(obj);
    traverseproc tp_traverse = NULL;
    while (type && __Pyx_PyType_GetSlot(type, tp_traverse, traverseproc) != current_tp_traverse)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    while (type && (tp_traverse = __Pyx_PyType_GetSlot(type, tp_traverse, traverseproc)) == current_tp_traverse)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    if (type && tp_traverse)
        return tp_traverse(obj, v, a);
    return 0;
}

/* CallNextTpClear */
  static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_clear) {
    PyTypeObject* type = Py_TYPE(obj);
    inquiry tp_clear = NULL;
    while (type && __Pyx_PyType_GetSlot(type, tp_clear, inquiry) != current_tp_clear)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    while (type && (tp_clear = __Pyx_PyType_GetSlot(type, tp_clear, inquiry)) == current_tp_clear)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    if (type && tp_clear)
        tp_clear(obj);
}

/* PyObjectCallMethod0 */
  static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* ValidateBasesTuple */
  #if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases) {
    Py_ssize_t i, n;
#if CYTHON_ASSUME_SAFE_MACROS
    n = PyTuple_GET_SIZE(bases);
#else
    n = PyTuple_Size(bases);
    if (n < 0) return -1;
#endif
    for (i = 1; i < n; i++)
    {
#if CYTHON_AVOID_BORROWED_REFS
        PyObject *b0 = PySequence_GetItem(bases, i);
        if (!b0) return -1;
#elif CYTHON_ASSUME_SAFE_MACROS
        PyObject *b0 = PyTuple_GET_ITEM(bases, i);
#else
        PyObject *b0 = PyTuple_GetItem(bases, i);
        if (!b0) return -1;
#endif
        PyTypeObject *b;
#if PY_MAJOR_VERSION < 3
        if (PyClass_Check(b0))
        {
            PyErr_Format(PyExc_TypeError, "base class '%.200s' is an old-style class",
                         PyString_AS_STRING(((PyClassObject*)b0)->cl_name));
#if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(b0);
#endif
            return -1;
        }
#endif
        b = (PyTypeObject*) b0;
        if (!__Pyx_PyType_HasFeature(b, Py_TPFLAGS_HEAPTYPE))
        {
            __Pyx_TypeName b_name = __Pyx_PyType_GetName(b);
            PyErr_Format(PyExc_TypeError,
                "base class '" __Pyx_FMT_TYPENAME "' is not a heap type", b_name);
            __Pyx_DECREF_TypeName(b_name);
#if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(b0);
#endif
            return -1;
        }
        if (dictoffset == 0)
        {
            Py_ssize_t b_dictoffset = 0;
#if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
            b_dictoffset = b->tp_dictoffset;
#else
            PyObject *py_b_dictoffset = PyObject_GetAttrString((PyObject*)b, "__dictoffset__");
            if (!py_b_dictoffset) goto dictoffset_return;
            b_dictoffset = PyLong_AsSsize_t(py_b_dictoffset);
            Py_DECREF(py_b_dictoffset);
            if (b_dictoffset == -1 && PyErr_Occurred()) goto dictoffset_return;
#endif
            if (b_dictoffset) {
                {
                    __Pyx_TypeName b_name = __Pyx_PyType_GetName(b);
                    PyErr_Format(PyExc_TypeError,
                        "extension type '%.200s' has no __dict__ slot, "
                        "but base type '" __Pyx_FMT_TYPENAME "' has: "
                        "either add 'cdef dict __dict__' to the extension type "
                        "or add '__slots__ = [...]' to the base type",
                        type_name, b_name);
                    __Pyx_DECREF_TypeName(b_name);
                }
#if !(CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY)
              dictoffset_return:
#endif
#if CYTHON_AVOID_BORROWED_REFS
                Py_DECREF(b0);
#endif
                return -1;
            }
        }
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(b0);
#endif
    }
    return 0;
}
#endif

/* PyType_Ready */
  static int __Pyx_PyType_Ready(PyTypeObject *t) {
#if CYTHON_USE_TYPE_SPECS || !(CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API) || defined(PYSTON_MAJOR_VERSION)
    (void)__Pyx_PyObject_CallMethod0;
#if CYTHON_USE_TYPE_SPECS
    (void)__Pyx_validate_bases_tuple;
#endif
    return PyType_Ready(t);
#else
    int r;
    PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
    if (bases && unlikely(__Pyx_validate_bases_tuple(t->tp_name, t->tp_dictoffset, bases) == -1))
        return -1;
#if PY_VERSION_HEX >= 0x03050000 && !defined(PYSTON_MAJOR_VERSION)
    {
        int gc_was_enabled;
    #if PY_VERSION_HEX >= 0x030A00b1
        gc_was_enabled = PyGC_Disable();
        (void)__Pyx_PyObject_CallMethod0;
    #else
        PyObject *ret, *py_status;
        PyObject *gc = NULL;
        #if PY_VERSION_HEX >= 0x030700a1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM+0 >= 0x07030400)
        gc = PyImport_GetModule(__pyx_kp_u_gc);
        #endif
        if (unlikely(!gc)) gc = PyImport_Import(__pyx_kp_u_gc);
        if (unlikely(!gc)) return -1;
        py_status = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_isenabled);
        if (unlikely(!py_status)) {
            Py_DECREF(gc);
            return -1;
        }
        gc_was_enabled = __Pyx_PyObject_IsTrue(py_status);
        Py_DECREF(py_status);
        if (gc_was_enabled > 0) {
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_disable);
            if (unlikely(!ret)) {
                Py_DECREF(gc);
                return -1;
            }
            Py_DECREF(ret);
        } else if (unlikely(gc_was_enabled == -1)) {
            Py_DECREF(gc);
            return -1;
        }
    #endif
        t->tp_flags |= Py_TPFLAGS_HEAPTYPE;
#if PY_VERSION_HEX >= 0x030A0000
        t->tp_flags |= Py_TPFLAGS_IMMUTABLETYPE;
#endif
#else
        (void)__Pyx_PyObject_CallMethod0;
#endif
    r = PyType_Ready(t);
#if PY_VERSION_HEX >= 0x03050000 && !defined(PYSTON_MAJOR_VERSION)
        t->tp_flags &= ~Py_TPFLAGS_HEAPTYPE;
    #if PY_VERSION_HEX >= 0x030A00b1
        if (gc_was_enabled)
            PyGC_Enable();
    #else
        if (gc_was_enabled) {
            PyObject *tp, *v, *tb;
            PyErr_Fetch(&tp, &v, &tb);
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_enable);
            if (likely(ret || r == -1)) {
                Py_XDECREF(ret);
                PyErr_Restore(tp, v, tb);
            } else {
                Py_XDECREF(tp);
                Py_XDECREF(v);
                Py_XDECREF(tb);
                r = -1;
            }
        }
        Py_DECREF(gc);
    #endif
    }
#endif
    return r;
#endif
}

/* PyObject_GenericGetAttr */
  #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetupReduce */
  #if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStrNoError(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_n_s_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred()) {
        __Pyx_TypeName type_obj_name =
            __Pyx_PyType_GetName((PyTypeObject*)type_obj);
        PyErr_Format(PyExc_RuntimeError,
            "Unable to initialize pickling for " __Pyx_FMT_TYPENAME, type_obj_name);
        __Pyx_DECREF_TypeName(type_obj_name);
    }
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}
#endif

/* SetVTable */
  static int __Pyx_SetVtable(PyTypeObject *type, void *vtable) {
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
    if (unlikely(!ob))
        goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(PyObject_SetAttr((PyObject *) type, __pyx_n_s_pyx_vtable, ob) < 0))
#else
    if (unlikely(PyDict_SetItem(type->tp_dict, __pyx_n_s_pyx_vtable, ob) < 0))
#endif
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* GetVTable */
  static void* __Pyx_GetVtable(PyTypeObject *type) {
    void* ptr;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *ob = PyObject_GetAttr((PyObject *)type, __pyx_n_s_pyx_vtable);
#else
    PyObject *ob = PyObject_GetItem(type->tp_dict, __pyx_n_s_pyx_vtable);
#endif
    if (!ob)
        goto bad;
    ptr = PyCapsule_GetPointer(ob, 0);
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* MergeVTables */
  #if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_MergeVtables(PyTypeObject *type) {
    int i;
    void** base_vtables;
    __Pyx_TypeName tp_base_name;
    __Pyx_TypeName base_name;
    void* unknown = (void*)-1;
    PyObject* bases = type->tp_bases;
    int base_depth = 0;
    {
        PyTypeObject* base = type->tp_base;
        while (base) {
            base_depth += 1;
            base = base->tp_base;
        }
    }
    base_vtables = (void**) malloc(sizeof(void*) * (size_t)(base_depth + 1));
    base_vtables[0] = unknown;
    for (i = 1; i < PyTuple_GET_SIZE(bases); i++) {
        void* base_vtable = __Pyx_GetVtable(((PyTypeObject*)PyTuple_GET_ITEM(bases, i)));
        if (base_vtable != NULL) {
            int j;
            PyTypeObject* base = type->tp_base;
            for (j = 0; j < base_depth; j++) {
                if (base_vtables[j] == unknown) {
                    base_vtables[j] = __Pyx_GetVtable(base);
                    base_vtables[j + 1] = unknown;
                }
                if (base_vtables[j] == base_vtable) {
                    break;
                } else if (base_vtables[j] == NULL) {
                    goto bad;
                }
                base = base->tp_base;
            }
        }
    }
    PyErr_Clear();
    free(base_vtables);
    return 0;
bad:
    tp_base_name = __Pyx_PyType_GetName(type->tp_base);
    base_name = __Pyx_PyType_GetName((PyTypeObject*)PyTuple_GET_ITEM(bases, i));
    PyErr_Format(PyExc_TypeError,
        "multiple bases have vtable conflict: '" __Pyx_FMT_TYPENAME "' and '" __Pyx_FMT_TYPENAME "'", tp_base_name, base_name);
    __Pyx_DECREF_TypeName(tp_base_name);
    __Pyx_DECREF_TypeName(base_name);
    free(base_vtables);
    return -1;
}
#endif

/* TypeImport */
  #ifndef __PYX_HAVE_RT_ImportType_3_0_11
#define __PYX_HAVE_RT_ImportType_3_0_11
static PyTypeObject *__Pyx_ImportType_3_0_11(PyObject *module, const char *module_name, const char *class_name,
    size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_0_11 check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
    Py_ssize_t itemsize;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_basicsize;
    PyObject *py_itemsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#if !CYTHON_COMPILING_IN_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
    itemsize = ((PyTypeObject *)result)->tp_itemsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
    py_itemsize = PyObject_GetAttrString(result, "__itemsize__");
    if (!py_itemsize)
        goto bad;
    itemsize = PyLong_AsSsize_t(py_itemsize);
    Py_DECREF(py_itemsize);
    py_itemsize = 0;
    if (itemsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (itemsize) {
        if (size % alignment) {
            alignment = size % alignment;
        }
        if (itemsize < (Py_ssize_t)alignment)
            itemsize = (Py_ssize_t)alignment;
    }
    if ((size_t)(basicsize + itemsize) < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize+itemsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error_3_0_11 &&
            ((size_t)basicsize > size || (size_t)(basicsize + itemsize) < size)) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd-%zd from PyObject",
            module_name, class_name, size, basicsize, basicsize+itemsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn_3_0_11 && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* ImportDottedModule */
  #if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx__ImportDottedModule_Error(PyObject *name, PyObject *parts_tuple, Py_ssize_t count) {
    PyObject *partial_name = NULL, *slice = NULL, *sep = NULL;
    if (unlikely(PyErr_Occurred())) {
        PyErr_Clear();
    }
    if (likely(PyTuple_GET_SIZE(parts_tuple) == count)) {
        partial_name = name;
    } else {
        slice = PySequence_GetSlice(parts_tuple, 0, count);
        if (unlikely(!slice))
            goto bad;
        sep = PyUnicode_FromStringAndSize(".", 1);
        if (unlikely(!sep))
            goto bad;
        partial_name = PyUnicode_Join(sep, slice);
    }
    PyErr_Format(
#if PY_MAJOR_VERSION < 3
        PyExc_ImportError,
        "No module named '%s'", PyString_AS_STRING(partial_name));
#else
#if PY_VERSION_HEX >= 0x030600B1
        PyExc_ModuleNotFoundError,
#else
        PyExc_ImportError,
#endif
        "No module named '%U'", partial_name);
#endif
bad:
    Py_XDECREF(sep);
    Py_XDECREF(slice);
    Py_XDECREF(partial_name);
    return NULL;
}
#endif
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx__ImportDottedModule_Lookup(PyObject *name) {
    PyObject *imported_module;
#if PY_VERSION_HEX < 0x030700A1 || (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400)
    PyObject *modules = PyImport_GetModuleDict();
    if (unlikely(!modules))
        return NULL;
    imported_module = __Pyx_PyDict_GetItemStr(modules, name);
    Py_XINCREF(imported_module);
#else
    imported_module = PyImport_GetModule(name);
#endif
    return imported_module;
}
#endif
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_ImportDottedModule_WalkParts(PyObject *module, PyObject *name, PyObject *parts_tuple) {
    Py_ssize_t i, nparts;
    nparts = PyTuple_GET_SIZE(parts_tuple);
    for (i=1; i < nparts && module; i++) {
        PyObject *part, *submodule;
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        part = PyTuple_GET_ITEM(parts_tuple, i);
#else
        part = PySequence_ITEM(parts_tuple, i);
#endif
        submodule = __Pyx_PyObject_GetAttrStrNoError(module, part);
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
        Py_DECREF(part);
#endif
        Py_DECREF(module);
        module = submodule;
    }
    if (unlikely(!module)) {
        return __Pyx__ImportDottedModule_Error(name, parts_tuple, i);
    }
    return module;
}
#endif
static PyObject *__Pyx__ImportDottedModule(PyObject *name, PyObject *parts_tuple) {
#if PY_MAJOR_VERSION < 3
    PyObject *module, *from_list, *star = __pyx_n_s__8;
    CYTHON_UNUSED_VAR(parts_tuple);
    from_list = PyList_New(1);
    if (unlikely(!from_list))
        return NULL;
    Py_INCREF(star);
    PyList_SET_ITEM(from_list, 0, star);
    module = __Pyx_Import(name, from_list, 0);
    Py_DECREF(from_list);
    return module;
#else
    PyObject *imported_module;
    PyObject *module = __Pyx_Import(name, NULL, 0);
    if (!parts_tuple || unlikely(!module))
        return module;
    imported_module = __Pyx__ImportDottedModule_Lookup(name);
    if (likely(imported_module)) {
        Py_DECREF(module);
        return imported_module;
    }
    PyErr_Clear();
    return __Pyx_ImportDottedModule_WalkParts(module, name, parts_tuple);
#endif
}
static PyObject *__Pyx_ImportDottedModule(PyObject *name, PyObject *parts_tuple) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030400B1
    PyObject *module = __Pyx__ImportDottedModule_Lookup(name);
    if (likely(module)) {
        PyObject *spec = __Pyx_PyObject_GetAttrStrNoError(module, __pyx_n_s_spec);
        if (likely(spec)) {
            PyObject *unsafe = __Pyx_PyObject_GetAttrStrNoError(spec, __pyx_n_s_initializing);
            if (likely(!unsafe || !__Pyx_PyObject_IsTrue(unsafe))) {
                Py_DECREF(spec);
                spec = NULL;
            }
            Py_XDECREF(unsafe);
        }
        if (likely(!spec)) {
            PyErr_Clear();
            return module;
        }
        Py_DECREF(spec);
        Py_DECREF(module);
    } else if (PyErr_Occurred()) {
        PyErr_Clear();
    }
#endif
    return __Pyx__ImportDottedModule(name, parts_tuple);
}

/* PatchInspect */
  static PyObject* __Pyx_patch_inspect(PyObject* module) {
#if defined(__Pyx_Generator_USED) && (!defined(CYTHON_PATCH_INSPECT) || CYTHON_PATCH_INSPECT)
    static int inspect_patched = 0;
    if (unlikely((!inspect_patched) && module)) {
        module = __Pyx_Coroutine_patch_module(
            module, ""
"old_types = getattr(_module.isgenerator, '_cython_generator_types', None)\n"
"if old_types is None or not isinstance(old_types, set):\n"
"    old_types = set()\n"
"    def cy_wrap(orig_func, type=type, cython_generator_types=old_types):\n"
"        def cy_isgenerator(obj): return type(obj) in cython_generator_types or orig_func(obj)\n"
"        cy_isgenerator._cython_generator_types = cython_generator_types\n"
"        return cy_isgenerator\n"
"    _module.isgenerator = cy_wrap(_module.isgenerator)\n"
"old_types.add(_cython_generator_type)\n"
        );
        inspect_patched = 1;
    }
#else
    if ((0)) return __Pyx_Coroutine_patch_module(module, NULL);
#endif
    return module;
}

/* PatchAsyncIO */
  static PyObject* __Pyx_patch_asyncio(PyObject* module) {
#if PY_VERSION_HEX < 0x030500B2 &&\
        (defined(__Pyx_Coroutine_USED) || defined(__Pyx_Generator_USED)) &&\
        (!defined(CYTHON_PATCH_ASYNCIO) || CYTHON_PATCH_ASYNCIO)
    PyObject *patch_module = NULL;
    static int asyncio_patched = 0;
    if (unlikely((!asyncio_patched) && module)) {
        PyObject *package;
        package = __Pyx_Import(__pyx_n_s_asyncio_coroutines, NULL, 0);
        if (package) {
            patch_module = __Pyx_Coroutine_patch_module(
                PyObject_GetAttrString(package, "coroutines"), ""
"try:\n"
"    coro_types = _module._COROUTINE_TYPES\n"
"except AttributeError: pass\n"
"else:\n"
"    if _cython_coroutine_type is not None and _cython_coroutine_type not in coro_types:\n"
"        coro_types = tuple(coro_types) + (_cython_coroutine_type,)\n"
"    if _cython_generator_type is not None and _cython_generator_type not in coro_types:\n"
"        coro_types = tuple(coro_types) + (_cython_generator_type,)\n"
"_module._COROUTINE_TYPES = coro_types\n"
            );
        } else {
            PyErr_Clear();
            package = __Pyx_Import(__pyx_n_s_asyncio_tasks, NULL, 0);
            if (unlikely(!package)) goto asyncio_done;
            patch_module = __Pyx_Coroutine_patch_module(
                PyObject_GetAttrString(package, "tasks"), ""
"if hasattr(_module, 'iscoroutine'):\n"
"    old_types = getattr(_module.iscoroutine, '_cython_coroutine_types', None)\n"
"    if old_types is None or not isinstance(old_types, set):\n"
"        old_types = set()\n"
"        def cy_wrap(orig_func, type=type, cython_coroutine_types=old_types):\n"
"            def cy_iscoroutine(obj): return type(obj) in cython_coroutine_types or orig_func(obj)\n"
"            cy_iscoroutine._cython_coroutine_types = cython_coroutine_types\n"
"            return cy_iscoroutine\n"
"        _module.iscoroutine = cy_wrap(_module.iscoroutine)\n"
"    if _cython_coroutine_type is not None:\n"
"        old_types.add(_cython_coroutine_type)\n"
"    if _cython_generator_type is not None:\n"
"        old_types.add(_cython_generator_type)\n"
            );
        }
        Py_DECREF(package);
        if (unlikely(!patch_module)) goto ignore;
asyncio_done:
        PyErr_Clear();
        asyncio_patched = 1;
#ifdef __Pyx_Generator_USED
        {
            PyObject *inspect_module;
            if (patch_module) {
                inspect_module = PyObject_GetAttr(patch_module, __pyx_n_s_inspect);
                Py_DECREF(patch_module);
            } else {
                inspect_module = __Pyx_Import(__pyx_n_s_inspect, NULL, 0);
            }
            if (unlikely(!inspect_module)) goto ignore;
            inspect_module = __Pyx_patch_inspect(inspect_module);
            if (unlikely(!inspect_module)) {
                Py_DECREF(module);
                module = NULL;
            }
            Py_XDECREF(inspect_module);
        }
#else
        if ((0)) return __Pyx_patch_inspect(module);
#endif
    }
    return module;
ignore:
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch asyncio package with custom generator type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    if ((0)) return __Pyx_patch_inspect(__Pyx_Coroutine_patch_module(module, NULL));
#endif
    return module;
}

/* PyVectorcallFastCallDict */
  #if CYTHON_METH_FASTCALL
static PyObject *__Pyx_PyVectorcall_FastCallDict_kw(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    PyObject *res = NULL;
    PyObject *kwnames;
    PyObject **newargs;
    PyObject **kwvalues;
    Py_ssize_t i, pos;
    size_t j;
    PyObject *key, *value;
    unsigned long keys_are_strings;
    Py_ssize_t nkw = PyDict_GET_SIZE(kw);
    newargs = (PyObject **)PyMem_Malloc((nargs + (size_t)nkw) * sizeof(args[0]));
    if (unlikely(newargs == NULL)) {
        PyErr_NoMemory();
        return NULL;
    }
    for (j = 0; j < nargs; j++) newargs[j] = args[j];
    kwnames = PyTuple_New(nkw);
    if (unlikely(kwnames == NULL)) {
        PyMem_Free(newargs);
        return NULL;
    }
    kwvalues = newargs + nargs;
    pos = i = 0;
    keys_are_strings = Py_TPFLAGS_UNICODE_SUBCLASS;
    while (PyDict_Next(kw, &pos, &key, &value)) {
        keys_are_strings &= Py_TYPE(key)->tp_flags;
        Py_INCREF(key);
        Py_INCREF(value);
        PyTuple_SET_ITEM(kwnames, i, key);
        kwvalues[i] = value;
        i++;
    }
    if (unlikely(!keys_are_strings)) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        goto cleanup;
    }
    res = vc(func, newargs, nargs, kwnames);
cleanup:
    Py_DECREF(kwnames);
    for (i = 0; i < nkw; i++)
        Py_DECREF(kwvalues[i]);
    PyMem_Free(newargs);
    return res;
}
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    if (likely(kw == NULL) || PyDict_GET_SIZE(kw) == 0) {
        return vc(func, args, nargs, NULL);
    }
    return __Pyx_PyVectorcall_FastCallDict_kw(func, vc, args, nargs, kw);
}
#endif

/* CythonFunctionShared */
  #if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc) {
    if (__Pyx_CyFunction_Check(func)) {
        return PyCFunction_GetFunction(((__pyx_CyFunctionObject*)func)->func) == (PyCFunction) cfunc;
    } else if (PyCFunction_Check(func)) {
        return PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
    }
    return 0;
}
#else
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc) {
    return __Pyx_CyOrPyCFunction_Check(func) && __Pyx_CyOrPyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
}
#endif
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj) {
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    __Pyx_Py_XDECREF_SET(
        __Pyx_CyFunction_GetClassObj(f),
            ((classobj) ? __Pyx_NewRef(classobj) : NULL));
#else
    __Pyx_Py_XDECREF_SET(
        ((PyCMethodObject *) (f))->mm_class,
        (PyTypeObject*)((classobj) ? __Pyx_NewRef(classobj) : NULL));
#endif
}
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, void *closure)
{
    CYTHON_UNUSED_VAR(closure);
    if (unlikely(op->func_doc == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_doc = PyObject_GetAttrString(op->func, "__doc__");
        if (unlikely(!op->func_doc)) return NULL;
#else
        if (((PyCFunctionObject*)op)->m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
#endif
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_doc, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_name == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_name = PyObject_GetAttrString(op->func, "__name__");
#elif PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_name, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_qualname, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_dict, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(op);
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = __Pyx_PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = __Pyx_PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyTuple_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__defaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->defaults_tuple, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->defaults_tuple;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__kwdefaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->defaults_kwdict, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->defaults_kwdict;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value || value == Py_None) {
        value = NULL;
    } else if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_annotations, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->func_annotations;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine(__pyx_CyFunctionObject *op, void *context) {
    int is_coroutine;
    CYTHON_UNUSED_VAR(context);
    if (op->func_is_coroutine) {
        return __Pyx_NewRef(op->func_is_coroutine);
    }
    is_coroutine = op->flags & __Pyx_CYFUNCTION_COROUTINE;
#if PY_VERSION_HEX >= 0x03050000
    if (is_coroutine) {
        PyObject *module, *fromlist, *marker = __pyx_n_s_is_coroutine;
        fromlist = PyList_New(1);
        if (unlikely(!fromlist)) return NULL;
        Py_INCREF(marker);
#if CYTHON_ASSUME_SAFE_MACROS
        PyList_SET_ITEM(fromlist, 0, marker);
#else
        if (unlikely(PyList_SetItem(fromlist, 0, marker) < 0)) {
            Py_DECREF(marker);
            Py_DECREF(fromlist);
            return NULL;
        }
#endif
        module = PyImport_ImportModuleLevelObject(__pyx_n_s_asyncio_coroutines, NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
        if (unlikely(!module)) goto ignore;
        op->func_is_coroutine = __Pyx_PyObject_GetAttrStr(module, marker);
        Py_DECREF(module);
        if (likely(op->func_is_coroutine)) {
            return __Pyx_NewRef(op->func_is_coroutine);
        }
ignore:
        PyErr_Clear();
    }
#endif
    op->func_is_coroutine = __Pyx_PyBool_FromLong(is_coroutine);
    return __Pyx_NewRef(op->func_is_coroutine);
}
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *
__Pyx_CyFunction_get_module(__pyx_CyFunctionObject *op, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_GetAttrString(op->func, "__module__");
}
static int
__Pyx_CyFunction_set_module(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_SetAttrString(op->func, "__module__", value);
}
#endif
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {(char *) "_is_coroutine", (getter)__Pyx_CyFunction_get_is_coroutine, 0, 0, 0},
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", (getter)__Pyx_CyFunction_get_module, (setter)__Pyx_CyFunction_set_module, 0, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
#if !CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), 0, 0},
#endif
#if CYTHON_USE_TYPE_SPECS
    {(char *) "__dictoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_dict), READONLY, 0},
#if CYTHON_METH_FASTCALL
#if CYTHON_BACKPORT_VECTORCALL
    {(char *) "__vectorcalloffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_vectorcall), READONLY, 0},
#else
#if !CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__vectorcalloffset__", T_PYSSIZET, offsetof(PyCFunctionObject, vectorcall), READONLY, 0},
#endif
#endif
#endif
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_weakreflist), READONLY, 0},
#else
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(PyCFunctionObject, m_weakreflist), READONLY, 0},
#endif
#endif
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, PyObject *args)
{
    CYTHON_UNUSED_VAR(args);
#if PY_MAJOR_VERSION >= 3
    Py_INCREF(m->func_qualname);
    return m->func_qualname;
#else
    return PyString_FromString(((PyCFunctionObject*)m)->m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) (((PyCFunctionObject*)cyfunc)->m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
#if !CYTHON_COMPILING_IN_LIMITED_API
    PyCFunctionObject *cf = (PyCFunctionObject*) op;
#endif
    if (unlikely(op == NULL))
        return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    op->func = PyCFunction_NewEx(ml, (PyObject*)op, module);
    if (unlikely(!op->func)) return NULL;
#endif
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
#if !CYTHON_COMPILING_IN_LIMITED_API
    cf->m_ml = ml;
    cf->m_self = (PyObject *) op;
#endif
    Py_XINCREF(closure);
    op->func_closure = closure;
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_XINCREF(module);
    cf->m_module = module;
#endif
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    op->func_classobj = NULL;
#else
    ((PyCMethodObject*)op)->mm_class = NULL;
#endif
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults_size = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    op->func_is_coroutine = NULL;
#if CYTHON_METH_FASTCALL
    switch (ml->ml_flags & (METH_VARARGS | METH_FASTCALL | METH_NOARGS | METH_O | METH_KEYWORDS | METH_METHOD)) {
    case METH_NOARGS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_NOARGS;
        break;
    case METH_O:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_O;
        break;
    case METH_METHOD | METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD;
        break;
    case METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS;
        break;
    case METH_VARARGS | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = NULL;
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        Py_DECREF(op);
        return NULL;
    }
#endif
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(m->func);
#else
    Py_CLEAR(((PyCFunctionObject*)m)->m_module);
#endif
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
#if PY_VERSION_HEX < 0x030900B1
    Py_CLEAR(__Pyx_CyFunction_GetClassObj(m));
#else
    {
        PyObject *cls = (PyObject*) ((PyCMethodObject *) (m))->mm_class;
        ((PyCMethodObject *) (m))->mm_class = NULL;
        Py_XDECREF(cls);
    }
#endif
#endif
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    Py_CLEAR(m->func_is_coroutine);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    __Pyx_PyHeapTypeObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(m->func);
#else
    Py_VISIT(((PyCFunctionObject*)m)->m_module);
#endif
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(__Pyx_CyFunction_GetClassObj(m));
#endif
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    Py_VISIT(m->func_is_coroutine);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *f = ((__pyx_CyFunctionObject*)func)->func;
    PyObject *py_name = NULL;
    PyCFunction meth;
    int flags;
    meth = PyCFunction_GetFunction(f);
    if (unlikely(!meth)) return NULL;
    flags = PyCFunction_GetFlags(f);
    if (unlikely(flags < 0)) return NULL;
#else
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    int flags = f->m_ml->ml_flags;
#endif
    Py_ssize_t size;
    switch (flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 0))
                return (*meth)(self, NULL);
#if CYTHON_COMPILING_IN_LIMITED_API
            py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
            if (!py_name) return NULL;
            PyErr_Format(PyExc_TypeError,
                "%.200S() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                py_name, size);
            Py_DECREF(py_name);
#else
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
#endif
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = __Pyx_PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
#if CYTHON_COMPILING_IN_LIMITED_API
            py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
            if (!py_name) return NULL;
            PyErr_Format(PyExc_TypeError,
                "%.200S() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                py_name, size);
            Py_DECREF(py_name);
#else
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
#endif
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        return NULL;
    }
#if CYTHON_COMPILING_IN_LIMITED_API
    py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
    if (!py_name) return NULL;
    PyErr_Format(PyExc_TypeError, "%.200S() takes no keyword arguments",
                 py_name);
    Py_DECREF(py_name);
#else
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
#endif
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *self, *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)func)->func);
    if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
    self = ((PyCFunctionObject*)func)->m_self;
#endif
    result = __Pyx_CyFunction_CallMethod(func, self, arg, kw);
    return result;
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
#if CYTHON_METH_FASTCALL
     __pyx_vectorcallfunc vc = __Pyx_CyFunction_func_vectorcall(cyfunc);
    if (vc) {
#if CYTHON_ASSUME_SAFE_MACROS
        return __Pyx_PyVectorcall_FastCallDict(func, vc, &PyTuple_GET_ITEM(args, 0), (size_t)PyTuple_GET_SIZE(args), kw);
#else
        (void) &__Pyx_PyVectorcall_FastCallDict;
        return PyVectorcall_Call(func, args, kw);
#endif
    }
#endif
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
#if CYTHON_ASSUME_SAFE_MACROS
        argc = PyTuple_GET_SIZE(args);
#else
        argc = PyTuple_Size(args);
        if (unlikely(!argc) < 0) return NULL;
#endif
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
#if PY_MAJOR_VERSION > 2
            PyErr_Format(PyExc_TypeError,
                         "unbound method %.200S() needs an argument",
                         cyfunc->func_qualname);
#else
            PyErr_SetString(PyExc_TypeError,
                            "unbound method needs an argument");
#endif
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE int __Pyx_CyFunction_Vectorcall_CheckArgs(__pyx_CyFunctionObject *cyfunc, Py_ssize_t nargs, PyObject *kwnames)
{
    int ret = 0;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        if (unlikely(nargs < 1)) {
            PyErr_Format(PyExc_TypeError, "%.200s() needs an argument",
                         ((PyCFunctionObject*)cyfunc)->m_ml->ml_name);
            return -1;
        }
        ret = 1;
    }
    if (unlikely(kwnames) && unlikely(PyTuple_GET_SIZE(kwnames))) {
        PyErr_Format(PyExc_TypeError,
                     "%.200s() takes no keyword arguments", ((PyCFunctionObject*)cyfunc)->m_ml->ml_name);
        return -1;
    }
    return ret;
}
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 0)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
            def->ml_name, nargs);
        return NULL;
    }
    return def->ml_meth(self, NULL);
}
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 1)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
            def->ml_name, nargs);
        return NULL;
    }
    return def->ml_meth(self, args[0]);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))def->ml_meth)(self, args, nargs, kwnames);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
    PyTypeObject *cls = (PyTypeObject *) __Pyx_CyFunction_GetClassObj(cyfunc);
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCMethod)(void(*)(void))def->ml_meth)(self, cls, args, (size_t)nargs, kwnames);
}
#endif
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_CyFunctionType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_CyFunction_dealloc},
    {Py_tp_repr, (void *)__Pyx_CyFunction_repr},
    {Py_tp_call, (void *)__Pyx_CyFunction_CallAsMethod},
    {Py_tp_traverse, (void *)__Pyx_CyFunction_traverse},
    {Py_tp_clear, (void *)__Pyx_CyFunction_clear},
    {Py_tp_methods, (void *)__pyx_CyFunction_methods},
    {Py_tp_members, (void *)__pyx_CyFunction_members},
    {Py_tp_getset, (void *)__pyx_CyFunction_getsets},
    {Py_tp_descr_get, (void *)__Pyx_PyMethod_New},
    {0, 0},
};
static PyType_Spec __pyx_CyFunctionType_spec = {
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if (defined(_Py_TPFLAGS_HAVE_VECTORCALL) && CYTHON_METH_FASTCALL)
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    __pyx_CyFunctionType_slots
};
#else
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
#if !CYTHON_METH_FASTCALL
    0,
#elif CYTHON_BACKPORT_VECTORCALL
    (printfunc)offsetof(__pyx_CyFunctionObject, func_vectorcall),
#else
    offsetof(PyCFunctionObject, vectorcall),
#endif
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if defined(_Py_TPFLAGS_HAVE_VECTORCALL) && CYTHON_METH_FASTCALL
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_PyMethod_New,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if __PYX_NEED_TP_PRINT_SLOT
    0,
#endif
#if PY_VERSION_HEX >= 0x030C0000
    0,
#endif
#if PY_VERSION_HEX >= 0x030d00A4
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
    0,
#endif
};
#endif
static int __pyx_CyFunction_init(PyObject *module) {
#if CYTHON_USE_TYPE_SPECS
    __pyx_CyFunctionType = __Pyx_FetchCommonTypeFromSpec(module, &__pyx_CyFunctionType_spec, NULL);
#else
    CYTHON_UNUSED_VAR(module);
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
#endif
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    m->defaults_size = size;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
  static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* CLineInTraceback */
  #ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    CYTHON_MAYBE_UNUSED_VAR(tstate);
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStrNoError(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
  #if !CYTHON_COMPILING_IN_LIMITED_API
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}
#endif

/* AddTraceback */
  #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyCode_Replace_For_AddTraceback(PyObject *code, PyObject *scratch_dict,
                                                       PyObject *firstlineno, PyObject *name) {
    PyObject *replace = NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_firstlineno", firstlineno))) return NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_name", name))) return NULL;
    replace = PyObject_GetAttrString(code, "replace");
    if (likely(replace)) {
        PyObject *result;
        result = PyObject_Call(replace, __pyx_empty_tuple, scratch_dict);
        Py_DECREF(replace);
        return result;
    }
    PyErr_Clear();
    #if __PYX_LIMITED_VERSION_HEX < 0x030780000
    {
        PyObject *compiled = NULL, *result = NULL;
        if (unlikely(PyDict_SetItemString(scratch_dict, "code", code))) return NULL;
        if (unlikely(PyDict_SetItemString(scratch_dict, "type", (PyObject*)(&PyType_Type)))) return NULL;
        compiled = Py_CompileString(
            "out = type(code)(\n"
            "  code.co_argcount, code.co_kwonlyargcount, code.co_nlocals, code.co_stacksize,\n"
            "  code.co_flags, code.co_code, code.co_consts, code.co_names,\n"
            "  code.co_varnames, code.co_filename, co_name, co_firstlineno,\n"
            "  code.co_lnotab)\n", "<dummy>", Py_file_input);
        if (!compiled) return NULL;
        result = PyEval_EvalCode(compiled, scratch_dict, scratch_dict);
        Py_DECREF(compiled);
        if (!result) PyErr_Print();
        Py_DECREF(result);
        result = PyDict_GetItemString(scratch_dict, "out");
        if (result) Py_INCREF(result);
        return result;
    }
    #else
    return NULL;
    #endif
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyObject *code_object = NULL, *py_py_line = NULL, *py_funcname = NULL, *dict = NULL;
    PyObject *replace = NULL, *getframe = NULL, *frame = NULL;
    PyObject *exc_type, *exc_value, *exc_traceback;
    int success = 0;
    if (c_line) {
        (void) __pyx_cfilenm;
        (void) __Pyx_CLineForTraceback(__Pyx_PyThreadState_Current, c_line);
    }
    PyErr_Fetch(&exc_type, &exc_value, &exc_traceback);
    code_object = Py_CompileString("_getframe()", filename, Py_eval_input);
    if (unlikely(!code_object)) goto bad;
    py_py_line = PyLong_FromLong(py_line);
    if (unlikely(!py_py_line)) goto bad;
    py_funcname = PyUnicode_FromString(funcname);
    if (unlikely(!py_funcname)) goto bad;
    dict = PyDict_New();
    if (unlikely(!dict)) goto bad;
    {
        PyObject *old_code_object = code_object;
        code_object = __Pyx_PyCode_Replace_For_AddTraceback(code_object, dict, py_py_line, py_funcname);
        Py_DECREF(old_code_object);
    }
    if (unlikely(!code_object)) goto bad;
    getframe = PySys_GetObject("_getframe");
    if (unlikely(!getframe)) goto bad;
    if (unlikely(PyDict_SetItemString(dict, "_getframe", getframe))) goto bad;
    frame = PyEval_EvalCode(code_object, dict, dict);
    if (unlikely(!frame) || frame == Py_None) goto bad;
    success = 1;
  bad:
    PyErr_Restore(exc_type, exc_value, exc_traceback);
    Py_XDECREF(code_object);
    Py_XDECREF(py_py_line);
    Py_XDECREF(py_funcname);
    Py_XDECREF(dict);
    Py_XDECREF(replace);
    if (success) {
        PyTraceBack_Here(
            (struct _frame*)frame);
    }
    Py_XDECREF(frame);
}
#else
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}
#endif

/* CIntFromPyVerify */
  #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* Declarations */
  #if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
  #if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabs(b.real) >= fabs(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                double r = b.imag / b.real;
                double s = (double)(1.0) / (b.real + b.imag * r);
                return __pyx_t_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            double r = b.real / b.imag;
            double s = (double)(1.0) / (b.imag + b.real * r);
            return __pyx_t_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_double(a, a);
                    case 3:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, a);
                    case 4:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = pow(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_double(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int64_t(int64_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int64_t neg_one = (int64_t) -1, const_zero = (int64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int64_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int64_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int64_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int64_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int64_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int64_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int64_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
  static CYTHON_INLINE int64_t __Pyx_PyInt_As_int64_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int64_t neg_one = (int64_t) -1, const_zero = (int64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(int64_t) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(int64_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int64_t) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        int64_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int64_t) -1;
        val = __Pyx_PyInt_As_int64_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int64_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int64_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) >= 2 * PyLong_SHIFT)) {
                            return (int64_t) (((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int64_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) >= 3 * PyLong_SHIFT)) {
                            return (int64_t) (((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int64_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) >= 4 * PyLong_SHIFT)) {
                            return (int64_t) (((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int64_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int64_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int64_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(int64_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int64_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int64_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int64_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (int64_t) (((int64_t)-1)*(((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int64_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (int64_t) ((((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (int64_t) (((int64_t)-1)*(((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int64_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (int64_t) ((((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (int64_t) (((int64_t)-1)*(((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int64_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (int64_t) ((((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int64_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int64_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(int64_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int64_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        int64_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int64_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int64_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int64_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int64_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int64_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int64_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int64_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int64_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int64_t) 1) << (sizeof(int64_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int64_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int64_t");
    return (int64_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int64_t");
    return (int64_t) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(long) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 2 * PyLong_SHIFT)) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 3 * PyLong_SHIFT)) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 4 * PyLong_SHIFT)) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (long) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(long) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(long) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(long) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(long) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(long) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        long val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (long) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (long) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (long) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (long) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(long) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((long) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(long) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((long) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((long) 1) << (sizeof(long) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (long) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
  static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(int) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 2 * PyLong_SHIFT)) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 3 * PyLong_SHIFT)) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 4 * PyLong_SHIFT)) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(int) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(int) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        int val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int) 1) << (sizeof(int) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(long));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
  static CYTHON_INLINE uint64_t __Pyx_PyInt_As_uint64_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint64_t neg_one = (uint64_t) -1, const_zero = (uint64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(uint64_t) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(uint64_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (uint64_t) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        uint64_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (uint64_t) -1;
        val = __Pyx_PyInt_As_uint64_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(uint64_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(uint64_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) >= 2 * PyLong_SHIFT)) {
                            return (uint64_t) (((((uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(uint64_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) >= 3 * PyLong_SHIFT)) {
                            return (uint64_t) (((((((uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(uint64_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) >= 4 * PyLong_SHIFT)) {
                            return (uint64_t) (((((((((uint64_t)digits[3]) << PyLong_SHIFT) | (uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (uint64_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(uint64_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint64_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(uint64_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint64_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(uint64_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(uint64_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (uint64_t) (((uint64_t)-1)*(((((uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(uint64_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (uint64_t) ((((((uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(uint64_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (uint64_t) (((uint64_t)-1)*(((((((uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(uint64_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (uint64_t) ((((((((uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(uint64_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (uint64_t) (((uint64_t)-1)*(((((((((uint64_t)digits[3]) << PyLong_SHIFT) | (uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(uint64_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (uint64_t) ((((((((((uint64_t)digits[3]) << PyLong_SHIFT) | (uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(uint64_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint64_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(uint64_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint64_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        uint64_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (uint64_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (uint64_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (uint64_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (uint64_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(uint64_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((uint64_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(uint64_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((uint64_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((uint64_t) 1) << (sizeof(uint64_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (uint64_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint64_t");
    return (uint64_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint64_t");
    return (uint64_t) -1;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint64_t(uint64_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint64_t neg_one = (uint64_t) -1, const_zero = (uint64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint64_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(uint64_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint64_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint64_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint64_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(uint64_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(uint64_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* FormatTypeName */
  #if CYTHON_COMPILING_IN_LIMITED_API
static __Pyx_TypeName
__Pyx_PyType_GetName(PyTypeObject* tp)
{
    PyObject *name = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_n_s_name);
    if (unlikely(name == NULL) || unlikely(!PyUnicode_Check(name))) {
        PyErr_Clear();
        Py_XDECREF(name);
        name = __Pyx_NewRef(__pyx_n_s__52);
    }
    return name;
}
#endif

/* FastTypeChecks */
  #if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = __Pyx_PyType_GetSlot(a, tp_base, PyTypeObject*);
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (cls == a || cls == b) return 1;
    mro = cls->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            PyObject *base = PyTuple_GET_ITEM(mro, i);
            if (base == (PyObject *)a || base == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(cls, a) || __Pyx_InBases(cls, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    if (exc_type1) {
        return __Pyx_IsAnySubtype2((PyTypeObject*)err, (PyTypeObject*)exc_type1, (PyTypeObject*)exc_type2);
    } else {
        return __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
  static unsigned long __Pyx_get_runtime_version(void) {
#if __PYX_LIMITED_VERSION_HEX >= 0x030B00A4
    return Py_Version & ~0xFFUL;
#else
    const char* rt_version = Py_GetVersion();
    unsigned long version = 0;
    unsigned long factor = 0x01000000UL;
    unsigned int digit = 0;
    int i = 0;
    while (factor) {
        while ('0' <= rt_version[i] && rt_version[i] <= '9') {
            digit = digit * 10 + (unsigned int) (rt_version[i] - '0');
            ++i;
        }
        version += factor * digit;
        if (rt_version[i] != '.')
            break;
        digit = 0;
        factor >>= 8;
        ++i;
    }
    return version;
#endif
}
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer) {
    const unsigned long MAJOR_MINOR = 0xFFFF0000UL;
    if ((rt_version & MAJOR_MINOR) == (ct_version & MAJOR_MINOR))
        return 0;
    if (likely(allow_newer && (rt_version & MAJOR_MINOR) > (ct_version & MAJOR_MINOR)))
        return 1;
    {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compile time Python version %d.%d "
                      "of module '%.100s' "
                      "%s "
                      "runtime version %d.%d",
                       (int) (ct_version >> 24), (int) ((ct_version >> 16) & 0xFF),
                       __Pyx_MODULE_NAME,
                       (allow_newer) ? "was newer than" : "does not match",
                       (int) (rt_version >> 24), (int) ((rt_version >> 16) & 0xFF)
       );
        return PyErr_WarnEx(NULL, message, 1);
    }
}

/* VoidPtrImport */
  #ifndef __PYX_HAVE_RT_ImportVoidPtr_3_0_11
#define __PYX_HAVE_RT_ImportVoidPtr_3_0_11
static int __Pyx_ImportVoidPtr_3_0_11(PyObject *module, const char *name, void **p, const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, name);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C variable %.200s",
                PyModule_GetName(module), name);
        goto bad;
    }
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    *p = PyCapsule_GetPointer(cobj, sig);
    if (!(*p))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
  #if PY_MAJOR_VERSION >= 3
static int __Pyx_InitString(__Pyx_StringTabEntry t, PyObject **str) {
    if (t.is_unicode | t.is_str) {
        if (t.intern) {
            *str = PyUnicode_InternFromString(t.s);
        } else if (t.encoding) {
            *str = PyUnicode_Decode(t.s, t.n - 1, t.encoding, NULL);
        } else {
            *str = PyUnicode_FromStringAndSize(t.s, t.n - 1);
        }
    } else {
        *str = PyBytes_FromStringAndSize(t.s, t.n - 1);
    }
    if (!*str)
        return -1;
    if (PyObject_Hash(*str) == -1)
        return -1;
    return 0;
}
#endif
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION >= 3
        __Pyx_InitString(*t, t->p);
        #else
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        #endif
        ++t;
    }
    return 0;
}

#include <string.h>
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s) {
    size_t len = strlen(s);
    if (unlikely(len > (size_t) PY_SSIZE_T_MAX)) {
        PyErr_SetString(PyExc_OverflowError, "byte string is too long");
        return -1;
    }
    return (Py_ssize_t) len;
}
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return __Pyx_PyUnicode_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return PyByteArray_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY && !CYTHON_COMPILING_IN_LIMITED_API) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
    __Pyx_TypeName result_type_name = __Pyx_PyType_GetName(Py_TYPE(result));
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ").  "
                "The ability to return an instance of a strict subclass of int is deprecated, "
                "and may be removed in a future version of Python.",
                result_type_name)) {
            __Pyx_DECREF_TypeName(result_type_name);
            Py_DECREF(result);
            return NULL;
        }
        __Pyx_DECREF_TypeName(result_type_name);
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type " __Pyx_FMT_TYPENAME ")",
                 type_name, type_name, result_type_name);
    __Pyx_DECREF_TypeName(result_type_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(__Pyx_PyLong_IsCompact(b))) {
        return __Pyx_PyLong_CompactValue(b);
    } else {
      const digit* digits = __Pyx_PyLong_Digits(b);
      const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(b);
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


/* #### Code section: utility_code_pragmas_end ### */
#ifdef _MSC_VER
#pragma warning( pop )
#endif



/* #### Code section: end ### */
#endif /* Py_PYTHON_H */
